/****************************************************************************
!C

!File: alloc_mem.c

!Description:
  Contains routines for allocating and freeing 2D and 3D memory.

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original January 1998. Version 1.0 Copied from MODLAND product source code
  library
  Modified January 1999
  Modified March 2000:  Added function "alloc_whole_sds" which allocates a
                        bulk storage block for an sds.  This block can be
                        used to copy and entire sds from one file to another.
                        (Curt Crandall   QSS   03/30/2000)

  Modified 06/12/2018: Removed "alloc_whole_sds" since functions in sds_rw.c
                       can handle all the needs, but just keep Calloc?D()
                       and Free?D(). Also added Calloc4D() and Free4D().
		       also removed #include <mfhdf.h>
		       (Maosheng Zhao, SSAI 06/12/2018)

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center.

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/
#include "alloc_mem.h"

/* Maosheng Zhao, SSAI, 06/12/2018, add Calloc4D and Free4D */
void ****Calloc4D(size_t nobj1, size_t nobj2, size_t nobj3, size_t nobj4, size_t size)
{
  void ****p1;
  void ***p2;
  void **p3;
  void *p4;
  char *c4;
  size_t iobj1, iobj2, iobj3;
  void ***Calloc3D(size_t nobj1, size_t nobj2, size_t nobj3, size_t size);

  p4 = calloc((nobj1 *nobj2 * nobj3 * nobj4), size);
  if (p4 == NULL) return NULL;

  p1 = (void ****) Calloc3D(nobj1, nobj2, nobj3, sizeof(void *));
  if (p1 == NULL) {
    free(p4);
    return NULL;
  }

  c4 = (char *) p4;
  for (iobj1 = 0; iobj1 < nobj1; iobj1++) {
    p2 = p1[iobj1];
    for (iobj2=0; iobj2<nobj2; iobj2++) {
      p3 = p2[iobj2];
    	for (iobj3=0; iobj3<nobj3; iobj3++) {
	  p3[iobj3] = (void *) c4;
      	  c4 += (nobj4 * size);
    	}
    }
  }

  return p1;
}

void ***Calloc3D(size_t nobj1, size_t nobj2, size_t nobj3, size_t size)
{
  void ***p1;
  void **p2;
  void *p3;
  char *c3;
  size_t iobj1, iobj2;
  void **Calloc2D(size_t nobj1, size_t nobj2, size_t size);

  p3 = calloc((nobj1 *nobj2 * nobj3), size);
  if (p3 == NULL) return NULL;

  p1 = (void ***) Calloc2D(nobj1, nobj2, sizeof(void *));
  if (p1 == NULL) {
    free(p3);
    return NULL;
  }

  c3 = (char *) p3;
  for (iobj1 = 0; iobj1 < nobj1; iobj1++) {
    p2 = p1[iobj1];
    for (iobj2=0; iobj2<nobj2; iobj2++) {
      p2[iobj2] = (void *) c3;
      c3 += (nobj3 * size);
    }
  }

  return p1;
}

void **Calloc2D(size_t nobj1, size_t nobj2, size_t size)
{
  void **p1;
  void *p2;
  size_t iobj1;
  char *c2;

  p2 = calloc((nobj1 * nobj2), size);
  if (p2 == NULL) return NULL;

  p1 = (void **) calloc(nobj1, sizeof(void *));
  if (p1 == NULL) {
    free(p2);
    return NULL;
  }

  c2 = (char *) p2;
  for (iobj1 = 0; iobj1 < nobj1; iobj1++) {
    p1[iobj1] = (void *) c2;
    c2 += (nobj2 * size);
  }

  return p1;
}

void Free2D(void **p1)
{
  if (*p1 != NULL) free(*p1);
  if (p1 != NULL) free(p1);
  return;
}

void Free3D(void ***p1)
{
  if (**p1 != NULL) free(**p1);
  if (*p1 != NULL) free(*p1);
  free(p1);
  return;
}

void Free4D(void ****p1)
{
  if (***p1 != NULL) free(***p1);
  if (**p1 != NULL) free(**p1);
  if (*p1 != NULL) free(*p1);
  free(p1);
}
