/****************************************************************************
!C

!File: comp_sds_diff_imgc

!Description:
  Compute SDS difference image in hdf format 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original Nov/Dec 2009. Version 1.0
           Jan 2010.     Version 1.1 (Updated for 3D dataset, 
				      Output scaled difference images)
           Jun 2010.     Version 1.1 (Updated for 3D dataset - MCD43A1) 
           Jul 2010.     Version 1.1 (include an option to generate the
					true difference image, not as
					relative percentage ) 
	   Oct 2010	 Version 1.2 (include an option to generate the
					percentage of data within a 
					threshold of difference) 
			 	by Xiaoping Zhang, Sigma Space Inc.			
	   Nov 2010	 Version 1.2 (changed the difference value
	   		 	from the absolute difference to the signed relative difference) 
			 	by Xiaoping Zhang, Sigma Space Inc.	
	   Jan 2011	 Version 1.2.2 (add CLOUD as esdt) 
			 	by Xiaoping Zhang, Sigma Space Inc.				 	
	   Jul 2012.     Version 1.2.3 
	   			by Sadashiva Devadiga, Sigma Space Inc.
	   			Fix the code for Fire Coarse granules MOD14CRS.
	   Jan 2013.     Version 1.2.4 
	   			by Sadashiva Devadiga, Sigma Space Inc.
	   			Set the fill value of fire SDS to 255 for Fire Coarse granules MOD14CRS
				if no fill value is available for this fire SDS.
	   Mar 2013	 Version 3.0
	   		 	Modified by Xiaoping Zhang, Sigma Space Corp.	
	   		 	write floating fill value(NPP_VMAE_L1) to output.			 		  
	   Jul 2014	 Version 3.1 
	   			Modified by Xiaoping Zhang, Sigma Space Corp.
				added a function of comp_sds_diff_npp to deat with NPP fill value issue.  	   
	   Mar 2016	 Version 3.1 Modified by Xiaoping Zhang, SSAI	   			
				fixed a bug to read a specific layer of a 3D SDS.    			
  				changed the members of sds_t: scalef to fscale, offsetf to foffset, scalef_type to fscale_type
				to work for all data including the converted h4 data from h5.							
	   Apr 2018      Version 3.5 Modified by Maosheng Zhao, SSAI. to update the sections on 
		         MOD10 and MOD29 from a relative old version but updated late for snow
	                 mismatch from the Sudpita's path
	                 /LANDQA/f18005/ssarkar/codes_c7/code/src/comp_sds_diff_img.c

  Modified in April, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c

   Modified February, 2020 Version 3.6 by Maosheng Zhao, SSAI
1) changed strstr to strcasecmp in the function copy_geolocation() so that hard-coded
   geolocation SDSs "Latitide" and "Longitude" can accept "latitude" and "longitude".
2) fix a bug in the function comp_sds_diff_npp() in comp_sds_diff_img.c. 
   Check MSZ for details. The bug causes problem for LST ESDTs.
					   
!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)
	The output SDS content
	For all MOD09 type esdt inputs
	  sds_val = 2000; one of the input is a fill value
	            -1000 - 1000 (percentage relative difference * 10)
	  if requested for absolute difference the range is 0 - 10000

	For MOD10 esdt type
	  sds_val = 2000;  one of the input is a fill value
		    0      match
		    1      missed snow
		    2      false snow
		    3      missed lake ice
		    4      false lake ice
		    5      other difference		    
		    3000   mismatch

	For MOD14/MOD14CRS/VAFIP esdt types
	  sds_val = 2000;  one of the input is a fill value
		    0:     match
		    1:     missed fire
		    2:     false fire
		    5:     other difference

	For QA sds and CLOUD
	  sds_val = 2000; one of the input is a fill value
		    0:    match
		    5:    no match

	For NPP VSCM and VAFIP esdt type
	  sds_val = 2000;  one of the input is a fill value
		    0      match
		    1      missed snow
		    2      false snow
		    5      other difference


	The fill value attribute is copied from the input. However the output
        difference SDS may never contain this value. 

!END
*****************************************************************************/         

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "alloc_mem.h"
#include "str_op.h"
#include "sds_rw.h"
#include "main_util.h"
#include "meta.h"

#define SCALE_DIFF 10 /* the percentage relative differece will be in the range 0 - 1000 */
#define SDS_FILL 2000 /* one of the input has fill value */
#define RELSCALE_DIFF 10 /* the percentage relative differece will be in the range -1000 - 1000 */
#define ACTSCALE_DIFF 10000 /* scaling factor for actual differeces */
#define INVALID -999.
#define SNOW_ERR 3000

#define HELP \
"NAME \n" \
"     comp_sds_diff_img  - Compute the difference between the two input SDSs and store the difference or relative difference \n" \
"     in the output hdf file and display percentage(s) of observations within relative difference threshold(s) on screen for \n" \
"     the continuous ESDT. Compute the counts of match, false and missed detections between the two input SDSs in the output \n" \
"     hdf file and display the numbers and percentages of match, false and missed detections on screen for the discrete ESDT.\n" \
"     Compute only the difference between the two input qa SDSs and store the difference in the output hdf file for all ESDTs.\n" \
"\n" \
"SYNOPSIS \n" \
" \n" \
"     comp_sds_diff_img  -help \n" \
" \n" \
"     comp_sds_diff_img  -esdt=<ESDT> -sds1=<f1,sds1> -sds2=<f2,sds2> -sds3=<f3,sds3> [-geo] [-meta] [-qa] \n" \
"     			[-abs] [-stat=threshold list]\n" \
" \n" \
" \n" \
"DESCRIPTION\n" \
" \n" \
"     Compute the difference between the two input SDSs and store the difference or relative difference in the output hdf \n" \
"     file and display percentage(s) of observations within relative difference threshold(s) on screen for the continuous \n" \
"     ESDT. Only one of the options of -abs and -stat can be chosen for MOD09. Option of -abs is for the difference and \n" \
"     -stat for the relative difference percentage.  \n" \
" \n" \
"     Compute the counts of match, false and missed detections between the two input SDSs in the output hdf file and \n" \
"     display the numbers and percentages of match, false and missed detections on screen for the discrete ESDTs. \n" \
" \n" \
"     The output SDS content:\n" \
"	For all MOD09 type esdt inputs\n" \
"	  sds_val = 2000; one of the input is a fill value\n" \
"                   -1000 - 1000 (percentage relative difference * 10) \n" \
"\n" \
"	For MOD10 esdt type\n" \
"         sds_val = 2000; one of the input is a fill value\n" \
"		    0      match\n" \
"                   3000   mismatch\n" \
"\n" \
"	For MOD14/MOD14CRS esdt types\n" \
"         sds_val = 2000; one of the input is a fill value\n" \
"		    0:     match\n" \
"		    1:     missed fire\n" \
"		    2:     false fire\n" \
"		    5:     other difference\n" \
"\n" \
"	For QA sds and CLOUD\n" \
"         sds_val = 2000; one of the input is a fill value\n" \
"		    0:    match\n" \
"		    5:    no match\n" \
"\n" \
"	The fill value attribute is copied from the input. However the output\n" \
"       difference SDS may never contain this value.\n" \
" \n" \
"     Compute only the difference between the two input qa SDSs and store the difference in the output hdf file for all\n" \
"     ESDT.    \n" \
" \n" \
"     This tool supports 2D/3D SDSs.\n" \
" \n" \
"     The tool command arguments can be specified in any order.\n" \
" \n" \
" \n" \
"ARGUMENTS \n" \
" \n" \
"     -help                         Display this help message \n" \
" \n" \
"     -sds1=<f1,sds1>               The first filename and sdsname are seperated by commas with no space.\n" \
"     				    To process a specific layer of a 3D SDS, specify the element number of the third \n" \
"				    dimension as a dot extension of the SDS name: sds_name.n \n" \
"				    (e.g., sur_refl_b02.2 = the layer defined by the second element of the 3rd dimension \n" \
"				    of the 3D SDS sur_refl_b02).\n" \
" 				    The 1st element of the 3rd dimension of the 3D SDS will be used if no element number \n" \
" 				    is specified.\n" \
" \n" \
"     -sds2=<f2,sds2>               The second filename and sdsname are seperated by commas with no space.\n" \
"     				    To process a specific layer of a 3D SDS, using the same way as the first SDS.\n" \
" \n" \
"     [-sds3=<f3,sds3>]             Output filename and sdsname are seperated by commas with no space.\n" \
"     				    The dimension of the output SDS is 2D.\n" \
" \n" \
"     [-scale1=<scale1>]	    scale factor for sds1\n"\
" \n" \
"     [-scale2=<scale2>]	    scale factor for sds2\n"\
" \n" \
"     [-off1=<off1>]	     	    offset factor for sds1\n"\
" \n" \
"     [-off2=<off2>]	     	    offset factor for sds2\n"\
" \n" \
"     -esdt=<ESDT>                  ESDT = MOD09|MOD10|MOD14|MOD29|CLOUD \n" \
"     				    Use esdt of MOD09 for all continuous products and use other esdts for the discrete \n" \
"				    products.\n" \
" \n" \
"     [-geo]                        Copy geolocation SDS from the first input file to the output file if geolocation SDS is\n" \
"     				    exist in the first input file.\n" \
" \n" \
"     [-meta]                       Copy metadata from the first input file to the output file.\n" \
" \n" \
"     [-qa]                         Input is a qa SDS and output is difference of two input qa SDSs with no summary results \n" \
"     				    printed on screen. This option can be used for all ESDTs.\n" \
" \n" \
"     [-abs]                        Ouptut difference (-ve to +ve) with no summary results printed on screen. This option is \n" \
"     				    only for MOD09. \n" \
" \n" \
"     [-stat=threshold list]        Output the percentage of relative difference and print percentage(s) of observations \n" \
"     				    within relative difference threshold(s) on screen. Thresholds are separated by comma \n" \
"				    if more than one threshold are used. This option is only for MOD09.\n" \
" \n" \
"Example: \n" \
" 	comp_sds_diff_img -esdt=MOD09 -meta -geo -stat=1\n" \
"		-sds1='input_data/MOD13A1.A2003001.h09v05.005.2010232090847_base.hdf,500m 16 days NDVI' \n" \
"		-sds2='input_data/MOD13A1.A2003001.h09v05.005.2010231195858_new.hdf,500m 16 days NDVI'  \n" \
"		-sds3=output_data/diff.MOD13A1.A2003001.h09v05.005.hdf,500m_16_days_NDVI_diff  \n" \
" \n" \
" 	comp_sds_diff_img -esdt=MOD09 -meta -geo -stat=1\n" \
" 		-sds1='input_data/MODHDFSR.A2003001.h09v05.005.2010231112702_base.hdf,sur_refl_b01' \n" \
" 		-sds2='input_data/MODHDFSR.A2003001.h09v05.005.2010231114502_new.hdf,sur_refl_b01' \n" \
"		 -sds3=output_data/diff.MODHDFSR.A2003001.h09v05.005.hdf,sur_refl_b01_diff \n" \
" \n" \
" 	comp_sds_diff_img -esdt=MOD10 -meta -geo\n" \
"		 -sds1=input_data/MOD10A1.A2003001.h09v05.005.2010231100856_base.hdf,Snow_Cover_Daily_Tile \n" \
"		-sds2=input_data/MOD10A1.A2003001.h09v05.005.2010231154155_new.hdf,Snow_Cover_Daily_Tile \n" \
"		-sds3=output_data/diff.MOD10A1.A2003001.h09v05.005.hdf,Snow_Cover_Daily_Tile  \n" \
" \n" \
"	comp_sds_diff_img -esdt=MOD09 -meta -geo -abs \n" \
"		-sds1='input_data/MODHDFSR.A2003001.h09v05.005.2010231112702_base.hdf,sur_refl_b01' \n" \
"		-sds2='input_data/MODHDFSR.A2003001.h09v05.005.2010231114502_new.hdf,sur_refl_b01' \n" \
"		-sds3=output_data/diff_abs.MODHDFSR.A2003001.h09v05.005.hdf,sur_refl_b01_diff \n" \
" \n" \
"	comp_sds_diff_img -esdt=CLOUD -meta -geo \n" \
"		-sds1='MOD35_L2.A2003208.2355.005.2010224033927.hdf,Cloud_Mask_sub' \n" \
"		-sds2='MOD35_L2.A2003208.2355.006.2010326093436.hdf,Cloud_Mask_sub' \n" \
" \n" \
"	comp_sds_diff_img -esdt=MOD14CRS -meta -geo  \n" \
"		-sds1='MOD14CRS.A2003001.0815.005.2012294001408.hdf,CoarseFireCounts' \n" \
"		-sds2='MOD14CRS.A2003001.0815.006.2012349130458.hdf,CoarseFireCounts' \n" \
"		-sds3='MOD14CRS.A2003001.0815.diff.hdf,CoarseFireCounts'\n" \
"	comp_sds_diff_img -esdt=MOD09   \n" \
"		-sds1='/LANDQA/f003/xzhang/src/NPP/codes/COMP_SDS_DIFF_IMG/v1.2.2/input_data/MOD13A1.A2003001.h09v05.005.2010232090847_base.hdf,500m 16 days NDVI' \n" \
"		-sds2='/LANDQA/f003/xzhang/src/NPP/codes/COMP_SDS_DIFF_IMG/v1.2.2/input_data/MOD13A1.A2003001.h09v05.005.2010231195858_new.hdf,500m 16 days NDVI' \n" \
"		-sds3='data_out/diff.MOD13A1.A2003001.h09v05.005.hdf,500m_16_days_NDVI_diff'\n" \
"	comp_sds_diff_img -esdt=MOD09 -meta -geo -stat=1    \n" \
"		-sds1='/LANDQA/f003/xzhang/src/NPP/data/input/NPP_SRFLMIP_L2.A2010251.2220.P1_00753.2011120134238.hdf,M10_SurfRefl' \n" \
"		-sds2='/LANDQA/f003/xzhang/src/NPP/data/input/NPP_SRFLMIP_L2.A2010251.2225.P1_00753.2011120134238.hdf,M10_SurfRefl' \n" \
"		-sds3='data_out/NPP_SRFLMIP_L2.A2010251.2220_2225_diff1.hdf,M10_SurfRefl'\n" \
"	comp_sds_diff_img -esdt=MOD09 -meta -geo -stat=1   \n" \
"		-sds1='/LANDQA/f003/xzhang/src/NPP/codes/L1B_Tests/NPP_VMAE_L1/2011/348/NPP_VMAE_L1.A2011348.0000.AGG_03000.2011348213939.hdf,Radiance_Mod_M4' \n" \
"		-sds2='/LANDQA/f003/xzhang/src/NPP/codes/L1B_Tests/NPP_VMAE_L1/2011/348/NPP_VMAE_L1.A2011348.0000.P1_03001.2011348185918.hdf,Radiance_Mod_M4' \n" \
"		-sds3='data_out/NPP_VMAE_L1.A2011348.0000.IDPS_LPEATE_diff.hdf,Radiance_Mod_M4'\n" \
"	comp_sds_diff_img -esdt=MOD09 -meta -stat=1   \n" \
"		-sds1=VNP02MOD.A2015262.1106.1100.2015321173151.hdf,M04 \n" \
"		-sds2=VNP02MOD.A2015262.1106.1100.2015321173151.hdf,M05 \n" \
"		-sds3=VNP02MOD.A2015262.1106.1100.2015321173151_m05_m04diff.hdf,M04-M05 \n" \
"	comp_sds_diff_img -esdt=MOD09 -meta -stat=1   \n" \
"		-sds1=VLSTO_npp_d20140905_t2359522_e0001164_b14806_c20140906063046978337_noaa_ops.hdf,LandSurfaceTemperature \n" \
"		-sds2=VLSTO_npp_d20150822_t2359424_e0001066_b19786_c20150823062748304334_noaa_ops.hdf,LandSurfaceTemperature \n" \
"		-sds3=VLSTO_npp_lst_diff.hdf,LandSurfaceTemperature \n" \
"	comp_sds_diff_img -esdt=MOD09 -meta -stat=1   \n" \
"		-sds1=SVM01_npp_d20151018_t2359219_e0000461_b20595_c20151019062325072157_noaa_ops.hdf,Reflectance \n" \
"		-sds2=SVM01_npp_d20151008_t2359247_e0000489_b20453_c20151009062436593243_noaa_ops.hdf,Reflectance \n" \
"		-sds3=SVM01_npp_Reflectance_diff.hdf,Reflectance_diff \n" \
"	comp_sds_diff_img -esdt=MOD09 -meta    \n" \
"		-sds1='MOD02SSH.A2016061.0905.006.2016074125951.hdf,EV_500_Aggr1km_RefSB.3' \n" \
"		-sds2='MOD02SSH.A2016061.0905.006.2016075123426.hdf,EV_500_Aggr1km_RefSB.3' \n" \
"		-sds3='MOD02SSH.A2016061.0905.006.EV_500_Aggr1km_RefSB.3.diff.hdf,EV_500_Aggr1km_RefSB.3'\n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:\tcomp_sds_diff_img -sds1=<f1,sds1> -sds2=<f2,sds2> -sds3=<f3,sds3> -esdt=<ESDT> [geo] [-meta] [-qa] [-abs] [-stat=threshold list]\n" \
" \n" \
"OPTIONS \n" \
"     -help                          Display this help message \n" \
"     -sds1=<f1,sds1>                filename and sdsname\n" \
"     -sds2=<f2,sds2>                filename and sdsname\n" \
"     -sds3=<f3,sds3>                output filename and sdsname\n" \
"     [-scale1=<scale1>]	     scale factor for sds1\n"\
"     [-scale2=<scale2>]	     scale factor for sds2\n"\
"     [-off1=<off1>]	     	     offset factor for sds1\n"\
"     [-off2=<off2>]	     	     offset factor for sds2\n"\
"     -esdt=<ESDT>                   ESDT = MOD09|MOD10|MOD14/MOD14CRS|MOD29|CLOUD \n" \
"     [-geo]                         copy geolocation SDS \n" \
"     [-meta]                        copy metadata \n" \
"     [-qa]                          input is a qa SDS \n" \
"     [-abs]                         ouptut absolute difference (-ve to +ve) (only for MOD09) \n" \
"                                    (default: relative percentage)\n" \
"     [-stat=threshold list]         output the percentage of relative difference(only for MOD09) \n" \
" 				     display percentage(s) of observations within relative difference threshold(s) \n" \
" 				     thresholds are separated by comma if more than one threshold are used. \n" \
" \n"


int parse_cmd_comp_sds_diff(int argc, char **argv, char *f1, char *sds1, char *f2, char *sds2, char *f3, char *sds3, 
	char *esdt, int *g_opt, int *m_opt, int *is_qa, int *is_rel, float *sc1, float *sc2, float *off1, float *off2, 
	float *err_th, int *tcnt, int *st_opt);
void comp_sds_diff(sds_t *sds1_info, sds_t *sds2_info, sds_t *sds3_info, char *sds1, char *esdt, int is_qa, int is_rel, char *gid,
	float *err_th, int tcnt, int st_opt);	
void comp_sds_diff_npp(sds_t *sds1_info, sds_t *sds2_info, sds_t *sds3_info, char *sds1, char *esdt, int is_qa, int is_rel, char *gid, 
	float *err_th, int tcnt, int st_opt);		
void copy_geolocation(sds_t *in_sds_info, sds_t *out_sds_info, char **sds_names, int nsds);
int main(int argc, char *argv[])
{  
  float err_th[10];  
  int len, cnt, tcnt;
  int g_opt, m_opt, is_qa, is_rel, nsds, st_opt, p1, p2;  
  sds_t sds1_info, sds2_info, sds3_info;
  char esdt[MAX_GID_LEN], Esdt[MAX_GID_LEN], gid[MAX_PATH_LENGTH], **sds_names;
  char sds1[MAX_SDS_NAME_LEN], f1[MAX_PATH_LENGTH];
  char sds2[MAX_SDS_NAME_LEN], f2[MAX_PATH_LENGTH];
  char sds3[MAX_SDS_NAME_LEN], f3[MAX_PATH_LENGTH]; 
  float f_fval;
  int f_val;
  int32 attr_type, attr_cnt; 
  void *attr_buf;
  float sc1=-999.,sc2=-999.,off1=-999.,off2=-999.;   
  char  tmp[20], dss[20];  
  char *meta_val1, *meta_val2;    
  char meta_name[MAX_META_NAME_LEN];
  int status;
  char *metadata;   
  
  if ((argc == 2) && ((strcmp(argv[1], "-h") == 0) || (strcmp(argv[1], "-help") == 0))) {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }         
  if (parse_cmd_comp_sds_diff(argc, argv, f1, sds1, f2, sds2, f3, sds3, esdt, &g_opt, &m_opt, &is_qa, &is_rel,
  	 &sc1, &sc2, &off1, &off2, err_th, &tcnt, &st_opt) == -1)
    fprintf(stderr, "%s\n", USAGE);
  else 
  {
    /* MSZ 2/6/2018 silent version mismatch warning standard print out */
    setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

    /* MSZ 2/6/2018 Disable error reporting */
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

    sds1_info.sds_id = -1;
    sds1_info.sd_id = -1;
    sds2_info.sds_id = -1;
    sds2_info.sd_id = -1;
    sds1_info.fill_val = 0;
    sds1_info.fill_uval = 0;
    sds1_info.fill_fval = 0.;
    sds2_info.fill_val = 0;
    sds2_info.fill_uval = 0;
    sds2_info.fill_fval = 0.;  
   
    if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
      fprintf(stderr, "Error allocating memory for sds_names in main()\n");
    if ((nsds = get_sds_names(f1, sds_names)) <= 0)
      fprintf(stderr, "Error reading the sds names in file %s\n", f1);
   
    if (strstr(f1, "/") != NULL) {
      len = (int)strlen(f1);
      for (p1=len-1, cnt=0; p1>0; p1--, cnt++) {
	if (f1[p1] == '/' || f1[p1] == '\\') break;
      }
      strmid(f1, p1+1, cnt, gid);
    } 
    else strcpy(gid, f1);

    if ((metadata = get_attr_metadata(f1, "CoreMetadata.0")) != NULL) status = 0;   /* MODIS */
    else status = 1;                                                                /* VIIRS */
    printf("status: %d\n", status);  
/*
    int32 attr_type, attr_cnt;

    strcpy(meta_name, "sensor");
    meta_val1 = (char*)get_attr_sel_metadata(f1, meta_name, &attr_type, &attr_cnt)
    if (meta_val1[0] != '\0') printf("sensor: %s\n", meta_val1); 
    strcpy(meta_name, "LongName");
    meta_val2 = (char*)get_attr_sel_metadata(f1, meta_name, &attr_type, &attr_cnt)
    if (meta_val2[0] != '\0') printf("LongName: %s\n", meta_val2);      
    if ((strstr(gid, "NPP_") != NULL) || (strstr(gid, "npp_") != NULL) 
     || (strcmp(meta_val1, "VIIRS") == 0) || (strstr(meta_val2, "NPP") != NULL) || (strstr(meta_val2, "VNP") ) 
    { 	
      free(meta_val1);  
      free(meta_val2); */
      
    if (status == 1) 
    { 	
                
/*      
printf("input file1: %s\n", f1);  
printf("input file2: %s\n", f2);       
printf("output file: %s\n", f3);
printf("input sds1: %s\n", sds1);  
printf("input sds2: %s\n", sds2);       
printf("output sds3: %s\n", sds3);
printf("sc1 : sc2 %f:%f \n",sc1, sc2);
printf("off1 : off2 %f:%f \n",off1, off2);      
*/
      strcpy(dss, "NPP");      	  
      strcpy(sds1_info.name, sds1);
      sds1_info.sd_id = sds1_info.sds_id = -1;
      if (get_sds_info(f1, &sds1_info) != -1)       
      {
	strcpy(sds2_info.name, sds2);
	sds2_info.sd_id = sds2_info.sds_id = -1;
	if (get_sds_info(f2, &sds2_info) != -1) 
	{
/*  printf(" sds1_info.data_type  =%d\n",   sds1_info.data_type);
  printf(" sds1_info.fill_fval  =%f\n",   sds1_info.fill_fval);	  
  printf(" sds1_info.fill_val  =%d\n",   sds1_info.fill_val);	
  printf(" sds1_info.fill_uval  =%d\n",   sds1_info.fill_uval);	  */	  
	  if ((sds1_info.dim_size[0] != sds2_info.dim_size[0]) ||
	      (sds1_info.dim_size[1] != sds2_info.dim_size[1]))
	  {
	      fprintf(stderr,"Dataset size mismatch for sds1: %s and sds2: %s\n", f1,f2);
	      exit(EXIT_FAILURE);
	  }
	  if (sc1 != -999.) sds1_info.fscale = sc1;
          if (sc2 != -999.) sds2_info.fscale = sc2;
          if (off1 != -999.) sds1_info.offset = off1;
          if (off2 != -999.) sds2_info.offset = off2;
          strcpy(sds3_info.name, sds3);
          sds3_info.sd_id = sds3_info.sds_id = -1;
	  sds3_info.rank = 2;
          if (sds2_info.rank == 2) 
	  {
	    sds3_info.dim_size[0] = sds2_info.dim_size[0];
	    sds3_info.dim_size[1] = sds2_info.dim_size[1];
          }
	  else 
	  {
	    if (sds1_info.dim_size[0] > sds1_info.dim_size[2]) 
	    {
	      sds3_info.dim_size[0] = sds2_info.dim_size[0];
	      sds3_info.dim_size[1] = sds2_info.dim_size[1];
	    }
	    else 
	    {
	      sds3_info.dim_size[0] = sds2_info.dim_size[1];
	      sds3_info.dim_size[1] = sds2_info.dim_size[2];
	    }
	  }
	  sds3_info.data_type = DFNT_INT32;
	  /*sds3_info.data_type = DFNT_INT16;*/
	  sds3_info.data_size = DFKNTsize(sds3_info.data_type); 
	  sds3_info.H5OUT = 0;   /* so far, output to HDF4 */
	  sds3_info.sd_id = sds3_info.sds_id = -1;
	 
	  if(open_sd(f3, &sds3_info, 'W') == FAIL) 
	    fprintf(stdout, "Cannot open output file %s\n", f3);
          if (open_sds((char *)NULL, &sds3_info, 'W') != FAIL) 
	    write_attr_fval(sds3_info.H5OUT, sds3_info.sds_id, sds3_info.data_type, 1, SDS_FILL);

	  fprintf(stderr, "computing and writing sds diff, please wait...\n");
	  comp_sds_diff_npp(&sds1_info, &sds2_info, &sds3_info, sds1, esdt, is_qa, is_rel, gid, err_th, tcnt, st_opt);     	    
	  close_sds(&sds3_info, sds3_info.H5OUT);
          if (m_opt == 1)
	    copy_metadata(sds1_info.H5IN, sds1_info.sd_id, sds3_info.H5OUT, sds3_info.sd_id);
          if (g_opt == 1)
	    copy_geolocation(&sds1_info, &sds3_info, sds_names, nsds);	    
	  close_sd(&sds3_info, sds3_info.H5OUT);
	  fprintf(stderr, "Done!\n");
	}
      }
    }   
    else
    {  
      strcpy(sds1_info.name, sds1);
      sds1_info.sd_id = sds1_info.sds_id = -1;
      if (get_sds_info(f1, &sds1_info) != -1) 
      {
//  printf(" sds1_info.data_type  =%d\n",   sds1_info.data_type);
//  printf(" sds1_info.fill_val  =%d\n",   sds1_info.fill_val);
	if (strcmp(esdt, "MOD14CRS") == 0) 
	{
            if ((attr_buf = (void *)get_sds_attr(&sds1_info, "_FillValue", &attr_type,
		  &attr_cnt)) != NULL)
            {
              switch(sds1_info.data_type)
              {
        	case 5: f_fval = ((float32 *)attr_buf)[0]; break;
		case 6: f_fval = ((float64 *)attr_buf)[0]; break;
		case 20: f_val = ((int8 *)attr_buf)[0]; break;
        	case 21: f_val = ((uint8 *)attr_buf)[0]; break;
        	case 22: f_val = ((int16 *)attr_buf)[0]; break;
        	case 23: f_val = ((uint16 *)attr_buf)[0]; break;
        	case 24: f_val = ((int32 *)attr_buf)[0]; break;
        	case 25: f_val = ((uint32 *)attr_buf)[0]; break;
              }
              free(attr_buf);
	      if (sds1_info.data_type == 5 || sds1_info.data_type == 6)
		fprintf(stdout, "\t%s (Fill_Value = %f)\n", sds1_info.name, f_fval);
	      else
		fprintf(stdout, "\t%s (Fill_Value = %d)\n", sds1_info.name, f_val);
            }    
    	    else
	    {
	      fprintf(stdout, "\tNo fill value for %s, set 255 as its fill value\n", sds1_info.name);
	      sds1_info.fill_val = 255;
	    }  
	}
	strcpy(sds2_info.name, sds2);                  
	sds2_info.sd_id = sds2_info.sds_id = -1;
	if (get_sds_info(f2, &sds2_info) != -1) 
	{      
	  if ((sds1_info.dim_size[0] != sds2_info.dim_size[0]) ||
	      (sds1_info.dim_size[1] != sds2_info.dim_size[1]))
	  {
	      fprintf(stderr,"Dataset size mismatch for sds1: %s and sds2: %s\n", f1,f2);
	      exit(EXIT_FAILURE);
	  }
	  if (sc1 != -999.) sds1_info.fscale = sc1;
          if (sc2 != -999.) sds2_info.fscale = sc2;
          if (off1 != -999.) sds1_info.offset = off1;
          if (off2 != -999.) sds2_info.offset = off2;	

	  if (strcmp(esdt, "MOD14CRS") == 0) 
	  {
              if ((attr_buf = (void *)get_sds_attr(&sds2_info, "_FillValue", &attr_type,
		    &attr_cnt)) != NULL)
              {
        	switch(sds2_info.data_type)
        	{
        	  case 5: f_fval = ((float32 *)attr_buf)[0]; break;
		  case 6: f_fval = ((float64 *)attr_buf)[0]; break;
		  case 20: f_val = ((int8 *)attr_buf)[0]; break;
        	  case 21: f_val = ((uint8 *)attr_buf)[0]; break;
        	  case 22: f_val = ((int16 *)attr_buf)[0]; break;
        	  case 23: f_val = ((uint16 *)attr_buf)[0]; break;
        	  case 24: f_val = ((int32 *)attr_buf)[0]; break;
        	  case 25: f_val = ((uint32 *)attr_buf)[0]; break;
        	}
        	free(attr_buf);
		if (sds2_info.data_type == 5 || sds2_info.data_type == 6)
		  fprintf(stdout, "\t%s (Fill_Value = %f)\n", sds2_info.name, f_fval);
		else
		  fprintf(stdout, "\t%s (Fill_Value = %d)\n", sds2_info.name, f_val);
              }    
    	      else
	      {
		fprintf(stdout, "\tNo fill value for %s, set 255 as its fill value\n", sds2_info.name);
		sds2_info.fill_val = 255;
	      }  
	  }      
          strcpy(sds3_info.name, sds3);
          sds3_info.sd_id = sds3_info.sds_id = -1;
	  sds3_info.rank = 2;
          if (sds2_info.rank == 2) {
	    sds3_info.dim_size[0] = sds2_info.dim_size[0];
	    sds3_info.dim_size[1] = sds2_info.dim_size[1];
          }
	  else {
	    if (sds1_info.dim_size[0] > sds1_info.dim_size[2]) {
	      sds3_info.dim_size[0] = sds2_info.dim_size[0];
	      sds3_info.dim_size[1] = sds2_info.dim_size[1];
	    }
	    else {
	      sds3_info.dim_size[0] = sds2_info.dim_size[1];
	      sds3_info.dim_size[1] = sds2_info.dim_size[2];
	    }
	  }

	  sds3_info.data_type = DFNT_INT32;
	  /*sds3_info.data_type = DFNT_INT16;*/
	  sds3_info.data_size = DFKNTsize(sds3_info.data_type);
	  sds3_info.H5OUT = 0;   /* so far, output to HDF4 */
          sds3_info.sd_id = sds3_info.sds_id = -1;

          if(open_sd(f3, &sds3_info, 'W') == FAIL)
	    fprintf(stdout, "Cannot open output file %s\n", f3);  
          if (open_sds((char *)NULL, &sds3_info, 'W') != -1) 
	  {  
	    write_attr_fval(sds3_info.H5OUT, sds3_info.sds_id, sds3_info.data_type, 1, SDS_FILL);
	    fprintf(stderr, "computing and writing sds diff, please wait...\n");
	    comp_sds_diff(&sds1_info, &sds2_info, &sds3_info, sds1, esdt, is_qa, is_rel, gid, err_th, tcnt, st_opt);                
	    close_sds(&sds3_info, sds3_info.H5OUT);
            if (m_opt == 1)
	      copy_metadata(sds1_info.H5IN, sds1_info.sd_id, sds3_info.H5OUT, sds3_info.sd_id);
            if (g_opt == 1)
	      copy_geolocation(&sds1_info, &sds3_info, sds_names, nsds);
	    close_sd(&sds3_info, sds3_info.H5OUT);
	    fprintf(stderr, "Done!\n");
	  }
	}
      }
    }
    if (sds1_info.sds_id != -1) close_sds(&sds1_info, sds1_info.H5IN);
    if (sds1_info.sd_id != -1) close_sd(&sds1_info, sds1_info.H5IN);
    if (sds2_info.sds_id != -1) close_sds(&sds2_info, sds2_info.H5IN);
    if (sds2_info.sd_id != -1) close_sd(&sds2_info, sds2_info.H5IN);
    if (sds_names != NULL) Free2D((void **)sds_names);
  }
}

int parse_cmd_comp_sds_diff(int argc, char **argv, char *f1, char *sds1, char *f2, char *sds2, char *f3, char *sds3, 
	char *esdt, int *g_opt, int *m_opt, int *is_qa, int *is_rel, float *sc1, float *sc2, float *off1, float *off2,
	float *err_th, int *tcnt, int *st_opt)
{
  int i, j, n;
  int st = 1;
  char tmp_str[10];
  char **tmp_arr, scale_str[15];
  
  *tcnt = 0;
  *is_qa = 0;
  *is_rel = 1;
  *m_opt = *g_opt = *st_opt = 0;
  f1[0] = f2[0] = f3[0] = '\0';
  sds1[0] = sds2[0] = sds3[0] = esdt[0] = '\0';

  if ((tmp_arr = (char **)Calloc2D(10, MAX_PATH_LENGTH, sizeof(char))) == NULL) 
  {
    fprintf(stdout, "Error allocating memory for tmp_arr in parse_cmd_comp_sds_diff \n");
    st = -1;
  }
  for (i=1; i<argc; i++)
  {
    if (is_arg_id(argv[i], "-sds1=") == 0) 
    {
      n = 0;
      get_arg_val_arr(argv[i], tmp_arr, &n);
      if (n == 2) 
      {
        strcpy(f1, tmp_arr[0]);
        strcpy(sds1, tmp_arr[1]);
      }
      else 
        fprintf(stderr, "Ignoring invalid option %s\n", argv[i]);
    }
    else if (is_arg_id(argv[i], "-sds2=") == 0) 
    {
      n = 0;
      get_arg_val_arr(argv[i], tmp_arr, &n);
      if (n == 2) 
      {
        strcpy(f2, tmp_arr[0]);
        strcpy(sds2, tmp_arr[1]);
      }
      else
        fprintf(stderr, "Ignoring invalid option %s\n", argv[i]);
    }
    else if (is_arg_id(argv[i], "-sds3=") == 0) 
    {
      n = 0;
      get_arg_val_arr(argv[i], tmp_arr, &n);
      if (n == 2) 
      {
	strcpy(f3, tmp_arr[0]);
        strcpy(sds3, tmp_arr[1]);
      }
      else
        fprintf(stderr, "Ignoring invalid option %s\n", argv[i]);
    }
    else if (is_arg_id(argv[i], "-scale1=") == 0)
    {
        get_arg_val(argv[i],scale_str);
        *sc1 = (float)atof(scale_str);
    }
    else if (is_arg_id(argv[i], "-scale2=") == 0)
    {
        get_arg_val(argv[i],scale_str);
        *sc2 = (float)atof(scale_str);
    }
    else if (is_arg_id(argv[i], "-off1=") == 0)
    {
        get_arg_val(argv[i],scale_str);
        *off1 = (float)atof(scale_str);
    }
    else if (is_arg_id(argv[i], "-off2=") == 0)
    {
        get_arg_val(argv[i],scale_str);
        *off2 = (float)atof(scale_str);
    }    
    else if (strcmp(argv[i], "-meta") == 0)
      *m_opt = 1;
    else if (strcmp(argv[i], "-geo") == 0)
      *g_opt = 1;
    else if (strcmp(argv[i], "-qa") == 0)
      *is_qa = 1;
    else if (strcmp(argv[i], "-abs") == 0)
      *is_rel = 0;
    else if (is_arg_id(argv[i], "-stat=") == 0) {
      *st_opt = 1;
      get_arg_val_arr(argv[i], tmp_arr, tcnt);
      for(j=0; j<*tcnt; j++)    
        err_th[j] = (float)atof(tmp_arr[j]);
    }      
    else if (is_arg_id(argv[i], "-esdt=") == 0)
      get_arg_val(argv[i], esdt);
    else if (argv[i][0] == '-') 
      fprintf(stderr, "Ignoring unknown option %s\n", argv[i]);
  }
  if ((f1[0] == '\0') || (sds1[0] == '\0')) 
  {
    st = -1;
    fprintf(stderr, "Missing input filename/sdsname for -sds1 . . \n");
  }
  if ((f2[0] == '\0') || (sds2[0] == '\0')) 
  {
    st = -1;
    fprintf(stderr, "Missing input filename/sdsname for -sds2 . . \n");
  }
  if ((f3[0] == '\0') || (sds3[0] == '\0')) 
  {
    st = -1;
    fprintf(stderr, "Missing output filename/sdsname for -sds3 . . \n");
  }
  if (*is_qa == 1) 
  {
    if (strlen(esdt) > 0) 
    { 
      fprintf(stderr, "The -qa option will override the -esdt option\n");
    }
  }
  Free2D((void **)tmp_arr);
  return st;
}

void comp_sds_diff(sds_t *sds1_info, sds_t *sds2_info, sds_t *sds3_info, char *sds1, char *esdt, int is_qa, int is_rel, 
	char *gid, float *err_th, int tcnt, int st_opt)
{
  char ib_str[10];
  int irow, icol, ic;
  int p1, len, ib=0, nb, i;
  int nrows, ncols, ndata;
  int val1_st, val2_st;
  int val1, val2, val3, per_err;
  float abs_fval1, diff, fper_err;
  void *data1, *data2, *data3;
  int32 in_start[3] = {0, 0, 0};
  int32 in_edge[3] = {0, 0, 0};
  int32 out_start[2] = {0, 0};
  int32 out_edge[2] = {0, 0};
  float fval1, fval2;
  double per_pix_err;
  double per_pix_match, per_pix_fill;
  double per_pix_miss, per_pix_false, per_pix_nsnow_fire;
  long int tot_pix;
  long int cnt_match[10];
  long int cnt_err_th[10];
  long int cnt_fill;
  long int missed_cnt = 0;
  long int false_cnt = 0;
  long int other_cnt = 0;
  long int nsnow_fire = 0;  
  int n, m, st_c, offset, rank;
  char sds_name[MAX_SDS_NAME_LEN];  
   
  get_sdsname_dim(sds1, sds_name, &n, &m); 
  rank = sds1_info->rank;
  if (sds1_info->rank == 2)
  {
    nrows = sds1_info->dim_size[0];
    ncols = ndata = sds1_info->dim_size[1];
  }                                                                            
  else if (sds1_info->rank > 2)
  {
    if (sds1_info->dim_size[0] > sds1_info->dim_size[rank-1])
    {
      nrows = sds1_info->dim_size[0];
      ncols = ndata = sds1_info->dim_size[1];
      for (i=2; i<rank; i++)
        ndata *= sds1_info->dim_size[i];
    }
    else
    {
      nrows = sds1_info->dim_size[rank-2];
      ncols = ndata = sds1_info->dim_size[rank-1];
      for (i=0; i<rank-2; i++)
        ndata *= sds1_info->dim_size[i];
    }
  }  
       
  for (i=0; i<10; i++) 
  {
    cnt_match[i] = 0;
    cnt_err_th[i] = 0;
  }
  cnt_fill = 0;
/*  printf("esdt:%s\n",esdt);*/

  if ((data1 = (void *)calloc(ndata, sds1_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data1 in comp_sds_stat\n");
  if ((data2 = (void *)calloc(ndata, sds2_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data2 in comp_sds_stat\n");
  if ((data3 = (void *)calloc(ncols, sds3_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data3 in comp_sds_stat\n");
  if ((data1 != NULL) && (data2 != NULL) && (data3 != NULL)) 
  { 
    out_edge[0] = 1;
    out_edge[1] = ncols;
    if (sds1_info->rank == 2) 
    {
      in_edge[0] = 1;
      in_edge[1] = ncols;
    }
    else if (rank > 2)
    {
      for (i=0; i<rank; i++)
        in_edge[i] = sds1_info->dim_size[i];
      if ((n == -1) && (m == -1))
      {
        if (sds1_info->dim_size[0] > sds1_info->dim_size[rank-1])
          in_edge[0] = 1;
        else
          in_edge[rank-2] = 1; 
      }
      else
      {
        if (sds1_info->dim_size[0] > sds1_info->dim_size[rank-1])
          in_edge[0] = 1;
        else
          in_edge[rank-2] = 1;
      }
    }                
    compute_sds_start_offset(sds1_info, n, m, &st_c, &offset); 
/*printf("ndata = %d\n", ndata);  
printf("ncols = %d\n", ncols);       
printf("n = %d\n", n);   
printf("m = %d\n", m);      
printf("st_c = %d\n", st_c);   
printf("offset = %d\n", offset);   */  
    for (irow=0; irow<nrows; irow++) 
    {
      if (sds1_info->rank == 2)
        in_start[0] = irow; 
      else 
      {
        if (sds1_info->dim_size[0] > sds1_info->dim_size[2])
          in_start[0] = irow; 
	else
          in_start[1] = irow; 
      }
      out_start[0] = irow;

      if (read_sel_sds_data(sds1_info, in_start, in_edge, data1) == FAIL)
      {
        fprintf(stderr, "Error reading data line %d from SDS %s\n", irow, sds1_info->name);
        break;
      }    
      if (read_sel_sds_data(sds2_info, in_start, in_edge, data2) == FAIL)
      {
        fprintf(stderr, "Error reading data line %d from SDS %s\n", irow, sds2_info->name);
        break;
      }        

      for (icol=0, ic=st_c; icol<ncols; icol++) 
      {
        if (sds1_info->rank == 2) 
	{
          switch(sds1_info->data_type) 
	  {
            case  5: fval1 = ((float32 *)data1)[icol];
                     fval2 = ((float32 *)data2)[icol];
		     break;
            case 20: val1 = ((int8 *)data1)[icol];
                     val2 = ((int8 *)data2)[icol];
		     break;
            case 21: val1 = ((uint8 *)data1)[icol];
                     val2 = ((uint8 *)data2)[icol];
		     break;
            case 22: val1 = ((int16 *)data1)[icol];
                     val2 = ((int16 *)data2)[icol];
		     break;
            case 23: val1 = ((uint16 *)data1)[icol];
                     val2 = ((uint16 *)data2)[icol];
		     break;
            case 24: val1 = ((int32 *)data1)[icol];
                     val2 = ((int32 *)data2)[icol];
		     break;
            case 25: val1 = ((uint32 *)data1)[icol];
                     val2 = ((uint32 *)data2)[icol];
		     break;
	  }
        }
        else 
	{
          switch(sds1_info->data_type) 
	  {
            case  5: fval1 = ((float32 *)data1)[ic];
                     fval2 = ((float32 *)data2)[ic];
		     break;
            case 20: val1 = ((int8 *)data1)[ic];
                     val2 = ((int8 *)data2)[ic];
		     break;
            case 21: val1 = ((uint8 *)data1)[ic];
                     val2 = ((uint8 *)data2)[ic];
		     break;
            case 22: val1 = ((int16 *)data1)[ic];
                     val2 = ((int16 *)data2)[ic];
		     break;
            case 23: val1 = ((uint16 *)data1)[ic];
                     val2 = ((uint16 *)data2)[ic];
		     break;
            case 24: val1 = ((int32 *)data1)[ic];
                     val2 = ((int32 *)data2)[ic];
		     break;
            case 25: val1 = ((uint32 *)data1)[ic];
                     val2 = ((uint32 *)data2)[ic];
		     break;
	  }
	  ic += offset;
	}
        if (is_qa == 1) 
	{
	  /* The qa data overrides the esdt name */
          if (((val1 != sds1_info->fill_val) && (val2 != sds2_info->fill_val)) ||
             ((val1 == sds1_info->fill_val) && (val2 == sds2_info->fill_val))) 
	  {
            val3 = val1 - val2;
            per_err = (val3 == 0) ? 0 : 5;
	  }
          else 
	    per_err = SDS_FILL;
        }
        else 
	{
	  /* The float data type overrides the esdt */
          if (sds1_info->data_type == 5) 
	  {
	    if (((fval1 != sds1_info->fill_fval) && (fval2 != sds2_info->fill_fval)) || 
	       ((fval1 == sds1_info->fill_fval) && (fval2 == sds2_info->fill_fval))) 
	    {
	      diff = fval2 - fval1;
	      abs_fval1 = fabsf(fval1);
	      if (is_rel == 1) 
	      {
		if (abs_fval1 != 0)
		{	      
	          fper_err = diff*100.0/abs_fval1;
	          if (fper_err > 100) fper_err = 100.0;
	          else if (fper_err < -100) fper_err = -100.0;
		  for (i=0; i<tcnt; i++) 
		  {
	            if (fabsf(fper_err) > err_th[i]) cnt_err_th[i]++;
		    else cnt_match[i]++;
		  }		
	          per_err = (int)(fper_err*SCALE_DIFF);		
	        }
		else if (diff == 0)	
		{
		  per_err = 0;
		  for (i=0; i<tcnt; i++) 
	            cnt_match[i]++;		    
		}
		else 
		{
		  for (i=0; i<tcnt; i++)
		    cnt_err_th[i]++;
		  per_err = (int)(100*SCALE_DIFF);		  
		}
	      }
	      else 
		per_err = diff;      
	    }
	    else 
	    {
	      cnt_fill++;
	      per_err = SDS_FILL;
	    }  
          }
	  else 
	  {
            if (strcmp(esdt, "MOD09") == 0) 
	    {	    	      
	      if (((val1 != sds1_info->fill_val) && (val2 != sds2_info->fill_val)) || 
	         ((val1 == sds1_info->fill_val) && (val2 == sds2_info->fill_val))) 
	      {
		diff = val2 - val1;
	        if (is_rel == 1) 
		{
	          abs_fval1 = fabsf((float)val1);
		  if (abs_fval1 != 0)
		  {
		    //if ((diff >= -2) && (diff <= 2)) diff = 0;              /* ignore a difference of 2 count values */
		    fper_err = diff*100.0/abs_fval1;		  
	            if (fper_err > 100) fper_err = 100.0;
	            else if (fper_err < -100) fper_err = -100.0;
		    per_err = (int)(fper_err*SCALE_DIFF);
		    for (i=0; i<tcnt; i++) {
	              if (fabsf(fper_err) > err_th[i]) cnt_err_th[i]++;
		      else cnt_match[i]++;
		    }		  
		  }
		  else if (diff == 0)	
		  {
		    per_err = 0;
		    for (i=0; i<tcnt; i++) 
	              cnt_match[i]++;		    
		  }
		  else 
		  {		     
		    for (i=0; i<tcnt; i++)
		      cnt_err_th[i]++;
		    per_err = (int)(100*SCALE_DIFF);		    
		  }
		}
		else 
		  per_err = diff;
	      }
	      else 
	      {
	        cnt_fill++;
	        per_err = SDS_FILL;
	      }		
	    }
	    else if (strcmp(esdt, "MOD10") == 0)
        {
            if ((val1 != sds1_info->fill_val) && (val2 != sds2_info->fill_val))
            {
              /* for proper snow areas compute usual relative diff */
              if ((val1 > 0 && val1 <= 100) && (val2 > 0 && val2 <= 100)) 
              {
                  diff = (float)val2 - (float)val1;
                  if (is_rel == 1)
                  {
                      abs_fval1 = (float)val1;
                      if (abs_fval1 != 0)
                      {
                        /* if ((diff >= -2) && (diff <= 2)) diff = 0;  */            /* ignore a difference of 2 count values */
                        fper_err = diff*100.0/abs_fval1;		  
                        if (fper_err > 100) fper_err = 100.0;
                        else if (fper_err < -100) fper_err = -100.0;
                        per_err = (int32)(fper_err*RELSCALE_DIFF);
                      }
                      else
                      {
                        if (diff == 0)
                          per_err = 0;		     
                        else
                          per_err = (int32)(100*RELSCALE_DIFF);
                      }
                  }  		  
                  else
                  {
                      per_err = (int32)(diff*ACTSCALE_DIFF);		  
                  }
              }
              else
              {
                /* for all other classes compute match and mismatch as before */
                if (val1 == val2) { val3 = 0; } /* match */
                else if (val1 != val2) val3 = SNOW_ERR; /* mismatch */
                per_err = val3;
              }
            }
            else 
              per_err = SDS_FILL;
        }
        else if (strcmp(esdt, "MOD29") == 0) 
	    {		     		     
    /*	      if (((val1 != sds1_info->fill_val) && (val2 != sds2_info->fill_val)) || 
                 ((val1 == sds1_info->fill_val) && (val2 == sds2_info->fill_val))) {*/
            if ((val1 == 200) || (val1 == 100)) nsnow_fire++;
                if (val1 == val2) { val3 = 0; cnt_match[0]++;} /* match */
                else if ((val1 == 200) && (val2 != 200)) {val3 = 1; missed_cnt++;}/* missed snow */
                else if ((val2 == 200) && (val1 != 200)) {val3 = 2; false_cnt++;}/* false snow */
                else if ((val1 == 100) && (val2 != 100)) {val3 = 3; missed_cnt++;}/* missed lake ice */
                else if ((val2 == 100) && (val1 != 100)) {val3 = 4; false_cnt++;} /* false lake ice */
                else {val3 = 5; other_cnt++;}/* other difference */
    /*		if ((val1 == 200) && (val2 == 200)) cnt_match[0]++;
            else if ((val1 == 100) && (val2 == 100)) cnt_match[0]++;*/		
                    per_err = val3;
    /*	        if (val3 == 0) printf("row=%d   col=%d  val3=%d  val1=%d   val2=%d\n", irow, icol, val3, val1, val2);
            if (val3 == 1 || val3 == 3) printf("row=%d   col=%d  val3=%d   val1=%d   val2=%d\n", irow, icol, val3, val1, val2);
            if (val3 == 2 || val3 == 4) printf("row=%d   col=%d  val3=%d   val1=%d   val2=%d\n", irow, icol, val3, val1, val2);*/	
    /*	      }
              else 
            per_err = SDS_FILL;*/
        }
	    else if (strcmp(esdt, "MOD14") == 0) 
	    {		 
/*	      if (((val1 != sds1_info->fill_val) && (val2 != sds2_info->fill_val)) || 
	         ((val1 == sds1_info->fill_val) && (val2 == sds2_info->fill_val))) { */
                if ((val1 >= 7) && (val1 <= 9)) nsnow_fire++;		 
                val1_st = ((val1 >= 7) && (val1 <= 9)) ? 1 : 0; 
                val2_st = ((val2 >= 7) && (val2 <= 9)) ? 1 : 0; 
	        if (val1 == val2) { val3 = 0; } /* match */
	        else if ((val1_st == 1) && (val2_st == 1)) {val3 = 0; cnt_match[0]++;}/* fire match */
	        else if ((val1_st == 1) && (val2_st == 0)) {val3 = 1; missed_cnt++;}/* missed fire */
	        else if ((val1_st == 0) && (val2_st == 1)) {val3 = 2; false_cnt++;}/* false fire */
	        else {val3 = 5; other_cnt++;}/* other difference */
                per_err = val3;
/*	      }
	      else
		per_err = SDS_FILL;*/
            }
            else if (strcmp(esdt, "MOD14CRS") == 0) 
	    {
              if (((val1 != sds1_info->fill_val) && (val2 != sds2_info->fill_val)) ||
                 ((val1 == sds1_info->fill_val) && (val2 == sds2_info->fill_val))) 	 		 
	      {
                if ((val1 >= 1) && (val1 <= 3)) nsnow_fire++;	
		val1_st = ((val1 >= 1) && (val1 <= 3)) ? 1 : 0;
                val2_st = ((val2 >= 1) && (val2 <= 3)) ? 1 : 0;
		if (val1 == val2) { val3 = 0;cnt_match[0]++;} /* match */
                else if ((val1_st == 1) && (val2_st == 1)) {val3 = 0; } /* fire match */
                else if ((val1_st == 1) && (val2_st == 0)) {val3 = 1; missed_cnt++;} /* missed fire */
                else if ((val1_st == 0) && (val2_st == 1)) {val3 = 2; false_cnt++;} /* false fire */
                else {val3 = 5; other_cnt++;} /* other difference */
                per_err = val3;
              }
              else
                per_err = SDS_FILL;
            }	    	    
            else if (strcmp(esdt, "CLOUD") == 0) 
	    {
              if (((val1 != sds1_info->fill_val) && (val2 != sds2_info->fill_val)) ||
        	 ((val1 == sds1_info->fill_val) && (val2 == sds2_info->fill_val))) 
	      {
                val1_st = ((val1 >= 0) && (val1 <= 1)) ? 1 : 0; 
                val2_st = ((val2 >= 0) && (val2 <= 1)) ? 1 : 0; 
		val3 = val1_st - val2_st;
   		per_err = (val3 == 0) ? 0 : 5;
	      }
              else 
		per_err = SDS_FILL;
            }	    	    
	  }
	}
        switch(sds3_info->data_type) 
	{
          case  5: ((float32 *)data3)[icol] = (float32)per_err; break;
          case 20: ((int8 *)data3)[icol] = (int8)per_err; break;
          case 21: ((uint8 *)data3)[icol] = (uint8)per_err; break;
          case 22: ((int16 *)data3)[icol] = (int16)per_err; break;
          case 23: ((uint16 *)data3)[icol] = (uint16)per_err; break;
          case 24: ((int32 *)data3)[icol] = (int32)per_err; break;
          case 25: ((uint32 *)data3)[icol] = (uint32)per_err; break;
        }
      } /* for (icol=0; . . . ) */
      if(set_sel_sds_data(sds3_info, out_start, out_edge, data3) == FAIL)
      {
        fprintf(stderr, "Error reading data line %d from SDS %s\n", irow, sds3_info->name);
        break;
      }
    } /* for (irow=0;  . . .) */
    
    if (st_opt == 1 && is_qa != 1) 
    {
      tot_pix = nrows*ncols;
      if (strcmp(esdt, "MOD09") == 0) 
      {
        if (is_rel == 1) 
	{	  	  
          fprintf(stdout, "File_name   	                             threshold(%) pix_match(%) pix_err(%) pix_fill(%)\n" );            	
	  for (i=0; i<tcnt; i++) 
	  { 
	    per_pix_err = (double)cnt_err_th[i]*100.0/(double)tot_pix;
            per_pix_match = (double)cnt_match[i]*100.0/(double)tot_pix;
            per_pix_fill = (double)cnt_fill*100.0/(double)tot_pix;
            fprintf(stdout, "%s\t%.2f\t   %.2f\t %.2f\t    %.2f\n", gid, err_th[i], per_pix_match, per_pix_err, per_pix_fill);
	  }
	}
      }
    }  
    else if ((is_qa != 1) && (strcmp(esdt, "MOD09") != 0) && (strcmp(esdt, "CLOUD") != 0))  
    {	  
	tot_pix = nrows*ncols;
	per_pix_miss = (nsnow_fire> 0) ? (double)missed_cnt*100.0/(double)nsnow_fire : 0;
	per_pix_false = (nsnow_fire> 0) ? (double)false_cnt*100.0/(double)nsnow_fire : 0;
	per_pix_match = (double)cnt_match[0]*100.0/(double)tot_pix;
	per_pix_nsnow_fire = (nsnow_fire > 0) ? (double)nsnow_fire*100.0/(double)tot_pix : 0.0;
        per_pix_err = (nsnow_fire> 0) ? ((double)missed_cnt + (double)false_cnt)*100.0/(double)nsnow_fire : 0.0;
	
	fprintf(stdout, "File_name  			                      missed_cnt  false_cnt  other_cnt    nsnow_fire        tot_pix\n");
        fprintf(stdout, "%s\t%ld\t\t%ld\t %ld\t     %ld\t     %ld\n", gid, missed_cnt, false_cnt, other_cnt, nsnow_fire, tot_pix); 	
	fprintf(stdout, "     			                              missed_cnt(%) false_cnt(%) per_pix_err(%) cnt_match(%) nsnow_fire(%)\n");
        fprintf(stdout, "\t\t\t\t\t\t\t %.3f\t\t %.3f\t %.3f\t\t %.3f\t\t %.3f\n", per_pix_miss, per_pix_false, per_pix_err, per_pix_match, per_pix_nsnow_fire); 
    } 
       
  }
  free(data1);
  free(data2);
  free(data3);
  return;
}

void comp_sds_diff_npp(sds_t *sds1_info, sds_t *sds2_info, sds_t *sds3_info, char *sds1, char *esdt, int is_qa, int is_rel, char *gid, 
	float *err_th, int tcnt, int st_opt)

{
  char ib_str[10];
  int irow, icol, ic;
  int p1, p2, i, len, ib, nb;
  int nrows, ncols, ndata;
  int val1_st, val2_st;
  int val1, val2, val3;
  float abs_fval1, diff, fper_err;
  void *data1, *data2, *data3;
  int32 in_start[3] = {0, 0, 0};
  int32 in_edge[3] = {0, 0, 0};
  int32 out_start[2] = {0, 0};
  int32 out_edge[2] = {0, 0};
  int per_err;  
  float fval1, fval2, fval3;  
  double per_pix_err;
  double per_pix_match, per_pix_fill;
  double per_pix_miss, per_pix_false, per_pix_nsnow_fire;
  long int tot_pix;
  long int cnt_match[10];
  long int cnt_err_th[10];
  long int cnt_fill=0;
  long int missed_cnt = 0;
  long int false_cnt = 0;
  long int other_cnt = 0;
  long int nsnow_fire = 0; 
  char Esdt[10], tmp[20];
  int n, m, st_c, offset, rank;
  char sds_name[MAX_SDS_NAME_LEN];   
/*
printf("sds1_info->data_type : sds2_info->data_type is %d:%d \n",sds1_info->data_type,sds2_info->data_type);    
printf("sds1_info->fill_fval : sds2_info->fill_fval is %f:%f\n",sds1_info->fill_fval,sds2_info->fill_fval);
printf("sds1_info->fill_val : sds2_info->fill_val is %d:%d \n",sds1_info->fill_val,sds2_info->fill_val);    
printf("sds1_info->scalef : sds1_info->offsetf is %f:%f \n",sds1_info->scalef, sds1_info->offsetf);  
*/      
  if ((p1 = charpos(gid,'.',0)) != -1)
  {
    strmid(gid, 0, p1, tmp);
    if ((p1 = charpos(tmp,'_',0)) != -1)
    {
      p2 = charpos(tmp,'_',p1+1);
      strmid(tmp,p1+1,(p2-p1)-1,Esdt);
    }
  } 
//printf("Esdt=%s\n",Esdt);  
  for (i=0; i<10; i++) {
    cnt_match[i] = 0;
    cnt_err_th[i] = 0;
  }  
  
  get_sdsname_dim(sds1, sds_name, &n, &m);   
  rank = sds1_info->rank;
  if (sds1_info->rank == 2)
  {
    nrows = sds1_info->dim_size[0];
    ncols = ndata = sds1_info->dim_size[1];
  }                                                                            
  else if (sds1_info->rank > 2)
  {
    if (sds1_info->dim_size[0] > sds1_info->dim_size[rank-1])
    {
      nrows = sds1_info->dim_size[0];
      ncols = ndata = sds1_info->dim_size[1];
      for (i=2; i<rank; i++)
        ndata *= sds1_info->dim_size[i];
    }
    else
    {
      nrows = sds1_info->dim_size[rank-2];
      ncols = ndata = sds1_info->dim_size[rank-1];
      for (i=0; i<rank-2; i++)
        ndata *= sds1_info->dim_size[i];
    }
  }    

  if ((data1 = (void *)calloc(ndata, sds1_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data1 in comp_sds_stat\n");
  if ((data2 = (void *)calloc(ndata, sds2_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data2 in comp_sds_stat\n");
  if ((data3 = (void *)calloc(ncols, sds3_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data3 in comp_sds_stat\n");
  if ((data1 != NULL) && (data2 != NULL) && (data3 != NULL)) 
  {
    out_edge[0] = 1;
    out_edge[1] = ncols;
    if (sds1_info->rank == 2) 
    {
      in_edge[0] = 1;
      in_edge[1] = ncols;
    }
    else if (rank > 2)
    {
      for (i=0; i<rank; i++)
        in_edge[i] = sds1_info->dim_size[i];
      if ((n == -1) && (m == -1))
      {
        if (sds1_info->dim_size[0] > sds1_info->dim_size[rank-1])
          in_edge[0] = 1;
        else
          in_edge[rank-2] = 1; 
      }
      else
      {
        if (sds1_info->dim_size[0] > sds1_info->dim_size[rank-1])
          in_edge[0] = 1;
        else
          in_edge[rank-2] = 1;
      }
    }                    
    compute_sds_start_offset(sds1_info, n, m, &st_c, &offset); 
/*printf("ndata = %d\n", ndata);  
printf("ncols = %d\n", ncols);       
printf("n = %d\n", n);   
printf("m = %d\n", m);      
printf("st_c = %d\n", st_c);   
printf("offset = %d\n", offset);   */    
    for (irow=0; irow<nrows; irow++) 
    {
      if (sds1_info->rank == 2)
        in_start[0] = irow; 
      else 
      {
        if (sds1_info->dim_size[0] > sds1_info->dim_size[2])
          in_start[0] = irow; 
	else
          in_start[1] = irow; 
      }
      out_start[0] = irow;
      if (read_sel_sds_data(sds1_info, in_start, in_edge, data1) == FAIL)
      {
        fprintf(stderr, "Error reading data line %d from SDS %s\n", irow, sds1_info->name);
        break;
      }
      if (read_sel_sds_data(sds2_info, in_start, in_edge, data2) == FAIL)
      {
        fprintf(stderr, "Error reading data line %d from SDS %s\n", irow, sds2_info->name);
        break;
      }
      for (icol=0, ic=st_c; icol<ncols; icol++) 
      {
        if ((sds1_info->rank == 2) || (sds1_info->dim_size[0] < sds1_info->dim_size[2])) 
	{
          switch(sds1_info->data_type) 
	  {
            case  5: fval1 = ((float32 *)data1)[icol];
		     break;
            case 20: val1 = ((int8 *)data1)[icol];
		     break;
            case 21: val1 = ((uint8 *)data1)[icol];
		     break;
            case 22: val1 = ((int16 *)data1)[icol];
		     break;
            case 23: val1 = ((uint16 *)data1)[icol];
		     break;
            case 24: val1 = ((int32 *)data1)[icol];
		     break;
            case 25: val1 = ((uint32 *)data1)[icol];
		     break;
	  }
	  switch(sds2_info->data_type) 
	  {
            case  5: fval2 = ((float32 *)data2)[icol];
		     break;
            case 20: val2 = ((int8 *)data2)[icol];
		     break;
            case 21: val2 = ((uint8 *)data2)[icol];
		     break;
            case 22: val2 = ((int16 *)data2)[icol];
		     break;
            case 23: val2 = ((uint16 *)data2)[icol];
		     break;
            case 24: val2 = ((int32 *)data2)[icol];
		     break;
            case 25: val2 = ((uint32 *)data2)[icol];
		     break;
	  }
        }
        else {
          switch(sds1_info->data_type) 
	  {
            case  5: fval1 = ((float32 *)data1)[ic];
		     break;
            case 20: val1 = ((int8 *)data1)[ic];
		     break;
            case 21: val1 = ((uint8 *)data1)[ic];
		     break;
            case 22: val1 = ((int16 *)data1)[ic];
		     break;
            case 23: val1 = ((uint16 *)data1)[ic];
		     break;
            case 24: val1 = ((int32 *)data1)[ic];
		     break;
            case 25: val1 = ((uint32 *)data1)[ic];
		     break;
	  }
	  switch(sds2_info->data_type) 
	  {
            case  5: fval2 = ((float32 *)data2)[ic];
		     break;
            case 20: val2 = ((int8 *)data2)[ic];
		     break;
            case 21: val2 = ((uint8 *)data2)[ic];
		     break;
            case 22: val2 = ((int16 *)data2)[ic];
		     break;
            case 23: val2 = ((uint16 *)data2)[ic];
		     break;
            case 24: val2 = ((int32 *)data2)[ic];
		     break;
            case 25: val2 = ((uint32 *)data2)[ic];
		     break;
	  }
	  ic += offset;
	}
        if ((is_qa == 1) || (strcmp(Esdt,"CMIP") == 0)) 
	{
	  /* The qa data overrides the esdt name  and the CMIP esdt is treated like a qa layer*/
          if (((val1 < sds1_info->fill_uval) && (val2 < sds2_info->fill_uval)) ||
             ((val1 >= sds1_info->fill_uval) && (val2 >= sds2_info->fill_uval))) 
	  {
            val3 = val1 - val2;
            per_err = (val3 == 0) ? 0 : 5;
	  }
          else 
	    per_err = SDS_FILL;
        }
        else 
	{
	  /* The float data type overrides the esdt */
          if (sds1_info->data_type == 5 && sds2_info->data_type == 5) 
	  {
	    if ((fval1 > sds1_info->fill_fval) && (fval2 > sds2_info->fill_fval)) 
	    {
/*
printf("fval1:fval2 is %f:%f \n", fval1, fval2);
printf("sds1_info->fill_fval : sds2_info->fill_fval is %f:%f \n",sds1_info->fill_fval,sds2_info->fill_fval );
printf("sds1_info->fscale : sds1_info->fscale is %f:%f \n",sds1_info->fscale, sds1_info->fscale);	    
*/
	      diff = fval2 - fval1;
	      abs_fval1 = fabsf(fval1);    
	      if (is_rel == 1) 	      	 
	      {
		if (abs_fval1 != 0)
		{
		  fper_err = diff*100.0/abs_fval1;		  
	          if (fper_err > 100) fper_err = 100.0;
	          else if (fper_err < -100) fper_err = -100.0;
		  for (i=0; i<tcnt; i++) {
	            if (fabsf(fper_err) > err_th[i]) cnt_err_th[i]++;
		    else cnt_match[i]++;
		  }		  
		  per_err = (int)(fper_err*RELSCALE_DIFF);	  
		}
		else 
		{
		  if (diff == 0)
		  {
		    per_err = 0;		     
		    for (i=0; i<tcnt; i++) 
	              cnt_match[i]++;		  
		  }
		  else
		  {
		    for (i=0; i<tcnt; i++)
		      cnt_err_th[i]++;		    
		    per_err = 100.0;
		    per_err = (int)(100*RELSCALE_DIFF);
		  }	      
	        }
	      }
	      else 
	      {
		per_err = (int)(diff*ACTSCALE_DIFF);		
	      }
	    }
	    else  if ((fval1 <= sds1_info->fill_fval) && (fval2 <= sds2_info->fill_fval))
	    {
	      per_err = 0;
	      cnt_fill++;
	    }
	    else
	    {
	      cnt_fill++;
	      per_err = SDS_FILL;
	    }  
          }
	  else if (strcmp(Esdt, "VSCM") == 0) 
	  {
	    if (((val1 < sds1_info->fill_uval) && (val2 < sds2_info->fill_uval)) ||
		((val1 >= sds1_info->fill_uval) && (val2 >= sds2_info->fill_uval))) 
	    {
	      if (val1 == val2 ) 
	      { 
		val3 = 0; 
		cnt_match[0]++;/* match */
		if ((val1 == 1) && (val2 == 1)) {nsnow_fire++;cnt_match[0]++;} /* snow count*/
	      } 
	      else if ((val1 == 1) && (val2 != 1)) {val3 = 1; missed_cnt++;} /* missed snow */
	      else if ((val2 == 1) && (val1 != 1)) {val3 = 2; false_cnt++;} /* false snow */
	      else {val3 = 5; other_cnt++;} /* other difference */
              per_err = val3;
	    }
	    else 
	      per_err = SDS_FILL;
          }
	  else if (strcmp(Esdt, "VAFIP") == 0) 
	  {
	    if (((val1 < sds1_info->fill_uval) && (val2 < sds2_info->fill_uval)) ||
		((val1 >= sds1_info->fill_uval) && (val2 >= sds2_info->fill_uval))) 
	    {
              val1_st = ((val1 >= 7) && (val1 <= 9)) ? 1 : 0; 
              val2_st = ((val2 >= 7) && (val2 <= 9)) ? 1 : 0; 
	      if (val1_st == val2_st) { val3 = 0; nsnow_fire++; cnt_match[0]++;} /* match */
	      else if ((val1_st == 1) && (val2_st == 0)) {val3 = 1; missed_cnt++;} /* missed fire */
	      else if ((val1_st == 0) && (val2_st == 1)) {val3 = 2; false_cnt++;} /* false fire */
	      else val3 = 5; /* other difference */
              per_err = val3;
	    }
	    else
	      per_err = SDS_FILL;
          }
	  else if (strcmp(esdt, "MOD09") == 0)
	  {	  	  
	    /* Default case for other integer types
	      First scale the values */
	    if (sds1_info->fill_val < 0)
	      fval1 = (val1 > sds1_info->fill_val)?(val1*sds1_info->fscale)+sds1_info->offset:INVALID;
	    else if (sds1_info->fill_val > 0)
	      fval1 = (val1 < sds1_info->fill_val)?(val1*sds1_info->fscale)+sds1_info->offset:INVALID;
	    else if (sds1_info->fill_uval != 0)
	      fval1 = (val1 < sds1_info->fill_uval)?(val1*sds1_info->fscale)+sds1_info->offset:INVALID;
	    else
	      /* MSZ corrected bug on 9/26/2018. So far, this is the case for LST */
	      /*fval1 = (fval1 > sds1_info->fill_fval)?fval1:INVALID;*/
	      fval1 = (val1 > sds1_info->fill_fval)?(val1*sds1_info->fscale)+sds1_info->offset:INVALID;

	    if (sds2_info->fill_val < 0)
	      fval2 = (val2 > sds2_info->fill_val)?(val2*sds2_info->fscale)+sds2_info->offset:INVALID;
	    else if (sds2_info->fill_val > 0)
		fval2 = (val2 < sds2_info->fill_val)?(val2*sds2_info->fscale)+sds2_info->offset:INVALID;
	    else if (sds2_info->fill_uval != 0)
		fval2 = (val2 < sds2_info->fill_uval)?(val2*sds2_info->fscale)+sds2_info->offset:INVALID;
	    else
		/* MSZ corrected bug on 9/26/2018. So far, this is the case for LST */
		/*fval2 = (fval2 > sds2_info->fill_fval)?fval2:INVALID;*/
		fval2 = (val2 > sds2_info->fill_fval)?(val2*sds2_info->fscale)+sds2_info->offset:INVALID;
/*if (irow == 4 && icol == 0) 
{
  printf("fval1:fval2 is %f:%f \n", fval1, fval2);
  printf("sds1_info->fill_val : sds2_info->fill_val is %d:%d \n",sds1_info->fill_val,sds2_info->fill_val );
  printf("sds1_info->fill_fval : sds2_info->fill_fval is %f:%f \n",sds1_info->fill_fval,sds2_info->fill_fval );
  printf("sds1_info->fill_uval : sds2_info->fill_uval is %d:%d \n",sds1_info->fill_uval,sds2_info->fill_uval );  
  printf("sds1_info->fscale : sds2_info->fscale is %f:%f \n",sds1_info->fscale, sds2_info->fscale);
  printf("sds1_info->scale : sds2_info->scale is %d:%d \n",sds1_info->scale, sds2_info->scale);
  printf("sds1_info->foffset : sds2_info->foffset is %f:%f \n",sds1_info->foffset, sds2_info->foffset);
  printf("sds1_info->offset : sds2_info->offset is %d:%d \n",sds1_info->offset, sds2_info->offset);    
}	*/

	    if ((fval1 != INVALID) && (fval2 != INVALID)) 
	    {
	      diff = fval2 - fval1;
	      if (is_rel == 1) {
		abs_fval1 = fabsf(fval1);
		if (abs_fval1 != 0)
		{
		  /* if ((diff >= -2) && (diff <= 2)) diff = 0;  */            /* ignore a difference of 2 count values */
		  fper_err = diff*100.0/abs_fval1;		  
		  if (fper_err > 100) fper_err = 100.0;
		  else if (fper_err < -100) fper_err = -100.0;
		  for (i=0; i<tcnt; i++) {
	            if (fabsf(fper_err) > err_th[i]) cnt_err_th[i]++;
		    else cnt_match[i]++;
		  }		  		  
		  per_err = (int)(fper_err*RELSCALE_DIFF);
		}
		else 
		{
		  if (diff == 0)
		  {
		    per_err = 0;	
		    for (i=0; i<tcnt; i++) 
	              cnt_match[i]++;			    	     
		  }
		  else
		  {
		    for (i=0; i<tcnt; i++)
		      cnt_err_th[i]++;		    		    
		    per_err = (int)(100*RELSCALE_DIFF);		    
		  }
		}
	      }  		  
	      else 
	      {
		per_err = (int)(diff*ACTSCALE_DIFF);		  
	      }
	      /* printf("fval1 is %f and fval2 is %f and diff is %f and per_err is %d\n",fval1,fval2,diff,per_err);*/
	    }
	    /* MSZ 10/31/2018, turn on the comment out for the case of both are fill values */
	    else if ((fval1 == INVALID) && (fval2 == INVALID))
	    {
	      per_err = 0;	    
	      for (i=0; i<tcnt; i++) 
		cnt_match[i]++;		    
	    }
	    else	    
	    {
	      cnt_fill++;
	      per_err = SDS_FILL;
	    }  	  	    		  		  		  
	  }
	}	
        switch(sds3_info->data_type) 
	{
          case  5: ((float32 *)data3)[icol] = (float32)per_err; break;
          case 20: ((int8 *)data3)[icol] = (int8)per_err; break;
          case 21: ((uint8 *)data3)[icol] = (uint8)per_err; break;
          case 22: ((int16 *)data3)[icol] = (int16)per_err; break;
          case 23: ((uint16 *)data3)[icol] = (uint16)per_err; break;
          case 24: ((int32 *)data3)[icol] = (int32)per_err; break;
          case 25: ((uint32 *)data3)[icol] = (uint32)per_err; break;
        }
      } /* for (icol=0; . . . ) */
      if(set_sel_sds_data(sds3_info, out_start, out_edge, data3) == FAIL)
      {
        fprintf(stderr, "Error reading data line %d from SDS %s\n", irow, sds3_info->name);
        break;
      }
    } /* for (irow=0;  . . .) */

    if (st_opt == 1 && is_qa != 1) 
    {
      tot_pix = nrows*ncols;
      if (strcmp(esdt, "MOD09") == 0) 
      {
        if (is_rel == 1) 
	{	  	  
          fprintf(stdout, "File_name   	                                    threshold(%) pix_match(%) pix_err(%) pix_fill(%)\n" );            	
	  for (i=0; i<tcnt; i++) 
	  { 
	    per_pix_err = (double)cnt_err_th[i]*100.0/(double)tot_pix;
            per_pix_match = (double)cnt_match[i]*100.0/(double)tot_pix;
            per_pix_fill = (double)cnt_fill*100.0/(double)tot_pix;
            fprintf(stdout, "%s\t%.2f\t   %.2f\t %.2f\t    %.2f\n", gid, err_th[i], per_pix_match, per_pix_err, per_pix_fill);
	  }
	}
      }
    }  
    else if (strcmp(Esdt, "VSCM") == 0 || strcmp(Esdt, "VAFIP") == 0)  
    {	  
	tot_pix = nrows*ncols;
	per_pix_miss = (double)missed_cnt*100.0/(double)cnt_match[0];
	per_pix_false = (double)false_cnt*100.0/(double)cnt_match[0];
	per_pix_match = (double)cnt_match[0]*100.0/(double)tot_pix;
	per_pix_nsnow_fire = (nsnow_fire > 0) ? (double)nsnow_fire*100.0/(double)tot_pix : 0.0;
        per_pix_err = (cnt_match[0] > 0) ? ((double)missed_cnt + (double)false_cnt)*100.0/(double)cnt_match[0] : 0.0;
	fprintf(stdout, "File_name  			                      missed_cnt  false_cnt  other_cnt    nsnow_fire        tot_pix\n");
        fprintf(stdout, "%s\t%ld\t\t%ld\t %ld\t     %ld\t     %ld\n", gid, missed_cnt, false_cnt, other_cnt, nsnow_fire, tot_pix); 	
	fprintf(stdout, "     			                              missed_cnt(%) false_cnt(%) per_pix_err(%) cnt_match(%) nsnow_fire(%)\n");
        fprintf(stdout, "\t\t\t\t\t\t\t %.3f\t\t %.3f\t %.3f\t\t %.3f\t\t %.3f\n", per_pix_miss, per_pix_false, per_pix_err, per_pix_match, per_pix_nsnow_fire); 
    }     
    
  }
  free(data1);
  free(data2);
  free(data3);
  return;
}

void copy_geolocation(sds_t *in_sds_info, sds_t *out_sds_info, char **sds_names, int nsds)
{
  void *data;
  int32 edge[2] = {0, 0};
  int32 start[2] = {0, 0};
  int i, geo_st, isds, irow, nrows, ncols;
  sds_t geo_sds_info, out_geo_sds_info;
  char *geo_sds_names[] = {"Latitude", "Longitude"};
  int fire_st = 0;
    
  geo_sds_info.sd_id = in_sds_info->sd_id;
  geo_sds_info.H5IN = in_sds_info->H5IN;
  out_geo_sds_info.sd_id = out_sds_info->sd_id;
  out_geo_sds_info.H5OUT = out_sds_info->H5OUT;

  for (isds=0; isds<nsds; isds++)   
    if (strstr(sds_names[isds], "CoarseFireCounts") != NULL) 
      fire_st = 1;
  for (isds=0; isds<nsds; isds++) {
    geo_st = 0;
    /* MSZ 10-24-2018, make it case insensitive. Note that strcasestr is not standard function so use strstr_case */
    /*if (strstr(sds_names[isds], geo_sds_names[0]) != NULL) {*/
    if (strstr_case(sds_names[isds], geo_sds_names[0]) != NULL) {
      strcpy(geo_sds_info.name, sds_names[isds]);
      geo_st = 1;
    }

    /* MSZ 10-24-2018, make it case insensitive. Note that strcasestr is not standard function so use strstr_case */	
    /*else if (strstr(sds_names[isds], geo_sds_names[1]) != NULL) {*/
    else if (strstr_case(sds_names[isds], geo_sds_names[1]) != NULL) {
      strcpy(geo_sds_info.name, sds_names[isds]);
      geo_st = 1;
    }
    if (geo_st == 1) {
      geo_sds_info.sds_id = -1;
      if (get_sds_info((char *)NULL, &geo_sds_info) != -1)
      {
        out_geo_sds_info.sds_id = -1;
        strcpy(out_geo_sds_info.name, geo_sds_info.name);
        out_geo_sds_info.rank = geo_sds_info.rank;
        out_geo_sds_info.data_type = geo_sds_info.data_type;
        out_geo_sds_info.data_size = geo_sds_info.data_size;
        for (i=0; i<geo_sds_info.rank; i++)
          out_geo_sds_info.dim_size[i] = geo_sds_info.dim_size[i];
        if (open_sds((char *)NULL, &out_geo_sds_info, 'W') != -1)
        {
          if (fire_st == 0)
	    write_attr_fval(out_geo_sds_info.H5OUT, out_geo_sds_info.sds_id, DFNT_FLOAT32, 1, geo_sds_info.fill_fval);
          nrows = geo_sds_info.dim_size[0];
          ncols = geo_sds_info.dim_size[1];
          if ((data = (void *)calloc(ncols, geo_sds_info.data_size)) == NULL)
            fprintf(stderr, "Cannot allocate memory for data in copy_geolocation\n");
          if (data != NULL)
          {
            edge[0] = 1;
            edge[1] = ncols;
            for (irow=0; irow<nrows; irow++)
            {
              start[0] = irow;
	      if (read_sel_sds_data(&geo_sds_info, start, edge, data) == FAIL)
              {
                fprintf(stderr, "Cannot read dataline from SDS %s\n", geo_sds_info.name);
                break;
              }
	      if (set_sel_sds_data(&out_geo_sds_info, start, edge, data) == FAIL)
              {
                fprintf(stderr, "Cannot write data line for SDS %s\n", geo_sds_info.name);
                break;
              }
            } /* for (irow=0; . . . . . . */
          }
          if (data != NULL) free(data);
        }
        if (out_geo_sds_info.sds_id != -1) close_sds(&out_geo_sds_info, out_geo_sds_info.H5OUT);
      }
      if (geo_sds_info.sds_id != -1) close_sds(&geo_sds_info, geo_sds_info.H5IN);
    }
  } /* for (isds=0; . . . */
}


