/****************************************************************************
!C

!File: copy_sds.c

!Description:
  Contains routines copy SDS(s) from input file to an output HDF file (appends to existing HDF file)

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original July 2001 Version 
  Modified June 2011 by Xiaoping Zhang, Sigma Space Corp.
  	added the feature of scaling only the user specified SDS.
	added "-out_flag" of the command argument for output integer data type if input data type is float.  
  Modified December 06, 2011 by Xiaoping Zhang, Sigma Space Corp.
  	changed fill value from 32720 to 65535 for output integer data type if input data type is float.
  Modified January 12, 2012 Version 1.0 by Xiaoping Zhang, Sigma Space Corp.
  	assigned the values of -999.6 and -999.7 to 65532 and the rest < -999.0 values to 65535 
	for output integer data type if input data type is float.  		
  Modified February, 2012 Version 2.0 by Xiaoping Zhang, Sigma Space Corp.
  	added the fillValue_in and fillValue_out options.
  Modified June, 2012 Version 2.1 by Xiaoping Zhang, Sigma Space Corp.
  	fixed a bug with the fill values(< -999.) being scaled. 
  Modified July 16, 2012 Version 2.2 by Xiaoping Zhang, Sigma Space Corp.
        changed output integer data type from unsigned integer to signed integer and
  	changed fill value from 65535 to 32720 for output integer data type if input data type is float.	
  Modified August 31, 2012 Version 2.3 by Xiaoping Zhang, Sigma Space Corp.
        compiled copy_sds.c code with the other libraries code from the LP-DAAC for delivering this tool. 
  Modified September 20 2012 Version 2.4 by Xiaoping Zhang, Sigma Space Corp.
  	Added an option of user specified data type as output data type, removed the options  
	of input/output fill values and made input/output fill values hard coded.
	The input data type will be output data type if this option is not used. 	
  Modified June 28, 2013 Version 2.5 by Xiaoping Zhang, Sigma Space Corp.
        based on v2.3 August 31 2012, removed the round up of 0.5 to solve the scaling problem for SDS of Radiance_1 (20 km) 
	in the product of NPPDNNEW.NPP_DDNBN_L2GD and fixed the problem of fill value containing the 
	non SDS values(-995, -996, -997 ...) which is wrongly scaled.	
  Modified July 25, 2014 Version 3.1 by Xiaoping Zhang, Sigma Space Corp.	  	
	set fill values of 65535, 65534 ... 65528 to output if -outflag option selected and 
	input data tye is floating.  
  Modified February, 2016 Version 3.2 by Xiaoping Zhang, SSAI	
        added write_all_sds_attr() and set correct value for nrows when rank = 1 in copy_an_sds().
  Modified January, 2018 Version 3.5.1 by Maosheng Zhao, SSAI
	1)modify it so that HDF5 can be read
	2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
	3)to speedup for dealing with internal compressed chunk data in HDF5, 
	  copy all data and write out in 20 times with multiply lines not line by line

  Modified in Feb., 2020, Version 3.6 by Maosheng Zhao, SSAI.
        The output HDF4 now has chunked internal compression.
  
!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga 		      Sigma Space Corp.
    LDOPE                             
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <math.h>

#include "mfhdf.h"
#include "hdf5.h"
#include "qa_tool.h"
#include "sds_rw.h"
#include "str_op.h"
#include "meta.h"
#include "alloc_mem.h"
#include "main_util.h"     

#define HELP \
"NAME \n" \
"       copy_sds - copy SDS(s) from input file to an output HDF file (appends to existing HDF file) \n" \
" \n" \
"SYNOPSIS \n" \
"       copy_sds [-help] \n" \
"       copy_sds [-sds=<old name>,<new name>] [-scale=val] -ncp [-meta] -out_flag \n" \
"       	[-fillValue_in=<fval1s>] [-fillValue_out==<fval2s>] -of=<output filename> filename\n" \
" \n" \
"DESCRIPTION \n" \
"       Copy one or more SDS from an input HDF file to an output HDF file. SDSs are appended\n"  \
"       to an existing HDF file. User may choose not to copy remaining SDSs else they will be copied. \n\n" \
"\n" \
" \n" \
"OPTIONS \n" \
"       -help                       Display this help message \n" \
"       -sds=<old name,new name>    Input and Output SDS names \n" \
"                                   To specify additional repeat the option as many times as \n" \
"                                   the number of SDSs to be renamed \n" \
"       -scale=<val>                Scale the SDS values(2D input) \n" \
"       -of=<filename>              Output filename \n" \
"       -ncp                        Do not copy remaining SDS(s) \n" \
"       -meta                       Copy all metadata from input to output \n" \
"       -out_flag                   Flag for out data type of integer(2D input) \n" \
"       -fillValue_in=<fval1>       input fill values separated by comma(2D input) \n" \
"       -fillValue_out=<fval2>      output fill values separated by comma(2D input) \n" \
" \n" \
"Examples:  \n" \
"	copy_sds -sds=M3_SurfRefl,M3_SurfRefl -sds=M4_SurfRefl,M4_SurfRefl -sds=M5_SurfRefl,M5_SurfRefl -scale=10000 -meta \n" \
"		 -of=copy_sds.hdf NPP_SRFLMIP_L2C.A2007193.0025.P1_00610.2011158132504.hdf \n" \
"	copy_sds -sds=IceFraction_Reproj,IceFraction_Reproj -scale=10000 -meta -out_flag -fillValue_in=-900,-800 -fillValue_out=-32720,-32720  \n" \
"		 -of=IDPS.NPP_VICIP_L2C.A2011340.20km.copy.hdf /home/ldope/NPP_browse/GLOBAL/IDPS/NPP_VICIP_L2C/IDPS.NPP_VICIP_L2C.A2011340.20km.hdf \n" \
"\n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:\tcopy_sds [-sds=<old name>,<new name>] [-scale=val] -ncp [-meta] -of=<output filename> filename\n" \
" \n" \
"OPTIONS \n" \
"       -help                       Display this help message \n" \
"       -sds=<old name,new name>    Input and Output SDS names \n" \
"                                   To specify additional repeat the option as many times as \n" \
"                                   the number of SDSs to be renamed \n" \
"       -scale=<val>                Scale the SDS values \n" \
"       -of=<filename>              Output filename \n" \
"       -meta                       Copy all metadata from input to output \n" \
"       -ncp                        Do not copy remaining SDS(s) \n" \
"       -out_flag                   Flag for out data type of integer \n" \
"       -fillValue_in               input fill values separated by comma \n" \
"       -fillValue_out              output fill values separated by comma \n" \
" \n" \
" \n"                                                      

int parse_cmd_rename_sds(int argc, char **argv, char *out_fname, char *in_fname, int *nsds, 
	char **in_sds_names, char **out_sds_names, int *m_opt, long *sc, int *nsds1, int *out_flag,
	char **in_fill_values, char **out_fill_values, int *fill_in_cnt, int *fill_out_cnt, sds_t *in_sds_info, sds_t *out_sds_info);
void copy_an_sds(sds_t *in_sds_info, sds_t *out_sds_info, long sc, int *nsds1, int out_flag,
		char **in_fill_values, char **out_fill_values, int fill_in_cnt, int fill_out_cnt, int jobs_num);

int main(int argc, char **argv)
{
    int nsds, isds, m_opt, nsds1, out_flag;
    long sc;
    char in_fname[MAX_PATH_LENGTH];
    char out_fname[MAX_PATH_LENGTH];
    sds_t in_sds_info, out_sds_info;
    char **in_sds_names, **out_sds_names;
    char **in_fill_values, **out_fill_values;
    int fill_in_cnt=0, fill_out_cnt=0;
    int jobs_num = 20;    /* MSZ 1/25/2018, speedup */

    /* MSZ 1/19/2018 silent version mismatch warning standard print out */
    setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

    /* MSZ 1/19/2018 Disable error reporting */
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  
    if ((argc== 1) || ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0))))
      fprintf(stderr, "%s\n", HELP);
    else
    {
        if ((in_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
            fprintf(stderr, "Cannot allocate memory for in_sds_names in rename_sds main()\n");
        if ((out_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
            fprintf(stderr, "Cannot allocate memory for out_sds_names in rename_sds main()\n");
        if ((in_fill_values = (char **)Calloc2D(MAX_NUM_SDS, 10, sizeof(char))) == NULL)             /* MSZ 1/26/1018, change 10 to MAX_NUM_SDS */
            fprintf(stderr, "Cannot allocate memory for in_fill_values in rename_sds main()\n");   
        if ((out_fill_values = (char **)Calloc2D(MAX_NUM_SDS, 10, sizeof(char))) == NULL)            /* MSZ 1/26/1018, change 10 to MAX_NUM_SDS */
            fprintf(stderr, "Cannot allocate memory for out_fill_values in rename_sds main()\n");            
        if ((in_sds_names != NULL) && (out_sds_names != NULL) && (in_fill_values != NULL) && (out_fill_values != NULL))
        {
            if (parse_cmd_rename_sds(argc, argv, out_fname, in_fname, &nsds, in_sds_names, out_sds_names,
                &m_opt, &sc, &nsds1, &out_flag, in_fill_values, out_fill_values, &fill_in_cnt, &fill_out_cnt, &in_sds_info, &out_sds_info) == -1)
              fprintf(stderr, "%s\n", USAGE);
            else
            {
                fprintf(stdout, "Processing file %s\n", in_fname);
                for (isds=0; isds<fill_in_cnt; isds++)
                  fprintf(stderr, "in_fill_values[%d] = %s  out_fill_values[%d] = %s\n", isds, in_fill_values[isds],
                    isds, out_fill_values[isds]);

		/* MSZ 1/19/2018 */
		/* Note that in_sds_info.H5IN and out_sds_info.H5OUT has been assigned the initial values in parse_cmd_rename_sds() */
		in_sds_info.H5OUT = 0;
		out_sds_info.H5IN = out_sds_info.H5OUT = 0;

		/* sd_id and sds_id must be initialized as -1 */
		in_sds_info.sd_id = out_sds_info.sd_id = -1;
		in_sds_info.sds_id = out_sds_info.sds_id = -1;

		/* MSZ 1/23/2018 use open_sd() to open in and out HDFs */
		if(open_sd(in_fname, &in_sds_info, 'R') == FAIL)
		{
		  fprintf(stderr, "cannot open %s\n", in_fname);
		  exit(0);
		}

		if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
		{
		  fprintf(stderr, "cannot create %s\n", out_fname);
		  exit(0);
		}

                if ((in_sds_info.sd_id != FAIL) && (out_sds_info.sd_id != FAIL))
                {
                    for (isds=0; isds<nsds; isds++)
                    {	    
                      strcpy(in_sds_info.name, in_sds_names[isds]);
                      strcpy(out_sds_info.name, out_sds_names[isds]);
		      if( !(strcasestr(in_sds_info.name, "metadata") != 0) )   /* MSZ 1/25/2018, not for metadata, DADASET of HDF5 */
		      {
                         fprintf(stdout, "\tPlease wait..., copying and renaming SDS \"%s\" as \"%s\"\n", in_sds_names[isds], out_sds_names[isds]);
                         copy_an_sds(&in_sds_info, &out_sds_info, sc, &nsds1, out_flag, 
                              in_fill_values, out_fill_values, fill_in_cnt, fill_out_cnt, jobs_num);
		      }
                      
                      nsds1 = nsds1 - 1;
                    }
                }
                if (m_opt == 1)
                  copy_metadata(in_sds_info.H5IN, in_sds_info.sd_id, out_sds_info.H5OUT, out_sds_info.sd_id);

		/* MSZ 1/24/2018 use close_sd() to close in and out HDFs */
		close_sd(&in_sds_info, in_sds_info.H5IN);
		close_sd(&out_sds_info, out_sds_info.H5OUT);
            }
        }
        if(in_sds_names) Free2D((void **)in_sds_names);
        if(out_sds_names) Free2D((void **)out_sds_names);
        if(in_fill_values) Free2D((void **)in_fill_values);
        if(out_fill_values) Free2D((void **)out_fill_values);
    }
    return 0;
}

int parse_cmd_rename_sds(int argc, char **argv, char *out_fname, char *in_fname, int *nsds, 
	char **in_sds_names, char **out_sds_names, int *m_opt, long *sc, int *nsds1, int *out_flag,
	char **in_fill_values, char **out_fill_values, int *fill_in_cnt, int *fill_out_cnt, sds_t *in_sds_info, sds_t *out_sds_info)
{
    int inc_sds;
    int st, i, p1;
    int ncp, sds_cnt, len;
    int isds, jsds, ksds, isds1;
    char **sds_names, tmp_str[MAX_STR_LEN];
    char* ptr1 = NULL;
    
    *nsds = 0;
    *sc = -99999;
    *nsds1 = 0;  
    *out_flag = 0;
    isds = ncp = isds1 = 0;
    in_fname[0] = out_fname[0] = '\0';
    for (i=1, st=1; i<argc; i++)
    {
        if ((is_arg_id(argv[i], "-of=") == 0) || (is_arg_id(argv[i], "-o=") == 0))
        {
	    get_arg_val(argv[i], out_fname);

	    if (((((ptr1 = strstr(out_fname,".h5")) != NULL)) ||
               (((ptr1 = strstr(out_fname,".he5")) != NULL)) ||
               (((ptr1 = strstr(out_fname,".hdf5")) != NULL)) ||
               (((ptr1 = strstr(out_fname,".nc")) != NULL))))
	      out_sds_info->H5OUT = 1;
	    else
	      out_sds_info->H5OUT = 0;
        }
        else if (is_arg_id(argv[i], "-sds=") == 0)
        {
            get_arg_val(argv[i], tmp_str);
            if ((p1 = charpos(tmp_str, ',', 0)) == -1)
                fprintf(stderr, "Ignoring invalid option assignment %s\n", argv[i]);
            else
            {
                len = (int)strlen(tmp_str);
                    strmid(tmp_str, 0, p1, in_sds_names[isds]);
                p1++;
                    strmid(tmp_str, p1, len-p1, out_sds_names[isds]);
                isds++;
            }
            *nsds1 = isds;
        }
        else if (is_arg_id(argv[i], "-scale=") == 0) {
            get_arg_val(argv[i], tmp_str);
            *sc = (long)atol(tmp_str);   
        }
        else if (strcmp(argv[i], "-ncp") == 0) ncp = 1; 
        else if (strcmp(argv[i], "-meta") == 0) *m_opt = 1; 
        else if (strcmp(argv[i], "-out_flag") == 0) *out_flag = 1; 
        else if (is_arg_id(argv[i], "-fillValue_in=") == 0)
            get_arg_val_arr(argv[i], in_fill_values, fill_in_cnt);
        else if (is_arg_id(argv[i], "-fillValue_out=") == 0)
            get_arg_val_arr(argv[i], out_fill_values, fill_out_cnt);          
        else if (argv[i][0] == '-')
            fprintf(stderr, "Unknown option %s\n", argv[i]);
        else
            strcpy(in_fname, argv[i]);
    }
    if (out_fname[0] == '\0') {
      st = -1;
      fprintf(stderr, "Missing output filename . .\n");
    }
    if (in_fname[0] == '\0') {
      st = -1;
      fprintf(stderr, "Missing input filename . .\n");
    }
    if (st == 1)
    {
        if (ncp == 0)
        {
            if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
              fprintf(stderr, "Cannot allocate memory for sds_names in parse_cmd_rename_sds\n");
            else
            {
                sds_cnt = get_sds_names(in_fname, sds_names);	
                for (jsds=0; jsds<sds_cnt; jsds++)
                {
                  inc_sds = 1;
                  for (ksds=0; ksds<isds; ksds++)
                    if (strcasecmp(sds_names[jsds], in_sds_names[ksds]) == 0)
                    {		      
                      inc_sds = 0;
                      break;
                    } 
                  if (inc_sds == 1)
                  {
                    strcpy(in_sds_names[isds], sds_names[jsds]);
                    strcpy(out_sds_names[isds], sds_names[jsds]);
                    isds++;
                  }
                }
              if(sds_names) Free2D((void **)sds_names);
            }
        }
        *nsds = isds;
    }
    return st;
}                                                   
	    
void copy_an_sds(sds_t *in_sds_info, sds_t *out_sds_info, long sc, int *nsds1,int out_flag, 
		char **in_fill_values, char **out_fill_values,  int fill_in_cnt, int fill_out_cnt, int jobs_num)
{
    void *sds_data;
    int i;
    int32 start[4], edge[4];
    int irow, icol;
    int nrows, ncols;
    int GROWS, last_rows;
    float tmpf;
    int tmpi, fcnt, nattrs=22;
    int nlines_a_copy = 1; /*MSZ 1/25/2018, speedup. Initialize to be 1 */
    int comp_chunk_rows;   /* MSZ 1/23/2020 internal compression */
    
    /*printf("out_flag=%d\n",out_flag);
    printf("nsds1=%d\n",*nsds1);
    printf("fill_in_cnt=%d\n",fill_in_cnt);*/

    out_sds_info->sds_id = in_sds_info->sds_id = -1;
    if (get_sds_info((char *)NULL, in_sds_info) != -1)
    {
        out_sds_info->rank = in_sds_info->rank;
	out_sds_info->interleave = in_sds_info->interleave;
	out_sds_info->row_id = in_sds_info->row_id;

        for (i=0; i<in_sds_info->rank; i++)
          out_sds_info->dim_size[i] = in_sds_info->dim_size[i]; 

        if (out_flag == 1 && *nsds1 > 0) 
        {
    // printf(" in_sds_info->data_type====%d\n",     in_sds_info->data_type);
          switch(in_sds_info->data_type) {
        case 5: out_sds_info->data_type = DFNT_UINT16; 
                out_sds_info->data_size = DFKNTsize(out_sds_info->data_type);
                break;
        case 6: out_sds_info->data_type = DFNT_UINT16; 
                out_sds_info->data_size = DFKNTsize(out_sds_info->data_type);
                break;
        case 20: out_sds_info->data_type = in_sds_info->data_type; 
                 out_sds_info->data_size = in_sds_info->data_size;
                 break;
        case 21: out_sds_info->data_type = in_sds_info->data_type; 
                 out_sds_info->data_size = in_sds_info->data_size;	         
		 break;
        case 22: out_sds_info->data_type = in_sds_info->data_type; 
                 out_sds_info->data_size = in_sds_info->data_size;	
                 break;
        case 23: out_sds_info->data_type = in_sds_info->data_type; 
                 out_sds_info->data_size = in_sds_info->data_size;	
                 break;
        case 24: out_sds_info->data_type = in_sds_info->data_type; 
                 out_sds_info->data_size = in_sds_info->data_size;	
                 break;
        case 25: out_sds_info->data_type = in_sds_info->data_type; 
                 out_sds_info->data_size = in_sds_info->data_size;	
                 break;
          }    
        }
        else 
        {
          out_sds_info->data_type = in_sds_info->data_type;
          out_sds_info->data_size = in_sds_info->data_size;
        }

	if (in_sds_info->rank != 1)
	{
	   /* MSZ 1/25/2018 speedup */
	   nlines_a_copy = (int)floor(in_sds_info->dim_size[in_sds_info->row_id] / jobs_num + 0.5);
	   if(in_sds_info->dim_size[in_sds_info->row_id] < nlines_a_copy) nlines_a_copy = in_sds_info->dim_size[in_sds_info->row_id];

	   /* MSZ 01/25/2020. It would be better to set 10 in case that 
	    * the HDF file will be used for mask. set 10 will not slow down
	    * the mask process which is done line by line */
	   /*comp_chunk_rows = nlines_a_copy;*/
	   comp_chunk_rows = 30;
	   if(comp_chunk_rows > in_sds_info->dim_size[in_sds_info->row_id]) 
		comp_chunk_rows = in_sds_info->dim_size[in_sds_info->row_id];
	}
		
        /*if (open_sds((char *)NULL, out_sds_info, 'W') != -1)*/
	if (open_sds_compression((char *)NULL, out_sds_info, comp_chunk_rows, 'W') != -1)
        {
            for (i=0; i<in_sds_info->rank; i++)
            {
                start[i] = 0;
                edge[i] = in_sds_info->dim_size[i];
            }

            if (in_sds_info->rank != 1) 
	    {
	      /* MSZ 1/25/2018 speedup */
              edge[in_sds_info->row_id] = 1 * nlines_a_copy;
	    }
            nrows = (in_sds_info->rank == 1) ? 1 : in_sds_info->dim_size[in_sds_info->row_id];
      //      nrows = in_sds_info->dim_size[in_sds_info->row_id];
            ncols = in_sds_info->dim_size[in_sds_info->row_id + 1];
            
            write_all_sds_attr(in_sds_info->H5IN, in_sds_info->sds_id, out_sds_info->H5OUT, out_sds_info->sds_id, nattrs);

	    /* MSZ 1/25/2018 */
	    GROWS = (int)ceil((double)nrows / (double)nlines_a_copy);
	    last_rows = nrows % nlines_a_copy;

	    /* MSZ, 1/25/2018, to speedup, copy lines by lines not line by line */
            for (irow=0; irow<GROWS; irow++)
            {
		start[in_sds_info->row_id] = irow * nlines_a_copy;

		/* MSZ, 1/25/2018, be cautious about the last group rows because it may be different from nlines_a_copy */
		if(last_rows > 0 && irow == GROWS - 1)
		   edge[in_sds_info->row_id] = last_rows;

                if ((sds_data = get_sel_sds_data(in_sds_info, start, edge)) == NULL)
                  fprintf(stderr, "Error reading lines from %d to %d from SDS %s\n", irow * nlines_a_copy, irow * nlines_a_copy + nlines_a_copy, in_sds_info->name);
                else
                {
                    if (sc != -99999 && *nsds1 > 0) {
                      if (fill_in_cnt == 0 && fill_out_cnt == 0) {

		    #if 1   /* MSZ 1/24/2018, for diagnosis purpose */
                    if (out_flag == 0) {
                      for (icol=0; icol<ncols * edge[in_sds_info->row_id]; icol++) {
                        if (((float32 *)sds_data)[icol] >= -999.) {
                          switch(in_sds_info->data_type) {
                            case 5: ((float32 *)sds_data)[icol] = ((float32 *)sds_data)[icol] * sc; break;
                            case 6: ((float64 *)sds_data)[icol] = ((float64 *)sds_data)[icol] * sc; break;
                            case 20: ((int8 *)sds_data)[icol] = ((int8 *)sds_data)[icol] * sc; break;
                            case 21: ((uint8 *)sds_data)[icol] = ((uint8 *)sds_data)[icol] * sc; break;
                            case 22: ((int16 *)sds_data)[icol] = ((int16 *)sds_data)[icol] * sc; break;
                            case 23: ((uint16 *)sds_data)[icol] = ((uint16 *)sds_data)[icol] * sc; break;
                            case 24: ((int32 *)sds_data)[icol] = ((int32 *)sds_data)[icol] * sc; break;
                            case 25: ((uint32 *)sds_data)[icol] = ((uint32 *)sds_data)[icol] * sc; break;
                          }
                        }
                      }
                    }
                    else {
                      for (icol=0; icol<ncols * edge[in_sds_info->row_id]; icol++) {
                        switch(in_sds_info->data_type) {
                          case 5: 
			    if (((float32 *)sds_data)[icol] <= -999.2 && ((float32 *)sds_data)[icol] >= -999.9005) {
                              switch((int)((((float32 *)sds_data)[icol] - 0.01)*10)) {
                                case -9992: ((uint16 *)sds_data)[icol] = 65528; break;
                                case -9993: ((uint16 *)sds_data)[icol] = 65529; break;
                                case -9994: ((uint16 *)sds_data)[icol] = 65530; break;
                                case -9995: ((uint16 *)sds_data)[icol] = 65531; break;
                                case -9996: ((uint16 *)sds_data)[icol] = 65532; break;
                                case -9997: ((uint16 *)sds_data)[icol] = 65533; break;
                                case -9998: ((uint16 *)sds_data)[icol] = 65534; break;
                                case -9999: ((uint16 *)sds_data)[icol] = 65535; break;
                              }
                            }  
                            else ((uint16 *)sds_data)[icol] = (uint16)(((float32 *)sds_data)[icol] *sc + 0.5); 
                            break;
                          case 6:
			    if (((float64 *)sds_data)[icol] <= -999.2 && ((float64 *)sds_data)[icol] >= -999.9005) {
                              switch((int)((((float64 *)sds_data)[icol] - 0.01)*10))
                              {
                                case -9992: ((uint16 *)sds_data)[icol] = 65528; break;
                                case -9993: ((uint16 *)sds_data)[icol] = 65529; break;
                                case -9994: ((uint16 *)sds_data)[icol] = 65530; break;
                                case -9995: ((uint16 *)sds_data)[icol] = 65531; break;
                                case -9996: ((uint16 *)sds_data)[icol] = 65532; break;
                                case -9997: ((uint16 *)sds_data)[icol] = 65533; break;
                                case -9998: ((uint16 *)sds_data)[icol] = 65534; break;
                                case -9999: ((uint16 *)sds_data)[icol] = 65535; break;
                              }
                            }
                            else ((uint16 *)sds_data)[icol] = (uint16)(((float64 *)sds_data)[icol] *sc + 0.5); 
                                break;		    
                          case 20:
			    ((int8 *)sds_data)[icol] = ((int8 *)sds_data)[icol] * sc; break;
                          case 21:
			    ((uint8 *)sds_data)[icol] = ((uint8 *)sds_data)[icol] * sc; break;
                          case 22:
			    ((int16 *)sds_data)[icol] = ((int16 *)sds_data)[icol] * sc; break;
                          case 23:
			    ((uint16 *)sds_data)[icol] = ((uint16 *)sds_data)[icol] * sc; break;
                          case 24:
			    ((int32 *)sds_data)[icol] = ((int32 *)sds_data)[icol] * sc; break;
                          case 25:
			    ((uint32 *)sds_data)[icol] = ((uint32 *)sds_data)[icol] * sc; break;
                        }
                      }	    
                    }
                      }
                      else {
                    if (out_flag == 0) {
			
                      for (icol=0; icol<ncols * edge[in_sds_info->row_id]; icol++) {
                        fcnt = 0;		    	    
                        switch(in_sds_info->data_type) {
                          case 5: for (i=0; i<fill_in_cnt; i++) {
                            tmpf = (float32)atof(in_fill_values[i]);			
                                if (((float32 *)sds_data)[icol] < tmpf+0.005 && ((float32 *)sds_data)[icol] > tmpf-0.005) {
                                  ((float32 *)sds_data)[icol] = (float32)atof(out_fill_values[i]);
                              fcnt =  fcnt + 1;
                            }  
                              }	 
                              if (fcnt == 0) ((float32 *)sds_data)[icol] = (float32)(((float32 *)sds_data)[icol] * sc ); 
                                  break;
                          case 6: for (i=0; i<fill_in_cnt; i++) {
                            tmpf = (float64)atof(in_fill_values[i]);
                                if (((float64 *)sds_data)[icol] < tmpf+0.005 && ((float64 *)sds_data)[icol] > tmpf-0.005) {
                                  ((float64 *)sds_data)[icol] = (float64)atof(out_fill_values[i]);
                              fcnt =  fcnt + 1;
                            }  
                              }	 
                              if (fcnt == 0) ((float64 *)sds_data)[icol] = (float64)(((float64 *)sds_data)[icol] * sc ); 
                                  break; 
                          case 20: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (int8)atoi(in_fill_values[i]);
                             if (((int8 *)sds_data)[icol] == tmpi) {		    		 
                                   ((int8 *)sds_data)[icol] = (int8)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1; 
                             } 
                               }	 
                               if (fcnt == 0) ((int8 *)sds_data)[icol] = (int8)(((int8 *)sds_data)[icol] * sc ); 
                                   break; 
                          case 21: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (uint8)atoi(in_fill_values[i]);
                             if (((uint8 *)sds_data)[icol] == tmpi) {		    		 
                                   ((uint8 *)sds_data)[icol] = (uint8)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1; 
                             } 
                               }	 
                               if (fcnt == 0) ((uint8 *)sds_data)[icol] = (uint8)(((uint8 *)sds_data)[icol] * sc ); 
                                   break; 
                          case 22: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (int16)atoi(in_fill_values[i]);
                             if (((int16 *)sds_data)[icol] == tmpi) {		    		 
                                   ((int16 *)sds_data)[icol] = (int16)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1; 
                             } 
                               }	 
                               if (fcnt == 0) ((int16 *)sds_data)[icol] = (int16)(((int16 *)sds_data)[icol] * sc ); 
                                   break; 
                          case 23: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (uint16)atoi(in_fill_values[i]);
                             if (((uint16 *)sds_data)[icol] == tmpi) {		    		 
                                   ((uint16 *)sds_data)[icol] = (uint16)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1;
                             } 
                               }	 
                               if (fcnt == 0) ((uint16 *)sds_data)[icol] = (uint16)(((uint16 *)sds_data)[icol] * sc ); 
                                   break; 
                          case 24: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (int32)atoi(in_fill_values[i]);
                             if (((int32 *)sds_data)[icol] == tmpi) {		    		 
                                   ((int32 *)sds_data)[icol] = (int32)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1; 
                             } 
                               }	 
                               if (fcnt == 0) ((int32 *)sds_data)[icol] = (int32)(((int32 *)sds_data)[icol] * sc ); 
                                   break; 
                          case 25: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (uint32)atoi(in_fill_values[i]);
                             if (((uint32 *)sds_data)[icol] == tmpi) {		    		 
                                   ((int32 *)sds_data)[icol] = (uint32)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1;
                             } 
                               }	 
                               if (fcnt == 0) ((uint32 *)sds_data)[icol] = (uint32)(((uint32 *)sds_data)[icol] * sc ); 
                                   break; 
                        }	  
                      }
                    }
                    else {
                      for (icol=0; icol<ncols * edge[in_sds_info->row_id]; icol++) {	
                        fcnt = 0;	     
                        switch(in_sds_info->data_type) {
                          case 5: for (i=0; i<fill_in_cnt; i++) {	      
                            tmpf = (float32)atof(in_fill_values[i]);
                                if (((float32 *)sds_data)[icol] < tmpf+0.005 && ((float32 *)sds_data)[icol] > tmpf-0.005) {
                                  ((uint16 *)sds_data)[icol] = (uint16)atof(out_fill_values[i]);
                              fcnt = fcnt + 1;
                            }  
                              }	 	      
                              if (fcnt == 0) ((uint16 *)sds_data)[icol] = (uint16)(((float32 *)sds_data)[icol] * sc ); 
                                  break;
                          case 6: for (i=0; i<fill_in_cnt; i++) {
                            tmpf = (float64)atof(in_fill_values[i]);
                                if (((float64 *)sds_data)[icol] < tmpf+0.005 && ((float64 *)sds_data)[icol] > tmpf-0.005) {
                                  ((uint16 *)sds_data)[icol] = (uint16)atof(out_fill_values[i]);
                              fcnt =  fcnt + 1;
                            }   
                              }	 
                              if (fcnt == 0) ((uint16 *)sds_data)[icol] = (uint16)(((float64 *)sds_data)[icol] * sc ); 
                                  break; 
                          case 20: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (int8)atoi(in_fill_values[i]);
                             if (((int8 *)sds_data)[icol] == tmpi) {		    		 
                                   ((int8 *)sds_data)[icol] = (int8)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1;
                             } 
                               }	 
                               if (fcnt == 0) ((int8 *)sds_data)[icol] = (int8)(((int8 *)sds_data)[icol] * sc ); 
                                   break; 
                          case 21: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (uint8)atoi(in_fill_values[i]);
                             if (((uint8 *)sds_data)[icol] == tmpi) {		    		 
                                   ((uint8 *)sds_data)[icol] = (uint8)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1; 
                             } 
                               }	 
                               if (fcnt == 0) ((uint8 *)sds_data)[icol] = (uint8)(((uint8 *)sds_data)[icol] * sc ); 
                                   break; 
                          case 22: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (int16)atoi(in_fill_values[i]);
                             if (((int16 *)sds_data)[icol] == tmpi) {		    		 
                                   ((int16 *)sds_data)[icol] = (int16)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1; 
                             } 
                               }	 
                               if (fcnt == 0) ((int16 *)sds_data)[icol] = (int16)(((int16 *)sds_data)[icol] * sc ); 
                                   break; 
                          case 23: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (uint16)atoi(in_fill_values[i]);
                             if (((uint16 *)sds_data)[icol] == tmpi) {		    		 
                                   ((uint16 *)sds_data)[icol] = (uint16)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1;
                             } 
                               }	 
                               if (fcnt == 0) ((uint16 *)sds_data)[icol] = (uint16)(((uint16 *)sds_data)[icol] * sc ); 
                                   break; 
                          case 24: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (int32)atoi(in_fill_values[i]);
                             if (((int32 *)sds_data)[icol] == tmpi) {		    		 
                                   ((int32 *)sds_data)[icol] = (int32)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1; 
                             } 
                               }	 
                               if (fcnt == 0) ((int32 *)sds_data)[icol] = (int32)(((int32 *)sds_data)[icol] * sc ); 
                                   break; 
                          case 25: for (i=0; i<fill_in_cnt; i++) {
                             tmpi = (uint32)atoi(in_fill_values[i]);
                             if (((uint32 *)sds_data)[icol] == tmpi) {		    		 
                                   ((int32 *)sds_data)[icol] = (uint32)atoi(out_fill_values[i]);
                               fcnt =  fcnt + 1; 
                             } 
                               }	 
                               if (fcnt == 0) ((uint32 *)sds_data)[icol] = (uint32)(((uint32 *)sds_data)[icol] * sc ); 
                                   break; 
                        }
                      }	      
                    }	      
			#endif
                      }
                    } 
                    else if (*nsds1 > 0) 
                    {
                      if (fill_in_cnt != 0 && fill_out_cnt !=0 && out_flag != 0) {
                    for (icol=0; icol<ncols * edge[in_sds_info->row_id]; icol++) {
                      fcnt = 0;	    
                      switch(in_sds_info->data_type) {
                        case 5: for (i=0; i<fill_in_cnt; i++) {
                              tmpf = (float32)atof(in_fill_values[i]);			
                                  if (((float32 *)sds_data)[icol] < tmpf+0.005 && ((float32 *)sds_data)[icol] > tmpf-0.005) {
                                ((uint16 *)sds_data)[icol] = (uint16)atof(out_fill_values[i]);
                            fcnt =  fcnt + 1;
                              }  
                            }	 
                            if (fcnt == 0) ((uint16 *)sds_data)[icol] = (uint16)(((float32 *)sds_data)[icol]); 
                                break;
                        case 6: for (i=0; i<fill_in_cnt; i++) {
                              tmpf = (float64)atof(in_fill_values[i]);
                                  if (((float64 *)sds_data)[icol] < tmpf+0.005 && ((float64 *)sds_data)[icol] > tmpf-0.005) {
                                ((uint16 *)sds_data)[icol] = (uint16)atof(out_fill_values[i]);
                            fcnt =  fcnt + 1;
                              }  
                            }	 
                            if (fcnt == 0) ((uint16 *)sds_data)[icol] = (uint16)(((float64 *)sds_data)[icol]); 
                                break; 
                        case 20: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (int8)atoi(in_fill_values[i]);
                               if (((int8 *)sds_data)[icol] == tmpi) {		    		 
                                 ((int8 *)sds_data)[icol] = (int8)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1; 
                               } 
                             }	 
                             if (fcnt == 0) ((int8 *)sds_data)[icol] = ((int8 *)sds_data)[icol]; 
                                 break; 
                        case 21: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (uint8)atoi(in_fill_values[i]);
                               if (((uint8 *)sds_data)[icol] == tmpi) {		    		 
                                 ((uint8 *)sds_data)[icol] = (uint8)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1; 
                               } 
                             }	 
                             if (fcnt == 0) ((uint8 *)sds_data)[icol] = ((uint8 *)sds_data)[icol]; 
                                 break; 
                        case 22: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (int16)atoi(in_fill_values[i]);
                               if (((int16 *)sds_data)[icol] == tmpi) {		    		 
                                 ((int16 *)sds_data)[icol] = (int16)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1; 
                               } 
                             }	 
                             if (fcnt == 0) ((int16 *)sds_data)[icol] = ((int16 *)sds_data)[icol]; 
                                 break; 
                        case 23: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (uint16)atoi(in_fill_values[i]);
                               if (((uint16 *)sds_data)[icol] == tmpi) {		    		 
                                 ((uint16 *)sds_data)[icol] = (uint16)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1;
                               } 
                             }	 
                             if (fcnt == 0) ((uint16 *)sds_data)[icol] = ((uint16 *)sds_data)[icol]; 
                                 break; 
                        case 24: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (int32)atoi(in_fill_values[i]);
                               if (((int32 *)sds_data)[icol] == tmpi) {		    		 
                                 ((int32 *)sds_data)[icol] = (int32)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1; 
                               } 
                             }	 
                             if (fcnt == 0) ((int32 *)sds_data)[icol] = ((int32 *)sds_data)[icol]; 
                                 break; 
                        case 25: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (uint32)atoi(in_fill_values[i]);
                               if (((uint32 *)sds_data)[icol] == tmpi) {		    		 
                                 ((int32 *)sds_data)[icol] = (uint32)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1;
                               } 
                             }	 
                             if (fcnt == 0) ((uint32 *)sds_data)[icol] = ((uint32 *)sds_data)[icol]; 
                                 break; 
                      }	  
                    }
                      }	
                      else if (fill_in_cnt != 0 && fill_out_cnt !=0 && out_flag == 0) {
                    for (icol=0; icol<ncols * edge[in_sds_info->row_id]; icol++) {
                      fcnt = 0;	    
                      switch(in_sds_info->data_type) {
                        case 5: for (i=0; i<fill_in_cnt; i++) {
                              tmpf = (float32)atof(in_fill_values[i]);			
                                  if (((float32 *)sds_data)[icol] < tmpf+0.005 && ((float32 *)sds_data)[icol] > tmpf-0.005) {
                                ((float32 *)sds_data)[icol] = (float32)atof(out_fill_values[i]);
                            fcnt =  fcnt + 1;
                              }  
                            }	 
                            if (fcnt == 0) ((float32 *)sds_data)[icol] = ((float32 *)sds_data)[icol]; 
                                break;
                        case 6: for (i=0; i<fill_in_cnt; i++) {
                              tmpf = (float64)atof(in_fill_values[i]);
                                  if (((float64 *)sds_data)[icol] < tmpf+0.005 && ((float64 *)sds_data)[icol] > tmpf-0.005) {
                                ((float64 *)sds_data)[icol] = (float64)atof(out_fill_values[i]);
                            fcnt =  fcnt + 1;
                              }  
                            }	 
                            if (fcnt == 0) ((float64 *)sds_data)[icol] = ((float64 *)sds_data)[icol]; 
                                break; 
                        case 20: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (int8)atoi(in_fill_values[i]);
                               if (((int8 *)sds_data)[icol] == tmpi) {		    		 
                                 ((int8 *)sds_data)[icol] = (int8)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1; 
                               } 
                             }	 
                             if (fcnt == 0) ((int8 *)sds_data)[icol] = ((int8 *)sds_data)[icol]; 
                                 break; 
                        case 21: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (uint8)atoi(in_fill_values[i]);
                               if (((uint8 *)sds_data)[icol] == tmpi) {		    		 
                                 ((uint8 *)sds_data)[icol] = (uint8)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1; 
                               } 
                             }	 
                             if (fcnt == 0) ((uint8 *)sds_data)[icol] = ((uint8 *)sds_data)[icol]; 
                                 break; 
                        case 22: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (int16)atoi(in_fill_values[i]);
                               if (((int16 *)sds_data)[icol] == tmpi) {		    		 
                                 ((int16 *)sds_data)[icol] = (int16)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1; 
                               } 
                             }	 
                             if (fcnt == 0) ((int16 *)sds_data)[icol] = ((int16 *)sds_data)[icol]; 
                                 break; 
                        case 23: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (uint16)atoi(in_fill_values[i]);
                               if (((uint16 *)sds_data)[icol] == tmpi) {		    		 
                                 ((uint16 *)sds_data)[icol] = (uint16)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1;
                               } 
                             }	 
                             if (fcnt == 0) ((uint16 *)sds_data)[icol] = ((uint16 *)sds_data)[icol]; 
                                 break; 
                        case 24: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (int32)atoi(in_fill_values[i]);
                               if (((int32 *)sds_data)[icol] == tmpi) {		    		 
                                 ((int32 *)sds_data)[icol] = (int32)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1; 
                               } 
                             }	 
                             if (fcnt == 0) ((int32 *)sds_data)[icol] = ((int32 *)sds_data)[icol]; 
                                 break; 
                        case 25: for (i=0; i<fill_in_cnt; i++) {
                               tmpi = (uint32)atoi(in_fill_values[i]);
                               if (((uint32 *)sds_data)[icol] == tmpi) {		    		 
                                 ((int32 *)sds_data)[icol] = (uint32)atoi(out_fill_values[i]);
                             fcnt =  fcnt + 1;
                               } 
                             }	 
                             if (fcnt == 0) ((uint32 *)sds_data)[icol] = ((uint32 *)sds_data)[icol]; 
                                 break; 
                      }	  
                    }	      
                      }	
                      else if (fill_in_cnt == 0 && fill_out_cnt ==0 && out_flag != 0) {
                    for (icol=0; icol<ncols * edge[in_sds_info->row_id]; icol++) {
                      switch(in_sds_info->data_type) {
                        case 5: if (((float32 *)sds_data)[icol] <= -999.2 && ((float32 *)sds_data)[icol] >= -999.9005) {
                              switch((int)((((float32 *)sds_data)[icol] - 0.01)*10))
                              {
                            case -9992: ((uint16 *)sds_data)[icol] = 65528; break;
                            case -9993: ((uint16 *)sds_data)[icol] = 65529; break;
                            case -9994: ((uint16 *)sds_data)[icol] = 65530; break;
                            case -9995: ((uint16 *)sds_data)[icol] = 65531; break;
                            case -9996: ((uint16 *)sds_data)[icol] = 65532; break;
                            case -9997: ((uint16 *)sds_data)[icol] = 65533; break;
                            case -9998: ((uint16 *)sds_data)[icol] = 65534; break;
                            case -9999: ((uint16 *)sds_data)[icol] = 65535; break;
                              }
                            }  
                            else ((uint16 *)sds_data)[icol] = (uint16)(((float32 *)sds_data)[icol] + 0.5); 
                            break;
                        case 6: if (((float64 *)sds_data)[icol] <= -999.2 && ((float64 *)sds_data)[icol] >= -999.9005) {
                              switch((int)((((float64 *)sds_data)[icol] - 0.01)*10))
                              {
                            case -9992: ((uint16 *)sds_data)[icol] = 65528; break;
                            case -9993: ((uint16 *)sds_data)[icol] = 65529; break;
                            case -9994: ((uint16 *)sds_data)[icol] = 65530; break;
                            case -9995: ((uint16 *)sds_data)[icol] = 65531; break;
                            case -9996: ((uint16 *)sds_data)[icol] = 65532; break;
                            case -9997: ((uint16 *)sds_data)[icol] = 65533; break;
                            case -9998: ((uint16 *)sds_data)[icol] = 65534; break;
                            case -9999: ((uint16 *)sds_data)[icol] = 65535; break;
                              }
                            }
                            else ((uint16 *)sds_data)[icol] = (uint16)(((float64 *)sds_data)[icol] + 0.5); 
                                  break;
                        case 20: ((int8 *)sds_data)[icol] = ((int8 *)sds_data)[icol]; break;
                        case 21: ((uint8 *)sds_data)[icol] = ((uint8 *)sds_data)[icol]; break;
                        case 22: ((int16 *)sds_data)[icol] = ((int16 *)sds_data)[icol]; break;
                        case 23: ((uint16 *)sds_data)[icol] = ((uint16 *)sds_data)[icol]; break;
                        case 24: ((int32 *)sds_data)[icol] = ((int32 *)sds_data)[icol]; break;
                        case 25: ((uint32 *)sds_data)[icol] = ((uint32 *)sds_data)[icol]; break;
                      }
                    }	 
                      }	      
                    }
		    /*if(in_sds_info->H5IN)
		      fprintf(stderr, "copying row=%d to row=%d in %d rows\n", start[in_sds_info->row_id], start[in_sds_info->row_id] + edge[in_sds_info->row_id] - 1, nrows);*/
                    /*if (set_sel_sds_data(out_sds_info, start, edge, sds_data) == FAIL)*/
		    if (set_sel_sds_data_compression(out_sds_info, start, edge, comp_chunk_rows, sds_data) == FAIL)
                      fprintf(stderr, "Error writing lines %d to %d to SDS %s\n", start[in_sds_info->row_id], start[in_sds_info->row_id] + edge[in_sds_info->row_id] - 1, out_sds_info->name);

		    if(sds_data) free(sds_data);
                }
            }
	    /* MSZ 1/23/2018, this was a bug. Note that get_sel_sds_data() calloced inside and should be freed within line by line loop above not here */
	    /* free(sds_data); */
        }    
	close_sds(in_sds_info, in_sds_info->H5IN);
	close_sds(out_sds_info, out_sds_info->H5OUT);
    }
    else  /* MSZ 4/21/2018, added to warn user for the wrong SDS name specified by user */
    {
      fprintf(stderr, "sds %s doesn\'t exist, please check the sds name using read_sds_attributes, exit...\n", in_sds_info->name);
      exit(0);
    }
}


