/****************************************************************************
!C

!File: cp_proj_param.c

!Description:
  Copy geolocation related metadata to a L3 HDF file. 

!Input Parameters: (none)

!Output Parameters: (none)

!Revision History:

  Version 1.0    May, 2003

  Modified in February, 2018, Version 3.6 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
        3)update all HDFEOS API to a higher-level APIs to handle 
          both HDFEOS2 and HDFEOS5 in eos_sds_rw.c

!Team-unique Header:

  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

    Yi (Sophyi) Zhang 
    LDOPE                             Science Systems and Applications Inc.
    sophyi@ltpmail.gsfc.nasa.gov      NASA/GSFC Code 922
    phone: 301-614-5497               Greenbelt, MD 20771    

    David Roy 
    LDOPE                             University of Maryland, 
    				      Department of Geography
    droy@kratmos.gsfc.nasa.gov        NASA/GSFC Code 922 (B32)
    phone: 301-614-5571               Greenbelt, MD 20771   

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/dir.h>         
#include <math.h>

#include "eos_sds_rw.h"
#include "qa_tool.h"
#include "sds_rw.h"
#include "str_op.h"
#include "meta.h"
#include "alloc_mem.h"
#include "main_util.h"         

#include "tile_bound_info.h"

#define MAX_GRID_NAME_LEN 40
#define MAX_NUM_TILES 25
#define TILE_ID_STR_LEN 10

#define HELP \
"NAME \n" \
"	cp_proj_param - copy projection parameter metadata to an HDF file\n" \
"\n" \
"SYNOPSIS \n" \
"	cp_proj_param -help \n" \
"\n" \
"	cp_proj_param -of=<output filename> [-ref=<reference>] \n" \
"		[-tile=<tile_id(s)>] [-proj=<projection type>] filename \n" \
"\n" \
"DESCRIPTION \n" \
"       Certain HDF files, created for example using the LDOPE QA tools or other \n" \
"       software, may not contain the HDF-EOS metadata that identify the projection \n" \
"       type and the geographic extent of the data. This tool creates an output \n" \
"       HDF-EOS file by copying all the SDS(s) from the input HDF file and adding the \n" \
"       projection parameters to the output file. The output file may then be reprojected \n" \
"       using the MODIS reprojection tool available from EDC-DAAC. \n" \
"\n" \
"       This tool is applicable only to geolocated L2G/L3/L4 MODLAND products or mosaics \n" \
"       of these products. There are two ways to use this tool: \n" \
"\n" \
"       1) The projection metadata may be copied from a reference HDF-EOS L2G/L3/L4 MODLAND \n" \
"       product. The reference product should have the same geographical extent as the input \n" \
"       HDF file. This method only works for single files (not mosaics). \n" \
"\n" \
"       2) The projection metadata may be copied by specifying the MODLAND product tile id \n" \
"       and the projection type. Multiple tile ids may be specified if the input file is a \n" \
"       mosaic of two or more files. \n" \
" \n" \
"       The tool command arguments can be specified in any order. \n" \
" \n" \
"       Note: To make this tool function transparently with the MODIS Reprojection Tool the \n" \
"       bounding rectangular coordinates are copied to the output file archive metadata without \n" \
"       using the SDP toolkit. \n" \
" \n" \
"       Note: Use of the -tile option may cause a geolocation imprecision of no more than \n" \
"       approximately 50m (which falls within the MODIS geolocation specification). \n" \
" \n" \
"       For more information on MODIS Land tile grid see \n" \
"       http://landweb.nascom.nasa.gov/cgi-bin/QA_WWW/newPage.cgi?fileName=prod_format \n" \
"\n" \
"OPTIONS \n" \
"       -help                   Display this help message. \n" \
"       -of=<filename>          Output filename. \n" \
"       -ref=<reference file>   Reference L2G/L3/L4 MODLAND product filename. \n" \
"                               If this is specified then do not use the -tile and -proj options.\n" \
"       -tile=<tile id(s)>      One or more tile id(s) separated by commas. The tile id is of \n" \
"                               the form hxxvyy where xx and yy are the two digit horizontal \n" \
"                               and vertical MODLAND product tile numbers. \n" \
"       -proj=<projection type> One of the following projection types: ISIN (Integerized \n" \
"                               Sinusoidal) or SIN (Sinusoidal) \n" \
"       filename                Input filename \n" \
"\n" \
"EXAMPLES \n" \
"       cp_proj_param -ref=MOD09A1.A2001145.h20v10.003.2001214125825.hdf \n" \
"       -of=zoo_coords.hdf zoo.hdf \n" \
"\n" \
"       cp_proj_param -tile=h20v10 <96>proj=ISIN -of= zoo_coords.hdf zoo.hdf \n" \
"\n" \
"       cp_proj_param -tile=h20v10,h21v10,h20v11,h21v11 proj=ISIN \n" \
"       -of=New.Four_Mosaiced_tiles.hdf Four_Mosaiced_tiles.hdf \n" \
"\n" \
"AUTHOR \n" \
"	Code: S. Devadiga and Yi Zhang \n" \
"	Documentation: S. Devadiga and D. Roy \n" \
"\n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"


#define USAGE \
"usage: \n" \
"	cp_proj_param - copy projection parameter metadata to an HDF file\n" \
"\n" \
"SYNOPSIS \n" \
"	cp_proj_param -help \n" \
"\n" \
"	cp_proj_param -of=<output filename> [-ref=<reference>] \n" \
"		[-tile=<tile_id(s)>] [-proj=<projection type>] filename \n" \
"\n" \
"OPTIONS \n" \
"       -help                   Display this help message. \n" \
"       -of=<filename>          Output filename. \n" \
"       -ref=<reference file>   Reference L2G/L3/L4 MODLAND product filename. \n" \
"       -tile=<tile id(s)>      One or more tile id(s) separated by commas. The tile id is of \n" \
"                               the form hxxvyy where xx and yy are the two digit horizontal \n" \
"                               and vertical MODLAND product tile numbers. \n" \
"       -proj=<projection type> One of the following projection types: ISIN (Integerized Sinusoidal) \n" \
"                               or SIN (Sinusoidal) \n" \
"       filename                Input filename \n" \
"\n" 

#define PSUEDO_ARCH_META \
"GROUP                  = ARCHIVEDMETADATA\n" \
"  GROUPTYPE            = MASTERGROUP\n" \
"\n" \
"  GROUP                  = BOUNDINGRECTANGLE\n" \
"\n" \
"    OBJECT                 = NORTHBOUNDINGCOORDINATE\n" \
"      NUM_VAL              = 1\n" \
"      VALUE                = xxxxxxxxx\n" \
"    END_OBJECT             = NORTHBOUNDINGCOORDINATE\n" \
"\n" \
"    OBJECT                 = SOUTHBOUNDINGCOORDINATE\n" \
"      NUM_VAL              = 1\n" \
"      VALUE                = xxxxxxxxx\n" \
"    END_OBJECT             = SOUTHBOUNDINGCOORDINATE\n" \
"\n" \
"    OBJECT                 = EASTBOUNDINGCOORDINATE\n" \
"      NUM_VAL              = 1\n" \
"      VALUE                = xxxxxxxxx\n" \
"    END_OBJECT             = EASTBOUNDINGCOORDINATE\n" \
"\n" \
"    OBJECT                 = WESTBOUNDINGCOORDINATE\n" \
"      NUM_VAL              = 1\n" \
"      VALUE                = xxxxxxxxx\n" \
"    END_OBJECT             = WESTBOUNDINGCOORDINATE\n" \
"\n" \
"  END_GROUP              = BOUNDINGRECTANGLE\n" \
"\n" \
"END_GROUP              = ARCHIVEDMETADATA\n" \
"\n" \
"END\n"

int parse_cmd_cp_proj_param(int argc, char **argv, char *in_fname, char *out_fname, char *ref_fname, 
			char **tile_id, int *n_tile, char* proj_name, int *H5OUT);
int get_hdfeos(char *file_name, grid_mapinfo *h, char *grid_name);
void add_hdfeos_meta(char *file_name, char *grid_name, grid_mapinfo *h, sds_t *sds_info, int nsds, int H5OUT);
void copy_sds_data(char *out_fname, char *grid_name, sds_t *sds_info, int nsds, int H5OUT);
int read_param_table(char **tile_id, int n_tile, grid_mapinfo *hdfeos, char* proj_name);
int get_bound_coord(char **tile_id, int n_tile, char *proj_name, double *bound_coord);
void copy_eos_meta(int32 in_sd_id, int32 out_sd_id);
void copy_psuedo_arch_meta(int32 out_sd_id, int H5OUT, double *bound_coord);
void insert_meta_value(char *meta_str, char *meta_name, char *meta_val);

int main(int argc, char **argv)
{
  int n_tile;
  int st, isds, nsds, H5OUT;
  grid_mapinfo hdfeos;
  sds_t *sds_info;
  double bound_coord[4];
  int32 ref_sd_id;
  sds_t out_sds_info;
  char **sds_names, **tile_ids;
  char ref_fname[MAX_PATH_LENGTH], in_fname[MAX_PATH_LENGTH];
  char out_fname[MAX_PATH_LENGTH], grid_name[MAX_GRID_NAME_LEN];
  char proj_name[10];  /* valid value are ISIN or SIN */
  int HDF_ver = 0;

  /* missing input file */
  if (argc == 1)
    {
      st = 1;
      fprintf(stderr, "Missing input file \n");
      fprintf(stderr, "%s\n", USAGE);
      exit(EXIT_FAILURE);
    }

  /* display help */

  if ((argc == 2) && ( strcmp(argv[1], "-help") == 0))
    {
      fprintf(stderr, "%s\n", HELP);
      exit(EXIT_SUCCESS);
    }
  else 
  {
    /* MSZ 2/10/2018 silent version mismatch warning standard print out */
    setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

    /* MSZ 2/10/2018 Disable error reporting */
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

    if ((tile_ids = (char **)Calloc2D(MAX_NUM_TILES, TILE_ID_STR_LEN, sizeof(char))) == NULL)
      fprintf(stderr, "Cannot allocate memory for tile_ids in main()\n");
    else
    {
      if (parse_cmd_cp_proj_param(argc, argv, in_fname, out_fname, ref_fname, tile_ids, &n_tile, 
			      proj_name, &H5OUT) == -1)
        fprintf(stderr, "%s\n", USAGE);
      else 
      {
        if (n_tile > 0) {
          st = read_param_table(tile_ids, n_tile, &hdfeos, proj_name);
	  if (st != -1)
	    st = get_bound_coord(tile_ids, n_tile, proj_name, bound_coord);
        }
        else
        {
          if ((st = get_grid_name(ref_fname, grid_name)) == 1)
	      st = get_hdfeos(ref_fname, &hdfeos, grid_name);
	}
        if (st == -1)
          fprintf(stderr, "Cannot read projection parameter\n");
        else
        {
          if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
	    fprintf(stderr, "Cannot allocate memory for sds_names in main() \n");
          else
          {
	    nsds = get_sds_names(in_fname, sds_names);
	    if ((sds_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)
	      fprintf(stderr, "Cannot allocate memory for sds_info in main()\n");
	    else
	    {
	      for (isds=0; isds<nsds; isds++)
	      {
	        sds_info[isds].sd_id = (isds == 0) ? -1 : sds_info[0].sd_id;
		sds_info[isds].H5IN = (isds == 0) ? 0 : sds_info[0].H5IN;
	        sds_info[isds].sds_id = -1;
	        strcpy(sds_info[isds].name, sds_names[isds]);	
		get_sds_info(in_fname, &sds_info[isds]);
	      }

	      /* add projection meta data */
              add_hdfeos_meta(out_fname, "SAMPLE_L3_GRID", &hdfeos, sds_info, nsds, H5OUT);

	      /* copy the sds from input file to output file */
	      copy_sds_data(out_fname, "SAMPLE_L3_GRID", sds_info, nsds, H5OUT);
	      for (isds=0; isds<nsds; isds++)
		close_sds(&sds_info[isds], sds_info[isds].H5IN);
	      close_sd(&sds_info[0], sds_info[0].H5IN);
	      free(sds_info);

	      /* open the output HDF file and copy the metadata/bounding coordinate to output */
	      out_sds_info.H5OUT = H5OUT;
	      out_sds_info.sd_id = out_sds_info.sds_id = -1;
	      if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
      		fprintf(stderr, "Cannot open the output HDF file for copying the metadata %s\n", out_fname);
	      else 
	      {
		if (n_tile > 0)
	          copy_psuedo_arch_meta(out_sds_info.sd_id, out_sds_info.H5OUT, bound_coord);
		else {
		  if(HDF_version_check(ref_fname) == 4 && out_sds_info.H5OUT == 0)  /* MSZ 2/10/2018. only when both ref_fname and out_fname are HDF4 */
	          {
	            if ((ref_sd_id = SDstart(ref_fname, DFACC_READ)) == FAIL)
      		      fprintf(stderr, "Cannot open the reference file for copying the metadata %s\n", ref_fname);
		    copy_eos_meta(ref_sd_id, out_sds_info.sd_id);
		    SDend(ref_sd_id);
		  }
		}
		close_sd(&out_sds_info, out_sds_info.H5OUT);	
	      }
	    }
	    Free2D((void **)sds_names);
          }
        }
      }  
      Free2D((void **)tile_ids);
    }
  }
  return 0;
}

int parse_cmd_cp_proj_param(int argc, char **argv, char *in_fname, char *out_fname, 
			char *ref_fname, char **tile_ids, int *n_tile, 
			char *proj_name, int *H5OUT)
{
    int i, cnt, st, iarg;
    int hn, vn;
    char tmp[10];
    double lon_min, lon_max;
    double lat_min, lat_max;
    int not_valid_tile_pair;   /* flag is set to 1(true) when both -tile and -proj 
                  options are specified and provided valid input, 
                  or both the -tile and -proj option is not specified. */ 
      
    int tileflag = 0;
    int projflag = 0;  
    char* ptr1 = NULL;
    
    *n_tile = 0;
    ref_fname[0] = '\0';
    in_fname[0] = out_fname[0] = '\0';
    *H5OUT = 0; /* Default is to assume that both I/O will be in hdf4 */
    for (st=1, iarg=1; iarg<argc; iarg++)
    {
        if (is_arg_id(argv[iarg], "-of") == 0)
        {
            get_arg_val(argv[iarg], out_fname);
            if (((((ptr1 = strstr(out_fname,".h5")) != NULL)) ||
              (((ptr1 = strstr(out_fname,".he5")) != NULL)) ||
              (((ptr1 = strstr(out_fname,".hdf5")) != NULL)) ||
              (((ptr1 = strstr(out_fname,".nc")) != NULL))))
               *H5OUT = 1;
            else
               *H5OUT = 0;
        }
        else if (is_arg_id(argv[iarg], "-ref") == 0)
          get_arg_val(argv[iarg], ref_fname);
        else if (is_arg_id(argv[iarg], "-tile") == 0)
        {
            get_arg_val_arr(argv[iarg], tile_ids, n_tile);
            if ( n_tile == 0 )
            {
                fprintf(stderr, "invalid tile_id entered. \n");
                exit(EXIT_FAILURE);
            }
            else
            {
                tileflag = 1;
            }
        }
        else if (is_arg_id(argv[iarg], "-proj") == 0)
        {
              get_arg_val(argv[iarg], proj_name);
              if ( proj_name == NULL )
                {
                  fprintf(stderr, "invalid projection type entered. Valid projection \n");
                  fprintf(stderr, "type is ISIN or SIN. \n");
                  exit(EXIT_FAILURE);
                }
              else 
                {
                  projflag = 1;
                }
        }
            
        else if (argv[iarg][0] == '-')
          fprintf(stderr, "Ignoring unknown option %s\n", argv[iarg]);
        else
            strcpy(in_fname, argv[iarg]);
    } 
  
    if ((tileflag == 1 && projflag == 0) || (tileflag == 0 && projflag == 1))
    {
      not_valid_tile_pair = 1;
    }
    else 
    {
      not_valid_tile_pair = 0;
    }
  
  
    if (in_fname[0] == '\0')
    {
      fprintf(stderr, "Missing input filename . . \n");
      st = -1;
    }
    
    if (out_fname[0] == '\0')
    {
      fprintf(stderr, "Missing output filename . . \n");
      st = -1;
    }
    
    if ((*n_tile == 0) && (strlen(ref_fname) <= 0))
    {
      fprintf(stderr, "Missing reference L3 filename or tile id \n");
      st = -1;
    }
    
    if (not_valid_tile_pair == 1)
    { 
      fprintf(stderr, "No valid -tile and -proj option entered. \n");
      st = -1;
    }
   
    if (st == 1)
    { 
      if ((*n_tile > 0) && (strlen(ref_fname) > 0))
    {
      fprintf(stderr, "Using reference L3 file %s only\n", ref_fname);
      *n_tile = 0;
    }
    }
    lon_min = lon_max = lat_min = lat_max = 0.0;
    if ((st == 1) && ( *n_tile > 0))
    {
        for (i=0, cnt=0; i< *n_tile; i++)
        {
            strmid(tile_ids[i], 1, 2, tmp);
            hn = (int)atoi(tmp);
            strmid(tile_ids[i], 4, 2, tmp);
            vn = (int)atoi(tmp);
            if (read_bound_table(hn, vn, proj_name, &lon_min, &lon_max, &lat_min, &lat_max) == 1)
            {
              if ((lon_min == -999.00) && (lon_max == -999.00) && (lat_max == -99.00) && (lat_min == -99.00)) {
                cnt++;
            fprintf(stderr, "Invalid tile id: %s\n", tile_ids[i]);
              }
            }
            else
          fprintf(stderr, "Invalid tile id: %s\n", tile_ids[i]);
        }
        if (cnt > 0) st = 0;
    }
    return st;
}

int get_hdfeos(char *file_name, grid_mapinfo *h, char *grid_name)
{
  int st, i;
  hid_t file_id, grid_id;
  int he5_fmt;
  int HDF_ver = 0;
  
  h->ns = (long)(-1);
  h->nl = (long)(-1);
  h->ulCornerGrid[0] = (float64)(-999.0);
  h->ulCornerGrid[1] = (float64)(-999.0);
  h->lrCornerGrid[0] = (float64)(-999.0);
  h->lrCornerGrid[1] = (float64)(-999.0);
  h->projCode = (int)(-1);
  h->sphereCode = (int)(-1);
  for (i = 0; i < 13; i++)
    h->projParameters[i] = (float64)(-999.0);
  h->zoneCode = (int)(-1);
  h->originCode = (int)(-1);

  HDF_ver = HDF_version_check(file_name);

  if( !(HDF_ver ==4 || HDF_ver == 5) )
  {
    printf("%s is neither HDF4 nor HDF5, exit...\n", file_name);
    exit(0);
  }

  if(HDF_ver == 4) he5_fmt = 0;
  else if(HDF_ver == 5) he5_fmt = 1;

  st = 1;
  if ((file_id = EOS_GDopen(file_name, 'R', he5_fmt)) == FAIL) 
  {
    fprintf(stderr, "Input file %s open failure (EOS_GDopen) in get_hdfeos\n", file_name);
    st = -1;
  }
  else
  {
    if ((grid_id = EOS_GDattach(file_id, grid_name, he5_fmt)) == FAIL)
    {
      fprintf(stderr, "Can't attach to %s (EOS_GDattach) in get_hdfeos\n", grid_name);
      st = -1;
    }
    else
    {
      if (EOS_GDgridinfo(grid_id, &h->ns, &h->nl, h->ulCornerGrid, h->lrCornerGrid, he5_fmt) == FAIL)
      {
        fprintf(stderr, "Can't get grid info (EOS_GDgridinfo) in get_hdfeos\n");
        st = -1;
      }
      
      if (EOS_GDprojinfo(grid_id, &h->projCode, &h->zoneCode, &h->sphereCode, h->projParameters, he5_fmt) == FAIL)
      {
        fprintf(stderr, "Can't get projection info (EOS_GDprojinfo) in get_hdfeos\n");
        st = -1;
      }
  
      if (EOS_GDorigininfo(grid_id, &h->originCode, he5_fmt) == FAIL)
      {
        fprintf(stderr,  "Can't get origin defintion (EOS_GDorigininfo) in get_hdfeos\n");
        st = -1;
      }
  
      if (EOS_GDdetach(grid_id, he5_fmt) == FAIL)
        fprintf(stderr, "Can't detach from grid (EOS_GDdetach) in get_hdfeos\n");
    }
  
    if (EOS_GDclose(file_id, he5_fmt) == FAIL)
      fprintf(stderr, "Can't close file (EOS_GDclose) in get_hdfeos\n");
  }
  return st;
}
      
void add_hdfeos_meta(char *file_name, char *grid_name, grid_mapinfo *h, sds_t *sds_info, int nsds, int H5OUT)
{
  int isds;
  hid_t gfid, gid;
  char dim_str[20];
  hid_t EOS_data_type;
  int he5_fmt = H5OUT;

  if ((gfid = EOS_GDopen(file_name, 'C', he5_fmt)) == FAIL)
    fprintf(stderr, "Can't open HDF file %s (EOS_GDopen) in add_hdfeos_meta\n", file_name);
  else
  {
    h->ns = sds_info[0].dim_size[0];
    h->nl = sds_info[0].dim_size[1];
    if ((gid = EOS_GDcreate(gfid, grid_name, h->ns, h->nl, h->ulCornerGrid, h->lrCornerGrid, he5_fmt)) == FAIL)
      fprintf(stderr, "Can't open output file (EOS_GDcreate) in add_hdfeos_meta\n");
    else
    {
      if (EOS_GDdefproj(gid, h->projCode, h->zoneCode, h->sphereCode, h->projParameters, he5_fmt) == FAIL)
        fprintf(stderr, "Cannot define projection parameter (EOS_GDdefproj) in add_hdfeos_meta\n");
    }

    for (isds=0; isds<nsds; isds++)  
    {
      if (sds_info[isds].rank == 2) 
        strcpy(dim_str, "YDim,XDim");
      else if (sds_info[isds].rank == 3) 
        strcpy(dim_str, "ZDim,YDim,XDim");

      if(he5_fmt == 0)
	EOS_data_type = (hid_t)sds_info[isds].data_type;
      else if(he5_fmt == 1)
      {
	if(convert_datatype_HDF4toHDFEOS5(sds_info[isds].data_type, &EOS_data_type) == FAIL)
	  fprintf(stderr, "Error in calling convert_datatype_HDF4toHDFEOS5()!!!\n");
      }

      if (EOS_GDdeffield(gid, sds_info[isds].name, dim_str, EOS_data_type, he5_fmt) == FAIL)
        fprintf(stderr, "Error defining the SDS field %s\n", sds_info[isds].name);
    }
  
    if (EOS_GDdetach(gid, he5_fmt) == FAIL)
      fprintf(stderr, "Error detaching from grid %s (EOS_GDdetach) in add_hdfeos_meta\n", grid_name);

    if (EOS_GDclose(gfid, he5_fmt) == FAIL)
      fprintf(stderr, "Cannot close file %s (EOS_GDclose) in add_hdfeos_meta() \n", file_name);
  }
}

void copy_sds_data(char *out_fname, char *grid_name, sds_t *sds_info, int nsds, int H5OUT)
{
  hid_t gfid, grid_id;
  int i, isds, irow;
  int nrows, ncols;
  int32 start[MAX_LDOPE_DIMS] = {0}, edge[MAX_LDOPE_DIMS] = {0};
  void *data = NULL;
  int he5_fmt = H5OUT;

  if ((gfid = EOS_GDopen(out_fname, 'W', he5_fmt)) == FAIL)
    fprintf(stderr, "Can't open HDF file %s (EOS_GDopen) in copy_sds_data\n", out_fname);
  else
  {
    if ((grid_id = EOS_GDattach(gfid, grid_name, he5_fmt)) == FAIL)
      fprintf(stderr, "Cannot attach to grid %s\n", grid_name);
    else
    {
      for (isds=0; isds<nsds; isds++)
      {
        nrows = sds_info[isds].dim_size[0];
        ncols = sds_info[isds].dim_size[1];
        if ((data = (void *)calloc(ncols, sds_info[isds].data_size)) == NULL)
          fprintf(stderr, "Cannot allocate memory for data in copy_sds_data\n");
        else
        {
	  start[0] = start[1] = 0;
	  edge[0] = 1; edge[1] = ncols;
	  for (irow=0; irow<nrows; irow++)
	  {
	    start[0] = irow;
	   
	    if (read_sel_sds_data(&sds_info[isds], start, edge, data) == FAIL) 
	      fprintf(stderr, "Cannot read line %d from SDS %s\n", irow, sds_info[isds].name); 
	    else
	    {
              hssize_t start_tmp[MAX_LDOPE_DIMS];
	      hsize_t edge_tmp[MAX_LDOPE_DIMS];
              for(i = 0; i < MAX_LDOPE_DIMS; i++)
              {
                start_tmp[i] = (hssize_t)start[i];
                edge_tmp[i] = (hsize_t)edge[i];
              }

	      if (EOS_GDwritefield(grid_id, sds_info[isds].name, start_tmp, edge_tmp, data, he5_fmt) == FAIL)
	        fprintf(stderr, "Cannot write line %d to SDS %s\n", irow, sds_info[isds].name); 
	    }
          }
	  free(data);
        }
      }                                            
      EOS_GDdetach(grid_id, he5_fmt);
    }
    EOS_GDclose(gfid, he5_fmt);
  }
}

int read_param_table(char **tile_id, int n_tile, grid_mapinfo *h, char* proj_name)
{
  FILE *fp;
  int d1, d2, d3, d4;
  int nt_fnd, c, i, id;
  double c1, c2, c3, c4, f[13];
  char *anc_path, in_tile_id[10];
  char param_fname[MAX_PATH_LENGTH];
  int flag[700];

  for (i=0; i<700; i++)
    {
      flag[i]= 0;
    }

  anc_path = getenv("ANCPATH");
  if (anc_path == NULL)
    {
      fprintf(stderr, "Please set up your environment variable for ANCPATH first.\n");
      exit(EXIT_FAILURE);
    }

  if (strcmp(proj_name, "ISIN") == 0)
    {
      sprintf(param_fname, "%s/%s", anc_path, ISIN_PARAM_FNAME);
    }
  else if (strcmp(proj_name, "SIN") == 0)
    {
      sprintf(param_fname, "%s/%s", anc_path, SIN_PARAM_FNAME);
    }
  else
    {
      fprintf(stderr, "No valid projection parameter entered. \n");
      exit(EXIT_FAILURE);
    }

  if ((fp = fopen(param_fname, "r")) == NULL)
    fprintf(stderr, "Cannot open projection parameter file %s\n", param_fname);
  else
  {
    nt_fnd = 0;
    c = fgetc(fp);
    while ((nt_fnd < n_tile) && (c != EOF))
    {
      ungetc(c, fp);
      fscanf(fp, "%s", in_tile_id);
      fscanf(fp, "%lf %lf %lf %lf", &c1, &c2, &c3, &c4);
      fscanf(fp, "%d %d %d %d", &d1, &d2, &d3, &d4);
      for (i=0; i<13; i++)
        fscanf(fp, "%lf", &f[i]);  
      for (id=0; id<n_tile; id++)
      {
        if (strcmp(in_tile_id, tile_id[id]) == 0)
        {
	  flag[id] = 1;
	  if (nt_fnd == 0)
	  {
            h->ulCornerGrid[0] = c1; h->ulCornerGrid[1] = c2;
            h->lrCornerGrid[0] = c3; h->lrCornerGrid[1] = c4;
	    h->projCode = d1; h->sphereCode = d2;
	    h->zoneCode = d3; h->originCode = d4;
            for (i=0; i<13; i++)
              h->projParameters[i] = f[i];
	  }
	  else
	  {
	    if (c1 < h->ulCornerGrid[0]) h->ulCornerGrid[0] = c1;
            if (c2 > h->ulCornerGrid[1]) h->ulCornerGrid[1] = c2;
            if (c3 > h->lrCornerGrid[0]) h->lrCornerGrid[0] = c3;
	    if (c4 < h->lrCornerGrid[1]) h->lrCornerGrid[1] = c4;
	  }
	  ++nt_fnd;
        }
      }
      c = fgetc(fp);
    }    
   
    for (i = 0; i< n_tile; i++)
      {
	if (flag[i] == 0)
	  {
	    fprintf(stderr, "Input tileid %s is wrong.\n", tile_id[i]);
	    exit(EXIT_FAILURE);
	  }
      }
    
    fclose(fp);
  }
  if (nt_fnd != n_tile) return -1;
  else return 1;
}

void copy_eos_meta(int32 in_sd_id, int32 out_sd_id)
/* MSZ 2/10/2018. Note that there is no need to update it because there is no archive and core
 * meta data for HDFEOS. Calling the function has been controled in the calling env. */
/*********************************************************************************
!C
!Description:
        Copy only archive and core metadata from input to output

!Input Parameters:
  in_sd_id      - input SD interface identifer.
  out_sd_id     - output SD interface identifer.

!Output Parameters: (none)

!Revision History:
  Sadashiva Devadiga		April 2004  /SSAI

!Team-unique Header: (see file prolog)

!References and Credits: (see file prolog)

!Design Notes: (none)

!END
************************************************************************************/
{
  long attr_len;
  int32 nsds;                     /* number of data sets in the file */
  int32 nattr;                    /* number of global attributes in the file */
  int32 attr_cnt;                 /* count of number of attribute */
  int32 attr_index;               /* attribute index */
  int32 attr_type, attr_size;     /* attribute type and size */
  char attr_name[40];             /* attribute name */
  char *attr_buf;                 /* data buffer for attribute */
  int ret;                        /* return value for testing purpose */
  
  fprintf(stdout, "Copying metadata from reference input file to output\n");

  ret = SDfileinfo(in_sd_id, &nsds, &nattr);
  /* test to see if num_datasets and num_global_attr can be retrieved from in_sd_id */
  if (ret == -1)
    {
      fprintf(stdout, "cannot read information from input file \n");
      exit(EXIT_FAILURE);
    }
  else
    {
      /* loop through each global attributes */
      for (attr_index=0; attr_index<nattr; attr_index++)
        {
          /* test to see if the file or dataset do indeed contain attributes */
          if (SDattrinfo(in_sd_id, attr_index, attr_name, &attr_type, &attr_cnt) == FAIL)
            fprintf(stdout, "Cannot read information for attribute %d\n", attr_index);
          else
	  {
	    if ((strstr(attr_name, "CoreMetadata") != NULL) || (strstr(attr_name, "ArchiveMetadata") != NULL))
            {
              attr_size = DFKNTsize(attr_type);
              attr_len = attr_size*attr_cnt;
              if ((attr_type == DFNT_CHAR) || (attr_type == DFNT_CHAR8))
		attr_len++;
              if ((attr_buf = (void *)calloc(attr_size, attr_len)) == NULL)
                fprintf(stdout, "Cannot allocate memory for attr_buf in copy_metadata\n");
              else
                {
                  if (SDreadattr(in_sd_id, attr_index, (VOIDP)attr_buf) == FAIL)
                    fprintf(stdout, "Cannot read %s from input file\n", attr_name);
                  else
                    {
                      /* copy the metadata from attr_buf to out_sd_id. */
                      if (SDsetattr(out_sd_id, attr_name, attr_type, attr_cnt,
                                    (VOIDP)attr_buf) == FAIL)
                        {
                          fprintf(stdout, "Cannot write metadata %s to output file\n", attr_name);
                        }
                    }
                  free(attr_buf);
                }
            }
	  }
        } /* for (attr_index=0; . . .      */
    }
}

void copy_psuedo_arch_meta(int32 out_sd_id, int H5OUT, double *bound_coord)
/*********************************************************************************
!C
!Description:
        Copy only archive and core metadata from input to output

!Input Parameters:
  out_sd_id     - output SD interface identifer.

!Output Parameters: (none)

!Revision History:
  Sadashiva Devadiga            April 2004  /SSAI

!Team-unique Header: (see file prolog)

!References and Credits: (see file prolog)

!Design Notes: (none)

!END
**************************************************************************/
{
  int i;
  void *attr_buf;
  int32 attr_type, attr_cnt, attr_len, attr_size;
  char meta_name[40], meta_val[40], meta_str[5000], attr_name[80];
  char *meta_names[] = {"WESTBOUNDINGCOORDINATE", "EASTBOUNDINGCOORDINATE", "SOUTHBOUNDINGCOORDINATE", "NORTHBOUNDINGCOORDINATE"};

  fprintf(stdout, "Writing psudeo archive metadata to the output file\n");
  sprintf(meta_str, "%s", PSUEDO_ARCH_META);
  for (i=0; i<4; i++) {
    strcpy(meta_name, meta_names[i]);
    sprintf(meta_val, "%f", (float)bound_coord[i]);
    insert_meta_value(meta_str, meta_name, meta_val);
  } 

  strcpy(attr_name, "ArchiveMetadata.0");
  attr_type = DFNT_CHAR;
  attr_size = DFKNTsize(attr_type);
  attr_cnt = (int)strlen(meta_str) + 1;
  attr_len = attr_cnt + 1;
  if ((attr_buf = (void *)calloc(attr_size, attr_len)) == NULL)
    fprintf(stdout, "Cannot allocate memory for attr_buf in copy_psuedo_arch_meta\n");
  for (i=0; i<attr_cnt; i++)
    ((char *)attr_buf)[i] = meta_str[i];
  ((char *)attr_buf)[attr_cnt] = '\0';
  if(set_sd_attr(H5OUT, out_sd_id, attr_name, attr_type, attr_len, attr_buf) == FAIL)
    fprintf(stdout, "Cannot write metadata %s to output file\n", attr_name);
  free(attr_buf);
}

void insert_meta_value(char *meta_str, char *meta_name, char *meta_val)
{
  int p1, p2, p3;
  int i, k, len, cnt;
  char tmp[40];

  p1 = strpos(meta_str, meta_name, 0);
  if (p1 != -1)
  {
    p2 = strpos(meta_str, "VALUE", p1);
    if (p2 != -1)
    {
      p3 = charpos(meta_str, '=', p2);
      if (p3 != -1)
      {
	p3 += 2;
        len = strlen(meta_val);
	cnt = 9 - len;
	if (cnt < 0) {
	  len -= cnt;
	  meta_val[len] = '\0';
	}
	else {
          tmp[0] = '\0';
	  for (i=0; i<cnt; i++)
	    strcat(tmp, "0");
	  strcat(meta_val, tmp);
	}
	for (k=p3, i=0; i<9; i++, k++)
	  meta_str[k] = meta_val[i];
      }
    }
  }
}

int get_bound_coord(char **tile_id, int n_tile, char *proj_name, double *bound_coord)
{
  char tmp[10];
  int i, cnt, hn, vn;
  double lon_min, lon_max;
  double lat_min, lat_max;

  if (n_tile > 0) {
    for (i=0, cnt=0; i< n_tile; i++) {
      strmid(tile_id[i], 1, 2, tmp);
      hn = (int)atoi(tmp);
      strmid(tile_id[i], 4, 2, tmp);
      vn = (int)atoi(tmp);
      if (read_bound_table(hn, vn, proj_name, &lon_min, &lon_max, &lat_min, &lat_max) == 1)
      {
        if ((lon_min != -999.00) && (lon_max != -999.00) && (lat_max != -99.00) && (lat_min != -99.00)) {
	  if (cnt > 0) {
	    if (lon_min < bound_coord[0]) bound_coord[0] = lon_min;
	    if (lon_max > bound_coord[1]) bound_coord[1] = lon_max;
	    if (lat_min < bound_coord[2]) bound_coord[2] = lat_min;
	    if (lat_max > bound_coord[3]) bound_coord[3] = lat_max;
            cnt++;
	  }
	  else {
	    bound_coord[0] = lon_min;
	    bound_coord[1] = lon_max;
	    bound_coord[2] = lat_min;
	    bound_coord[3] = lat_max;
            cnt++;
	  }
	}
	else
          fprintf(stderr, "Invalid tile id: %s\n", tile_id[i]);
      }
      else
        fprintf(stderr, "Invalid tile id: %s\n", tile_id[i]);
    }
  }
  if (cnt == 0) return -1;
  else return 1;
}



