/* Developer: Maosheng Zhao, SSAI, 2/9/2018 */
/* wrapped functions for Grid HDFEOS2.x and HDFEOS5.x. using HDFEOS5 data type system */
/* !!! note that you must be very very cautious of the argument of field_data_type for
 * EOS_GDdeffield(), EOS_GDfieldinfo, and EOS_GDinqfields() because HDF4 and HDFEOS2 use
 * the same data type system but not HDF5 and HDFEOS5. convert_datatype_HDF4toHDFEOS5() 
 * can be used to convert data type system.
 */ 
#include "alloc_mem.h"   /* for Calloc2D */
#include "sds_rw.h"
#include "eos_sds_rw.h"

int EOS_version_check(const char *HDFEOS_name)
{
        /* written: Maosheng Zhao
         * date: Aug. 9, 2019
         * Requires HDFEOS2/HDFEOS5 library
         * Return 5 (2) as HDFEOS5 (HDFEOS2)
         * Return -1, fail and HDFEOS_name is not an HDFEOS
         */

	/* suggest usage as below style
        EOS_ver = EOS_version_check(HDFEOS_name);

        if(EOS_ver == 2)
           he5_fmt = 0;
        else if(EOS_ver == 5)
           he5_fmt = 1;
        else
        {
           fprintf(stderr, "%s is not an HDFEOS2 or HDFEOS5, exit...\n", HDFEOS_name);
           exit(1);
        }
	*/
 

        /* check the HDFEOS version of one HDFEOS filename */
	if( HE5_EHHEisHE5(HDFEOS_name) == 1 )
		return(5);                         /* HDFEOS5 file format */
	else
	{
		if( EHHEisHE2(HDFEOS_name) == 1 )
			return(2);                /* HDFEOS2 file format */
		else
		{
			fprintf(stderr, "%s is not an HDFEOS file...\n", HDFEOS_name);
			return(-1);                /* NOT HDFEOS file format */
		}
	}
}

hid_t EOS_GDopen(char *hdf_fname, char open_t, int he5_fmt)
{
	/* open_t has three access modes: 'R' for readonly, 'W' for read and write, and 'C' for create */
	/* negative return value is failure */
	hid_t eos_gdfID  = -1;

	if(hdf_fname[0] == '\0')
	{
	   fprintf(stderr, "hdf_fname is empty, exit...\n");
	   exit(0);
	}

	if(open_t == 'R')      /* read only */
	{
	  if(he5_fmt == 1)    /* HDFEOS5 */
	    eos_gdfID = HE5_GDopen((const char*)hdf_fname, H5F_ACC_RDONLY);
	  else                /* HDFEOS2 */
	    eos_gdfID = (hid_t)GDopen(hdf_fname, DFACC_READ);
	}
	else if(open_t == 'W') /* read and write */
	{
	  if(he5_fmt == 1)    /* HDFEOS5 */
	    eos_gdfID = HE5_GDopen((const char*)hdf_fname, H5F_ACC_RDWR);
	  else                /* HDFEOS2 */
	    eos_gdfID = (hid_t)GDopen(hdf_fname, DFACC_RDWR);
	}
	else if(open_t == 'C') /* create */
	{
	  if(he5_fmt == 1)    /* HDFEOS5 */
	    eos_gdfID = HE5_GDopen((const char*)hdf_fname, H5F_ACC_TRUNC);
	  else                /* HDFEOS2 */
	    eos_gdfID = (hid_t)GDopen(hdf_fname, DFACC_CREATE);
	}
	else
	{
	  fprintf(stderr, "Don't understand the access mode of open_t %c for HDFEOS file\n", open_t);
	  eos_gdfID = -1;
	}

	if(eos_gdfID < 0) eos_gdfID = -1;

	return eos_gdfID;
}

hid_t EOS_GDcreate(hid_t eos_gdfID, char* gridname, long xdim, long ydim, double upleft[], double lowright[], int he5_fmt)
{
	/* negative return value is failure */
	hid_t eos_gridID  = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  eos_gridID = HE5_GDcreate(eos_gdfID, (const char*)gridname, xdim, ydim, upleft, lowright);
	else                /* HDFEOS2 */
	  eos_gridID = (hid_t)GDcreate((int32)eos_gdfID, gridname, (int32)xdim, (int32)ydim, upleft, lowright);

	if(eos_gridID < 0) eos_gridID = -1;

	return eos_gridID;
}

hid_t EOS_GDattach(hid_t eos_gdfID, char *gridname, int he5_fmt)
{
	/* negative return value is failure */
	hid_t eos_gridID = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  eos_gridID = HE5_GDattach(eos_gdfID, (const char*)gridname);
	else                /* HDFEOS2 */
	  eos_gridID = (hid_t)GDattach((int32)eos_gdfID, gridname);

	if(eos_gridID < 0) eos_gridID = -1;

	return eos_gridID;
}

herr_t EOS_GDdetach(hid_t eos_gridID, int he5_fmt)
{
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDdetach(eos_gridID);
	else                /* HDFEOS2 */
	  status = (herr_t)GDdetach((int32)eos_gridID);

	if(status < 0) status = -1;

	return status;
}

herr_t EOS_GDclose(hid_t eos_gdfID, int he5_fmt)
{
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDclose(eos_gdfID);
	else                /* HDFEOS2 */
	  status = (herr_t)GDclose((int32)eos_gdfID);

	if(status < 0) status = -1;

	return status;
}

herr_t EOS_GDdefproj(hid_t eos_gridID, int projcode, int zonecode, int spherecode, double projparm[], int he5_fmt)
{
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDdefproj(eos_gridID, projcode, zonecode, spherecode, projparm);
	else                /* HDFEOS2 */
	  status = (herr_t)GDdefproj((int32)eos_gridID, (int32)projcode, (int32)zonecode, (int32)spherecode, projparm);

	if(status < 0) status = -1;

	return status;
}

herr_t EOS_GDprojinfo(hid_t eos_gridID, int *projcode, int *zonecode, int *spherecode, double projparm[], int he5_fmt)
{
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDprojinfo(eos_gridID, projcode, zonecode, spherecode, projparm);
	else                /* HDFEOS2 */
	{
	  /* deal with pointers with different types */
	  int32 projcode_tmp, zonecode_tmp, spherecode_tmp;
	  status = (herr_t)GDprojinfo((int32)eos_gridID, &projcode_tmp, &zonecode_tmp, &spherecode_tmp, projparm);
	  *projcode = (int)projcode_tmp;
	  *zonecode = (int)zonecode_tmp;
	  *spherecode = (int)spherecode_tmp;
	}

	if(status < 0) status = -1;

	return status;
}

herr_t EOS_GDdeforigin(hid_t eos_gridID, int origincode, int he5_fmt)
{
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDdeforigin(eos_gridID, origincode);
	else                /* HDFEOS2 */
	  status = (herr_t)GDdeforigin((int32)eos_gridID, (int32)origincode);

	if(status < 0) status = -1;

	return status;
}

herr_t EOS_GDorigininfo(hid_t eos_gridID, int *origincode, int he5_fmt)
{
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDorigininfo(eos_gridID, origincode);
	else                /* HDFEOS2 */
	{
	  /* deal with pointers with different types */
	  int32 origincode_tmp;
	  status = (herr_t)GDorigininfo((int32)eos_gridID, &origincode_tmp);
	  *origincode = (int)origincode_tmp;
	}

	if(status < 0) status = -1;

	return status;
}

herr_t EOS_GDdefpixreg(hid_t eos_gridID, int pixregcode, int he5_fmt)
{
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDdefpixreg(eos_gridID, pixregcode);
	else                /* HDFEOS2 */
	  status = (herr_t)GDdefpixreg((int32)eos_gridID, (int32)pixregcode);

	if(status < 0) status = -1;

	return status;
	
}

herr_t EOS_GDpixreginfo(hid_t eos_gridID, int *pixregcode, int he5_fmt)
{
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDpixreginfo(eos_gridID, pixregcode);
	else                /* HDFEOS2 */
	{
	  /* deal with pointers with different types */
	  int32 pixregcode_tmp;
	  status = (herr_t)GDpixreginfo((int32)eos_gridID, &pixregcode_tmp);
	  *pixregcode = (int)pixregcode_tmp;
	}

	if(status < 0) status = -1;

	return status;
}

herr_t EOS_GDdefdim(hid_t eos_gridID, char *dimname, hsize_t dim, int he5_fmt)
{
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDdefdim(eos_gridID, dimname, dim);
	else                /* HDFEOS2 */
	  status = (herr_t)GDdefdim((int32)eos_gridID, dimname, (int32)dim);

	if(status < 0) status = -1;

	return status;
}

herr_t EOS_GDdeffield(hid_t eos_gridID, char *fieldname, char *dimlist, hid_t field_data_type, int he5_fmt)
{
/* !!! note that you must be very very cautious the argument of field_data_type for
 * EOS_GDdeffield() because HDF4 and HDFEOS2 use the same data type system but not
 * HDF5 and HDFEOS5. convert_datatype_HDF4toHDFEOS5() can be used to convert
 * data type system before calling this function. Plz check how it is used in the tools. */
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDdeffield(eos_gridID, (const char*)fieldname, dimlist, NULL, field_data_type, 0);
	else                /* HDFEOS2 */
	  status = (herr_t)GDdeffield((int32)eos_gridID, fieldname, dimlist, (int32)field_data_type, HDFE_NOMERGE);

	if(status < 0) status = -1;

	return status;
}

herr_t EOS_GDfieldinfo(hid_t eos_gridID, char *fieldname, int *rank, hsize_t dims[], hid_t *ntype, char *dimlist, int he5_fmt)
{
	/* Note that for the same C data types, derived ntype code is different between HDFEOS2
	 * and HDFEOS5, please check the tables of the data types in the appendix of HDF-EOS 
	 * Users Guides,
	 *  Volume 2: Function Reference Guide */
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	{
	  long ndims, str_len;
	  char *maxdimlist = NULL;

	  ndims = HE5_GDnentries(eos_gridID, 0, &str_len);
	  if(ndims == -1)
	  {
	     fprintf(stderr, "fail to call HE5_GDnentries, exit...\n");
	     exit(0);
	  }

	  maxdimlist = (char*)calloc(str_len + 1, sizeof(char));
	  
	  status = HE5_GDfieldinfo(eos_gridID, (const char*)fieldname, rank, dims, ntype, dimlist, maxdimlist);
	  if(status == -1)
	  {
	     fprintf(stderr, "fail to call HE5_GDfieldinfo, exit...\n");
	     exit(0);
	  }

	  if(maxdimlist) free(maxdimlist);
	}
	else                /* HDFEOS2 */
	{
	  /* deal with pointers with different types */
	  int32 ntype_tmp, ndims, strbufsize;
	  int32 *dims_tmp = NULL;
	  char *dimlist_tmp = NULL;
	  int i, number_comma;

	  if(dims != NULL)
	  {

	    /* NOTE THAT the returned ndims for GDnentries with HDFE_NENTDIM can be wrong for API
 	     * GDnentries() which I found out when I tested it with MCDLCHKM data. For 2D land cover
 	     * it return ndims as 1, which is ridiculously wrong. I then found out that first using strbufsize
 	     * and then calling GDfieldinfo() to get rank and dimlist for the specified fieldname. The rank
 	     * will be the correct ndims for the fieldname but may not for the gridname because the grid
 	     * may have more ndims than the field in the grid. For example, MCD43A1 */
	    ndims = GDnentries((int32)eos_gridID, 0, &strbufsize);
	    if(ndims == -1)
	    {
	      fprintf(stderr, "fail to GDnentries, exit...\n");
	      exit(0);
	    }

	    /* because ndims from the above maybe wrong, so I use H4_MAX_NC_DIMS for dims_tmp */
	    dims_tmp    = (int32*)calloc((size_t)H4_MAX_NC_DIMS, sizeof(int32));

	    dimlist_tmp = (char*)calloc((size_t)(strbufsize+1), sizeof(char));

	    status = (herr_t)GDfieldinfo((int32)eos_gridID, fieldname, rank, dims_tmp, &ntype_tmp, dimlist_tmp);
	    if(status == -1)
	    {
	      fprintf(stderr, "fail to call GDfieldinfo, exit...\n");
	      exit(0);
	    }

	    if(ntype != NULL)
	      *ntype = (hid_t)ntype_tmp;

	    /* assign dim size for each dim and convert to the correct data type for dims */
	    /* if you incorrectly defined dims having less than rank outside, the below could be a potential bug */
	    for(i = 0; i < *rank; i++)
	      dims[i] = (hsize_t)dims_tmp[i];

	    if(dims_tmp) free(dims_tmp);
	    if(dimlist_tmp) free(dimlist_tmp);
	  }
	  else
	  {
	     status = (herr_t)GDfieldinfo((int32)eos_gridID, fieldname, rank, NULL, &ntype_tmp, dimlist);
	     if(status == -1)
	     {
		fprintf(stderr, "fail to call GDfieldinfo, exit...\n");
		exit(0);
	     }

	     if(ntype != NULL)
		*ntype = (hid_t)ntype_tmp;
	  }
	}

	return status;
}

herr_t EOS_GDreadfield(hid_t eos_gridID, char *fieldname, hssize_t start[], hsize_t edge[], void *data, int he5_fmt)
{
	/* INPUTS: eos_gridID, fieldname, start[], edge[], data, he5_fmt
 	 * OUTPUTS: to update the memory of data
 	 * RETURN: -1 fail others success */ 
	/* start[] and edge[] can be NULL */
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDreadfield(eos_gridID, (const char*)fieldname, (const hssize_t*)start, NULL, (const hsize_t*)edge, data);
	else                /* HDFEOS2 */
	{
	  /* deal with pointers with different types */
	  /* Note that MAX_VAR_DIMS is 32, but MAX_LDOPE_DIMS is 4 */
	  int32 i, start_tmp[MAX_LDOPE_DIMS] = {0}, edge_tmp[MAX_LDOPE_DIMS] = {0};
	  int32 *start_p = NULL, *edge_p = NULL;

	  /* get rank only */
	  /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
	  /* get rank */
	  int32 ntype_tmp, ndims, strbufsize;
	  int32 *dims_tmp = NULL;
	  char *dimlist_tmp = NULL;
	  int number_comma;
	  int rank;

	  /* NOTE THAT the returned ndims for GDnentries with HDFE_NENTDIM can be wrong for API
 	   * GDnentries() which I found out when I tested it with MCDLCHKM data. For 2D land cover
 	   * it return ndims as 1, which is ridiculously wrong. I then found out that first using strbufsize
 	   * and then calling GDfieldinfo() to get rank and dimlist for the specified fieldname. The rank
 	   * will be the correct ndims for the fieldname but may not for the gridname because the grid
 	   * may have more ndims than the field in the grid. For example, MCD43A1 */
	  ndims = GDnentries((int32)eos_gridID, 0, &strbufsize);
	  if(ndims == -1)
	  {
	      fprintf(stderr, "fail to GDnentries, exit...\n");
	      exit(0);
	  }

	  /* because ndims from the above maybe wrong, so I use H4_MAX_NC_DIMS for dims_tmp */
	  dims_tmp    = (int32*)calloc((size_t)H4_MAX_NC_DIMS, sizeof(int32));
	  dimlist_tmp = (char*)calloc((size_t)(strbufsize+1), sizeof(char));

	  status = (herr_t)GDfieldinfo((int32)eos_gridID, fieldname, &rank, dims_tmp, &ntype_tmp, dimlist_tmp);

	  if(dims_tmp) free(dims_tmp);
	  if(dimlist_tmp) free(dimlist_tmp);
	  /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/

	  if(status == -1)
	  {
	    fprintf(stderr, "fail to call GDfieldinfo, exit...\n");
	    exit(0);
	  }

	  if(rank > (int)MAX_LDOPE_DIMS)
	  {
	      fprintf(stderr, "The rank %d exceeds MAX_LDOPE_DIMS %d, exit...\n", rank, (int)MAX_LDOPE_DIMS);
	      exit(0);
	  }

	  if(start != NULL)
	  {
	     for(i = 0; i < rank; i++)
		start_tmp[i] = (int32)start[i];

	     start_p = start_tmp;
	  }

	  if(edge != NULL)
	  {
	     for(i = 0; i < rank; i++)
		edge_tmp[i] = (int32)edge[i];

	     edge_p = edge_tmp;
	  }
		  
	  status = (herr_t)GDreadfield((int32)eos_gridID, fieldname, start_p, NULL, edge_p, data);
	}

	if(status < 0) status = -1;

	return status;
}

herr_t EOS_GDwritefield(hid_t eos_gridID, char *fieldname, hssize_t start[], hsize_t edge[], void *data, int he5_fmt)
{
	/* INPUTS: eos_gridID, fieldname, start[], edge[], data, rank, he5_fmt
 	 * OUTPUTS: to write (or update) the data to the fieldname under eos_gridID
 	 * RETURN: -1 fail others success */ 
	/* rank is the dimensions of the data */
	/* if start[] and edge[] are NULL, rank can be just any value or zero */
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDwritefield(eos_gridID, (const char*)fieldname, (const hssize_t*)start, NULL, (const hsize_t*)edge, data);
	else                /* HDFEOS2 */
	{
	  /* deal with pointers with different types */
	  /* Note that MAX_VAR_DIMS is 32, but MAX_LDOPE_DIMS is 4 */
	  int32 i, start_tmp[MAX_LDOPE_DIMS] = {0}, edge_tmp[MAX_LDOPE_DIMS] = {0};
	  int32 *start_p = NULL, *edge_p = NULL;

	  /* get rank only */
	  /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
	  /* get rank */
	  int32 ntype_tmp, ndims, strbufsize;
	  int32 *dims_tmp = NULL;
	  char *dimlist_tmp = NULL;
	  int number_comma;
	  int rank;

	  /* NOTE THAT the returned ndims for GDnentries with HDFE_NENTDIM can be wrong for API
 	   * GDnentries() which I found out when I tested it with MCDLCHKM data. For 2D land cover
 	   * it return ndims as 1, which is ridiculously wrong. I then found out that first using strbufsize
 	   * and then calling GDfieldinfo() to get rank and dimlist for the specified fieldname. The rank
 	   * will be the correct ndims for the fieldname but may not for the gridname because the grid
 	   * may have more ndims than the field in the grid. For example, MCD43A1 */
	  ndims = GDnentries((int32)eos_gridID, 0, &strbufsize);
	  if(ndims == -1)
	  {
	      fprintf(stderr, "fail to GDnentries, exit...\n");
	      exit(0);
	  }

	  /* because ndims from the above maybe wrong, so I use H4_MAX_NC_DIMS for dims_tmp */
	  dims_tmp    = (int32*)calloc((size_t)H4_MAX_NC_DIMS, sizeof(int32));
	  dimlist_tmp = (char*)calloc((size_t)(strbufsize+1), sizeof(char));

	  status = (herr_t)GDfieldinfo((int32)eos_gridID, fieldname, &rank, dims_tmp, &ntype_tmp, dimlist_tmp);

	  if(dims_tmp) free(dims_tmp);
	  if(dimlist_tmp) free(dimlist_tmp);
	  /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/

	  if(rank > (int)MAX_LDOPE_DIMS)
	  {
	    fprintf(stderr, "The rank %d exceeds MAX_LDOPE_DIMS %d, exit...\n", rank, (int)MAX_LDOPE_DIMS);
	    exit(0);
	  }

	  if(start != NULL)
	  {
	    for(i = 0; i < rank; i++)
	      start_tmp[i] = (int32)start[i];

	    start_p = start_tmp;
	  }

	  if(edge != NULL)
	  {
	    for(i = 0; i < rank; i++)
	      edge_tmp[i] = (int32)edge[i];

	    edge_p = edge_tmp;
	  }

	  status = (herr_t)GDwritefield((int32)eos_gridID, fieldname, start_p, NULL, edge_p, data);
	    
	}

	if(status < 0) status = -1;

	return status;
}

herr_t EOS_GDgridinfo(hid_t eos_gridID, long *xdimsize, long *ydimsize, double upleft[], double lowright[], int he5_fmt)
{
	/* negative return value is failure */
	herr_t status = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  status = HE5_GDgridinfo(eos_gridID, xdimsize, ydimsize, upleft, lowright);
	else                /* HDFEOS2 */
	{
	  /* deal with pointers with different types */
	  int32 xdimsize_tmp, ydimsize_tmp;
	  status = (herr_t)GDgridinfo((int32)eos_gridID, &xdimsize_tmp, &ydimsize_tmp, upleft, lowright);
	  *xdimsize = (long)xdimsize_tmp;
	  *ydimsize = (long)ydimsize_tmp;
	}

	if(status < 0) status = -1;

	return status;
}

long EOS_GDinqgrid(char *filename, char *gridlist, long *strbufsize, int he5_fmt)
{
	/* negative return value is failure */
	long ngrid  = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	  ngrid = HE5_GDinqgrid((const char*)filename, gridlist, strbufsize);
	else                /* HDFEOS2 */
	{
	  /* deal with pointers with different types */
	  int32 strbufsize_tmp;
	  ngrid = (long)GDinqgrid(filename, gridlist, &strbufsize_tmp);

	  if(strbufsize != NULL)
	    *strbufsize = (long)strbufsize_tmp;
	}

	if(ngrid < 0) ngrid = -1;

	return ngrid;
}

long EOS_GDnentries(hid_t gridID, int entrycode, long *strbufsize, int he5_fmt)
{
	/* NOTE that for entrycode = 0 or HDFE_NENTDIM, the returned NDIM might be wrong but strbufsize could be right.
         * Please check comments in EOS_GDfieldinfo for more details */
	/* entrycode can be 0 (NENTDIM) or 4 (NENTDFLD), this can be used to get either the number of fields or number of dimensions under a grid */

	/* negative return value is failure */
	long ndims_flds  = -1;

	if(he5_fmt == 1)    /* HDFEOS5 */
	{
	  if((ndims_flds = HE5_GDnentries(gridID, entrycode, strbufsize)) == -1)
	  {
		fprintf(stderr, "fail to call HE5_GDnentries, exit...\n");
		exit(0);
	  }
	}
	else     /* HDFEOS2 */
	{
	   if(strbufsize != NULL)
	   {
	     int32 strbufsize_temp;
	     if((ndims_flds = (long)GDnentries((int32)gridID, entrycode, &strbufsize_temp)) == -1)
	     {
		fprintf(stderr, "fail to call GDnentries, exit...\n");
		exit(0);
	     }

	     if(strbufsize != NULL)
	       *strbufsize = (long)strbufsize_temp;
	   }
	   else
	     if((ndims_flds = (long)GDnentries((int32)gridID, entrycode, NULL)) == -1)
	     {
		fprintf(stderr, "fail to call GDnentries, exit...\n");
		exit(0);
	     }
	}
	
	if(ndims_flds < 0) ndims_flds = -1;

	return ndims_flds;
}

int  EOS_GDinqfields(hid_t eos_gridID, char *fieldlist, int rank[], hid_t ntype[], int he5_fmt)
{
	/* negative return value is failure */
	int nfield  = -1;
	int i;
	int32 *rank_hdfeos2 = NULL, *ntype_hdfeos2 = NULL;

	int32 num_Flds, entrycode, strbufsize;
	char *flds_list_str = NULL;

	if(he5_fmt == 1)    /* HDFEOS5 */
	{
          nfield = HE5_GDinqfields(eos_gridID, fieldlist, rank, ntype);
	}
	else                /* HDFEOS2 */
	{

	  /* we have to use flds_list_str and calloc flds_list_str first before calling GDinqfields() */
	  entrycode = 4; /* HDFE_NENTDFLD */
	  num_Flds = (size_t)GDnentries(eos_gridID, entrycode, &strbufsize);

	  if(strbufsize < 1)
	  {
	     fprintf(stderr, "strbufsize less than 1, and there is no data fields in the HDFEOS2 filem exit...\n");
	     exit(0);
	  }

	  flds_list_str = (char *)calloc(strbufsize + 1, sizeof(char)); /* need to account for terminator */

	  nfield = (int)GDinqfields(eos_gridID, flds_list_str, NULL, NULL);

	  if(flds_list_str) free(flds_list_str);

	  if(nfield < 1)
	  {
	     fprintf(stderr, "There is no data fields in the HDFEOS2 filem exit...\n");
	     exit(0);
	  }

	  /* to handle the differences in data types in the arguments in the DFHEOS2 */
	  if((rank_hdfeos2 = (int32*)calloc(nfield, sizeof(int32))) == NULL)
	  {
		fprintf(stderr, "cannot calloc rank_hdfeos2, exit...\n");
		exit(0);
	  }
	  if((ntype_hdfeos2 = (int32*)calloc(nfield, sizeof(int32))) == NULL)
	  {
		fprintf(stderr, "cannot calloc ntype_hdfeos2, exit...\n");
		exit(0);
	  }

	  nfield = (int)GDinqfields(eos_gridID, fieldlist, rank_hdfeos2, ntype_hdfeos2);

	  if(rank != NULL)
	    for(i=0; i<nfield; i++)
		rank[i] = (int)rank_hdfeos2[i];

	  if(rank_hdfeos2) free(rank_hdfeos2);

	  if(ntype != NULL)
	    for(i=0; i<nfield; i++)
		ntype[i] = (hid_t)ntype_hdfeos2[i];

	  if(ntype_hdfeos2) free(ntype_hdfeos2);

	} /* end else */

	if(nfield < 0) nfield = -1;

	return nfield;
}

int convert_datatype_HDF4toHDFEOS5(int32 H4_type, hid_t *HE5_type)
{
/* MSZ 2/9/2018 to convert data type in structure HDF4 sds_t to HDFEOS5 data type */
	int st = 1;
	switch(H4_type)
	{
	  case DFNT_CHAR8:
	    *HE5_type = HE5T_CHARSTRING;
	    break;
	  case DFNT_UCHAR8: 
	    *HE5_type = HE5T_CHARSTRING;
	    break;
	  case DFNT_INT8:
	    *HE5_type = HE5T_NATIVE_INT8;
	    break;
	  case DFNT_UINT8:
	    *HE5_type = HE5T_NATIVE_UINT8;
	    break;
	  case DFNT_INT16:
	    *HE5_type = HE5T_NATIVE_INT16;
	    break;
	  case DFNT_UINT16:
	    *HE5_type = HE5T_NATIVE_UINT16;
	    break;
	  case DFNT_INT32:
	    *HE5_type = HE5T_NATIVE_INT32;
	    break;
	  case DFNT_UINT32:
	    *HE5_type = HE5T_NATIVE_UINT32;
	    break;
	  case DFNT_FLOAT32:
	    *HE5_type = HE5T_NATIVE_FLOAT;
	    break;
	  case DFNT_FLOAT64:
	    *HE5_type = HE5T_NATIVE_DOUBLE;
	    break;
	  default:
	    fprintf(stderr, "Don't know that HDF4 data type %d\n", H4_type);
	    st = -1;
	}

   return st;
}



int convert_datatype_HDFEOS5toHDF4(hid_t HE5_type, int32 *HDF4_type)
{
/* MSZ 9/09/2019 to convert data type in structure HDF4 sds_t to HDFEOS5 data type */
	int st = 1;
	switch(HE5_type)
	{
	  case HE5T_CHARSTRING: case HE5T_NATIVE_SCHAR:
	    *HDF4_type = DFNT_CHAR8;
	    break;
	  case HE5T_NATIVE_INT8:
	    *HDF4_type = DFNT_INT8;
	    break;
	  case HE5T_NATIVE_UINT8: case HE5T_NATIVE_UCHAR:
	    *HDF4_type = DFNT_UINT8;
	    break;
	  case HE5T_NATIVE_INT16: case HE5T_NATIVE_SHORT:
	    *HDF4_type = DFNT_INT16;
	    break;
	  case HE5T_NATIVE_UINT16: case HE5T_NATIVE_USHORT:
	    *HDF4_type = DFNT_UINT16;
	    break;
	  case HE5T_NATIVE_INT32: case HE5T_NATIVE_INT:
	    *HDF4_type = DFNT_INT32;
	    break;
	  case HE5T_NATIVE_UINT32: case HE5T_NATIVE_UINT:
	    *HDF4_type = DFNT_UINT32;
	    break;
	  case HE5T_NATIVE_FLOAT:
	    *HDF4_type = DFNT_FLOAT32;
	    break;
	  case HE5T_NATIVE_DOUBLE:
	    *HDF4_type = DFNT_FLOAT64;
	    break;
	  default:
	    fprintf(stderr, "No HDF4 has that HDFEOS5 data type %d\n", HE5_type);
	    st = -1;
	}

   return st;
}

int GetHDFEOS(char *file_name, char *HDFEOSgridname, grid_mapinfo *hdfeos) 
/*
!C*****************************************************************************
!Description:
  GetHDFEOS retrives the HDF EOS metadata from a single file.

!Input Parameters:
  file_name     file name

!Input/output Parameters: (none)

!Output Parameters:
  hdfeos        HDF EOS metadata fields; all elements

!Revision History: (see file prolog)

!Team-unique Header: (see file prolog)

!References and Credits: (see file prolog)

!Design Notes:
   1. Any incomplete or corrupt fields are ignored.

MSZ 2/10/2018, move GetHDFEOS here as a library because several HDFEOS LDOPE tools use it

!END
*******************************************************************************/
{
  int i;		/* parameter index */
  hid_t file_id, grid_id;
  grid_mapinfo *h = NULL;	/* pointer to the HDF EOS metadata structure */
  int he5_fmt;
  int EOS_ver = 0;

  /* Get pointer and initialize fields to fill values */

  h = hdfeos; 
    
  h->ns = (long)(-1);
  h->nl = (long)(-1);
  h->ulCornerGrid[0] = (float64)(-999.0);
  h->ulCornerGrid[1] = (float64)(-999.0);
  h->lrCornerGrid[0] = (float64)(-999.0);
  h->lrCornerGrid[1] = (float64)(-999.0);
  h->projCode = (int)(-1);
  h->sphereCode = (int)(-1);
  for (i = 0; i < 13; i++)
    h->projParameters[i] = (float64)(-999.0);
  h->zoneCode = (int)(-1);
  h->originCode = (int)(-1);

  /* we have to use EOS_version_check() not HDF_version_check() because HDFEOS' API not HDF's are used here */
  EOS_ver = EOS_version_check(file_name);

  if(EOS_ver == 2)
     he5_fmt = 0;
  else if(EOS_ver == 5)
     he5_fmt = 1;
  else
  {
     fprintf(stderr, "%s is not an HDFEOS2 or HDFEOS5, exit...\n", file_name);
     exit(1);
  }
  
 /* Open the geolocation file */
  if ((file_id = EOS_GDopen(file_name, 'R', he5_fmt)) == FAIL) {
    fprintf(stderr, "Input file %s open failure (EOS_GDopen)\n", file_name);  
    return -1;
  }

  /* Read info for the first layer (get it from the right grid type) */
   
  if ((grid_id = EOS_GDattach(file_id, HDFEOSgridname, he5_fmt)) == FAIL) {
    fprintf(stderr, "can't attach to %s (EOS_GDattach) to GetHDFEOS!\n", HDFEOSgridname); 
    if (EOS_GDclose(file_id, he5_fmt) == FAIL)
      fprintf(stderr, "can't close file (EOS_GDclose)\n");
    return -1;
  }

  /* get grid map info */
  if (EOS_GDgridmapinfo(grid_id, h, he5_fmt) == FAIL) {
    fprintf(stderr, "can't get grid info (EOS_GDgridmapinfo)!\n");
    return -1;
  }

  /* Detach from the grid and close the file */
  if (EOS_GDdetach(grid_id, he5_fmt) == FAIL) {
    fprintf(stderr, "can't detach from grid (EOS_GDdetach)!\n");  
    return -1;
  }
  
  if (EOS_GDclose(file_id, he5_fmt) == FAIL) {
    fprintf(stderr, "can't close file (GDclose)!\n"); 
    return -1;
  }

  return 0;
}

herr_t EOS_GDgridmapinfo(hid_t gridID, grid_mapinfo* mapinfo, int he5_fmt)
{
  /* written: Maosheng Zhao
   * date: Aug. 11, 2019
   * Requires HDFEOS2/HDFEOS5 library
   */ 
  /* negative return value is failure */
  herr_t status = -1;

  if((status = EOS_GDgridinfo(gridID, &mapinfo->ns, &mapinfo->nl, mapinfo->ulCornerGrid, mapinfo->lrCornerGrid, he5_fmt)) == -1)
  {
     fprintf(stderr, "cannot EOS_GDgridinfo from gridID %d, exit\n", gridID);
     exit(1);	
  }

  if((status = EOS_GDprojinfo(gridID, &mapinfo->projCode, &mapinfo->zoneCode, &mapinfo->sphereCode, mapinfo->projParameters, he5_fmt)) == -1)
  {
     fprintf(stderr, "cannot EOS_GDprojinfo from gridID %d, exit\n", gridID);
     exit(1);
  }

  if((status = EOS_GDorigininfo(gridID, &mapinfo->originCode, he5_fmt)) == -1)
  {
     fprintf(stderr, "cannot EOS_GDorigininfo from %d, exit\n", gridID);
     /* in case some HDF missing this, we will not exit but continue */
     /*exit(1);*/
  }

  if(status < 0) status = -1;

  return status;
  

  return(0);
}

int get_EOS_all_grids_info(char *HE_name, EOS_Grids_info* Grids_info)
{
  /* Developer: Maosheng Zhao
   * Date: 8/16/2019
   * INPUTS: HE_name
   * OUTPUTS: Grids_info   (memebers are calloced in this function)
   * RETURN: -1 if fail.
   * Purpose: get all gridnames in an HDFEOS2 or HDFEOS5, include their each proj_info_vectors, respectively
   * Grids_info: (derived) the struct containing information for all girds in an HDFEOS
   * required functions:
   * int EOS_version_check(const char *HDF_name);
   * herr_t H5Fget_objs_links(hid_t fid, H5_OBJS_TYPES *h5_obj_types);
   * static herr_t H5L_get_num_objs_links(hid_t fid, hsize_t *num_objs_links);
   * static herr_t H5L_get_names_types_objs_links(hid_t fid, H5_OBJS_TYPES *h5_obj_types);
   * static herr_t cb_func_num_objs_links_L(hid_t loc_id, const char *name, const H5L_info_t *info, void *operator_data);
   * static herr_t cb_func_names_objs_links_L(hid_t loc_id, const char *name, const H5L_info_t *info, void *operator_data);
   * NOTE: the function free_EOS_all_grids_info_members(EOS_Grids_info* Grids_info) will free the struct calloced in this function
   */ 

  int EOS_ver, he5_fmt;
  int st = -1;
  size_t num_Grids = 0;
  size_t num_Flds = 0;
  int i_sds, i_grid, j;
  int entrycode;
  long strbufsize;
  char *grids_list_str = NULL, *flds_list_str = NULL;
  char *file_attr_temp = NULL;
  int trim_spaces = 0;
  char **gridnames = NULL;
  char **fieldnames= NULL;

  herr_t status;
  hid_t gdfID, gridID;

  /* 1. get he5_fmt value */
  EOS_ver = EOS_version_check((const char*)HE_name);

  if(EOS_ver == 2)
     he5_fmt = 0;
  else if(EOS_ver == 5)
     he5_fmt = 1;
  else
  {
     fprintf(stderr, "%s is not an HDFEOS2 or HDFEOS5, exit...\n", HE_name);
     exit(1);
  }

  /* 2. get ngrid and grids_list_str */
  num_Grids = (size_t)EOS_GDinqgrid(HE_name, NULL, &strbufsize, he5_fmt);
  if(num_Grids < 1)
  {
     st = -1;
     return st;
  }
  else
  {
     grids_list_str = (char*)calloc(strbufsize + 1, sizeof(char));
     num_Grids = (size_t)EOS_GDinqgrid(HE_name, grids_list_str, &strbufsize, he5_fmt);

     if(num_Grids < 1)
     {
	fprintf(stderr, "num_Grids = %d, less than 1, not an HDFEOS file??? exit...\n");
	exit(0);
     }

     /* get the elements of EOS_Grids_info */
     /* 1: num_grids */
     Grids_info->num_grids = (int)num_Grids;

     /* 2: gridnames */
     Grids_info->gridnames = (char **)Calloc2D(Grids_info->num_grids, H4_MAX_NC_NAME, sizeof(char));

     if(Grids_info->num_grids == 1)
	strcpy(Grids_info->gridnames[0], grids_list_str);
     else
     {
	/* split the grids_list_str using ',' as delimit into gridnames, and gridnames are calloc inside of the split_string_with_delimit */
	split_string_with_delimit(grids_list_str, (size_t)(strbufsize), ',', &gridnames, &num_Grids, trim_spaces);
	if(grids_list_str) free(grids_list_str);

	for(i_grid = 0; i_grid < (int)num_Grids; i_grid++)
	{
	   strcpy(Grids_info->gridnames[i_grid], gridnames[i_grid]);
	   if(gridnames[i_grid]) free(gridnames[i_grid]);
	}
	if(gridnames) free(gridnames);
     }

     /* 3: get map projection information for each grid by using higher level EOS_GD APIs */
     Grids_info->proj_info_vector = (grid_mapinfo*)calloc(Grids_info->num_grids, sizeof(grid_mapinfo));
     Grids_info->fields_names_struct_vector=(EOS_Grid_fields*)calloc(Grids_info->num_grids, sizeof(EOS_Grid_fields));

     /* 3.1. GDopen */
     if((gdfID = EOS_GDopen(HE_name, 'R', he5_fmt)) == -1)
     {
	fprintf(stderr, "cannot EOS_GDopen %s, exit\n", HE_name);
	exit(1);
     }

     /* 3.2. attach the gridname, to get grid mapinfo and EOS_Grid_fields, and then GDdetach */
     for(i_grid = 0; i_grid < (int)num_Grids; i_grid++)
     {
	if((gridID = EOS_GDattach(gdfID, Grids_info->gridnames[i_grid], he5_fmt)) == -1)
	{
	   fprintf(stderr, "cannot EOS_GDattach %s in %s, exit\n", Grids_info->gridnames[i_grid], HE_name);
	   exit(1);
	}

	if((status = EOS_GDgridmapinfo(gridID, &Grids_info->proj_info_vector[i_grid], he5_fmt)) == -1)
	{
	   fprintf(stderr, "cannot EOS_GDattach %s in %s, exit\n", Grids_info->gridnames[i_grid], HE_name);
	   exit(1);
	}

	/* get the number of fields and their names for each grid */
	entrycode = 4; /* HDFE_NENTDFLD or HE5_HDFE_NENTDFLD */
	num_Flds = (size_t)EOS_GDnentries(gridID, entrycode, &strbufsize, he5_fmt);

	if(num_Flds < 1)
	{
	   fprintf(stderr, "There is no field for the grid %s, exit...\n", Grids_info->gridnames[i_grid]);
	   exit(0);
	}


	flds_list_str = (char *)calloc(strbufsize + 1, sizeof(char)); /* need to account for terminator */

	/* get the number of fields and all fieldnames for the grid */
	Grids_info->fields_names_struct_vector[i_grid].num_fields = (int)num_Flds;

	num_Flds = (size_t)EOS_GDinqfields(gridID, flds_list_str, NULL, NULL, he5_fmt);

	split_string_with_delimit(flds_list_str, (size_t)(strbufsize), ',', &fieldnames, &num_Flds, trim_spaces);
	if(flds_list_str)free(flds_list_str);

	Grids_info->fields_names_struct_vector[i_grid].fieldnames = (char**)Calloc2D(num_Flds, H4_MAX_NC_NAME, sizeof(char));

	for(i_sds = 0; i_sds < (int)num_Flds; i_sds++)
	{
	   strcpy(Grids_info->fields_names_struct_vector[i_grid].fieldnames[i_sds], fieldnames[i_sds]);
	   if(fieldnames[i_sds]) free(fieldnames[i_sds]);
	}
	if(fieldnames)free(fieldnames);

	/* detach the grid */
	if((status = EOS_GDdetach(gridID, he5_fmt)) == -1)
	{
	   fprintf(stderr, "cannot EOS_GDdetach %s in %s, exit\n", Grids_info->gridnames[i_grid], HE_name);
	   exit(1);
	}

     }

     /* 3.1. GDclose */
     if((status = EOS_GDclose(gdfID, he5_fmt)) == -1)
     {
        fprintf(stderr, "cannot EOS_GDclose %s, exit\n", HE_name);
        exit(1);
     }

    st = 1;
    return st;
  }

}

int free_EOS_all_grids_info_members(EOS_Grids_info* Grids_info)
{
   /* developer: Maosheng Zhao, 9/05/2019
    * purpose: free members of Grids_info
    * RETURN: -1 if fail, 0 scuess
    * NOTE: the memory of the members of the struct of Grids_info is calloced
    * in get_EOS_all_grids_info()
    */

    /* free from kids to parents */
    if(Grids_info != NULL && Grids_info->num_grids > 0)
    {
      /* free memebers */
      if(Grids_info->fields_names_struct_vector != NULL)
      {
	if(Grids_info->fields_names_struct_vector->num_fields > 0)
	  Free2D((void**)Grids_info->fields_names_struct_vector->fieldnames);
      }

      if(Grids_info->proj_info_vector != NULL)
	 free(Grids_info->proj_info_vector);

      Free2D((void**)Grids_info->gridnames);

      return 0;

    }
    else
      return -1;
    
}

int get_EOS_datatype_size(hid_t HE_dtype_id, int he5_fmt)
{
    /* developer: Maosheng Zhao, 9/05/2019
     * get the size of the HDFEOS (not HDF) datatype in bytes.
     * INPUTS: HE_dtype_id, he5_fmt
     * OUTPUTS: none
     * RETURN: -1 for failure and others for size
     */

    hid_t HDF5_dtype_id = -1;

    if(he5_fmt == 1)   /* HDFEOS5 */
    {
      /* HDFEOS5 and HDF5 have inconsistent data type codes and
       * the HDFEOS5 datatype should be converted to HDF5 first before using HDF5 API
       * H5Tget_size() to get the size. 
       * I accidently found an very useful function, HE5_EHconvdatatype() in EHapi.c
       * of HDFEOS5 library which is not mentioned in HDFEOS5 reference or user guide */
      HDF5_dtype_id = HE5_EHconvdatatype((int)HE_dtype_id);

      if(HDF5_dtype_id != -1)
      {
	if(H5Tis_variable_str(HDF5_dtype_id) > 0)
	   /* treat the string H5T_C_S1 as character. MAYBE a bug and will check further */
	   return 1;
	else
	   return H5Tget_size(HDF5_dtype_id);
      }
      else
	return -1;
    }
    else               /* HDFEOS2 */
    {
      /* HDFEOS2 and HDF4 have consistent data type codes */
      if((HE_dtype_id >= 3 && HE_dtype_id <= 6) || (HE_dtype_id >= 20 && HE_dtype_id <= 25))
	return DFKNTsize((int32)HE_dtype_id);
      else
      {
	fprintf(stderr, "You need to expand this function as there is a new data type. Check Apendix A. Numbertype Codes in HDFEOS2 guide.\n");
	return -1;
      }
    }
}

int get_EOS_datatype_sign(int32 H4_type, int *data_sign)
{
/* MSZ 11/25/2019, to get data_sign, a member of the struct EOS_data_info.
 * Note that: for the data_sign,  0 for unsigned int and 1 for signed int and 2 for float or double.
 * The size_data is also a member of the struct EOS_data_info. Thus both data_sign and size_data
 * can be used together */
	int st = 1;
	switch(H4_type)
	{
	  case DFNT_UCHAR8: case DFNT_UINT8: case DFNT_UINT16: case DFNT_UINT32:
	    *data_sign = 0;
	    break;
	  case DFNT_CHAR8: case DFNT_INT8: case DFNT_INT16: case DFNT_INT32:
	    *data_sign = 1;
	    break;
	  case DFNT_FLOAT32: case DFNT_FLOAT64:
	    *data_sign = 2;
	    break;
	  default:
	    fprintf(stderr, "Don't know that HDF4 data type %d\n", H4_type);
	    st = -1;
	}

   return st;
}

int get_EOS_gridname_for_fieldname(char *EOS_FILE_NAME, char *fieldname, char *gridname)
{
   /* developer: Maosheng Zhao, 9/05/2019
    * purpose: get the gridname for a specified fieldname after calling get_EOS_all_grids_info().
    * INPUTS: Grids_info, fieldname
    * OUTPUTS: gridname
    * RETURN: -1 if no gridname is found; >=0 is success; and >1 is more than 1 gridname is found and the last one is used.
    */
    EOS_Grids_info Grids_info;
    int i_grd, i_fld;
    int n_exist = 0;

    /* STEP 1, get the information of all the grids */
    if(get_EOS_all_grids_info(EOS_FILE_NAME, &Grids_info) == -1)
    {
	fprintf(stderr, "fail to get_EOS_all_grids_info from %s, exit...\n", EOS_FILE_NAME);
	exit(0);
    }
   
    /* STEP 2, get the gridname */ 
    if(Grids_info.num_grids > 0)
    {
	/* global search to get the gridname(s) for the fieldname */
	for(i_grd = 0; i_grd < Grids_info.num_grids; i_grd++)
	{
	   if(Grids_info.fields_names_struct_vector->num_fields > 0)
	   {
	      for(i_fld = 0; i_fld < Grids_info.fields_names_struct_vector->num_fields; i_fld++)
	      {
		if(strcmp(fieldname, Grids_info.fields_names_struct_vector->fieldnames[i_fld]) == 0)
		{
		   /* get the gridname */
		   strcpy(gridname, Grids_info.gridnames[i_grd]);
		   n_exist++;

		   if(n_exist > 1)
		      fprintf(stderr, "There are more than one gridname for the fieldname %s, PROBLEM!!! the last one %s is used.\n", fieldname, gridname);
		}
	      }
	   }
	}

	if(n_exist == 0)
	{
	   fprintf(stderr, "cannot find the gridname for the fieldname %s, are you use the right fieldname???, exit...\n", fieldname);
	   return -1;
	}
	else
	   return n_exist;
    }
    else
    {
	fprintf(stderr, "there is no gridname...\n");
	return -1;
    }

    /* STEP 3, finally free members of Grids_info */
    free_EOS_all_grids_info_members(&Grids_info);
}

int split_string_with_delimit(char *long_string, size_t long_str_len, char delimit,  char ***sub_strs, size_t *num_sub_strs, int trim_spaces)
{
    /* Developer: Maosheng Zhao, 4/5/2019 */
    /* Note that long_str_len does not account for the string terminator in the end */
    /* Note that we assume that there is only one delimit between the two sub-strings.i.e., no duplicated delimits */
    /* split a long string with returns or new lines into lines with selection of weather triming
     * the leading and tailing spaces */
    /* Note that **sub_strs is calloced in this function */

   size_t i;
   size_t *str_start_loc = NULL, *str_end_loc = NULL;
   char *temp_char = NULL, *temp_char_end = NULL;
   size_t str_start, str_end;

   /* get the real start and end without delimit */
   for (i = 0; i < long_str_len; i++)
	if(long_string[i] != delimit)
	{
	   str_start = i;
	   break;
	}

   for(i = long_str_len - 1; i != (size_t)-1; i--)
	if(long_string[i] != delimit)
	{
	   str_end = i;
	   break;
	}

    /* check how many sub strings. Note that here we assume that there is only one delimit between the two sub-strings  */
    *num_sub_strs = 0;
    for (i = str_start; i <= str_end; i++)
	if( ((char*)long_string)[i] == delimit)
	   (*num_sub_strs)++;

    /* because now from str_start and str_end, there is no delimit in the end or front */
    (*num_sub_strs) += 1;

    /* record the location of start and end subscript for each string */
    str_start_loc = (size_t*)calloc(*num_sub_strs, sizeof(size_t));
    str_end_loc = (size_t*)calloc(*num_sub_strs, sizeof(size_t));

    *num_sub_strs = 0;
    /* the first string starts from the 0 */
    str_start_loc[*num_sub_strs] = str_start;
    for (i = str_start; i <= str_end; i++)
	if(((char*)long_string)[i] == delimit) {

	   /* next char is the start of the next string */
	   str_start_loc[*num_sub_strs + 1] = i + 1;

	   /* it the end of the current string */
	   str_end_loc[*num_sub_strs] = i - 1;

	   (*num_sub_strs)++;
	}

    /* because the above *num_sub_strs misses the last sub string */
    *num_sub_strs += 1;

    str_end_loc[*num_sub_strs - 1] = str_end;

    /* split the long_string into strings */
    /* as mentioned above, *num_sub_strs misses the last one */
    *sub_strs = (char**)calloc(*num_sub_strs, sizeof(char*));
    for (i = 0; i < *num_sub_strs; i++)
    {
	/* trim leading and trailing spaces and correct str_start_loc and str_end_loc */
	if(trim_spaces) {
	    /* trim leading spaces */
	    temp_char = long_string + str_start_loc[i];
	    while(isspace((unsigned char)*temp_char)) {
		temp_char++;
		str_start_loc[i]++;
	    }

	    /* also trailing spaces ? */
	    if(*temp_char == 0)
		str_end_loc[i] = str_start_loc[i];
	    else {
		temp_char_end = temp_char + strlen(temp_char) - 1;
		while(temp_char_end > temp_char && isspace((unsigned char)*temp_char_end)) {
		   temp_char_end--;
		   str_end_loc[i]--;
		}
		/* Write new null terminator */
		*(temp_char_end + 1) = '\0';
	    } /* end of else */
	}

	if(str_end_loc[i] <= str_start_loc[i])
	{
	    (*sub_strs)[i] = (char*)calloc(1, sizeof(char));
	    (*sub_strs)[i][0] = '\0';
	}
	else  /* note that str_end_loc not include the last null terminator */
	{
	    (*sub_strs)[i] = (char*)calloc(str_end_loc[i] - str_start_loc[i] + 1, sizeof(char)); 
	    memmove((*sub_strs)[i], long_string + str_start_loc[i], str_end_loc[i] - str_start_loc[i] + 1);
	}

	/*fprintf(stderr, "str=%s len=%lu str_start_loc=%lu str_end_loc=%lu\n", (*sub_strs)[i], strlen((*sub_strs)[i]), str_start_loc[i], str_end_loc[i]);*/
    }

    return(0);
}


EOS_data_info* get_EOS_field_data_info(char *EOS_FILE_NAME, char *EOS_dataset_name)
{
	/* developer: Maosheng Zhao, Oct 11, 2019
         * Required EOS_sds_rw.c and EOS_sds_rw.h higher level API
         * to get the key data information (data shape and type) of an
         *  EOS_dataset_name stored in either HDFEOS2 or HDFEOS5 file.
         * Note that for EOS_dataset_name of HDFEOS5, there is no need
         * to include the group path into the EOS_dataset_name or fieldname.
         * INOUTS: EOS_FILE_NAME, EOS_dataset_name
         * OUTPUT: struct EOS_field_data_info, dims[], data_type_HDF4, data_type_HDF5;  Obviously, rank is the length of the dims[] vector
         * RETURN: EOS_data_info pointer. Return NULL if failed.
         */

	int i;
	int EOS_ver, he5_fmt;
	char gridname[MAX_NC_NAME] = {'\0'}, fieldname[MAX_NC_NAME] = {'\0'};
	hid_t eos_gdfID, eos_gridID;

	hid_t data_type, data_type_HDFEOS5;

	int size_data = 0, data_sign;
	size_t numbers_data = 1;

	int rank, data_type_HDF4;
	hsize_t dims[H4_MAX_VAR_DIMS];

	EOS_data_info *data_info = NULL;

	/* string copy the fieldname */
	strcpy(fieldname, EOS_dataset_name);

	/* STEP 1, get the gridname for the fieldname */
	if((get_EOS_gridname_for_fieldname(EOS_FILE_NAME, fieldname, gridname)) == -1)
	{
	   fprintf(stderr, "fail to get_EOS_gridname_for_fieldname of %s...\n", fieldname);
	   return NULL;
	}
	else
	   fprintf(stderr, "the grdiname for %s is %s\n", fieldname, gridname);

	/* STEP 2, get the he5_fmt  */
	EOS_ver = EOS_version_check(EOS_FILE_NAME);

	if(EOS_ver == 2)
	   he5_fmt = 0;
	else if(EOS_ver == 5)
	   he5_fmt = 1;
	else
	{
	   fprintf(stderr, "%s is not an HDFEOS2 or HDFEOS5...\n", EOS_FILE_NAME);
	   return NULL;
	}

	/* STEP 3, open the HDFEOS file for reading */
	if((eos_gdfID = EOS_GDopen(EOS_FILE_NAME, 'R', he5_fmt)) == -1)
	{
	   fprintf(stderr, "fail to EOS_GDopen %s...\n", EOS_FILE_NAME);
	   return NULL;
	}

	/* STEP 4, attach the grid */
	if((eos_gridID = EOS_GDattach(eos_gdfID, gridname, he5_fmt)) == -1)
	{
	   fprintf(stderr, "fail to EOS_GDattach  %s...\n", gridname);
	   return NULL;
	}

	/* STEP 5, get the field data information */
	if((EOS_GDfieldinfo(eos_gridID, fieldname, &rank, dims, &data_type, NULL, he5_fmt)) == -1)
	{
	   fprintf(stderr, "fail to EOS_GDfieldinfo  %s...\n", gridname);
	   return NULL;
	}

	/* STEP 6, 1. get the size of the actual datatype in bytes
 	 *  	   2. get the numbers_data
 	 *  	   3. calloc data
 	 *  */
	if((size_data = get_EOS_datatype_size(data_type, he5_fmt)) == -1)
	{
	   fprintf(stderr, "fail to get_EOS_datatype_size...\n");
	   return NULL;
	}

	if(he5_fmt)
	{
	   data_type_HDFEOS5 = data_type;
	   if(convert_datatype_HDFEOS5toHDF4(data_type, &data_type_HDF4) == -1)
	      fprintf(stderr, "fail to call convert_datatype_HDFEOS5toHDF4()\n");
	}
	else
	{
	  data_type_HDF4 = (int32)data_type;
	  if(convert_datatype_HDF4toHDFEOS5(data_type, &data_type_HDFEOS5) == -1)
	     fprintf(stderr, "fail to call convert_datatype_HDF4toHDFEOS5()\n");
	}

	/* STEP 7, get the data_sign */
	get_EOS_datatype_sign(data_type_HDF4, &data_sign);

	/* added data_sign. Using data_type_HDF4 to determine data_sign. 
 	 * Note that for data_type_HDFEOS5, the method of just using data_type_HDF4
 	 * may result in miss some data types */
	switch(data_type_HDF4)

	if(rank < 1)
	{
	   fprintf(stderr, "rank for %s is less than 1!!!...\n", fieldname);
	   return NULL;
	}

	numbers_data = 1;
	for(i = 0; i < rank; i++)
	   numbers_data *= dims[i];

	if(numbers_data <= 0)
	{
	   fprintf(stderr, "There are zero length of data for some dimensions of data for %s...\n", fieldname);
	   return NULL;
	}

	/* STEP 7, detached the grid and close EOS file */
	if(EOS_GDdetach(eos_gridID, he5_fmt) == -1)
	{
	   fprintf(stderr, "can't detach from grid (EOS_GDdetach)!...\n");
	   return NULL;
	}

	if(EOS_GDclose(eos_gdfID, he5_fmt) == -1)
	{
	   fprintf(stderr, "can't close file (GDclose)!...\n");
	   return NULL;
	}

	/* STEP 8, assign all related values to all elements */
	if((data_info = (EOS_data_info*)calloc(1, sizeof(EOS_data_info))) == NULL)
	{
	   fprintf(stderr, "cannot calloc data_info...\n");
	   return NULL;
	}
	
	strcpy(data_info->field_name, EOS_dataset_name);
	strcpy(data_info->EOS_file_name, EOS_FILE_NAME);
	strcpy(data_info->gridname, gridname);
	data_info->he5_fmt = he5_fmt;
	data_info->rank = rank;
	if((data_info->dims = (hsize_t*)calloc((size_t)rank, sizeof(hsize_t))) == NULL)
	{
	   fprintf(stderr, "cannot calloc dims...\n");
	   return NULL;
	}
	for(i = 0; i < rank; i++)
	   data_info->dims[i] = (hsize_t)dims[i];

	data_info->data_type_HDF4 = data_type_HDF4;
	data_info->data_type_HDFEOS5 = data_type_HDFEOS5;
	data_info->size_data = size_data;
	data_info->data_sign = data_sign;
	data_info->numbers_data = numbers_data;

	return data_info;

}

void *Get_EOS_data_2D_subset(char *EOS_FILE_NAME, char *EOS_dataset_name, EOS_data_info *data_info, long col_start, long row_start, long cols, long rows)
{
	/* Developer: Maosheng Zhao, 11/22/2019.
 	 * modified from EOS_DUMP_BIN_MALLOC() for read chunked lines of 2D data only
 	 * from HDFEOS2/HDFEOS5 data field.
 	 * INPUTs: EOS_FILE_NAME, EOS_dataset_name, data_info, row (0-based), and rows (number of rows to be read)
 	 * OUTPUT: void data
 	 * Note that the data memory is calloc in this function and data_info can be first obtained from
 	 * calling the function get_EOS_field_data_info() */

	int i, he5_fmt;
	hid_t eos_gdfID, eos_gridID;
	size_t numbers_data = 0;
	hssize_t start[2] = {0}, edge[2] = {0};
	void *data = NULL;

	if(data_info->rank != 2)
	{
	   fprintf(stderr, "This function only works for 2D data field!!!...\n");
	   return NULL;
	}

	/* get he5_fmt, which is frequently used */
	he5_fmt = data_info->he5_fmt;

	/* STEP 1, open the HDFEOS file for reading */
	if((eos_gdfID = EOS_GDopen(EOS_FILE_NAME, 'R', he5_fmt)) == -1)
	{
	   fprintf(stderr, "fail to EOS_GDopen %s...\n", EOS_FILE_NAME);
	   return NULL;
	}

	/* STEP 2, attach the grid */
	if((eos_gridID = EOS_GDattach(eos_gdfID, data_info->gridname, he5_fmt)) == -1)
	{
	   fprintf(stderr, "fail to EOS_GDattach %s...\n", data_info->gridname);
	   return NULL;
	}

	/* STEP 3, calloc data */
	numbers_data = (size_t)cols * (size_t)rows; 
	if((data = calloc(numbers_data, (size_t)data_info->size_data)) == NULL)
	{
	   fprintf(stderr, "cannot calloc data...\n");
	   return NULL;
	}

	/* STEP 4, read the entire data for the fieldname once */
	start[0] = (hssize_t)row_start;
	start[1] = (hssize_t)col_start;
	edge[0] = (hsize_t)rows;
	edge[1] = (hsize_t)cols;
	if((EOS_GDreadfield(eos_gridID, EOS_dataset_name, start, edge, data, he5_fmt)) == -1)
	{
	   fprintf(stderr, "cannnot EOS_GDreadfield the fieldname %s...\n", EOS_dataset_name);
	   return NULL;
	}

	/* STEP 5, detached the grid and close EOS file */
	if(EOS_GDdetach(eos_gridID, he5_fmt) == -1)
	{
	   fprintf(stderr, "can't detach from grid (EOS_GDdetach)!...\n");
	   return NULL;
	}

	if(EOS_GDclose(eos_gdfID, he5_fmt) == -1)
	{
	   fprintf(stderr, "can't close file (GDclose)!...\n");
	   return NULL;
	}

	return data;

}

void *Get_EOS_data_2D_lines(char *EOS_FILE_NAME, char *EOS_dataset_name, EOS_data_info *data_info, int row, int rows)
{
	/* Developer: Maosheng Zhao, 10/13/2019.
 	 * modified from EOS_DUMP_BIN_MALLOC() for read chunked lines of 2D data only
 	 * from HDFEOS2/HDFEOS5 data field.
 	 * INPUTs: EOS_FILE_NAME, EOS_dataset_name, data_info, row (0-based), and rows (number of rows to be read)
 	 * OUTPUT: void data
 	 * Note that the data memory is calloc in this function and data_info can be first obtained from
 	 * calling the function get_EOS_field_data_info() */

	int i, he5_fmt;
	hid_t eos_gdfID, eos_gridID;
	size_t numbers_data = 0;
	hssize_t start[2] = {0}, edge[2] = {0};
	void *data = NULL;

	if(data_info->rank != 2)
	{
	   fprintf(stderr, "This function only works for 2D data field!!!...\n");
	   return NULL;
	}

	/* get he5_fmt, which is frequently used */
	he5_fmt = data_info->he5_fmt;

	/* STEP 1, open the HDFEOS file for reading */
	if((eos_gdfID = EOS_GDopen(EOS_FILE_NAME, 'R', he5_fmt)) == -1)
	{
	   fprintf(stderr, "fail to EOS_GDopen %s...\n", EOS_FILE_NAME);
	   return NULL;
	}

	/* STEP 2, attach the grid */
	if((eos_gridID = EOS_GDattach(eos_gdfID, data_info->gridname, he5_fmt)) == -1)
	{
	   fprintf(stderr, "fail to EOS_GDattach %s...\n", data_info->gridname);
	   return NULL;
	}

	/* STEP 3, calloc data */
	numbers_data = data_info->dims[1] * (size_t)rows; 
	if((data = calloc(numbers_data, (size_t)data_info->size_data)) == NULL)
	{
	   fprintf(stderr, "cannot calloc data...\n");
	   return NULL;
	}

	/* STEP 4, read the entire data for the fieldname once */
	start[0] = (hssize_t)row;
	edge[0] = (hsize_t)rows;
	edge[1] = data_info->dims[1];
	if((EOS_GDreadfield(eos_gridID, EOS_dataset_name, start, edge, data, he5_fmt)) == -1)
	{
	   fprintf(stderr, "cannnot EOS_GDreadfield the fieldname %s...\n", EOS_dataset_name);
	   return NULL;
	}

	/* STEP 5, detached the grid and close EOS file */
	if(EOS_GDdetach(eos_gridID, he5_fmt) == -1)
	{
	   fprintf(stderr, "can't detach from grid (EOS_GDdetach)!...\n");
	   return NULL;
	}

	if(EOS_GDclose(eos_gdfID, he5_fmt) == -1)
	{
	   fprintf(stderr, "can't close file (GDclose)!...\n");
	   return NULL;
	}

	return data;

}

void *EOS_DUMP_BIN_MALLOC(char *EOS_FILE_NAME, char *EOS_dataset_name, EOS_data_info *data_info)
{
	/* developer: Maosheng Zhao, Aug 30, 2019
         * Required EOS_sds_rw.c and EOS_sds_rw.h higher level API. data_info should be
         * obtained by calling get_EOS_field_data_info() before calling this function
         * to read the entire data for the specified dataset name
         * in either HDFEOS2 or HDFEOS5 file.
         * Note that for EOS_dataset_name of HDFEOS5, there is no need
         * to include the group path into the EOS_dataset_name or fieldname.
         * INOUTS: EOS_FILE_NAME, EOS_dataset_name and data_info
         * OUTPUT: data;  data_info is available using get_EOS_field_data_info()
         * RETURN: data void pointer. Return NULL if failed.
         * In one word, this function read all data for a dataset and data information
         * can be available using get_EOS_field_data_info() such as data shape, data type etc
         * prior calling this function
         */

	int i, he5_fmt;
	hid_t eos_gdfID, eos_gridID;
	void *data = NULL;

	/* get he5_fmt, which is frequently used */
	he5_fmt = data_info->he5_fmt;

	/* STEP 1, open the HDFEOS file for reading */
	if((eos_gdfID = EOS_GDopen(EOS_FILE_NAME, 'R', he5_fmt)) == -1)
	{
	   fprintf(stderr, "fail to EOS_GDopen %s...\n", EOS_FILE_NAME);
	   return NULL;
	}

	/* STEP 2, attach the grid */
	if((eos_gridID = EOS_GDattach(eos_gdfID, data_info->gridname, he5_fmt)) == -1)
	{
	   fprintf(stderr, "fail to EOS_GDattach %s...\n", data_info->gridname);
	   return NULL;
	}

	/* STEP 3, calloc data */
	if((data = calloc(data_info->numbers_data, (size_t)data_info->size_data)) == NULL)
	{
	   fprintf(stderr, "cannot calloc data...\n");
	   return NULL;
	}

	/* STEP 4, read the entire data for the fieldname once */
	if((EOS_GDreadfield(eos_gridID, EOS_dataset_name, NULL, NULL, data, he5_fmt)) == -1)
	{
	   fprintf(stderr, "cannnot EOS_GDreadfield the fieldname %s...\n", EOS_dataset_name);
	   return NULL;
	}

	/* STEP 5, detached the grid and close EOS file */
	if(EOS_GDdetach(eos_gridID, he5_fmt) == -1)
	{
	   fprintf(stderr, "can't detach from grid (EOS_GDdetach)!...\n");
	   return NULL;
	}

	if(EOS_GDclose(eos_gdfID, he5_fmt) == -1)
	{
	   fprintf(stderr, "can't close file (GDclose)!...\n");
	   return NULL;
	}

	return data;

}

void *reproj_GEO_data_to_SIN_data_2D(void *Geo_data, int Geo_data_elemt_bytes, int data_sign, double Geo_lon_min, double Geo_lon_max, double Geo_lat_min, double Geo_lat_max, long Geo_cols, long Geo_rows, double SIN_x_min, double SIN_x_max, double SIN_y_min, double SIN_y_max, long SIN_cols, long SIN_rows)
{
	/* Developer: Maosheng Zhao
 	 * Purpose: reproject the Geographic data to Sinusoidal data.
 	 * INPUTs:  all 14 arguments and they must be known prior use the function.
 	 * 	    Geo_data_elemt_bytes: number of bytes of each datum;
 	 * 	    data_sign: signed data or unsigned. a memeber of the struct EOS_data_info.
 	 * 	               0 for unsigned int and 1 for signed int and 2 for float or double.
 	 * OUTPUTs: returned reprojected data's pointer.
 	 * Notes: 1) Geo_lon_min, Geo_lon_max, Geo_lat_min, and Geo_lat_max are all in decimal degree, whereas
 	 * 	  SIN_x_min, SIN_x_max, SIN_y_min, and SIN_y_max are all in meters; More importantly, 
 	 * 	  these georeference ranges must be outer corner based.
 	 * 	  2) Both Geo_data and returned SIN_data must start from upleft.
 	 * 	  3) the returned pointer for *SIN_data will be calloced in this function and in 2D and thus 
 	 * 	  there is no need to calloc it outside. You'd better free it after using the data outside
 	 * 	  of the function.
 	 * 	  4) The fill values for the out of SIN map projetion: integer types are the maximum value of
 	 * 	  their types; float or double is -999999.
 	 * Date: 11-22-2019.
 	 */ 
	void *SIN_data = NULL;
	double sin_x_res, sin_y_res, geo_lon_res, geo_lat_res;
	double x_SIN, y_SIN, lon_Geo, lat_Geo;
	long col_Geo, row_Geo, col_SIN, row_SIN;
	size_t Geo_sub, SIN_sub;

	/* for the fill values of float and double out of the map projection */
	float flt_fill = -999999.;
	double dlb_fill= -999999.;

	/* for the fill values of integer out of the map projection */
        unsigned char  bits_8  = 0;

        unsigned char  ubyte_1;
        unsigned short ubyte_2;
        unsigned int   ubyte_4;
        unsigned long  ubyte_8;

        signed char  byte_1;
        signed short byte_2;
        signed int   byte_4;
        signed long  byte_8;

	/* unsigned fill value */
	ubyte_1 = ~bits_8;   /* take advantage of integer promotion for bitwise not operator */
	ubyte_2 = ~bits_8;   /* take advantage of integer promotion for bitwise not operator */
	ubyte_4 = ~bits_8;   /* take advantage of integer promotion for bitwise not operator */
	ubyte_8 = ~bits_8;   /* take advantage of integer promotion for bitwise not operator */

	/* signed fill value */
	byte_1 = 0.5*(ubyte_1 - 1);
	byte_2 = 0.5*(ubyte_2 - 1);
	byte_4 = 0.5*(ubyte_4 - 1);
	byte_8 = 0.5*(double)(ubyte_8 - 1);  /* avoid integer promotion and thus use bits_64 not bits_8 */

	/* check the validate of the input arguments */
	if(Geo_data == NULL)
	{
	   fprintf(stderr, "Geo_data don't exist, exit...\n");
	   exit(0);
	}

	if(Geo_data_elemt_bytes <= 0)
	{
	   fprintf(stderr, "Geo_data_elemt_bytes should be equal to or greater than 1, exit...\n");
	   exit(0); 
	}

	/* are the input values valid ? */
	if(Geo_cols <= 0 || Geo_rows <=0)
	{
	   fprintf(stderr, "The Geo_cols or Geo_rows are invalid as they are not positive values, exit...\n");
	   exit(0);
	}

        if(Geo_lon_min >= Geo_lon_max || Geo_lat_min >= Geo_lat_max)	
	{
	   fprintf(stderr, "The Geo_lon_min, Geo_lon_max, Geo_lat_min and Geo_lat_max are invalid as the minima values are greater than the maxima, exit...\n");
	   exit(0);
	}

	if( (Geo_lon_min < -180. || Geo_lon_min > 180.) || (Geo_lat_min < -90. || Geo_lat_min > 90.) ||
	    (Geo_lon_max < -180. || Geo_lon_max > 180.) || (Geo_lat_max < -90. || Geo_lat_max > 90.) )
	{
	   fprintf(stderr, "The range of Geo_lon_min, Geo_lon_max, Geo_lat_min, and Geo_lat_max are invalid in reproj_GEO_data_to_SIN_data_2D, exit...\n");
	   exit(0);
	}

	if(SIN_cols <= 0 || SIN_rows <=0)
	{
	   fprintf(stderr, "The SIN_cols or SIN_rows are invalid as they are not positive values, exit...\n");
	   exit(0);
	}

        if(SIN_x_min >= SIN_x_max || SIN_y_min >= SIN_y_max)	
	{
	   fprintf(stderr, "The SIN_x_min, SIN_x_max, SIN_y_min and SIN_y_max are invalid as the minima values are greater than the maxima, exit...\n");
	   exit(0);
	}

	/* get Geo and SIN resolutions */
	geo_lon_res = (Geo_lon_max - Geo_lon_min) / (double)Geo_cols;
	geo_lat_res = (Geo_lat_max - Geo_lat_min) / (double)Geo_rows;

	sin_x_res = (SIN_x_max - SIN_x_min) / (double)SIN_cols;
	sin_y_res = (SIN_y_max - SIN_y_min) / (double)SIN_rows;

	/* calloc SIN_data */
	if((SIN_data = (void*)calloc((size_t)SIN_cols * (size_t)SIN_rows, (size_t)Geo_data_elemt_bytes)) == NULL)
	{
	   fprintf(stderr, "cannot calloc SIN_data, exit...\n");
	   exit(0);
	}

	/* reproject Geo_data to SIN_data */
	for(row_SIN = 0; row_SIN < SIN_rows; row_SIN++)
	{
	    /* center y coordinate in meters */
	    y_SIN = SIN_y_max - (row_SIN + 0.5) * sin_y_res;
  
	    for(col_SIN = 0; col_SIN < SIN_cols; col_SIN++)
	    {
		/* center x coordinate in meters */
		x_SIN = SIN_x_min + (col_SIN + 0.5) * sin_x_res;

		/* get the corresponding lat and lon */
		lat_Geo = y_SIN / (double)R_earth;
		lon_Geo = x_SIN / ( (double)R_earth * cos(lat_Geo) );

		/* convert lat and lon to decimal degree */
		lat_Geo *= R2D;
		lon_Geo *= R2D;

		/* note that for SIN map projection system of MODIS/VIIRS tiles, latitude 
 		 * will never be out of the valid range but longitude will for boundary tiles */
		SIN_sub = row_SIN * SIN_cols + col_SIN;

		if(lon_Geo > 180. || lon_Geo < -180.)
		{
		   /* use the fill value for the pixel out of the map projection */
		   switch(data_sign)
		   {
		      case 0:      /* unsigned integer, maximum value is the fill value */
			memmove( (void*)(SIN_data + SIN_sub * (size_t)Geo_data_elemt_bytes),
				 (void*)&ubyte_8, (size_t)Geo_data_elemt_bytes);
			break;
		      case 1:     /* signed integer, maximum value is the fill value */
			switch(Geo_data_elemt_bytes)
			{
			   case 1:
				memmove( (void*)(SIN_data + SIN_sub * (size_t)Geo_data_elemt_bytes),
					 (void*)&byte_1, (size_t)Geo_data_elemt_bytes);
				break;
			   case 2:
				memmove( (void*)(SIN_data + SIN_sub * (size_t)Geo_data_elemt_bytes),
					 (void*)&byte_2, (size_t)Geo_data_elemt_bytes);
				break;
			   case 4:
				memmove( (void*)(SIN_data + SIN_sub * (size_t)Geo_data_elemt_bytes),
					 (void*)&byte_4, (size_t)Geo_data_elemt_bytes);
				break;
			   case 8:
				memmove( (void*)(SIN_data + SIN_sub * (size_t)Geo_data_elemt_bytes),
					 (void*)&byte_8, (size_t)Geo_data_elemt_bytes);
				break;
			   default:
				fprintf(stderr, "There is no such case %d for Geo_data_elemt_bytes, exit...\n", Geo_data_elemt_bytes);
				exit(0);
			}
			break;
		      case 2:  /* float or double, -999999. is the fill value */
			switch(Geo_data_elemt_bytes)
			{
			   case 4:
				memmove( (void*)(SIN_data + SIN_sub * (size_t)Geo_data_elemt_bytes),
					 (void*)&flt_fill, (size_t)Geo_data_elemt_bytes);
				break;
			   case 8:
				memmove( (void*)(SIN_data + SIN_sub * (size_t)Geo_data_elemt_bytes),
					 (void*)&dlb_fill, (size_t)Geo_data_elemt_bytes);
				break;
			   default:
				fprintf(stderr, "There is no such case %d for Geo_data_elemt_bytes, exit...\n", Geo_data_elemt_bytes);
				exit(0);	
			}
			break;
		      default:
			fprintf(stderr, "There is no such case %d for data_sign, exit...\n", data_sign);
			exit(0);
		   }
		}
		else
		{
		   /* note that we assume the data should be within the Geo range */
		   if(lon_Geo < Geo_lon_min || lon_Geo > Geo_lon_max ||
		      lat_Geo < Geo_lat_min || lat_Geo > Geo_lat_max)
		   {
			fprintf(stderr, "the pixel is not within the geographic data, exit...\n");
			exit(0);
		   }

		   /* get the nearest pixel in the Geographic map projection */
		   col_Geo = (long)( (lon_Geo - Geo_lon_min) / geo_lon_res );
		   row_Geo = (long)( (Geo_lat_max - lat_Geo) / geo_lat_res );

		   /* subscript for 1D memory array */ 
		   Geo_sub = row_Geo * Geo_cols + col_Geo;

		   /* copy the data value from the Geo to SIN */
		   memmove( (void*)(SIN_data + SIN_sub * (size_t)Geo_data_elemt_bytes), 
			    (void*)(Geo_data + Geo_sub * (size_t)Geo_data_elemt_bytes),
			    (size_t)Geo_data_elemt_bytes );
		}
		 
	    }
	}
	
	return SIN_data;
}

int HDFE_DEG_DMS_2_Decimal_DEG(double HDF_DMS, double* Angle_DDG)
/* written: Maosheng Zhao
 * purpose: convert HDFEOS DDDMMMSSS.SS to Decimal Degree
 * Mayh 27, 2008
 */
{
   double int_deg, fraction_min, fraction_sec;

   int_deg = (double)((int)floor(HDF_DMS*1.e-6 + 0.5));
   fraction_min= (double)(((double)(int)floor(HDF_DMS*1.e-3+0.5) - int_deg*1.e+3)/60.);
   fraction_sec= (double)((HDF_DMS - (double)(int)floor(HDF_DMS*1.e-3+0.5)*1.e+3)/(60.*60.));

   *Angle_DDG = int_deg+fraction_min+fraction_sec;

   return(0);
}

int Decimal_DEG_2_HDFE_DEG_DMS(double Angle_DDG, double* HDF_DMS)
/* written: Maosheng Zhao
 * purpose: convert Decimal degree to HDFEOS DDDMMMSSS.SS
 * May 27, 2008
 */
{
   int DDD, MMM;
   double SSS_SS;

   DDD = (int)Angle_DDG;
   MMM = (int)((Angle_DDG - (double)DDD)*60.);
   SSS_SS= (Angle_DDG - (double)DDD - (double)MMM/60.)*(60.*60.);

   *HDF_DMS=(double)DDD*1.e+6+(double)MMM*1.e+3+(double)SSS_SS;

   return(0);
}

EOS_PROJINFO* get_EOS_grid_projection_info_Malloc(char* EOS_FILE_NAME, char* GridName)
{
/* Developer: Maosheng Zhao, 9/15/2019. This is to enhance the previous 
 *    extraction_MODIS_SIN_projection_info_Malloc() developed by the same developer
 * purpose: get the map projection information for a gridname in an EOS_FILE_NAME
 * of either HDFEOS2/HDFEOS5.
 * INPUTS: EOS_FILE_NAME, GridName.
 * RETURN: pointer to a struct of HDF_EOS_PROJINFO.
 * NOTE: This is only for Grid HDF-EOS only, not point and swath.
 * for a fieldname, you may get the projection information in two steps as an example below:
 *   STEP 1: get the Gridname
 *
 *   char LC_gridname[255] = {0};
 *   EOS_PROJINFO *proj_info = NULL;
 *
 *   if((get_EOS_gridname_for_fieldname(inputs.MOD12Q1_name, "Land_Cover_Type_1", LC_gridname)) == -1)
 *   {
 *     fprintf(stderr, "cannot get_EOS_gridname_for_fieldname, exit...\n");
 *     exit(0);
 *   }
 *
 *   STEP 2: get the proj_info
 *
 *   if((proj_info = get_EOS_grid_projection_info_Malloc(inputs.MOD12Q1_name, LC_gridname)) == NULL)
 *   {
 *     fprintf(stderr, "cannot get the map projection infomation, exit...\n");
 *     exit(0);
 *   }
 *   ... ...
 *   FINALLY, free the proj_info
 *   if(proj_info)
 *      free(proj_info);
 */

  int i;                /* parameter index */
  hid_t file_id, grid_id;
  herr_t status = -1;
  grid_mapinfo *h;      /* pointer to the HDF EOS metadata structure */
  int he5_fmt;
  int EOS_ver = 0;
  double upleft_x_DDG, upleft_y_DDG, lowright_x_DDG, lowright_y_DDG;
  /* HDF interface variables */
  EOS_PROJINFO* proj_info=NULL;

  /* allocate dynamic memory to store proj_info */
  if((proj_info = (EOS_PROJINFO*)calloc(1, sizeof(EOS_PROJINFO))) == NULL)
  {
     fprintf(stderr, "cannot allocate memory for proj_info, exit...\n");
     exit(0);
  }
	
  /* we have to use EOS_version_check() not HDF_version_check() because HDFEOS' API not HDF's are used here */
  EOS_ver = EOS_version_check(EOS_FILE_NAME);

  if(EOS_ver == 2)
     he5_fmt = 0;
  else if(EOS_ver == 5)
     he5_fmt = 1;
  else
  {
     fprintf(stderr, "%s is not an HDFEOS2 or HDFEOS5, exit...\n", EOS_FILE_NAME);
     exit(1);
  }
  
 /* Open the geolocation file */
  if ((file_id = EOS_GDopen(EOS_FILE_NAME, 'R', he5_fmt)) == FAIL) {
    fprintf(stderr, "Input file %s open failure (EOS_GDopen)\n", EOS_FILE_NAME);  
    exit(0);
  }

  /* Read info for the first layer (get it from the right grid type) */
  if ((grid_id = EOS_GDattach(file_id, GridName, he5_fmt)) == FAIL) {
    fprintf(stderr, "can't attach to %s (EOS_GDattach) to GetHDFEOS!\n", GridName); 
    if (EOS_GDclose(file_id, he5_fmt) == FAIL)
      fprintf(stderr, "can't close file (EOS_GDclose)\n");
      exit(0);
  }

  /* get grid map info */
  if((status = EOS_GDgridinfo(grid_id, &proj_info->xdimsize, &proj_info->ydimsize, proj_info->upleft, proj_info->lowright, he5_fmt)) == -1)
  {
     fprintf(stderr, "cannot EOS_GDgridinfo from gridID %d, exit\n", grid_id);
     exit(1);	
  }

  if((status = EOS_GDprojinfo(grid_id, &proj_info->projcode, &proj_info->zonecode, &proj_info->spherecode, proj_info->projparm, he5_fmt)) == -1)
  {
     fprintf(stderr, "cannot EOS_GDprojinfo from gridID %d, exit\n", grid_id);
     exit(1);
  }

  /* derive upright and lowleft X and Y */
  proj_info->upright[0] = proj_info->lowright[0];
  proj_info->upright[1] = proj_info->upleft[1];
  proj_info->lowleft[0] = proj_info->upleft[0];
  proj_info->lowleft[1] = proj_info->lowright[1];

  /* derive x and y spatial resolution for non-geographic (not UTM and PGSd_SPCS but others non GCTP_GEO) */
  if(proj_info->projcode != GCTP_GEO)
  {
    /* in meters */
    proj_info->x_res = (proj_info->lowright[0] - proj_info->upleft[0])/(float64)proj_info->xdimsize;
    proj_info->y_res = (proj_info->upleft[1] - proj_info->lowright[1])/(float64)proj_info->ydimsize;

    /* derive minimum geographic map projection resolution for SIN map projection or ISIN */
    proj_info->min_geo_DD_res = ((proj_info->x_res < proj_info->y_res ? proj_info->x_res : proj_info->y_res)/proj_info->projparm[0]) * R2D;
  }
  else    /* Geographic */
  {
    /* convert to decimal degree */
    HDFE_DEG_DMS_2_Decimal_DEG(proj_info->upleft[0], &proj_info->upleft_DDG[0]);
    HDFE_DEG_DMS_2_Decimal_DEG(proj_info->upleft[1], &proj_info->upleft_DDG[1]);
    HDFE_DEG_DMS_2_Decimal_DEG(proj_info->lowright[0], &proj_info->lowright_DDG[0]);
    HDFE_DEG_DMS_2_Decimal_DEG(proj_info->lowright[1], &proj_info->lowright_DDG[1]);

    /* in decimal degree */
    proj_info->x_res = (proj_info->lowright_DDG[0] - proj_info->upleft_DDG[0])/(float64)proj_info->xdimsize;
    proj_info->y_res = (proj_info->upleft_DDG[1] - proj_info->lowright_DDG[1])/(float64)proj_info->ydimsize;

    /* derive minimum resolution for GCTP_GEO map projection */
    proj_info->min_geo_DD_res = proj_info->x_res < proj_info->y_res ? proj_info->x_res : proj_info->y_res;
  }


  if((status = EOS_GDorigininfo(grid_id, &proj_info->origincode, he5_fmt)) == -1)
  {
     fprintf(stderr, "cannot EOS_GDorigininfo from %d, exit\n", grid_id);
     /* in case some HDF missing this, we will not exit but continue */
     /*exit(1);*/
  }

  if((status = EOS_GDorigininfo(grid_id, &proj_info->origincode, he5_fmt)) == -1)
  {
     fprintf(stderr, "cannot EOS_GDorigininfo from %d, exit\n", grid_id);
     /* in case some HDF missing this, we will not exit but continue */
     /*exit(1);*/
  }

  if((status = EOS_GDpixreginfo(grid_id, &proj_info->pixregcode, he5_fmt)) == -1)
  {
     fprintf(stderr, "cannot EOS_GDpixreginfo from %d, exit\n", grid_id);
     /* in case some HDF missing this, we will not exit but continue */
     /*exit(1);*/
  }

  /* Detach from the grid and close the file */
  if (EOS_GDdetach(grid_id, he5_fmt) == FAIL) {
    fprintf(stderr, "can't detach from grid (EOS_GDdetach)!\n");  
    exit(0);
  }
  
  if (EOS_GDclose(file_id, he5_fmt) == FAIL) {
    fprintf(stderr, "can't close file (GDclose)!\n"); 
    exit(0);
  }

  return proj_info;

}

int get_tile_HV_code_from_HDFEOS_filename(char *HDFEOS_filename, int *H, int *V)
{
	/* Developer: Maosheng Zhao */
	/* INPUT:  HDFEOS_filename, with path is Ok
 	 * OUTPUTs:integer H and V values */

	char *bname = NULL;   /* for basename */
	char MODIS_tile_code[7]={0};
	char H_num_str[3]={0}, V_num_str[3]={0};
	int i, j, HV_check;

	/* get the file basename */
	bname = strrchr(HDFEOS_filename, '/');
	bname = bname ? bname + 1 : (char *)HDFEOS_filename;

	/* get H and V values */
	HV_check = 0;
	for(i = 0; i < strlen(bname); i++)
        {
            if((bname[i] == 'h' || bname[i] == 'H') && bname[i+1] >= '0' && bname[i+1] <= '9' && bname[i+2] >= '0' && bname[i+2] <= '9' && (bname[i+3] == 'v' || bname[i+3] == 'V') && bname[i+4] >= '0' && bname[i+4] <= '9' && bname[i+5] >= '0' && bname[i+5] <= '9')
            {
		HV_check = 1;
                for(j = 0; j < 6; j++)
                    MODIS_tile_code[j] = bname[i+j];

                break;
            }
        }

	if(!HV_check)
	{
	   fprintf(stderr, "The format of H and V is incorrect\n");
	   exit(0);
	}
	else  /* get H and V values */
	{
	   H_num_str[0] = MODIS_tile_code[1];
	   H_num_str[1] = MODIS_tile_code[2];

	   V_num_str[0] = MODIS_tile_code[4];
	   V_num_str[1] = MODIS_tile_code[5];

	   *H = atoi(H_num_str);
	   *V = atoi(V_num_str);
	}

	return 0;
}

int SIN_tile_UL_LR_coordinates(int H, int V, double *UL_x, double *UL_y, double *LR_x, double *LR_y)
{
	/* Developer: Maosheng Zhao */
	/* Purpose: To obtain upleft and lowerright Sinusoidal tile corner coordinates */
	/* INPUTs: H and V codes of MODIS Sinusoidal tile system
 	 * OUTPUTs: the corner X and Y values of SIN map projection
 	 */ 
	if(H > 35 || H < 0)
	{
	    fprintf(stderr, "H is out of range !!!\n");
	    exit(0);
	}
	if(V > 17 || V < 0)
	{
	    fprintf(stderr, "V is out of range !!!\n");
	    exit(0);
	}

	/* adjust H and V to get UL x/y and LR x/y */
	if(H <= 17)
	    H += -18;
	else
	    H -= 18;

	if(V <= 8)
	    V = 9 - V ;
	else
	    V = (V - 9) * (-1);
	
	*UL_x = (double)H * DIS;
	*UL_y = (double)V * DIS;

	*LR_x = (double)(H + 1) * DIS;
	*LR_y = (double)(V - 1) * DIS;

	return(0);
}

int Get_GEO_range_thr_SIN_XY_range(float64 earth_radius, float64 sin_x_min, float64 sin_x_max, float64 sin_y_min, float64 sin_y_max, float64* lon_min, float64* lon_max, float64* lat_min, float64* lat_max)
/* written: Maosheng Zhao
 * date: Mar 21, 2009
 * method: based on my in-depth understanding of Sinusoidal map projection, and described in this function
 * Note that input earth radius and x and y should be in meters, output lat and lon must be in decimal degree
 * This function can be used for any case with Sinusoidal map projection, not limited to 10-degree MODIS tiles
 * The output lon may be out of the normal range, and we have to correct the issue by constrain it within the
 * normal extreme values, i.e., -180 to 180.
 * INPUTs: earth_radius, sin_x_min, sin_x_max, sin_y_min, sin_y_max;
 * OUTPUTs: *lon_max, *lat_min, *lat_max;
 */
{
	int i;
	float64 lon[6]; /* lon[0] upleft, lon[1] upright, lon[2] lowright lon[3] lowleft lon[4] equator left, lon[5] equator right */
	if(sin_y_min * sin_y_max >= 0.)  /* in the same hemisphere, no need to account for equator */
	{
	    *lat_min = sin_y_min/earth_radius;
	    *lat_max = sin_y_max/earth_radius;

	    /* upleft, sin_x_min and sin_y_max */
	    lon[0] = sin_x_min/(earth_radius * cos(*lat_max));

	    /* upright, sin_x_max and sin_y_max */
	    lon[1] = sin_x_max/(earth_radius * cos(*lat_max));

	    /* lowright, sin_x_max and sin_y_min */
	    lon[2] = sin_x_max/(earth_radius * cos(*lat_min));

	    /* lowleft, sin_x_min, sin_y_min */
	    lon[3] = sin_x_min/(earth_radius * cos(*lat_min));
	    
	}
	else  /* in different hemispheres, i.e., the north and the south, equation has the longest sin_x */
	{
	    *lat_min = sin_y_min/earth_radius;
	    *lat_max = sin_y_max/earth_radius;

	    /* upleft, sin_x_min and sin_y_max */
	    lon[0] = sin_x_min/(earth_radius * cos(*lat_max));

	    /* upright, sin_x_max and sin_y_max */
	    lon[1] = sin_x_max/(earth_radius * cos(*lat_max));

	    /* lowright, sin_x_max and sin_y_min */
	    lon[2] = sin_x_max/(earth_radius * cos(*lat_min));

	    /* lowleft, sin_x_min, sin_y_min */
	    lon[3] = sin_x_min/(earth_radius * cos(*lat_min));

	    /* equtorial left, sin_x_min and lat = 0 */
	    lon[4] = sin_x_min/earth_radius;

	    /* equtorial right, sin_x_max and lat = 0 */
	    lon[5] = sin_x_max/earth_radius;
	}

	/* obtain the extreme longitude */
	*lon_min = PI_val;
	*lon_max = PI_val * -1.;

	if(sin_y_min * sin_y_max >= 0.)
	    for(i=0;i<4;i++)
	    {
	    	if(lon[i] < *lon_min) *lon_min = lon[i];
		if(lon[i] > *lon_max) *lon_max = lon[i];
	    }
	else
	    for(i=0;i<6;i++)
	    {
	    	if(lon[i] < *lon_min) *lon_min = lon[i];
		if(lon[i] > *lon_max) *lon_max = lon[i];
	    }

	/* convert radius to degree */
	*lon_min *= R2D;
	*lon_max *= R2D;
	*lat_min *= R2D;
	*lat_max *= R2D;

	/* constrain distorted unrealistic logitude caused by SIN map projection */
	if(*lon_min > 180. ) *lon_min =180.;
	if(*lon_min < -180.) *lon_min = -180.;

	if(*lon_max > 180.)  *lon_max = 180.;
	if(*lon_max < -180.) *lon_max = -180.;

	return(0);
}

int Get_GEO_subset_geolocation(double glb_lon_min, double glb_lon_max, double glb_lat_min, double glb_lat_max, long glb_cols, long glb_rows, double *sub_lon_min, double *sub_lon_max, double *sub_lat_min, double *sub_lat_max, long *col_start, long *col_end, long *row_start, long *row_end, long *col_len, long *row_len)
{
	/* developer: Maosheng Zhao
 	 * date: 11/22/2019
 	 * INPUTs: glb_lon_min, glb_lon_max, glb_lat_min, glb_lat_max, glb_cols, glb_rows, *sub_lon_min, *sub_lon_max, *sub_lat_min, *sub_lat_max in decimal degree.
 	 * OUTPUTs: *sub_lon_min, *sub_lon_max, *sub_lat_min, *sub_lat_max, *col_start, *col_end, *row_start, *row_end.
 	 *    You can see that *sub_lon_min, *sub_lon_max, *sub_lat_min, and *sub_lat_max are both INPUT and OUTPUT 
 	 *    since they are slightly corrected using the global georeference system.
 	 * Notes: We assume that 1) The map projection is Geographic and both input and output for lon and lat should be in decimal degree;
 	 *  2) the coordinates are corner not center and the data starts from upleft; and
 	 *  3) the subset region is within the global region.
 	 *  4) the col_start, col_start, col_end, row_start and row_end are zero-based.
 	 * purpose: to correct the specified longitudes and latitudes of the subset based on the global georeferences.
 	 *  Note that the global georeferences may not be global but can be regional. This function is to 
 	 *  avoid the issue of the shift of the location of subset using nearest resample method and to use
 	 *  the original global data georeference system to correct the specified geographic range specified
 	 *  by users. At the same time to get the 0-based start and end column and rows of the global data
 	 *  for the specified subset. 
 	 *  In one words, the user-specified subset region *sub_lon_min, *sub_lon_max, *sub_lat_min, *sub_lat_max 
 	 *  are corrected to avoid the shift of pixel and also get the start and end of the column and row for the seubset.
 	 */  

	double lon_res, lat_res;

	/* are the input values valid ? */
	if(glb_cols <= 0 || glb_rows <=0)
	{
	   fprintf(stderr, "The glb_cols or glb_rows are invalid as they are not positive values, exit...\n");
	   exit(0);
	}

        if(glb_lon_min >= glb_lon_max || glb_lat_min >= glb_lat_max)	
	{
	   fprintf(stderr, "The glb_lon_min, glb_lon_max, glb_lat_min and glb_lat_max are invalid as the minima values are greater than the maxima, exit...\n");
	   exit(0);
	}

        if(*sub_lon_min >= *sub_lon_max || *sub_lat_min >= *sub_lat_max)	
	{
	   fprintf(stderr, "The sub_lon_min, sub_lon_max, sub_lat_min and sub_lat_max are invalid as the minima values are greater than the maxima, exit...\n");
	   exit(0);
	}

	if( (glb_lon_min < -180. || glb_lon_min > 180.) || (glb_lat_min < -90. || glb_lat_min > 90.) ||
	    (glb_lon_max < -180. || glb_lon_max > 180.) || (glb_lat_max < -90. || glb_lat_max > 90.) )
	{
	   fprintf(stderr, "The range of glb_lon_min, glb_lon_max, glb_lat_min, and glb_lat_max are invalid, exit...\n");
	   exit(0);
	}

	if( (*sub_lon_min < -180. || *sub_lon_min > 180.) || (*sub_lat_min < -90. || *sub_lat_min > 90.) ||
	    (*sub_lon_max < -180. || *sub_lon_max > 180.) || (*sub_lat_max < -90. || *sub_lat_max > 90.) )
	{
	   fprintf(stderr, "The range of sub_lon_min, sub_lon_max, sub_lat_min, and sub_lat_max are invalid, exit...\n");
	   exit(0);
	}

	/* check if the georeference range of the subset are within the global data */
	/* solve the float point issue */
	if(*sub_lon_min <= -180.) *sub_lon_min = -179.9999999;
	if(*sub_lon_max >=  180.) *sub_lon_max =  179.9999999;

	if(*sub_lat_min <=  -90.) *sub_lat_min = -89.9999999;
	if(*sub_lat_max >=   90.) *sub_lat_max =  89.9999999;

	if(*sub_lon_min < glb_lon_min || *sub_lon_max > glb_lon_max ||
	   *sub_lat_min < glb_lat_min || *sub_lat_max > glb_lat_max)
	{
	   fprintf(stderr, "The specified range of the subset georeference range is beyond the global georeference range, exit...\n");
	   exit(0);
	}
	
	/* get the resolution of the data */
	lon_res = (glb_lon_max - glb_lon_min) / (double)glb_cols;
	lat_res = (glb_lat_max - glb_lat_min) / (double)glb_rows;

	/* get the zero-based start and end in column and row */
	*col_start = (long)( (*sub_lon_min - glb_lon_min) / lon_res );
	*col_end   = (long)( (*sub_lon_max - glb_lon_min) / lon_res );

	*row_start = (long)( (glb_lat_max - *sub_lat_max) / lat_res );
	*row_end   = (long)( (glb_lat_max - *sub_lat_min) / lat_res );

	if(*col_start < 0 || *col_end < 0 || *row_start < 0 || *row_end < 0)
	{
	   fprintf(stderr, "Something is wrong for the start and end, exit...\n");
	   exit(0);
	}

	if(*col_end < *col_start || *row_end < *row_start)
	{
	   fprintf(stderr, "col_end should not be greater than col_start or row_end should not be greater than row_start.\n");
	   exit(0);
	}

	*col_len   = (*col_end - *col_start) + 1;
	*row_len   = (*row_end - *row_start) + 1;

	/* correct the input sub_lon_min, sub_lon_max, sub_lat_min, and sub_lat_max
 	 * to be consistent with the global data georeference frame. Again, note that
 	 * the coordinates are corner based */
	*sub_lon_min = glb_lon_min + (double)*col_start     * lon_res;
	*sub_lon_max = glb_lon_min + (double)(*col_end + 1) * lon_res;

	*sub_lat_max = glb_lat_max - (double)*row_start     * lat_res;
	*sub_lat_min = glb_lat_max - (double)(*row_end + 1) * lat_res;
}

int get_grid_name(char *ref_fname, char *grid_name)
{
  /* MSZ 2/9/2018, add the case for HDFEOS5 */
  /* MSZ 2/10/2018, move it from cp_hdfeos.c to here because I found out that several tools
   * use the function so I treat it as a library function here */
  int st, len, p1;
  int32 nsds, nattr;
  int32 sd_id, sds_id;
  int32 rank, dim_size[4];
  int32 data_type, dim_id;
  int32 d_size, d_type, d_nattr;
  char dim_name[(int)(0.5*H4_MAX_NC_NAME)];
  char sds_name[H4_MAX_NC_NAME];
  int HDF_ver = 0;

  /* we have to EOS_version_check() not HDF_version_check() because HDF' API not HDFEOS's are used here */
  HDF_ver = HDF_version_check(ref_fname);

  if(!HDF_ver)
  {
    printf("%s is neither HDF4 nor HDF5, exit...\n", ref_fname);
    exit(0);
  }

  st = 1;
  if(HDF_ver == 4)
  {
    if ((sd_id = SDstart(ref_fname, DFACC_READ)) != FAIL)
    {
      if (SDfileinfo(sd_id, &nsds, &nattr) != FAIL)
      {
        if ((sds_id = SDselect(sd_id, 0)) != FAIL)
        {
	  if (SDgetinfo(sds_id, sds_name, &rank, dim_size, &data_type, &nattr) != FAIL)     
	  {
	    if ((dim_id = SDgetdimid(sds_id, 0)) != FAIL)
	    {
	      SDdiminfo(dim_id, dim_name, &d_size, &d_type, &d_nattr); 
	      len = (int)strlen(dim_name);
	      if ((p1 = charpos(dim_name, ':', 0)) != -1)
	        strmid(dim_name, p1+1, len-p1-1, grid_name);
	      else st = -1;
	    }
	  }
          SDendaccess(sds_id);
        }
      }
      SDend(sd_id);
    }
  }
  else if(HDF_ver == 5)  /* MSZ 2/9/2018, add the case for HDFEOS5 */
  {
    int nsds = 0;
    char **sds_names = NULL;

    if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, H4_MAX_NC_NAME, sizeof(char))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for sds_names in get_grid_name(), exit...\n");
      exit(1);
    }
    else
    {
      /* get all numerical SDSs' names in standard HDFEOS5 */
      nsds = get_H5_full_sds_names(ref_fname, sds_names);

      if(nsds)
      { 
        char *start_p = NULL;
        char *shift_p = NULL;
        char *end_p = NULL;
        size_t len_grid = 0;

        /* just use the first one */
        start_p = strstr(sds_names[0], "/HDFEOS/GRIDS/");
        shift_p = start_p + strlen("/HDFEOS/GRIDS/");
        end_p = strstr(sds_names[0], "/Data Fields/");

        if(start_p == NULL || end_p == NULL)
	    st = -1;
        else
        {
	    st = 1;
            len_grid = end_p - shift_p + 1;   /* add extra 1 byte for the terminator */
            strncpy(grid_name, shift_p, len_grid);
            grid_name[len_grid - 1] = '\0';   /* make terminator to be terminator */
        }
      }
      else
	st = -1;

      /* free 2D */
      if(sds_names) Free2D((void**)sds_names);
    } /* end else */
  } /*end else if(HDF_ver == 5) */

  return st;
}
