/*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
/* Developer: Maosheng Zhao, SSAI, 2/9/2018 */
/* wrapped functions for Grid HDFEOS2.x and HDFEOS5.x. using HDFEOS5 data type system */
/* !!! note that you must be very very cautious the argument of field_data_type for
 * EOS_GDdeffield() because HDF4 and HDFEOS2 use the same data type system but not
 * HDF5 and HDFEOS5. convert_datatype_HDF4toHDFEOS5() can be used to convert
 * data type system before calling this function. Plz check how it is used in the tools. */

#include "mfhdf.h"          /* HDF4 */
#include "HdfEosDef.h"      /* HDFEOS2 */
#include "hdf5.h"           /* HDF5 */
#include "hdf5_hl.h"        /* HDF5 H5LT, H5 light */
#include "HE5_HdfEosDef.h"  /* HDFEOS5 */
#define MAX_LDOPE_DIMS  4   /* Note that MAX_VAR_DIMS is 32 */
#define MAX_NUM_SDS 200
#define MAX_NUM_GRID 20     /* MSZ added on 4/04/2019 */
#define R_earth 6371007.181 /* earth radius used by MODIS/VIIRS data */
#define PI_val  3.141592653589793238  /* follow the GCTP3 */
#define DIS (double)((10./180.)*(double)PI_val*(double)R_earth)       /* because MODIS HDF-EOS is 10 degree tiles */
#define R2D (180./(double)PI_val)
#define D2R ((double)PI_val/180.)

/* MSZ 2/9/2018, update int32 to int and long for HDFEOS5 */
/* MSZ 2/10/2018, move it from meta.h to here */
typedef struct {
  long ns, nl;    /* xdim, ydim */
  float64 ulCornerGrid[2], lrCornerGrid[2];
  int projCode;
  int sphereCode;
  float64 projParameters[13];
  int zoneCode;
  int originCode;
} grid_mapinfo;

/* 9/15/2019, for HDFEOS map projection to be compatible with MOD16 PGE code */
/* similar to the above but have more items */
typedef struct hdf_eos_projinfo
{
    char gridname[H4_MAX_NC_NAME];
    int projcode;
    int zonecode;
    int spherecode;
    double projparm[13];
    int origincode;
    int pixregcode;
    long xdimsize;
    long ydimsize;
    double upleft[2];
    double lowright[2];
    double upleft_DDG[2];   /* effective for GCTP_GEO in decimal degree */
    double lowright_DDG[2]; /* effective for GCTP_GEO in decimal degree */
    double upright[2];      /* derived in this code */
    double lowleft[2];      /* deribed in this code */
    double x_res;           /* x axis resolution derived in this code. in meters for non-GCTP_GEO and decimal degree for GCTP_GEO */
    double y_res;           /* y axis resolution derived in this code. in meters for non-GCTP_GEO and decimal degree for GCTP_GEO */
    double min_geo_DD_res;  /* minimum corresponding geographic projection resolution */
}
EOS_PROJINFO;

/* MSZ 8/16/2019, add the below structs so that eos_sds_rw can deal with multiple grids in an HDFEOS file */
typedef struct
{
    int num_fields;
    char **fieldnames;   /* Calloc2D [num_fields][H4_MAX_NC_NAME] for a given grid */
}
EOS_Grid_fields;      /* for a given Grid */

typedef struct
{
    int num_grids;               /* number of all grids in an HDFEOS */
    char **gridnames;            /* all gridnames Calloc2D[num_grids][H4_MAX_NC_NAME]*/
    grid_mapinfo *proj_info_vector;    /* vector, corresponding proj_info_vector, georeference info. and XDim, YDim, etc for each different gridname */

    EOS_Grid_fields *fields_names_struct_vector; /* vector, corresponding all fields */
}
EOS_Grids_info;    /* for all Grids in an HDFEOS */

/* MSZ 10/11/2019, add the below for a general struct for HDFEOS field data information */
typedef struct EOS_field_data_info
{
    char field_name[MAX_NC_NAME];
    char EOS_file_name[1024];        /* field_name's filename */
    char gridname[MAX_NC_NAME];      /* field_name's gridname */
    int he5_fmt;                     /* format of the original data field. 0 for HDFEOS2 and 1 for HDFEOS5 */
    int rank;                        /* rank is for the length of the dims */
    hsize_t *dims;                   /* dims has rank length */
    int data_type_HDF4;
    hid_t data_type_HDFEOS5;         /* note that unlike HDFEOS2 and HDF4, HDFEOS5 data type code is different from HDF5 */
    int data_sign;                   /* 0 for unsigned int and 1 for signed int and 2 for float or double. using function get_EOS_datatype_sign() */
    int size_data;                   /* byte size or number of byte of an element datum in the data array. using function get_EOS_datatype_size() */
    size_t numbers_data;             /* the total number of data is the data array flattened */
}EOS_data_info;

hid_t EOS_GDopen(char *hdf_fname, char open_t, int he5_fmt);
hid_t EOS_GDcreate(hid_t eos_gdfID, char *gridname, long xdim, long ydim, double upleft[], double lowright[], int he5_fmt);
hid_t EOS_GDattach(hid_t eos_gdfID, char *gridname, int he5_fmt);
herr_t EOS_GDdetach(hid_t eos_gridID, int he5_fmt);
herr_t EOS_GDclose(hid_t eos_gdfID, int he5_fmt);

herr_t EOS_GDdefproj(hid_t eos_gridID, int projcode, int zonecode, int spherecode, double projparm[], int he5_fmt);
herr_t EOS_GDprojinfo(hid_t eos_gridID, int *projcode, int *zonecode, int *spherecode, double projparm[], int he5_fmt);
herr_t EOS_GDdeforigin(hid_t eos_gridID, int origincode, int he5_fmt);
herr_t EOS_GDorigininfo(hid_t eos_gridID, int *origincode, int he5_fmt);
herr_t EOS_GDdefpixreg(hid_t eos_gridID, int pixregcode, int he5_fmt);
herr_t EOS_GDpixreginfo(hid_t eos_gridID, int *pixregcode, int he5_fmt);
herr_t EOS_GDdefdim(hid_t eos_gridID, char *dimname, hsize_t dim, int he5_fmt);
herr_t EOS_GDdeffield(hid_t eos_gridID, char *fieldname, char *dimlist, hid_t field_data_type, int he5_fmt);
herr_t EOS_GDreadfield(hid_t eos_gridID, char *fieldname, hssize_t start[], hsize_t edge[], void *data, int he5_fmt);
herr_t EOS_GDwritefield(hid_t eos_gridID, char *fieldname, hssize_t start[], hsize_t edge[], void *data, int he5_fmt);
herr_t EOS_GDgridinfo(hid_t eos_gridID, long *xdimsize, long *ydimsize, double upleft[], double lowright[], int he5_fmt);
long EOS_GDinqgrid(char *filename, char *gridlist, long *strbufsize, int he5_fmt);

/* MSZ 8/14/2019, added and tested the below five EOS functions */ 
int EOS_version_check(const char *HDF_name);
herr_t EOS_GDgridmapinfo(hid_t eos_gridID, grid_mapinfo* mapinfo, int he5_fmt);
long EOS_GDnentries(hid_t eos_gridID, int entrycode, long *strbufsize, int he5_fmt);
int  EOS_GDinqfields(hid_t eos_gridID, char *fieldlist, int rank[], hid_t ntype[], int he5_fmt);
herr_t EOS_GDfieldinfo(hid_t eos_gridID, char *fieldname, int *rank, hsize_t dims[], hid_t *ntype, char *dimlist, int he5_fmt);

int convert_datatype_HDF4toHDFEOS5(int32 H4_type, hid_t *HE5_type);
int convert_datatype_HDFEOS5toHDF4(hid_t HE5_type, int32 *HDF4_type);
int get_grid_name(char *ref_fname, char *grid_name);

/* MSZ 8/14/2019, updated GetHDFEOS so that the previous meta_hdfeos_t changed to grid_mapinfo */
int GetHDFEOS(char *file_name, char *HDFEOSgridname, grid_mapinfo *hdfeos);

/* MSZ 8/16/2019, added the below two functions */
int get_EOS_all_grids_info(char *HE_name, EOS_Grids_info *Grids_info);

int split_string_with_delimit(char *long_string, size_t long_str_len, char delimit,  char ***sub_strs, size_t *num_sub_strs, int trim_spaces);

/* MSZ 9/05/2019, add the function to get the gridname for a fieldname after calling get_EOS_all_grids_info() */
int get_EOS_gridname_for_fieldname(char *EOS_FILE_NAME, char *fieldname, char *gridname);
int free_EOS_all_grids_info_members(EOS_Grids_info* Grids_info);
int get_EOS_datatype_size(hid_t dtype_id, int he5_fmt);
/* MSZ 11/25/2019. add the function for a general function get_EOS_field_data_info() and reproj_GEO_data_to_SIN_data_2D() */
int get_EOS_datatype_sign(int32 H4_type, int *data_sign);

/* MSZ 9/09/2019, added the function to read the entire fieldname */
EOS_data_info* get_EOS_field_data_info(char *EOS_FILE_NAME, char *EOS_dataset_name);
void *EOS_DUMP_BIN_MALLOC(char *EOS_FILE_NAME, char *EOS_dataset_name, EOS_data_info *data_info);

/* MSZ 10/13/2019, added the below function for HDFEOS5 VNP16/VNP17 */
void *Get_EOS_data_2D_lines(char *EOS_FILE_NAME, char *EOS_dataset_name, EOS_data_info *data_info, int row, int rows);

/* MSZ 9/15/2019, added this function for PGE117/118/119 revision */
int HDFE_DEG_DMS_2_Decimal_DEG(double HDF_DMS, double* Angle_DDG);
int Decimal_DEG_2_HDFE_DEG_DMS(double Angle_DDG, double* HDF_DMS);
EOS_PROJINFO* get_EOS_grid_projection_info_Malloc(char* EOS_FILE_NAME, char* GridName);

/* MSZ 11/22/2019, added the below function for get the subset data from the global geographicl CMG data */
int get_tile_HV_code_from_HDFEOS_filename(char *HDFEOS_filename, int *H, int *V);
int SIN_tile_UL_LR_coordinates(int H, int V, double *UL_x, double *UL_y, double *LR_x, double *LR_y);
int Get_GEO_range_thr_SIN_XY_range(float64 earth_radius, float64 sin_x_min, float64 sin_x_max, float64 sin_y_min, float64 sin_y_max, float64* lon_min, float64* lon_max, float64* lat_min, float64* lat_max);
int Get_GEO_subset_geolocation(double glb_lon_min, double glb_lon_max, double glb_lat_min, double glb_lat_max, long glb_cols, long glb_rows, double *sub_lon_min, double *sub_lon_max, double *sub_lat_min, double *sub_lat_max, long *col_start, long *col_end, long *row_start, long *row_end, long *col_len, long *row_len);
void *Get_EOS_data_2D_subset(char *EOS_FILE_NAME, char *EOS_dataset_name, EOS_data_info *data_info, long col_start, long row_start, long cols, long rows);
void *reproj_GEO_data_to_SIN_data_2D(void *Geo_data, int Geo_data_elemt_bytes, int data_sign, double Geo_lon_min, double Geo_lon_max, double Geo_lat_min, double Geo_lat_max, long Geo_cols, long Geo_rows, double SIN_x_min, double SIN_x_max, double SIN_y_min, double SIN_y_max, long SIN_cols, long SIN_rows);
