/****************************************************************************
!C

!File: meta.c

!Description:
  Contains routines for reading metadata.

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original March 1998. Version 1.0
  Modified June 1998. Version 1.1 (Updated to support batch processing. None
	of the routines abruptly exit, instead returns to calling routines
	with error messages.
  Modified June 2011. by Xiaoping Zhang, Sigma space corp.
 	added the function of get_attr_npp_sel_metadata to meta value for NPP meta attribute	
  Modified August 2014
  	changed attr_name string length from 40 to 100 in copy_metadata.		
  Modified Jan 2018. by Maosheng Zhao, SSAI
	added five functions:
	1) HDF5_read_one_dataset_STRING(hid_t h5_file_id, const char *h5_dataset_name, void** rdata, size_t *len_strs)
	2) void copy_H5_all_metedata_strings_to_H4(int32 H5_sd_id, int32 H4_sd_id)
	3) void H5_print_all_attr_groups(int32 H5_sd_id, const char* key_str_control)
	4) int HDF5_group_attrs(hid_t h5_file_id, const char *h5_group_name, const char* key_str_control)
	5) herr_t HDF5_attr_info(hid_t loc_id, const char *name, const H5A_info_t *ainfo, void *opdata)
	6) modified copy_metadata() so that HDF5 dataset object with H5T_STRING can be copied to HDF4.
	 HDF5 is unlike HDF4, which treats metadata as DATASETs with datatype_class H5T_STRING
	 except other attributs under the root and other groups.
	7) add function H5_get_OBJ_one_attrinfo() to read each attribute for an object. It is applied
	  to get each attribute of a dataset of HDF5 in the QA tool read_sds_attributes.c
  Modified Apr, 2019. by Maosheng Zhao, SSAI
        1) modified get_all_metadata() in serveral places and major change it that replacing strpos() with 
           strpos_wb(). Also update the else in the function, too.

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "meta.h"
#include "hdf5.h"
#include "mfhdf.h"
#include "qa_tool.h"
#include "str_op.h"
#include "sds_rw.h"

int GetNumberDays(int year, int month);

char *get_attr_metadata(char *hdf_fname, char *meta_gname)
/**********************************************************************************
	Get core or archive metadata from the hdf file hdf_fname.
	On success the functions returns a pointer to the character string 
	containing the metadata else returns a NULL pointer on failure
Maosheng Zhao. SSAI. modified so that for metedata if HDF5 in DATASET in H5T_string
	meta_gname can get the name of the DATASET.
***********************************************************************************/
{
    int status = 1;
    int32 sd_id;
    int32 attr_cnt;
    int32 attr_size;
    int32 attr_type;
    int32 attr_index;
    char attr_name[20];
    char *meta_attr_buf;
    
    /* HDF5 addition */
    H5_ATTR h5attr;
    int HDF_ver = 0;
    int H5IN;
    size_t len_strs;

    hid_t           h5_file_id;
    herr_t          h5_err;
    hsize_t i;
    int meta_sub;
    H5_OBJS_TYPES *h5_obj_types = NULL;

    /* MSZ 1/29/2018 HDF5 or HDF4 */
    HDF_ver = HDF_version_check(hdf_fname);

    if(HDF_ver == 4)
        H5IN = 0;
    else if(HDF_ver == 5)
        H5IN = 1;
    else
    {
        printf("%s is neither HDF4 nor HDF5, exit...\n", hdf_fname);
        exit(0);
    }

    if (H5IN)
    {
        if ((h5_file_id = H5Fopen(hdf_fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
        {
          fprintf(stderr, "Cannot open the input HDF file %s\n", hdf_fname);
          status = -1;
        }
        else
        {
            h5attr.rdata = NULL;
            strcpy(h5attr.name,meta_gname);

            /* MSZ 1/30/2018, note that the below commneted out method was wrong for HDF5 because HDF5 treats metadata as DATASET with H5T_STRING,
 	     * which is unlike HDF4 and HDF4 treat meta data as file attributes */
	    /*
            H5Aiterate2(sd_id,H5_INDEX_CRT_ORDER,H5_ITER_INC,NULL,(H5A_operator2_t)H5_attr_read,(void *)&h5attr);
            if ((meta_attr_buf = (char *)calloc(strlen(h5attr.rdata[0]), sizeof(char))) == NULL)
            {
                fprintf(stderr, "Cannot allocate memory for metadata in get_attr_metadata\n");
                status = -1;
            }
            else
                strcpy(meta_attr_buf, h5attr.rdata[0]);
	    */

	    /* MSZ 1/30/2018, add the below for the new method to get the metadata */
	    /* I didn't use sds_get_info or open_sds is that their related wrapped functions
 	     * just deal with DATASET of H5T_INTEGER or H5T_FLOAT not H5T_STRING */
	    h5_obj_types = (H5_OBJS_TYPES*)calloc(1, sizeof(H5_OBJS_TYPES));

	    /* get objs_links */
	    h5_err = H5Fget_objs_links(h5_file_id, h5_obj_types);

	    /* check if there is meta_gname */
	    meta_sub = -1;
	    for(i = 0; i < h5_obj_types->num_objs_links; i++)
	      if(h5_obj_types->obj_type[i] == H5O_TYPE_DATASET)
		if(strcasestr(h5_obj_types->obj_name[i], meta_gname) != '\0')
		{
		   meta_sub = i;
		   break;
		}

	    if(meta_sub == -1)
	    {
		fprintf(stderr, "There is no metadata %s in %s\n", meta_gname, hdf_fname);
		status = -1;
	    }
	    else
		HDF5_read_one_dataset_STRING(h5_file_id, h5_obj_types->obj_name[meta_sub], (void**)&meta_attr_buf, &len_strs);

	    /* close HDF5 */
	    h5_err = H5Fclose (h5_file_id);

	    for(i = 0; i < h5_obj_types->num_objs_links; i++)
	    if(h5_obj_types->obj_name[i]) free(h5_obj_types->obj_name[i]);

	    if(h5_obj_types->obj_name) free(h5_obj_types->obj_name);
	    if(h5_obj_types->obj_type) free(h5_obj_types->obj_type);
	    if(h5_obj_types)free(h5_obj_types);
        }
    }
    else
    {
        if ((sd_id = SDstart(hdf_fname, DFACC_READ)) == FAIL)
        {
          fprintf(stderr, "Cannot open the input HDF file %s\n", hdf_fname);
          status = -1;
        }
        else
        {
            if ((attr_index = SDfindattr(sd_id, meta_gname)) == FAIL)
            {
              fprintf(stderr, "Cannot find %s in %s\n", meta_gname, hdf_fname);
              status = -1;
            }
            else
            {
              if (SDattrinfo(sd_id, attr_index, attr_name, &attr_type, &attr_cnt) == FAIL)
              {
                fprintf(stderr, "Cannot read %s in %s\n", meta_gname, hdf_fname);
                status = -1;
              }
              else
              {
                attr_size = DFKNTsize(attr_type) * attr_cnt + 1;
                if ((meta_attr_buf = (char *)calloc(attr_size, sizeof(char))) == NULL)
                {
                  fprintf(stderr, "Cannot allocate memory for metadata in get_attr_metadata\n");
                  status = -1;
                }
                else if (SDreadattr(sd_id, attr_index, (VOIDP)meta_attr_buf) == FAIL)
                {
                  fprintf(stderr, "Cannot read %s from %s\n", meta_gname, hdf_fname);
              free(meta_attr_buf);
                  status = -1;
                }
              }
            }
            SDend(sd_id);
        }
    }
    if (status == -1) return (char *)NULL;
    else return meta_attr_buf;
}


char *get_attr_metadata_byID(int H5IN, int32 sd_id, char *meta_gname)
/**********************************************************************************
	Get core or archive metadata from the hdf file hdf_fname.
	On success the functions returns a pointer to the character string 
	containing the metadata else returns a NULL pointer on failure
***********************************************************************************/
{
    int status = 1;
    int32 attr_cnt;
    int32 attr_size;
    int32 attr_type;
    int32 attr_index;
    char attr_name[20];
    char *meta_attr_buf;
  
    /* HDF5 addition */
    H5_ATTR h5attr;
    
    if (sd_id == -1)
    {
        fprintf(stderr, "Invalid SD ID: %d\n", sd_id);
        status = -1;
    }
    else
    {
        if (H5IN)
        {
            h5attr.rdata = NULL;
            strcpy(h5attr.name,meta_gname);
            H5Aiterate2(sd_id,H5_INDEX_CRT_ORDER,H5_ITER_INC,NULL,(H5A_operator2_t)H5_attr_read,(void *)&h5attr);
           if ((meta_attr_buf = (char *)calloc(strlen(h5attr.rdata[0]), sizeof(char))) == NULL)
            {
                fprintf(stderr, "Cannot allocate memory for metadata in get_attr_metadata\n");
                status = -1;
            }
            else
                strcpy(meta_attr_buf, h5attr.rdata[0]);
        }
        else
        {
            if ((attr_index = SDfindattr(sd_id, meta_gname)) == FAIL)
            {
              fprintf(stderr, "Failure to fine attr_index from %s \n", meta_gname);
              status = -1;
            }
            else
            {
              if (SDattrinfo(sd_id, attr_index, attr_name, &attr_type, &attr_cnt) == FAIL)
              {
                fprintf(stderr, "Cannot read %s\n", meta_gname);
                status = -1;
              }
              else
              {
                attr_size = DFKNTsize(attr_type) * attr_cnt + 1;
                if ((meta_attr_buf = (char *)calloc(attr_size, sizeof(char))) == NULL)
                {
                  fprintf(stderr, "Cannot allocate memory for metadata in get_attr_metadata\n");
                  status = -1;
                }
                else if (SDreadattr(sd_id, attr_index, (VOIDP)meta_attr_buf) == FAIL)
                {
                  fprintf(stderr, "Cannot read %s\n", meta_gname);
              free(meta_attr_buf);
                  status = -1;
                }
              }
            }
            /*    SDend(sd_id);*/
        }
    }
    if (status == -1) return (char *)NULL;
    else return meta_attr_buf;
}

void get_all_metadata(char *meta_str, char **meta_name, char **meta_val, int *meta_cnt)
/***************************************************************************************
	Read all individual metadata names into the character array meta_name and the 
	corresponding values into the character array meta_val from the single character 
	string meta_str containing all the core or archive metadata. meta_cnt contains
	the number of metadata in meta_str.
****************************************************************************************/
{
  int i, cnt;
  int len, len_t;
  int pos22, pos33;
  int pos1, pos2, pos3;

  /* MSZ 4/10/2019 */
  char OBJ_name[] = "OBJECT";
  char NUM_VAL_name[] = "NUM_VAL";
  char CLASS_name[] = "CLASS";

  cnt = *meta_cnt;
  pos3 = 0;
  len = (int)strlen(meta_str);
  while (pos3 < len)
  {
    /* MSZ 4/12/2019, replace strpos with strpos_wb */
    pos1 = strpos_wb(meta_str, OBJ_name, pos3);
    if (pos1 != -1)
    {
      pos2 = charpos(meta_str, '=', pos1) + 2;
      pos3 = charpos(meta_str, '\n', pos2);
      i = pos3 + 1;
      while (meta_str[i] == ' ') i++;
     
      /* MSZ 4/10/2019, using strstr makes more sense */ 
      /*if (meta_str[i] == 'N')*/
      if(strstr((char*)(meta_str + i), NUM_VAL_name))
      {
        strmid(meta_str, pos2, pos3-pos2, meta_name[cnt]);
        pos3 = i;
	/* MSZ 4/12/2019, replace strpos with strpos_wb */
        pos1 = strpos_wb(meta_str, "VALUE", pos3);
	if (pos1 != -1)
	{
          pos2 = charpos(meta_str, '=', pos1) + 2;
          if (meta_str[pos2] == '(')
            pos3 = charpos(meta_str, ')', pos2) + 1;
          else
            pos3 = charpos(meta_str, '\n', pos2);
          strmid(meta_str, pos2, pos3-pos2, meta_val[cnt]);
	  if (meta_str[pos3-1] == ')')
	    rm_ln_in_str(meta_val[cnt]);
          if (strcmp(meta_name[cnt], "ADDITIONALATTRIBUTENAME") == 0)
          {
	    len_t = (int)strlen(meta_val[cnt]);
	    strmid(meta_val[cnt], 1, len_t-2, meta_name[cnt]);
	    cnt++;
          }
          else if (strcmp(meta_name[cnt], "PARAMETERVALUE") == 0)
	    strcpy(meta_val[cnt-1], meta_val[cnt]);
          else cnt++;
	}
      }
      /* MSZ 4/10/2019, using strstr makes more sense */
      /*else if (meta_str[i] == 'C')*/
      else if (strstr((char*)(meta_str + i), CLASS_name))
      {
        pos22 = charpos(meta_str, '=', i);
        pos33 = charpos(meta_str, '\n', pos22);
        i = pos33 + 1;
        while (meta_str[i] == ' ') i++;
        /*if (meta_str[i] == 'N')*/
	if(strstr((char*)(meta_str + i), NUM_VAL_name))
        {
          strmid(meta_str, pos2, pos3-pos2, meta_name[cnt]);
	  /* MSZ 4/12/2019, replace strpos with strpos_wb */
          pos1 = strpos_wb(meta_str, "VALUE", pos33);
	  if(pos1 != -1)
	  {
            pos2 = charpos(meta_str, '=', pos1) + 2;
            if (meta_str[pos2] == '(')
              pos3 = charpos(meta_str, ')', pos2) + 1;
            else
              pos3 = charpos(meta_str, '\n', pos2);
            strmid(meta_str, pos2, pos3-pos2, meta_val[cnt]);
	    if (meta_str[pos3-1] == ')')
	      rm_ln_in_str(meta_val[cnt]);
            if (strcmp(meta_name[cnt], "ADDITIONALATTRIBUTENAME") == 0)
            {
	      len_t = (int)strlen(meta_val[cnt]);
	      strmid(meta_val[cnt], 1, len_t-2, meta_name[cnt]);
              cnt++;
            }
            else if (strcmp(meta_name[cnt], "PARAMETERVALUE") == 0)
              strcpy(meta_val[cnt-1], meta_val[cnt]);
            else cnt++;
	  }
	}
      }
    }
    /* MSZ 4/12/2019 changed the below */
    /*else pos3 = len + 1;*/
    else pos3 += 1;
  }
  *meta_cnt = cnt;
}

void get_sel_metadata(char *meta_str, char *meta_name, char **meta_val, int *meta_cnt, int case_ch)
/***************************************************************************************
	Read all instances of metadata meta_name and read the metadata values into the 
	character array meta_val from the single character string meta_str containing 
	all the core or archive metadata. meta_cnt contains the number of instances of 
	metadata meta_name found in meta_str.
****************************************************************************************/
{
  int cnt, st;
  char tmp_name[MAX_STR_LEN];
  int pos1, pos2, pos3, pos4;

  cnt = *meta_cnt;
  pos1 = (case_ch == 0) ? strpos(meta_str, meta_name, 0) : strcasepos(meta_str, meta_name, 0); 
  while (pos1 != -1)
  {
    pos1--;
    if ((meta_str[pos1] == '\"') || (meta_str[pos1] == ' ')) pos1++;
    pos2 = charpos(meta_str, '\n', pos1);
    if (meta_str[pos2-1] == '\"') pos2--; 
    strmid(meta_str, pos1, pos2-pos1, tmp_name);
    st = (case_ch == 0) ? strcmp(meta_name, tmp_name) : strcasecmp(meta_name, tmp_name);
    if (st == 0)
    {
      if (cnt == *meta_cnt) strcpy(meta_name, tmp_name);
      pos3 = strpos(meta_str, "END_OBJECT", pos2);
      pos3 = charpos(meta_str, '=', pos3) + 2;
      pos4 = charpos(meta_str, '\n', pos3);
      strmid(meta_str, pos3, pos4-pos3, tmp_name);
      if (strcmp(tmp_name, "ADDITIONALATTRIBUTENAME") == 0)
      {
	pos2 = strpos(meta_str, "PARAMETERVALUE", pos4);
	pos2 = charpos(meta_str, '\n', pos2);
      }
      pos2 = strpos(meta_str, "VALUE", pos2);
      pos2 = charpos(meta_str, '=', pos2) + 2;
      if (meta_str[pos2] == '(')
        pos3 = charpos(meta_str, ')', pos2) + 1;
      else
        pos3 = charpos(meta_str, '\n', pos2);
      strmid(meta_str, pos2, pos3-pos2, meta_val[cnt]);
      cnt++;
      pos1 = strpos(meta_str, "END_OBJECT", pos3);
      pos1 = charpos(meta_str, '\n', pos1);
      pos1 = strpos(meta_str, meta_name, pos1);
    } 
    else
      pos1 = strpos(meta_str, meta_name, pos2);
  }
  *meta_cnt = cnt;
}

void copy_metadata(int H5IN, int32 in_sd_id, int H5OUT, int32 out_sd_id)
/********************************************************************************************
  Copy all global metadata from input to output 
********************************************************************************************/
{
    long attr_len;
    int32 nsds, nattr;
    int32 attr_cnt, attr_index;
    int32 attr_type, attr_size;
    char attr_name[100], *attr_buf;
    
    /* HDF5 addition */
    H5_ATTR h5attr;
    
    fprintf(stdout, "\tWriting metadata to output\n");
    if (H5IN)
    {
        h5attr.loc_id = out_sd_id;
        h5attr.rdata = NULL;
        h5attr.name[0] = '\0';
        if (H5OUT)
            H5Aiterate2(in_sd_id,H5_INDEX_CRT_ORDER,H5_ITER_INC,NULL,(H5A_operator2_t)H5_attr_write_hdf5,(void *)&h5attr);
        else
            H5Aiterate2(in_sd_id,H5_INDEX_CRT_ORDER,H5_ITER_INC,NULL,(H5A_operator2_t)H5_attr_write_hdf4,(void *)&h5attr);

        /* MSZ 1/30/2018, add the below for metedata in HDF5 as DATASET of H5T_STRING */
	if(H5OUT) /* do nothing so far */
	  ;
	else
	  copy_H5_all_metedata_strings_to_H4(in_sd_id, out_sd_id);
    }
    else
    {
        SDfileinfo(in_sd_id, &nsds, &nattr);
        for (attr_index=0; attr_index<nattr; attr_index++)
        {
            if (SDattrinfo(in_sd_id, attr_index, attr_name, &attr_type, &attr_cnt) == FAIL)
                fprintf(stdout, "Cannot read information for attribute %d\n", attr_index);
            else
            {
                attr_size = DFKNTsize(attr_type);
                attr_len = attr_size*attr_cnt;
                if ((attr_type == DFNT_CHAR) || (attr_type == DFNT_CHAR8))
                    attr_len++;
                if ((attr_buf = (void *)calloc(attr_size, attr_len)) == NULL)
                    fprintf(stdout, "Cannot allocate memory for attr_buf in copy_metadata\n");
                else
                {
                    if (SDreadattr(in_sd_id, attr_index, (VOIDP)attr_buf) == FAIL)
                      fprintf(stdout, "Cannot read %s from input file\n", attr_name);
                    else
                    {
                        if (H5OUT && setH5Attr(attr_type,attr_name,attr_cnt,(VOIDP)attr_buf, out_sd_id) == FAIL)
                            fprintf(stderr, "Cannot write metadata %s to output HDF5 file\n", attr_name);
                        else if (!H5OUT && SDsetattr(out_sd_id, attr_name, attr_type, attr_cnt, (VOIDP)attr_buf) == FAIL)
                        fprintf(stdout, "Cannot write metadata %s to output file\n", attr_name);
                    }
                    free(attr_buf);
                }
            }
        } /* for (attr_index=0; . . .      */
    }
    return;
}

void write_modss_metadata(int32 in_sd_id, int32 out_sd_id, char **meta_names, char **meta_vals)
/********************************************************************************************
    Copy metadata from input to output L2G subset. Update all neccessary metadata.
********************************************************************************************/
{
    int imeta;
    int32 attr_cnt, attr_index;
    int32 attr_size, attr_type;
    char attr_name[20], *meta_attr_buf, *new_mbuf;
    char *meta_gname[] = {"CoreMetadata.0", "ArchiveMetadata.0"};
  
    for (imeta=0; imeta<2; imeta++)
    {
      if ((attr_index = SDfindattr(in_sd_id, meta_gname[imeta])) == FAIL)
        fprintf(stderr, "Cannot find %s in input HDF file\n", meta_gname[imeta]);
      else
      {
        if (SDattrinfo(in_sd_id, attr_index, attr_name, &attr_type, &attr_cnt) == FAIL)
          fprintf(stderr, "Cannot read information for %s from input HDF file", meta_gname[imeta]);
        else
        {
          attr_size = DFKNTsize(attr_type) * attr_cnt + 1;
          if ((meta_attr_buf = (char *)calloc(attr_size, sizeof(char))) == NULL)
            fprintf(stderr, "Cannot allocate memory for metadata in write_modss_metadata()\n");
          else
          {
            if (SDreadattr(in_sd_id, attr_index, (VOIDP)meta_attr_buf) == FAIL)
              fprintf(stderr, "Cannot read %s from input HDF file \n", meta_gname[imeta]);
            else
            {
          new_mbuf = update_modss_metadata(meta_attr_buf, attr_cnt, imeta, meta_names, meta_vals);
          attr_cnt = (int)strlen(new_mbuf);
              if (SDsetattr(out_sd_id, meta_gname[imeta], attr_type, attr_cnt, (VOIDP)new_mbuf) == FAIL)
                fprintf(stderr, "Cannot write metadata %s to output file \n", meta_gname[imeta]);
              free(new_mbuf);
            }
            free(meta_attr_buf);
          }
        }
      }
    } /* for (imeta=0; . . . )      */
}                                      

char *update_modss_metadata(char *attr_buf, int org_len, int imeta, char **meta_names, char **meta_vals)
/****************************************************************************************
   Create the new metadata string by modifying the values of required metadata
****************************************************************************************/
{
  char end_c, *meta_str;
  int m1, m2, m_cnt;
  int i, j, k, ii, jj;
  int p1, p2, p3, len;
  int st_id[10], st_p[10], end_p[10];

  /* get index of core metadata and esdt metadata to be updated */

  if (imeta == 0) { m1 = 0; m2 = 6; m_cnt = 6; }
  else { m1 = 6; m2 = 10; m_cnt = 4; }

  /* Find the starting and ending positions of metadata values in the string */

  for (i=m1; i<m2; i++)
  {
    if ((p1 = strpos(attr_buf, meta_names[i], 0)) != -1)
    {
      p2 = strpos(attr_buf, "VALUE", p1);
      p3 = charpos(attr_buf, '=', p2);
      p3 += 2;
      st_p[i-m1] = p3;
      end_c = (attr_buf[p3] == '(') ? ')' : '\n';
      while (attr_buf[p3] != end_c) p3++;
      end_p[i-m1] = p3;
    }                                    
    else st_p[i-m1] = end_p[i-m1] = -1;
  }

  /* sort the metadata based on their position in the string */

  for (i=0; i<m_cnt; i++)
    st_id[i] = i;
  for (i=0; i<m_cnt; i++)
  {
    ii = st_id[i];
    for (j=i+1; j<m_cnt; j++)
    {
      jj = st_id[j];
      if (st_p[ii] > st_p[jj])
        ii = st_id[j];
    }
    if (ii != st_id[i])
    {
      k = st_id[i];
      st_id[i] = ii;
      st_id[ii] = k;
    }
  }

  /* obtain the new length */

  for (i=0, k=0, len=0; i<m_cnt; i++)
  {
    ii = st_id[i];
    if (i == 0) p1 = 0;
    else
    {
      jj = st_id[i-1];
      p1 = end_p[jj];
    }
    p2 = st_p[ii];
    len = len + p2 - p1 + 1;

    ii += m1;
    len += (int)strlen(meta_vals[ii]);
  }                                  
  ii = st_id[i-1];
  len = len + (org_len - end_p[ii]) + 1;

  /* update metadata and shift the string appropriately */ 

  if ((meta_str = (char *)calloc(len, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for meta_str in update_modss_metadata()\n");
  else
  {
    for (i=0, k=0; i<m_cnt; i++)
    {
      ii = st_id[i];
      if (i == 0) p1 = 0;
      else
      {
        jj = st_id[i-1];
        p1 = end_p[jj];
      }
      p2 = st_p[ii];
      for (j=p1; j<=p2; j++, k++)
        meta_str[k] = attr_buf[j];
      ii += m1;
      len = (int)strlen(meta_vals[ii]);
      for (j=0; j<len; j++, k++)
        meta_str[k] = meta_vals[ii][j];
    }
    jj = st_id[i-1];
    p1 = end_p[jj]; p2 = org_len;
    for (j=p1; j<=p2; j++, k++)
      meta_str[k] = attr_buf[j];
  }                                                    
  return meta_str;
}

#if 0
/* MSZ 3/3/2018, Note that get_attr_npp_sel_metadata() cannot deal with HDF5 and
 * is deprecated and void *get_attr_sel_metadata() and 
 * void find_get_sd_attr_to_str(char *fname, char *attr_name, char *meta_val)
 * should be used instead.*/
void get_attr_npp_sel_metadata(char *hdf_fname, char *meta_gname, char *meta_val)
/**********************************************************************************
	Get core or archive metadata from the hdf file hdf_fname.
	On success the functions returns a pointer to the character string 
	containing the metadata else returns a NULL pointer on failure
***********************************************************************************/
{
  int status = 1;
  int32 sd_id;
  int32 attr_cnt;
  int32 attr_size;
  int32 attr_type;
  int32 attr_index;
  char attr_name[20];
  char *attr_buf;
  int i;
  long attr_len;
  char tmpData[200];  
  int32 nsds, nattr;
  
  if ((sd_id = SDstart(hdf_fname, DFACC_READ)) == FAIL)
  {
    fprintf(stderr, "Cannot open the input HDF file %s\n", hdf_fname);
    status = -1;
  }
  else
  {
    SDfileinfo(sd_id, &nsds, &nattr);
    for (attr_index=0; attr_index<nattr; attr_index++)
    {	      		    
      if (SDattrinfo(sd_id, attr_index, attr_name, &attr_type, &attr_cnt) == FAIL)
	fprintf(stdout, "Cannot read information for attribute %d\n", attr_index);
      else
      {	      	
	status = strcasecmp(meta_gname, attr_name);
	if (status == 0)
	{    
	  attr_size = DFKNTsize(attr_type);
	  attr_len = attr_size*attr_cnt;
	  if ((attr_type == DFNT_CHAR) || (attr_type == DFNT_CHAR8))
            attr_len++;
	  if ((attr_buf = (void *)calloc(attr_size, attr_len)) == NULL) 
	  {
            fprintf(stdout, "Cannot allocate memory for attr_buf in copy_metadata\n");
	    status = -1;
	  }
	  else
	  {
            if (SDreadattr(sd_id, attr_index, (VOIDP)attr_buf) == FAIL)
	    {
              fprintf(stdout, "Cannot read %s from input file\n", attr_name);        	      
              status = -1;
	    }
	    else 
            {    
	      switch (attr_type)
	      {			 			        
		case 4: /*fprintf(stdout, "%s:\"%s\"\n",attr_name, attr_buf);*/
	        	sprintf(meta_val, "%s", attr_buf);
/meta_val
			break; 		
		case 5:  /*fprintf(stdout, "%s:%f\n", attr_name, ((float32 *)attr_buf)[0]);*/
	      		sprintf(meta_val, "%f", ((float32 *)attr_buf)[0]);
			break; 
		case 6:  /*fprintf(stdout, "%s:%f\n", attr_name, ((float64 *)attr_buf)[0]);*/
	        	sprintf(meta_val, "%f", ((float64 *)attr_buf)[0]);
			break; 	     
		case 20:  /*fprintf(stdout, "%s:%d\n", attr_name, ((int8 *)attr_buf)[0]);*/
	        	 sprintf(meta_val, "%d", ((int8 *)attr_buf)[0]);	      
			 break; 
		case 21:  /*fprintf(stdout," %s:%d\n", attr_name, ((uint8 *)attr_buf)[0]);*/
	        	 sprintf(meta_val, "%d", ((uint8 *)attr_buf)[0]);	      
			 break; 
		case 22:  /*fprintf(stdout, "%s:%d\n", attr_name, ((int16 *)attr_buf)[0]);*/
                	 sprintf(meta_val, "%d", ((int16 *)attr_buf)[0]);	      
			 break; 
		case 23:  /*fprintf(stdout, "%s:%d\n", attr_name, ((uint16 *)attr_buf)[0]);*/
                	 sprintf(meta_val, "%d", ((uint16 *)attr_buf)[0]);	      
			 break; 
		case 24:  /*fprintf(stdout, "%s: ", attr_name);*/
			 if (attr_cnt == 1) 
			 {
		            /*fprintf(stdout, "%d\n", ((int32 *)attr_buf)[0]);*/
			   sprintf(meta_val, "%d", ((int32 *)attr_buf)[0]);
			 }
			 else
			 {
			   sprintf(meta_val, "%d,", ((int32 *)attr_buf)[0]);
			   for (i=1; i<attr_cnt-1; i++)
			   {
			     sprintf(tmpData, "%d,", ((int32 *)attr_buf)[i]);
			     strcat(meta_val, tmpData); 
			   }  
			   sprintf(tmpData, "%d", ((int32 *)attr_buf)[attr_cnt-1]);			   
			   strcat(meta_val, tmpData); 			 				  	
			 }			    			    
			 break; 			    
		case 25: /*fprintf(stdout, "%s:%d\n", attr_name, ((uint32 *)attr_buf)[0]);*/
	        	 sprintf(meta_val, "%d", ((uint32 *)attr_buf)[0]);
			 break; 
	      }			
            }
	  free(attr_buf);  
	  }
	}	        	
      }
    }
    SDend(sd_id);
  }
}
#endif

void *get_attr_sel_metadata(const char* hdf_fname, char *attr_name, int32 *attr_type, int32 *attr_cnt)
{
    /* MSZ 2/8/2018 modified from void *find_get_sd_attr() but this function first finds the
     * exactly attrname using strcasestr() if the input attr_name is not the exact attr_name.
     * Note that attr_name can be not exactly same as that in the file and this function will 
     * find and use the correct attr_name */
    int32 nsds, nattr;
    int32 attr_size;
    int32 attr_index;
    int32 sd_id;
    void *attr_buf = NULL;
    char name[80] = {'\0'};
    char real_attr_name[80] = {'\0'};

    herr_t  ret;                /* Return value */
    H5O_info_t oinfo;           /* Object info */
  
    hid_t attrh5_type, attr_space;
    hid_t attr_id, memtype;
    H5T_class_t type_class;
    size_t size;
    unsigned h5_idx;
    ssize_t name_len;
    int HDF_ver = 0;
    int H5IN;
    int find_it = 0;

    /* MSZ HDF5 or HDF4 */
    HDF_ver = HDF_version_check(hdf_fname);

    if(HDF_ver == 4)
        H5IN = 0;
    else if(HDF_ver == 5)
        H5IN = 1;
    else
    {
        printf("%s is neither HDF4 nor HDF5, exit...\n", hdf_fname);
        exit(0);
    }

    if (H5IN)
    {
      if ((sd_id = (int32)H5Fopen(hdf_fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
      {
	fprintf(stderr, "Cannot open the HDF file %s\n", hdf_fname);
	return attr_buf;
      }
      /* find the real_attr_name */
      ret = H5Oget_info((hid_t)sd_id, &oinfo);
      if(oinfo.num_attrs > 0)
      {
        for(h5_idx = 0; h5_idx < (unsigned)oinfo.num_attrs; h5_idx++)
	{
	   /* for the current id */
	   /* MSZ 2/21/2018  Note that sometimes H5Aget_name_by_idx() fails to work to get the attribute names */
	   /*name_len = H5Aget_name_by_idx((hid_t)sd_id, ".", H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)h5_idx, real_attr_name, 0, H5P_DEFAULT);*/

	   attr_id = H5Aopen_by_idx((hid_t)sd_id, ".", H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)h5_idx, H5P_DEFAULT, H5P_DEFAULT);
	   name_len = H5Aget_name(attr_id, 0, real_attr_name);

	   if(name_len < 0) real_attr_name[0] = '\0';
	   else
	     if(strcasestr(real_attr_name, attr_name) != '\0')
	     {
	       find_it = 1;
	       break;
	     }

	   ret = H5Aclose(attr_id);
	}
      }
      else
	return attr_buf;
      
      if( find_it == 0 || (attr_id = H5Aopen((hid_t)sd_id, real_attr_name, H5P_DEFAULT)) < 0 )
      {
        fprintf(stderr, "Could not find the attribute %s for HDF file\n", real_attr_name);
	return attr_buf;
      }
      else
      {
        /* Get attribute datatype, dataspace, rank, and dimensions.  */
        attr_space = H5Aget_space(attr_id);

	attrh5_type= (int32)H5Aget_type(attr_id);
	*attr_type = H5toH4type(attrh5_type);

	type_class = H5Tget_class(attrh5_type);

	/* MSZ 8/15/17, largely modified the below code */
	size       = H5Tget_size(attrh5_type);

	if(type_class == H5T_STRING)
	  size++; /* Make room for null terminator */

	/*
	if(type_class != -1)
	    printf("ATTR_NAME=%s type_class=%d H5T_INTEGER=%d H5T_FLOAT=%d size = %d\n", attr_name, type_class, H5T_INTEGER, H5T_FLOAT, size);
	*/

	if(type_class == H5T_INTEGER || type_class == H5T_FLOAT)  /* numerical types */
	{
	    /* Get attribute datatype, dataspace, rank, and dimensions.  */
	    *attr_cnt = H5Sget_simple_extent_npoints(attr_space);

	    if ((attr_buf = (void *)calloc(*attr_cnt, size)) == NULL)
		fprintf(stderr, "Cannot allocate memory for attr_buf in get_sd_attr\n");
	    else
	        if (H5Aread(attr_id, attrh5_type, attr_buf) < 0)
		    fprintf(stderr, "Cannot read sd_attr %s in get_sd_attr\n", attr_name);
		
	}
	else if(type_class == H5T_STRING)        /* string */
	{
	    *attr_cnt = size;

	    if ((attr_buf = (void *)calloc(*attr_cnt, sizeof(char))) == NULL)
		fprintf(stderr, "Cannot allocate memory for attr_buf in get_sd_attr\n");
	    else
		if (H5Aread(attr_id, attrh5_type, attr_buf) < 0)
		    fprintf(stderr, "CANnot read sd_attr %s in get_sd_attr\n", attr_name);
	}
	else  /* we just interested in scale_factor, offset fill value etc not others */
	{
	   *attr_cnt = 0;
	   *attr_type = -1;
	   return attr_buf;
	}

	H5Tclose(attrh5_type);
        H5Sclose(attr_space);
        H5Aclose(attr_id);
	H5Fclose((hid_t)sd_id);
      } 
    }
    else
    {
	if ((sd_id = SDstart(hdf_fname, DFACC_READ)) == FAIL)
	{
	  fprintf(stderr, "Cannot open the HDF file %s\n", hdf_fname);
	  return attr_buf;
	}

	/* find the real_attr_name */
	SDfileinfo(sd_id, &nsds, &nattr);
	if(nattr < 1)
	   return attr_buf;

	for (attr_index=0; attr_index<nattr; attr_index++)
	{
	  if (SDattrinfo(sd_id, attr_index, real_attr_name, attr_type, attr_cnt) == FAIL)
	  {
	    fprintf(stdout, "Cannot read information for attribute %d\n", attr_index);
	    real_attr_name[0] = '\0';
	  }
	  else
	    if(strcasestr(real_attr_name, attr_name) != '\0')
	    {
		find_it = 1;
		break;
	    }
	}

        if (find_it ==0 || (attr_index = SDfindattr(sd_id, real_attr_name)) == FAIL)
	{
          fprintf(stderr, "Could not find the attribute %s for HDF file\n", real_attr_name);
	  return attr_buf;
	}
        else
        {
          if (SDattrinfo(sd_id, attr_index, name, attr_type, attr_cnt) == FAIL)
	  {
            fprintf(stderr, "Cannot read the attribute information for %s\n", real_attr_name);
	    return attr_buf;
	  }
          else
          {
            attr_size = DFKNTsize(*attr_type);
            if ((attr_buf = (void *)calloc(*attr_cnt, attr_size)) == NULL)
              fprintf(stderr, "Cannot allocate memory for attr_buf in get_sd_attr\n");
            else
            {
		int status;
		status = SDreadattr(sd_id, attr_index, attr_buf);
		if(status == -1)
		{
		    fprintf(stderr, "Cannot read sds_attr %s in get_sd_attr\n", real_attr_name);
		    if(attr_buf) free(attr_buf);
		    return NULL;
		}
		else
		  return attr_buf;
            }
          }
        }
    }

    return attr_buf;
}

/* MSZ */
int GetJulianDay(int year, int month, int day)
{
  int jday = 0, i;
  
  for (i=1; i<month; i++)
    jday += GetNumberDays(year, i); 
  jday = jday + day;
  return(jday);
}  

void jultomdy(int julday, int *month, int *day, int *year)
{
  /*return month, day, year from specified julian day Uses getnumberdays */
  int i, juldaytemp, daynumber, numbermonths;
  juldaytemp = 0;
  daynumber = 0;
  numbermonths = 12;
  for (i=0; i<numbermonths; i++)
  {
    daynumber = GetNumberDays(*year, i);
    if ((julday >=juldaytemp) && (julday <= juldaytemp+daynumber))
    {
      *month = i;
      *day = julday - juldaytemp;
      return;
    }
    else
      juldaytemp = juldaytemp + daynumber;
  }
}  
  
int GetNumberDays(int year, int month)
{
  int numberofdays;
  int four = 4.;
  
  switch(month)
  {
    case 1:
      numberofdays = 31;
      break;
    case 2:
      if ((year % four) == 0)
        numberofdays = 29;
      else
        numberofdays = 28;
      break;
    case 3:
      numberofdays = 31;
      break;      
    case 4:
      numberofdays = 30;
      break;      
    case 5:
      numberofdays = 31;
      break;      
    case 6:
      numberofdays = 30;
      break;
    case 7:
      numberofdays = 31;
      break;      
    case 8:
      numberofdays = 31;
      break;      
    case 9:
      numberofdays = 30;
      break;      
    case 10:
      numberofdays = 31;
      break; 
    case 11:
      numberofdays = 30;
      break;      
    case 12:
      numberofdays = 31;
      break;                  		  
  }
  return (numberofdays);  
} 

int HDF5_read_one_dataset_STRING(hid_t h5_file_id, const char *h5_dataset_name, void** rdata, size_t *len_strs)
{
    /* written: Maosheng Zhao, SSAI */
    /* Jan, 30, 2018 */
    /* Note that h5_dataset_name is the full name starting from the root group */
    int status = 1;
    hid_t       sds_id, dtype, memtype, dspace, dataset;
    H5T_class_t type_class;
    herr_t      ret;
    hsize_t     dims[32] = {1};
    size_t      stringlen;
    int         i, rank = 0;
    H5T_cset_t cset;  /* MSZ 10/25/2018, UTF8 */

    /* Open dataset. */
    if((sds_id = H5Dopen2(h5_file_id, h5_dataset_name, H5P_DEFAULT)) < 0)
    {
      fprintf(stderr, "fail to open dataset %s\n", h5_dataset_name);
      status = -1;
    }

    dtype = H5Dget_type(sds_id);
    type_class = H5Tget_class(dtype);
    dspace = H5Dget_space(sds_id);
    rank = H5Sget_simple_extent_ndims(dspace);
    ret = H5Sget_simple_extent_dims(dspace, dims, NULL);
    *len_strs = 0;

    if(type_class == H5T_STRING) {
    	    /* assign memtype for reading dataset */
	    stringlen = H5Tget_size(dtype);
	    stringlen++;  /* Make room for null terminator */
	    /* different from numerical data types, you must specify size for H5T_C_S1, memtype */
	    memtype = H5Tcopy(H5T_C_S1);

	    /* Maosheng Zhao, 10/25/2018 get and set character systrem of either ASCII or UTF8*/
            cset = H5Tget_cset(dtype);
            H5Tset_cset(memtype, cset);

	    /* here size is the length of the string with end null */
	    ret = H5Tset_size (memtype, stringlen);
    	if(rank == 0) /* a string */
	{
	    /* calloc memory for rdata with H5T_C_S1 type of a string, dataset_info.sdim[0] = 0 */
	    *len_strs = stringlen;
	    *rdata = (void *)calloc(*len_strs, sizeof(char));
	}
	else if(rank == 1) /* string vector */
	{
	    /* calloc memory for rdata with H5T_C_S1 type. different strings have same string_len */
	    *len_strs = dims[0] * stringlen;
	    *rdata = (void *)calloc(*len_strs, sizeof (char));
	}
	else
	{
	    fprintf(stderr, "don't deal with H5T_STRING beyond string vector, exit...\n");
	    exit(0);
	}
        /* Read the data. */
        ret = H5Dread (sds_id, memtype, H5S_ALL, H5S_ALL, H5P_DEFAULT, *rdata);

	if(ret < 0)
	   status = -1;

        /* Close and release resources.*/
        ret = H5Dclose (sds_id);

        ret = H5Tclose (memtype);

	status = 1;
    }
    else
    {
	fprintf(stderr, "%s is not in H5T_STRING!\n", h5_dataset_name);
	status = -1;
    }

    H5Tclose(dtype);
    H5Tclose(type_class);
    H5Sclose(dspace);
    H5Dclose(sds_id);

    return status;
}

void copy_H5_all_metedata_strings_to_H4(int32 H5_sd_id, int32 H4_sd_id)
{
  /* Developer: Maosheng Zhao, SSAI
   * date: 1/30/2018
   * Purpose: read the metadata in DATASET in H5T_STRING in HDF5
   *          and copy them to the file attrs of HDF4 */

  hid_t   h5_sds_id, h5dtype;
  H5T_class_t h5_type_class;
  herr_t  ret;
  hsize_t i;
  size_t len_strs;
  H5_OBJS_TYPES *h5_obj_types = NULL;
  void *meta_attr_buf = NULL;

  /* get objs_links */
  h5_obj_types = (H5_OBJS_TYPES*)calloc(1, sizeof(H5_OBJS_TYPES));

  /* The reason I don't use the existing
   * H5Ovisit(sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5O_iterate_t)H5dataset_search, (void *)&h5_sds);
   * is that call-back function H5dataset_search removed the groups from the name of DATASETs but
   * I need full name of DATASETs here. So I used my developed H5Fget_objs_links */ 
  ret = H5Fget_objs_links((hid_t)H5_sd_id, h5_obj_types);

  /* check if there is meta_name */
  for(i = 0; i < h5_obj_types->num_objs_links; i++)
    if(h5_obj_types->obj_type[i] == H5O_TYPE_DATASET)
    {
      /* if the dataset type class is H5T_STRING */
      if((h5_sds_id = H5Dopen2(H5_sd_id, h5_obj_types->obj_name[i], H5P_DEFAULT)) < 0)
	fprintf(stderr, "cannot H5Dopen2 %s\n", h5_obj_types->obj_name[i]);
      else
      {
	h5dtype = H5Dget_type(h5_sds_id);
	if(H5Tget_class(h5dtype) == H5T_STRING)
	{
	  if(HDF5_read_one_dataset_STRING(H5_sd_id, h5_obj_types->obj_name[i], &meta_attr_buf, &len_strs) == -1)
	  {
	    fprintf(stderr, "cannot read %s\n", h5_obj_types->obj_name[i]);
	    exit(0);
	  }
    	  else
	  {
	    char *grp_name = NULL, *meta_name = NULL;
	    HDF5_split_abs_name_into_path_name(h5_obj_types->obj_name[i], &grp_name, &meta_name);
	    if ( SDsetattr(H4_sd_id, meta_name, DFNT_CHAR8, (int32)len_strs, meta_attr_buf) == FAIL)
               fprintf(stderr, "WARNING: cannot copy HDF5 attribute %s (H5_attr_write_hdf4)\n", meta_name);

	    /* free memory */
	    if(grp_name) free(grp_name);
	    if(meta_name) free(meta_name);
	    if(meta_attr_buf) free(meta_attr_buf);
	  }
	}
	H5Tclose(h5dtype);
	H5Dclose(h5_sds_id);
      }
    }

  for(i = 0; i < h5_obj_types->num_objs_links; i++)
  if(h5_obj_types->obj_name[i]) free(h5_obj_types->obj_name[i]);

  if(h5_obj_types->obj_name) free(h5_obj_types->obj_name);
  if(h5_obj_types->obj_type) free(h5_obj_types->obj_type);
  if(h5_obj_types)free(h5_obj_types);
}

void H5_print_all_attr_groups(int32 H5_sd_id, const char* key_str_control)
{
  /* Developer: Maosheng Zhao, SSAI
   * date: 1/30/2018
   * Purpose: read the metadata in DATASET in H5T_STRING in HDF5
   *          and copy them to the file attrs of HDF4 */

  herr_t  ret;
  hsize_t i;
  H5_OBJS_TYPES *h5_obj_types = NULL;
  void *meta_attr_buf = NULL;

  /* get objs_links */
  h5_obj_types = (H5_OBJS_TYPES*)calloc(1, sizeof(H5_OBJS_TYPES));

  /* The reason I don't use the existing
   * H5Ovisit(sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5O_iterate_t)H5dataset_search, (void *)&h5_sds);
   * is that call-back function H5dataset_search removed the groups from the name of DATASETs but
   * I need full name of DATASETs here. So I used my developed H5Fget_objs_links */ 
  ret = H5Fget_objs_links((hid_t)H5_sd_id, h5_obj_types);

  /* check if there is meta_name */
  for(i = 0; i < h5_obj_types->num_objs_links; i++)
    if(h5_obj_types->obj_type[i] == H5O_TYPE_GROUP)
      HDF5_group_attrs((hid_t)H5_sd_id, h5_obj_types->obj_name[i], key_str_control);

  for(i = 0; i < h5_obj_types->num_objs_links; i++)
  if(h5_obj_types->obj_name[i]) free(h5_obj_types->obj_name[i]);

  if(h5_obj_types->obj_name) free(h5_obj_types->obj_name);
  if(h5_obj_types->obj_type) free(h5_obj_types->obj_type);
  if(h5_obj_types)free(h5_obj_types);
}

int HDF5_group_attrs(hid_t h5_file_id, const char *h5_group_name, const char* key_str_control)
{
    /* written: Maosheng */
    /* Jan, 15, 2017 */
    /* Note that this function can read both numerical and string types HDF5 dataset */
    /* Note that the routine can also be followed and modified to get the attrs for other objects such as group */

    hid_t   group_id;
    hsize_t index_offset;

    hid_t   attr;
    hid_t   atype;    /* Attribute type */
    H5T_class_t  type_class;

    herr_t  ret;                /* Return value */
    H5O_info_t oinfo;           /* Object info */
    unsigned i;              /* Counters */

    /* Open the group */
    group_id = H5Gopen2(h5_file_id, h5_group_name, H5P_DEFAULT);

    #if 0
    /* get attribute by idx, I didn't use it but iteration function below */
    ret = H5Oget_info(group_id, &oinfo);
    for(i = 0; i < (unsigned)oinfo.num_attrs; i++) {
	attr = H5Aopen_by_idx(group_id, ".", H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)i, H5P_DEFAULT, H5P_DEFAULT);
	atype = H5Aget_type(attr);
	type_class = H5Tget_class(atype);
	if (type_class == H5T_STRING) {
	/* do something for attribute */
	}
	ret = H5Aclose(attr);
	ret = H5Tclose(atype);
    }
    #endif

    /* Get attribute info using iteration function. */
    ret = H5Oget_info(group_id, &oinfo);
    if(oinfo.num_attrs > 0) {
	/*printf("\"%s\" has %llu attributes, they are:\n", h5_group_name, oinfo.num_attrs);*/
	/* start from the first attribute */
	index_offset = 0;
	ret = H5Aiterate2(group_id, H5_INDEX_CRT_ORDER, H5_ITER_INC, &index_offset, HDF5_attr_info, (void*)key_str_control);
    } else {
	/*printf("\"%s\" has no attribute.\n", h5_group_name);*/
    }

    /* Close the group_id.*/
    H5Gclose(group_id);

    return 0;
}

herr_t HDF5_attr_info(hid_t loc_id, const char *name, const H5A_info_t *ainfo, void *opdata)
{
/* written: Maosheng Zhao
 * Date: June, 9, 2017
 * modified H5A_operator2_t operator function by HDF group
 * requires functions:
 * HDF5_get_data_type_info()
 */ 
    hid_t attr, atype, aspace, atype_mem;  /* Attribute, datatype and dataspace identifiers */
    int  rank, num_scales;
    hsize_t sdim[H5S_MAX_RANK];
    herr_t ret;
    int i;
    size_t npoints, stringlen;             /* Number of elements in the array attribute. */
    void *data_array = NULL;               /* Pointer to the array attribute. */
    char *string_out = NULL;
    HDF5_DATA_TYPE_INFO data_type_info;
    H5T_cset_t cset;   /* MSZ 10/25/2018 UTF8 */

    /* get the opotional data */
    opdata = opdata;

    /* Open the attribute using its name.*/
    attr = H5Aopen(loc_id, name, H5P_DEFAULT);

    /* Display attribute name.*/
    /*printf("\nName : %s\n", name);*/
    if(strcasestr(name, (char*)opdata) != '\0')
      fprintf(stdout, "%s = ", name);

    /* Get attribute datatype, dataspace, rank, and dimensions.*/
    atype  = H5Aget_type(attr);
    aspace = H5Aget_space(attr);
    rank = H5Sget_simple_extent_ndims(aspace);
    ret = H5Sget_simple_extent_dims(aspace, sdim, NULL);

    /* Read array attribute and display its type and values.
     * H5Tget_class() returns class names: H5T_INTEGER, H5T_FLOAT, H5T_STRING, or H5T_BITFIELD.*/
    data_type_info = HDF5_get_data_type_info(atype);

    /* Read array attribute and display its type and values.
     * H5Tget_class() returns class names: H5T_INTEGER, H5T_FLOAT, H5T_STRING, or H5T_BITFIELD.*/
  if(strcasestr(name, (char*)opdata) != '\0')
  {  
    if(data_type_info.type_class == H5T_INTEGER || data_type_info.type_class == H5T_FLOAT)
    {
	/* read the attr to the buffer */
	npoints = H5Sget_simple_extent_npoints(aspace);
	if((data_array = (void*)calloc(npoints, data_type_info.size)) == NULL)
	{
	    fprintf(stderr, "cannot calloc data_array in HDF5_attr_info(), exit...\n");
	    exit(0);
	}
	ret = H5Aread(attr, data_type_info.num_type, data_array);
	switch(data_type_info.HDF5_type) {
	    case DFNT_INT8:
		for( i = 0; i < (int)npoints; i++)    
		    fprintf(stdout, "%d ", ((int8*)(data_array))[i]);
		fprintf(stdout, "\n");
		break;
	   case DFNT_INT16:
		for( i = 0; i < (int)npoints; i++)    
		    fprintf(stdout, "%d ", ((int16*)(data_array))[i]);
		fprintf(stdout, "\n");
		break;
	   case DFNT_INT32:
		for( i = 0; i < (int)npoints; i++)    
		    fprintf(stdout, "%d ", ((int32*)(data_array))[i]);
		fprintf(stdout, "\n");
		break;
	   case DFNT_INT64:
		fprintf(stdout, "datatype : int64\n");
		/*printf("value : ");
		for( i = 0; i < (int)npoints; i++)    
		    printf("%ld ", ((int64*)(data_array))[i]);
		printf("\n"); */
		break;
	   case DFNT_UINT8:
		for( i = 0; i < (int)npoints; i++)    
		    fprintf(stdout, "%u ", ((uint8*)(data_array))[i]);
		fprintf(stdout, "\n");
		break;
	   case DFNT_UINT16:
		for( i = 0; i < (int)npoints; i++)    
		    fprintf(stdout, "%u ", ((uint16*)(data_array))[i]);
		fprintf(stdout, "\n");
		break;
	   case DFNT_UINT32:
		for( i = 0; i < (int)npoints; i++)    
		    fprintf(stdout, "%u ", ((uint32*)(data_array))[i]);
		fprintf(stdout, "\n");
		break;
	   case DFNT_UINT64:
		fprintf(stdout, "datatype : uint64\n");
		/*printf("value : ");
		for( i = 0; i < (int)npoints; i++)    
		    printf("%lu ", ((uint64*)(data_array))[i]);
		printf("\n");*/
		break;
	   case DFNT_FLOAT32:
		for( i = 0; i < (int)npoints; i++)    
		    fprintf(stdout, "%f ", ((float32*)(data_array))[i]);
		fprintf(stdout, "\n");
		break;
	   case DFNT_FLOAT64:
		for( i = 0; i < (int)npoints; i++)    
		    fprintf(stdout, "%lf ", ((float64*)(data_array))[i]);
		fprintf(stdout, "\n");
		break;
	   default:
		fprintf(stdout, "undefined HDF5 data types in HDF5_attr_info()\n");
	}

	/* free data_array */
	if(data_array)
	    free(data_array);

    } else {  /* non-numerical class types */
	switch(data_type_info.type_class)
	{
	    case H5T_STRING:
		stringlen = H5Tget_size(atype);
		stringlen++;  /* Make room for null terminator */
		if((string_out = (char*)calloc(stringlen, sizeof(char))) == NULL)
		{
		    fprintf(stdout, "cannot calloc string_out, exit...\n");
		    exit(0);
		}
		atype_mem = H5Tcopy(H5T_C_S1);

		/* Maosheng Zhao, 10/25/2018 get and set character systrem of either ASCII or UTF8*/
		cset = H5Tget_cset(atype);
		H5Tset_cset(atype_mem, cset);

		ret = H5Tset_size (atype_mem, stringlen);
		ret = H5Aread(attr, atype_mem, string_out);
		fprintf(stdout, "%s\n", string_out);
		ret = H5Tclose(atype_mem);
		if(string_out)
		    free(string_out);
		break;
	    case H5T_TIME:
		fprintf(stdout, "%s %s\n", "", "datatype class : H5T_TIME\n");
		break;
	    case H5T_BITFIELD:
		fprintf(stdout, "%s %s\n", "", "datatype class : H5T_BITFIELD\n");
		break;
	    case H5T_OPAQUE:
		fprintf(stdout, "%s %s\n", "", "datatype class : H5T_OPAQUE\n");
		break;
	    case H5T_COMPOUND:
		fprintf(stdout, "%s %s\n", "", "datatype class : H5T_COMPOUND\n");
		break;
	    case H5T_REFERENCE:
		fprintf(stdout, "%s %s\n", "", "datatype class : H5T_REFERENCE");
		break;
	    case H5T_ENUM:
		fprintf(stdout, "%s %s\n", "", "datatype class : H5T_ENUM\n");
		break;
	    case H5T_VLEN:
		fprintf(stdout, "%s %s\n", "", "datatype class : H5T_ENUM\n");
		break;
	    case H5T_ARRAY:
		fprintf(stdout, "%s %s\n", "", "datatype class : H5T_ARRAY\n");
		break;
	    case H5T_NO_CLASS:
		fprintf(stdout, "%s %s\n", "", "datatype class : H5T_NO_CLASS\n");
		break;
	    case H5T_NCLASSES:
		fprintf(stdout, "%s %s\n", "", "datatype class : H5T_NCLASSES\n");
		break;
	    default:
		fprintf(stdout, "undefined datatype class name in HDF5_attr_info()\n");
	} /* end switch */
    }	
  }

    /* Release all identifiers.*/
    H5Tclose(atype);
    H5Sclose(aspace);
    H5Aclose(attr);

    return 0;
}

herr_t H5_SDattrinfo(hid_t loc_id, const char *obj_name, int idx, char *attr_name, int *attr_H4_type, int *attr_cnt)
{
/* Maosheng Zhao, SSAI, 4/9/2018
 * Usage: H5_SDattrinfo(sds_info[isds].sds_id, ".", k, attr_name, &attr_type, &attr_cnt);
 * requires functions:
 * get_HDF5_data_type_info()
 * attr_cnt is the number of numerical values for numertical types but the length of the strings for string class
 * attr_H4_type includes numerical types of DFNT_INT8, DFNT_UINT8, DFNT_INT16, DFNT_UINT16, DFNT_INT32, DFNT_UINT32, DFNT_FLOAT32, DFNT_FLOAT64
 * DFNT_CHAR8 for string only
 */ 
    hid_t attr_id, atype, aspace, atype_mem;  /* Attribute, datatype and dataspace identifiers */
    int  rank, num_scales;
    hsize_t sdim[H5S_MAX_RANK];
    herr_t ret;
    int i, j;
    size_t npoints, stringlen;             /* Number of elements in the array attribute. */
    DATA_TYPE_INFO data_type_info;
    ssize_t name_size;
    hsize_t attr_storage_size;
    H5T_cset_t cset;
    H5S_class_t space_type;

    /* get the attr_name */
    name_size = H5Aget_name_by_idx(loc_id, obj_name, H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)idx, NULL, 0, H5P_DEFAULT);
    if(name_size < 0) return -1;
    name_size += 1;
    H5Aget_name_by_idx(loc_id, obj_name, H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)idx, attr_name, (size_t)name_size, H5P_DEFAULT);

    /* Open the attribute using its name */
    attr_id = H5Aopen(loc_id, attr_name, H5P_DEFAULT);
    if(attr_id < 0) return -1;

    /* Get attribute datatype, dataspace, rank, and dimensions.*/
    atype  = H5Aget_type(attr_id);
    aspace = H5Aget_space(attr_id);
    if(atype < 0 || aspace < 0) return -1;
    rank = H5Sget_simple_extent_dims(aspace, sdim, NULL);
    space_type = H5Sget_simple_extent_type(aspace);
    if(space_type < 0) return -1;

    /*  Display rank and dimension sizes for the array attribute.*/
    /*
    if(rank > 0) {
	printf("rank : %d \n", rank);
	printf("dimension sizes : ");
	for (i=0; i< rank; i++)
	  printf("%d ", (int)sdim[i]);
	printf("\n");
    }
    */

    /* initialize output arguments */
    *attr_H4_type = 99;
    *attr_cnt = 0;

    if(space_type == H5S_SCALAR || space_type == H5S_SIMPLE)
    {
	/* Read array attribute and display its type and values.
	 * H5Tget_class() returns class names: H5T_INTEGER, H5T_FLOAT, H5T_STRING, or H5T_BITFIELD.*/
	data_type_info = get_HDF5_data_type_info(atype, attr_name);

	/* Read array attribute and display its type and values.
	* H5Tget_class() returns class names: H5T_INTEGER, H5T_FLOAT, H5T_STRING, or H5T_BITFIELD.*/
	if(data_type_info.type_class == H5T_INTEGER || data_type_info.type_class == H5T_FLOAT){
	    /* read the attr to the buffer */
	    npoints = H5Sget_simple_extent_npoints(aspace);
	    *attr_cnt = (int)npoints;

	    if(data_type_info.num_type == H5T_NATIVE_SCHAR)
		*attr_H4_type = DFNT_INT8;
	    else if(data_type_info.num_type == H5T_NATIVE_SHORT)
	  	*attr_H4_type = DFNT_INT16;
	    else if(data_type_info.num_type == H5T_NATIVE_INT)
	  	*attr_H4_type = DFNT_INT32;
	    else if(data_type_info.num_type == H5T_NATIVE_LONG)
	  	/*printf("datatype : int64\n")*/;
	    else if(data_type_info.num_type == H5T_NATIVE_UCHAR)
	  	*attr_H4_type = DFNT_UINT8;
	    else if(data_type_info.num_type == H5T_NATIVE_USHORT)
	  	*attr_H4_type = DFNT_UINT16;
	    else if(data_type_info.num_type == H5T_NATIVE_UINT)
	  	*attr_H4_type = DFNT_UINT32;
	    else if(data_type_info.num_type == H5T_NATIVE_ULONG)
	  	/*printf("datatype : uint64\n")*/;
	    else if(data_type_info.num_type == H5T_NATIVE_FLOAT)
	  	*attr_H4_type = DFNT_FLOAT32;
	    else if(data_type_info.num_type == H5T_NATIVE_DOUBLE)
	  	*attr_H4_type = DFNT_FLOAT64;
	    else
	    {
	  	/*printf("undefined HDF5 data types in attr_info(), exit...\n");
	  	exit(0)*/;
	    }

	} else {  /* non-numerical class types */
	    switch(data_type_info.type_class)
	    {
		case H5T_STRING:
		    *attr_H4_type = DFNT_CHAR8;
		    /*printf("datatype class : H5T_STRING\n");
		    printf("datatype : H5T_C_S1\n");*/
		    stringlen = H5Tget_size(atype);
		    /* stringlen++; make room for terminator */
		    /* Maosheng Zhao, 4/11/2018. I identify a weird problem for HDF5 API. Sometimes,
		     * for data class H5T_STRING, H5Tget_size(atype) returns the length of the
		     * string with 1 character short. We may think adding 1 to include the string
		     * terminator. But it is not and you have to add 2 to include the terminator.
		     * To be safe, after reading the string, scan the string and make all following
		     * characters to be '\0' from the first '\0' location. */
		    stringlen += 2;
		    *attr_cnt = (int)stringlen;

		    break;
		case H5T_TIME:
		    /*printf(" %s\n", "datatype class : H5T_TIME\n");*/
		    break;
		case H5T_BITFIELD:
		    /*printf(" %s\n", "datatype class : H5T_BITFIELD\n");*/
		    break;
		case H5T_OPAQUE:
		    /*printf(" %s\n", "datatype class : H5T_OPAQUE\n");*/
		    break;
		case H5T_COMPOUND:
		    /*printf(" %s\n", "datatype class : H5T_COMPOUND\n");*/
		    break;
		case H5T_REFERENCE:
		    /*printf(" %s\n", "datatype class : H5T_REFERENCE");*/
		    break;
		case H5T_ENUM:
		    /*printf(" %s\n", "datatype class : H5T_ENUM\n");*/
		    break;
		case H5T_VLEN:
		    /*printf(" %s\n", "datatype class : H5T_ENUM\n");*/
		    break;
		case H5T_ARRAY:
		    /*printf(" %s\n", "datatype class : H5T_ARRAY\n");*/
		    break;
		case H5T_NO_CLASS:
		    /*printf(" %s\n", "datatype class : H5T_NO_CLASS\n");*/
		    break;
		case H5T_NCLASSES:
		    /*printf(" %s\n", "datatype class : H5T_NCLASSES\n");*/
		    break;
		default:
		    /*printf("undefined datatype class name in HDF5_attr_info(), exit...\n");
		    exit(0)*/;
	   } /* end switch */
	}	
    }
    else   /* H5S_NO_CLASS or H5S_NULL */ 
	/*printf("\n")*/;

    /* Release all identifiers.*/
    H5Tclose(atype);
    H5Sclose(aspace);
    H5Aclose(attr_id);

    return 0;
}

herr_t H5_SDreadattr(hid_t loc_id, const char *obj_name, int idx, int attr_cnt, void *attr_buf)
{
/* Maosheng Zhao, SSAI, 4/9/2018
 * Usage: H5_SDreadattr(sds_info[isds].sds_id, ".", k, attr_cnt, &attr_buf);
 * requires functions:
 * get_HDF5_data_type_info()
 * attr_cnt is the number of numerical values for numertical types but the length of the strings for string class
 * attr_H4_type includes numerical types of DFNT_INT8, DFNT_UINT8, DFNT_INT16, DFNT_UINT16, DFNT_INT32, DFNT_UINT32, DFNT_FLOAT32, DFNT_FLOAT64
 * DFNT_CHAR8 for string only
 */ 
    hid_t attr_id, atype, aspace, atype_mem;  /* Attribute, datatype and dataspace identifiers */
    int  rank, num_scales;
    hsize_t sdim[H5S_MAX_RANK];
    herr_t ret;
    int i, j;
    size_t npoints, stringlen;             /* Number of elements in the array attribute. */
    DATA_TYPE_INFO data_type_info;
    ssize_t name_size;
    hsize_t attr_storage_size;
    H5T_cset_t cset;
    H5S_class_t space_type;
    int attr_H4_type;
    char attr_name[MAX_SDS_NAME_LEN]; 

    /* get the attr_name */
    name_size = H5Aget_name_by_idx(loc_id, obj_name, H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)idx, NULL, 0, H5P_DEFAULT);
    if(name_size < 0) return -1;
    name_size += 1;
    H5Aget_name_by_idx(loc_id, obj_name, H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)idx, attr_name, (size_t)name_size, H5P_DEFAULT);

    /* Open the attribute using its name */
    attr_id = H5Aopen(loc_id, attr_name, H5P_DEFAULT);
    if(attr_id < 0) return -1;

    /* Get attribute datatype, dataspace, rank, and dimensions.*/
    atype  = H5Aget_type(attr_id);
    aspace = H5Aget_space(attr_id);
    if(atype < 0 || aspace < 0) return -1;
    rank = H5Sget_simple_extent_dims(aspace, sdim, NULL);
    space_type = H5Sget_simple_extent_type(aspace);
    if(space_type < 0) return -1;

    /*  Display rank and dimension sizes for the array attribute.*/
    /*
    if(rank > 0) {
	printf("rank : %d \n", rank);
	printf("dimension sizes : ");
	for (i=0; i< rank; i++)
	  printf("%d ", (int)sdim[i]);
	printf("\n");
    }
    */

    /* initialize output arguments */
    attr_H4_type = 99;
    if(space_type == H5S_SCALAR || space_type == H5S_SIMPLE)
    {
	/* Read array attribute and display its type and values.
	 * H5Tget_class() returns class names: H5T_INTEGER, H5T_FLOAT, H5T_STRING, or H5T_BITFIELD.*/
	data_type_info = get_HDF5_data_type_info(atype, attr_name);

	/* Read array attribute and display its type and values.
	* H5Tget_class() returns class names: H5T_INTEGER, H5T_FLOAT, H5T_STRING, or H5T_BITFIELD.*/
	if(data_type_info.type_class == H5T_INTEGER || data_type_info.type_class == H5T_FLOAT) {
	    /* read the attr to the buffer */
	    if(attr_cnt > 0)
	    {
	      ret = H5Aread(attr_id, data_type_info.num_type, attr_buf);
	      if(ret < 0)
	      {
		/*fprintf(stderr, "cannot read using H5Aread() for H5T_INTEGER or H5T_FLOAT type with len %d.\n", attr_cnt);*/
		return -1;
	      }
	    }

	    if(data_type_info.num_type == H5T_NATIVE_SCHAR)
		attr_H4_type = DFNT_INT8;
	    else if(data_type_info.num_type == H5T_NATIVE_SHORT)
	  	attr_H4_type = DFNT_INT16;
	    else if(data_type_info.num_type == H5T_NATIVE_INT)
	  	attr_H4_type = DFNT_INT32;
	    else if(data_type_info.num_type == H5T_NATIVE_LONG)
	  	/*printf("datatype : int64\n")*/;
	    else if(data_type_info.num_type == H5T_NATIVE_UCHAR)
	  	attr_H4_type = DFNT_UINT8;
	    else if(data_type_info.num_type == H5T_NATIVE_USHORT)
	  	attr_H4_type = DFNT_UINT16;
	    else if(data_type_info.num_type == H5T_NATIVE_UINT)
	  	attr_H4_type = DFNT_UINT32;
	    else if(data_type_info.num_type == H5T_NATIVE_ULONG)
	  	/*printf("datatype : uint64\n")*/;
	    else if(data_type_info.num_type == H5T_NATIVE_FLOAT)
	  	attr_H4_type = DFNT_FLOAT32;
	    else if(data_type_info.num_type == H5T_NATIVE_DOUBLE)
	  	attr_H4_type = DFNT_FLOAT64;
	    else
	    {
	  	/*printf("undefined HDF5 data types in attr_info(), exit...\n");
	  	exit(0)*/;
	    }

	} else {  /* non-numerical class types */
	    if(data_type_info.type_class == H5T_STRING)
	    {
		if(attr_cnt > 0)
		{
		  /*attr_cnt += 1;*/ /* for terminator */
		  atype_mem = H5Tcopy(H5T_C_S1);

		  /* Maosheng Zhao, 2/27/2018 get and set character systrem of either ASCII or UTF8*/
		  cset = H5Tget_cset(atype);
		  H5Tset_cset(atype_mem, cset);

		  ret = H5Tset_size (atype_mem, attr_cnt);
		  ret = H5Aread(attr_id, atype_mem, attr_buf);
		  if(ret < 0)
		  {
		     /*fprintf(stderr, "cannot read using H5Aread() for H5T_STRING type with len %d.\n", attr_cnt);*/
		     return -1;
		  }
		  ret = H5Tclose(atype_mem);

		  /* scan the string and make all following characters to be '\0' 
 		   * from the first '\0' location */
		  j = -1;
		  for(i=0; i<attr_cnt; i++)
		  {
		     if(((char*)attr_buf)[i] == '\0')
		     {
			j = i;
			if(j >= 0)
			  ((char*)attr_buf)[i] = '\0';
		     }
		  }
		}
	    }
	    else;
	}
    }
    else ;/* H5S_NO_CLASS or H5S_NULL */ 

    /* Release all identifiers.*/
    H5Tclose(atype);
    H5Sclose(aspace);
    H5Aclose(attr_id);

    return 0;
}

DATA_TYPE_INFO get_HDF5_data_type_info(hid_t tid, char *name)
{
/* requires funtcions:
 * byte_order()
 */ 
    hid_t       p_type=-1;
    H5T_class_t type_class;
    size_t      size;
    H5T_sign_t  sign;
    DATA_TYPE_INFO data_type_info;

    type_class = H5Tget_class(tid);
    size       = H5Tget_size(tid);
    sign       = H5Tget_sign(tid);

    if(!byte_order())  /* little-endian */
        data_type_info.endian = 0;
    else
        data_type_info.endian = 1;

    switch(type_class) {
	case H5T_INTEGER:
	    if ( size == 1 && sign == H5T_SGN_2) {
		p_type=H5T_NATIVE_SCHAR;
	    }	
	    else if ( size == 2 && sign == H5T_SGN_2) {
		p_type=H5T_NATIVE_SHORT;
	    }	
	    else if ( size == 4 && sign == H5T_SGN_2) {
		p_type=H5T_NATIVE_INT;
	    }	
	    else if ( size == 8 && sign == H5T_SGN_2) {
		p_type=H5T_NATIVE_LONG;
	    }
	    else if ( size == 1 && sign == H5T_SGN_NONE) {
		p_type=H5T_NATIVE_UCHAR;
	    }	
	    else if ( size == 2 && sign == H5T_SGN_NONE) {
		p_type=H5T_NATIVE_USHORT;
	    }	
	    else if ( size == 4 && sign == H5T_SGN_NONE) {
		p_type=H5T_NATIVE_UINT;
	    }	
	    else if ( size == 8 && sign == H5T_SGN_NONE) {
		p_type=H5T_NATIVE_ULONG;
	    }	
	    else {
		fprintf(stderr, "unknown data type in H5T_INTEGER, exit...\n");
		exit(0);
	    }
	    break;

	case H5T_FLOAT:
	    if ( size == 4) {
		p_type=H5T_NATIVE_FLOAT;
	    }	
	    else if ( size == 8) {
		p_type=H5T_NATIVE_DOUBLE;
	    }
	    else {
	        fprintf(stderr, "unknown data type in H5T_FLOAT, exit...\n");
		exit(0);
	    }
	    break;

	case H5T_TIME:
	case H5T_BITFIELD:
	case H5T_OPAQUE:
	case H5T_STRING:
	    size++; /* Make room for null terminator */
   	    break;
	case H5T_COMPOUND:
	case H5T_REFERENCE:
	case H5T_ENUM:
	case H5T_VLEN:
	case H5T_ARRAY:
	    break;

	case H5T_NO_CLASS:
	case H5T_NCLASSES:
	default:
	    fprintf(stderr, "undefined datatype class in get_HDF5_data_type_info() for %s, exit...\n", name);
	    exit(0);

    } /* end switch */

    /*return p_type;*/
    data_type_info.num_type = p_type;
    data_type_info.type_class = type_class;
    data_type_info.size = size;
    data_type_info.sign = sign;

    return data_type_info;
}
unsigned char byte_order(void)
{
    /* written: Maosheng Zhao */
    /* Date: Jan. 09, 2017 */
    /* check how the bytes of the datatype are layed out in memory for the machine you are using */
    unsigned short Endian = 1;
    char *little = NULL;
    unsigned char bigEndian;

    /* check the byte oder of the native machine */
    little = (char*)&Endian;
    if( *little==0)
    {
	bigEndian=1;  /* it is bigEndian machine, which stores the most significant byte in the smallest address */
	/*printf("This is a big Endian machine\n");*/
    }
    else
    {
	bigEndian=0;  /* it is little Endian machine, which stores the least significant byte in the smallest address */
	/*printf("This is a little Endian machine\n");*/
    }	

    return(bigEndian);
}


