/****************************************************************************
!C

!File: read_l2g.c

!Description:
  Contains routines for reading l2g.

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original January/February 1998. Version 1.0
  Modified March/April 1998 (Incorporated new command line format)
			    (Few bugs fixed)
  Modified June 1998 Version 1.1 (Changes to overall organization. Chnages to 
		improve the speedup.)
  Modified September 1999. (included -meta option)
  Modified March 2000 (included -ptr option)
  Modified September 2001 (fixed bug while importing to Linux)
  Updated March 2013 by Xiaoping Zhang, Sigma Space Corp. 
  	(made it work for NPP)

  Modified February, 2018 Version 3.5.1 by Maosheng Zhao, SSAI
        1)modify it so that HDF5 can be read
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c and read_l2g_comp.c

Maosheng Zhao, SSAI. Note on 3/06/2018
Unlike C6 L2G MxD09, L2G VNP09GHKI and VNP09GHKM in HDF4 don't have ArchiveMetadata.0
and CoreMetadata.0 in the file attribute. Thus the previous method to determine L2G file or not 
won't work for a general purpose in read_l2g and read_l2g_lite tool. The new
method is below
1) use L2G_format_check() to check if it is L2G lite or L2G.
2) get number of layers from maximum_observations SDSs not from meta or attribute data
3) users can output SDSs and the corresponding number layers with -help filenames
4) use whether there is existence of ArchiveMetadata.0 or CoreMetadata.0 to determine
   MODIS versus VNP or old way vs new way in HDF file meta data system
5) treat -layer and -g* option differently. For -layer option, there is
   no need to get the input point file for obtaining meta data but instead
   to get max_obs by using maximum of SDS named "num_observations".
6) turn on -layer option and all layers will be decompacted if without =num
7) In the end, I found out there is no hard rule to absolutely distinguish between
   L2G heavy and L2G lite for -layer option without prior knowledge of ESDTs.
   Thus, to facilitate average users, for read_l2g, if there are more than one SDS
   with "num_observations" in the SDS names' profile, the tool exits and asks users
   to turn to read_l2g_lite. This is because current read_l2g just has one max_obs
   whereas read_l2g_lite can have three resolutions: 1km, 500m and 250m.
8) solve the issue that when there was no compact layers or just one layer, add_obs SDS
   should not be read at all but current versions prints out lots of error information
   due to failure of reading add_obs SDSs.
9) make the -gid works and works correctly for both MODIS and VIIRS.
10) make -ptr= works for -gid case by using arg_list[5] to store the point file name,
    either derived and existing or specified and existing. The the previous arg_list[5+*]
    has been changed to arg_list[6+*]

# MSZ 3/7/2018, Note on difference and same between read_l2g and read_l2g_lite:
For L2G heavy or L2G lite file with just one SDS with "num_observation" and under -layer option,
both read_l2g and read_l2g_lite tools can handle it.  For just -layer option, read_l2g_lite
can handle both L2G heavy and L2G lite files. Only for granule option -g*, just read_l2g can do it.

  Modified February, 2020 Version 3.6 by Maosheng Zhao, SSAI

1) correct a bug error when running read_l2g and the bug is in main_util.c.
   In function find_file(), the 
     char tmp_fname[80]; is updated to char tmp_fname[1024];
   which is called by get_pnt_fname(), and get_pnt_fname is called by
   get_l2g_datac_by_gran() in read_l2g_comp.c where as get_l2g_datac_by_gran()
   is called in read_l2g.c
2) correct issue for get_sds_names_num_observations() and
   read_input_sds_names_ress() because for some VNP l2g, the dimension of the tile
   data is not 1200, 2400, or/and 4800 but others.
3) increase defined string length to avoid crash as below
   in read_l2g.c, get_npp_gran_id()
   if ((iptr_fnames = (char **)Calloc2D(20, MAX_META_VAL_LEN, sizeof(char))) == NULL) changed to
   if ((iptr_fnames = (char **)Calloc2D(127, MAX_META_VAL_LEN, sizeof(char))) == NULL)

   127 has been change to MAX_NUM_GRAN
   l2g.h:#define MAX_NUM_GRAN 20
     to 127
   change several 20 in read_l2g_comp.c to MAX_NUM_GRAN

4) enable output HDF4 with internal compression


!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "hdf5.h"
#include "l2g.h"
#include "meta.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "str_op.h"
#include "qa_tool.h"
#include "main_util.h"
#include "alloc_mem.h"

#define HELP \
"NAME\n" \
"     read_l2g - read data from L2G files\n" \
"\n" \
"SYNOPSIS\n" \
"     read_l2g -help\n" \
"     read_l2g -help [filenames]\n" \
"     read_l2g -layer|glist|gpidx=item1[,item2...]] -of=output_file [-meta]\n" \
"              [-sds=SDSname1[,SDSname2...]]] [-ptr=pointer file] filename\n" \
"\n" \
"     echo -layer|glist|gpidx=item1[,item2...]] -of=output_file [-meta]\n" \
"          [-sds=SDSname1[,SDSname2...]]] [-ptr=pointer file] filename | read_l2g\n" \
" or  read_l2g < argument_file\n" \
"\n" \
"DESCRIPTION\n" \
"     Extracts a subset of parameters (full format or compact format) from a\n" \
"     L2G file, selecting pixels according to the layer number, or the identity\n" \
"     of the L2 granule associated to the pixels. The output is an HDF file\n" \
"     where all the selected items are stored as separate SDS. \n" \
"     To read observations by identity of granules user can provide the pointer \n" \
"     filename using the -ptr option else directory paths in the environment variable \n" \
"     MODISPATH will be searched for the required pointer file. \n" \
"\n" \
"OPTIONS\n" \
"     -help                          Display this help message or all -sds names when file name is also used\n" \
"     -layer[=layer1[,layer2...]]    Select by layer number (0-based)\n" \
"     -glist=orbit,gran1[,gran2...]  Select by granule number\n" \
"                                    Enter orbit first, then granule list\n" \
"                                    Repeat option to select multiple orbits\n" \
"     -gpidx=index1[,index2...]      Select by granule pointer index\n" \
"     -gid=granule1[,granule2...]    Select by granule ID (e.g., Ayyyyddd.hhmm)\n" \
"     -sds=<SDS list>                List of SDS to read (separated by commas)\n" \
"                                    Enter all to read all SDS\n" \
"     -of=filename                   Output file\n" \
"     -ptr=pointer filename          Pointer filename with full path or the\n" \
"                                    directory containing the pointer file. \n" \
"                                    Required only if reading parameters by\n" \
"                                    the identity of the L2 granule. \n" \
"                                    This option is not needed for NPP \n" \
"     -meta                          Copy metadata from input file to output \n" \
" \n" \
"Examples: \n" \
"       read_l2g -layer ./DATA/MODL2G/MOD09GHK.A1996214.h12v04.002.hdf -of=l2g_layers.hdf \n" \
"       read_l2g -layer -sds=sur_refl_b03 ./DATA/MODL2G/MOD09GHK.A1996214.h12v04.002.hdf -of=l2g_layers.hdf \n" \
"	read_l2g -layer=0-5 -sds=sur_refl_b03 ./DATA/MODL2G/MOD09GHK.A1996214.h12v04.002.hdf -of=l2g_layers.hdf \n" \
"	read_l2g -glist -sds=sur_refl_b01 -of=test_b01_gt.hdf MOD09GHK.A2018005.h12v02.006.2018009190349.hdf \n" \
"	read_l2g -glist=96017,232,233 -sds=sur_refl_b01 -of=test_b01_gt.hdf MOD09GHK.A2018005.h12v02.006.2018009190349.hdf \n" \
"	read_l2g -gpidx=0,1,2 -sds=sur_refl_b03 ./DATA/MODL2G/MOD09GHK.A1996214.h12v04.002.hdf -of=l2g_granules.hdf \n" \
"	read_l2g -gid -sds=SurfReflect_M1 -of=test_G1KI.hdf VNP09G1KI.A2018005.h12v02.001.2018008142514.hdf \n" \
"	read_l2g -gid=A1996214.1650 -sds=sur_refl_b03 ./DATA/MODL2G/MOD09GHK.A199614.h12v04.002.hdf -of=l2g_granules.hdf \n" \
"	read_l2g -gpidx -sds=sur_refl_b03 /YDAY/MOD09GHK/MOD09GHK.A1999053.h08v06.001.1999230135638.hdf -of=l2g_gran.hdf -ptr=/YDAY/MODPNTR/1999053/MODPTHKM.A1999053.h08v06.001.1999230133414.hdf \n" \
"	read_l2g -layer=0-8 -sds=SurfReflect_M1  -of=l2g_npp_layers.hdf VNP09G1KM.A2018005.h12v02.001.2018008142514.hdf \n" \
"	read_l2g -gpidx -sds=SurfReflect_M1 -of=test_G1KM_gpidx.hdf VNP09G1KM.A2018005.h12v02.001.2018008142514.hdf \n" \
"	read_l2g -gid=A2012366.1925,A2012366.1930 -sds=SurfReflect_Mod_b02 -of=l2g_NPP_DSRF1KD_L2GD_gid.hdf NPP_DSRF1KD_L2GD.A2012366.h09v05.C1_03001.2013005080512.hdf \n" \
"	read_l2g -glist -sds=SurfReflect_M1 -of=test_G1KI.hdf VNP09G1KI.A2018005.h12v02.001.2018008142514.hdf \n" \
"	read_l2g -glist=32084,1906,1912 -sds=SurfReflect_M1 -of=test_G1KI.hdf VNP09G1KI.A2018005.h12v02.001.2018008142514.hdf \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"Usage:      read_l2g -layer|glist|gpidx=item1[,item2...]] -of=output_file \n" \
"            [-sds=SDSname1[,SDSname2...]]] [-ptr=pointer file] -meta filename \n" \
" \n" \
"OPTIONS \n" \
"     -help                          Display this help message or all -sds names and their max number of layers if file name is also used\n" \
"     -layer[=layer1[,layer2...]]    Select by layer number (0-based) \n" \
"     -glist=orbit,gran1[,gran2...]  Select by granule number \n" \
"                                    Enter orbit first, then granule list \n" \
"                                    Repeat option to select multiple orbits \n" \
"     -gpidx=index1[,index2...]      Select by granule pointer index \n" \
"     -gid=granule1[,granule2...]    Select by granule ID (e.g., Ayyyyddd.hhmm)\n" \
"     -sds=<SDS list>                List of SDS to read (separated by commas) \n" \
"                                    Enter all to read all SDS \n" \
"     -of=filename                   Output file \n" \
"     -ptr=pointer filename          Pointer filename \n" \
"                                    Required only if reading parameters by\n" \
"     -meta                          Copy metadata from input file to output \n"


int parse_cmd_read_l2g(int argc, char **argv, char **sds_names, int *sds_cnt, int *if_cnt, char ***num_str,
	int *run_cnt, char *out_fname, char *ptr_fname, int *m_opt, int *ptr_fname_existence);
int get_sds_names_num_observations(char *fname, char **SDS_names_num_obs);
int check_param(char *fname, char **sds_names, int sds_cnt, char *r_mode, 
	char **num_str, int run_cnt, char **arg_list, int *ptr_fname_existence, char *ptr_fname);
int check_param_npp(char *fname, char **sds_names, int sds_cnt, char *r_mode, 
	char **num_str, int run_cnt, char **arg_list, int *ptr_fname_existence, char *ptr_fname);	
int read_input_sds_names_ress(char *fname, char **in_sds_names, int sds_cnt, char **out_sds_names, int *res);
int get_l2g_max_obs_layers(char *hdf_fname, char *max_obs_sds_name);
void read_input_sds_names(char *fname, char **sds_names, int *sds_cnt);
void get_layer_or_granptr_numbers(char **arg_list, int max_nobs, int run_cnt, 
	char **num_str, char *r_mode, int m_cnt);
void get_granorb_numbers(char **arg_list, int max_nobs, char *orb_narr_str, 
	char *gran_narr_str, char *pnt_narr_str, int run_cnt, char **num_str);
void get_gran_id(char **arg_list, int max_nobs, char *in_ptr, int run_cnt, 
	char **num_str, int *granulepointerarr, int num_in_gran);
void get_npp_gran_id(char **arg_list, int max_nobs, char *in_ptr, int run_cnt, 
	char **num_str, int *granulepointerarr, int num_in_gran, char* pnt_file_name);	
int check_gran_orb_num(int *gn, int on, int *pn, char *gran_narr_str, 
	char *orb_narr_str, char *pnt_narr_str, int max_nobs);
void get_l2g_dataf_by_obs(char **arg_list, int32 *out_sd_id, int if_cnt, int *m_opt, int H5OUT);
void get_l2g_datac_by_obs(char **arg_list, int32 *out_sd_id, int if_cnt, int *m_opt, int H5OUT, int max_nobs);
void get_l2g_dataf_by_gran(char **arg_list, char *ptr_fname, int32 *out_sd_id, int if_cnt, int *m_opt, int H5OUT);
void get_l2g_datac_by_gran(char **arg_list, char *ptr_fname, int32 *out_sd_id, int if_cnt, int *m_opt, int H5OUT, int max_nobs);
int get_npp_GranulePointerArray(char *fname, int *iptr_array);
void get_npp_granorb_numbers(char **arg_list, int max_nobs, char *orb_narr_str, 
	char *gran_narr_str, char *pnt_narr_str, int run_cnt, char **num_str);
int check_npp_gran_orb_num(int *gn, int on, int *pn, char *gran_narr_str, 
	char *orb_narr_str, char *pnt_narr_str, int max_nobs);
	
int main(int argc, char *argv[])
{
  sds_t out_sds_info;
  int i, ir;
  int fid, isds;
  int m_opt, status, st;
  int if_cnt, sds_cnt;
  int run_cnt[4], stdin_argc;
  char r_mode[10], ***num_str;
  char **stdin_argv = NULL, **arg_list = NULL;
  char **sds_names = NULL, **tmp_sds_names = NULL;
  char out_fname[MAX_PATH_LENGTH] = {'\0'}, ptr_fname[MAX_PATH_LENGTH] = {'\0'};
  int nsds, max_nobs, res[MAX_NUM_SDS];
  int H5OUT = 0; /* so far, output to HDF4 */
  char fname_tmp[MAX_PATH_LENGTH] = {'\0'};
  int ptr_fname_existence;
  char **SDS_names_num_obs = NULL;         /* MSZ 3/6/2018 */
  int nsds_num_observations;               /* MSZ 3/6/2018 */
  char SDS_name_num_obs[MAX_SDS_NAME_LEN] = {'\0'};

  /* MSZ 2/21/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 2/21/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if ( (argc==1) || ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0))))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  else
  {
    /* Print all SDS names when -help option is input with the input filename */
    for (i=0, status=0; i<argc; i++)
      if ((strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "-help") == 0)) status = i;

    if (status == 1)
    {
      /* MSZ added the below on 2/21/2018 to display the file content of L2G */
      /* Print number of layers in the input L2G file */
      for (fid=1; fid<argc; fid++)
        if (argv[fid][0] != '-')
        {
	  /* just for L2G file name patterns */
	  if (L2G_format_check(argv[fid]) >= 1)
	  {
	    if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
	      fprintf(stderr, "Cannot allocate memory for sds_names in read_l2g_lite(main)\n");
	    if ((tmp_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
	      fprintf(stderr, "Cannot allocate memory for tmp_sds_names in read_l2g_lite(main)\n");

	    sds_cnt = 0;
	    nsds = read_input_sds_names_ress(argv[fid], sds_names, sds_cnt, tmp_sds_names, res);

	    fprintf(stderr, "\n\n##############################################################################\n");
	    fprintf(stderr, "Below are the details on the SDSs and their number layers in this L2G file %s:\n", argv[fid]);

	    for (isds=0; isds<nsds; isds++)
	    {
	      max_nobs = 0;
	      if(res[isds] == 1)
	      {
		max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs_1km);
		if(max_nobs == -1)
		  max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs);

		fprintf(stderr, "1km SDS: %s, max_number_layers: %d\n", tmp_sds_names[isds], max_nobs);
	      }
	      else if (res[isds] == 2)
	      {
		max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs_500m);
		if(max_nobs == -1)
		  max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs);

		fprintf(stderr, "500m SDS: %s, max_number_layers: %d\n", tmp_sds_names[isds], max_nobs);
	      }
	      else if (res[isds] == 4)
	      {
		max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs);
		if(max_nobs == -1)
		  max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs);

		fprintf(stderr, "250m SDS: %s, max_number_layers: %d\n", tmp_sds_names[isds], max_nobs);
	      }
	    } /* for (isds=0; isds<nsds; isds++) */

	    if(sds_names) Free2D((void**)sds_names);
	    if(tmp_sds_names) Free2D((void**)tmp_sds_names); 
	    fprintf(stderr, "##############################################################################\n\n\n");
	  } /* if ((strstr(argv[fid], "09GA.") != NULL) || (strstr(argv[fid], "09GQ.") != NULL)) */
	  else
	  {
	    fprintf(stderr, "%s is not L2G data product\n", argv[fid]);
	    exit(0);
	  }
        } /* if (argv[fid][0] != '-') */

      exit(0);	
    }
  }

  /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
  /* start the real work */
  if ((stdin_argv = (char **)Calloc2D(MAX_NUM_PARAM, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for stdin_argv in data_fedex(main)\n");
    exit(1);
  }

  if (argc == 1)
  {
    if (parse_stdin(&stdin_argc, stdin_argv) == -1)
    {
      Free2D((void *)stdin_argv);
      exit(0);
    }
  }
  else
  {
    if (argc >= MAX_NUM_PARAM)
    {
      stdin_argc = MAX_NUM_PARAM;
      fprintf(stderr, "Too many input arguments. Using first %d arguments only . . \n", stdin_argc);
    }
    else stdin_argc = argc;

    for (i=0; i<stdin_argc; i++)
      strcpy(stdin_argv[i], argv[i]);
  }

  if ((SDS_names_num_obs = (char **)Calloc2D(4, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for SDS_names_num_obs\n");
  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_names in read_l2g(main)\n");
  if ((num_str = (char ***)Calloc3D(4, MAX_NUM_LAYERS, 50, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for num_str in read_l2g(main)\n");
  if ((sds_names != NULL) || (num_str != NULL))
  {
    if ((status = parse_cmd_read_l2g(stdin_argc, stdin_argv, sds_names, &sds_cnt, 
		&if_cnt, num_str, run_cnt, out_fname, ptr_fname, &m_opt, &ptr_fname_existence)) == -1)
      fprintf(stderr, "%s\n", USAGE);
    else if (status != 0)
    {
      if ((arg_list = (char **)Calloc2D(MAX_NUM_PARAM, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
        fprintf(stderr, "Cannot allocate memory for arg_list in read_l2g\n");
      if ((tmp_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
        fprintf(stderr, "Cannot allocate memory for tmp_sds_names in read_l2g\n");
      out_sds_info.H5OUT = 0;   /* so far, output to HDF4 */
      out_sds_info.sd_id = out_sds_info.sds_id = -1;
      if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
	fprintf(stderr, "Cannot create output HDF file %s in read_l2g(main)\n", out_fname);
      if ((arg_list != NULL) && (tmp_sds_names != NULL) && (out_sds_info.sd_id != FAIL))
      {
        for (fid=1; fid<stdin_argc; fid++)
        {
   	  if (stdin_argv[fid][0] == '-') continue;

	  if((nsds_num_observations = get_sds_names_num_observations(argv[fid], SDS_names_num_obs)) == 1) /* MSZ 3/6/2018, add this condition */
	  {
	    /* get max_obs from that one SDS */
	    for(i=0; i<4; i++)
	      if(SDS_names_num_obs[i][0] != '\0')
		strcpy(SDS_name_num_obs, SDS_names_num_obs[i]);

	    max_nobs = get_l2g_max_obs_layers(argv[fid], SDS_name_num_obs);
	
            fprintf(stderr, "\nProcessing input file %s\n", stdin_argv[fid]);
            fprintf(stderr, "---------------------------------------------------------------\n");
	    /* MSZ 3/3/2018, MODIS or VNP L2G */
	    /*st = strpos(stdin_argv[fid], "_L2G", 0); */
	    if(file_attr_name_existence(stdin_argv[fid], "ArchiveMetadata") ||
		 file_attr_name_existence(stdin_argv[fid], "CoreMetadata"))
		st = -1;  /* MODIS */
	    else
		st = 1;   /* VNP or some MODIS */

            for (isds=0; isds<sds_cnt; isds++)
	      strcpy(tmp_sds_names[isds], sds_names[isds]);

	    for (ir=0; ir<4; ir++)
	    {
	      if (run_cnt[ir] == 0) continue;
	      switch (ir)
	      {
	        case 0: strcpy(r_mode, "-layer=");
		      break;
	        case 1: strcpy(r_mode, "-glist=");
		      break;
	        case 2: strcpy(r_mode, "-gpidx=");
		      break;
	        case 3: strcpy(r_mode, "-gid=");
		      break;
	      }
              if (st == -1) /* MODIS */
	        status = check_param(stdin_argv[fid], tmp_sds_names, sds_cnt, r_mode, 
			             num_str[ir], run_cnt[ir], arg_list, &ptr_fname_existence, ptr_fname);
	      else          /* VNP */
		status = check_param_npp(stdin_argv[fid], tmp_sds_names, sds_cnt, r_mode,
		                         num_str[ir], run_cnt[ir], arg_list, &ptr_fname_existence, ptr_fname);

	      /* MSZ 3/13/2018, very critical step */
	      strcpy(arg_list[5], ptr_fname);

              if (status != 0)
              {
                strcpy(arg_list[0], stdin_argv[fid]);
	        strcpy(arg_list[1], stdin_argv[fid]);
	        rm_path(arg_list[1]);
	        arg_list[1][24] = '\0'; 
	        fprintf(stderr, "Reading from HDF file %s\n", stdin_argv[fid]);
                if (strcmp(r_mode, "-layer=") == 0)
	        {
	          if (strcmp(arg_list[2], "full") == 0)
                    get_l2g_dataf_by_obs(arg_list, &out_sds_info.sd_id, if_cnt, &m_opt, out_sds_info.H5OUT);
	          else
                    get_l2g_datac_by_obs(arg_list, &out_sds_info.sd_id, if_cnt, &m_opt, out_sds_info.H5OUT, max_nobs);
	        }
                else
	        {
	          if (st != -1) strcpy(ptr_fname, "all");
		  if (strcmp(arg_list[2], "full") == 0)
                    get_l2g_dataf_by_gran(arg_list, ptr_fname, &out_sds_info.sd_id, if_cnt, &m_opt, out_sds_info.H5OUT); 
	          else
		  {
                    get_l2g_datac_by_gran(arg_list, ptr_fname, &out_sds_info.sd_id, if_cnt, &m_opt, out_sds_info.H5OUT, max_nobs); 
		  }
	        }
              } /* if (status = . . .  */
	    }  /* for (r_mode=0; . . . */
	  } /* if((nsds_num_observations = get_sds_names_num_observations(argv[fid], SDS_names_num_obs)) == 1) */
	  else
          {
            fprintf(stderr, "Please use the other tool, read_l2g_lite, because for L2G, there should be just one SDS with \"num_observations\" as whole or part of name for L2G, %s is not L2G, exit...\n", argv[fid]);
            exit(0);
          }
        } /* for (fid=0 . . . . */
	close_sd(&out_sds_info, out_sds_info.H5OUT);
      }
      if(tmp_sds_names) Free2D((void **)tmp_sds_names);
      if(arg_list) Free2D((void **)arg_list);
    }
  }
  if(SDS_names_num_obs) Free2D((void **)SDS_names_num_obs);
  if(num_str) Free3D((void ***)num_str);
  if(sds_names) Free2D((void **)sds_names);
  if(stdin_argv) Free2D((void **)stdin_argv);
  fprintf(stderr, "Process read_l2g completed \n");
  return 0;
}

int parse_cmd_read_l2g(int argc, char **argv, char **sds_names, int *sds_cnt, int *if_cnt, 
		char ***num_str, int *run_cnt, char *out_fname, char *ptr_fname, int *m_opt, int* ptr_fname_existence)
{
  int opt_st;
  int status;
  int i, j, k;
  int tmp_run_cnt;
  char **tmp_num_str;
  status = 1;
  *m_opt = 0;
  *if_cnt = 0;
  *sds_cnt = 0;
  out_fname[0] = '\0';
  ptr_fname[0] = '\0';
  for (i=0; i<4; i++)
    run_cnt[i] = 0;

  if ((tmp_num_str = (char **)Calloc2D(MAX_NUM_LAYERS, 50, sizeof(char))) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for tmp_num_str in parse_cmd_read_l2g\n");
    return 0;
  }

  *ptr_fname_existence = 0;
  for (i=1; i<argc; i++)
  {
    if ((is_arg_id(argv[i], "-of=") == 0) || (is_arg_id(argv[i], "-o=") == 0))
      get_arg_val(argv[i], out_fname);

    else if (is_arg_id(argv[i], "-sds=") == 0)
      get_arg_val_arr(argv[i], sds_names, sds_cnt);
    else if (is_arg_id(argv[i], "-sds") == 0) ;

    else if (is_arg_id(argv[i], "-layer=") == 0)
    {
      get_arg_val_arr(argv[i], num_str[0], &run_cnt[0]);
      if (run_cnt[0] == 0) strcpy(num_str[0][0], "all");
    }
    else if (strcmp(argv[i], "-layer") == 0) strcpy(num_str[0][0], "all");

    else if (is_arg_id(argv[i], "-glist=") == 0)
    {
      tmp_run_cnt = 0;
      get_arg_val_arr(argv[i], tmp_num_str, &tmp_run_cnt);
      if (tmp_run_cnt == 0) strcpy(num_str[1][0], "all");
      else if (tmp_run_cnt == 1)
      {
	sprintf(num_str[1][0], "-1 %s", tmp_num_str[0]);
        run_cnt[1] = run_cnt[1] + 1;
      }
      else if (tmp_run_cnt > 1)
      {
        for (k=0, j=1; j<tmp_run_cnt; k++, j++)
	  sprintf(num_str[1][run_cnt[1] + k], "%s %s", tmp_num_str[j], tmp_num_str[0]);	  
	run_cnt[1] = run_cnt[1] + tmp_run_cnt - 1;	
      }
    }
    else if (strcmp(argv[i], "-glist") == 0) strcpy(num_str[1][0], "all");

    else if (is_arg_id(argv[i], "-gpidx=") == 0)
    {
      get_arg_val_arr(argv[i], num_str[2], &run_cnt[2]);
      if (run_cnt[2] == 0) strcpy(num_str[2][0], "all");
    }
    else if (strcmp(argv[i], "-gpidx") == 0) strcpy(num_str[2][0], "all");

    else if (is_arg_id(argv[i], "-gid=") == 0)
    {
      get_arg_val_arr(argv[i], num_str[3], &run_cnt[3]);
      if (run_cnt[3] == 0) strcpy(num_str[3][0], "all");
    }
    else if (strcmp(argv[i], "-gid") == 0) strcpy(num_str[3][0], "all");

    else if (is_arg_id(argv[i], "-ptr=") == 0)
    {
      *ptr_fname_existence = 1;
      get_arg_val(argv[i], ptr_fname);
    }

    else if (strcmp(argv[i], "-meta") == 0) *m_opt = 1;

    else if (argv[i][0] == '-')
      fprintf(stderr, "Unknown option %s\n", argv[i]);
    else
      ++*if_cnt; 
  }

  if (*sds_cnt == 0)
  {
    fprintf(stderr, "No SDS name input. Reading all SDS . . \n");
    *sds_cnt = 1;
    strcpy(sds_names[0], "all");
  }

  for (i=0, opt_st=0; i<4; i++)
  {
    if (run_cnt[i] != 0) opt_st++;
    else if (strcmp(num_str[i][0], "all") == 0)
    {
      opt_st++;
      switch(i)
      {
	case 0: fprintf(stderr, "No values input for option -layer. Reading all layers\n");
	        run_cnt[0] = 1;
		break;
	case 1: fprintf(stderr, "No values input for option -glist. Reading all granules\n");
	        run_cnt[1] = 1;
		break;
	case 2: fprintf(stderr, "No values input for option -gpidx. Reading all granules\n");
	        run_cnt[2] = 1;
		break;
	case 3: fprintf(stderr, "No values input for option -gid. Reading all granules\n");
	        run_cnt[3] = 1;
		break;
      }
    }
  }

  if ((*if_cnt == 0) || (out_fname[0] == '\0') || (opt_st == 0))
    status = -1;
  if (*if_cnt == 0) 
    fprintf(stderr, "Missing input file name \n");
  if (out_fname[0] == '\0')
    fprintf(stderr, "Missing output file name \n");
  if (opt_st == 0)
    fprintf(stderr, "Missing l2g read mode option \n");

  if ((status != -1) && (*sds_cnt == 0))
  {
    fprintf(stderr, "No SDS name input. Reading all SDS . . \n");
    *sds_cnt = 1;
    strcpy(sds_names[0], "all");
  }

  Free2D((void **)tmp_num_str);  
  return status;
}

int check_param(char *fname, char **sds_names, int sds_cnt, char *r_mode, 
	char **num_str, int run_cnt, char **arg_list, int *ptr_fname_existence, char *ptr_fname)
{
  /* MSZ 3/5/2018, SSAI 
   * treat -layer and -g* options differently so that for -layer option, there is
   * no need to get the input point file and read meta data from the point file but
   * get max_obs by using maximum of SDS named "num_observations". */
  int len, i, k;
  int m_cnt, status;
  int max_nobs, tot_num_sds;
  char orb_narr_str[MAX_STR_LEN*4];      /* MSZ 3/8/2017 */
  char pnt_narr_str[MAX_STR_LEN*4];      /* MSZ 3/8/2017 */
  char gran_narr_str[MAX_STR_LEN*8];     /* MSZ 3/8/2017 */
  char meta_name[MAX_META_NAME_LEN];
  char *ameta_str = NULL, *cmeta_str = NULL, **meta_val = NULL;
  int *granulepointerarr = NULL;         /* MSZ 3/12/2018 */
  int num_in_gran = 0;                   /* MSZ 3/12/2018 */

  /* MSZ 3/5/2018 */
  sds_t in_sds_info;
  int32 attr_type, attr_cnt;
  char *l2g_storage_format = NULL;

  /* get arg_list[2] */
  /* get the l2g storage format */
  in_sds_info.sd_id = in_sds_info.sds_id = -1;
  if(open_sd(fname, &in_sds_info, 'R') == FAIL)
    fprintf(stderr, "Cannot open output HDF file %s\n", fname);
  else
  {
    if( !(l2g_storage_format = (char*)find_get_sd_attr(&in_sds_info, "storage_format", &attr_type, &attr_cnt)) )
      l2g_storage_format = (char*)find_get_sd_attr(&in_sds_info, "storageformat", &attr_type, &attr_cnt);

    /* Note that even if meta data for l2g_storage_format is unavailable, it is Ok because arg_list[2] will not be "full"
     * and we assume there are uncompact layers if meta data for storageformat is unavailable */
    if(l2g_storage_format)
    {
      strcpy(arg_list[2], l2g_storage_format);
      if(l2g_storage_format) free(l2g_storage_format);
    }
    else
    {
      if(sds_name_existence_soft(fname, sds_name_nobs) == 1)
	strcpy(arg_list[2], "compact");
      else
	fprintf(stderr, "No information on l2g_storageformat, and assuming compact by default.\n");
    }
  }
  close_sd(&in_sds_info, in_sds_info.H5IN);

  /* get arg_list[3] and arg_list[6+k] */
  if (strcmp(sds_names[0], "all") == 0) sds_cnt = 0; 
  read_input_sds_names(fname, sds_names, &sds_cnt);
  for (i=0, k=0; i<sds_cnt; i++)
    if (sds_names[i][0] != '\0')
    {
      strcpy(arg_list[6+k], sds_names[i]);
      k++;
    }
  if (k == 0) status = 0;
  sprintf(arg_list[3], "%d", k);

  status = 0;
  if (strcasestr(r_mode, "-layer") != '\0')
  {
     m_cnt = 1;

     max_nobs = get_l2g_max_obs_layers(fname, sds_name_nobs);
     fprintf(stderr, "\nMaximumObservations=%d\n\n",max_nobs);
     if(max_nobs == -1) 
     {
	m_cnt = 0;
	fprintf(stderr, "max_nobs is -1, exit...\n");
	exit(0);
	sprintf(arg_list[4], "%d", 0);
     }

     if (strcmp(num_str[0], "all") == 0) run_cnt = 0;
     get_layer_or_granptr_numbers(arg_list, max_nobs, run_cnt, num_str, r_mode, m_cnt);

     status = 1;

     if (atoi(arg_list[4]) == 0)
	 status = 0;
  }
  else  /* for non "-layer" option or "-g*=" options */
  {
    status = 1;
    if ((meta_val = (char **)Calloc2D(5, MAX_META_VAL_LEN, sizeof(char))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for metaval in parse_cmd_read_l2g");
      status =  0;
    }
    if ((ameta_str = get_attr_metadata(fname, "ArchiveMetadata.0")) == NULL)
    {
      fprintf(stderr, "Cannot read L2G without reading archive metadata\n");
      status = 0;
    }
    else if (strpos(ameta_str, "NUMBEROFOVERLAPGRANULES", 0) == -1)
    {
      fprintf(stderr, "%s is not an L2G file\n", fname);
      status = 0;
    }

    if (status != 0)
    {
      if (strcmp(r_mode, "-gpidx=") == 0)
      {
        m_cnt = 0;
        strcpy(meta_name, "NUMBEROFOVERLAPGRANULES");
        get_sel_metadata(ameta_str, meta_name, meta_val, &m_cnt, 0);
        if (m_cnt == 0)
          fprintf(stderr, "Metadata NUMBEROFOVERLAPGRANULES missing in %s\n", fname);
        else max_nobs = atoi(meta_val[0]);
        if (strcmp(num_str[0], "all") == 0) run_cnt = 0;
        get_layer_or_granptr_numbers(arg_list, max_nobs, run_cnt, num_str, r_mode, m_cnt);
      }
      else if (strcmp(r_mode, "-glist=") == 0)
      {
        max_nobs = -1;
        orb_narr_str[0] = gran_narr_str[0] = '\0';
        m_cnt = 0;
        strcpy(meta_name, "NUMBEROFOVERLAPGRANULES");
        get_sel_metadata(ameta_str, meta_name, meta_val, &m_cnt, 0);
        if (m_cnt == 0) 
	  fprintf(stderr, "Missing metadata NUMBEROFOVERLAPGRANULES in %s\n", fname);
        else max_nobs = atoi(meta_val[0]); 
        m_cnt = 0;
        strcpy(meta_name, "ORBITNUMBERARRAY");
        get_sel_metadata(ameta_str, meta_name, meta_val, &m_cnt, 0);
        if (m_cnt == 0) 
          fprintf(stderr, "Missing metadata ORBITNUMBERARRAY in %s\n", fname);
        else strcpy(orb_narr_str, meta_val[0]);
        m_cnt = 0;
        strcpy(meta_name, "GRANULENUMBERARRAY");
        get_sel_metadata(ameta_str, meta_name, meta_val, &m_cnt, 0);
        if (m_cnt == 0) 
          fprintf(stderr, "Missing metadata GRANULENUMBERARRAY in %s\n", fname);
        else strcpy(gran_narr_str, meta_val[0]);
        m_cnt = 0;
        strcpy(meta_name, "GRANULEPOINTERARRAY");
        get_sel_metadata(ameta_str, meta_name, meta_val, &m_cnt, 0);
        if (m_cnt == 0) 
          fprintf(stderr, "Missing metadata GRANULEPOINTERARRAY in %s\n", fname);
        else
	{
	  /* MSZ 3/9/2018, add meta_val[1] to meta_val[4] to the list */
	  strcpy(pnt_narr_str, meta_val[0]);
	  strcat(pnt_narr_str, meta_val[1]);
	  strcat(pnt_narr_str, meta_val[2]);
	  strcat(pnt_narr_str, meta_val[3]);
	  strcat(pnt_narr_str, meta_val[4]);
	}
        if (strcmp(num_str[0], "all") == 0) run_cnt = 0;
        get_granorb_numbers(arg_list, max_nobs, orb_narr_str, gran_narr_str, pnt_narr_str, run_cnt, num_str); 
      }
      else if (strcmp(r_mode, "-gid=") == 0)
      {
        num_in_gran = 0;
        m_cnt = 0;
        strcpy(meta_name, "NUMBEROFINPUTGRANULES");
        get_sel_metadata(ameta_str, meta_name, meta_val, &m_cnt, 0);
        if (m_cnt == 0)
          fprintf(stderr, "Metadata NUMBEROFINPUTGRANULES missing in %s\n", fname);
        else
	{
	  num_in_gran = atoi(meta_val[0]);
	  fprintf(stderr, "NUMBEROFINPUTGRANULES=%s\n", meta_val[0]);
	}

        max_nobs = 0;
        m_cnt = 0;
        strcpy(meta_name, "NUMBEROFOVERLAPGRANULES");
        get_sel_metadata(ameta_str, meta_name, meta_val, &m_cnt, 0);
        if (m_cnt == 0)
          fprintf(stderr, "Metadata NUMBEROFOVERLAPGRANULES missing in %s\n", fname);
        else
	{
	  max_nobs = atoi(meta_val[0]);
	  fprintf(stderr, "NUMBEROFOVERLAPGRANULES=%s\n", meta_val[0]);
	}

        m_cnt = 0;
        if ((cmeta_str = get_attr_metadata(fname, "CoreMetadata.0")) == NULL)
          fprintf(stderr, "Cannot find CoreMetadata in %s", fname);
        else
        {
	  /* MSZ 3/12/2018, added GRANULEPOINTERARRAY info */
          strcpy(meta_name, "GRANULEPOINTERARRAY");
          get_sel_metadata(ameta_str, meta_name, meta_val, &m_cnt, 0);
          if (m_cnt == 0) 
            fprintf(stderr, "Missing metadata GRANULEPOINTERARRAY in %s\n", fname);
          else
	  {
	    /* MSZ 3/9/2018, add meta_val[1] to meta_val[4] to the list */
	    strcpy(pnt_narr_str, meta_val[0]);
	    strcat(pnt_narr_str, meta_val[1]);
	    strcat(pnt_narr_str, meta_val[2]);
	    strcat(pnt_narr_str, meta_val[3]);
	    strcat(pnt_narr_str, meta_val[4]);

	    rm_ln_in_str(pnt_narr_str);
	    rm_character_in_str(pnt_narr_str, '(');
	    rm_character_in_str(pnt_narr_str, ')');
	    rm_character_in_str(pnt_narr_str, ' ');

	    char **arr_strs = NULL;
	    int num_eles = 0;
	    if ((arr_strs = (char **)Calloc2D(MAX_VNP_GRANArray*2, 5, sizeof(char))) == NULL)
	    {
	      fprintf(stderr, "cannot calloc arr_strs, exit\n");
	      exit(0);
	    }
	    else
	      split_string(pnt_narr_str, arr_strs, &num_eles);
	
	    if(num_eles > 0)
	    {
	      if((granulepointerarr = calloc(num_in_gran, sizeof(int))) == NULL)
	      {
	        fprintf(stderr, "cannot calloc granulepointerarr, exit\n");
	        exit(0);
	      }
	      else   /* convert string array to integer array */
	      {
		printf("granule point array:%d\n", num_in_gran);
		for(i = 0; i < num_in_gran; i++)
		{
		  granulepointerarr[i] = atoi(arr_strs[i]);
		  printf(" %d ", granulepointerarr[i]);
		}
		printf("\n");
	      }
	      if(arr_strs) Free2D((void**)arr_strs);
	    }
	    else
	    {
	      fprintf(stderr, "Cannot get integer array for granulepointerarr, exit.\n");
	      exit(0);
	    }
	  } /* else */

	  strcpy(meta_name, "INPUTPOINTER");
          get_sel_metadata(cmeta_str, meta_name, meta_val, &m_cnt, 0);
          if (m_cnt == 0)
          {
            fprintf(stderr, "Metadata INPUTPOINTER missing in %s\n", fname);
	    sprintf(arg_list[4], "%d", 0);
          }
          else
	  {
            if (strcmp(num_str[0], "all") == 0) run_cnt = 0;
	    /* MSZ 3/9/2018, to solve possible missing lines in meta data */
	    /*get_gran_id(arg_list, max_nobs, meta_val[0], run_cnt, num_str);*/
	    gran_narr_str[0] = '\0';
	    strcpy(gran_narr_str, meta_val[0]);
	    strcat(gran_narr_str, meta_val[1]);
	    strcat(gran_narr_str, meta_val[2]);
	    strcat(gran_narr_str, meta_val[3]);
	    strcat(gran_narr_str, meta_val[4]);
            get_gran_id(arg_list, max_nobs, gran_narr_str, run_cnt, num_str, granulepointerarr, num_in_gran);

	    if(granulepointerarr) free(granulepointerarr);
	  }
	  if(cmeta_str) free(cmeta_str);
        }
      }
      if (atoi(arg_list[4]) == 0) status = 0;
    }

    if(ameta_str) free(ameta_str);
    if(meta_val) Free2D((void **)meta_val);
  } /* else */ /* for non "-layer" option or "-g*=" options */

  if (status != 0)
  {
    sds_cnt = atoi(arg_list[3]);
    run_cnt = atoi(arg_list[4]);
    tot_num_sds = sds_cnt * run_cnt;
    fprintf(stderr, "\nNumber of dataset output: %d\n", tot_num_sds);
    if (tot_num_sds > 30)
    {
      fprintf(stderr, "If you want to continue press 'y': ");
      if (getchar() != 'y') status = 0;
    }
  }
  return status;
}

int check_param_npp(char *fname, char **sds_names, int sds_cnt, char *r_mode, 
	char **num_str, int run_cnt, char **arg_list, int *ptr_fname_existence, char *ptr_fname)
{
  /* MSZ 3/5/2018, SSAI 
   * treat -layer and -g* option differently so that for -layer option, there is
   * no need to get the input point file and read meta data from the point file but
   * get max_obs by using maximum of SDS named "num_observations". */

  int len, i, k, ip;
  int m_cnt, status, l2g_form;
  int max_nobs, tot_num_sds;
  char orb_narr_str[MAX_STR_LEN*4];
  char pnt_narr_str[MAX_STR_LEN*4];
  char gran_narr_str[MAX_STR_LEN*8];
  char meta_name[MAX_META_NAME_LEN];
  char *meta_val;
  char **iptr_fnames;
  int iptr_array[1240];
  int iptr_fcnt;
  int iptr_cnt;  
  char pnt_file_name[MAX_PATH_LENGTH] = {'\0'};
  int *granulepointerarr = NULL;       /* MSZ 3/12/2018 */
  int num_in_gran = 0;                 /* MSZ 3/12/2018 */


  /* MSZ 3/3/2018 */
  sds_t in_sds_info;
  int32 attr_type, attr_cnt;
  char *l2g_storage_format = NULL;

  /* get arg_list[2] */
  /* get the l2g storage format */
  in_sds_info.sd_id = in_sds_info.sds_id = -1;
  if(open_sd(fname, &in_sds_info, 'R') == FAIL)
    fprintf(stderr, "Cannot open output HDF file %s\n", fname);
  else
  {
    if( !(l2g_storage_format = (char*)find_get_sd_attr(&in_sds_info, "storage_format", &attr_type, &attr_cnt)) )
      l2g_storage_format = (char*)find_get_sd_attr(&in_sds_info, "storageformat", &attr_type, &attr_cnt);

    /* Note that even if meta data for l2g_storage_format is unavailable, it is Ok because arg_list[2] will not be "full"
     * and we assume there are uncompact layers if meta data for storageformat is unavailable */
    if(l2g_storage_format)
    {
      strcpy(arg_list[2], l2g_storage_format);
      if(l2g_storage_format) free(l2g_storage_format);
    }
    else
    {
      if(sds_name_existence_soft(fname, sds_name_nobs) == 1)
	strcpy(arg_list[2], "compact");
      else
	fprintf(stderr, "No information on l2g_storageformat, and assuming compact by default.\n");
    }
  }
  close_sd(&in_sds_info, in_sds_info.H5IN);

  /* get arg_list[3] and arg_list[6+k] */
  if (strcmp(sds_names[0], "all") == 0) sds_cnt = 0; 
  read_input_sds_names(fname, sds_names, &sds_cnt);
  for (i=0, k=0; i<sds_cnt; i++)
    if (sds_names[i][0] != '\0')
    {
      strcpy(arg_list[6+k], sds_names[i]);
      k++;
    }
  if (k == 0) status = 0;
  sprintf(arg_list[3], "%d", k);

  status = 0;
  if (strcasestr(r_mode, "-layer") != '\0')
  {
     m_cnt = 1;

     max_nobs = get_l2g_max_obs_layers(fname, sds_name_nobs);
     fprintf(stderr, "\nMaximumObservations=%d\n\n",max_nobs);
     if(max_nobs == -1) 
     {
	m_cnt = 0;
	fprintf(stderr, "max_nobs is -1, exit...\n");
	exit(0);
	sprintf(arg_list[4], "%d", 0);
     }

     if (strcmp(num_str[0], "all") == 0) run_cnt = 0;
     get_layer_or_granptr_numbers(arg_list, max_nobs, run_cnt, num_str, r_mode, m_cnt);

     status = 1;

     if (atoi(arg_list[4]) == 0)
	 status = 0;
  }
  else   /* for non "-layer" option or "-g*=" options */
  { 
    status = 0;
    if ((meta_val = (char *)calloc(MAX_META_VAL_LEN, sizeof(char))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for metaval in parse_cmd_read_l2g");
      return 0; 
    }  
    if ((iptr_fnames = (char **)Calloc2D(MAX_NUM_GRAN, MAX_META_VAL_LEN, sizeof(char))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for iptr_fnames in parse_cmd_read_l2g");
      return 0; 
    }
    if ((iptr_fcnt = get_npp_input_files(fname, "all", iptr_fnames)) == 0)
    {
      fprintf(stderr, "No matching input files found in the InputPointer, exit...\n"); 
      exit(0);
    }
    else
    {

      if(*ptr_fname_existence == 1)
        strcpy(pnt_file_name, ptr_fname);
      else
	/* get the grid point file name for -gid option, 
         * which is the first of the item in file attr InputPointer */
        strcpy(pnt_file_name, iptr_fnames[0]);

      /* MSZ 3/12/2018 */
      FILE *sfp = NULL;
	
      if( (sfp = fopen(pnt_file_name, "rb")) == NULL )
      {
	fprintf(stderr, "Point file %s doesn't exists, please copy it to the current fold or specify it using -ptr= option, exit...\n", pnt_file_name);
	exit(0);
      }
      else
      {
	fclose(sfp);
	*ptr_fname_existence = 1;
	strcpy(ptr_fname, pnt_file_name);
      }

      strcpy(meta_name, "NumberofOverlapGranules");
      find_get_sd_attr_to_str(pnt_file_name, meta_name, meta_val);
      if (meta_val != NULL)  
      {
          max_nobs = atoi(meta_val);
	  status = 1;
      }	
      else 
          fprintf(stderr, "Metadata MaximumObservations missing in %s\n", pnt_file_name);
    }  

    if (status == 0) /* MSZ 3/5/2018 */
    {
      if(iptr_fcnt == 0)
	fprintf(stderr, "Input file %s failed because it cannot find the InputPointer file, exit...\n", fname);
      else	
        fprintf(stderr, "Input file %s failed because Metadata MaximumObservations missing in InputPointer file %s, exit...\n", fname, pnt_file_name);

      exit(0);

      return 0;
    }

    if (status != 0)
    {
      if (strcmp(r_mode, "-gpidx=") == 0)
      {
        m_cnt = 1;
        strcpy(meta_name, "NumberofOverlapGranules");
	find_get_sd_attr_to_str(pnt_file_name, meta_name, meta_val);
        if (meta_val != NULL) 
        { 
          max_nobs = atoi(meta_val);	
	  /*fprintf(stderr, "NumberofOverlapGranules=%s\n",meta_val);*/
        }
        else
        {
          m_cnt = 0;
	  fprintf(stderr, "Metadata NumberofOverlapGranules missing in %s\n", pnt_file_name);
        }
        if (strcmp(num_str[0], "all") == 0) run_cnt = 0;
        get_layer_or_granptr_numbers(arg_list, max_nobs, run_cnt, num_str, r_mode, m_cnt);
      }
      else if (strcmp(r_mode, "-glist=") == 0)
      {
        max_nobs = -1;
        orb_narr_str[0] = gran_narr_str[0] = '\0';

        m_cnt = 1;
        strcpy(meta_name, "NumberofOverlapGranules");
	find_get_sd_attr_to_str(pnt_file_name, meta_name, meta_val);
        if (meta_val != NULL) 
        { 
          max_nobs = atoi(meta_val);	
	  /*fprintf(stderr, "NumberofOverlapGranules=%s\n",meta_val);*/
        }
        else
        {
          m_cnt = 0;
	  fprintf(stderr, "Metadata NumberofOverlapGranules missing in %s\n", pnt_file_name);       
        }

        m_cnt = 1;
        strcpy(meta_name, "OrbitNumberArray");
	find_get_sd_attr_to_str(pnt_file_name, meta_name, meta_val);
        if (meta_val != NULL) 
          strcpy(orb_narr_str, meta_val);
        else
        {
          m_cnt = 0;
	  fprintf(stderr, "Missing metadata OrbitNumberArray in %s\n", pnt_file_name);	 
        } 

        m_cnt = 1;
        strcpy(meta_name, "InputPointer");              
	find_get_sd_attr_to_str(pnt_file_name, meta_name, meta_val);
        if (meta_val != NULL) 
          strcpy(gran_narr_str, meta_val);
        else
        {
          m_cnt = 0;
	  fprintf(stderr, "Missing metadata InputPointer in %s\n", pnt_file_name);		 
        }

        m_cnt = 1;
        strcpy(meta_name, "GranulePointerArray");
	find_get_sd_attr_to_str(pnt_file_name, meta_name, meta_val);
        if (meta_val != NULL) 
          strcpy(pnt_narr_str, meta_val);
        else
        {
          m_cnt = 0;
	  fprintf(stderr, "Missing metadata GranulePointerArray in %s\n", pnt_file_name);	              
        }
        if (strcmp(num_str[0], "all") == 0) run_cnt = 0;  
        get_npp_granorb_numbers(arg_list, max_nobs, orb_narr_str, gran_narr_str, pnt_narr_str, run_cnt, num_str);  
      }
      else if (strcmp(r_mode, "-gid=") == 0)
      {
        num_in_gran = 0;
        m_cnt = 1;
        strcpy(meta_name, "NumberofInputGranules");
        find_get_sd_attr_to_str(pnt_file_name, meta_name, meta_val);
        if (meta_val != NULL)
        {
          num_in_gran = atoi(meta_val);
          fprintf(stderr, "NumberofInputGranules=%s\n",meta_val);
        }
        else
        {
          m_cnt = 0;
          fprintf(stderr, "Metadata NumberofInputGranules missing in %s\n", pnt_file_name);
        }

        max_nobs = 0;
        m_cnt = 1;
        strcpy(meta_name, "NumberofOverlapGranules");
        find_get_sd_attr_to_str(pnt_file_name, meta_name, meta_val);
        if (meta_val != NULL)
        {
          max_nobs = atoi(meta_val);
          fprintf(stderr, "NumberofOverlapGranules=%s\n",meta_val);
        }
        else
        {
          m_cnt = 0;
          fprintf(stderr, "Metadata NumberofOverlapGranules missing in %s\n", pnt_file_name);
        }

	/* MSZ 3/12/2018, added GranulePointerArray info */
        m_cnt = 1;
        strcpy(meta_name, "GranulePointerArray");
	find_get_sd_attr_to_str(pnt_file_name, meta_name, meta_val);
        if (meta_val != NULL) 
	{
            strcpy(pnt_narr_str, meta_val);
	    char **arr_strs = NULL;
	    int num_eles = 0;
	    if ((arr_strs = (char **)Calloc2D(MAX_VNP_GRANArray*2, 5, sizeof(char))) == NULL)
	    {
	      fprintf(stderr, "cannot calloc arr_strs, exit\n");
	      exit(0);
	    }
	    else
	      split_string(pnt_narr_str, arr_strs, &num_eles);
	
	    if(num_eles > 0)
	    {
	      if((granulepointerarr = calloc(num_in_gran, sizeof(int))) == NULL)
	      {
	        fprintf(stderr, "cannot calloc granulepointerarr, exit\n");
	        exit(0);
	      }
	      else   /* convert string array to integer array */
	      {
		printf("granule point array: %d\n", num_in_gran);
		for(i = 0; i < num_in_gran; i++)
		{
		  granulepointerarr[i] = atoi(arr_strs[i]);
		  printf(" %d ", granulepointerarr[i]);
		}
		printf("\n");
	      }
	      if(arr_strs) Free2D((void**)arr_strs);
	    }
	    else
	    {
	      fprintf(stderr, "Cannot get integer array for granulepointerarr, exit.\n");
	      exit(0);
	    }
	}
        else
        {
          m_cnt = 0;
	  fprintf(stderr, "Missing metadata GranulePointerArray in %s\n", pnt_file_name);	              
        }

        if (iptr_fcnt == 0)
        {
          fprintf(stderr, "No matching input files found in the InputPointer\n");
          sprintf(arg_list[4], "%d", 0);
        }
        else
        {
          if (strcmp(num_str[0], "all") == 0) run_cnt = 0;
	  get_npp_gran_id(arg_list, max_nobs, fname, run_cnt, num_str, granulepointerarr, num_in_gran, pnt_file_name);

	  if(granulepointerarr) free(granulepointerarr);
        }
      }
      if (atoi(arg_list[4]) == 0) status = 0;
    }

    free(meta_val);
    Free2D((void **)iptr_fnames);
  } /* else */ /* for non "-layer" option or "-g*=" options */

  if (status != 0)
  {
    sds_cnt = atoi(arg_list[3]);
    run_cnt = atoi(arg_list[4]);
    tot_num_sds = sds_cnt * run_cnt;
    fprintf(stderr, "\nNumber of dataset output: %d\n", tot_num_sds);
    if (tot_num_sds > 30)
    {
      fprintf(stderr, "If you want to continue press 'y': ");
      if (getchar() != 'y') status = 0;
    }
  }
  return status;
}

void read_input_sds_names(char *fname, char **sds_names, int *sds_cnt)
{
  int i, cnt;
  int32 sd_id;
  char tmp_sds_name[MAX_SDS_NAME_LEN];

  if (*sds_cnt == 0)
  {
    *sds_cnt = get_l2g_sds_names(fname, sds_names);
    if (*sds_cnt == 0)
      fprintf(stderr, "No SDS found in hdf file %s\n", fname);
  }
  else  /* user specifies without _1 or _c in the names */
  {
    /* MSZ 2/21/2018, modify to use sds_name_existence() */
    for (i=0, cnt = 0; i < *sds_cnt; i++)
    {
      sprintf(tmp_sds_name, "%s_1", sds_names[i]);
      if(sds_name_existence(fname, tmp_sds_name) == 0)
      {
        fprintf(stderr, "Ignoring invalid SDS name %s\n", sds_names[i]);
	sds_names[i][0] = '\0';
      }
      else
        cnt++;
    }

    if (cnt == 0)
    {
      *sds_cnt = 0;
      fprintf(stderr, "None of the input SDS names are valid \n");
      fprintf(stderr, "Valid SDS names are: \n");
      cnt = get_l2g_sds_names(fname, sds_names);
      if (cnt == 0)
        fprintf(stderr, "No SDS found in %s\n", fname);
      else
        for (i=0; i<cnt; i++)
          fprintf(stderr, "          %s\n", sds_names[i]);
    }
  }
}

void get_layer_or_granptr_numbers(char **arg_list, int max_nobs, int run_cnt, 
	char **num_str, char *r_mode, int m_cnt)
{
  int sds_cnt;
  int i, k, p1, ln;
  int ln1, ln2, len;
  char ln_str[10];

  sds_cnt = atoi(arg_list[3]);
  if (run_cnt == 0)
  {
    if (m_cnt != 0)
    {
      for (k=0; k<max_nobs; k++)
        sprintf(arg_list[6+sds_cnt+k], "%d", k);
      run_cnt = max_nobs;
    }
  }
  else
  {
    if (m_cnt == 0)
    {
      for (i=0; i < run_cnt; i++)
        strcpy(arg_list[6+sds_cnt+i], num_str[i]);
    }
    else
    {
      for (i=0, k=0; i<run_cnt; i++)
      {
	if ((p1 = charpos(num_str[i], '-', 0)) == -1)
	{
          ln = atoi(num_str[i]);
          if((ln < max_nobs) && (ln >= 0))
          {
            strcpy(arg_list[6+sds_cnt+k], num_str[i]);
            k++;
          }
          else
          {
            if (strcmp(r_mode, "-layer=") == 0)
              fprintf(stderr, "Ignoring invalid layer %d (0-based max_layers = %d)\n", ln, max_nobs-1);
            else
              fprintf(stderr, "Ignoring invalid granule index %d (0-based max_index = %d)\n", ln, max_nobs-1);
          }
	}
	else
	{
	  len = (int)strlen(num_str[i]);
	  strmid(num_str[i], 0, p1, ln_str);
	  ln1 = (int)atoi(ln_str);
	  strmid(num_str[i], p1+1, len-p1-1, ln_str);
	  ln2 = (int)atoi(ln_str);
	  for (ln=ln1; ln<=ln2; ln++)
	  {
	    if((ln < max_nobs) && (ln >= 0))
	    {
	      sprintf(arg_list[6+sds_cnt+k], "%d", ln);
	      k++;
	    }
            else
            {
              if (strcmp(r_mode, "-layer=") == 0)
                fprintf(stderr, "Ignoring invalid layer %d (0-based max_layers = %d)\n", ln, max_nobs-1);
              else
                fprintf(stderr, "Ignoring invalid granule index %d (0-based max_index = %d)\n", ln, max_nobs-1);
            }
	  }
	}
      }
      run_cnt = k;
    }
  }
  sprintf(arg_list[4], "%d", run_cnt);
}

void get_granorb_numbers(char **arg_list, int max_nobs, char *orb_narr_str, 
	char *gran_narr_str, char *pnt_narr_str, int run_cnt, char **num_str)
{
  int gn[2], on, pn[2];
  int sds_cnt;
  int i, k, len;
  int p1_g, p1_o, p1_p;
  int p2_g, p2_o, p2_p;
  char gn_str[20];
  char on_str[20];
  char pn_str[20];

  sds_cnt = atoi(arg_list[3]);
  if (run_cnt == 0)
  {
    k = 0;
    p1_g = p1_o = p1_p = 1;
    while (k<max_nobs)
    {
      p2_o = charpos(orb_narr_str, ',', p1_o);
      p2_g = charpos(gran_narr_str, ',', p1_g);
      p2_p = charpos(pnt_narr_str, ',', p1_p);
      if ((p2_g != -1) && (p2_o != -1) && (p2_p != -1))
      {
        strmid(gran_narr_str, p1_g, p2_g-p1_g, gn_str);
        strmid(orb_narr_str, p1_o, p2_o-p1_o, on_str);
        strmid(pnt_narr_str, p1_p, p2_p-p1_p, pn_str);
	if (atoi(pn_str) != -1)
        {
          sprintf(arg_list[6+sds_cnt+k], "%s %s %s", pn_str, gn_str, on_str);
          k++;
	}
        p1_g = p2_g + 1;
        p1_o = p2_o + 1;
        p1_p = p2_p + 1;
      }
      else break;
    }
  }
  else for (i=0, k=0; i<run_cnt; i++)
  {
    len = sscanf(num_str[i], "%d %d", &gn[0], &on);
//printf("num_str[%d]=%s\n",i,    num_str[i]);
//printf("gn[0]=%d   on=%d\n",gn[0], on);
//printf("len=%d\n",len);
    if (len == 0) 
      fprintf(stderr, "Ignoring invalid gran_orb number %s\n", num_str[i]);
    else
    {
      gn[1] = pn[0] = pn[1] = -1;
      if (check_gran_orb_num(gn, on, pn, gran_narr_str, orb_narr_str, pnt_narr_str, max_nobs) == -1)
        fprintf(stderr, "Ignoring invalid gran_orb number %s\n", num_str[i]);
      else
      {
        sprintf(arg_list[6+sds_cnt+k], "%d %d %d", pn[0], gn[0], on);
        k++;
	if ((pn[1] != -1) && (gn[1] != -1))
	{
          sprintf(arg_list[6+sds_cnt+k], "%d %d %d", pn[1], gn[1], on);
          k++;
	}
      }
    }
  }
  sprintf(arg_list[4], "%d", k);
  if (k == 0)
  {
    fprintf(stderr, "Valid granule numbers and corresponding orbit numbers are: \n");
    fprintf(stderr, "Granules: %s\n", gran_narr_str);
    fprintf(stderr, "Orbits: %s\n", orb_narr_str);
  }
}

int check_gran_orb_num(int *gn, int on, int *pn, char *gran_narr_str, 
	char *orb_narr_str, char *pnt_narr_str, int max_nobs)
{
  int i, j;
  int status;
  int p1_g, p1_o, p1_p;
  int p2_g, p2_o, p2_p;
  char gn_v[20], on_v[20], pn_v[20];

  i = j = 0;
  status = -1;
  p1_g = p1_o = p1_p = 1;
  while (i<max_nobs)
  {
    p2_g = charpos(gran_narr_str, ',', p1_g);
    p2_o = charpos(orb_narr_str, ',', p1_o);
    p2_p = charpos(pnt_narr_str, ',', p1_p);
    if ((p2_g != -1) && (p2_o != -1) && (p2_p != -1))
    {
      strmid(gran_narr_str, p1_g, p2_g-p1_g, gn_v);
      strmid(orb_narr_str, p1_o, p2_o-p1_o, on_v);
      strmid(pnt_narr_str, p1_p, p2_p-p1_p, pn_v);
      strtrim(gn_v);
      strtrim(on_v);
      strtrim(pn_v);
      if (atoi(pn_v) != -1)
      {
	i++;
        if ((atoi(gn_v) == gn[0]) && (atoi(on_v) == on))
        {
	  pn[0] = atoi(pn_v);
	  status = 1;
	  break;
        }
        else if (gn[j] == -1)
	{
	  if ((on >= 0) && (atoi(on_v) == on))
          {
	    gn[j] = atoi(gn_v);
	    pn[j] = atoi(pn_v);
	    status = 1;
	    j++;
	    if (j == 2) break;
          }
	}
      }
      p1_g = p2_g + 1;
      p1_o = p2_o + 1;
      p1_p = p2_p + 1;
    }
    else break;
  }
  return status;
}

void get_gran_id(char **arg_list, int max_nobs, char *in_ptr, int run_cnt, char **num_str, int *granulepointerarr, int num_in_gran)
{
  int p1, p2;
  int i, j, k;
  int sds_cnt;
  int gran_cnt;
  char **gran_id;
  char gid[MAX_NAME_LENGTH];

  if ((gran_id = (char **)Calloc2D(MAX_NUM_GRAN, MAX_NAME_LENGTH, sizeof(char))) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for gran_id in get_gran_id\n");
    sprintf(arg_list[4], "%d", 0);
  } 
  else
  {
    gran_cnt = 0;
    p1 = charpos(in_ptr, '"', 0);
    while (p1 != -1)
    {
      p2 = charpos(in_ptr, '"', p1+1); 
      if (p2 != -1)
      {
        strmid(in_ptr, p1+1, p2-p1-1, gran_id[gran_cnt]);
        strtrim(gran_id[gran_cnt]);
        gran_cnt++;
        p1 = charpos(in_ptr, '"', p2+1); 
      }
      else p1 = -1;
    }

    /* MSZ 3/12/2018 */
    /* assume the first is not granule but point file if 
     * gran_cnt == num_in_gran + 1
     * and remove the first from the input gran_id */
    if(gran_cnt == num_in_gran + 1)
    {
      fprintf(stderr, "\nInput Pointer File: %s\n", gran_id[0]);
      for(i = 0; i < gran_cnt - 1; i++)
      {
	strcpy(gran_id[i], gran_id[i+1]);
	fprintf(stderr, "Input Granuel File: %2d %s\n", granulepointerarr[i], gran_id[i]);
      }
      fprintf(stderr, "\n");

      gran_id[gran_cnt - 1][0] = '\0';
    }
    else if(gran_cnt == num_in_gran)
	;
    else
    {
      fprintf(stderr, "Make sure that you specified a correct pointer file, because there should be no such case, exit...\n");
      exit(0);
    }

    /* MSZ 3/12/2018 */
    /* just keep these overlaped granules based on granulepointerarr */
    j = 0;
    for(i = 0; i < num_in_gran; i++)
      if(granulepointerarr[i] != -1)
	strcpy(gran_id[j++], gran_id[i]);

    /* MSZ 3/12/2018, now gran_id is just the names list of the overlapgranules */
    for(i = max_nobs; i < gran_cnt; i++)
      gran_id[i][0] = '\0';

    gran_cnt = max_nobs;
    sds_cnt = atoi(arg_list[3]);
    if (run_cnt == 0)
    {
      for (k=0; k<max_nobs; k++)
      {
        p1 = charpos(gran_id[k], '.', 0); 
        p2 = charpos(gran_id[k], '.', p1+1); 
        p2 = charpos(gran_id[k], '.', p2+1); 
        strmid(gran_id[k], p1+1, p2-p1-1, gid);
        sprintf(arg_list[6+sds_cnt+k], "%d %s", k, gid);
      }
      sprintf(arg_list[4], "%d", max_nobs);
    }
    else
    {
      for (i=0, k=0; i<run_cnt; i++)
      {
        for (j=0; j<gran_cnt; j++)
          if (strpos(gran_id[j], num_str[i], 0) != -1)
          {
	    sprintf(arg_list[6+sds_cnt+k], "%d %s", j, num_str[i]);
	    k++;
	    break;
	  }
	if (j == gran_cnt)
	  fprintf(stderr, "Ignoring invalid granule %s (may not exist or not overlap)\n", num_str[i]);
      }
      sprintf(arg_list[4], "%d", k);
    }
    if(gran_id) Free2D((void **)gran_id);
  }
}

void get_npp_gran_id(char **arg_list, int max_nobs, char *in_ptr, int run_cnt, char **num_str, int *granulepointerarr, int num_in_gran, char *pnt_file_name)
{
  int p1, p2;
  int i, j, k;
  int sds_cnt;
  int gran_cnt;
  char **gran_id;
  char gid[MAX_NAME_LENGTH];
  char **iptr_fnames;
  int iptr_fcnt, iptr_cnt, status;  
  int iptr_array[MAX_VNP_GRANArray];   /* for VNP */

  if ((gran_id = (char **)Calloc2D(MAX_NUM_GRAN, MAX_NAME_LENGTH*2, sizeof(char))) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for gran_id in get_gran_id\n");
    sprintf(arg_list[4], "%d", 0);
  } 
  else
  {
    status = 0;
    if ((iptr_fnames = (char **)Calloc2D(MAX_NUM_GRAN, MAX_META_VAL_LEN, sizeof(char))) == NULL) /* MSZ 8/2/2018, increase from 20 to 127 */
    {
      fprintf(stderr, "Cannot allocate memory for iptr_fnames in parse_cmd_read_l2g");
      return; 
    }
    if ((iptr_fcnt = get_npp_input_files(in_ptr, "all", iptr_fnames)) == 0)
      fprintf(stderr, "No input files found in the InputPointer\n");

    /*if ((iptr_cnt = get_npp_GranulePointerArray(in_ptr, iptr_array)) == 0)  */
    /*if(iptr_fnames[0] != '\0') fprintf(stderr, "trying to read GranulePointerArray from %s...\n", iptr_fnames[0]);*/

    /* Note that the returned iptr_cnt is equal to the length of MAX_VNP_GRANArray */
    /*if ((iptr_cnt = get_npp_GranulePointerArray(iptr_fnames[0], iptr_array)) == 0)*/
    if ((iptr_cnt = get_npp_GranulePointerArray(pnt_file_name, iptr_array)) == 0)
      fprintf(stderr, "No GranulePointerArray found in the InputPointer\n"); 
    else
      /*fprintf(stderr, "GranulePointerArray is successfully read from %s\n", iptr_fnames[0])*/;

    if (iptr_fcnt == 0 || iptr_cnt == 0) 
    {
      fprintf(stderr, "Input file %s failed\n", in_ptr);    
      return;
    }
    else
    {
      j = 0;
      for (i=0; i<iptr_fcnt; i++)
      {
	rm_path(iptr_fnames[i]);
	strtrim(iptr_fnames[i]);	
	if(i==0)  fprintf(stderr, "\nInput Pointer File: %s\n", iptr_fnames[i]);
        strcpy(gran_id[j], iptr_fnames[i]); 
	j++;
      }
      status = 1; 
    }    
    if (status == 0) 
    {
      fprintf(stderr, "Input file %s failed\n", in_ptr);
      return;
    }   

    /* MSZ 3/12/2018 */
    /* assume the first is not granule but point file if 
     * iptr_fcnt == num_in_gran + 1
     * and remove the first from the input gran_id */
    if(iptr_fcnt == num_in_gran + 1)
    {
      for(i = 0; i < iptr_fcnt - 1; i++)
      {
	strcpy(gran_id[i], gran_id[i+1]);
	fprintf(stderr, "Input Granuel File: %2d %s\n", granulepointerarr[i], gran_id[i]);
      }
      fprintf(stderr, "\n");

      gran_id[iptr_fcnt - 1][0] = '\0';
    }

    /* MSZ 3/12/2018 */
    /* just keep these overlaped granules based on granulepointerarr */
    j = 0;
    for(i = 0; i < num_in_gran; i++)
      if(granulepointerarr[i] != -1)
	strcpy(gran_id[j++], gran_id[i]);

    /* MSZ 3/12/2018, now gran_id is just the names list of the overlapgranules */
    for(i = max_nobs; i < iptr_fcnt; i++)
      gran_id[i][0] = '\0';


    gran_cnt = max_nobs;
    sds_cnt = atoi(arg_list[3]);
    if (run_cnt == 0)
    {
      for (k=0; k<max_nobs; k++)
      {
        p1 = charpos(gran_id[k], '.', 0); 
        p2 = charpos(gran_id[k], '.', p1+1); 
        p2 = charpos(gran_id[k], '.', p2+1); 
        strmid(gran_id[k], p1+1, p2-p1-1, gid);
        sprintf(arg_list[6+sds_cnt+k], "%d %s", k, gid);
      }
      sprintf(arg_list[4], "%d", max_nobs);
    }
    else
    {
      for (i=0, k=0; i<run_cnt; i++)
      {
        for (j=0; j<gran_cnt; j++)
          if (strpos(gran_id[j], num_str[i], 0) != -1)
          {
	    sprintf(arg_list[6+sds_cnt+k], "%d %s", j, num_str[i]);
	    k++;
	    break;
	  }
	if (j == gran_cnt)
	  fprintf(stderr, "Ignoring invalid granule %s (may not exist or not overlap)\n", num_str[i]);
      }
      sprintf(arg_list[4], "%d", k);
    }
    if(gran_id) Free2D((void **)gran_id);
    if(iptr_fnames) Free2D((void **)iptr_fnames);   
  }
}

int get_npp_GranulePointerArray(char *fname, int *iptr_array)
/***********************************************************************************
	Returns number of input files used in the production of fname that matches
	the string iptr_id. iptr_fname contains the resulting filenames. if iptr_id 
	== 'all', all filenames are read from metadata INPUTPOINTER. else matching 
	filenames (match to iptr_id) are returned. else filenames from INPUTPOINTER
	with matching SHORTNAME (match to iptr_id) are returned.
***********************************************************************************/ 
{
  int i;
  int pos1, pos2;
  int m_cnt;
  int iptr_cnt;
  char **modis_fpath;
  char *meta_str = NULL;
  char tmp[MAX_STR_LEN], meta_name[MAX_META_NAME_LEN];
  int attr_type, attr_cnt;
  int meta_str_len = 600;

  if ((modis_fpath = (char **)Calloc2D(MAX_NUM_PATH, MAX_PATH_LENGTH, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for modis_fpath in get_npp_GranulePointerArray \n");

  /* MSZ 3/9/2018 */
  if ((meta_str = (char *)calloc(meta_str_len, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for meta_str in get_npp_GranulePointerArray \n");

  if (modis_fpath == NULL)
  {
    Free2D((void **)modis_fpath);
    return -1;
  }

  iptr_cnt = 0;

  m_cnt = 0;
  strcpy(meta_name, "GranulePointerArray");
 
  /* MSZ 3/9/2018 */ 
  find_get_sd_attr_to_str(fname, meta_name, meta_str);
  if (meta_str[0] == '\0')
  {
    fprintf(stderr, "Metadata GranulePointerArray not found in %s\n", fname);
    exit(1);
  }
  else
  {
    get_qa_tool_env(MODIS_ENV, modis_fpath);

    /*fprintf(stderr, "\nReading all the granule index in the InputPointer\n");*/
    pos1 = 0;
    pos2 = charpos(meta_str, ',', 0);
    while (pos2 != -1) 
    {
      strmid(meta_str, pos1, pos2 - pos1, tmp);
      strtrim(tmp);  

      iptr_array[iptr_cnt++] = atoi(tmp); 	
      pos1 = pos2 + 1;
      pos2 = charpos(meta_str, ',', pos1);
    }
    strmid(meta_str, pos1, meta_str_len - pos1 -1, tmp);
    strtrim(tmp);         
    iptr_array[iptr_cnt++] = atoi(tmp);                     
    if (iptr_cnt == 0)
      fprintf(stderr, "No granule pointer array index found in the InputPointer \n");        		

    if(meta_str) free(meta_str);
  }    
  Free2D((void **)modis_fpath);
  return iptr_cnt;
}

void get_npp_granorb_numbers(char **arg_list, int max_nobs, char *orb_narr_str, 
	char *gran_narr_str, char *pnt_narr_str, int run_cnt, char **num_str)
{
  int gn[2], on, pn[2];
  int sds_cnt;
  int i, k, len;
  int p1_g, p1_o, p1_p;
  int p2_g, p2_o, p2_p;
  char gn_str[20];
  char on_str[20];
  char pn_str[20];
  int p1, p2, p3;
  char gid[255];


  fprintf(stderr, "max_nobs===%d\n",max_nobs);
  sds_cnt = atoi(arg_list[3]);
  if (run_cnt == 0)
  {
    k = 0;
    p1_o = p1_p = 0;
    p1_g = 1;    
    while (k<max_nobs)
    {
      p2_o = charpos(orb_narr_str, ',', p1_o);
      p2_g = charpos(gran_narr_str, ':', p1_g);
      p2_p = charpos(pnt_narr_str, ',', p1_p);
      if ((p2_g != -1) && (p2_o != -1) && (p2_p != -1))
      {
        strmid(gran_narr_str, p1_g, p2_g-p1_g, gid);
        strmid(orb_narr_str, p1_o, p2_o-p1_o, on_str);
        strmid(pnt_narr_str, p1_p, p2_p-p1_p, pn_str);
	if (atoi(pn_str) != -1)
        {
          p1 = charpos(gid, '.', 0); 
          p2 = charpos(gid, '.', p1+1); 
          p3 = charpos(gid, '.', p2+1); 
          strmid(gid, p2+1, p3-p2-1, gn_str); 
	  strtrim(gn_str);       
//printf("pn_str=:%s\n",pn_str);	
//printf("gn_str=:%s\n",gn_str);
//printf("on_str=:%s\n",on_str);  
	  sprintf(arg_list[6+sds_cnt+k], "%s %s %s", pn_str, gn_str, on_str);
          k++;
	}
        p1_g = p2_g + 1;
        p1_o = p2_o + 1;
        p1_p = p2_p + 1;
      }
      else if ((p2_g == -1) && (p2_o != -1) && (p2_p != -1))
      {
	len = strlen(gran_narr_str);
	strmid(gran_narr_str, p1_g, len-p1_g-2, gid);	
        strmid(orb_narr_str, p1_o, p2_o-p1_o, on_str);
        strmid(pnt_narr_str, p1_p, p2_p-p1_p, pn_str);
	if (atoi(pn_str) != -1)
        {
          p1 = charpos(gid, '.', 0); 
          p2 = charpos(gid, '.', p1+1); 
          p3 = charpos(gid, '.', p2+1); 
          strmid(gid, p2+1, p3-p2-1, gn_str); 
	  strtrim(gid);       
//printf("pn_str=:%s\n",pn_str);	
//printf("gn_str=:%s\n",gn_str);
//printf("on_str=:%s\n",on_str);  
	  sprintf(arg_list[6+sds_cnt+k], "%s %s %s", pn_str, gn_str, on_str);
          k++;
	}
        p1_g = p2_g + 1;
        p1_o = p2_o + 1;
        p1_p = p2_p + 1;      
      }
      else break;
    }
//printf("k=:%d\n",k);	    
  }
  else for (i=0, k=0; i<run_cnt; i++)
  {
    len = sscanf(num_str[i], "%d %d", &gn[0], &on);
//printf("num_str[%d]=%s\n",i,    num_str[i]);
//printf("gn[0]=%d   on=%d\n",gn[0], on);
//printf("len=%d\n",len);
    if (len == 0) 
      fprintf(stderr, "Ignoring invalid gran_orb number %s\n", num_str[i]);
    else
    {
      gn[1] = pn[0] = pn[1] = -1;
      if (check_npp_gran_orb_num(gn, on, pn, gran_narr_str, orb_narr_str, pnt_narr_str, max_nobs) == -1)
        fprintf(stderr, "Ignoring invalid gran_orb number %s\n", num_str[i]);
      else
      {
        sprintf(arg_list[6+sds_cnt+k], "%d %d %d", pn[0], gn[0], on);
        k++;
	if ((pn[1] != -1) && (gn[1] != -1))
	{
          sprintf(arg_list[6+sds_cnt+k], "%d %d %d", pn[1], gn[1], on);
          k++;
	}
      }
    }
  }
  sprintf(arg_list[4], "%d", k);
  if (k == 0)
  {
    fprintf(stderr, "Valid granule numbers and corresponding orbit numbers are: \n");
    fprintf(stderr, "Granules: %s\n", gran_narr_str);
    fprintf(stderr, "Orbits: %s\n", orb_narr_str);
  }
}

int check_npp_gran_orb_num(int *gn, int on, int *pn, char *gran_narr_str, 
	char *orb_narr_str, char *pnt_narr_str, int max_nobs)
{
  int i, j;
  int status;
  int p1_g, p1_o, p1_p;
  int p2_g, p2_o, p2_p;
  char gn_v[20], on_v[20], pn_v[20];
  char gn_str[255];
  int p1, p2, p3;
  int len;
  
  i = j = 0;
  status = -1;
  p1_o = p1_p = 0;
  p1_g = 1;      
  while (i<max_nobs)
  {
    p2_g = charpos(gran_narr_str, ':', p1_g);
    p2_o = charpos(orb_narr_str, ',', p1_o);
    p2_p = charpos(pnt_narr_str, ',', p1_p);
    if ((p2_g != -1) && (p2_o != -1) && (p2_p != -1))
    {
//      strmid(gran_narr_str, p1_g, p2_g-p1_g, gn_v);
      

      strmid(gran_narr_str, p1_g, p2_g-p1_g, gn_str);
      strmid(orb_narr_str, p1_o, p2_o-p1_o, on_v);
      strmid(pnt_narr_str, p1_p, p2_p-p1_p, pn_v);
      strtrim(on_v);
      strtrim(pn_v);      
      if (atoi(pn_v) != -1)
      {
	p1 = charpos(gn_str, '.', 0); 
	p2 = charpos(gn_str, '.', p1+1); 
	p3 = charpos(gn_str, '.', p2+1); 
	strmid(gn_str, p2+1, p3-p2-1, gn_v); 
	strtrim(gn_v);        
	fprintf(stderr, "gn_v=:%s\n",gn_v);	
 
	i++;
        if ((atoi(gn_v) == gn[0]) && (atoi(on_v) == on))
        {
	  pn[0] = atoi(pn_v);
	  status = 1;
	  break;
        }
        else if (gn[j] == -1)
	{
	  if ((on >= 0) && (atoi(on_v) == on))
          {
	    gn[j] = atoi(gn_v);
	    pn[j] = atoi(pn_v);
	    status = 1;
	    j++;
	    if (j == 2) break;
          }
	}   
      }   
      p1_g = p2_g + 1;
      p1_o = p2_o + 1;
      p1_p = p2_p + 1;
    }
    else if ((p2_g == -1) && (p2_o != -1) && (p2_p != -1))
    {
      len = strlen(gran_narr_str);
      strmid(gran_narr_str, p1_g, len-p1_g-2, gn_str);	
      strmid(orb_narr_str, p1_o, p2_o-p1_o, on_v);
      strmid(pnt_narr_str, p1_p, p2_p-p1_p, pn_v);
      strtrim(on_v);
      strtrim(pn_v);      
      if (atoi(pn_v) != -1)
      {
        p1 = charpos(gn_str, '.', 0); 
        p2 = charpos(gn_str, '.', p1+1); 
        p3 = charpos(gn_str, '.', p2+1); 
        strmid(gn_str, p2+1, p3-p2-1, gn_v); 
	strtrim(gn_v);       
	fprintf(stderr, "gn_v=:%s\n",gn_v); 
	i++;
        if ((atoi(gn_v) == gn[0]) && (atoi(on_v) == on))
        {
	  pn[0] = atoi(pn_v);
	  status = 1;
	  break;
        }
        else if (gn[j] == -1)
	{
	  if ((on >= 0) && (atoi(on_v) == on))
          {
	    gn[j] = atoi(gn_v);
	    pn[j] = atoi(pn_v);
	    status = 1;
	    j++;
	    if (j == 2) break;
          }
	} 
      }
      p1_g = p2_g + 1;
      p1_o = p2_o + 1;
      p1_p = p2_p + 1;          
    }
    else break;
  }
  return status;
}

int get_sds_names_num_observations(char *fname, char **SDS_names_num_obs)
{
   /* Maosheng Zhao 3/6/2018, SSAI, get the sds names with "num_observations" and also put them in a string array 0 to 3,
    * with 0 1km, 1 500m and 3 250m to correspond res 1, 2, and 4 so that res - 1 matches the correct subscript of 
    * sds names with "num_observations"
    * Fail is the return value is 0.
    * SDS_names_num_obs is 4 strings defined outside of the function */

   char **sds_names = NULL;
   int nsds = 0, nsds_num_observations = 0;
   int i;
   sds_t in_sds_info;

   for(i=0; i < 4; i++) /* corresponding to res 1 to 4 minus 1 */
     SDS_names_num_obs[i][0] = '\0';

   if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
   {
     fprintf(stderr, "Cannot allocate memory for sds_names, exit...\n");
     exit(0);
   }
   else
   {
     if ((nsds = get_sds_names(fname, sds_names)) == 0)
     {
       fprintf(stderr, "No sds found in file %s, exit...\n", fname);
       exit(0);
     }
     else
     {
       int RESs_num = 0;
       for(i = 0, nsds_num_observations = 0; i < nsds; i++)
	  if(strcasestr(sds_names[i], "num_observations") != '\0')
	  {
	    strcpy(in_sds_info.name, sds_names[i]);
	    in_sds_info.sd_id = in_sds_info.sds_id = -1;
	    if(get_sds_info(fname, &in_sds_info) != FAIL)
	    {
	      if(in_sds_info.rank == 2 && in_sds_info.dim_size[0] == in_sds_info.dim_size[1])
	      {
		/* for MODIS 1200, 2400 and 4800 system */
		if(in_sds_info.dim_size[0] % 1200 == 0)  /* for MODIS */
		{
		  if(in_sds_info.dim_size[0] == 1200)
		  {
		    strcpy(SDS_names_num_obs[0], sds_names[i]);
		    nsds_num_observations++;
		  }
		  else if(in_sds_info.dim_size[0] == 2400)
		  {
		    strcpy(SDS_names_num_obs[1], sds_names[i]);
		    nsds_num_observations++;
		  }
		  else if(in_sds_info.dim_size[0] == 4800)
		  {
		    strcpy(SDS_names_num_obs[3], sds_names[i]);
		    nsds_num_observations++;
		  }
		}
		else   /* for non-MODIS system */
		{
		  if((int)(in_sds_info.dim_size[0] * 0.001) == 1)
		  {
		    strcpy(SDS_names_num_obs[0], sds_names[i]);
		    nsds_num_observations++;
		  }
		  else if((int)(in_sds_info.dim_size[0] * 0.001) == 2)
		  {
		    strcpy(SDS_names_num_obs[1], sds_names[i]);
		    nsds_num_observations++;
		  }
		  else if((int)(in_sds_info.dim_size[0] * 0.001) == 4)
		  {
		    strcpy(SDS_names_num_obs[3], sds_names[i]);
		    nsds_num_observations++;
		  }
		}
	      }
	    }
	  }
       if(sds_names) Free2D((void**)sds_names);
     }
   }

   return nsds_num_observations;
}
