/****************************************************************************
!C

!File: read_pixvals.c

!Description:
  Read observations at a specific pixel in a list of files.  

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Version 1.0    May, 2003
  Modified from read_pixvals.c in Nov, 2012, Version 2.0 by Xiaoping Zhang, Sigma Space Corp.	
	fixed the problem for reading floating data type SDS in l2 and l3 data files.
	fixed the problem for reading SDS data with bsq equal to 1 in l2 and l3 data files.
  Updated read_pixvals.c in February, 2013, Version 2.1 by Xiaoping Zhang, Sigma Space Corp.
	updated to work for MODIS l2g lite data(MOD09GA and MOD09GQ)	
  Updated read_pixvals.c in March, 2013, by Xiaoping Zhang, Sigma Space Corp.
	updated to make it work for NPP all products.
  Modifieded read_pixvals.c in April, 2013, Version 3.0 by Xiaoping Zhang, Sigma Space Corp.
	fixed to read_l2l3_obs_at_pts() and printL2gSdsVal().	
  Modified in February 2015 Version 3.1 by Xiaoping Zhang, SSAI		
	fixed a error which the offset is set incorrectly in the function of 
	read_l2l3_obs_at_pts for 3D/4D when bsq is equal to 1.
  Modified in February 2016 Version 3.2 by Xiaoping Zhang, SSAI	
  	to read in and print out one dimmension sds data.

  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
        3) replace get_attr_npp_sel_metadata() with get_attr_sel_metadata() and add meta.h
	 in the pre-include head files
	  
!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Sigma Space Corp. 
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "meta.h"
#include "qa_tool.h"
#include "main_util.h"
#include "alloc_mem.h"
#include "str_op.h"
#include "l2g.h"

#define HELP \
"NAME \n" \
"    read_pixvals - read MODIS Land HDF-EOS data product values at specified\n"\
"    pixel locations.\n" \
"\n" \
"SYNOPSIS \n" \
"    read_pixvals [-help] [filename] \n" \
"\n" \
"    read_pixvals -xy=col[.cs[.cs]],row[.rs[.rs]]|<coordinates filename> \n" \
"                 [-res=qkm|hkm|1km] filename(s) \n" \
" or    read_pixvals < argument file \n" \
"\n" \
"DESCRIPTION \n" \
"    Read the pixel values at specified locations in one or more input MODIS\n"\
"    Land data product and output to stdout. The pixel values for each SDS\n"\
"    in each file are output as separate lines. \n" \
"\n" \
"    If more than one input MODIS Land HDF-EOS data product is specified\n"\
"    then they must be all L2 products or all L2G/L3/L4 MODIS Land data\n"\
"    products. \n" \
"\n" \
"    The MODIS Land data products may contain SDSs with different spatial\n"\
"    dimensions corresponding to the 250m, 500m and 1km MODIS Land pixel\n"\
"    resolutions. In this case, the -res option is used to specify which of\n"\
"    the 1km, 500m or 250m pixel resolutions is referenced. If -res is not\n"\
"    specified then the xy location is assumed to reference the coarsest\n"\
"    spatial resolution of the different SDSs. \n" \
"\n" \
"    Sub pixel locations may be output by specifying a sub pixel offset (0\n"\
"    or 1 in the x and/or y axes). If not specified a 0 pixel offset is\n"\
"    assumed. See examples below. \n" \
"\n" \
"    This tool supports 2D/3D/4D and L2G compact and full format SDSs. Pixel\n"\
"    values for each layer of the 3D/4D and L2G SDSs are output. \n" \
"\n" \
"OPTIONS \n" \
"    -help [filename]         Display this help message. If the input\n"\
"                             filename is specified with this option, then\n"\
"                             the names of all the SDSs in the file are\n" \
"                             displayed. \n" \
"    -res=qkm|hkm|1km         Reference SDS resolution (qkm=250m, hkm=500m,\n"\
"                             1km=1000m) of the pixel location specified in\n"\
"                             the <96>xy argument. If unspecified the\n" \
"                             coarsest resolution of all the SDSs in the\n"\
"                             input file list is used. \n" \
"    -xy=col[.cs[.cs]],row[.rs[.rs]]|<coordinates filename> \n" \
"                             Column and row pixel locations (0-based) or\n"\
"                             name of an ASCII coordinates file containing\n"\
"                             the column and row pixel locations.  Multiple\n"\
"                             locations may be specified by repeating the xy\n"\
"                             option or by specifying the  x and y\n"\
"                             coordinates on different lines in the ASCII\n"\
"                             coordinates file.  \n" \
"                             Sub pixel offsets for higher spatial\n"\
"                             resolution SDS may be specified as col.cs\n"\
"                             row.rs. The offsets refer to the top left\n"\
"                             corner pixel. For example: \n" \
"                                -res=hkm -xy=100.0,200.0 \n" \
"                                  (read values at pixel 100,200 from the\n"\
"                                   500m SDS, and at pixel 200,400 from the\n"\
"                                   250m SDS) \n" \
"                                -res=hkm -xy=100.1,200.1 \n" \
"                                  (read values at pixel 100,200 from the\n"\
"                                   500m SDS, and at pixel 201,401 from the\n"\
"                                   250m SDS) \n" \
"                                -res=1km -xy=100.0.1,200.0.1 \n" \
"                                  (read values at pixel 100,200 from the\n"\
"                                   1km SDS, at pixel 200,400 from the 500m\n"\
"                                   SDS, and at pixel 401,801 from the 250m\n"\
"                                   SDS)\n" \
"    filename(s)              Input filename(s) \n" \
"\n" \
"Examples: \n" \
"	read_pixvals -xy=840,360 -res=1km MOD09GHK.A2010001.h20v11.005.2010007014814.hdf /LANDQA/f003/xzhang/QA_TOOLS/NPP/data_in/MOD09GQK.A2002134.h12v05.005.2012072205914.hdf /LANDQA/f003/xzhang/QA_TOOLS/NPP/data_in/MODAGAGG.A2012199.h09v05.005.2012201062210.hdf \n" \
"	read_pixvals -xy=840,360 -res=1km MOD09GA.A2012248.h20v11.005.2012250180349.hdf \n" \
"	read_pixvals -xy=840,360 -res=1km MOD09GA.A2012248.h20v11.005.2012250180349.hdf \n" \
"	read_pixvals -xy=840,360 -res=qkm MOD09GQK.A2002134.h12v05.005.2012072205914.hdf \n" \
"	read_pixvals -xy=840,360 -res=qkm MOD09GQ.A2010013.h35v10.005.2010015104431.hdf \n" \
"	read_pixvals -xy=1000,200 MOD09.A2010015.0015.005.2010020231418.hdf	\n" \
"	read_pixvals -xy=0,0 MOD11_L2.A2012309.2015.005.2012311022852.hdf	\n" \
"	read_pixvals -xy=840,360 -res=hkm NPP_SRFLIIP_L2.A2012259.1720.P1_03001.2012260065614.hdf \n" \
"	read_pixvals -xy=840,360 -res=1km NPP_DSRF1KD_L2GD.A2003193.h09v05.C1_00602.2009252165253.hdf \n" \
"	read_pixvals -xy=840,360 -res=hkm NPP_CMIP_L2C.A2012225.1620.P1_01164.2012250122910.hdf	\n" \
"	read_pixvals -xy=840,360 -res=hkm NPP_D17BRDFIP_L3.A2012222.h09v05.P1_01110.2012305165838.hdf	\n" \
"	read_pixvals -xy=840,360 -res=1km NPP_DSRFLD_L2GD.A2012225.h20v11.C1_01164.2013080104440.hdf	\n" \
"	read_pixvals -xy=0,1199 -res=1km MOD14A1.A2010001.h10v04.005.2010010211836.hdf	\n" \
"	read_pixvals -xy=4,5 SVM01_npp_d20141114_t2359374_e0001016_b15799_c20141115062014393469_noaa_ops.hdf\n" \
"	read_pixvals -xy=6,2 VNP02DNB.A2015262.1106.1100.2015321173151.hdf\n" \
"\n" \
"AUTHOR \n" \
"    Code: S. Devadiga and X. Zhang \n" \
"    Documentation: S. Devadiga and X. Zhang \n" \
"\n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:	\n" \
"   read_pixvals -xy=col[.cs[.cs]],row[.rs[.rs]]|<coordinates filename> \n" \
"                [-res=qkm|hkm|1km] filename(s) \n" \
"\n" \
"OPTIONS \n" \
"    -help [filename]         Display this help message. If the input\n"\
"                             filename is specified with this option, then\n"\
"                             the names of all the SDSs in the file are\n" \
"                             displayed. \n" \
"    -res=qkm|hkm|1km         Reference SDS resolution (qkm=250m, hkm=500m,\n"\
"                             1km=1000m) of the pixel location specified in\n"\
"                             the <96>xy argument. If unspecified the\n" \
"                             coarsest resolution of all the SDSs in the\n"\
"                             input file list is used. \n" \
"    -xy=col[.cs[.cs]],row[.rs[.rs]]|<coordinates filename> \n" \
"                             Column and row pixel locations (0-based) or\n"\
"                             name of an ASCII coordinates file containing\n"\
"                             the column and row pixel locations.  Multiple\n"\
"                             locations may be specified by repeating the xy\n"\
"                             option or by specifying the  x and y\n"\
"                             coordinates on different lines in the ASCII\n"\
"                             coordinates file.  \n" \
"                             Sub pixel offsets for higher spatial\n"\
"                             resolution SDS may be specified as col.cs\n"\
"                             row.rs. The offsets refer to the top left\n"\
"                             corner pixel. For example: \n" \
"                                -res=hkm -xy=100.0,200.0 \n" \
"                                  (read values at pixel 100,200 from the\n"\
"                                   500m SDS, and at pixel 200,400 from the\n"\
"                                   250m SDS) \n" \
"                                -res=hkm -xy=100.1,200.1 \n" \
"                                  (read values at pixel 100,200 from the\n"\
"                                   500m SDS, and at pixel 201,401 from the\n"\
"                                   250m SDS) \n" \
"                                -res=1km -xy=100.0.1,200.0.1 \n" \
"                                  (read values at pixel 100,200 from the\n"\
"                                   1km SDS, at pixel 200,400 from the 500m\n"\
"                                   SDS, and at pixel 401,801 from the 250m\n"\
"                                   SDS)\n" \
"    filename(s)              Input filename(s) \n" \
"\n"


int parse_cmd_read_pixvals(char **argv, int argc, char **xy_str, int *pt_cnt, int *res,
		char *xy_fname);
void get_xy_pts(char *xy_str, int *xy_pt, int *xy_sh, int *xy_sq);
void read_l2l3_obs_at_pts(char *fname, char **xy_str, char *xy_fname, int npt, int res, int gran_st);
void read_l2g_obs_at_pts(char *fname, char **xy_str, char *xy_fname, int npt, int res);
void print_l2g_sds_val(int **out_pnts, int32 **sds_val, sds_t *sds_info, sds_t *sdsc_info, int nsds, int l2g_st);
void printSdsVal(int **out_pnts, void **sdsVal, sds_t *sds_info, int nsds, int bsq, int l2g_st);
void printL2gSdsVal(int **out_pnts, int32 **sds_val, sds_t *sds_info, sds_t *sdsc_info, int nsds, float *rf);

int main(int argc, char **argv)
{
  int st, iarg, status, status1;
  int res, npt;
  int stdin_argc;
  int l2g_st, gran_st;
  char **xy_str, **stdin_argv;
  char mask_str[MAX_STR_LEN];
  char *ameta_str, *cmeta_str;
  char xy_fname[MAX_PATH_LENGTH];  
  int k, isds;
  int32 sd_id, sds_id, nsds;
  int32 rank, dim_size[4], dt, nattr;
  char name[MAX_SDS_NAME_LEN];
  char dim_str[MAX_STR_LEN], sds_name[MAX_SDS_NAME_LEN];
  char *meta_val1 = NULL;  
  char meta_name[MAX_META_NAME_LEN]; 
   
  if ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  if ((stdin_argv = (char **)Calloc2D(MAX_NUM_PARAM, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for stdin_argv in data_fedex(main)\n");
    exit(1);
  }
  if (argc == 1)
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);    
/*    if (parse_stdin(&stdin_argc, stdin_argv) == -1)
    {
      Free2D((void **)stdin_argv);
      exit(0);
    }*/
  }
  else
  {
    if (argc >= MAX_NUM_PARAM)
    {
      stdin_argc = MAX_NUM_PARAM;
      fprintf(stderr, "Too many input arguments. Using first %d arguments only . \n", stdin_argc);
    }
    else stdin_argc = argc;
    for (iarg=0; iarg<stdin_argc; iarg++)
      strcpy(stdin_argv[iarg], argv[iarg]);
  }

  /* MSZ 2/8/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 2/8/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  
  nsds = 0;
  for (iarg=1; iarg<argc; iarg++)
  {
    if (argv[iarg][0] != '-')
    {
      display_sds_info_of_file(argv[iarg]);
      nsds = get_nsds(argv[iarg]);
      if (nsds == 0)
	fprintf(stdout, "Input %s do not contain any valid SDS\n", argv[iarg]);
      break;
    }
  } /* for (i=2; . .  ) */  
    
  if ((xy_str = (char **)Calloc2D(MAX_NUM_PNTS, 25, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for xy_str in read_pixvals(main)\n");
  if (xy_str != NULL)
  {
    if ((st = parse_cmd_read_pixvals(stdin_argv, stdin_argc, xy_str, &npt, &res, xy_fname)) == -1)
      fprintf(stderr, "%s\n", USAGE);
    else if (st != 0)
    {
      iarg = 1;
      while (iarg < stdin_argc)
      {
        if (stdin_argv[iarg][0] != '-')
        {
	  if(L2G_format_check(stdin_argv[iarg]) > 0)
	    l2g_st = 1;
	  else
	    l2g_st = 0;
	  
	  if (l2g_st == 1)
	    read_l2g_obs_at_pts(stdin_argv[iarg], xy_str, xy_fname, npt, res);
	  else 
	    read_l2l3_obs_at_pts(stdin_argv[iarg], xy_str, xy_fname, npt, res, gran_st);
        }
        iarg++;
      }
    }
  }
  if(xy_str) Free2D((void **)xy_str);
  if(stdin_argv) Free2D((void **)stdin_argv);
  return 0;
}

int parse_cmd_read_pixvals(char **argv, int argc, char **xy_str, int *pt_cnt, int *res, 
		char *xy_fname)

/*
!C************************************************************************************
       
!Description:
  Parse the command line arguments.

!Input Parameters:
  argc: number of arguments in the command line.
  argv: string array containing the input arguments.

!Input/Output Paramenters: (none)

!Output Parameters: 
  xy_str:    String array containing the input column and row number.
  pt_cnt:    Count of input pixels.
  res:       Reference resolution for xy. Default is the coarsest resolution of all the SDSs
             in the input files.
  xy_fname:  input file name of the file containing the input column and row number(optional)

  (returns)  Completion status: 
             0 - successful completion
             1 - error exit
    
!Revision History: (none)

!Team-unique Header: (see file prolog)
 
!References and Credits: (see file prolog)

!Design Notes: (none)

!END
**************************************************************************************/     

{
  int st = 0;
  int i, fcnt, cnt, npt;
  int len_fstr;
  char res_str[10];
  
  st = 1;
  *res = npt = cnt = fcnt = 0;
  res_str[0] = xy_fname[0] = '\0';
  for (i=1; i<argc; i++)
    {
      if (is_arg_id(argv[i], "-res") == 0) get_arg_val(argv[i], res_str);
      else if (is_arg_id(argv[i], "-xy") == 0)
	{
	  if ((cnt == 0) && (charpos(argv[i], ',', 0) == -1))
	    get_arg_val(argv[i], xy_fname);
	  else
	    {
	      get_arg_val_arr(argv[i], xy_str, &npt);
	      if (npt == cnt+2) cnt = npt;
	      else 
		{
		  fprintf(stderr, "Ignoring invalid entry %s\n", argv[i]);
		  npt = cnt;
		}
	    }
	}
      else if (argv[i][0] == '-') fprintf(stderr, "Ignoring invalid option %s\n", argv[i]);
      else fcnt++;
    } /* for (i=1, . . ) */
    
  if (fcnt == 0)
    {
      st = -1;
      fprintf(stderr, "Missing input files \n");
    }
  else
    {
      if (strcmp(res_str, "qkm") == 0) *res=250;
      else if (strcmp(res_str, "hkm") == 0) *res=500;
      else if (strcmp(res_str, "1km") == 0) *res=1000;
      else fprintf(stderr, "Reference resolution will be set to coarse SDS resolution\n");
      len_fstr = (int)strlen(xy_fname);
      if ((npt == 0) && (len_fstr <= 0))
	{
	  st = -1;
	  fprintf(stderr, "Input point sets or a point set file is required . . \n");
	}
    }
  *pt_cnt = npt/2;
  return st;
}

void get_xy_pts(char *xy_str, int *pt, int *xy_sh, int *xy_sq)
{
  int len;
  int p1, p2;
  char tmp_str[10];

  *xy_sh = *xy_sq = 0;
  if ((p1 = charpos(xy_str, '.', 0)) == -1)
    *pt = (int)atoi(xy_str);
  else
  {
    len = (int)strlen(xy_str);
    strmid(xy_str, 0, p1, tmp_str);
    *pt = (int)atoi(tmp_str);
    p2 = charpos(xy_str, '.', p1+1);
    if (p2 == -1)
    {
      strmid(xy_str, p1+1, len-p1-1, tmp_str);
      *xy_sh = (int)atoi(tmp_str);
    }
    else
    {
      strmid(xy_str, p1+1, p2-p1-1, tmp_str);
      *xy_sh = (int)atoi(tmp_str);
      strmid(xy_str, p2+1, len-p2-1, tmp_str);
      *xy_sq = (int)atoi(tmp_str);
    }
  }
  if (*xy_sh > 1) 
  {
    *xy_sh = 0;
    fprintf(stderr, "Invalid sub sample specified in %s. Set to default 0\n", xy_str);
  }
  if (*xy_sq > 1) 
  {
    *xy_sq = 0;
    fprintf(stderr, "Invalid sub sample specified in %s. Set to default 0\n", xy_str);
  }
}

void read_l2l3_obs_at_pts(char *fname, char **xy_str, char *xy_fname, int npt, int res, int gran_st)
{
  FILE *fp;
  float *rf, f_fval;
  int done = 1, bsq, rank, st=1;
  int min_rows, min_cols;
  int max_ndata, max_data_size, max_isds;
  int x_sh, y_sh, x_sq, y_sq;
  int nobs, iobs, irow, icol;
  int pt_x, pt_y, f_val, cres;
  int id, idim, ipt, isds, i, nsds;
  int ndata[100], offset, **out_pnts;
  sds_t *sds_info;
  void *attr_buf, *data_in, **sdsVal;
  int32 edge[4] = {0, 0, 0, 0};
  int32 start[4] = {0, 0, 0, 0};
  int32 attr_type, attr_cnt, **sds_val;
  char **sds_names = NULL;
  char **sds_names_exclusion = NULL;
  char x_str[15], y_str[15];

  fprintf(stdout, "\nReading input file %s\n", fname);
  fprintf(stdout, "-----------------------------------------------------------------------\n");
  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for sds_names in read_l2l3_obs_at_pts\n");
    return;
  }
  /* for HDF5 input, excluded StructMetadata.0 for sds_names */
  if ((sds_names_exclusion = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
  {
      fprintf(stderr, "Cannot allocate memory for sds_names_exclusion in read_l2l3_obs_at_pts(). \n");
      return;
  }
  strcpy(sds_names_exclusion[0], "StructMetadata.0");
  strcpy(sds_names_exclusion[1], "CoreMetadata.0");
  strcpy(sds_names_exclusion[2], "ArchiveMetadata.0");

  if ((nsds = get_sds_names_exclusions(fname, sds_names, sds_names_exclusion, 3)) > 0)
  {
    if ((sds_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)
      fprintf(stderr, "Cannot allocate memory for sds_info in read_l2l3_obs_at_pts\n");
    if ((rf = (float *)calloc(nsds, sizeof(float))) == NULL)
      fprintf(stderr, "Cannot allocate memory for rf in read_l2l3_obs_at_pts\n");
    if ((sds_info != NULL) && (rf != NULL))
    {
      fprintf(stdout, "SDS number = %d\n", nsds);
      for (isds=0; isds<nsds; isds++)
      {
        sds_info[isds].sds_id = -1;
        strcpy(sds_info[isds].name, sds_names[isds]);
        sds_info[isds].sd_id = (isds == 0) ? -1 : sds_info[0].sd_id;
	sds_info[isds].H5IN = (isds == 0) ? -1 : sds_info[0].H5IN;
        if (get_sds_info(fname, &sds_info[isds]) != -1)
        {         
	  if ((attr_buf = get_sds_attr(&sds_info[isds], "_FillValue", &attr_type,
		&attr_cnt)) != NULL)
          {
            switch(sds_info[isds].data_type)
            {
              case 5: f_fval = ((float32 *)attr_buf)[0]; break;
	      case 6: f_fval = ((float64 *)attr_buf)[0]; break;
	      case 20: f_val = ((int8 *)attr_buf)[0]; break;
              case 21: f_val = ((uint8 *)attr_buf)[0]; break;
              case 22: f_val = ((int16 *)attr_buf)[0]; break;
              case 23: f_val = ((uint16 *)attr_buf)[0]; break;
              case 24: f_val = ((int32 *)attr_buf)[0]; break;
              case 25: f_val = ((uint32 *)attr_buf)[0]; break;
            }
            free(attr_buf);
	    if (sds_info[isds].data_type == 5 || sds_info[isds].data_type == 6)
	      fprintf(stdout, "\t%s (Fill_Value = %f)\n", sds_info[isds].name, f_fval);
	    else
	      fprintf(stdout, "\t%s (Fill_Value = %d)\n", sds_info[isds].name, f_val);
          }
        }
      } /* for (isds=0; . . ) */
      max_data_size = max_isds = 0;
      for (i=0; i<nsds; i++)
      { 	  
	rank = sds_info[i].rank;
	bsq = ((rank == 2) || (sds_info[i].dim_size[0] < sds_info[i].dim_size[rank-1])) ? 1 : 0; 	 
   	     
	if (res == 0)
	{
	  if (rank >= 2)
	  {
	    if (bsq == 0)
	    {	
	      min_rows = cres = sds_info[i].dim_size[0];
	      min_cols = sds_info[i].dim_size[1];
              for (isds=1; isds<nsds; isds++)
		if (cres > sds_info[isds].dim_size[0]) 
		{
		  min_rows = cres = sds_info[isds].dim_size[0];
		  min_cols =  sds_info[isds].dim_size[1];
		}
	    }
	    else
	    {
	      min_rows = cres = sds_info[i].dim_size[rank-2];
	      min_cols = sds_info[i].dim_size[rank-1];
              for (isds=1; isds<nsds; isds++)
		if (cres > sds_info[isds].dim_size[rank-2]) 
		{
		  min_rows = cres = sds_info[isds].dim_size[rank-2];
		  min_cols =  sds_info[isds].dim_size[rank-1];
		}            
	    }
	  }
	  else 
	  {
	    min_rows = sds_info[i].dim_size[0];
	    cres = sds_info[i].dim_size[0];
	    min_cols = 1;
	  }
	}
	else
	{
	  switch(res)
	  {
	    case 250: cres = (gran_st == 1) ? 8120 : 4800; break;
	    case 500: cres = (gran_st == 1) ? 4060 : 2400; break;
	    case 1000: cres = (gran_st == 1) ? 2030 : 1200; break;
	    default: fprintf(stderr, "Wrong resolution %d encountered\n", res); break;
	  }
	} 

//printf("cres=%d\n",cres);
//printf("sds_info[i].dim_size[rank-2]=%f\n",(float)sds_info[i].dim_size[rank-2]);	
	if (bsq == 0)
	  rf[i] = (float)sds_info[i].dim_size[0]/(float)cres; 
	else  rf[i] = (float)sds_info[i].dim_size[rank-2]/(float)cres; 	
	if (res == 0) rf[i] = 1;
//printf("rf[%d]=====%f\n",i, rf[i]);	     
      }
//printf(" min_rows  == %d\n",   min_rows);
//printf(" min_cols  == %d\n",   min_cols);
      max_data_size = max_isds = 0;
      for (isds=0; isds<nsds; isds++)
      {	
	if (sds_info[isds].sds_id != -1)
	{
	  if (sds_info[isds].data_size > max_data_size) { max_data_size = sds_info[isds].data_size; max_isds = isds; }
	}   
      }

      if ((sdsVal = (void **)Calloc2D(nsds, MAX_NUM_OBS, sds_info[max_isds].data_size)) == NULL)
	fprintf(stderr, "Cannot allocate memory for sds_val in read_l2l3_obs_at_pts\n");
      if ((out_pnts = (int **)Calloc2D(nsds, 2, sizeof(int))) == NULL)
	fprintf(stderr, "Cannot allocate memory for out_pnts in read_l2l3_obs_at_pts\n");
      if ((sdsVal != NULL) && (out_pnts != NULL))
      {
	max_ndata = max_isds = 0;
	
        for (isds=0; isds<nsds; isds++)
	{
	  rank = sds_info[isds].rank;
	  ndata[isds] = 1;
	  if (rank >= 2)
	  {   
	    bsq = ((rank == 2) || (sds_info[isds].dim_size[0] < sds_info[isds].dim_size[rank-1])) ? 1 : 0; 	         
	    if (bsq == 1) 
	    {
	      if (sds_info[isds].sds_id != -1)
              {
		for (idim=0, ndata[isds]=1; idim<rank-2; idim++)
		  ndata[isds] *= (int)sds_info[isds].dim_size[idim];
		ndata[isds] *= (int)sds_info[isds].dim_size[rank-1];	     	      
	      }   	    
	    } 
	    else 
	    {
	      if (sds_info[isds].sds_id != -1)
              {
		for (idim=1, ndata[isds]=1; idim<sds_info[isds].rank; idim++)
		  ndata[isds] *= (int)sds_info[isds].dim_size[idim];
	      }
	    }
	  }
	  else if (rank == 1)
	  {
	    if (sds_info[isds].sds_id != -1)
	      ndata[isds] = (int)sds_info[isds].dim_size[0];   
	  }			   	  
	}  
  	  
	if (npt == 0)
	{
	  if ((fp = fopen(xy_fname, "r")) == NULL)
	    fprintf(stderr, "Cannot open file %s\n", xy_fname);
	}
	ipt = 0;
	while (done)
	{
	  if (npt > 0)
	  { 
	    id = ipt*2;
	    get_xy_pts(xy_str[id], &pt_x, &x_sh, &x_sq);
	    get_xy_pts(xy_str[id+1], &pt_y, &y_sh, &y_sq);
	    ++ipt;
	    if (ipt > npt) break;
	  }
	  else 
	  {
	    if (fscanf(fp, "%s %s", x_str, y_str) == 2)
	    {
	      get_xy_pts(x_str, &pt_x, &x_sh, &x_sq);
	      get_xy_pts(y_str, &pt_y, &y_sh, &y_sq);
	    }
	    if (feof(fp) != 0) { fclose(fp); break; }
	  }
	  for (isds=0; isds<nsds; isds++)
	  {
            if ((data_in = (void *)calloc(ndata[isds], sds_info[isds].data_size)) == NULL)
              fprintf(stderr, "Cannot allocate memory for data_in in read_l2l3_obs_at_pts\n");
	    else
	    {
	      for (i=0; i<4; i++)
	      {
	        start[i] = 0;
	        edge[i] = 0;
	      }
	      	      
	      irow = (int)((float)pt_y*rf[isds]);
	      icol = (int)((float)pt_x*rf[isds]);
	
//printf("irow=%d\n",irow);
//printf("icol=%d\n",icol);
/*printf("isds=%d sds_info[%d].name=%s\n", isds, isds, sds_info[isds].name);		
printf("sds_info[%d].data_size=%d\n", isds, sds_info[isds].data_size);
printf("sds_info[%d].data_type=%d\n", isds, sds_info[isds].data_type);
printf("sds_info[%d].dim_size[0]=%d\n", isds, sds_info[isds].dim_size[0]);
printf("ndata[%d]=%d\n", isds, ndata[isds]);	*/

	      if (sds_info[isds].rank >= 2)
	      {			
		if (bsq == 0)
		{
		  if (sds_info[isds].dim_size[0]-1 < irow) {
		    fprintf(stderr, "Invalid row range for sds %s\n", sds_info[isds].name);
		    st = -1;
		    continue;
                  }
		  if (sds_info[isds].dim_size[1]-1 < icol) {
		    fprintf(stderr, "Invalid column range for sds %s\n", sds_info[isds].name); 
		    st = -1;
		    continue;
		  }
		}
		else
		{
		  if (sds_info[isds].dim_size[rank-2]-1 < irow) {
		    fprintf(stderr, "Invalid row range for sds %s\n", sds_info[isds].name);
                    st = -1;
		    continue; 
		  }
		  if (sds_info[isds].dim_size[rank-1]-1 < icol) {
		    fprintf(stderr, "Invalid column range for sds %s\n", sds_info[isds].name); 
		    st = -1; 
		    continue;  		  
		  }       	              
		}		
	      }

	      
/*printf("sds_info[%d].name=%s  sds_info[%d].data_type=%d ndata[%d]=%d,sds_info[%d].rank=%d\n",
  isds,sds_info[isds].name,isds,sds_info[isds].data_type,isds, ndata[isds], isds,sds_info[isds].rank);  */  
	      if (rf[isds] == 2.0) { irow += y_sh; icol += x_sh; }
	      else if (rf[isds] == 4.0) { irow += y_sq; icol += x_sq; }
	      rank = sds_info[isds].rank;
	      bsq = ((rank == 2) || (sds_info[isds].dim_size[0] < sds_info[isds].dim_size[rank-1])) ? 1 : 0; 						
	      if (rank >= 2)
	      {
		if (bsq == 0)
		{
	          start[0] = irow;
		  edge[0] = 1;
		  for (idim=1; idim<sds_info[isds].rank; idim++)
                    edge[idim] = sds_info[isds].dim_size[idim];
		}
		else
		{
		  start[rank-2] = irow;
		  for (idim=0; idim<sds_info[isds].rank; idim++)
                    edge[idim] = sds_info[isds].dim_size[idim];
		  edge[rank-2] = 1;   	        	              
		}
	      }
	      else 
	      {
	        start[0] = 0;
		edge[0] = sds_info[isds].dim_size[0];	  
	      }
	      if(read_sel_sds_data(&sds_info[isds], start, edge, data_in) == FAIL)
                fprintf(stderr, "Cannot read data line for sds %s\n", sds_info[isds].name);
	      if (rank >= 2)
	      {
		if (bsq == 0)
		{
	          for (idim=2, nobs=1; idim<sds_info[isds].rank; idim++)
		    nobs *= sds_info[isds].dim_size[idim];	
		  offset = icol*nobs;
		  out_pnts[isds][0] = icol;
		  out_pnts[isds][1] = irow;
		  for (iobs=0; iobs<nobs; iobs++, offset++) 
		  {          			
		    switch(sds_info[isds].data_type)
                    {
                      case 5: ((float32 *)sdsVal[isds])[iobs] = ((float32 *)data_in)[offset]; break;
		      case 6: ((float64 *)sdsVal[isds])[iobs] = ((float64 *)data_in)[offset]; break;
		      case 20: ((int8 *)sdsVal[isds])[iobs] = ((int8 *)data_in)[offset]; break;
                      case 21: ((uint8 *)sdsVal[isds])[iobs] = ((uint8 *)data_in)[offset]; break;
                      case 22: ((int16 *)sdsVal[isds])[iobs] = ((int16 *)data_in)[offset]; break;
                      case 23: ((uint16 *)sdsVal[isds])[iobs] = ((uint16 *)data_in)[offset]; break;
                      case 24: ((int32 *)sdsVal[isds])[iobs] = ((int32 *)data_in)[offset]; break;
                      case 25: ((uint32 *)sdsVal[isds])[iobs] = ((uint32 *)data_in)[offset]; break;
                    }
		  } /* for (iobs=0; . . ) */	      
		}	
		else
		{
	          for (idim=0, nobs=1; idim<sds_info[isds].rank-2; idim++)
		    nobs *= sds_info[isds].dim_size[idim];
		  offset = icol;
		  out_pnts[isds][0] = icol;
		  out_pnts[isds][1] = irow;
		  for (iobs=0; iobs<nobs; iobs++) 
		  {          			
		    switch(sds_info[isds].data_type)
                    {
                      case 5: ((float32 *)sdsVal[isds])[iobs] = ((float32 *)data_in)[offset]; break;
		      case 6: ((float64 *)sdsVal[isds])[iobs] = ((float64 *)data_in)[offset]; break;
		      case 20: ((int8 *)sdsVal[isds])[iobs] = ((int8 *)data_in)[offset]; break;
                      case 21: ((uint8 *)sdsVal[isds])[iobs] = ((uint8 *)data_in)[offset]; break;
                      case 22: ((int16 *)sdsVal[isds])[iobs] = ((int16 *)data_in)[offset]; break;
                      case 23: ((uint16 *)sdsVal[isds])[iobs] = ((uint16 *)data_in)[offset]; break;
                      case 24: ((int32 *)sdsVal[isds])[iobs] = ((int32 *)data_in)[offset]; break;
                      case 25: ((uint32 *)sdsVal[isds])[iobs] = ((uint32 *)data_in)[offset]; break;
                    }
		    offset = offset + sds_info[isds].dim_size[rank-1];
		  } /* for (iobs=0; . . ) */	      	        
		}
	      }
	      else 
	      {	 
		nobs = sds_info[isds].dim_size[0];		
		offset = 0;
		for (iobs=0; iobs<nobs; iobs++, offset++) 
		{          			
		  switch(sds_info[isds].data_type)
                  {
                    case 5: ((float32 *)sdsVal[isds])[iobs] = ((float32 *)data_in)[offset]; break;
		    case 6: ((float64 *)sdsVal[isds])[iobs] = ((float64 *)data_in)[offset]; break;
		    case 20: ((int8 *)sdsVal[isds])[iobs] = ((int8 *)data_in)[offset]; break;
                    case 21: ((uint8 *)sdsVal[isds])[iobs] = ((uint8 *)data_in)[offset]; break;
                    case 22: ((int16 *)sdsVal[isds])[iobs] = ((int16 *)data_in)[offset]; break;
                    case 23: ((uint16 *)sdsVal[isds])[iobs] = ((uint16 *)data_in)[offset]; break;
                    case 24: ((int32 *)sdsVal[isds])[iobs] = ((int32 *)data_in)[offset]; break;
                    case 25: ((uint32 *)sdsVal[isds])[iobs] = ((uint32 *)data_in)[offset]; break;	
                  }		  		  
		} /* for (iobs=0; . . ) */	  				  		      
	      }
	    }  
	    free(data_in);     	     
          } /* for (isds=0; . . . ) */
  
	  if (st != -1) 
	  {	  
	    if (rank >= 2 && res != 0)
	    {
	      if (res == 0)
  		fprintf(stdout, "Observations: pixel (%d %d) at ref resolution (%d x %d)\n", 
			  pt_x, pt_y, min_cols, min_rows);
	      else
  		fprintf(stdout, "Observations: pixel (%d %d) at ref resolution %dm\n", pt_x, pt_y, res);  
	    }      
	    printSdsVal(out_pnts, sdsVal, sds_info, nsds, bsq, 0);
	  }
	} /* for (ipt=0; . . ) */
	for (isds=0; isds<nsds; isds++)
	  close_sds(&sds_info[isds], sds_info[0].H5IN);
	if(out_pnts) Free2D((void **)out_pnts);
        if(sdsVal) Free2D((void **)sdsVal);
      }
      close_sd(&sds_info[0], sds_info[0].H5IN);
    }
    if(rf) free(rf);
    if(sds_info) free(sds_info); 
  } /* if ((nsds =  . . ) */
  if(sds_names) Free2D((void **)sds_names);
  if(sds_names_exclusion) Free2D((void **)sds_names_exclusion);
}

void read_l2g_obs_at_pts(char *fname, char **xy_str, char *xy_fname, int npt, int res)
{
  FILE *fp;
  float *rf;
  int done = 1;
  int ic, ir, nobs_ic;
  int min_rows, min_cols;
  int id, ipt, iobs, nobs, nobs_500m;
  int st1, st2, f_val, cres;
  int pt_x, pt_y, **out_pnts;
  int ndata, ndata_500m, max_size, offset;
  int x_sh, y_sh, x_sq, y_sq;
  int irow, icol, irow_1km, icol_1km, irow_500m, icol_500m, isds, nsds;
  int32 edge[2], start[2];
  void *attr_buf, *data_in, *data_in_500m;
  int8 *data_nobs, *data_nobs_500m;
  int32 *data_nadd_obs, *data_nadd_obs_1km, *data_nadd_obs_500m;
  int32 attr_type, attr_cnt, **sds_val;
  char **sds_names, x_str[15], y_str[15];
  sds_t *sds1_info, *sdsc_info, sds_nadd_obs_info;
  sds_t sds_nadd_obs_1km_info, sds_nadd_obs_500m_info;
  
  fprintf(stdout, "\nReading input file %s\n", fname);
  fprintf(stdout, "-----------------------------------------------------------------------\n");
  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for sds_names in read_l2g_obs_at_pts\n");
    return;
  }
  if ((nsds = get_l2g_sds_names(fname, sds_names)) > 0)
  {
    if ((strstr(fname, "MOD09GA") == NULL) && (strstr(fname, "MYD09GA") == NULL) &&
        (strstr(fname, "NPP_DSRFLD_L2GD") == NULL))
    
    {        
      nsds++;
      if ((sds1_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)
	fprintf(stderr, "Cannot allocate memory for sds1_info in read_l2g_obs_at_pts\n");
      if ((sdsc_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)
	fprintf(stderr, "Cannot allocate memory for sdsc_info in read_l2g_obs_at_pts\n");
      if ((rf = (float *)calloc(nsds, sizeof(float))) == NULL)
	fprintf(stderr, "Cannot allocate memory for rf in read_l2g_obs_at_pts\n");
      if ((sds1_info != NULL) && (sdsc_info != NULL) && (rf != NULL))
      {
//	for (isds=0; isds<nsds-1; isds++)
//	  fprintf(stdout, "\tsds_names[%d] = %s\n", isds, sds_names[isds]);       
//	fprintf(stdout, "SDS in file and Fill values\n");
	for (isds=0; isds<nsds; isds++)
	{       
	  sds1_info[isds].sds_id = sdsc_info[isds].sds_id = -1;
	  if (isds == 0)
	  {
	    sds1_info[isds].sd_id = -1;
	    strcpy(sds1_info[isds].name, sds_name_nobs);
            st1 = st2 = get_sds_info(fname, &sds1_info[isds]);
	  }
	  else
	  {
            sprintf(sds1_info[isds].name, "%s_1", sds_names[isds-1]);
            sprintf(sdsc_info[isds].name, "%s_c", sds_names[isds-1]);
	    sds1_info[isds].sd_id = sdsc_info[isds].sd_id = sds1_info[0].sd_id;
	    sds1_info[isds].H5IN = sdsc_info[isds].H5IN = sds1_info[0].H5IN;
            st1 = get_sds_info(fname, &sds1_info[isds]);
	    st2 = get_sds_info(fname, &sdsc_info[isds]);
	  }
  //	if ((st1 != -1) && (st2 != -1))
	  if (st1 != -1) 	
          {
            if ((attr_buf = get_sds_attr(&sds1_info[isds], "_FillValue", &attr_type,
		  &attr_cnt)) != NULL)
            {
              switch(sds1_info[isds].data_type)
              {
        	case 20: f_val = ((int8 *)attr_buf)[0]; break;
        	case 21: f_val = ((uint8 *)attr_buf)[0]; break;
        	case 22: f_val = ((int16 *)attr_buf)[0]; break;
        	case 23: f_val = ((uint16 *)attr_buf)[0]; break;
        	case 24: f_val = ((int32 *)attr_buf)[0]; break;
        	case 25: f_val = ((uint32 *)attr_buf)[0]; break;
              }
              free(attr_buf);
            }
//	    if (isds == 0)
//	      fprintf(stdout, "\t%s (Fill_Value = %d)\n", sds1_info[isds].name, f_val);
//	    else
//	      fprintf(stdout, "\t%s(c) (Fill_Value = %d)\n", sds1_info[isds].name, f_val);
          }
	} /* for (isds=0; . . ) */

	sds_nadd_obs_info.sds_id = -1;
	sds_nadd_obs_info.sd_id = sds1_info[0].sd_id;
	sds_nadd_obs_info.H5IN = sds1_info[0].H5IN;
	strcpy(sds_nadd_obs_info.name, sds_name_nadd_obs);     
	if (get_sds_info(fname, &sds_nadd_obs_info) == -1)
          fprintf(stderr, "Result may be in error \n");
	if ((data_nadd_obs = (int32 *)calloc(sds_nadd_obs_info.dim_size[0], sizeof(int32))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for data_nadd_obs in read_l2g_obs_at_pts\n");
	start[0] = 0; edge[0] = sds_nadd_obs_info.dim_size[0];
	start[1] = edge[1] = 0;
	if (read_sel_sds_data(&sds_nadd_obs_info, start, edge, data_nadd_obs) == FAIL)
          fprintf(stderr, "Cannot read data line for sds %s\n", sds_nadd_obs_info.name);

	if (res == 0)
	{
	  min_rows = cres = sds1_info[0].dim_size[0];
	  min_cols = sds1_info[0].dim_size[1];
          for (isds=1; isds<nsds; isds++)
	    if (cres > sds1_info[isds].dim_size[0]) 
	    {
	      min_rows = cres = sds1_info[isds].dim_size[0];
	      min_cols = sds1_info[isds].dim_size[1];
	    }
	}
	else
	{
	  if (res == 250) cres = 4800;
	  else if (res == 500) cres = 2400;
	  else if (res == 1000) cres = 1200;
	  else fprintf(stderr, "Wrong resolution %d encountered\n", res);
	}
	for (isds=0; isds<nsds; isds++)
	  rf[isds] = (float)sds1_info[isds].dim_size[0]/(float)cres;

	if ((sds_val = (int32 **)Calloc2D(nsds, MAX_NUM_OBS, sizeof(int32))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for sds_val in read_l2g_obs_at_pts\n");
	if ((out_pnts = (int **)Calloc2D(nsds, 2, sizeof(int))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for out_pnts in read_l2l3_obs_at_pts\n");
	if ((sds_val != NULL) && (out_pnts != NULL))
	{
          ndata = sds1_info[0].dim_size[1];
          if ((data_nobs = (int8 *)calloc(ndata, sizeof(int8))) == NULL)
            fprintf(stderr, "Cannot allocate memory for data_nobs in read_l2g_obs_at_pts\n");
	  max_size = sds1_info[0].data_size;
	  for (isds=1; isds<nsds; isds++)
	    if (sds1_info[isds].data_size > max_size)
	      max_size = sds1_info[isds].data_size;
	  if ((data_in = (void *)calloc(ndata, max_size)) == NULL)
	    fprintf(stderr, "Cannot allocate memory for data_in in read_l2g_obs_at_pts\n");

	  if ((data_nobs != NULL) && (data_in != NULL))
	  {
            if (npt == 0)
            {
              if ((fp = fopen(xy_fname, "r")) == NULL)
        	fprintf(stderr, "Cannot open file %s\n", xy_fname);
            }
            ipt = 0;
            while (done)
            {
              if (npt > 0)
              {
        	id = ipt*2;
        	get_xy_pts(xy_str[id], &pt_x, &x_sh, &x_sq);
        	get_xy_pts(xy_str[id+1], &pt_y, &y_sh, &y_sq);
        	++ipt;
        	if (ipt > npt) break;
              }
              else
              {
        	if (fscanf(fp, "%s %s", x_str, y_str) == 2)
        	{
                  get_xy_pts(x_str, &pt_x, &x_sh, &x_sq);
                  get_xy_pts(y_str, &pt_y, &y_sh, &y_sq);
        	}
        	if (feof(fp) != 0) { fclose(fp); break; }
              }
	      for (isds=1; isds<nsds; isds++)
	      { 
/*		if ((sds1_info[isds].sds_id != -1) && (sdsc_info[isds].sds_id != -1 && (sds1_info[0].sds_id != -1))*/
		if ((sds1_info[isds].sds_id != -1) && (sds1_info[0].sds_id != -1))			
		{
	          irow = (int)((float)pt_y*rf[isds]);
	          icol = (int)((float)pt_x*rf[isds]);
	          if (rf[isds] == 2.0) { irow += y_sh; icol += x_sh; }
	          else if (rf[isds] == 4.0) { irow += y_sq; icol += x_sq; }
                  start[0] = irow; edge[0] = 1;
	          start[1] = 0; edge[1] = ndata;
	          out_pnts[isds][0] = icol;
	          out_pnts[isds][1] = irow;
	          if (isds == 1)
	          {
		    if (read_sel_sds_data(&sds1_info[0], start, edge, data_nobs) == FAIL)
                      fprintf(stderr, "Cannot read data line for sds %s\n", sds1_info[0].name);
	            nobs = ((int8 *)data_nobs)[icol];
/*		    printf("nobs=%d\n",nobs);		*/
	            sds_val[0][0] = nobs;
		    out_pnts[0][0] = icol;
		    out_pnts[0][1] = irow;
	          }
	          if (nobs > 0)
	          {
		    if (read_sel_sds_data(&sds1_info[isds], start, edge, data_in) == FAIL)
                      fprintf(stderr, "Cannot read data line for sds %s\n", sds1_info[isds].name);
                    switch(sds1_info[isds].data_type)
                    {
                      case 20: sds_val[isds][0] = ((int8 *)data_in)[icol]; break;
                      case 21: sds_val[isds][0] = ((uint8 *)data_in)[icol]; break;
                      case 22: sds_val[isds][0] = ((int16 *)data_in)[icol]; break;
                      case 23: sds_val[isds][0] = ((uint16 *)data_in)[icol]; break;
                      case 24: sds_val[isds][0] = ((int32 *)data_in)[icol]; break;
                      case 25: sds_val[isds][0] = ((uint32 *)data_in)[icol]; break;
                    }
	          }
	          if (nobs > 1 && sdsc_info[isds].sds_id != -1 )
	          {
	            for (ir=0, offset=0; ir<irow; ir++)
		      offset += ((int32 *)data_nadd_obs)[ir];
		    for (ic=0; ic<icol; ic++)
		    {
		      nobs_ic = ((int8 *)data_nobs)[ic];
		      if (nobs_ic > 1) offset = offset + nobs_ic - 1;
		    }
		    start[0] = offset; edge[0] = nobs-1;
		    start[1] = edge[1] = 0;
		    if (read_sel_sds_data(&sdsc_info[isds], start, edge, data_in) == FAIL)
                      fprintf(stderr, "Cannot read data line for sds %s\n", sdsc_info[isds].name);
		    for (iobs=1, offset=0; iobs<nobs; iobs++, offset++)
                      switch(sdsc_info[isds].data_type)
                      {
                	case 20: sds_val[isds][iobs] = ((int8 *)data_in)[offset]; break;
                	case 21: sds_val[isds][iobs] = ((uint8 *)data_in)[offset]; break;
                	case 22: sds_val[isds][iobs] = ((int16 *)data_in)[offset]; break;
                	case 23: sds_val[isds][iobs] = ((uint16 *)data_in)[offset]; break;
                	case 24: sds_val[isds][iobs] = ((int32 *)data_in)[offset]; break;
                	case 25: sds_val[isds][iobs] = ((uint32 *)data_in)[offset]; break;
                      }
	          } /* if (nobs > 1) */
        	} /* if (sds1_info . . . ) */	          
	      } /* for (isds=1; . . . ) */
	      if (res == 0)
  		fprintf(stdout, "Observations: pixel (%d %d) at ref resolution (%d x %d)\n", 
			  icol, irow, min_cols, min_rows);
	      else
  		fprintf(stdout, "Observations: pixel (%d %d) at ref resolution %dm\n", icol, irow, res);
              print_l2g_sds_val(out_pnts, sds_val, sds1_info, sdsc_info, nsds, 1);
	    } /* for (ipt=0; . . . ) */
	    free(data_in);
	    free(data_nobs);
	  } /* if ((data_nobs != . . .  )) */
	  Free2D((void **)out_pnts);
	  Free2D((void **)sds_val);
	} /* if (sds_val != . . . ) */
	for (isds=0; isds<nsds; isds++)
	{
	  close_sds(&sds1_info[isds], sds1_info[0].H5IN);
	  close_sds(&sdsc_info[isds], sdsc_info[0].H5IN);
	} 
	close_sd(&sds1_info[0], sds1_info[0].H5IN);
	free(rf);
	free(sds1_info); 
	free(sdsc_info); 
	free(data_nadd_obs);	
      } /* if ((sds1_info != . . . )) */
    }
    else
    {
      nsds = nsds + 2;
      if ((sds1_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)
	fprintf(stderr, "Cannot allocate memory for sds1_info in read_l2g_obs_at_pts\n");
      if ((sdsc_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)
	fprintf(stderr, "Cannot allocate memory for sdsc_info in read_l2g_obs_at_pts\n");
      if ((rf = (float *)calloc(nsds, sizeof(float))) == NULL)
	fprintf(stderr, "Cannot allocate memory for rf in read_l2g_obs_at_pts\n");
      if ((sds1_info != NULL) && (sdsc_info != NULL) && (rf != NULL))
      {
//        for (isds=0; isds<nsds-2; isds++)
//          fprintf(stdout, "\tsds_names[%d] = %s\n", isds, sds_names[isds]); 	
//	fprintf(stdout, "SDS in file and Fill values\n");
	for (isds=0; isds<nsds; isds++)
	{    
	  sds1_info[isds].sds_id = sdsc_info[isds].sds_id = -1;
	  if (isds == 0)
	  {
	    sds1_info[isds].sd_id = -1;
	    strcpy(sds1_info[isds].name, sds_name_nobs_1km);
            st1 = st2 = get_sds_info(fname, &sds1_info[isds]);
	  }
	  else if (isds == 1)
	  {
	    sds1_info[isds].sd_id = -1;
	    strcpy(sds1_info[isds].name, sds_name_nobs_500m);
            st1 = st2 = get_sds_info(fname, &sds1_info[isds]);			
	  }	  
	  else
	  {
            sprintf(sds1_info[isds].name, "%s_1", sds_names[isds-2]);
            sprintf(sdsc_info[isds].name, "%s_c", sds_names[isds-2]);
	    sds1_info[isds].sd_id = sdsc_info[isds].sd_id = sds1_info[0].sd_id;
	    sds1_info[isds].H5IN = sdsc_info[isds].H5IN = sds1_info[0].H5IN;
            st1 = get_sds_info(fname, &sds1_info[isds]);
	    st2 = get_sds_info(fname, &sdsc_info[isds]);
	  }
	  if (st1 != -1) 	
          {
            if ((attr_buf = get_sds_attr(&sds1_info[isds], "_FillValue", &attr_type,
		  &attr_cnt)) != NULL)
            {
              switch(sds1_info[isds].data_type)
              {
        	case 20: f_val = ((int8 *)attr_buf)[0]; break;
        	case 21: f_val = ((uint8 *)attr_buf)[0]; break;
        	case 22: f_val = ((int16 *)attr_buf)[0]; break;
        	case 23: f_val = ((uint16 *)attr_buf)[0]; break;
        	case 24: f_val = ((int32 *)attr_buf)[0]; break;
        	case 25: f_val = ((uint32 *)attr_buf)[0]; break;
              }
              free(attr_buf);
            }
//	    if (isds < 2)
//	      fprintf(stdout, "\t%s (Fill_Value = %d)\n", sds1_info[isds].name, f_val);
//	    else
//	      fprintf(stdout, "\t%s(c) (Fill_Value = %d)\n", sds1_info[isds].name, f_val);
          }
	} /* for (isds=0; . . ) */

	sds_nadd_obs_1km_info.sds_id = sds_nadd_obs_500m_info.sds_id = -1;
	sds_nadd_obs_1km_info.sd_id = sds_nadd_obs_500m_info.sd_id = sds1_info[0].sd_id;
	sds_nadd_obs_1km_info.H5IN = sds_nadd_obs_500m_info.H5IN = sds1_info[0].H5IN;
	strcpy(sds_nadd_obs_1km_info.name, sds_name_nadd_obs_1km);      
	if (get_sds_info(fname, &sds_nadd_obs_1km_info) == -1) 
	  fprintf(stderr, "Result may be in error for %s \n", sds_nadd_obs_1km_info.name);	  
	if ((data_nadd_obs_1km = (int32 *)calloc(sds_nadd_obs_1km_info.dim_size[0], sizeof(int32))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for data_nadd_obs_1km in read_l2g_obs_at_pts\n");      
	start[0] = 0; edge[0] = sds_nadd_obs_1km_info.dim_size[0];
	start[1] = edge[1] = 0;
	if (read_sel_sds_data(&sds_nadd_obs_1km_info, start, edge, data_nadd_obs_1km) == FAIL)
          fprintf(stderr, "Cannot read data line for sds %s\n", sds_nadd_obs_1km_info.name);	
	
	strcpy(sds_nadd_obs_500m_info.name, sds_name_nadd_obs_500m);   
	if (get_sds_info(fname, &sds_nadd_obs_500m_info) == -1) 
	  fprintf(stderr, "Result may be in error for %s \n", sds_nadd_obs_500m_info.name);	
	if ((data_nadd_obs_500m = (int32 *)calloc(sds_nadd_obs_500m_info.dim_size[0], sizeof(int32))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for data_nadd_obs_500m in read_l2g_obs_at_pts\n");      
	start[0] = 0; edge[0] = sds_nadd_obs_500m_info.dim_size[0];
	start[1] = edge[1] = 0;
	if (read_sel_sds_data(&sds_nadd_obs_500m_info, start, edge, data_nadd_obs_500m) == FAIL)
          fprintf(stderr, "Cannot read data line for sds %s\n", sds_nadd_obs_500m_info.name);			
	
	if (res == 0)
	{
	  min_rows = cres = sds1_info[0].dim_size[0];
	  min_cols = sds1_info[0].dim_size[1];
          for (isds=1; isds<nsds; isds++)
	    if (cres > sds1_info[isds].dim_size[0]) 
	    {
	      min_rows = cres = sds1_info[isds].dim_size[0];
	      min_cols = sds1_info[isds].dim_size[1];
	    }	    
	}
	else
	{
	  if (res == 250) cres = 4800;
	  else if (res == 500) cres = 2400;
	  else if (res == 1000) cres = 1200;
	  else fprintf(stderr, "Wrong resolution %d encountered\n", res);
	}

	for (isds=0; isds<nsds; isds++)
	{
	  rf[isds] = (float)sds1_info[isds].dim_size[0]/(float)cres;
//          fprintf(stdout, "rf[%d]=%f\n", isds, rf[isds]);
	}
	if ((sds_val = (int32 **)Calloc2D(nsds, MAX_NUM_OBS, sizeof(int32))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for sds_val in read_l2g_obs_at_pts\n");
	if ((out_pnts = (int **)Calloc2D(nsds, 2, sizeof(int))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for out_pnts in read_l2l3_obs_at_pts\n");
	if ((sds_val != NULL) && (out_pnts != NULL))
	{
          ndata = sds1_info[0].dim_size[1];
          if ((data_nobs = (int8 *)calloc(ndata, sizeof(int8))) == NULL)
            fprintf(stderr, "Cannot allocate memory for data_nobs in read_l2g_obs_at_pts\n");
	  max_size = sds1_info[0].data_size;
	  for (isds=2; isds<nsds; isds++)
	    if (sds1_info[isds].data_size > max_size)
	      max_size = sds1_info[isds].data_size;
	  if ((data_in = (void *)calloc(ndata, max_size)) == NULL)
	    fprintf(stderr, "Cannot allocate memory for data_in in read_l2g_obs_at_pts\n");

          ndata_500m = sds1_info[1].dim_size[1];
          if ((data_nobs_500m = (int8 *)calloc(ndata_500m, sizeof(int8))) == NULL)
            fprintf(stderr, "Cannot allocate memory for data_nobs_500m in read_l2g_obs_at_pts\n");
	  max_size = sds1_info[1].data_size;
	  for (isds=2; isds<nsds; isds++)
	    if (sds1_info[isds].data_size > max_size)
	      max_size = sds1_info[isds].data_size;
	  if ((data_in_500m = (void *)calloc(ndata_500m, max_size)) == NULL)
	    fprintf(stderr, "Cannot allocate memory for data_in_500m in read_l2g_obs_at_pts\n");


	  if ((data_nobs != NULL) && (data_in != NULL) && (data_nobs_500m != NULL) && (data_in_500m != NULL))
	  {
            if (npt == 0)
            {
              if ((fp = fopen(xy_fname, "r")) == NULL)
        	fprintf(stderr, "Cannot open file %s\n", xy_fname);
            }
            ipt = 0;
            while (done)
            {
              if (npt > 0)
              {
        	id = ipt*2;
        	get_xy_pts(xy_str[id], &pt_x, &x_sh, &x_sq);
        	get_xy_pts(xy_str[id+1], &pt_y, &y_sh, &y_sq);
        	++ipt;
        	if (ipt > npt) break;
              }
              else
              {
        	if (fscanf(fp, "%s %s", x_str, y_str) == 2)
        	{
                  get_xy_pts(x_str, &pt_x, &x_sh, &x_sq);
                  get_xy_pts(y_str, &pt_y, &y_sh, &y_sq);
        	}
        	if (feof(fp) != 0) { fclose(fp); break; }
              }
	      
	      if (res == 1000)
	      {
		irow_1km = (int)((float)pt_y); 
		icol_1km = (int)((float)pt_x);		  
		irow_500m = (int)((float)pt_y*2); 
		icol_500m = (int)((float)pt_x*2);	  
	      }   
	      else if (res == 500)
	      {
		irow_1km = (int)((float)pt_y/2.); 
		icol_1km = (int)((float)pt_x/2.);		  
		irow_500m = (int)((float)pt_y); 
		icol_500m = (int)((float)pt_x);	      
	      }
	      else if (res == 250)
	      {
		irow_1km = (int)((float)pt_y/4.); 
		icol_1km = (int)((float)pt_x/4.);		  
		irow_500m = (int)((float)pt_y/2.); 
		icol_500m = (int)((float)pt_x/2.);	      
	      }	       	      
	      for (isds=2; isds<nsds; isds++)
	      { 
		if ((sds1_info[isds].sds_id != -1) && (sds1_info[0].sds_id != -1) && (sds1_info[1].sds_id != -1))			
		{	  	  
	          if (isds == 2)
	          {		    		    
	            irow = (int)((float)pt_y*rf[0]);
	            icol = (int)((float)pt_x*rf[0]);		  
		    if (sds1_info[0].dim_size[0] == 2400) { irow += y_sh; icol += x_sh; }
	            else if (sds1_info[0].dim_size[0] == 4800) { irow += y_sq; icol += x_sq; }		  
                    start[0] = irow; edge[0] = 1;
	            start[1] = 0; 		    		    
		    edge[1] = ndata;
		    if (read_sel_sds_data(&sds1_info[0], start, edge, data_nobs) == FAIL)
                      fprintf(stderr, "Cannot read data line for sds %s\n", sds1_info[0].name);
	            nobs = ((int8 *)data_nobs)[icol];
		    sds_val[0][0] = nobs;
		    
	            irow = (int)((float)pt_y*rf[1]);
	            icol = (int)((float)pt_x*rf[1]);		  
		    if (sds1_info[1].dim_size[0] == 2400) { irow += y_sh; icol += x_sh; }
	            else if (sds1_info[1].dim_size[0] == 4800) { irow += y_sq; icol += x_sq; }		  
                    start[0] = irow; edge[0] = 1;
	            start[1] = 0; 		  		    
		    edge[1] = ndata_500m;	
		    if (read_sel_sds_data(&sds1_info[1], start, edge, data_nobs_500m) == FAIL)
                      fprintf(stderr, "Cannot read data line for sds %s\n", sds1_info[1].name);
	            nobs_500m = ((int8 *)data_nobs_500m)[icol];
		    sds_val[1][0] = nobs_500m;	            		   
		    out_pnts[0][0] = icol;
		    out_pnts[0][1] = irow;
	          }
		  
	          irow = (int)((float)pt_y*rf[isds]);
	          icol = (int)((float)pt_x*rf[isds]);		  
		  if (sds1_info[isds].dim_size[0] == 2400) { irow += y_sh; icol += x_sh; }
	          else if (sds1_info[isds].dim_size[0] == 4800) { irow += y_sq; icol += x_sq; }		  
                  start[0] = irow; edge[0] = 1;
	          start[1] = 0; 
	          out_pnts[isds][0] = icol;
	          out_pnts[isds][1] = irow;		  

		  if (sds1_info[isds].dim_size[0] == 1200)
		  {
	            if (nobs > 0)
	            {
                      edge[1] = ndata;
		      if (read_sel_sds_data(&sds1_info[isds], start, edge, data_in) == FAIL)
                	fprintf(stderr, "Cannot read data line for sds %s\n", sds1_info[isds].name);
                      switch(sds1_info[isds].data_type)
                      {
                	case 20: sds_val[isds][0] = ((int8 *)data_in)[icol]; break;
                	case 21: sds_val[isds][0] = ((uint8 *)data_in)[icol]; break;
                	case 22: sds_val[isds][0] = ((int16 *)data_in)[icol]; break;
                	case 23: sds_val[isds][0] = ((uint16 *)data_in)[icol]; break;
                	case 24: sds_val[isds][0] = ((int32 *)data_in)[icol]; break;
                	case 25: sds_val[isds][0] = ((uint32 *)data_in)[icol]; break;
                      }
	            }
	            if (nobs > 1)
	            {
	              for (ir=0, offset=0; ir<irow; ir++)
			offset += ((int32 *)data_nadd_obs_1km)[ir];
		      for (ic=0; ic<icol; ic++)
		      {
			nobs_ic = ((int8 *)data_nobs)[ic];
			if (nobs_ic > 1) offset = offset + nobs_ic - 1;
		      }
		      start[0] = offset; edge[0] = nobs - 1;
		      start[1] = edge[1] = 0;
		      if (read_sel_sds_data(&sdsc_info[isds], start, edge, data_in) == FAIL)
                	fprintf(stderr, "Cannot read data line for sds %s\n", sdsc_info[isds].name);
		      for (iobs=1, offset=0; iobs<nobs; iobs++, offset++)
                	switch(sdsc_info[isds].data_type)
                	{
                	  case 20: sds_val[isds][iobs] = ((int8 *)data_in)[offset]; break;
                	  case 21: sds_val[isds][iobs] = ((uint8 *)data_in)[offset]; break;
                	  case 22: sds_val[isds][iobs] = ((int16 *)data_in)[offset]; break;
                	  case 23: sds_val[isds][iobs] = ((uint16 *)data_in)[offset]; break;
                	  case 24: sds_val[isds][iobs] = ((int32 *)data_in)[offset]; break;
                	  case 25: sds_val[isds][iobs] = ((uint32 *)data_in)[offset]; break;
                	}
	            } /* if (nobs > 1) */
		  }
		  else
		  {
	            if (nobs_500m > 0)
	            {
                      edge[1] = ndata_500m;
		      if (read_sel_sds_data(&sds1_info[isds], start, edge, data_in_500m) == FAIL)
                	fprintf(stderr, "Cannot read data line for sds %s\n", sds1_info[isds].name);
                      switch(sds1_info[isds].data_type)
                      {
                	case 20: sds_val[isds][0] = ((int8 *)data_in_500m)[icol]; break;
                	case 21: sds_val[isds][0] = ((uint8 *)data_in_500m)[icol]; break;
                	case 22: sds_val[isds][0] = ((int16 *)data_in_500m)[icol]; break;
                	case 23: sds_val[isds][0] = ((uint16 *)data_in_500m)[icol]; break;
                	case 24: sds_val[isds][0] = ((int32 *)data_in_500m)[icol]; break;
                	case 25: sds_val[isds][0] = ((uint32 *)data_in_500m)[icol]; break;
                      }
	            }
	            if (nobs_500m > 1)
	            {
	              for (ir=0, offset=0; ir<irow; ir++)
			offset += ((int32 *)data_nadd_obs_500m)[ir];
		      for (ic=0; ic<icol; ic++)
		      {
			nobs_ic = ((int8 *)data_nobs_500m)[ic];
			if (nobs_ic > 1) offset = offset + nobs_ic - 1;
		      }
		      start[0] = offset; edge[0] = nobs_500m - 1;
		      start[1] = edge[1] = 0;
		      if (read_sel_sds_data(&sdsc_info[isds], start, edge, data_in_500m) == FAIL)
                	fprintf(stderr, "Cannot read data line for sds %s\n", sdsc_info[isds].name);
		      for (iobs=1, offset=0; iobs<nobs_500m; iobs++, offset++)
                	switch(sdsc_info[isds].data_type)
                	{
                	  case 20: sds_val[isds][iobs] = ((int8 *)data_in_500m)[offset]; break;
                	  case 21: sds_val[isds][iobs] = ((uint8 *)data_in_500m)[offset]; break;
                	  case 22: sds_val[isds][iobs] = ((int16 *)data_in_500m)[offset]; break;
                	  case 23: sds_val[isds][iobs] = ((uint16 *)data_in_500m)[offset]; break;
                	  case 24: sds_val[isds][iobs] = ((int32 *)data_in_500m)[offset]; break;
                	  case 25: sds_val[isds][iobs] = ((uint32 *)data_in_500m)[offset]; break;
                	}
	            } /* if (nobs > 1) */		  		  
		  }
         	} /* if (sds1_info . . . ) */	      	      
	      } /* for (isds=0; . . . ) */
	      if (res == 0)
	      {
  		fprintf(stdout, "Observations: pixel (%d %d) at ref resolution (%d x %d)\n", 
			  icol, irow, min_cols, min_rows);			  	      
	      }
	      else
	      {
		fprintf(stdout, "Observations: pixel (%d %d) at ref resolution %dm for 1km SDS\n", icol_1km, irow_1km, res); 
		fprintf(stdout, "Observations: pixel (%d %d) at ref resolution %dm for 500m SDS\n", icol_500m, irow_500m, res);		
	      }
	      if ((strstr(fname, "MOD09GA") == NULL) && (strstr(fname, "MYD09GA") == NULL) &&
	          (strstr(fname, "NPP_DSRFLD_L2GD") == NULL))
                print_l2g_sds_val(out_pnts, sds_val, sds1_info, sdsc_info, nsds, 1);
	      else
	        printL2gSdsVal(out_pnts, sds_val, sds1_info, sdsc_info, nsds, rf);		      
	    } /* for (ipt=0; . . . ) */
	    free(data_in);
	    free(data_nobs);
	    free(data_in_500m);
	    free(data_nobs_500m);	    
	  } /* if ((data_nobs != . . .  )) */
	  Free2D((void **)out_pnts);
	  Free2D((void **)sds_val);
	} /* if (sds_val != . . . ) */
	for (isds=0; isds<nsds; isds++)
	{
	  close_sds(&sds1_info[isds], sds1_info[0].H5IN);
	  close_sds(&sdsc_info[isds], sdsc_info[0].H5IN);
	} 
	close_sd(&sds1_info[0], sds1_info[0].H5IN);	
	free(rf);
	free(sds1_info); 
	free(sdsc_info); 
	free(data_nadd_obs);	
	free(data_nadd_obs_1km);	
	free(data_nadd_obs_500m);
      } /* if ((sds1_info != . . . )) */    
    }    
  } /* if ((nsds =  . . ) */
  Free2D((void **)sds_names);
}

void printSdsVal(int **out_pnts, void **sdsVal, sds_t *sds_info, int nsds, int bsq, int l2g_st)
{
  int isds;
  int jobs, kobs, iobs;

  if (l2g_st == 1)
  {
    kobs = ((int *)sdsVal[0])[0];
    fprintf(stdout, "\t%s: %d\n", sds_info[0].name, kobs);
    for (isds=1; isds<nsds; isds++)
    {
      fprintf(stdout, "\t%s(c) at (%d %d): ", sds_info[isds].name, out_pnts[isds][0], out_pnts[isds][1]);
      for (iobs=0; iobs<kobs; iobs++)
      {	       					
	switch(sds_info[isds].data_type)
        {
          case 5: fprintf(stdout, "%f ", ((float32 *)sdsVal[isds])[iobs]); break;
          case 6: fprintf(stdout, "%f ", ((float64 *)sdsVal[isds])[iobs]); break;	  
	  case 20: fprintf(stdout, "%d ", ((int8 *)sdsVal[isds])[iobs]); break;
          case 21: fprintf(stdout, "%d ", ((uint8 *)sdsVal[isds])[iobs]); break;
          case 22: fprintf(stdout, "%d ", ((int16 *)sdsVal[isds])[iobs]); break;
          case 23: fprintf(stdout, "%d ", ((uint16 *)sdsVal[isds])[iobs]); break;
          case 24: fprintf(stdout, "%d ", ((int32 *)sdsVal[isds])[iobs]); break;
          case 25: fprintf(stdout, "%d ", ((uint32 *)sdsVal[isds])[iobs]); break;
        }	  	  
      }	  
      fprintf(stdout, "\n");
    }
  }
  else for (isds=0; isds<nsds; isds++)
  {
    if (sds_info[isds].rank == 1 && strstr(sds_info[isds].name, "Factors") != NULL)
    {
/*      fprintf(stdout, "%s: \n", sds_info[isds].name);    
      for (iobs=0; iobs<sds_info[isds].dim_size[0]; iobs++)
      {	       					
	switch(sds_info[isds].data_type)
        {
          case 5: fprintf(stdout, "%f ", ((float32 *)sdsVal[isds])[iobs]); break;
          case 6: fprintf(stdout, "%f ", ((float64 *)sdsVal[isds])[iobs]); break;	  
	  case 20: fprintf(stdout, "%d ", ((int8 *)sdsVal[isds])[iobs]); break;
          case 21: fprintf(stdout, "%d ", ((uint8 *)sdsVal[isds])[iobs]); break;
          case 22: fprintf(stdout, "%d ", ((int16 *)sdsVal[isds])[iobs]); break;
          case 23: fprintf(stdout, "%d ", ((uint16 *)sdsVal[isds])[iobs]); break;
          case 24: fprintf(stdout, "%d ", ((int32 *)sdsVal[isds])[iobs]); break;
          case 25: fprintf(stdout, "%d ", ((uint32 *)sdsVal[isds])[iobs]); break;
        }	  	  
      }	        
      fprintf(stdout, "\n");   */
    } 
    else if (sds_info[isds].rank == 2)    
    {
      fprintf(stdout, "%s at (%d %d): ", sds_info[isds].name, out_pnts[isds][0], out_pnts[isds][1]);      
      switch(sds_info[isds].data_type)
      {
        case 5: fprintf(stdout, "%f ", ((float32 *)sdsVal[isds])[0]); break;
        case 6: fprintf(stdout, "%f ", ((float64 *)sdsVal[isds])[0]); break;	
	case 20: fprintf(stdout, "%d ", ((int8 *)sdsVal[isds])[0]); break;
        case 21: fprintf(stdout, "%d ", ((uint8 *)sdsVal[isds])[0]); break;
        case 22: fprintf(stdout, "%d ", ((int16 *)sdsVal[isds])[0]); break;
        case 23: fprintf(stdout, "%d ", ((uint16 *)sdsVal[isds])[0]); break;
        case 24: fprintf(stdout, "%d ", ((int32 *)sdsVal[isds])[0]); break;
        case 25: fprintf(stdout, "%d ", ((uint32 *)sdsVal[isds])[0]); break;
      } 
      fprintf(stdout, "\n");   
    }  
    else if (sds_info[isds].rank == 3)
    {
      fprintf(stdout, "%s at (%d %d): ", sds_info[isds].name, out_pnts[isds][0], out_pnts[isds][1]);        
      if (bsq == 0)
	for (iobs=0; iobs<sds_info[isds].dim_size[2]; iobs++)
	{
	  switch(sds_info[isds].data_type)
          {
            case 5: fprintf(stdout, "%f ", ((float32 *)sdsVal[isds])[iobs]); break;
            case 6: fprintf(stdout, "%f ", ((float64 *)sdsVal[isds])[iobs]); break;		  
	    case 20: fprintf(stdout, "%d ", ((int8 *)sdsVal[isds])[iobs]); break;
            case 21: fprintf(stdout, "%d ", ((uint8 *)sdsVal[isds])[iobs]); break;
            case 22: fprintf(stdout, "%d ", ((int16 *)sdsVal[isds])[iobs]); break;
            case 23: fprintf(stdout, "%d ", ((uint16 *)sdsVal[isds])[iobs]); break;
            case 24: fprintf(stdout, "%d ", ((int32 *)sdsVal[isds])[iobs]); break;
            case 25: fprintf(stdout, "%d ", ((uint32 *)sdsVal[isds])[iobs]); break;
          }	  	  
	}
      else
	for (iobs=0; iobs<sds_info[isds].dim_size[0]; iobs++)
	{
	  switch(sds_info[isds].data_type)
          {
            case 5: fprintf(stdout, "%f ", ((float32 *)sdsVal[isds])[iobs]); break;
            case 6: fprintf(stdout, "%f ", ((float64 *)sdsVal[isds])[iobs]); break;		  
	    case 20: fprintf(stdout, "%d ", ((int8 *)sdsVal[isds])[iobs]); break;
            case 21: fprintf(stdout, "%d ", ((uint8 *)sdsVal[isds])[iobs]); break;
            case 22: fprintf(stdout, "%d ", ((int16 *)sdsVal[isds])[iobs]); break;
            case 23: fprintf(stdout, "%d ", ((uint16 *)sdsVal[isds])[iobs]); break;
            case 24: fprintf(stdout, "%d ", ((int32 *)sdsVal[isds])[iobs]); break;
            case 25: fprintf(stdout, "%d ", ((uint32 *)sdsVal[isds])[iobs]); break;
          }	  	  
	}              	  		
      fprintf(stdout, "\n");
    }
    else if (sds_info[isds].rank == 4)
    {      
      fprintf(stdout, "%s at (%d %d): ", sds_info[isds].name, out_pnts[isds][0], out_pnts[isds][1]);  
      fprintf(stdout, "\n");
      if (bsq == 0)  
	for (jobs=0, iobs=0; jobs<sds_info[isds].dim_size[2]; jobs++)
	{
	  fprintf(stdout, "\t\t\t");
	  for (kobs=0; kobs<sds_info[isds].dim_size[3]; kobs++, iobs++)
	  {
	    switch(sds_info[isds].data_type)
            {
              case 5: fprintf(stdout, "%f ", ((float32 *)sdsVal[isds])[iobs]); break;
              case 6: fprintf(stdout, "%f ", ((float64 *)sdsVal[isds])[iobs]); break;		    
	      case 20: fprintf(stdout, "%d ", ((int8 *)sdsVal[isds])[iobs]); break;
              case 21: fprintf(stdout, "%d ", ((uint8 *)sdsVal[isds])[iobs]); break;
              case 22: fprintf(stdout, "%d ", ((int16 *)sdsVal[isds])[iobs]); break;
              case 23: fprintf(stdout, "%d ", ((uint16 *)sdsVal[isds])[iobs]); break;
              case 24: fprintf(stdout, "%d ", ((int32 *)sdsVal[isds])[iobs]); break;
              case 25: fprintf(stdout, "%d ", ((uint32 *)sdsVal[isds])[iobs]); break;
            }	  
	  } 	 
	  fprintf(stdout, "\n");
	}
      else
	for (jobs=0, iobs=0; jobs<sds_info[isds].dim_size[0]; jobs++)
	{
	  fprintf(stdout, "\t\t\t");
	  for (kobs=0; kobs<sds_info[isds].dim_size[1]; kobs++, iobs++)
	  {
	    switch(sds_info[isds].data_type)
            {
              case 5: fprintf(stdout, "%f ", ((float32 *)sdsVal[isds])[iobs]); break;
              case 6: fprintf(stdout, "%f ", ((float64 *)sdsVal[isds])[iobs]); break;		    
	      case 20: fprintf(stdout, "%d ", ((int8 *)sdsVal[isds])[iobs]); break;
              case 21: fprintf(stdout, "%d ", ((uint8 *)sdsVal[isds])[iobs]); break;
              case 22: fprintf(stdout, "%d ", ((int16 *)sdsVal[isds])[iobs]); break;
              case 23: fprintf(stdout, "%d ", ((uint16 *)sdsVal[isds])[iobs]); break;
              case 24: fprintf(stdout, "%d ", ((int32 *)sdsVal[isds])[iobs]); break;
              case 25: fprintf(stdout, "%d ", ((uint32 *)sdsVal[isds])[iobs]); break;
            }	  
	  } 	 
	  fprintf(stdout, "\n");
	}        
    } /* else */
  } /* for (isds=0; . . . ) */
}

void print_l2g_sds_val(int **out_pnts, int32 **sds_val, sds_t *sds_info, sds_t *sdsc_info, int nsds, int l2g_st)
{
  int isds;
  int jobs, kobs, iobs;

  if (l2g_st == 1)
  {
    kobs = sds_val[0][0];
    fprintf(stdout, "\t%s: %d\n", sds_info[0].name, kobs);  
    if (kobs >0)
    {
      for (isds=1; isds<nsds; isds++)
      {
	for (iobs=0; iobs<kobs; iobs++)
	{
          if (iobs == 0)
	  {
	    fprintf(stdout, "\t%s(1) at (%d %d): ", sds_info[isds].name, out_pnts[isds][0], out_pnts[isds][1]);
	    fprintf(stdout, "%d ", sds_val[isds][iobs]);
	  }
	  else if (sdsc_info[isds].sds_id != -1)
	  {
	    fprintf(stdout, "\t%s(c) at (%d %d): ", sdsc_info[isds].name, out_pnts[isds][0], out_pnts[isds][1]);
	    fprintf(stdout, "%d ", sds_val[isds][iobs]);
	  }  
	}
	fprintf(stdout, "\n");
      }
    }
    else
      fprintf(stdout, "No observation found.\n");
  }
  else for (isds=0; isds<nsds; isds++)
  {
    fprintf(stdout, "\t%s at (%d %d): ", sds_info[isds].name, out_pnts[isds][0], out_pnts[isds][1]);
    if (sds_info[isds].rank == 2)
      fprintf(stdout, "%d\n", sds_val[isds][0]);
    else if (sds_info[isds].rank == 3)
    {
      for (iobs=0; iobs<sds_info[isds].dim_size[2]; iobs++)
	fprintf(stdout, "%d ", sds_val[isds][iobs]);
      fprintf(stdout, "\n");
    }
    else if (sds_info[isds].rank == 4)
    {
      fprintf(stdout, "\n");
      for (jobs=0, iobs=0; jobs<sds_info[isds].dim_size[2]; jobs++)
      {
	fprintf(stdout, "\t\t\t");
	for (kobs=0; kobs<sds_info[isds].dim_size[3]; kobs++, iobs++)
	  fprintf(stdout, "%d ", sds_val[isds][iobs]);
	fprintf(stdout, "\n");
      }
    } /* else */
  } /* for (isds=0; . . . ) */
}

void printL2gSdsVal(int **out_pnts, int32 **sds_val, sds_t *sds_info, sds_t *sdsc_info, int nsds, float *rf)
{
  int isds;
  int jobs, kobs, kobs_500m, iobs;

  kobs = sds_val[0][0];
  kobs_500m = sds_val[1][0];  
  fprintf(stdout, "\t%s: %d\n", sds_info[0].name, kobs); 
  fprintf(stdout, "\t%s: %d\n", sds_info[1].name, kobs_500m);    
  for (isds=2; isds<nsds; isds++)
  {
    if (sds_info[isds].dim_size[0] == 1200)
    {
      if (kobs >= 0)
      {
	for (iobs=0; iobs<kobs; iobs++)
	{
	  if (iobs == 0)
	  {
	    fprintf(stdout, "\t%s(c) at (%d %d): ", sds_info[isds].name, out_pnts[isds][0], out_pnts[isds][1]);
	    fprintf(stdout, "%d ", sds_val[isds][iobs]);
	  }
	  else if (sdsc_info[isds].sds_id != -1)
	  {
	    fprintf(stdout, "\t%s(c) at (%d %d): ", sdsc_info[isds].name, out_pnts[isds][0], out_pnts[isds][1]);
	    fprintf(stdout, "%d ", sds_val[isds][iobs]);
	  }  
	}
      }
      else
        fprintf(stdout, "\tNo observation found for %s.\n", sds_info[0].name); 
    }
    else
    {
      if (kobs_500m >= 0)
      {
	for (iobs=0; iobs<kobs_500m; iobs++)
	{
	  if (iobs == 0)
	  {
	    fprintf(stdout, "\t%s(c) at (%d %d): ", sds_info[isds].name, out_pnts[isds][0], out_pnts[isds][1]);
	    fprintf(stdout, "%d ", sds_val[isds][iobs]);
	  }
	  else if (sdsc_info[isds].sds_id != -1)
	  {
	    fprintf(stdout, "\t%s(c) at (%d %d): ", sdsc_info[isds].name, out_pnts[isds][0], out_pnts[isds][1]);
	    fprintf(stdout, "%d ", sds_val[isds][iobs]);
	  }  
	}
      }  
      else
        fprintf(stdout, "\tNo observation found for %s.\n", sds_info[1].name);       
    }
    fprintf(stdout, "\n");
  }
}


