/*
  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/dir.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "qa_tool.h"
#include "sds_rw.h"
#include "str_op.h"
#include "meta.h"
#include "alloc_mem.h"
#include "main_util.h"     


#define HELP \
"NAME \n" \
"       rename_sds - Rename SDS(s) in an HDF files (creates new HDF file) \n" \
" \n" \
"SYNOPSIS \n" \
"       rename_sds [-help] \n" \
"       rename_sds [-sds=<old name>,<new name>] -ncp [-meta] -of=<output filename> filename\n" \
" \n" \
"DESCRIPTION \n" \
"       Rename one or more SDS in an HDF files. A new HDF file with user specified SDS(s) is \n"  \
"       created. SDSs not renamed by the user are copied to the output file with the original \n" \
"       SDS names. User may choose not to copy remaining SDSs. \n\n" \
"\n" \
" \n" \
"OPTIONS \n" \
"       -help                       Display this help message \n" \
"       -sds=<old name,new name>    Input and Output SDS names \n" \
"                                   To specify additional repeat the option as many times as \n" \
"                                   the number of SDSs to be renamed \n" \
"       -of=<filename>              Output filename \n" \
"       -ncp                        Do not copy remaining SDS(s) \n" \
"       -meta                       Copy all metadata from input to output \n" \
" \n" \
"Examples: \n" \
"\n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:\trename_sds [-sds=<old name>,<new name>] -ncp [-meta] -of=<output filename> filename\n" \
" \n" \
"OPTIONS \n" \
"       -help                       Display this help message \n" \
"       -sds=<old name,new name>    Input and Output SDS names \n" \
"                                   To specify additional repeat the option as many times as \n" \
"                                   the number of SDSs to be renamed \n" \
"       -of=<filename>              Output filename \n" \
"       -meta                       Copy all metadata from input to output \n" \
"       -ncp                        Do not copy remaining SDS(s) \n" \
" \n" \
" \n"                                                      

int parse_cmd_rename_sds(int argc, char **argv, char *out_fname, char *in_fname, int *nsds, 
	char **in_sds_names, char **out_sds_names, int *H5OUT, int *m_opt);
void copy_an_sds(sds_t *in_sds_info, sds_t *out_sds_info);

int main(int argc, char **argv)
{
    int nsds, isds, m_opt;
    char in_fname[MAX_PATH_LENGTH];
    char out_fname[MAX_PATH_LENGTH];
    sds_t in_sds_info, out_sds_info;
    char **in_sds_names, **out_sds_names;
    int H5OUT;
  
    if ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
      fprintf(stderr, "%s\n", HELP);
    else
    { 
        if ((in_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
            fprintf(stderr, "Cannot allocate memory for in_sds_names in rename_sds main()\n");
        if ((out_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
            fprintf(stderr, "Cannot allocate memory for out_sds_names in rename_sds main()\n");
        if ((in_sds_names != NULL) && (out_sds_names != NULL))
        {
            if (parse_cmd_rename_sds(argc, argv, out_fname, in_fname, &nsds, in_sds_names, out_sds_names, &H5OUT, &m_opt) == -1)
              fprintf(stderr, "%s\n", USAGE);
            else
            {
		/* MSZ 1/31/2018 silent version mismatch warning standard print out */
		setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

		/* MSZ 1/31/2018 Disable error reporting */
		H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

		in_sds_info.sd_id = in_sds_info.sds_id = -1;

		fprintf(stdout, "Processing file %s\n", in_fname);

		if(open_sd(in_fname, &in_sds_info, 'R') == FAIL)
		{
		  fprintf(stderr, "cannot open %s\n", in_fname);
		  exit(0);
		}
	
		out_sds_info.H5OUT = H5OUT; /* get from parse_cmd_rename_sds */
		out_sds_info.sd_id = out_sds_info.sds_id = -1;

		if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
		{
		  fprintf(stderr, "cannot create %s\n", out_fname);
		  exit(0);
		}

                if ((in_sds_info.sd_id != FAIL) && (out_sds_info.sd_id != FAIL))
                {
                    printf("nsds = %d\n", nsds);
                    for (isds=0; isds<nsds; isds++)
                    {
                      strcpy(in_sds_info.name, in_sds_names[isds]);
                      strcpy(out_sds_info.name, out_sds_names[isds]);
                      fprintf(stdout, "\tCopying and renaming SDS %s as %s\n", in_sds_names[isds], out_sds_names[isds]);
                      copy_an_sds(&in_sds_info, &out_sds_info);
                    }
                }
                if (m_opt == 1)
		  copy_metadata(in_sds_info.H5IN, in_sds_info.sd_id, out_sds_info.H5OUT, out_sds_info.sd_id);

		close_sd(&in_sds_info, in_sds_info.H5IN);
		close_sd(&out_sds_info, out_sds_info.H5OUT);
            }
        }
        if (in_sds_names != NULL)
          Free2D((void **)in_sds_names);
        if (out_sds_names != NULL)
          Free2D((void **)out_sds_names);
    }
    return 0;
}

int parse_cmd_rename_sds(int argc, char **argv, char *out_fname, char *in_fname, int *nsds, 
	char **in_sds_names, char **out_sds_names, int *H5OUT, int *m_opt)
{
  int inc_sds;
  int st, i, p1;
  int ncp, sds_cnt, len;
  int isds, jsds, ksds;
  char **sds_names, tmp_str[MAX_STR_LEN];
  /* Sudipta-09/16: hdf5 addition */
  char* ptr1 = NULL;
  
  *nsds = 0;
  isds = ncp = 0;
  in_fname[0] = out_fname[0] = '\0';
  for (i=1, st=1; i<argc; i++)
  {
    if ((is_arg_id(argv[i], "-of=") == 0) || (is_arg_id(argv[i], "-o=") == 0))
    {
      get_arg_val(argv[i], out_fname);
      if (((((ptr1 = strstr(out_fname,".h5")) != NULL)) ||
           (((ptr1 = strstr(out_fname,".he5")) != NULL)) ||
           (((ptr1 = strstr(out_fname,".hdf5")) != NULL)) ||
           (((ptr1 = strstr(out_fname,".nc")) != NULL))))
            *H5OUT = 1;
        else
            *H5OUT = 0;
    }
    else if (is_arg_id(argv[i], "-sds=") == 0)
    {
      get_arg_val(argv[i], tmp_str);
      if ((p1 = charpos(tmp_str, ',', 0)) == -1)
        fprintf(stderr, "Ignoring invalid option assignment %s\n", argv[i]);
      else
      {
	len = (int)strlen(tmp_str);
        strmid(tmp_str, 0, p1, in_sds_names[isds]);
	p1++;
        strmid(tmp_str, p1, len-p1, out_sds_names[isds]);
	isds++;
      }
    }
    else if (strcmp(argv[i], "-ncp") == 0) ncp = 1; 
    else if (strcmp(argv[i], "-meta") == 0) *m_opt = 1; 
    else if (argv[i][0] == '-')
      fprintf(stderr, "Unknown option %s\n", argv[i]);
    else
      strcpy(in_fname, argv[i]);
  }
  if (out_fname[0] == '\0') {
    st = -1;
    fprintf(stderr, "Missing output filename . .\n");
  }
  if (in_fname[0] == '\0') {
    st = -1;
    fprintf(stderr, "Missing input filename . .\n");
  }
  if (st == 1)
  {
    if (ncp == 0)
    {
      if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
        fprintf(stderr, "Cannot allocate memory for sds_names in parse_cmd_rename_sds\n");
      else
      {
        sds_cnt = get_sds_names(in_fname, sds_names);
        for (jsds=0; jsds<sds_cnt; jsds++)
	{
	  inc_sds = 1;
	  for (ksds=0; ksds<isds; ksds++)
	    if (strcasecmp(sds_names[jsds], in_sds_names[ksds]) == 0)
	    {
	      inc_sds = 0;
	      break;
	    } 
	  if (inc_sds == 1)
	  {
	    strcpy(in_sds_names[isds], sds_names[jsds]);
	    strcpy(out_sds_names[isds], sds_names[jsds]);
	    isds++;
	  }
        }
        Free2D((void **)sds_names);
      }
    }
    *nsds = isds;
  }
  return st;
}                                                   
	    
void copy_an_sds(sds_t *in_sds_info, sds_t *out_sds_info)
{
    void *sds_data;
    int i, rank;
    int32 start[4] = {0}, edge[4] = {0};
    int irow, nrows, ndata;
  
    out_sds_info->sds_id = in_sds_info->sds_id = -1;
    if (get_sds_info((char *)NULL, in_sds_info) != -1)
    {
        rank = in_sds_info->rank;
        out_sds_info->rank = rank;
        for (i=0; i<in_sds_info->rank; i++)
            out_sds_info->dim_size[i] = in_sds_info->dim_size[i];
        out_sds_info->data_type = in_sds_info->data_type;
        out_sds_info->data_size = in_sds_info->data_size;
        if (open_sds((char *)NULL, out_sds_info, 'W') != -1)
        {
            for (i=0; i<rank; i++)
            {
                start[i] = 0;
                edge[i] = in_sds_info->dim_size[i];
            }

            if (rank != 1) 
              edge[in_sds_info->row_id] = 1;
            nrows = in_sds_info->dim_size[in_sds_info->row_id];
            for (irow=0; irow<nrows; irow++)
            {
                start[in_sds_info->row_id] = irow;
                if ((sds_data = get_sel_sds_data(in_sds_info, start, edge)) != NULL)
                {
                    if (set_sel_sds_data(out_sds_info, start, edge, sds_data) == FAIL)
                    {
                        fprintf(stderr, "Error writing line %d to SDS %s\n", irow, out_sds_info->name);
                        exit(0);
                    }
                        
                    if(sds_data) free(sds_data);
                }
                else
                    fprintf(stderr, "Error reading line %d from SDS %s\n", irow, in_sds_info->name);
                if (rank == 1) /* for 1D data there is no scope for looping over rest of rows */
                    break; 
            }
        }
       
	close_sds(in_sds_info, in_sds_info->H5IN);
	close_sds(out_sds_info, out_sds_info->H5OUT); 
    }
}


