/****************************************************************************
!C

!File: sds_range.c

!Description:
  Read sds data range.  

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original September 1999. QA Tool Version 1.2
  Modified January 2000. Inlcude -layer option
  Modified March 2000.   Include -scale option
  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c


!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "qa_tool.h"
#include "main_util.h"
#include "alloc_mem.h"
#include "str_op.h"
#include "l2g.h"

#define MIN_DEF 2147483647 
#define MAX_DEF -2147483648 

#define HELP \
"NAME \n" \
"     sds_range - Print the observed range of an SDS in an HDF file\n" \
" \n" \
"SYNOPSIS \n" \
"       sds_range [-help] \n" \
"       sds_range [-sds=<sds_name>] [-layer] [-scale] filename \n" \
" \n" \
"       echo [-sds=sds_name] [-layer] [-scale] filename | sds_range \n" \
" \n" \
"DESCRIPTION \n" \
"	Read the SDS and print the minimum and maximum value in the SDS that \n" \
"is not a fill value and also print the fill value. Tool accepts 2D/3D/4D SDS.\n" \
"In case of 3D and 4D SDS user can request the detail for every SDS layer/slice. \n" \
"Output is directed to stdout and contains following information: \n" \
"\tSDS name, Layer id, Minimum, Maximum, fill value and out of range indicator. \n" \
"Layer id is provided only for 3D and 4D SDS upon request by the user. If an SDS \n" \
"contains only fill value then minimum and maximum field will contain a string of *. \n" \
"SDS attributes scale_factor and add_offset are printed when -scale option is used. \n" \
" \n" \
"OPTIONS \n" \
"	-help                  Display this help message \n" \
"	-sds=<sds_name>        SDS name \n" \
"	-layer                 Report details for every layer \n" \
"	-scale                 Scale the range values \n" \
"	filename               One or more input filenames \n" \
" \n" \
"Examples: \n" \
"\tsds_range -layer /DATA/YDAY/MOD09A1/1999049/MOD09A1.A1999049.h13v13.001.1999268235744.hdf \n" \
"\tsds_range -layer /DATA/JDAY/MOD43/1999066/MOD43B1.A1999064.h10v04.001.1999356235305.hdf \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:		sds_range [-sds=<sds name>] [-layer] [-scale] filename\n" \
" \n" \
"OPTIONS \n" \
"	-help                  Display this help message \n" \
"	-sds=<sds_name>        SDS name \n" \
"	-layer                 Report details for every layer \n" \
"	-scale                 Scale the range values \n" \
"	filename               One or more input filenames \n" \
" \n"

void read_sds_range(char *fname, char **sds_names, int nsds, int nd_info, int sc); 
void read_sds_range_for_row(sds_t *sds_info, void *data_in, int ncol, int layer_id, int st_c, int offset, 
			int *min, int *max, float *fmin, float *fmax);
void print_sds_range(sds_t *sds_info, int *min, int *max, float *fmin, float *fmax, int n_layer, int sc);

int main(int argc, char **argv)
{
  char **sds_names;
  int nsds = 0;
  int nd_info = 0;
  int scale_st = 0;
  int i, fcnt, sds_cnt;

  if ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_names in sds_range: main()\n");
  else
  {
    for (i=1, fcnt=0; i<argc; i++)
    {
      if (strcmp(argv[i], "-layer") == 0) nd_info = 1;
      else if (strcmp(argv[i], "-scale") == 0) scale_st = 1;
      else if (is_arg_id(argv[i], "-sds") == 0)
        get_arg_val_arr(argv[i], sds_names, &nsds);
      else if (argv[i][0] == '-')
        fprintf(stderr, "Ignoring unknown option %s\n", argv[i]);
      else fcnt++;
    }
    if (fcnt == 0) 
    {
      fprintf(stderr, "No file input \n");
      fprintf(stderr, "%s\n", USAGE);
    }
    else
    {

      /* MSZ 1/31/2018 silent version mismatch warning standard print out */
      setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

      /* MSZ 1/31/2018 Disable error reporting */
      H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

      for (i=1; i<argc; i++)
      {
        if (argv[i][0] != '-')
        {
	  fprintf(stdout, "Reading from %s\n", argv[i]);
	  if (nsds == 0) {
	    sds_cnt = get_sds_names(argv[i], sds_names);    
	  }  
          else sds_cnt = nsds;	  
	  read_sds_range(argv[i], sds_names, sds_cnt, nd_info, scale_st); 
        }
      }
    }
    Free2D((void **)sds_names);
  }
  return 0;
}

void read_sds_range(char *fname, char **sds_names, int nsds, int nd_info, int sc)
{
  void *data_in;
  sds_t sds_info;
  int rank;
  int n, m, max_m;
  int st_c, offset;
  int layer_id, n_layer;
  int irow, irank, isds;
  int nrows, ncols, ndata;
  int min[MAX_NUM_LAYERS], max[MAX_NUM_LAYERS];
  float32 fmin[MAX_NUM_LAYERS], fmax[MAX_NUM_LAYERS];
  int32 start[4], edge[4];

  sds_info.sd_id = -1;
  for (isds=0; isds<nsds; isds++)
  {
    sds_info.sds_id = -1;
    strcpy(sds_info.name, sds_names[isds]);
    for (irank=0; irank<4; irank++)
      start[irank] = edge[irank] = 0;
    if (get_sds_info(fname, &sds_info) != -1)
    {
      rank = sds_info.rank;
      n_layer = 1;
      if (rank == 1)
      {
        nrows = 1;
	edge[0] = ncols = ndata = sds_info.dim_size[0];
      }
      else
      {
        if (sds_info.interleave == 1)
        {
          nrows = sds_info.dim_size[rank-2];
          ncols = ndata = edge[rank-1] = sds_info.dim_size[rank-1];
          for (irank=0; irank<rank-2; irank++)
          {
	    n_layer *= sds_info.dim_size[irank];
            ndata *= sds_info.dim_size[irank];
            edge[irank] = sds_info.dim_size[irank];
          }
          edge[rank-2] = 1;
	  max_m = sds_info.dim_size[1];
        }
        else
        {
	  nrows = sds_info.dim_size[0];
	  ncols = ndata = sds_info.dim_size[1];
	  for (irank=2; irank<rank; irank++)
	  {
	    n_layer *= sds_info.dim_size[irank];
	    ndata *= sds_info.dim_size[irank];
	    edge[irank] = sds_info.dim_size[irank];
	  }
	  edge[0] = 1;
	  edge[1] = sds_info.dim_size[1];
	  max_m = sds_info.dim_size[3];
        }
      }

      for (layer_id=0; layer_id<MAX_NUM_LAYERS; layer_id++)
      {
        min[layer_id] = MIN_DEF;
        max[layer_id] = MAX_DEF;
        fmin[layer_id] = MIN_DEF;
        fmax[layer_id] = MAX_DEF;
      }

      if ((nd_info == 0) || (rank <= 2))
      {
	  st_c = 0;
	  offset = 1;
	  n_layer = 1;
      }
      for (irow=0; irow<nrows; irow++)
      {
	  if (sds_info.interleave == 1) start[rank-2] = irow;
	  else start[0] = irow;
	  if( ( data_in = get_sel_sds_data(&sds_info, start, edge) ) == NULL ) 
	    fprintf(stderr, "Failed to read data row for SDS %s\n", sds_info.name);
	  else
          {
	    for (layer_id=0; layer_id<n_layer; layer_id++)
	    {
	      if (nd_info == 1) 
	      {
		m = (rank == 3) ? -1 : layer_id%max_m;
		n = (rank == 3) ? layer_id : layer_id/max_m;
		compute_sds_start_offset(&sds_info, n, m, &st_c, &offset);
	      }
	      else ncols = ndata;
	      read_sds_range_for_row(&sds_info, data_in, ncols, layer_id, st_c, offset, min, max, fmin, fmax);
	    }
	    if(data_in) free(data_in);
	  }
      } /* for (irow=0;  . .  ) */
      print_sds_range(&sds_info, min, max, fmin, fmax, n_layer, sc);
      close_sds(&sds_info, sds_info.H5IN);
    }
  } /* for (isds=0;  . .  ) */
  close_sd(&sds_info, sds_info.H5IN);
}
	        
void read_sds_range_for_row(sds_t *sds_info, void *data_in, int ncol, int layer_id, int st_c, int offset, 
			int *min, int *max, float *fmin, float *fmax)
{
  float sds_fval;
  int i, icol, sds_val;
 
  switch(sds_info->data_type)
  {
    case 5 : for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_fval = ((float32 *)data_in)[icol]; 
      	       if (sds_fval != sds_info->fill_fval)
      	       {
      	         if (sds_fval < fmin[layer_id]) fmin[layer_id] = sds_fval;
                 if (sds_fval > fmax[layer_id]) fmax[layer_id] = sds_fval;
      	       }
             }
    	     break;
    case 20: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((int8 *)data_in)[icol]; 
      	       if (sds_val != sds_info->fill_val)
      	       {
      	         if (sds_val < min[layer_id]) min[layer_id] = sds_val;
                 if (sds_val > max[layer_id]) max[layer_id] = sds_val;
      	       }
             }
    	     break;
    case 21: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((uint8 *)data_in)[icol]; 
      	       if ((uint8)sds_val != (uint8)sds_info->fill_val)
      	       {
      	         if (sds_val < min[layer_id]) min[layer_id] = sds_val;
                 if (sds_val > max[layer_id]) max[layer_id] = sds_val;
      	       }
             }
    	     break;
    case 22: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((int16 *)data_in)[icol]; 
      	       if (sds_val != sds_info->fill_val)
      	       {
      	         if (sds_val < min[layer_id]) min[layer_id] = sds_val;
                 if (sds_val > max[layer_id]) max[layer_id] = sds_val;
      	       }
             }
    	     break;
    case 23: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((uint16 *)data_in)[icol]; 
      	       if ((uint16)sds_val != (uint16)sds_info->fill_val)
      	       {
      	         if (sds_val < min[layer_id]) min[layer_id] = sds_val;
                 if (sds_val > max[layer_id]) max[layer_id] = sds_val;
      	       }
             }
    	     break;
    case 24: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((int32 *)data_in)[icol]; 
      	       if (sds_val != sds_info->fill_val)
      	       {
      	         if (sds_val < min[layer_id]) min[layer_id] = sds_val;
                 if (sds_val > max[layer_id]) max[layer_id] = sds_val;
      	       }
             }
    	     break;
    case 25: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((uint32 *)data_in)[icol]; 
      	       if (sds_val != sds_info->fill_val)
      	       {
      	         if (sds_val < min[layer_id]) min[layer_id] = sds_val;
                 if (sds_val > max[layer_id]) max[layer_id] = sds_val;
      	       }
             }
    	     break;
  }
}
        
void print_sds_range(sds_t *sds_info, int *min, int *max, float *fmin, float *fmax, int n_layer, int sc)
{
  void *attr_val;
  int scale_offset[2];
  int i, m, n, layer_id;
  float fscale_offset[2];
  int32 attr_type, attr_cnt;
  char **layer_str, dim_str[255];
  char *attr_name[] = {"scale_factor", "add_offset"};
  char *attr_name1[] = {"Scale", "Offset"};  

  fscale_offset[0] = fscale_offset[1] = 0;
  scale_offset[0] = scale_offset[1] = 0;
  if (sc == 1)
  {
    for (i=0; i<2; i++)
    {
      if ((attr_val = (void *)get_sds_attr(sds_info, attr_name[i], &attr_type, &attr_cnt)) != NULL)
      {
        switch(attr_type)
        {
          case 5 : fscale_offset[i] = ((float32 *)attr_val)[0]; break;
          case 6 : fscale_offset[i] = ((double *)attr_val)[0]; break;
          case 20: scale_offset[i] = ((int8 *)attr_val)[0]; break;
          case 21: scale_offset[i] = ((uint8 *)attr_val)[0]; break;
          case 22: scale_offset[i] = ((int16 *)attr_val)[0]; break;
          case 23: scale_offset[i] = ((uint16 *)attr_val)[0]; break;
          case 24: scale_offset[i] = ((int32 *)attr_val)[0]; break;
          case 25: scale_offset[i] = ((uint32 *)attr_val)[0]; break;
        }
        free(attr_val);
      } 
      else
      {
	if ((attr_val = (void *)get_sds_attr(sds_info, attr_name1[i], &attr_type, &attr_cnt)) != NULL)
	{
          switch(attr_type)
          {
            case 5 : fscale_offset[i] = ((float32 *)attr_val)[0]; break;
            case 6 : fscale_offset[i] = ((double *)attr_val)[0]; break;
            case 20: scale_offset[i] = ((int8 *)attr_val)[0]; break;
            case 21: scale_offset[i] = ((uint8 *)attr_val)[0]; break;
            case 22: scale_offset[i] = ((int16 *)attr_val)[0]; break;
            case 23: scale_offset[i] = ((uint16 *)attr_val)[0]; break;
            case 24: scale_offset[i] = ((int32 *)attr_val)[0]; break;
            case 25: scale_offset[i] = ((uint32 *)attr_val)[0]; break;
          }
          free(attr_val);
	}       
      }                                                          
    }                                 
  }
  if (sds_info->rank == 2) 
    sprintf(dim_str, "(%d x %d)", sds_info->dim_size[0], sds_info->dim_size[1]);
  else if (sds_info->rank == 3) 
    sprintf(dim_str, "(%d x %d x %d)", sds_info->dim_size[0], sds_info->dim_size[1], sds_info->dim_size[2]);
  else if (sds_info->rank == 4) 
    sprintf(dim_str, "(%d x %d x %d x %d)", sds_info->dim_size[0], sds_info->dim_size[1], sds_info->dim_size[2], sds_info->dim_size[3]);
  if ((layer_str = (char **)Calloc2D(MAX_NUM_LAYERS, 10, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for layer_str in print_sds_range()\n");
  else if (n_layer > 1)
  {
    if (sds_info->rank == 3)
      for (layer_id=0; layer_id<n_layer; layer_id++)
        sprintf(layer_str[layer_id], "%d:\t", layer_id+1);
    else if (sds_info->rank == 4)
      for (layer_id=0; layer_id<n_layer; layer_id++)
      {
        m = (sds_info->dim_size[0] < sds_info->dim_size[3]) ? layer_id/sds_info->dim_size[1]+1 : layer_id/sds_info->dim_size[3]+1;
        n = (sds_info->dim_size[0] < sds_info->dim_size[3]) ? layer_id%sds_info->dim_size[1]+1 : layer_id%sds_info->dim_size[3]+1;
        sprintf(layer_str[layer_id], "%d.%d:\t", m, n);
      }
  }
  if (n_layer == 1) fprintf(stdout, "\t%s %s", sds_info->name, dim_str);
  else fprintf(stdout, "\t%s %s\n", sds_info->name, dim_str);

  switch(sds_info->data_type)
  {
    case 5 : for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (fmin[layer_id] != MIN_DEF) fprintf(stdout, "\t\t%s%f", layer_str[layer_id], fmin[layer_id]);
	       else fprintf(stdout, "\t\t%s*******", layer_str[layer_id]);
	       if (fmax[layer_id] != MAX_DEF) fprintf(stdout, "\t%f", fmax[layer_id]);
	       else fprintf(stdout, "\t*******");
	       fprintf(stdout, "\t%f", sds_info->fill_fval);
	       if (attr_type <= 6) {
		 if (fscale_offset[0] != 0) fprintf(stdout, "\t%f\t%f", fscale_offset[0], fscale_offset[1]);
	       }
	       else {
		 if (scale_offset[0] != 0) fprintf(stdout, "\t%d\t%d", scale_offset[0], scale_offset[1]);
	       }
	       if (((fmin[layer_id] != MIN_DEF) && (fmin[layer_id] < sds_info->frange[0])) ||
	           ((fmax[layer_id] != MAX_DEF) && (fmax[layer_id] > sds_info->frange[1])))
	         fprintf(stdout, "\tout_of_range!");
	       fprintf(stdout, "\n");
	     }
	     break;
    case 20: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (min[layer_id] != MIN_DEF) fprintf(stdout, "\t\t%s%d", layer_str[layer_id], min[layer_id]);
	       else fprintf(stdout, "\t\t%s*******", layer_str[layer_id]);
	       if (max[layer_id] != MAX_DEF) fprintf(stdout, "\t%d", max[layer_id]);
	       else fprintf(stdout, "\t*******");
	       fprintf(stdout, "\t%d", sds_info->fill_val);
	       if (attr_type <= 6) {
		 if (fscale_offset[0] != 0) fprintf(stdout, "\t%f\t%f", fscale_offset[0], fscale_offset[1]);
	       }
	       else {
		 if (scale_offset[0] != 0) fprintf(stdout, "\t%d\t%d", scale_offset[0], scale_offset[1]);
	       }
	       if (((min[layer_id] != MIN_DEF) && (min[layer_id] < sds_info->range[0])) ||
	           ((max[layer_id] != MAX_DEF) && (max[layer_id] > sds_info->range[1])))
	         fprintf(stdout, "\tout_of_range!");
	       fprintf(stdout, "\n");
	     }
	     break;
    case 21: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (min[layer_id] != MIN_DEF) fprintf(stdout, "\t\t%s%u", layer_str[layer_id], (unsigned int)min[layer_id]);
	       else fprintf(stdout, "\t\t%s*******", layer_str[layer_id]);
	       if (max[layer_id] != MAX_DEF) fprintf(stdout, "\t%u", (unsigned int)max[layer_id]);
	       else fprintf(stdout, "\t*******");
	       fprintf(stdout, "\t%u", (unsigned int)sds_info->fill_val);
	       if (attr_type <= 6) {
		 if (fscale_offset[0] != 0) fprintf(stdout, "\t%f\t%f", fscale_offset[0], fscale_offset[1]);
	       }
	       else {
		 if (scale_offset[0] != 0) fprintf(stdout, "\t%d\t%d", scale_offset[0], scale_offset[1]);
	       }
	       if (((min[layer_id] != MIN_DEF) && ((uint8)min[layer_id] < (uint8)sds_info->range[0])) ||
	           ((max[layer_id] != MAX_DEF) && ((uint8)max[layer_id] > (uint8)sds_info->range[1])))
	         fprintf(stdout, "\tout_of_range!");
	       fprintf(stdout, "\n");
	     }
	     break;
    case 22: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (min[layer_id] != MIN_DEF) fprintf(stdout, "\t\t%s%d", layer_str[layer_id], min[layer_id]);
	       else fprintf(stdout, "\t\t%s*******", layer_str[layer_id]);
	       if (max[layer_id] != MAX_DEF) fprintf(stdout, "\t%d", max[layer_id]);
	       else fprintf(stdout, "\t*******");
	       fprintf(stdout, "\t%d", sds_info->fill_val);
	       if (attr_type <= 6) {
		 if (fscale_offset[0] != 0) fprintf(stdout, "\t%f\t%f", fscale_offset[0], fscale_offset[1]);
	       }
	       else {
		 if (scale_offset[0] != 0) fprintf(stdout, "\t%d\t%d", scale_offset[0], scale_offset[1]);
	       }
	       if (((min[layer_id] != MIN_DEF) && (min[layer_id] < sds_info->range[0])) ||
	           ((max[layer_id] != MAX_DEF) && (max[layer_id] > sds_info->range[1])))
	         fprintf(stdout, "\tout_of_range!");
	       fprintf(stdout, "\n");
	     }
	     break;
    case 23: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (min[layer_id] != MIN_DEF) fprintf(stdout, "\t\t%s%u", layer_str[layer_id], (unsigned int)min[layer_id]);
	       else fprintf(stdout, "\t\t%s*******", layer_str[layer_id]);
	       if (max[layer_id] != MAX_DEF) fprintf(stdout, "\t%u", (unsigned int)max[layer_id]);
	       else fprintf(stdout, "\t*******");
	       fprintf(stdout, "\t%u", (unsigned int)sds_info->fill_val);
	       if (attr_type <= 6) {
		 if (fscale_offset[0] != 0) fprintf(stdout, "\t%f\t%f", fscale_offset[0], fscale_offset[1]);
	       }
	       else {
		 if (scale_offset[0] != 0) fprintf(stdout, "\t%d\t%d", scale_offset[0], scale_offset[1]);
	       }
	       if (((min[layer_id] != MIN_DEF) && ((uint16)min[layer_id] < (uint16)sds_info->range[0])) ||
	           ((max[layer_id] != MAX_DEF) && ((uint16)max[layer_id] > (uint16)sds_info->range[1])))
	         fprintf(stdout, "\tout_of_range!");
	       fprintf(stdout, "\n");
	     }
	     break;
    case 24: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (min[layer_id] != MIN_DEF) fprintf(stdout, "\t\t%s%d", layer_str[layer_id], min[layer_id]);
	       else fprintf(stdout, "\t\t%s*******", layer_str[layer_id]);
	       if (max[layer_id] != MAX_DEF) fprintf(stdout, "\t%d", max[layer_id]);
	       else fprintf(stdout, "\t*******");
	       fprintf(stdout, "\t%d", sds_info->fill_val);
	       if (attr_type <= 6) {
		 if (fscale_offset[0] != 0) fprintf(stdout, "\t%f\t%f", fscale_offset[0], fscale_offset[1]);
	       }
	       else {
		 if (scale_offset[0] != 0) fprintf(stdout, "\t%d\t%d", scale_offset[0], scale_offset[1]);
	       }
	       if (((min[layer_id] != MIN_DEF) && (min[layer_id] < sds_info->range[0])) ||
	           ((max[layer_id] != MAX_DEF) && (max[layer_id] > sds_info->range[1])))
	         fprintf(stdout, "\tout_of_range!");
	       fprintf(stdout, "\n");
	     }
	     break;
    case 25: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (min[layer_id] != MIN_DEF) fprintf(stdout, "\t\t%s%u", layer_str[layer_id], (unsigned int)min[layer_id]);
	       else fprintf(stdout, "\t\t%s*******", layer_str[layer_id]);
	       if (max[layer_id] != MAX_DEF) fprintf(stdout, "\t%u", (unsigned int)max[layer_id]);
	       else fprintf(stdout, "\t*******");
	       fprintf(stdout, "\t%u", (unsigned int)sds_info->fill_val);
	       if (attr_type == 6) {
		 if (fscale_offset[0] != 0) fprintf(stdout, "\t%f\t%f", fscale_offset[0], fscale_offset[1]);
	       }
	       else {
		 if (scale_offset[0] != 0) fprintf(stdout, "\t%d\t%d", scale_offset[0], scale_offset[1]);
	       }
	       if ((((uint32)min[layer_id] != MIN_DEF) && ((uint32)min[layer_id] < (uint32)sds_info->range[0])) ||
	           ((max[layer_id] != MAX_DEF) && ((uint32)max[layer_id] > (uint32)sds_info->range[1])))
	         fprintf(stdout, "\tout_of_range!");
	       fprintf(stdout, "\n");
	     }
	     break;
  }
  Free2D((void **)layer_str);
}


