/****************************************************************************
!C

!File: sds_values.c

!Description:
   Print a list of different values present in an SDS.  

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original March 2000. QA Tool Version 1.2
  Modified August 06, 2001
  Modified January, 2018 Version 3.5.1 by Maosheng Zhao, SSAI
        1)modify it so that HDF5 can be read
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c


!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "qa_tool.h"
#include "main_util.h"
#include "alloc_mem.h"
#include "str_op.h"
#include "l2g.h"

#define MAX_NUM_VALUES 16000

#define HELP \
"NAME \n" \
"     sds_values - Print the list of different values present in one or more SDS of an HDF file\n" \
"     and optionally produce the ENVI compatible density slice color table. \n" \
" \n" \
"SYNOPSIS \n" \
"       sds_values [-help] \n" \
"       sds_values [-sds=<sds_name>] [-layer] [-clr=<nc,f1,f2>] filename \n" \
" \n" \
"       echo [-sds=sds_name] [-layer] [-clr=<nc,f1,f2>] filename | sds_values \n" \
" \n" \
"DESCRIPTION \n" \
"	Read the SDS and print all different values present in one or more SDS \n" \
"of an HDF file. Tool accepts 2D/3D/4D SDS. In case of 3D and 4D SDS user can \n" \
"request the detail for every SDS layer/slice. Output is directed to stdout. \n" \
"Output contains SDS name with dimension size and fill value followed by list of \n" \
"SDS values. In case of 3D and 4D SDSs fill value and list of SDS values are \n" \
"listed for each layer. SDS layers are identified by the dimension index of that \n" \
" layer. \n" \
"	An ENVI compatible density slice color table can be produced. Requested \n" \
"number of colors at equal intervals are selected from the input color table \n" \
"covering the full range of sds values. Output color table is written to an ASCII file.\n" \
"If the number of distinct sds values are less than or equal to specified number \n" \
"of output colors distinct colors are assigned to distinct SDS values, otherwise \n" \
"the SDS value range is divided into number of colors and colors are assigend to \n" \
"range of SDS values. \n" \
" \n" \
"OPTIONS \n" \
"	-help                  Display this help message \n" \
"	-sds=<sds_name>        SDS name \n" \
"	-layer                 Report details for every layer \n" \
"	-clr=<nc,f1,f2>        Number of output colors, input file f1 containing full \n" \
"	                       range color table, and output file f2 for ENVI compatible\n" \
"	                       density slice color table. \n" \
"	filename               One or more input filenames \n" \
" \n" \
"Examples: \n" \
"\tsds_values -sds=\"500m Reflectance Band Quality\"  /DATA/MOD09/2000064/MOD09.A2000064.1500.002.2000069220633.hdf \n" \
"\tsds_values -layer -sds=Band_QC -layer /DATA/MODAGAGG/2000065/MODAGAGG.A2000065.h13v02.002.2000075160322.hdf\n" \
" \n" \
"Note: \n" \
"\tThis tool is mainly for use when an SDS contains few number of different values such as \n" \
"QA SDS. The tool limits the maximum number of output values to 16000. The tool takes  \n" \
"enormously long time to complete is run on regular Science Data that contains wide range of \n" \
"values. \n\n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:		sds_values [-sds=<sds name>] [-layer] [-clr=<nc,file>] filename\n" \
" \n" \
"OPTIONS \n" \
"	-help                  Display this help message \n" \
"	-sds=<sds_name>        SDS name \n" \
"	-layer                 Report details for every layer \n" \
"	-clr=<nc,f1,f2>        Number of output colors, input file f1 containing full \n" \
"	                       range color table, and output file f2 for ENVI compatible\n" \
"	                       density slice color table. \n" \
"	filename               One or more input filenames \n" \
" \n"

void read_sds_values(char *fname, char **sds_names, int nsds, int nd_info, int nc, char *in_clr_fname, char *out_clr_fname); 
void read_sds_values_in_row(sds_t *sds_info, void *data_in, int ncol, int layer_id, int st_c, int offset, 
			int **sds_values, float **sds_fvalues, int *num_val);
void print_sds_values(sds_t *sds_info, int **sds_values, float **sds_fvalues, int *num_val, int n_layer);
int create_color_table(int32 data_type, int **sds_values, float **sds_fvalues, int *num_val, int n_layer,
	int nc, char *in_clr_fname, char *out_clr_fname);

int main(int argc, char **argv)
{
  char clr_str[MAX_STR_LEN];
  char **sds_names, tmp_str[10];
  char in_clr_fname[MAX_PATH_LENGTH];
  char out_clr_fname[MAX_PATH_LENGTH];
  int i, fcnt, nd_info, nc;
  int nsds, sds_cnt, p1, p2, len;

  /* MSZ 1/26/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 1/26/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_names in sds_range: main()\n");
  else
  {
    nc = 0;
    in_clr_fname[0] = '\0';
    out_clr_fname[0] = '\0';
    for (i=1, fcnt=nsds=nd_info=0; i<argc; i++)
    {
      if (strcmp(argv[i], "-layer") == 0) 
	nd_info = 1;
      else if (is_arg_id(argv[i], "-sds") == 0)
        get_arg_val_arr(argv[i], sds_names, &nsds);
      else if (is_arg_id(argv[i], "-clr") == 0)
      {
        get_arg_val(argv[i], clr_str);
        p1 = charpos(clr_str, ',', 0);
        p2 = charpos(clr_str, ',', p1+1);
	if ((p1 == -1) || (p2 == -1))
	  fprintf(stderr, "Ignoring invalid color option %s\n", argv[i]);
	else 
	{
          len = (int)strlen(clr_str);
	  strmid(clr_str, 0, p1, tmp_str);
	  nc = (int)atoi(tmp_str);
	  p1++;
	  strmid(clr_str, p1, p2-p1, in_clr_fname);
	  p2++;
	  strmid(clr_str, p2, len-p2, out_clr_fname);
	}
      }
      else if (argv[i][0] == '-')
        fprintf(stderr, "Ignoring unknown option %s\n", argv[i]);
      else fcnt++;
    }
    if (fcnt == 0) 
    {
      fprintf(stderr, "No file input \n");
      fprintf(stderr, "%s\n", USAGE);
    }
    else
    {
      for (i=1; i<argc; i++)
      {
        if (argv[i][0] != '-')
        {
	  fprintf(stdout, "Reading from %s\n", argv[i]);
	  if (nsds == 0) 
	    sds_cnt = get_sds_names(argv[i], sds_names);
          else sds_cnt = nsds;
	  read_sds_values(argv[i], sds_names, sds_cnt, nd_info, nc, in_clr_fname, out_clr_fname); 
        }
      }
    }
    Free2D((void **)sds_names);
  }
  return 0;
}

void read_sds_values(char *fname, char **sds_names, int nsds, int nd_info, int nc, char *in_clr_fname, char *out_clr_fname)
{
  void *data_in;
  sds_t sds_info;
  int rank;
  int n, m, max_m;
  int st_c, offset;
  int layer_id, n_layer;
  int irow, irank, isds;
  int nrows, ncols, ndata;
  int *num_val, **sds_values;
  float **sds_fvalues;
  int32 start[4], edge[4];

  /* initialize as -1 for sd_id and sds_id */
  sds_info.sd_id = sds_info.sds_id = -1;

  if(open_sd(fname, &sds_info, 'R') == FAIL)
    fprintf(stderr, "Cannot open input HDF file %s\n", fname);
  else
  for (isds=0; isds<nsds; isds++)
  {
    strcpy(sds_info.name, sds_names[isds]);
    for (irank=0; irank<4; irank++)
      start[irank] = edge[irank] = 0;
    if (get_sds_info((char *)NULL, &sds_info) != -1)
    {
      rank = sds_info.rank;
      n_layer = 1;
      if (rank == 1)
      {
        nrows = 1;
	edge[0] = ncols = ndata = sds_info.dim_size[0];
      }
      else
      {
        if (sds_info.interleave == 1)
        {
          nrows = sds_info.dim_size[rank-2];
          ncols = ndata = edge[rank-1] = sds_info.dim_size[rank-1];
          for (irank=0; irank<rank-2; irank++)
          {
	    n_layer *= sds_info.dim_size[irank];
            ndata *= sds_info.dim_size[irank];
            edge[irank] = sds_info.dim_size[irank];
          }
          edge[rank-2] = 1;
	  max_m = sds_info.dim_size[1];
        }
        else
        {
	  nrows = sds_info.dim_size[0];
	  ncols = ndata = sds_info.dim_size[1];
	  for (irank=2; irank<rank; irank++)
	  {
	    n_layer *= sds_info.dim_size[irank];
	    ndata *= sds_info.dim_size[irank];
	    edge[irank] = sds_info.dim_size[irank];
	  }
	  edge[0] = 1;
	  edge[1] = sds_info.dim_size[1];
	  max_m = sds_info.dim_size[3];
        }
      }

      if ((data_in = (void *)calloc(ndata, sds_info.data_size)) == NULL)
	fprintf(stderr, "Cannot allocate memory for data_in in read_sds_range()\n");
      else
      {
	if ((nd_info == 0) || (rank <= 2))
	{
	  st_c = 0;
	  offset = 1;
	  n_layer = 1;
	}

        if (sds_info.data_type == 5)
        {
	  if ((sds_fvalues = (float **)Calloc2D(n_layer, MAX_NUM_VALUES, sizeof(float))) == NULL)
	    fprintf(stderr, "Cannot allocate memory for sds_values in read_sds_values\n");	
        }
        else
        {
	  if ((sds_values = (int **)Calloc2D(n_layer, MAX_NUM_VALUES, sizeof(int))) == NULL)
	    fprintf(stderr, "Cannot allocate memory for sds_values in read_sds_values\n");	
        } 
        if ((num_val = (int *)calloc(n_layer, sizeof(int))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for num_val in read_sds_values()\n");
	for (layer_id=0; layer_id<n_layer; layer_id++)
	  num_val[layer_id] = 0;

        for (irow=0; irow<nrows; irow++)
        {
          if (rank == 1) start[0] = 0;
	  else
	  {
	    if (sds_info.interleave == 1) start[rank-2] = irow;
	    else start[0] = irow;
	  }
	  if (read_sel_sds_data(&sds_info, start, edge, data_in) == FAIL)
	    fprintf(stderr, "Failed to read data row for SDS %s\n", sds_info.name);
	  else
          {
	    for (layer_id=0; layer_id<n_layer; layer_id++)
	    {
	      if (nd_info == 1) 
	      {
		m = (rank == 3) ? -1 : layer_id%max_m;
		n = (rank == 3) ? layer_id : layer_id/max_m;
		compute_sds_start_offset(&sds_info, n, m, &st_c, &offset);
	      }
	      else ncols = ndata;
	      read_sds_values_in_row(&sds_info, data_in, ncols, layer_id, st_c, offset, sds_values, sds_fvalues, &num_val[layer_id]);
	    }
	  }
        } /* for (irow=0;  . .  ) */
        if (sds_info.data_type == 5) 
        {
	  for (layer_id=0; layer_id<n_layer; layer_id++)
	    sort_fvalues(sds_fvalues[layer_id], num_val[layer_id]);
        }
	else
        {
	  for (layer_id=0; layer_id<n_layer; layer_id++)
	    sort_values(sds_values[layer_id], num_val[layer_id]);
	}
        print_sds_values(&sds_info, sds_values, sds_fvalues, num_val, n_layer);
        if ((in_clr_fname[0] != '\0') && (out_clr_fname[0] != '\0'))
        {
	  if (create_color_table(sds_info.data_type, sds_values, sds_fvalues, num_val, n_layer, nc, 
		in_clr_fname, out_clr_fname) == -1)
	    fprintf(stderr, "Output color table file not created\n");
	}
	if (sds_info.data_type == 5)
	  Free2D((void **)sds_fvalues);
	else
	  Free2D((void **)sds_values);
	free(num_val);
        free(data_in);
      }  
      close_sds(&sds_info, sds_info.H5IN);
    }
  } /* for (isds=0;  . .  ) */
  close_sd(&sds_info, sds_info.H5IN);
}
	        
void read_sds_values_in_row(sds_t *sds_info, void *data_in, int ncol, int layer_id, int st_c, int offset, 
			int **sds_values, float **sds_fvalues, int *num_val)
{
  float sds_fval;
  int i, j, icol, sds_val;

  switch(sds_info->data_type)
  {
    case 5 : for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
	       if (*num_val == MAX_NUM_VALUES - 1) break;
      	       sds_fval = ((float *)data_in)[icol]; 
      	       if (sds_fval != sds_info->fill_fval)
	       {
		 for (j=0; j < *num_val; j++)
		   if (sds_fval == sds_fvalues[layer_id][j]) break;
		 if (j >= *num_val)
		 {
		   sds_fvalues[layer_id][*num_val] = sds_fval;
		   ++*num_val;
		 }
	       }	
             }
    	     break;
    case 20: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
	       if (*num_val == MAX_NUM_VALUES - 1) break;
      	       sds_val = ((int8 *)data_in)[icol]; 
      	       if (sds_val != sds_info->fill_val)
      	       {
		 for (j=0; j < *num_val; j++)
		   if (sds_val == sds_values[layer_id][j]) break;
		 if (j >= *num_val)
		 {
		   sds_values[layer_id][*num_val] = sds_val;
		   ++*num_val;
		 }
      	       }
             }
    	     break;
    case 21: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
	       if (*num_val == MAX_NUM_VALUES - 1) break;
      	       sds_val = ((uint8 *)data_in)[icol]; 
      	       if ((uint8)sds_val != (uint8)sds_info->fill_val)
      	       {
		 for (j=0; j < *num_val; j++)
		   if (sds_val == sds_values[layer_id][j]) break;
		 if (j >= *num_val)
		 {
		   sds_values[layer_id][*num_val] = sds_val;
		   ++*num_val;
		 }
      	       }
             }
    	     break;
    case 22: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
	       if (*num_val == MAX_NUM_VALUES - 1) break;
      	       sds_val = ((int16 *)data_in)[icol]; 
      	       if (sds_val != sds_info->fill_val)
      	       {
		 for (j=0; j < *num_val; j++)
		   if (sds_val == sds_values[layer_id][j]) break;
		 if (j >= *num_val)
		 {
		   sds_values[layer_id][*num_val] = sds_val;
		   ++*num_val;
		 }
      	       }
             }
    	     break;
    case 23: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
	       if (*num_val == MAX_NUM_VALUES - 1) break;
      	       sds_val = ((uint16 *)data_in)[icol]; 
      	       if ((uint16)sds_val != (uint16)sds_info->fill_val)
      	       {
		 for (j=0; j < *num_val; j++)
		   if (sds_val == sds_values[layer_id][j]) break;
		 if (j >= *num_val)
		 {
		   sds_values[layer_id][*num_val] = sds_val;
		   ++*num_val;
		 }
      	       }
             }
    	     break;
    case 24: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
	       if (*num_val == MAX_NUM_VALUES - 1) break;
      	       sds_val = ((int32 *)data_in)[icol]; 
      	       if (sds_val != sds_info->fill_val)
      	       {
		 for (j=0; j < *num_val; j++)
		   if (sds_val == sds_values[layer_id][j]) break;
		 if (j >= *num_val)
		 {
		   sds_values[layer_id][*num_val] = sds_val;
		   ++*num_val;
		 }
      	       }
             }
    	     break;
    case 25: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
	       if (*num_val == MAX_NUM_VALUES - 1) break;
      	       sds_val = ((uint32 *)data_in)[icol]; 
      	       if (sds_val != sds_info->fill_val)
      	       {
		 for (j=0; j < *num_val; j++)
		   if (sds_val == sds_values[layer_id][j]) break;
		 if (j >= *num_val)
		 {
		   sds_values[layer_id][*num_val] = sds_val;
		   ++*num_val;
		 }
      	       }
             }
    	     break;
  }
}
        
void print_sds_values(sds_t *sds_info, int **sds_values, float **sds_fvalues, int *num_val, int n_layer)
{
  int i, j;
  int m, n, layer_id;
  char **layer_str, dim_str[255];

  if (sds_info->rank == 2) 
    sprintf(dim_str, "(%d x %d)", sds_info->dim_size[0], sds_info->dim_size[1]);
  else if (sds_info->rank == 3) 
    sprintf(dim_str, "(%d x %d x %d)", sds_info->dim_size[0], sds_info->dim_size[1], sds_info->dim_size[2]);
  else if (sds_info->rank == 4) 
    sprintf(dim_str, "(%d x %d x %d x %d)", sds_info->dim_size[0], sds_info->dim_size[1], sds_info->dim_size[2], sds_info->dim_size[3]);
  if ((layer_str = (char **)Calloc2D(n_layer, 10, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for layer_str in print_sds_range()\n");
  else if (n_layer > 1)
  {
    if (sds_info->rank == 3)
      for (layer_id=0; layer_id<n_layer; layer_id++)
        sprintf(layer_str[layer_id], "%d:\t", layer_id+1);
    else if (sds_info->rank == 4)
      for (layer_id=0; layer_id<n_layer; layer_id++)
      {
        m = (sds_info->dim_size[0] < sds_info->dim_size[3]) ? layer_id/sds_info->dim_size[1]+1 : layer_id/sds_info->dim_size[3]+1;
        n = (sds_info->dim_size[0] < sds_info->dim_size[3]) ? layer_id%sds_info->dim_size[1]+1 : layer_id%sds_info->dim_size[3]+1;
        sprintf(layer_str[layer_id], "%d.%d:\t", m, n);
      }
  }
  if (n_layer == 1) fprintf(stdout, "\t%s %s", sds_info->name, dim_str);
  else fprintf(stdout, "\t%s %s\n", sds_info->name, dim_str);
  switch(sds_info->data_type)
  {
    case 5 : for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (n_layer == 1) fprintf(stdout, "\t\t%s%f\n\t\t", layer_str[layer_id], sds_info->fill_fval);
	       else fprintf(stdout, "\t\t%s%f\n\t\t\t", layer_str[layer_id], sds_info->fill_fval);
	       for (i=0, j=0; i<num_val[layer_id]; i++, j++)
	       {
		 fprintf(stdout, "%f ", sds_fvalues[layer_id][i]);
		 if (j == 9)
	         {
		   if (n_layer == 1) fprintf(stdout, "\n\t\t");
		   else fprintf(stdout, "\n\t\t\t");
		   j = -1;
		 }
	       }
	       fprintf(stdout, "\n");
	     }
	     break;
    case 20: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (n_layer == 1) fprintf(stdout, "\t\t%s%d\n\t\t", layer_str[layer_id], sds_info->fill_val);
	       else fprintf(stdout, "\t\t%s%d\n\t\t\t", layer_str[layer_id], sds_info->fill_val);
	       for (i=0, j=0; i<num_val[layer_id]; i++, j++)
	       {
		 fprintf(stdout, "%d ", sds_values[layer_id][i]);
		 if (j == 9)
	         {
		   if (n_layer == 1) fprintf(stdout, "\n\t\t");
		   else fprintf(stdout, "\n\t\t\t");
		   j = -1;
		 }
	       }
	       fprintf(stdout, "\n");
	     }
	     break;
    case 21: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (n_layer == 1) fprintf(stdout, "\t\t%s%d\n\t\t", layer_str[layer_id], sds_info->fill_val);
	       else fprintf(stdout, "\t\t%s%d\n\t\t\t", layer_str[layer_id], sds_info->fill_val);
	       for (i=0, j=0; i<num_val[layer_id]; i++, j++)
	       {
		 fprintf(stdout, "%d ", sds_values[layer_id][i]);
		 if (j == 9)
	         {
		   if (n_layer == 1) fprintf(stdout, "\n\t\t");
		   else fprintf(stdout, "\n\t\t\t");
		   j = -1;
		 }
	       }
	       fprintf(stdout, "\n");
	     }
	     break;
    case 22: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (n_layer == 1) fprintf(stdout, "\t\t%s%d\n\t\t", layer_str[layer_id], sds_info->fill_val);
	       else fprintf(stdout, "\t\t%s%d\n\t\t\t", layer_str[layer_id], sds_info->fill_val);
	       for (i=0, j=0; i<num_val[layer_id]; i++, j++)
	       {
		 fprintf(stdout, "%d ", sds_values[layer_id][i]);
		 if (j == 9)
	         {
		   if (n_layer == 1) fprintf(stdout, "\n\t\t");
		   else fprintf(stdout, "\n\t\t\t");
		   j = -1;
		 }
	       }
	       fprintf(stdout, "\n");
	     }
	     break;
    case 23: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (n_layer == 1) fprintf(stdout, "\t\t%s%d\n\t\t", layer_str[layer_id], sds_info->fill_val);
	       else fprintf(stdout, "\t\t%s%d\n\t\t\t", layer_str[layer_id], sds_info->fill_val);
	       for (i=0, j=0; i<num_val[layer_id]; i++, j++)
	       {
		 fprintf(stdout, "%d ", sds_values[layer_id][i]);
		 if (j == 9)
	         {
		   if (n_layer == 1) fprintf(stdout, "\n\t\t");
		   else fprintf(stdout, "\n\t\t\t");
		   j = -1;
		 }
	       }
	       fprintf(stdout, "\n");
	     }
	     break;
    case 24: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (n_layer == 1) fprintf(stdout, "\t\t%s%d\n\t\t", layer_str[layer_id], sds_info->fill_val);
	       else fprintf(stdout, "\t\t%s%d\n\t\t\t", layer_str[layer_id], sds_info->fill_val);
	       for (i=0, j=0; i<num_val[layer_id]; i++, j++)
	       {
		 fprintf(stdout, "%d ", sds_values[layer_id][i]);
		 if (j == 9)
	         {
		   if (n_layer == 1) fprintf(stdout, "\n\t\t");
		   else fprintf(stdout, "\n\t\t\t");
		   j = -1;
		 }
	       }
	       fprintf(stdout, "\n");
	     }
	     break;
    case 25: for (layer_id=0; layer_id<n_layer; layer_id++)
	     {
	       if (n_layer == 1) fprintf(stdout, "\t\t%s%d\n\t\t", layer_str[layer_id], sds_info->fill_val);
	       else fprintf(stdout, "\t\t%s%d\n\t\t\t", layer_str[layer_id], sds_info->fill_val);
	       for (i=0, j=0; i<num_val[layer_id]; i++, j++)
	       {
		 fprintf(stdout, "%d ", sds_values[layer_id][i]);
		 if (j == 9)
	         {
		   if (n_layer == 1) fprintf(stdout, "\n\t\t");
		   else fprintf(stdout, "\n\t\t\t");
		   j = -1;
		 }
	       }
	       fprintf(stdout, "\n");
	     }
	     break;
  }
  Free2D((void **)layer_str);
}
	  
int create_color_table(int32 data_type, int **sds_values, float **sds_fvalues, int *num_val, int n_layer,
	int nc, char *in_clr_fname, char *out_clr_fname)
{
  FILE *fp;
  float slope_clr;
  int nvals, tot_num;
  int sds_val, min, max; 
  int slope_range, i, j;
  int dm, in_nc, clr_id, clr_st;
  int out_clr_table[5], in_clr_table[256*3];
  int sds_descr_vals[256];

  for (i=0, tot_num=0; i<n_layer; i++)
  {
    nvals = num_val[i];
    if (data_type == 5)
    {
      for (j=0; j<nvals; j++)
      {
        sds_val = (int)ceil(sds_fvalues[i][j]);
	if (tot_num < 256) sds_descr_vals[tot_num] = sds_val;
        if (tot_num == 0) min = max = sds_val;
        else 
        {
          if (sds_val < min) min = sds_val;
          if (sds_val > max) max = sds_val;
        }
	++tot_num;
      } 
    }
    else
    {
      for (j=0; j<nvals; j++)
      {
        sds_val = sds_values[i][j];
	if (tot_num < 256) sds_descr_vals[tot_num] = sds_val;
        if (tot_num == 0) min = max = sds_val;
        else
        {
          if (sds_val < min) min = sds_val;
          if (sds_val > max) max = sds_val;
        }
        ++tot_num;
      }                               
    }
  } 
  clr_st = 0;
  if (nc == 0) {
    if (tot_num <= MAX_NUM_CLR) { nc = tot_num; clr_st = 1; }
    else nc = MAX_NUM_CLR;
  }
  else {
    if (nc > tot_num) { nc = tot_num; clr_st = 1; }
  }
   
  if ((in_nc = read_clr_table(in_clr_fname, in_clr_table)) != -1)
  {
    fprintf(stderr, "Number of colors output: %d\n", nc);
    if ((fp = fopen(out_clr_fname, "w")) != NULL)
    {
      slope_clr = (float)ceil((float)in_nc/(float)(nc));
      if (clr_st == 0)
      {
        dm = nc;
        for (i=0; i<nc; i++)
        {
          slope_range = (int)ceil((float)(max - min)/(float)dm);
          out_clr_table[0] = (i == 0) ? min : out_clr_table[1] + 1;
          out_clr_table[1] = (i == nc-1) ? max : out_clr_table[0] + slope_range - 1;
          clr_id = (int)ceil(slope_clr*i);
          clr_id *= 3;
          out_clr_table[2] = in_clr_table[clr_id];
          out_clr_table[3] = in_clr_table[clr_id + 1];
          out_clr_table[4] = in_clr_table[clr_id + 2];
          fprintf(fp, "%d %d %d %d %d\n", out_clr_table[0], out_clr_table[1], out_clr_table[2], out_clr_table[3], out_clr_table[4]);
          --dm;
          min = out_clr_table[1];
        }                                            
      }
      else 
      {
        for (i=0; i<nc; i++)
        {
          out_clr_table[0] = out_clr_table[1] = sds_descr_vals[i];
          clr_id = (int)ceil(slope_clr*i);
          clr_id *= 3;
          out_clr_table[2] = in_clr_table[clr_id];
          out_clr_table[3] = in_clr_table[clr_id + 1];
          out_clr_table[4] = in_clr_table[clr_id + 2];
          fprintf(fp, "%d %d %d %d %d\n", out_clr_table[0], out_clr_table[1], out_clr_table[2], out_clr_table[3], out_clr_table[4])
;
	}
      }
    }
    else
    {
      in_nc = -1;
      fprintf(stderr, "Cannot open output file %s\n", out_clr_fname);
    }
  }
  return in_nc;
}


