/****************************************************************************
!C

!File: subset_sds.c

!Description:
  Create spatial subset SDSs of one or more SDS from an MODIS Land HDF-EOS 
  data product.

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision 1.3 - Xiaoping Zhang  January 2009 / 
    Adding grid option, GetHDFEOS will retrives the HDF EOS metadata from a single file 
    	and write HDF EOS metadata to output hdf file if this option is selected.
!Revision 1.2 - Xiaoping Zhang  November 2008. / 
    Extending to 4D subset and create a subset hdf file with grid information and SDS attributes.
!Revision 1.1 - Xiaoping Zhang  September 2008. / 
    Extending to 3D subset and create a subset hdf file with grid information and SDS attributes.

  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
        3)update all HDFEOS API to a higher-level APIs to handle 
          both HDFEOS2 and HDFEOS5 in eos_sds_rw.c
        4) correct a bug when -gname= flag is used in function
           hdf_OpenWriteGridInfo(). The created HDFEOS file must
           be closed before return. See details in the function.
	5) add warning information for -gname= flag when gridname
	   specified by user is different from that in the input file.
        6) for -meta flag. avoid copying StructMetadata.0 because it will
          overwrite the computed correct georeference


!Revision History:
  December 2003. Version 1.0

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

    Yi (Sophyi) Zhang 
    LDOPE                             Science Systems and Applications Inc.
    sophyi@ltpmail.gsfc.nasa.gov      NASA/GSFC Code 922
    phone: 301-614-5497               Greenbelt, MD 20771    

    David Roy 
    LDOPE                             University of Maryland, Department of Geography
    droy@kratmos.gsfc.nasa.gov        NASA/GSFC Code 922 (B32)
    phone: 301-614-5571               Greenbelt, MD 20771   

!Design Notes: (none)

!END
*****************************************************************************/         

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "eos_sds_rw.h"
#include "alloc_mem.h"
#include "str_op.h"
#include "sds_rw.h"
#include "main_util.h"
#include "meta.h"

#define HELP \
"NAME \n" \
"     subset_sds - Subset one or more SDS in an HDF file (granule)\n" \
" \n" \
"SYNOPSIS \n" \
"     subset_sds -help \n" \
"     subset_sds -sds=<sds_names> -row=<st,end> -col=<st,end> -gname=<gridName> -of=<output file> <file> -meta \n" \
" \n" \
"DESCRIPTION \n" \
"     Subset one or more SDS in an input HDF file \n" \
" \n" \
"OPTIONS \n" \
"     -help                          Display this help message \n" \
"     -sds=<sds names>               SDS names in the input file separated by comma \n" \
"     -gname=<gridName>              Grid name in hdf file \n"   \
"     -row=<st,end>                  Subset row range (st and end are inclusive) \n" \
"     -col=<st,end>                  Subset column range (st and end are inclusive) \n" \
"     -of=<out filename>             Output filename \n" \
"     -meta                          copy meta data from <file> to <outfile> \n" \
" \n" \
"Example: \n" \
"subset_sds -col=0,400 -row=0,400 -gname=MODIS_Grid_Daily_Fire -of=testMod14A.hdf MOD14A1.A2000305.h09v05.005.2008210044004.hdf -meta \n" \
"\n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:\tsubset_sds -sds=<sds_names> -row=<st,end> -col=<st,end> -gname=<gridName> -of=<output filename> <file> \n" \
" \n" \
"OPTIONS \n" \
"     -help                          Display this help message \n" \
"     -sds=<sds names>               SDS names in the input file separated by comma \n" \
"     -gname=<gridName>              Grid name in hdf file \n"   \
"     -row=<st,end>                  Subset row range (st and end are inclusive) \n" \
"     -col=<st,end>                  Subset column range (st and end are inclusive) \n" \
"     -of=<out filename>             Output filename \n" \
"     -meta                          copy meta data from <file> to <outfile> \n" \
" \n"


/**************************************************************************************************
                            		Prototypes.
**************************************************************************************************/
int parse_cmd_subset_sds(int argc, char **argv, char **sds_names, char *HDFEOSgridname, int *nsds, int *row_range,
                int *col_range, char *out_fname, char *in_fname, int *m_opt, int *g_opt);

void subset_an_sds(sds_t *in_sds_info, sds_t *out_sds_info, int *row_range, int *col_range);
int hdf_OpenWriteGridInfo(char *filename, char *HDFEOSgridname, float64 *upleftpt, float64 *lowrightpt, 
    int32 xdimsize, int32 ydimsize, int32 projcode, int32 spherecode, int32 zonecode, 
    float64 *projparm, sds_t *sds_info, int nsds, int H5OUT); 
    
    
int main(int argc, char *argv[])
/******************************************************************************
!C

!Description:
  Main function for subset_sds.

!Input Parameters: (none)
  command line arguments: see help for details.

!Output Parameters: (none)
  return 0 on successful completion of the process

!Revision History:
  See file prologue..

!Team-unique Header:
  See file prologue.

!References and Credits: (see file prologue)

!Design Notes: (none)

!END
********************************************************************************/
{
  sds_t *in_sds_info;
  sds_t out_sds_info;
  sds_t *sds_info;
  int i, st, nsds, isds, m, n;
  int m_opt = 0, g_opt = 0, rank;
  int row_range[5], col_range[5];
  int nattrs = 22;  
  int32 out_dim_id, in_dim_id;
  int32 n_values, data_type, n_attrs;
  char  in_dim_name[64];
  char **sds_names;
  char in_fname[MAX_PATH_LENGTH];
  char out_fname[MAX_PATH_LENGTH];
  int DFACC_MODE;
  char HDFEOSgridname[MAX_PATH_LENGTH]; 
  char gridname[MAX_PATH_LENGTH];  /* MSZ 4/4/2018 */
  grid_mapinfo *hdfeos, *h;
 
  int32 tdimsize = 0, xdimsize, ydimsize, dimSize;
  int32 projCode, sphereCode, zoneCode, originCode; 
  
  int32 ns, nl, Xdimsize, Ydimsize, ProjCode, SphereCode, ZoneCode, OriginCode;  
  float64 incx, incy, ulCornerGrid[2], lrCornerGrid[2], ulCorner[2], lrCorner[2];  
  float64 *upleftpt, *lowrightpt;    
  float64 *projparm;
  
  hdfeos = (grid_mapinfo *)calloc(1, sizeof(grid_mapinfo));
  upleftpt = (float64 *)calloc(2, sizeof(float64));
  lowrightpt = (float64 *)calloc(2, sizeof(float64));
  projparm = (float64 *)calloc(13, sizeof(float64));
  sds_info = (sds_t *)calloc(nsds, sizeof(sds_t));
  
  
  if (argc == 1)
    {
      st = -1;
      fprintf(stderr, "Missing input file \n");
      fprintf(stderr, "%s\n", USAGE);
      exit(EXIT_FAILURE);
    }   
  
  if ((argc==2) && (strcmp(argv[1], "-help")==0))
    {
      fprintf(stderr, "%s\n", HELP);
      exit(0);
    }
  
  /* MSZ 2/10/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 2/10/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  /*  Display SDS names of input HDF file */
  
  if ((argc>=3) && (strcmp(argv[1], "-help")==0))
    {
      /* Print all the SDS names in the input file */
      for (i=2; i<argc; i++)
        {
          if (argv[i][0] != '-')
              display_sds_info_of_file(argv[i]);
        } 
      exit(EXIT_SUCCESS);
    }

  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for sds_names in subset_sds: main()\n");
      exit(EXIT_FAILURE);
    }
  else
    {

      if ((st = parse_cmd_subset_sds(argc, argv, sds_names, HDFEOSgridname, &nsds, 
           row_range, col_range, out_fname, in_fname, &m_opt, &g_opt)) == -1)
	fprintf(stderr, "%s\n", USAGE);    
/*      if ((st = parse_cmd_subset_sds(argc, argv, sds_names, &nsds, row_range, col_range, 
				     out_fname, in_fname)) == -1)
	fprintf(stderr, "%s\n", USAGE);  */
      else if (st != 0)
	{
	  if (g_opt == 1)
	  {	  
	    /* MSZ 4/4/2018 */
	    if(get_grid_name(in_fname, gridname) == FAIL)
	    {
		printf("fail to get gridname, %s is not HDFEOS, flag -gname= should not be used, exit...\n", in_fname);
		exit(0);
	    } 
	    else
	    {
	      if(strcmp(gridname, HDFEOSgridname) !=0)
	      {
		fprintf(stderr,"########################################################################################\n");
		fprintf(stderr,"Warning!!! %s should be used as gridname not %s for flag -gname=, \n-999.0 will be used as georeference parameters in StructMetadata.0\n", gridname, HDFEOSgridname);
		fprintf(stderr,"########################################################################################\n");
	      }
	    }

	    GetHDFEOS(in_fname, HDFEOSgridname, hdfeos); 			       

	    h = hdfeos;
	    ns = h->ns;
	    nl = h->nl;
	    ulCornerGrid[0] = h->ulCornerGrid[0];
	    ulCornerGrid[1] = h->ulCornerGrid[1];
	    lrCornerGrid[0] = h->lrCornerGrid[0];
	    lrCornerGrid[1] = h->lrCornerGrid[1];
	    projCode = h->projCode;
	    sphereCode = h->sphereCode;
	    zoneCode = h->zoneCode;
	    originCode = h->originCode;  
	    for (i=0; i<13; i++) {
	      projparm[i] = h->projParameters[i];
	    }       
	    incx = (lrCornerGrid[0] - ulCornerGrid[0])/((float64)ns);
	    incy = (lrCornerGrid[1] - ulCornerGrid[1])/((float64)nl);
	    ulCorner[0] = ulCornerGrid[0] + incx * (float64)col_range[0];
	    lrCorner[0] = ulCornerGrid[0] + incx * (float64)(col_range[1]+1);
	    if (sphereCode == (int32)(-1)) {
	      ulCorner[1] = ulCornerGrid[1] + incy * (float64)row_range[0];
	      lrCorner[1] = ulCornerGrid[1] + incy * (float64)(row_range[1]+1);     
	    } 
	    else {
	      ulCorner[1] = lrCornerGrid[1] - incy * (float64)(row_range[1]+1);
	      lrCorner[1] = lrCornerGrid[1] - incy * (float64)row_range[0];       
	    }    
	    upleftpt[0] = ulCorner[0];
	    upleftpt[1] = ulCorner[1];
	    lowrightpt[0] = lrCorner[0];
	    lowrightpt[1] = lrCorner[1];     
	  }
	  
	    xdimsize = (int32)(col_range[1] - col_range[0] + 1);
	    ydimsize = (int32)(row_range[1] - row_range[0] + 1);	  
	  		
	  if (nsds == 0)
	    nsds = get_sds_names(in_fname, sds_names);
          in_sds_info = (sds_t *)calloc(nsds, sizeof(sds_t));

	  in_sds_info[0].sd_id = -1;
	  if (in_sds_info[0].sd_id == -1)
	  {
	    if(open_sd(in_fname, &in_sds_info[0], 'R') == FAIL)
	    {
		fprintf(stderr, "Cannot open the HDF file %s\n", in_fname);
		return -1;
	    }
 	    else
	    {			
              for (isds=0; isds<nsds; ++isds) 
              {
        	strcpy(in_sds_info[isds].name, sds_names[isds]);
		in_sds_info[isds].file_name = in_fname;
		in_sds_info[isds].sd_id = in_sds_info[0].sd_id;
		in_sds_info[isds].H5IN = in_sds_info[0].H5IN;
        	in_sds_info[isds].sds_id = -1;
        	in_sds_info[isds].sds_index = -1;
        	if (get_sds_info(in_fname, &in_sds_info[isds]) == -1)
        	{	      
		  fprintf(stderr, "Cannot get sds info for %s\n", in_sds_info[isds].name);
		  return -1;
		}
	      } 	   
            }
	  }	  

	  /* open output HDF file and set global attribute input pointer
	     to contain names of all input files */
	  out_sds_info.H5OUT = 0;  /* so far, output to HDF4 */
	  out_sds_info.H5IN = out_sds_info.H5OUT;  /* this is a must because the empty output HDF is opened for sds_info inquiring */
	  out_sds_info.sd_id = -1;
	  out_sds_info.file_name = out_fname;

	  if (g_opt == 1) {
            hdf_OpenWriteGridInfo(out_fname, HDFEOSgridname, upleftpt, lowrightpt, xdimsize, ydimsize, 
                        projCode, sphereCode, zoneCode, projparm, in_sds_info, nsds, out_sds_info.H5OUT);		
	  }

          fprintf(stderr, "g_opt: %d\n", g_opt); 
	  if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
	  {
	    fprintf(stderr, "Cannot create the output hdf file %s\n", out_fname);
	    exit(EXIT_FAILURE);
	  }
	  else 
	    {
	      if (m_opt == 1)
		/* MSZ 4/4/2018. note that this function will not copy StructMetadata.0 to the output 
                 * file so that to computed georeference in the output StructMetadata.0 not be overwritten */
                writeMetadata(in_sds_info[0].H5IN, in_sds_info[0].sd_id, out_sds_info.H5OUT, out_sds_info.sd_id);

              fprintf(stderr, "%d SDS below to be processed:\n", nsds);	      	    
	      for (isds=0; isds<nsds; ++isds)
		{
		  fprintf(stdout, "Processing SDS %s\n", sds_names[isds]);
		 
		  get_sdsname_dim(sds_names[isds], in_sds_info[isds].name, &n, &m);
/*		  fprintf(stdout, "in_sds_info[%d].name: %s\n",isds, in_sds_info[isds].name); */

		  if ((n != -1) || (m != -1))
		    fprintf(stdout, "A 2D slice of the 3D/4D SDS can't be selected to output\n");		    		    		  
		  out_sds_info.sds_id = -1;	    		    	
        	  /*strcpy(out_sds_info.name, in_sds_info[isds].name); */
		  sprintf(out_sds_info.name, "Subset_%s", in_sds_info[isds].name); 
		  
		  if (g_opt == 1) {				
		    if(open_sds((char*)NULL, &out_sds_info, 'R') == FAIL)
		    {
		      fprintf(stderr, "Cannot open the SDS %s \n", out_sds_info.name);
		      return -1;
		    }			      
	      	  }     
		  else {
		    rank = out_sds_info.rank = in_sds_info[isds].rank;
		    out_sds_info.data_type = in_sds_info[isds].data_type;
		    for (i = 0; i < rank; i++) { 
		      out_sds_info.dim_size[i] = in_sds_info[isds].dim_size[i];
		    }
		    if (in_sds_info[isds].dim_size[0] < in_sds_info[isds].dim_size[rank-1]) {    
		      out_sds_info.dim_size[rank-2] = ydimsize;
		      out_sds_info.dim_size[rank-1] = xdimsize; 
		    } else {
		      out_sds_info.dim_size[0] = ydimsize;
		      out_sds_info.dim_size[1] = xdimsize;    
		    }	    		          		    

		    if(open_sds((char*)NULL, &out_sds_info, 'W') == FAIL)
		    {
        	      fprintf(stderr, "Cannot create the SDS %s\n", out_sds_info.name);
        	      return -1;
		    }	

		    #if 0  /* MSZ 2/10/2018 marco if out the below */
		    /* Name the Dimensions */
		    for (i=0; i<rank; i++)
		    {
        		out_dim_id = SDgetdimid(out_sds_info.sds_id, i);
        		in_dim_id  = SDgetdimid(in_sds_info[isds].sds_id, i);
        		if (SDdiminfo(in_dim_id, in_dim_name, &n_values, &data_type, &n_attrs) == FAIL)
        		   fprintf(stderr, "Could not Retrieve Dimension Info from Input\n");
        		if (SDsetdimname(out_dim_id, in_dim_name) == FAIL)
        		   fprintf(stderr, "Could not write dimensions names\n");
		    }		    	  
		    #endif
		  }    
		   	      
		  subset_an_sds(&in_sds_info[isds], &out_sds_info, row_range, col_range); 
		 
		  write_all_sds_attr(in_sds_info[isds].H5IN, in_sds_info[isds].sds_id, out_sds_info.H5OUT, out_sds_info.sds_id, nattrs); 
		 
		  close_sds(&out_sds_info, out_sds_info.H5OUT); 
		  close_sds(&in_sds_info[isds], in_sds_info[isds].H5IN);
		}
		/* close input and output file just once respectivley */
	        close_sd(&in_sds_info[0], in_sds_info[0].H5IN);
	        close_sd(&out_sds_info, out_sds_info.H5OUT);
	    }
	}
      Free2D((void **)sds_names);
      fprintf(stderr, "Processing done ! \n");
    }
  return 0;
}

int parse_cmd_subset_sds(int argc, char **argv, char **sds_names, char *HDFEOSgridname, int *nsds, int *row_range, 
		int *col_range, char *out_fname, char *in_fname, int *m_opt, int *g_opt)
{
  int i, cnt, p1;
  int st = 1;
  int len, pos1;
  char **tmp_str, *tmpName;

  if ((tmp_str = (char **)Calloc2D(5, 20, sizeof(char))) == NULL)
  {
    st = 0;
    fprintf(stderr, "Cannot allocate memory for tmp_str in parse_cmd_subset_sds\n");
  }
  else 
  {
    *nsds = 0;
    in_fname[0] = out_fname[0] = HDFEOSgridname[0] = '\0';
    row_range[0] = row_range[1] = 0;
    col_range[0] = col_range[1] = 0;
    for (i=1; i<argc; i++)
    {
      cnt = 0;
      if (is_arg_id(argv[i], "-sds=") == 0)
        get_arg_val_arr(argv[i], sds_names, nsds);
      else if (is_arg_id(argv[i], "-gname=") == 0) {  
        get_arg_val(argv[i], HDFEOSgridname); 
	*g_opt = 1;
      }     
      else if (strcmp(argv[i], "-meta") == 0) 
        *m_opt = 1;	
      else if (is_arg_id(argv[i], "-of=") == 0)  
        get_arg_val(argv[i], out_fname);      
      else if (is_arg_id(argv[i], "-row=") == 0) {
        get_arg_val_arr(argv[i], tmp_str, &cnt);
	if (cnt != 2) {
	  st = -1;
	  fprintf(stderr, "Invalid value for option -row: %s\n", argv[i]);
	}
	else {
	  row_range[0] = (int)atoi(tmp_str[0]);
	  row_range[1] = (int)atoi(tmp_str[1]);
        }
      }
      else if (is_arg_id(argv[i], "-col=") == 0) {
        get_arg_val_arr(argv[i], tmp_str, &cnt);
	if (cnt != 2) {
	  st = -1;
	  fprintf(stderr, "Invalid value for option -col: %s\n", argv[i]);
	}
	else {
	  col_range[0] = (int)atoi(tmp_str[0]);
	  col_range[1] = (int)atoi(tmp_str[1]);
        }
      }
      else if (argv[i][0] == '-') 
        fprintf(stderr, "Ignoring unknown option %s\n", argv[i]);
      else strcpy(in_fname, argv[i]);
    }
    if (in_fname[0] == '\0') 
    {
      st = -1;
      fprintf(stderr, "Missing input filename . . \n");
    }
    if (out_fname[0] == '\0') 
    {
      len = (int)strlen(in_fname);
      pos1 = strpos(in_fname, ".hdf", 0);
      if (pos1 != -1) {      
	strmid(in_fname, 0, len-4, out_fname);
	sprintf(out_fname, "%s_subset.hdf", out_fname);
      }		
      else {
        st = -1;
        fprintf(stderr, "Failing to create output filename!\n");
      }                
    } 
       
/*    if (HDFEOSgridname[0] == '\0')
    {
      st = -1;
      fprintf(stderr, "Missing input gridName! \n"); 
    }  */ 

    if ((row_range[0] >= row_range[1]) || (row_range[0] < 0) || (row_range[1] < 0)) 
    {
      st = -1;
      fprintf(stderr, "Invalid subset row range . . \n");
    }
    if ((col_range[0] >= col_range[1]) || (col_range[0] < 0) || (col_range[1] < 0)) 
    {
      st = -1;
      fprintf(stderr, "Invalid subset column range . . \n");
    }
  }
  return st;
}

void subset_an_sds(sds_t *in_sds_info, sds_t *out_sds_info, int *row_range, int *col_range)
/*
!C******************************************************************************

!Function: subset_an_sds
       
!Description:

  create a subset of the original SDS to output. Original interleaving is preserved.

!Input Parameters: 
  in_sds_info    Array of input SDS information structure
  out_sds_info   Array of output SDS information structure
  row_range      Row range of input SDS to subset
  col_range      Column range of input SDS to subset

!Revision History: (see file prolog) 

!Team-unique Header: (see file prolog)
 
!References and Credits: (see file prolog)

!Design Notes: (none)

!END
*****************************************************************************/
{
  int i, k, irow, ncols, nn, in, nm, im; 
  int rank, bsq, st_c, offset;
  int32 in_start[4] = {0, 0, 0, 0};
  int32 in_edge[4] = {0, 0, 0, 0};
  int32 out_start[4] = {0, 0, 0, 0};
  int32 out_edge[4] = {0, 0, 0, 0};
  int32 attr_type, attr_cnt;
  void *sds_data, *sds_data1, *attr_val;
  int n, m, j, nrows, icol, ic;
  int start_row, start_col, end_row, end_col;
  int ndata_sm, ndata_sm1;

  rank = in_sds_info->rank;

  n = m = -1;  /* no layer can be selected */
  
  /*create output SDS with proper dimension and other attributes */
  if ((attr_val = (void *)get_sds_attr(in_sds_info, "_FillValue", &attr_type, &attr_cnt)) == NULL)
    fprintf(stderr, "Attribute "_FillValue" not defined for output\n");
  
  bsq = ((rank == 2) || (in_sds_info->dim_size[0] < in_sds_info->dim_size[rank-1])) ? 1 : 0;
  
  /* initialize the dimension size of output sds */
  for (j=0; j<4; j++)
    out_sds_info->dim_size[j] = 0;

  for (j=0; j<rank; j++)
    out_sds_info->dim_size[j] = in_sds_info->dim_size[j];

  start_row = row_range[0] < row_range[1] ? row_range[0] : row_range[1];
  end_row = row_range[1] > row_range[0] ? row_range[1] : row_range[0];
  start_col = col_range[0] < col_range[1] ? col_range[0] : col_range[1];
  end_col = col_range[1] > col_range[0] ? col_range[1] : col_range[0];

  nrows = end_row - start_row + 1;
  ncols = end_col - start_col + 1;
  ndata_sm = ncols;
  
  if (rank == 2)
    {
      if ((row_range[0] > in_sds_info->dim_size[0]) ||
	  (row_range[1] > in_sds_info->dim_size[0]) ||
	  (col_range[0] > in_sds_info->dim_size[1]) ||
	  (col_range[1] > in_sds_info->dim_size[1]))
	{
	  fprintf(stderr, "Input subsetting range is incorrect\n");
	  exit(EXIT_FAILURE);
	}
    }
  else if (rank > 2)
    {      
      if ((n == -1) && (m == -1))
	{
	  if (bsq == 0)
	    {
	      if ((row_range[0] > in_sds_info->dim_size[0]) ||
		  (row_range[1] > in_sds_info->dim_size[0]) ||
		  (col_range[0] > in_sds_info->dim_size[1]) ||
		  (col_range[1] > in_sds_info->dim_size[1]))
		{
		  fprintf(stderr, "Input subsetting range is incorrect\n");
		  exit(EXIT_FAILURE);
		}    	      
	    }
	  else
	    {
	      if ((row_range[0] > in_sds_info->dim_size[rank-2]) ||
		  (row_range[1] > in_sds_info->dim_size[rank-2]) ||
		  (col_range[0] > in_sds_info->dim_size[rank-1]) ||
		  (col_range[1] > in_sds_info->dim_size[rank-1]))
		{
		  fprintf(stderr, "Input subsetting range is incorrect\n");
		  exit(EXIT_FAILURE);
		}  		 
	    }
	}
    }

  out_sds_info->data_type = in_sds_info->data_type;
  out_sds_info->data_size = in_sds_info->data_size;
  out_sds_info->fill_val = in_sds_info->fill_val;
  out_sds_info->rank = in_sds_info->rank;
  
  if ((sds_data = (void *)calloc(ndata_sm, out_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_in in subset_an_sds \n");  
     	      
  if (attr_val != NULL)
    {
      if(set_sds_attr(out_sds_info->H5OUT, out_sds_info->sds_id, "_FillValue", attr_type, 1, (VOIDP)attr_val) == FAIL) 
	fprintf(stderr, " Cannot write sds attribute _FillValue in subset_an_sds \n");
      free(attr_val);
    }

  if (rank == 2)
    {    
      if ((sds_data = (void *)calloc(ncols, in_sds_info->data_size)) == NULL)
	 fprintf(stderr, "Cannot allocate memory for sds_data in subset_an_sds\n");    
      out_edge[0] = in_edge[0] = 1;
      out_edge[1] = in_edge[1] = ncols;  
      in_start[1] = col_range[0];
      out_start[1] = 0;    
      out_sds_info->dim_size[0] = nrows;
      out_sds_info->dim_size[1] = ncols;    
      for (irow = start_row; irow <end_row +1; irow++)
	{	      
	  in_start[0] = irow;
	  out_start[0] = irow - row_range[0]; 
	  if (read_sel_sds_data(in_sds_info, in_start, in_edge, sds_data) == FAIL) {
	    fprintf(stderr, "Error reading data line %d from SDS %s\n", irow, in_sds_info->name);
	    break;
	  }
	  else 
	  {
	    if(set_sel_sds_data(out_sds_info, out_start, out_edge, sds_data) == FAIL) {
              fprintf(stderr, "Error writing data line %d to SDS %s\n", out_start[0], out_sds_info->name);
              break;
	    }
	  }	      
	}
     }
   else if (rank == 3)
     {	    
       if (bsq == 1)
	 {
	   if ((sds_data = (void *)calloc(ncols, in_sds_info->data_size)) == NULL)
	     fprintf(stderr, "Cannot allocate memory for sds_data in subset_an_sds\n");
	   out_edge[0] = in_edge[0] = 1;
	   out_edge[1] = in_edge[1] = 1;  
	   out_edge[2] = in_edge[2] = ncols;
	   in_start[2] = col_range[0];
	   out_start[2] = 0;	 
	   nn = out_sds_info->dim_size[0];
	   for (in=0; in<nn; in++)
	   {
	     in_start[0] = in;  
	     out_start[0] = in;
	     for (irow=row_range[0]; irow<=row_range[1]; irow++)
	     {
	       in_start[1] = irow;      
	       out_start[1] = irow - row_range[0]; 
	       if (read_sel_sds_data(in_sds_info, in_start, in_edge, sds_data) == FAIL) {
		 fprintf(stderr, "Error reading data from SDS %s\n", in_sds_info->name);
		 break;
	       }
	       else 
	       {
		 if(set_sel_sds_data(out_sds_info, out_start, out_edge, sds_data) == FAIL) {
        	   fprintf(stderr, "Error writing data to SDS %s\n", out_sds_info->name);
        	   break;
		 }
	       }
	     } /* for (irow=0;  . . .) */
	   }/* for (in=0;  . . .) */	
	 }
       else
	 {
	   nn = out_sds_info->dim_size[2];	 
	   out_edge[0] = in_edge[0] = 1;
	   out_edge[1] = in_edge[1] = 1;  
	   out_edge[2] = in_edge[2] = nn;	 
	 		  
	   if ((sds_data1 = (void *)calloc(nn, out_sds_info->data_size)) == NULL)
	     fprintf(stderr, "Cannot allocate memory for data_in in subset_an_sds \n");	

	   in_start[2] = 0;
           out_start[2] = 0;	  		
	   for (irow=row_range[0]; irow<=row_range[1]; irow++)
	   {
	     in_start[0] = irow;      
	     out_start[0] = irow - row_range[0]; 
	     for (icol=col_range[0]; icol<=col_range[1]; icol++)
	     {	
	       in_start[1] = icol;  
	       out_start[1] = icol - col_range[0];		      	    
	       if (read_sel_sds_data(in_sds_info, in_start, in_edge, sds_data1) == FAIL) {
		 fprintf(stderr, "Error reading data from SDS %s\n", in_sds_info->name);
		 break;
	       }
	       else 
	       {
		 if(set_sel_sds_data(out_sds_info, out_start, out_edge, sds_data1) == FAIL) {
        	   fprintf(stderr, "Error writing data to SDS %s\n", out_sds_info->name);
        	   break;
		 }
	       }
	     }
	   } /* for (irow=0;  . . .) */	
	   if (sds_data1 != NULL) free(sds_data1);	
	 }                	    
     }
   else if (rank == 4)
     {
       if (bsq == 1)
	 {
	   if ((sds_data = (void *)calloc(ncols, in_sds_info->data_size)) == NULL)
	     fprintf(stderr, "Cannot allocate memory for sds_data in subset_an_sds\n");
	   out_edge[0] = in_edge[0] = 1;
	   out_edge[1] = in_edge[1] = 1;  
	   out_edge[2] = in_edge[2] = 1;
	   out_edge[3] = in_edge[3] = ncols;	   
	   in_start[3] = col_range[0];
	   out_start[3] = 0;	 
	   nn = out_sds_info->dim_size[0];
	   nm = out_sds_info->dim_size[1];
	   for (in=0; in<nn; in++)
	   {
	     in_start[0] = in;  
	     out_start[0] = in;
	     for (im=0; im<nm; im++)
	     {
	       in_start[1] = im;  
	       out_start[1] = im;	       
	       for (irow=row_range[0]; irow<=row_range[1]; irow++)
	       {
		 in_start[2] = irow;      
		 out_start[2] = irow - row_range[0]; 
		 if (read_sel_sds_data(in_sds_info, in_start, in_edge, sds_data) == FAIL) {
		   fprintf(stderr, "Error reading data from SDS %s\n", in_sds_info->name);
		   break;
		 }
		 else 
		 {
		   if(set_sel_sds_data(out_sds_info, out_start, out_edge, sds_data) == FAIL) {
        	     fprintf(stderr, "Error writing data to SDS %s\n", out_sds_info->name);
        	     break;
		   }
		 }
	       } /* for (irow=0;  . . .) */
	     } /* for (im=0;  . . .) */
	   }/* for (in=0;  . . .) */	
	 }
       else
	 {
	   nn = out_sds_info->dim_size[2];
	   nm = out_sds_info->dim_size[3];	 
	   out_edge[0] = in_edge[0] = 1;
	   out_edge[1] = in_edge[1] = 1;  
	   out_edge[2] = in_edge[2] = 1;
	   out_edge[3] = in_edge[3] = nm;	 
	 		  
	   if ((sds_data1 = (void *)calloc(nm, out_sds_info->data_size)) == NULL)
	     fprintf(stderr, "Cannot allocate memory for data_in in subset_an_sds \n");		     	    
	   
	   in_start[3] = 0;
           out_start[3] = 0;	  		
	   for (irow=row_range[0]; irow<=row_range[1]; irow++)
	   {
	     in_start[0] = irow;      
	     out_start[0] = irow - row_range[0]; 
	     for (icol=col_range[0]; icol<=col_range[1]; icol++)
	     {	
	       in_start[1] = icol;  
	       out_start[1] = icol - col_range[0];	     
	       for (in=0; in<nn; in++)
	       {
		 in_start[2] = in;
        	 out_start[2] = in; 	       		      	    
		 if (read_sel_sds_data(in_sds_info, in_start, in_edge, sds_data1) == FAIL) {
		   fprintf(stderr, "Error reading data from SDS %s\n", in_sds_info->name);
		   break;
		 }
		 else 
		 {
		   if(set_sel_sds_data(out_sds_info, out_start, out_edge, sds_data1) == FAIL) {
        	     fprintf(stderr, "Error writing data to SDS %s\n", out_sds_info->name);
        	     break;
		   }
		 }
	       } /* for (in=0;  . . .) */
	     } /* for (icol=0;  . . .) */	
	   } /* for (irow=0;  . . .) */	 
	   if (sds_data1 != NULL) free(sds_data1);	   	
	 }              
     }	       	      	      
  if (sds_data != NULL) free(sds_data);      	  
}

int hdf_OpenWriteGridInfo(char *filename, char *HDFEOSgridname, float64 *upleftpt, float64 *lowrightpt, 
    int32 xdimsize, int32 ydimsize, int32 projcode, int32 spherecode, int32 zonecode, 
    float64 *projparm, sds_t *sds_info, int nsds, int H5OUT)    
    
/*
!C******************************************************************************
!Description: 
	Opens a new HDF file for write access.

!Input Parameters:
 filename       name of subset output file to open
 upleftpt       location of upper left corner pixel center (meters)
 lowrightpt     location of lower right corner pixel center (meters)
 projcode       map projection type code
 spherecode     sphere code; typically -1
 zoneCode       map projection zone code
 projparm	projection parameters
 data_types	data type for each sds
 sds_info	sds_info from original file
 xdimsize	col numbers of subset
 ydimsize	row numbers of subset
 HDFEOSgridname grid name("MODIS_Grid_Daily_Fire", "MOD_Grid_500m_Surface_Reflectance")

!Output:        out subset file with new geolocation infomation for global meta date 
		and subset data.

 (returns)      status  
		  0: failure;  1: success
                
!END
******************************************************************************/
{
  int good_status;
  int p1, k, k1, isds, len, rank;
  int nattrs = 22; 
  hid_t gfid, gid;
  int32 sds_index, data_type, dim_size[4]; 
  char dimstring[255], param_name[255];
  int32 i; 
  int he5_fmt;
  
  char **dim_names, **short_dim_names;

  he5_fmt = H5OUT;

  if ((dim_names = (char **)Calloc2D(sds_info->rank, MAX_NAME_LENGTH, sizeof(char))) 
      == NULL)
    fprintf(stderr, "Cannot allocate memory for dim_names\n");

  if ((short_dim_names = (char **)Calloc2D(sds_info->rank, MAX_NAME_LENGTH, sizeof(char))) 
      == NULL)
    fprintf(stderr, "Cannot allocate memory for short_dim_names\n");    
  
/*  
  char *HDFEOSgridname = "MODIS_Grid_Daily_Fire" (3D)
  char *HDFEOSgridname = "MOD_Grid_500m_Surface_Reflectance" (2D) 
  char *HDFEOSgridname = "MOD_Grid_Aggregation"  (4D)
*/
 
  HDF_CHUNK_DEF cdef;
  
  if ((gfid = EOS_GDopen(filename, 'C', he5_fmt)) == FAIL)
  {
    fprintf(stderr, "can't open file: %s (EOS_GDopen)!\n",filename); 
    return -1;
  }

  if ((gid = EOS_GDcreate(gfid, HDFEOSgridname, xdimsize, ydimsize, upleftpt, lowrightpt, he5_fmt)) == FAIL)
  { 
    fprintf(stderr, "can't create grid id from file: %s (EOS_GDcreate)!\n",filename);   
    if (EOS_GDclose(gfid, he5_fmt) == FAIL)
      fprintf(stderr, "can't close gfid for file: %s (EOS_GDclose)!\n",filename); 
    return -1;
  }
  
  good_status = 1;
  if (EOS_GDdefproj(gid, projcode, zonecode, spherecode, projparm, he5_fmt)==FAIL)
  {
    good_status = 0;
    fprintf(stderr, "can't define projection for file %s (EOS_GDdefproj)!\n",filename); 
    return -1;
  }
  
  /* Define the dimensions (YDim and XDim are implicitly created by EOS_GDCreate). */   
  for (isds = 0; isds < nsds; isds++) {  
    rank = sds_info[isds].rank;  
    get_sds_dim_name(&sds_info[isds], dim_names, short_dim_names);      
    strcpy(dimstring,"\0");
    for (i = 0; i < rank; i++) { 
      strcat(dimstring, short_dim_names[i]); 
      if (i != rank-1) strcat(dimstring, ","); 
      dim_size[i] = sds_info[isds].dim_size[i];
    }
    if (sds_info[isds].dim_size[0] < sds_info[isds].dim_size[rank-1]) {    
      dim_size[rank-2] = ydimsize;
      dim_size[rank-1] = xdimsize; 
    } else {
      dim_size[0] = ydimsize;
      dim_size[1] = xdimsize;    
    }        
    
/*    fprintf(stdout, "dimstring:%s\n", dimstring);   */    
    for (i = 0; i < rank; i++) {                      
      if (good_status){
	if (EOS_GDdefdim(gid, short_dim_names[i], dim_size[i], he5_fmt) == FAIL) {
            fprintf(stderr, "can't define dimension for file %s (EOS_GDdefproj)!\n",filename);
            good_status = 0;
	}
      } 
    }

   /* Define the SDSs all in one grid */    
      sprintf(param_name,"Subset_%s", sds_info[isds].name); 
/*      fprintf(stdout, "param_name[%d]:%s\n", isds,param_name);  */
    data_type = sds_info[isds].data_type;  
    if (good_status) 
    {     
      if (EOS_GDdeffield(gid, param_name, dimstring, data_type, he5_fmt) == FAIL)        
      {
	fprintf(stdout, "HDFE_NOMERGE:%d\n",HDFE_NOMERGE);
	fprintf(stderr, "can't define field for file: %s (EOS_GDdeffield)!\n",filename); 
	good_status = 0;
      }
    }
  }
  
  /* Detach from the grid and close the file */
 
  if (EOS_GDdetach(gid, he5_fmt) == FAIL)
  {
    fprintf(stderr, "can't detach gid for file: %s (EOS_GDdetach)!\n",filename); 
    good_status = 0;
    return -1;
  }

  /* MSZ 4/4/2018, correct a bug by previous developers, which use "else if" not if here. 
   * This empty HDFEOS file must be closed before inquiring in the main code using open_sds(NULL...) */
  if (EOS_GDclose(gfid, he5_fmt) == FAIL)
  {
    fprintf(stderr, "can't close gfid for file: %s (EOS_GDclose)!\n",filename); 
    good_status = 0;
    return -1;
  }
  return 1;
}


