/****************************************************************************
!C

!File: transpose_sds.c

!Description:
	Transpose SDS in an HDF file

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original May/June 2002 
  Modified Mar 2016, Xiaoping Zhang, SSAI
  	to transpose sds in floating data type.
  Modified January, 2018 Version 3.5.1 by Maosheng Zhao, SSAI
        1)modify it so that HDF5 can be read
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "meta.h"
#include "hdf5.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "str_op.h"
#include "qa_tool.h"
#include "main_util.h"
#include "alloc_mem.h"

#define HELP \
"NAME\n" \
"     transpose_sds - Transpose SDS in an HDF file \n" \
"\n" \
"SYNOPSIS\n" \
"     transpose_sds -help\n" \
"     tranpose -sds=[SDS1[,SDS2[,.]]] -of=output_file [-meta] filename\n" \
"\n" \
"     echo -sds=[SDS1[,SDS2[,.]]] -of=output_file [-meta] filename | tranpose\n" \
"\n" \
"DESCRIPTION\n" \
"     Transpose one or more SDS in an HDF file. Output is an HDF file containing \n" \
"one or more transposed SDS. The tool supports 3D and 4D SDSs. All SDSs are output \n" \
"in the original dimension and format. i.e. a 3D/4D SDS is output as 3D/4D SDS. A \n" \
"2D slice of a 3D/4D SDS cannot be selected to output. SDS attributes are copied \n" \
"from input to output. User can optionally select to output the metadata. \n" \
"\n" \
"OPTIONS\n" \
"     -help                          Display this help message\n" \
"     -sds=<SDS names>               SDS names separated by comma\n" \
"     -of=filename                   Output file\n" \
"     -meta                          Copy metadata from input file to output \n" \
" \n" \
"Examples: \n" \
"\ttranspose_sds -sds='500m Surface Reflectance Band 1,500m Surface Reflectance Band 3,500m Surface Reflectance Band 4' MOD09.A2002123.0040.003.2002125174437.hdf -of=test.hdf -meta \n" \
"\ttranspose_sds -sds=Cloud_Mask,Quality_Assurance -of=test.hdf MOD35_L2.A2002123.0040.003.2002124023706.hdf -meta\n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"Usage:      transnose -sds=[SDS1[,SDS2[,.]]] -of=output_file [-meta] filename\n" \
" \n" \
"OPTIONS \n" \
"     -help                          Display this help message\n" \
"     -sds=<SDS names>               SDS names separated by comma\n" \
"     -of=filename                   Output file\n" \
"     -meta                          Copy metadata from input file to output \n"

int parse_cmd_transpose_sds(int argc, char **argv, char **sds_names, int *nsds, char *in_fname, char *out_fname, int *m_opt);
void transpose_an_sds(sds_t *in_sds_info, sds_t *out_sds_info);

int main(int argc, char *argv[])
{
  int isds, nsds, m_opt, i;
  sds_t in_sds_info, out_sds_info;
  char **sds_names;
  char in_fname[MAX_PATH_LENGTH];
  char out_fname[MAX_PATH_LENGTH];

  /* MSZ 1/26/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 1/26/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
    fprintf(stderr, "%s\n", HELP);
  else
  {
    if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
      fprintf(stderr, "Error allocating memory for sds_names in transpose_sds: main()\n");
    else 
    {
      if (parse_cmd_transpose_sds(argc, argv, sds_names, &nsds, in_fname, out_fname, &m_opt) == -1)
        fprintf(stderr, "%s\n", USAGE);
      else if (nsds == 0)
	fprintf(stderr, "No SDS to process . . . \n");
      else
      {
        /* initialize as -1 for sd_id and sds_id */
        in_sds_info.sd_id = in_sds_info.sds_id = -1;

        out_sds_info.H5OUT= 0;  /* so far, output to HDF4 */
        out_sds_info.sd_id = out_sds_info.sds_id = -1;

	if(open_sd(in_fname, &in_sds_info, 'R') == FAIL)
	  fprintf(stderr, "Cannot open input HDF file %s\n", in_fname);
        else
        {
	  if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
	    fprintf(stderr, "Cannot open output HDF file %s\n", out_fname);
	  else
	  {
	    fprintf(stdout, "Processing file %s\n", in_fname);
	    for (isds=0; isds<nsds; isds++) 
	    {
	      strcpy(in_sds_info.name, sds_names[isds]);
	      in_sds_info.sds_id = -1;
	      if (get_sds_info(in_fname, &in_sds_info) != -1)
	      {
		strcpy(out_sds_info.name, in_sds_info.name);
		out_sds_info.rank = in_sds_info.rank;
		for (i=0; i<in_sds_info.rank; i++)
		  out_sds_info.dim_size[i] = in_sds_info.dim_size[i];
		out_sds_info.data_type = in_sds_info.data_type;
		out_sds_info.data_size = in_sds_info.data_size;
		out_sds_info.sds_id = -1;
		if (open_sds((char *)NULL, &out_sds_info, 'W') != -1)
		{
                  transpose_an_sds(&in_sds_info, &out_sds_info);
		  write_all_sds_attr(in_sds_info.H5IN, in_sds_info.sds_id, out_sds_info.H5OUT, out_sds_info.sds_id, in_sds_info.nattr);
		  close_sds(&out_sds_info, out_sds_info.H5OUT);
		}
		close_sds(&in_sds_info, in_sds_info.H5IN);
	      }
	    } /* for (isds=0; . . .) */
	    if (m_opt == 1) 
	      write_metadata(in_sds_info.H5IN, in_sds_info.sd_id, out_sds_info.H5OUT, out_sds_info.sd_id);
	    close_sd(&out_sds_info, out_sds_info.H5OUT);
	  }
	  close_sd(&in_sds_info, in_sds_info.H5IN);
	}
      }
    }
  }
  return 0;
}

int parse_cmd_transpose_sds(int argc, char **argv, char **sds_names, int *nsds, char *in_fname, char *out_fname, int *m_opt)
{
  int i, st;

  st = 1;
  *m_opt = *nsds = 0;
  in_fname[0] = '\0';
  out_fname[0] = '\0';
  for (i=0, st=1; i<argc; i++)
  {
    if (is_arg_id(argv[i], "-of") == 0) 
      get_arg_val(argv[i], out_fname);
    else if (is_arg_id(argv[i], "-sds") == 0) 
      get_arg_val_arr(argv[i], sds_names, nsds);
    else if (strcmp(argv[i], "-meta") == 0)
      *m_opt = 1;
    else if (argv[i][0] == '-')
      fprintf(stderr, "Ignoring invalid input argument %s\n", argv[i]);
    else strcpy(in_fname, argv[i]);
  }
  if (in_fname[0] == '\0') { fprintf(stderr, "Missing input file \n"); st = -1; }
  if (out_fname[0] == '\0') { fprintf(stderr, "Missing output file \n"); st = -1; }
  if (st == 1)
  {
    if (*nsds == 0) {
      fprintf(stdout, "No SDS names specified. Processing all SDS\n");
      *nsds = get_sds_names(in_fname, sds_names);
    }
  }
  return st;
}

void transpose_an_sds(sds_t *in_sds_info, sds_t *out_sds_info)
{
  int ic_in, ic_out;
  int irow, icol, iobs;
  int nrows, ncols, nobs;
  int i, rank, ndata;
  int32 in_edge[4], out_edge[4]; 
  int32 in_start[4], out_start[4]; 
  void *data_in, *data_out;

  fprintf(stdout, "\tProcessing SDS %s\n", in_sds_info->name);
  rank = in_sds_info->rank;
  if (in_sds_info->interleave == 1)
  {
    nrows = in_sds_info->dim_size[rank-2];
    ncols = ndata = in_sds_info->dim_size[rank-1];
    for (i=0, nobs=1; i<rank-2; i++)
    {
      ndata *= in_sds_info->dim_size[i];
      nobs *= in_sds_info->dim_size[i];
    }
  }
  else
  {
    nrows = in_sds_info->dim_size[0];
    ncols = ndata = in_sds_info->dim_size[1];
    for (i=2, nobs=1; i<rank; i++)
    {
      ndata *= in_sds_info->dim_size[i];
      nobs *= in_sds_info->dim_size[i];
    }
  }

  if ((data_in = (void *)calloc(ndata, in_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_in in transpose_an_sds\n");
  if ((data_out = (void *)calloc(ndata, out_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_out in transpose_an_sds\n");
  if ((data_in != NULL) && (data_out != NULL))
  {
    for (i=0; i<rank; i++)
    {
      in_edge[i] = out_edge[i] = in_sds_info->dim_size[i];
      in_start[i] = out_start[i] = 0;
    }
    if (in_sds_info->interleave == 0) in_edge[0] = out_edge[0] = 1;
    else in_edge[rank-2] = out_edge[rank-2] = 1;

    for (irow=0; irow<nrows; irow++)
    {
/*      if (irow % 100 == 0) fprintf(stdout, "Processing row %d\n", irow); */
      if (in_sds_info->interleave == 1) { 
	in_start[rank-2] = irow;
	out_start[rank-2] = nrows - 1 - irow;
      }
      else {
	in_start[0] = irow;
	out_start[0] = nrows - 1- irow;
      }
     
      if (read_sel_sds_data(in_sds_info, in_start, in_edge, data_in) == FAIL) 
      {
        fprintf(stderr, "Cannot read data line %d from SDS %s in transpose_an_sds\n", irow, in_sds_info->name);
        break;
      }
      else
      {
	if (in_sds_info->interleave == 1)
	{
	  for (iobs=0, ic_out=0; iobs<nobs; iobs++)
	  {
	    ic_in = iobs*ncols + ncols - 1; 
	    switch(in_sds_info->data_type)
	    {
	      case 5: for (icol=0; icol<ncols; icol++, ic_in--, ic_out++)
			 ((float32 *)data_out)[ic_out] = ((float32 *)data_in)[ic_in]; 
		       break;	      
	      case 20: for (icol=0; icol<ncols; icol++, ic_in--, ic_out++)
			 ((int8 *)data_out)[ic_out] = ((int8 *)data_in)[ic_in]; 
		       break;
	      case 21: for (icol=0; icol<ncols; icol++, ic_in--, ic_out++)
			 ((uint8 *)data_out)[ic_out] = ((uint8 *)data_in)[ic_in];
		       break;
	      case 22: for (icol=0; icol<ncols; icol++, ic_in--, ic_out++)
			 ((int16 *)data_out)[ic_out] = ((int16 *)data_in)[ic_in];
		       break;
	      case 23: for (icol=0; icol<ncols; icol++, ic_in--, ic_out++)
			 ((uint16 *)data_out)[ic_out] = ((uint16 *)data_in)[ic_in];
		       break;
	      case 24: for (icol=0; icol<ncols; icol++, ic_in--, ic_out++)
			 ((int32 *)data_out)[ic_out] = ((int32 *)data_in)[ic_in];
		       break;
	      case 25: for (icol=0; icol<ncols; icol++, ic_in--, ic_out++)
			 ((uint32 *)data_out)[ic_out] = ((uint32 *)data_in)[ic_in];
		       break;
	    }
	  } /* for (iobs=0; . . ) */
	}
	else
	{
	  for (icol=0, ic_out=0; icol<ncols; icol++)
	  {
	    ic_in = (ncols - 1 - icol)*nobs;; 
	    switch(in_sds_info->data_type)
	    {
	      case 5: for (iobs=0; iobs<nobs; iobs++, ic_out++, ic_in++)
			 ((float32 *)data_out)[ic_out] = ((float32 *)data_in)[ic_in]; 
		       break;	      
	      case 20: for (iobs=0; iobs<nobs; iobs++, ic_out++, ic_in++)
			 ((int8 *)data_out)[ic_out] = ((int8 *)data_in)[ic_in]; 
		       break;
	      case 21: for (iobs=0; iobs<nobs; iobs++, ic_out++, ic_in++)
			 ((uint8 *)data_out)[ic_out] = ((uint8 *)data_in)[ic_in]; 
		       break;
	      case 22: for (iobs=0; iobs<nobs; iobs++, ic_out++, ic_in++)
			 ((int16 *)data_out)[ic_out] = ((int16 *)data_in)[ic_in]; 
		       break;
	      case 23: for (iobs=0; iobs<nobs; iobs++, ic_out++, ic_in++)
			 ((uint16 *)data_out)[ic_out] = ((uint16 *)data_in)[ic_in]; 
		       break;
	      case 24: for (iobs=0; iobs<nobs; iobs++, ic_out++, ic_in++)
			 ((int32 *)data_out)[ic_out] = ((int32 *)data_in)[ic_in]; 
		       break;
	      case 25: for (iobs=0; iobs<nobs; iobs++, ic_out++, ic_in++)
			 ((uint32 *)data_out)[ic_out] = ((uint32 *)data_in)[ic_in]; 
		       break;
	    }
	  } /* for (icol=0; . .  ) */
	}
	if(set_sel_sds_data(out_sds_info, out_start, out_edge, data_out) == FAIL)
        {
          fprintf(stderr, "Cannot write data line %d to SDS %s in transpose_an_sds\n", irow, out_sds_info->name);
          break;
        }
      }
    } /* for (irow=0; . . .) */
  }
  if (data_in != NULL) free(data_in);
  if (data_out != NULL) free(data_out);
}


