/****************************************************************************
!C

!File: sds_rw.c

!Description:
  Routines to support SDS read and write 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original January/February 1998. Version 1.0
  Modified May/June 1998. Version 1.1 (Changes to support batch processing. 
		Routines do not exit abruptly, instead returns to main 
		with appropriate error messages)
  Modified July 1999 Version 1.2 (Handle higher dimensions specified in
		the sds names input)
  Modified March 2000 Version 1.3 (Added function copy_sds which is used to
                                   copy SDS fields from one file to another.)
  Modified March 2000 (Use sds_index in get_sds_info, get_sds_data)
  Modified March 2013 by Xiaoping Zhang Sigma Space Corp.
  	(handle the SDS name contains "." in get_sds_info, update_nd_sdsnames)
  Modified 2017 and 2018 by Sudipta Sarkar and Maosheng Zhao for dealing with
	HDF4/HDF5.

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <string.h>
#include <strings.h>

#include "mfhdf.h"
#include "hdf5.h"
#include "str_op.h"
#include "sds_rw.h"
#include "qa_tool.h"
#include "alloc_mem.h"
#include "main_util.h"

/* all static support functions for HDF4 */
static int HDF4_scan_SDSs_info(HDF4_SDSs_INFO*);
static int HDF4_free_SDSs_info(HDF4_SDSs_INFO*);

/* all static support functions for HDF5 */
herr_t H5Fget_objs_links(hid_t, H5_OBJS_TYPES*);
static herr_t H5L_get_num_objs_links(hid_t fid, hsize_t *num_objs_links);
static herr_t cb_func_num_objs_links_L(hid_t loc_id, const char *name, const H5L_info_t *info, void *operator_data);
static herr_t H5L_get_names_types_objs_links(hid_t fid, H5_OBJS_TYPES *h5_obj_types);
static herr_t cb_func_names_objs_links_L(hid_t loc_id, const char *name, const H5L_info_t *info, void *operator_data);

/* this supports HDF5_get_data_type_info() */
static unsigned char byte_order(void);

int get_sds_info(const char *hdf_fname, sds_t *sds_info)  
/*******************************************************************************
	Read the sds information about the names sds sds_name and store
	them in the structure sds_info. If the file and the sds are not open, 
	then open the file and then the sds. Return 1 on success and return -1 
	on failure.
*******************************************************************************/
{
    int p1, p11;
    int32 attr_type, attr_cnt;
    char sds_name[MAX_SDS_NAME_LEN];
    char attr_name[MAX_ATTR_NAME_LEN];  
    void *attr_val;
    char gran_id[MAX_GID_LEN];    
    int32 nsds, nattr;
    int i;
    int32 rank, dt, dim_size[4];
    char name[MAX_SDS_NAME_LEN];
    char *attr_names[] = {"scale_factor", "add_offset"};  /* MODIS */
    char *attr_names1[] = {"Scale", "Offset"};     /* VNP */
  
    /* H5 addition */
    H5SDS *h5_sds = NULL;
    hid_t grp_id;
    int sds_cnt;
    int HDF_ver = 0;
    herr_t ret;
    H5O_info_t oinfo;

    /* MSZ 4/9/2018 */
    sds_info->nattr = -1;

    /* MSZ 2/1/2018, to initialize use_L2G_max_nobs as 0.
     * Note that if you want to use it and the associated L2G_max_nobs, you must reset
     * use_L2G_max_nobs as 1 and set L2G_max_nobs after call get_sds_info */
    sds_info->use_L2G_max_nobs = 0;

    if (sds_info->sd_id == -1 && hdf_fname != NULL)
    {
	/* MSZ 8/17/17, to track the file name of the sds */
	sds_info->file_name = (char*)hdf_fname;

        /* MSZ 2/2/2018 Note that here the situation is different from the below
         * is that if hdf_fname is the input file, reset H5IN is fine; whereas
         * if hdf_fname is the output file, H5OUT is not reset here and H5IN has 
         * no effects on the output file */
	HDF_ver = HDF_version_check(hdf_fname);

	if(HDF_ver == 4)
	{
	  sds_info->H5IN = 0;
	  if((sds_info->sd_id = SDstart(hdf_fname, DFACC_READ)) == FAIL)
	  {
	    sds_info->sd_id = -1;
	    fprintf(stderr, "Cannot open the HDF file %s\n", hdf_fname);
	    return -1;
	  }
	}
	else if(HDF_ver == 5)
	{
	  sds_info->H5IN = 1;
	  if((sds_info->sd_id = (int32)H5Fopen(hdf_fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
	  {
	    sds_info->sd_id = -1;
	    fprintf(stderr, "Cannot open the HDF file %s\n", hdf_fname);
	    return -1;
	  }
	}
	else
	{
	  fprintf(stderr, "%s is neither HDF4 or HDF5\n", hdf_fname);
	  return -1;
	}
    }

    if (sds_info->sds_id == -1)
    {
        if (strlen(sds_info->name) > 0)
        {
	    /* MSZ note on 2/2/2018, though this function is often being used for the input HDF,
 	     * but if for the output, this could be a bug if input and output have different format.
 	     * possible solutions are:
 	     * 1) use sds_info->file_name to check HDF4 or HDF5 but not use H5IN and H5OUT.
 	     * 2) use H5IN or H5OUT but add open_t in the argument in a similar style as open_sd() and open_sds()
 	     */
	    if (sds_info->H5IN)
            {
		/* MSZ note: h5_dsetget scan HDF5 to find if short name sds_info->name exists */
		/* at the same time, get its abs_gnames or abs_path */
                if ((h5_sds = h5_dsetget(sds_info, sds_info->name)) == NULL)
                {
		  p1 = charpos(sds_info->name, '.', 0);
                  if ((p1 = charpos(sds_info->name, '.', 0)) != -1)    /* for VNP43 */
                  {
                    strmid(sds_info->name, 0, p1, sds_name);
                    if ((h5_sds = h5_dsetget(sds_info, sds_name)) != NULL)
		    {
		      for(i = 0; i < strlen(sds_info->name); i++) sds_info->name[i] = '\0';
		      strcpy(sds_info->name, sds_name);
		    }
		    else
                    {
                      fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_info->name);
                      return -1;
                    }
                  }
		  else  /* MSZ 9/9/17, for missing nadd_obs_row and compact data */
		    return -1;
                }

                if (!h5_sds->sds_found)
                {
                    fprintf(stderr, "Cannot find the SDS First try %s \n", sds_info->name);
                    free(h5_sds);
                    if ((p1 = charpos(sds_info->name, '.', 0)) != -1)
                    {
                        strmid(sds_info->name, 0, p1, sds_name);
                        if ((h5_sds = h5_dsetget(sds_info, sds_name)) == NULL)
                        {
                            fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_info->name);
                            return -1;
                        }
                        if (!h5_sds->sds_found)
                        {
                            free(h5_sds);
                            fprintf(stderr, "Cannot find the SDS Second try %s \n", sds_name);
                            p1++;
                            if ((p11 = charpos(sds_info->name, '.', p1)) != -1) 
                            {		      		      		      
                                strmid(sds_info->name, 0, p11, sds_name);
                                fprintf(stderr, "sds_name1: %s \n",sds_name);
                                if ((h5_sds = h5_dsetget(sds_info, sds_name)) == NULL)
                                {
                                    fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_info->name);
                                    return -1;
                                }
                            }
                            else
                            {
                                h5_sds = calloc(1,sizeof(H5SDS));
                                h5_sds->srch_type = 4; /* for getting sds count and all sds names */
                                h5_sds->nsds = 0;
                                H5Ovisit((hid_t)sds_info->sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5O_iterate_t)H5dataset_search, (void *)&h5_sds);
                                sds_cnt = h5_sds->nsds;
                                for (i=0; i<sds_cnt; i++)
                                {
                                    if (strcasecmp(h5_sds->h5sdsnames[i], sds_info->name) == 0)
                                    {
                                        strcpy(sds_info->name, h5_sds->h5sdsnames[i]);
                                        strcpy(h5_sds->name,h5_sds->h5sdsnames[i]);
                                        /*H5Ovisit((hid_t)sds_info->sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5O_iterate_t)H5lnk_handler, (void*)&h5_sds);*/
					H5Lvisit((hid_t)sds_info->sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5L_iterate_t)H5lnk_handler, (void*)&h5_sds);
                                        strcpy(sds_name, sds_info->name);
                                        fprintf(stderr, "sds_name2: %s \n", sds_name);    
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else
            {
                if ((sds_info->sds_index = SDnametoindex(sds_info->sd_id, sds_info->name)) == FAIL)
                {
                    if ((p1 = charpos(sds_info->name, '.', 0)) != -1)   /* for MCD43 */
                    {		  	  
                      strmid(sds_info->name, 0, p1, sds_name);
                          sds_info->sds_index = SDnametoindex(sds_info->sd_id, sds_name);
                      fprintf(stderr, "sds_info->sds_index: %d \n",sds_info->sds_index);	 	  
                      fprintf(stderr, "sds_name1: %s \n",sds_name);	  
                      if (sds_info->sds_index == FAIL)
                      {
                        fprintf(stderr, "Cannot find the SDS %s \n", sds_name);
                        p1++;
                        if ((p11 = charpos(sds_info->name, '.', p1)) != -1) 
                        {		      		      		      
                          strmid(sds_info->name, 0, p11, sds_name);
                          sds_info->sds_index = SDnametoindex(sds_info->sd_id, sds_name);
                          fprintf(stderr, "sds_name1: %s \n",sds_name);	      
                        }
                        else
                        {	 	    	        	     	     	     	      	      	      	      
                          if (SDfileinfo(sds_info->sd_id, &nsds, &nattr) != FAIL)
                            for (i=0; i<nsds; i++)
                              if ((sds_info->sds_id = SDselect(sds_info->sd_id, i)) != FAIL)
                                if (SDgetinfo(sds_info->sds_id, name, &rank, dim_size, &dt, &sds_info->nattr) != FAIL)
                                  if (strcasecmp(name, sds_info->name) == 0)
                                  {
                                    strcpy(sds_info->name, name);
                                    sds_info->sds_index = i;
                                    strcpy(sds_name, sds_info->name);
                                  }	

                          fprintf(stderr, "sds_name2: %s \n", sds_name);       
                          if (sds_info->sds_index == FAIL)
                          {
                            fprintf(stderr, "Cannot find the SDS %s \n", sds_info->name);
                            return -1;
                          }			      	      	      
                        }	    			    	    	    
                      }
                      else
                      {	    
                        strcpy(sds_info->name, sds_name);
			
                        fprintf(stderr, "sds_name2: %s \n",sds_name); 
                      }	  	   
                    }        	
		    else  /* MSZ 9/9/17, for missing nadd_obs_row and compact data */
                        return -1;

		    /*
                    else 
                    {	  
                      sds_info->sds_index = (int)atoi(sds_info->name);
                      strcpy(sds_name, sds_info->name);
                    }
			*/
                }
                else
                {
                  strcpy(sds_name, sds_info->name);
                }
            }
        } /* if (strlen(sds_info->name) > 0) */
  
        if (sds_info->H5IN)
        {
          if (!h5_sds->sds_found)
          {
            fprintf(stderr, "Cannot find the SDS %s \n", sds_info->name);
            free(h5_sds);
            return -1;
          }
          
          /* fprintf(stderr, "sdsname = %s, sd_id = %d abs_gnames = %s\n", sds_info->name, sds_info->sd_id, h5_sds->abs_gnames); */
          if ((grp_id = H5Gopen2(sds_info->sd_id, h5_sds->abs_gnames, H5P_DEFAULT)) < 0)
          {
            fprintf(stderr,"H5: Failed to open H5 group %s in get_sds_info()\n",h5_sds->abs_gnames);
            free(h5_sds);
            return -1;
          }
	  else
	    strcpy(sds_info->h5gname, h5_sds->abs_gnames);   /* MSZ 1/29/2018, add this to store the sds group name for the sds name */
         
	 
          if ((sds_info->sds_id = (int32)H5Dopen2(grp_id,h5_sds->name,H5P_DEFAULT)) < 0)
          {
            fprintf(stderr,"H5: Failed to retrieve dataset %s in group %s get_sds_info()\n",h5_sds->name, h5_sds->abs_gnames);
	    /* MSZ 8/15/17 exit here and print clear information since this is a serious error */
	    fprintf(stderr, "cannot get sds_id using H5Dopen2, exit...\n");
	    exit(0);
            free(h5_sds);
            return -1;
          }

	  /* MSZ 4/9/2018, add this to get nattr for dataset in HDF5 */
	  if((ret = H5Oget_info((hid_t)sds_info->sds_id, &oinfo)) >= 0)
	    sds_info->nattr = (int32)oinfo.num_attrs;
	  else
	    sds_info->nattr = -1;
        }
        else
        {
            if (sds_info->sds_index == FAIL)
            {
              fprintf(stderr, "Cannot find the SDS %s \n", sds_info->name);
              return -1;
            }
            if ((sds_info->sds_id = SDselect(sds_info->sd_id, sds_info->sds_index)) == FAIL)
            {
              fprintf(stderr, "Cannot open the SDS %s \n", sds_info->name);
              return -1;
            }
        }
    } /* if (sds_info->sds_id == -1) */
  
    if (sds_info->H5IN)
    {
        if (h5_dsetinfo(h5_sds, sds_info) == FAIL)
        {
            fprintf(stderr,"Failed to get h5_dsetinfo for sdsname %s\n", sds_info->name);
            H5Gclose(grp_id);
            exit(0);
            free(h5_sds);
            return -1;
        }
    }
    else
    {
        if (SDgetinfo(sds_info->sds_id, sds_name, &sds_info->rank, sds_info->dim_size,
              &sds_info->data_type, &sds_info->nattr) == FAIL)
        {   
          fprintf(stderr, "Cannot get information for the SDS %s \n", sds_name);
          return -1;
        }
        else {
          if (strlen(sds_info->name) <= 0)
            strcpy(sds_info->name, sds_name);
          if ((int)atoi(sds_info->name) == sds_info->sds_index)
          {
            if (strcmp(sds_info->name, sds_name) != 0)
            {
          fprintf(stderr, "Cannot find the SDS %s \n", sds_info->name);
          return -1;
            }	
            else
            {        
              strcpy(sds_info->name, sds_name);
              sds_info->data_size = DFKNTsize(sds_info->data_type);
            }
          }  	
          else
            sds_info->data_size = DFKNTsize(sds_info->data_type);  
        }
    }

  /* MSZ 2/5/2018, get row_id */
  if(sds_info->rank > 4)
  {
    fprintf(stderr, "The tool works for data rank <= 4, whereas %s has rank = %d, bad data dims!...\n", sds_info->name, sds_info->rank);
    return -1;
  }
  else
  {
    if(sds_info->rank <= 2)
      sds_info->row_id = 0;
    else 
      sds_info->row_id = (sds_info->dim_size[0] > sds_info->dim_size[sds_info->rank-1]) ? 0 : sds_info->rank-2;
  }

  /* MSZ 2/5/2018, get band interleave for rank >= 2. Here 1 is for bsq and 0 for bip */
  /* Note that the assumption is that Y or X dim have larger size or length than non-XY dimes. 
   * This is almost true for all cases for remote sensing data */
  sds_info->interleave = 0;
  if(sds_info->rank >= 2)
    sds_info->interleave = ((sds_info->rank == 2) || (sds_info->dim_size[0] < sds_info->dim_size[sds_info->rank - 1])) ? 1 : 0;

  /* MSZ 5/21/2019, get non_xy_dims_size for rank >= 2 */
  sds_info->non_xy_dims_size = 1;
  if(sds_info->rank >= 2)
  {
    if(sds_info->interleave)  /* bsq */
      for (i = 0; i < sds_info->rank-2; i++)
        sds_info->non_xy_dims_size *= sds_info->dim_size[i];
    else	              /* bip */
      for (i = 2; i < sds_info->rank; i++)
	sds_info->non_xy_dims_size *= sds_info->dim_size[i];
  }

  strcpy(attr_name, "_FillValue");
  if ((attr_val = (void *)get_sds_attr(sds_info, attr_name, &attr_type, &attr_cnt)) != NULL)
  {
    switch(attr_type)
    {
      case 5 : sds_info->fill_fval = ((float32 *)attr_val)[0]; break;
      case 6 : sds_info->fill_fval = ((float64 *)attr_val)[0]; break;
      case 20: sds_info->fill_val = ((int8 *)attr_val)[0]; break;
      case 21: sds_info->fill_val = ((uint8 *)attr_val)[0]; break;
      case 22: sds_info->fill_val = ((int16 *)attr_val)[0]; break;
      case 23: sds_info->fill_val = ((uint16 *)attr_val)[0]; break;
      case 24: sds_info->fill_val = ((int32 *)attr_val)[0]; break;
      case 25: sds_info->fill_val = ((uint32 *)attr_val)[0]; break;
    }
    free(attr_val);
/*
    if(attr_type == 5 || attr_type == 6)
	printf("ZMS0 %lf\n", sds_info->fill_fval);
    else
	printf("ZMS1 %d\n", sds_info->fill_val);
*/
  }
  else{
    /* Check for NPP product types  and assign the lower limit of fill value range */
     switch(sds_info->data_type)
    {
      case 5 : sds_info->fill_fval = SOUB_FLOAT32_FILL;
	sds_info->fill_val = sds_info->fill_uval = 0;
	break;
      case 6: sds_info->fill_fval = SOUB_FLOAT32_FILL;
	sds_info->fill_val = sds_info->fill_uval = 0;
	break;
      case 20: sds_info->fill_uval = SOUB_INT8_FILL;
	sds_info->fill_fval = sds_info->fill_val = 0;
	break;
      case 21: sds_info->fill_uval = SOUB_UINT8_FILL;
	sds_info->fill_fval = sds_info->fill_val = 0;
	break;
      case 22: sds_info->fill_val =  SOUB_INT16_FILL;
	sds_info->fill_fval = sds_info->fill_uval = 0;
	break;
      case 23: sds_info->fill_uval = SOUB_UINT16_FILL;
	sds_info->fill_fval = sds_info->fill_val = 0;
	break;
      case 24: sds_info->fill_val = SOUB_INT16_FILL;
	sds_info->fill_fval = sds_info->fill_uval = 0;
	break;
      case 25: sds_info->fill_uval = SOUB_UINT32_FILL;
	sds_info->fill_fval = sds_info->fill_val = 0;
	break;
    }
  }  

  strcpy(attr_name, "valid_range");
  if( !(strcasestr(sds_info->name, "num_observations")))
  {
    if ((attr_val = (void *)get_sds_attr(sds_info, attr_name, &attr_type, &attr_cnt)) != NULL)
    {
      if (attr_cnt != 2)
        /*fprintf(stderr, "SDS range value in error: Does not contain two values for %s\n", sds_info->name)*/;
      switch(attr_type)
      {
        case 5 : sds_info->frange[0] = ((float32 *)attr_val)[0]; 
	       sds_info->frange[1] = ((float32 *)attr_val)[1];
	       break;
        case 6 : sds_info->frange[0] = ((float64 *)attr_val)[0]; 
	       sds_info->frange[1] = ((float64 *)attr_val)[1];
	       break;	             
        case 20: sds_info->range[0] = ((int8 *)attr_val)[0]; 
	       sds_info->range[1] = ((int8 *)attr_val)[1];
	       break;
        case 21: sds_info->range[0] = ((uint8 *)attr_val)[0]; 
	       sds_info->range[1] = ((uint8 *)attr_val)[1];
	       break;
        case 22: sds_info->range[0] = ((int16 *)attr_val)[0]; 
	       sds_info->range[1] = ((int16 *)attr_val)[1];
	       break;
        case 23: sds_info->range[0] = ((uint16 *)attr_val)[0]; 
	       sds_info->range[1] = ((uint16 *)attr_val)[1];
	       break;
        case 24: sds_info->range[0] = ((int32 *)attr_val)[0]; 
	       sds_info->range[1] = ((int32 *)attr_val)[1];
	       break;
        case 25: sds_info->range[0] = ((uint32 *)attr_val)[0]; 
	       sds_info->range[1] = ((uint32 *)attr_val)[1];
	       break;
      }
      free(attr_val);
  }  
  else
    {sds_info->range[0] = -111; sds_info->range[1] = -111;}  
  }

  /* MSZ 8/18/17, don't get attr for name of sdss with obs, state and QC for L2G MODIS and NPP datasets
   * because there is no scale and offset for these SDSs and then there will be creen printed warning info
   * and it is ugly. */
  attr_type = 999;
  if( !(strcasestr(sds_info->name, "obs") ||          /* L2G */
        strcasestr(sds_info->name, "state") ||        /* MxD09 and VNP09 */
        strcasestr(sds_info->name, "qc") ||           /* MxD and VNP */
	strcasestr(sds_info->name, "reliability") ||  /* MxD13A */
	strcasestr(sds_info->name, "Quality")) )      /* MCD43 and VNP43 */
  {
    /* MSZ 7/10/2018, initialize it because there is weird results on Mac,
     * but update fscale or scale below if the other exists */
    
    sds_info->fscale = 1.;
    sds_info->scale  = 1;

    sds_info->foffset = 0.;
    sds_info->offset  = 0;

    if ((attr_val = (void *)get_sds_attr(sds_info, attr_names[0], &attr_type, &attr_cnt)) != NULL)
    {
      switch(attr_type)
      {
        case 5 : sds_info->fscale = ((float32 *)attr_val)[0]; sds_info->scale = 0; break;
        case 6 : sds_info->fscale = ((float64 *)attr_val)[0]; sds_info->scale = 0; break;
        case 20: sds_info->scale = ((int8 *)attr_val)[0]; sds_info->fscale = 0.; break;
        case 21: sds_info->scale = ((uint8 *)attr_val)[0]; sds_info->fscale = 0.; break;
        case 22: sds_info->scale = ((int16 *)attr_val)[0]; sds_info->fscale = 0.; break;
        case 23: sds_info->scale = ((uint16 *)attr_val)[0]; sds_info->fscale = 0.; break;
        case 24: sds_info->scale = ((int32 *)attr_val)[0]; sds_info->fscale = 0.; break;
        case 25: sds_info->scale = ((uint32 *)attr_val)[0]; sds_info->fscale = 0.; break;          
      }
      free(attr_val);
    } 
    else if ((attr_val = (void *)get_sds_attr(sds_info, attr_names1[0], &attr_type, &attr_cnt)) != NULL)
    {
      switch(attr_type)
      {
        case 5 : sds_info->fscale = ((float32 *)attr_val)[0]; sds_info->scale = 0; break;
        case 6 : sds_info->fscale = ((float64 *)attr_val)[0]; sds_info->scale = 0; break;
        case 20: sds_info->scale = ((int8 *)attr_val)[0]; sds_info->fscale = 0.; break;
        case 21: sds_info->scale = ((uint8 *)attr_val)[0]; sds_info->fscale = 0.; break;
        case 22: sds_info->scale = ((int16 *)attr_val)[0]; sds_info->fscale = 0.; break;
        case 23: sds_info->scale = ((uint16 *)attr_val)[0]; sds_info->fscale = 0.; break;
        case 24: sds_info->scale = ((int32 *)attr_val)[0]; sds_info->fscale = 0.; break;
        case 25: sds_info->scale = ((uint32 *)attr_val)[0]; sds_info->fscale = 0.; break;
      }
      free(attr_val);      
    }
    else
    {
      sds_info->fscale = FSO_FILL;
      sds_info->scale = (int)FSO_FILL;
      sds_info->fscale_type = SO_TYPE;      
    }

    if ((attr_val = (void *)get_sds_attr(sds_info, attr_names[1], &attr_type, &attr_cnt)) != NULL)
    {
      switch(attr_type)
      {
        case 5 : sds_info->foffset = ((float32 *)attr_val)[0]; sds_info->offset = 0; break;
        case 6 : sds_info->foffset = ((float64 *)attr_val)[0]; sds_info->offset = 0; break;
        case 20: sds_info->offset = ((int8 *)attr_val)[0]; sds_info->foffset = 0.; break;
        case 21: sds_info->offset = ((uint8 *)attr_val)[0]; sds_info->foffset = 0.; break;
        case 22: sds_info->offset = ((int16 *)attr_val)[0]; sds_info->foffset = 0.; break;
        case 23: sds_info->offset = ((uint16 *)attr_val)[0]; sds_info->foffset = 0.; break;
        case 24: sds_info->offset = ((int32 *)attr_val)[0]; sds_info->foffset = 0.; break;
        case 25: sds_info->offset = ((uint32 *)attr_val)[0]; sds_info->foffset = 0.; break;          
      }
      free(attr_val);
    } 
    else if ((attr_val = (void *)get_sds_attr(sds_info, attr_names1[1], &attr_type, &attr_cnt)) != NULL)
    {
      switch(attr_type)
      {
        case 5 : sds_info->foffset = ((float32 *)attr_val)[0]; sds_info->offset = 0; break;
        case 6 : sds_info->foffset = ((float64 *)attr_val)[0]; sds_info->offset = 0; break;
        case 20: sds_info->offset = ((int8 *)attr_val)[0]; sds_info->foffset = 0.; break;
        case 21: sds_info->offset = ((uint8 *)attr_val)[0]; sds_info->foffset = 0.; break;
        case 22: sds_info->offset = ((int16 *)attr_val)[0]; sds_info->foffset = 0.; break;
        case 23: sds_info->offset = ((uint16 *)attr_val)[0]; sds_info->foffset = 0.; break;
        case 24: sds_info->offset = ((int32 *)attr_val)[0]; sds_info->foffset = 0.; break;
        case 25: sds_info->offset = ((uint32 *)attr_val)[0]; sds_info->foffset = 0.; break;
      }
      free(attr_val);      
    }
    else
    {
      sds_info->foffset = 0.;  
      sds_info->offset = 0;                              
      sds_info->fscale_type = SO_TYPE;      
    } 
  } /* end if  if( !(strstr(sds_info.name, "obs") || ... */

    if (sds_info->H5IN)
    {
        H5Gclose(grp_id);
        if (h5_sds) free(h5_sds);
    }                                                 
    /*
    fprintf(stderr, "  sds_info.fscale = %f\n", sds_info->fscale);
    fprintf(stderr, "  sds_info.foffset = %f\n", sds_info->foffset);  
    fprintf(stderr, "  sds_info.fscale_type = %d\n", sds_info->fscale_type);   
    */
                              
    return 1;
}

int get_sds_data(sds_t *sds_info, void *data)
/*********************************************************************************
	Read the data for the named sds sds->name and return the SDS data in
	data. Return 1 on success and -1 on failure.
*********************************************************************************/
{
    int idim;
    int status;
    int32 *edge;
    int32 *start;
    hsize_t *he5_start = NULL;
    hsize_t *he5_count = NULL;
    H5SDS *h5_sds;
    hid_t grp_id;
    hid_t in_h5memspace_id;
    
    status = 1;
  
    if (sds_info->H5IN)
    {
        if (sds_info->sds_id == -1)
        {
            if ((h5_sds = h5_dsetget(sds_info, sds_info->name)) == NULL)
            {
                fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_info->name);
                return -1;
            }
            if (!h5_sds->sds_found)
            {
                fprintf(stderr, "Cannot find SDS %s in get_sds_data\n", sds_info->name);
                status = -1;
            }
            if ((grp_id = H5Gopen2(sds_info->sd_id, h5_sds->abs_gnames,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to open H5 group %s in get_sds_data()\n",h5_sds->abs_gnames);
              return -1;
            }
            
            if ((sds_info->sds_id = (int32)H5Dopen2(grp_id,h5_sds->name,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to retrieve dataset %s in group %s get_sds_data()\n",h5_sds->name, h5_sds->abs_gnames);
	      /* MSZ 8/15/17, exit here because this is a serious error */
	      fprintf(stderr, "cannot open the data using H5Dopen2, exit...\n");
	      exit(0);
              return -1;
            }
            
            if (h5_dsetinfo(h5_sds, sds_info) == FAIL)
            {
                fprintf(stderr,"Failed to get h5_dsetinfo for sdsname %s\n", sds_info->name);
                H5Gclose(grp_id);
                free(h5_sds);
                return -1;
            }
            
            if (grp_id >= 0) H5Gclose(grp_id);
            if (h5_sds->h5_dims) free(h5_sds->h5_dims);
            if (h5_sds) free(h5_sds);
        }
        he5_start= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        he5_count= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        if (he5_start == NULL || he5_count == NULL)
        {
            fprintf(stderr, "Cannot allocate memory for start/edge in get_sds_data\n");
            return -1;
        }
        for (idim = 0; idim < sds_info->rank; idim++)
        {
          he5_start[idim] = 0;
          he5_count[idim] = sds_info->dim_size[idim];
        }
        
        /* create the memory space for this dataset */
        if ((in_h5memspace_id = H5Screate_simple(sds_info->rank, he5_count, NULL)) < 0)
        {
            fprintf (stderr,"Failed to create in_h5memspace_id for SDS \"%s\". in get_sds_data \n", sds_info->name);
            /* fprintf(stderr, "rank:%d he5_count:%d,%d\n", sds->rank, sds->he5_count[0], sds->he5_count[1]); */
            free((size_t *)he5_start);
            free((size_t *)he5_count);
            return -1;
        }
        
        if (H5Sselect_hyperslab (sds_info->h5space_id, H5S_SELECT_SET, he5_start, NULL, he5_count,NULL) < 0)
        {
            fprintf(stderr,"Can't select hyperslab for SDS \"%s\". in get_sds_data\n", sds_info->name);
            H5Sclose(sds_info->h5space_id);
            free((size_t *)he5_start);
            free((size_t *)he5_count);
            return -1;
        }
    
        if (H5Dread((hid_t)sds_info->sds_id, sds_info->h5dmtype,in_h5memspace_id, sds_info->h5space_id, H5P_DEFAULT, data) == FAIL)
        {
            fprintf(stderr, "Error reading input HDF file in get_sds_data\n");
            free((size_t *)he5_start);
            free((size_t *)he5_count);
            return -1;
        }
        free((size_t *)he5_start);
        free((size_t *)he5_count);
        return 0;
    }
    else
    {
        if (sds_info->sds_id == -1)
        {
            if ((sds_info->sds_index = SDnametoindex(sds_info->sd_id, sds_info->name)) == FAIL)
            {
              fprintf(stderr, "Cannot find the SDS %s in get_sds_data\n", sds_info->name);
              status = -1;
            }
            if ((sds_info->sds_id = SDselect(sds_info->sd_id, sds_info->sds_index)) == FAIL)
            {
              fprintf(stderr, "Cannot read the SDS %s in get_sds_data\n", sds_info->name);
              status = -1;
            }
            if (status == -1)
              return -1;
        }
        if ((start = (int32 *)calloc(sds_info->rank, sizeof(int32))) == NULL)
        {
          fprintf(stderr, "Cannot allocate memory for start in get_sds_data\n");
          return -1;
        }
        if ((edge = (int32 *)calloc(sds_info->rank, sizeof(int32))) == NULL)
        {
          fprintf(stderr, "Cannot allocate memory for edge in get_sds_data\n");
          return -1;
        }
        for (idim = 0; idim < sds_info->rank; idim++)
        {
          start[idim] = 0;
          edge[idim] = sds_info->dim_size[idim];
        }
          
        if (SDreaddata(sds_info->sds_id, start, NULL, edge, data) == FAIL)
        {
          fprintf(stderr, "Cannot read the SDS data %s in get_sds_data\n", sds_info->name);
          status = -1;
        }
      
        free((int32 *)start);
        free((int32 *)edge);
        if (status == -1) return status;
        else return 0;
    }
}

void *get_sel_sds_data(sds_t *sds_info, int32 *start, int32 *edge)
{
    /* MSZ 07/18/2018. Note that you must call get_sds_info() before call this function */
    /* MSZ 8/29/17. Note the different between this get_sel_sds_data() and 
     * read_sel_sds_data() is that this function callocs data inside whereas latter outside */
    int idim;
    hsize_t nelem;
    hsize_t nelem_sel;  /* MSZ, 8/9/17 */
    int status;
    int32 nattr;
    void *data = NULL;
    
    /* H5 addition */
    H5SDS *h5_sds = NULL;
    hid_t grp_id = -1;
    hsize_t *he5_start = NULL;
    hsize_t *he5_count = NULL;
    hid_t in_h5memspace_id;

    /* MSZ 4/6/2018, add this error information to speedup bug detection */
    if(sds_info->sd_id == -1)
    {
        fprintf(stderr, "file %s is not open in get_sel_sds_data() because its sd_id = -1.\n", sds_info->file_name);
        fprintf(stderr, "SDS %s cannot be read, exit...\n", sds_info->name);
        exit(0);
    }

    status = 1;

    if (sds_info->H5IN)
    {
        if (sds_info->sds_id == -1)
        {
            if ((h5_sds = h5_dsetget(sds_info, sds_info->name)) == NULL)
            {
                fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_info->name);
                return (void *)NULL;
            }
            if (!h5_sds->sds_found)
            {
                fprintf(stderr, "Cannot find SDS %s in get_sel_sds_data\n", sds_info->name);
                status = -1;
            }
            if ((grp_id = H5Gopen2(sds_info->sd_id, h5_sds->abs_gnames,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to open H5 group %s in get_sel_sds_data()\n",h5_sds->abs_gnames);
              return (void *)NULL;
            }
    
            if ((sds_info->sds_id = (int32)H5Dopen2(grp_id,h5_sds->name,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to retrieve dataset %s in group %s get_sel_sds_data()\n",h5_sds->name, h5_sds->abs_gnames);
              return (void *)NULL;
            }

            if (h5_dsetinfo(h5_sds, sds_info) == FAIL)
            {
                fprintf(stderr,"Failed to get h5_dsetinfo for sdsname %s\n", sds_info->name);
                H5Gclose(grp_id);
                free(h5_sds);
                return (void *)NULL;
            }
            
            if (grp_id >= 0) H5Gclose(grp_id);
            if (h5_sds->h5_dims) free(h5_sds->h5_dims);
            if (h5_sds) free(h5_sds);
        }

        he5_start= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        he5_count= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        if (he5_start == NULL || he5_count == NULL)
        {
            fprintf(stderr, "Cannot allocate memory for start/edge in get_sel_sds_data\n");
            return (void *)NULL;
        }
            
        nelem = (sds_info->rank == 1) ? sds_info->dim_size[0] : 1;

	/* MSZ, 8/9/17 */
	nelem_sel = 1;
        for (idim=0; idim<sds_info->rank; idim++)
        {
            if ((sds_info->rank > 1) && (idim != sds_info->row_id))
                nelem *= sds_info->dim_size[idim];
            he5_start[idim] = start[idim];
            he5_count[idim] = edge[idim];

	    /* MSZ, 8/9/17 */
	    if(start[idim] + edge[idim] > sds_info->dim_size[idim])
	    {
		fprintf(stderr, "The specified start %d and edge %d for get_sel_sds_data() is out of %d dim range %d for %s, exit...\n", start[idim], edge[idim], idim, sds_info->dim_size[idim], sds_info->name);
		exit(0);
	    }
	    nelem_sel *= (hsize_t)he5_count[idim];
        }

	/* MSZ 8/18/17, change the nelem_sel for l2g (L2G) compact data */
	if(sds_info->use_L2G_max_nobs == 1)
	{
	    if(sds_info->rank > 1)
	    {
		fprintf(stderr, "for L2G compact data %s, rank should NOT > 1 in get_sel_sds_data(), exit...\n", sds_info->name);
		exit(0);
	    }

	    if(edge[0] > sds_info->L2G_max_nobs)
	    {
		fprintf(stderr, "edge %d exceeds L2G_max_nobs %zu for %s, exit...\n", edge[0], sds_info->L2G_max_nobs, sds_info->name);
		exit(0);
	    }

	    nelem_sel = sds_info->L2G_max_nobs;
	}

        /*if ((data = (void *)calloc(nelem, sds_info->data_size)) != NULL)*/
	if ((data = (void *)calloc(nelem_sel, sds_info->data_size)) != NULL)   /* MSZ, 8/9/17 */ /* selected hyperslab data */
        {   
            if ((in_h5memspace_id = H5Screate_simple(sds_info->rank, he5_count, NULL)) < 0)
            {
              fprintf (stderr,"Failed to create in_h5memspace_id for SDS \"%s\". in get_sds_data \n", sds_info->name);
              status = -1;
            }

            if (H5Sselect_hyperslab (sds_info->h5space_id, H5S_SELECT_SET, he5_start, NULL, he5_count, NULL) < 0)
            {
              fprintf(stderr,"Can't select hyperslab for SDS \"%s\". in get_sds_data\n", sds_info->name);
              H5Sclose(sds_info->h5space_id);
               status = -1;
            }
            
            if (H5Dread((hid_t)sds_info->sds_id, sds_info->h5dmtype, in_h5memspace_id, sds_info->h5space_id, H5P_DEFAULT, data) == FAIL)
            {
                fprintf(stderr, "Cannot read the SDS data %s in get_sel_sds_data\n", sds_info->name);
                status = -1;
            }
        }
        else
            status = -1;
        
        /* if (sds_info->sds_id != -1) H5Dclose(sds_info->sds_id); */
        free((size_t *)he5_start);
        free((size_t *)he5_count);
        if (status == -1)
            return (void *)NULL;
        else
            return data; 
    }
    else
    {
        if (sds_info->sds_id == -1)
        {
            if ((sds_info->sds_index = SDnametoindex(sds_info->sd_id, sds_info->name)) == FAIL)
            {
              fprintf(stderr, "Cannot SDnametoindex for SDS %s in get_sel_sds_data\n", sds_info->name);
              status = -1;
            }

	    /* MSZ 07/17/2018, correct a bug here. Here it should be if not else if */
            if ((sds_info->sds_id = SDselect(sds_info->sd_id, sds_info->sds_index)) == FAIL)
            {
              fprintf(stderr, "Cannot select the SDS %s in get_sel_sds_data\n", sds_info->name);
              status = -1;
            } 

            if (status == -1)
              return (void *)NULL;
        }

        if (SDgetinfo(sds_info->sds_id, sds_info->name, &sds_info->rank, sds_info->dim_size, &sds_info->data_type,
        &nattr) != FAIL)
        {
	    sds_info->data_size = DFKNTsize(sds_info->data_type);

            nelem = (sds_info->rank == 1) ? sds_info->dim_size[0] : 1;

	    /* MSZ, 8/9/17 */
	    nelem_sel = 1;            
            for (idim=0; idim<sds_info->rank; idim++)
	    {
                if ((sds_info->rank > 1) && (idim != sds_info->row_id))
                    nelem *= sds_info->dim_size[idim];

	        /* MSZ, 8/9/17 */
	        if(start[idim] + edge[idim] > sds_info->dim_size[idim])
	        {
		    fprintf(stderr, "The specified start %d and edge %d for get_sel_sds_data() is out of %d dim range %d for %s, exit...\n", start[idim], edge[idim], idim, sds_info->dim_size[idim], sds_info->name);
		    exit(0);
	        }
	        nelem_sel *= (hsize_t)edge[idim];
	    }

	    /* MSZ 8/18/17, change the nelem_sel for l2g (L2G) compact data */
	    if(sds_info->use_L2G_max_nobs == 1)
	    {
		if(sds_info->rank > 1)
		{
		    fprintf(stderr, "for L2G compact data %s, rank should NOT > 1 in get_sel_sds_data(), exit...\n", sds_info->name);
		    exit(0);
		}

	        if(edge[0] > sds_info->L2G_max_nobs)
	        {
	  	  fprintf(stderr, "edge exceeds L2G_max_nobs for %s, exit...\n", sds_info->name);
		  exit(0);
	        }

		nelem_sel = sds_info->L2G_max_nobs;
	    }
                    
            /* fprintf(stderr, "rank:%d edge:%d,%d\n", sds_info->rank, edge[0], edge[1]);
            fprintf(stderr, "start:%d,%d\n", start[0], start[1]);
            fprintf(stderr, "sds_id = %d\n", sds_info->sds_id); */
            /*if ((data = (void *)calloc(nelem, sds_info->data_size)) != NULL)*/
	    if ((data = (void *)calloc(nelem_sel, sds_info->data_size)) != NULL)   /* MSZ, 8/9/17 */ /* selected hyperslab data */
            {
                  if (SDreaddata(sds_info->sds_id, start, NULL, edge, data) == FAIL)
                      fprintf(stderr, "Cannot read the SDS data %s in get_sel_sds_data\n", sds_info->name);
            }
        }
        else data = (void *)NULL;

	if(data == NULL)
	   fprintf(stderr, "cannot read %s in %s\n", sds_info->name, sds_info->file_name);

        /* SDendaccess(sds_info->sds_id); */
        return data;
    }
}

int read_sel_sds_data(sds_t *sds_info, int32 *start, int32 *edge, void *data)
{
    /* MSZ 8/29/17. Note the different between this read_sel_sds_data() and 
 * get_sel_sds_data() is that this function needs data calloced outside whereas
 * latter inside. */
    int idim;
    hsize_t nelem;
    hsize_t nelem_sel;  /* MSZ, 8/9/17 */
    int32 nattr;
    
    /* H5 addition */
    H5SDS *h5_sds = NULL;
    hid_t grp_id = -1;
    hsize_t *he5_start = NULL;
    hsize_t *he5_count = NULL;
    hid_t in_h5memspace_id;

    /* MSZ 4/6/2018, add this error information to speedup bug detection */
    if(sds_info->sd_id == -1)
    {
	fprintf(stderr, "file %s is not open in read_sel_sds_data() because its sd_id = -1.\n", sds_info->file_name);
	fprintf(stderr, "SDS %s cannot be read, exit...\n", sds_info->name);
	exit(0);
    }

    if (sds_info->H5IN)
    {
        if (sds_info->sds_id == -1)
        {
            if ((h5_sds = h5_dsetget(sds_info, sds_info->name)) == NULL)
            {
                fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_info->name);
		return -1;
            }
            if (!h5_sds->sds_found)
            {
                fprintf(stderr, "Cannot find SDS %s read_sel_sds_data\n", sds_info->name);
		return -1;
            }
            if ((grp_id = H5Gopen2(sds_info->sd_id, h5_sds->abs_gnames,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to open H5 group %s read_sel_sds_data()\n",h5_sds->abs_gnames);
		return -1;
            }
            
            if ((sds_info->sds_id = (int32)H5Dopen2(grp_id,h5_sds->name,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to retrieve dataset %s in group %s get_sel_sds_data()\n",h5_sds->name, h5_sds->abs_gnames);
              return -1;
            }
            
            if (h5_dsetinfo(h5_sds, sds_info) == FAIL)
            {
                fprintf(stderr,"Failed to get h5_dsetinfo for sdsname %s\n", sds_info->name);
                H5Gclose(grp_id);
                free(h5_sds);
                return -1;
            }
            
            if (grp_id >= 0) H5Gclose(grp_id);
            if (h5_sds->h5_dims) free(h5_sds->h5_dims);
            if (h5_sds) free(h5_sds);
        }
        he5_start= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        he5_count= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        if (he5_start == NULL || he5_count == NULL)
        {
            fprintf(stderr, "Cannot allocate memory for start/edge read_sel_sds_data\n");
	    return -1;
        }
            
        nelem = (sds_info->rank == 1) ? sds_info->dim_size[0] : 1;

	/* MSZ, 8/9/17 */
	nelem_sel = 1;
        for (idim=0; idim<sds_info->rank; idim++)
        {
            if ((sds_info->rank > 1) && (idim != sds_info->row_id))
                nelem *= sds_info->dim_size[idim];
            he5_start[idim] = start[idim];
            he5_count[idim] = edge[idim];

	    /* MSZ, 8/9/17 */
	    if(start[idim] + edge[idim] > sds_info->dim_size[idim])
	    {
		fprintf(stderr, "The specified start %d and edge %d for read_sel_sds_data() is out of %d dim range %d for %s, exit...\n", start[idim], edge[idim], idim, sds_info->dim_size[idim], sds_info->name);
		exit(0);
	    }
	    nelem_sel *= (hsize_t)he5_count[idim];
        }

	/* MSZ 8/18/17, change the nelem_sel for l2g (L2G) compact data */
	if(sds_info->use_L2G_max_nobs == 1)
	{
	    if(sds_info->rank > 1)
	    {
		fprintf(stderr, "for L2G compact data %s, rank should NOT > 1 in read_sel_sds_data(), exit...\n", sds_info->name);
		exit(0);
	    }

	    if(edge[0] > sds_info->L2G_max_nobs)
	    {
		fprintf(stderr, "edge exceeds L2G_max_nobs for %s, exit...\n", sds_info->name);
		exit(0);
	    }

	    nelem_sel = sds_info->L2G_max_nobs;
	}

        if ((in_h5memspace_id = H5Screate_simple(sds_info->rank, he5_count, NULL)) < 0)
        {
              fprintf (stderr,"Failed to create in_h5memspace_id for SDS \"%s\". in get_sds_data \n", sds_info->name);
	      return -1;
        }

        if (H5Sselect_hyperslab (sds_info->h5space_id, H5S_SELECT_SET, he5_start, NULL, he5_count, NULL) < 0)
        {
              fprintf(stderr,"Can't select hyperslab for SDS \"%s\". in get_sds_data\n", sds_info->name);
              H5Sclose(sds_info->h5space_id);
	      return -1;
        }
            
        if (H5Dread((hid_t)sds_info->sds_id, sds_info->h5dmtype, in_h5memspace_id, sds_info->h5space_id, H5P_DEFAULT, data) == FAIL)
        {
              fprintf(stderr, "Cannot read the SDS data %s read_sel_sds_data\n", 
              sds_info->name);
	      return -1;
        }
        
        /* if (sds_info->sds_id != -1) H5Dclose(sds_info->sds_id); */
        free((size_t *)he5_start);
        free((size_t *)he5_count);
    }
    else
    {
        if (sds_info->sds_id == -1)
        {
            if ((sds_info->sds_index = SDnametoindex(sds_info->sd_id, sds_info->name)) == FAIL)
            {
              fprintf(stderr, "Cannot SDnametoindex for SDS %s read_sel_sds_data\n", sds_info->name);
	      return -1;
            }
            else if ((sds_info->sds_id = SDselect(sds_info->sd_id, sds_info->sds_index)) == FAIL)
            {
              fprintf(stderr, "Cannot select the SDS %s read_sel_sds_data\n", sds_info->name);
	      return -1;
            } 
        }

        if (SDgetinfo(sds_info->sds_id, sds_info->name, &sds_info->rank, sds_info->dim_size, &sds_info->data_type,
        &nattr) != FAIL)
        {
            nelem = (sds_info->rank == 1) ? sds_info->dim_size[0] : 1;

	    /* MSZ, 8/9/17 */
	    nelem_sel = 1;            
            for (idim=0; idim<sds_info->rank; idim++)
	    {
                if ((sds_info->rank > 1) && (idim != sds_info->row_id))
                    nelem *= sds_info->dim_size[idim];

	        /* MSZ, 8/9/17 */
	        if(start[idim] + edge[idim] > sds_info->dim_size[idim])
	        {
		    fprintf(stderr, "The specified start %d and edge %d for read_sel_sds_data() is out of %d dim range %d for %s, exit...\n", start[idim], edge[idim], idim, sds_info->dim_size[idim], sds_info->name);
		    exit(0);
	        }
	        nelem_sel *= (hsize_t)edge[idim];
	    }

	    /* MSZ 8/18/17, change the nelem_sel for l2g (L2G) compact data */
	    if(sds_info->use_L2G_max_nobs == 1)
	    {
		if(sds_info->rank > 1)
		{
		    fprintf(stderr, "for L2G compact data %s, rank should NOT > 1 in read_sel_sds_data(), exit...\n", sds_info->name);
		    exit(0);
		}

	        if(edge[0] > sds_info->L2G_max_nobs)
	        {
		   fprintf(stderr, "edge exceeds L2G_max_nobs for %s, exit...\n", sds_info->name);
		   exit(0);
	        }

		nelem_sel = sds_info->L2G_max_nobs;
	    }
                    
            /* fprintf(stderr, "rank:%d edge:%d,%d\n", sds_info->rank, edge[0], edge[1]);
            fprintf(stderr, "start:%d,%d\n", start[0], start[1]);
            fprintf(stderr, "sds_id = %d\n", sds_info->sds_id); */
            if (SDreaddata(sds_info->sds_id, start, NULL, edge, data) == FAIL)
	    {
              fprintf(stderr, "Cannot read the SDS data %s read_sel_sds_data\n", sds_info->name);
	      return -1;
	    }
        }
        else return -1;
    }

    return 0;
}

int set_sel_sds_data(sds_t *sds_info, int32 *start, int32 *edge, void *data)
{
    int idim;
    int status;
    int32 nattr;
    
    /* H5 addition */
    H5SDS *h5_sds = NULL;
    hid_t grp_id;
    hsize_t *he5_start = NULL;
    hsize_t *he5_count = NULL;
    hid_t h5memspace_id;
    
    status = 1;
    
    /* fprintf(stderr, "sds_info->H5IN = %d, sds_info->H5OUT = %d\n", sds_info->H5IN, sds_info->H5OUT); */
    
    if (sds_info->H5OUT)
    {
        if (sds_info->sds_id == -1)
        {
            if ((h5_sds = h5_dsetget(sds_info, sds_info->name)) == NULL)
            {
                fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_info->name);
                return -1;
            }
            if (!h5_sds->sds_found)
            {
                fprintf(stderr, "Cannot find SDS %s in set_sel_sds_data\n", sds_info->name);
                status = -1;
            }
            if ((grp_id = H5Gopen2(sds_info->sd_id, h5_sds->abs_gnames,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to open H5 group %s in set_sel_sds_data()\n",h5_sds->abs_gnames);
              return -1;
            }
            
            if ((sds_info->sds_id = (int32)H5Dopen2(grp_id,h5_sds->name,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to retrieve dataset %s in group %s set_sel_sds_data()\n",h5_sds->name, h5_sds->abs_gnames);
              return -1;
            }
            
            if (h5_dsetinfo(h5_sds, sds_info) == FAIL)
            {
                fprintf(stderr,"Failed to get h5_dsetinfo for sdsname %s\n", sds_info->name);
                H5Gclose(grp_id);
                free(h5_sds);
                return -1;
            }
            
            if (grp_id >= 0) H5Gclose(grp_id);
            if (h5_sds->h5_dims) free(h5_sds->h5_dims);
            if (h5_sds) free(h5_sds);
        }
        
        he5_start= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        he5_count= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        if (he5_start == NULL || he5_count == NULL)
        {
            fprintf(stderr, "Cannot allocate memory for start/edge in set_sel_sds_data\n");
            return -1;
        }
        for (idim=0; idim<sds_info->rank; idim++)
        {
            he5_start[idim] = start[idim];
            he5_count[idim] = edge[idim];
        }
        if ((h5memspace_id = H5Screate_simple(sds_info->rank, he5_count, NULL)) < 0)
        {
            fprintf(stderr,"Creation of h5memspace_id failed in set_sel_sds_data() sds_rw.c\n");
            H5Sclose(sds_info->h5space_id);
            free((size_t *)he5_start);
            free((size_t *)he5_count);
            return -1;
        }
        
        if (H5Sselect_hyperslab (sds_info->h5space_id, H5S_SELECT_SET, he5_start, NULL, he5_count,NULL) < 0)
        {
            fprintf(stderr,"Can't select hyperslab for SDS \"%s\". in set_sel_sds_data() sds_rw.c\n", sds_info->name);
            H5Sclose(sds_info->h5space_id);
            free((size_t *)he5_start);
            free((size_t *)he5_count);
          return -1;
        }
        
        if (H5Dwrite ((hid_t)sds_info->sds_id, sds_info->h5dmtype, h5memspace_id, sds_info->h5space_id, H5P_DEFAULT, (VOIDP)data) < 0)
        {
            fprintf(stderr, "Error writing output to the HDF file in set_sel_sds_data() sds_rw.c\n");
            free((size_t *)he5_start);
            free((size_t *)he5_count);
            return -1;
            /* fprintf(stderr, "sds_id = %d h5dmtype = %d h5memspace_id = %d h5space_id = %d\n",
                   out_sds_info.sds_id, out_sds_info.h5dmtype, out_h5memspace_id, out_sds_info.h5space_id); */
        }
        
        /* if (sds_info->sds_id != -1) H5Dclose(sds_info->sds_id); */
        free((size_t *)he5_start);
        free((size_t *)he5_count);
        return 0; 
    }
    else
    {

        if (sds_info->sds_id == -1)
        {
            if ((sds_info->sds_index = SDnametoindex(sds_info->sd_id, sds_info->name)) == FAIL)
            {
              fprintf(stderr, "Cannot find SDS %s in set_sel_sds_data\n", sds_info->name);
              status = -1;
            }
            else if ((sds_info->sds_id = SDselect(sds_info->sd_id, sds_info->sds_index)) == FAIL)
            {
              fprintf(stderr, "Cannot select the SDS %s in set_sel_sds_data\n", sds_info->name);
              status = -1;
            } 
            if (status == -1)
              return status;
        }

        if (SDwritedata(sds_info->sds_id, start, NULL, edge, data) == FAIL)
        {
            fprintf(stderr, "Cannot write the SDS data %s in set_sel_sds_data\n", sds_info->name);
            status = -1;
        }
        if (status == -1) return status;
        else return 0;
    }
}

int set_sel_sds_data_compression(sds_t *sds_info, int32 *start, int32 *edge, int comp_chunk_rows,  void *data)
{
    /* MSZ 1/21/2020, modified from set_sel_sds_data() and perform internal compression just for HDF4 output. 
     * 		     The chunk size is that chunk_cols for the entire cols whereas chunk_rows is specified by
     * 		     comp_chunk_rows. For other non-xy dims, the chunk size is set by default to be 1.
     * 		     This is because QA tools usually write data in one full line or many lines.
     * 		     Note that sds_info rank and dim_size[] must be known before using the function. */

    /* Note that for internal compression, sds_info->rank and sds_info->dim_size must be known besides dim_sizes */

    /* Note that this is used with open_sds_compression() */
  
    int idim;
    int status;
    int32 nattr;
    int i;
    
    /* H5 addition */
    H5SDS *h5_sds = NULL;
    hid_t grp_id;
    hsize_t *he5_start = NULL;
    hsize_t *he5_count = NULL;
    hid_t h5memspace_id;

    HDF_CHUNK_DEF c_def;
    
    status = 1;
    
    /* fprintf(stderr, "sds_info->H5IN = %d, sds_info->H5OUT = %d\n", sds_info->H5IN, sds_info->H5OUT); */
    
    if (sds_info->H5OUT)
    {
        if (sds_info->sds_id == -1)
        {
            if ((h5_sds = h5_dsetget(sds_info, sds_info->name)) == NULL)
            {
                fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_info->name);
                return -1;
            }
            if (!h5_sds->sds_found)
            {
                fprintf(stderr, "Cannot find SDS %s in set_sel_sds_data_compression\n", sds_info->name);
                status = -1;
            }
            if ((grp_id = H5Gopen2(sds_info->sd_id, h5_sds->abs_gnames,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to open H5 group %s in set_sel_sds_data_compression()\n",h5_sds->abs_gnames);
              return -1;
            }
            
            if ((sds_info->sds_id = (int32)H5Dopen2(grp_id,h5_sds->name,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to retrieve dataset %s in group %s set_sel_sds_data_compression()\n",h5_sds->name, h5_sds->abs_gnames);
              return -1;
            }
            
            if (h5_dsetinfo(h5_sds, sds_info) == FAIL)
            {
                fprintf(stderr,"Failed to get h5_dsetinfo for sdsname %s\n", sds_info->name);
                H5Gclose(grp_id);
                free(h5_sds);
                return -1;
            }
            
            if (grp_id >= 0) H5Gclose(grp_id);
            if (h5_sds->h5_dims) free(h5_sds->h5_dims);
            if (h5_sds) free(h5_sds);
        }
        
        he5_start= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        he5_count= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        if (he5_start == NULL || he5_count == NULL)
        {
            fprintf(stderr, "Cannot allocate memory for start/edge in set_sel_sds_data_compression\n");
            return -1;
        }
        for (idim=0; idim<sds_info->rank; idim++)
        {
            he5_start[idim] = start[idim];
            he5_count[idim] = edge[idim];
        }
        if ((h5memspace_id = H5Screate_simple(sds_info->rank, he5_count, NULL)) < 0)
        {
            fprintf(stderr,"Creation of h5memspace_id failed in set_sel_sds_data_compression() sds_rw.c\n");
            H5Sclose(sds_info->h5space_id);
            free((size_t *)he5_start);
            free((size_t *)he5_count);
            return -1;
        }
        
        if (H5Sselect_hyperslab (sds_info->h5space_id, H5S_SELECT_SET, he5_start, NULL, he5_count,NULL) < 0)
        {
            fprintf(stderr,"Can't select hyperslab for SDS \"%s\". in set_sel_sds_data_compression() sds_rw.c\n", sds_info->name);
            H5Sclose(sds_info->h5space_id);
            free((size_t *)he5_start);
            free((size_t *)he5_count);
          return -1;
        }
        
        if (H5Dwrite ((hid_t)sds_info->sds_id, sds_info->h5dmtype, h5memspace_id, sds_info->h5space_id, H5P_DEFAULT, (VOIDP)data) < 0)
        {
            fprintf(stderr, "Error writing output to the HDF file in set_sel_sds_data_compression() sds_rw.c\n");
            free((size_t *)he5_start);
            free((size_t *)he5_count);
            return -1;
            /* fprintf(stderr, "sds_id = %d h5dmtype = %d h5memspace_id = %d h5space_id = %d\n",
                   out_sds_info.sds_id, out_sds_info.h5dmtype, out_h5memspace_id, out_sds_info.h5space_id); */
        }
        
        /* if (sds_info->sds_id != -1) H5Dclose(sds_info->sds_id); */
        free((size_t *)he5_start);
        free((size_t *)he5_count);
        return 0; 
    }
    else
    {

        if (sds_info->sds_id == -1)
        {
            if ((sds_info->sds_index = SDnametoindex(sds_info->sd_id, sds_info->name)) == FAIL)
            {
              fprintf(stderr, "Cannot find SDS %s in set_sel_sds_data_compression\n", sds_info->name);
              status = -1;
            }
            else if ((sds_info->sds_id = SDselect(sds_info->sd_id, sds_info->sds_index)) == FAIL)
            {
              fprintf(stderr, "Cannot select the SDS %s in set_sel_sds_data_compression\n", sds_info->name);
              status = -1;
            } 
            if (status == -1)
              return status;

	    /* MSZ 1/21/2020 */
	    if(sds_info->rank != 1)
	    {
	      /* initialize the compression parameters */
	      for(i = 0; i < MAX_LDOPE_DIMS; i++)
		c_def.chunk_lengths[i] = c_def.comp.chunk_lengths[i] = 1;

	      /* use the below method so that there is no need to use sds_info->interleave in case it is not assigned */
	      if(sds_info->rank == 2 || sds_info->dim_size[0] > sds_info->dim_size[sds_info->rank - 1])   /* 2D or BIP */
	      {
		/* set rows and cols for the chunk size */
		c_def.chunk_lengths[0] = c_def.comp.chunk_lengths[0] = comp_chunk_rows;
		c_def.chunk_lengths[1] = c_def.comp.chunk_lengths[1] = edge[1];
	      }
	      else                                                                                        /* BSQ */
	      {
	        /* set rows and cols for the chunk size */
	        c_def.chunk_lengths[sds_info->rank - 2] = c_def.comp.chunk_lengths[sds_info->rank - 2] = comp_chunk_rows;
	        c_def.chunk_lengths[sds_info->rank - 1] = c_def.comp.chunk_lengths[sds_info->rank - 1] = edge[sds_info->rank - 1];
	      }

              c_def.comp.comp_type = COMP_CODE_DEFLATE;                   /* GZIP */
              c_def.comp.cinfo.deflate.level = 6;                         /* compression level */

	      status = SDsetchunk(sds_info->sds_id, c_def, HDF_CHUNK | HDF_COMP);
	
	      if (status == -1)
                 fprintf(stderr, "Error in SDsetchunk in set_sel_sds_data_compression(), maybe info. on rank, or dim_size is missing. Or output HDF SDS exists!\n");
	    }
	    else   /* for 1D compression */
	    {
	      c_def.chunk_lengths[0] = c_def.comp.chunk_lengths[0] = sds_info->dim_size[0];
	      c_def.comp.comp_type = COMP_CODE_DEFLATE;                   /* GZIP */
	      c_def.comp.cinfo.deflate.level = 6;                         /* compression level */

	      status = SDsetchunk(sds_info->sds_id, c_def, HDF_CHUNK | HDF_COMP);
	
	      if (status == -1)
                 fprintf(stderr, "Error in SDsetchunk in set_sel_sds_data_compression(), maybe info. on rank, or dim_size is missing. Or output HDF SDS exists!\n");
	      
	    }

	    if (status == -1)
	      return status;
        }

        if (SDwritedata(sds_info->sds_id, start, NULL, edge, data) == FAIL)
        {
            fprintf(stderr, "Cannot write the SDS data %s in set_sel_sds_data_compression\n", sds_info->name);
            status = -1;
        }
        if (status == -1) return status;
        else return 0;
    }
}

int set_sel_l2g_sds_2D_data_chunked_gzip(sds_t *sds_info, int32 *start, int32 *edge, void *data)
{
    /* MSZ 7/26/2018, modified from set_sel_sds_data to save the room for L2G laryers' output to dicrease the data volume */
    int idim;
    int status;
    int32 nattr;
    int32 row_len_compression = 0;
    
    /* H5 addition */
    H5SDS *h5_sds = NULL;
    hid_t grp_id;
    hsize_t *he5_start = NULL;
    hsize_t *he5_count = NULL;
    hid_t h5memspace_id;
    
    status = 1;

    /* MSZ 7/26/2018, HDF4 internal compression */
    HDF_CHUNK_DEF c_def; /* Chunking definitions */
    
    /* fprintf(stderr, "sds_info->H5IN = %d, sds_info->H5OUT = %d\n", sds_info->H5IN, sds_info->H5OUT); */
    
    if (sds_info->H5OUT)
    {
        if (sds_info->sds_id == -1)
        {
            if ((h5_sds = h5_dsetget(sds_info, sds_info->name)) == NULL)
            {
                fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_info->name);
                return -1;
            }
            if (!h5_sds->sds_found)
            {
                fprintf(stderr, "Cannot find SDS %s in set_sel_sds_data\n", sds_info->name);
                status = -1;
            }
            if ((grp_id = H5Gopen2(sds_info->sd_id, h5_sds->abs_gnames,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to open H5 group %s in set_sel_sds_data()\n",h5_sds->abs_gnames);
              return -1;
            }
            
            if ((sds_info->sds_id = (int32)H5Dopen2(grp_id,h5_sds->name,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"H5: Failed to retrieve dataset %s in group %s set_sel_sds_data()\n",h5_sds->name, h5_sds->abs_gnames);
              return -1;
            }
            
            if (h5_dsetinfo(h5_sds, sds_info) == FAIL)
            {
                fprintf(stderr,"Failed to get h5_dsetinfo for sdsname %s\n", sds_info->name);
                H5Gclose(grp_id);
                free(h5_sds);
                return -1;
            }
            
            if (grp_id >= 0) H5Gclose(grp_id);
            if (h5_sds->h5_dims) free(h5_sds->h5_dims);
            if (h5_sds) free(h5_sds);
        }
        
        he5_start= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        he5_count= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
        if (he5_start == NULL || he5_count == NULL)
        {
            fprintf(stderr, "Cannot allocate memory for start/edge in set_sel_sds_data\n");
            return -1;
        }
        for (idim=0; idim<sds_info->rank; idim++)
        {
            he5_start[idim] = start[idim];
            he5_count[idim] = edge[idim];
        }
        if ((h5memspace_id = H5Screate_simple(sds_info->rank, he5_count, NULL)) < 0)
        {
            fprintf(stderr,"Creation of h5memspace_id failed in set_sel_sds_data() sds_rw.c\n");
            H5Sclose(sds_info->h5space_id);
            free((size_t *)he5_start);
            free((size_t *)he5_count);
            return -1;
        }
        
        if (H5Sselect_hyperslab (sds_info->h5space_id, H5S_SELECT_SET, he5_start, NULL, he5_count,NULL) < 0)
        {
            fprintf(stderr,"Can't select hyperslab for SDS \"%s\". in set_sel_sds_data() sds_rw.c\n", sds_info->name);
            H5Sclose(sds_info->h5space_id);
            free((size_t *)he5_start);
            free((size_t *)he5_count);
          return -1;
        }
        
        if (H5Dwrite ((hid_t)sds_info->sds_id, sds_info->h5dmtype, h5memspace_id, sds_info->h5space_id, H5P_DEFAULT, (VOIDP)data) < 0)
        {
            fprintf(stderr, "Error writing output to the HDF file in set_sel_sds_data() sds_rw.c\n");
            free((size_t *)he5_start);
            free((size_t *)he5_count);
            return -1;
            /* fprintf(stderr, "sds_id = %d h5dmtype = %d h5memspace_id = %d h5space_id = %d\n",
                   out_sds_info.sds_id, out_sds_info.h5dmtype, out_h5memspace_id, out_sds_info.h5space_id); */
        }
        
        /* if (sds_info->sds_id != -1) H5Dclose(sds_info->sds_id); */
        free((size_t *)he5_start);
        free((size_t *)he5_count);
        return 0; 
    }
    else
    {

        if (sds_info->sds_id == -1)
        {
            if ((sds_info->sds_index = SDnametoindex(sds_info->sd_id, sds_info->name)) == FAIL)
            {
              fprintf(stderr, "Cannot find SDS %s in set_sel_sds_data\n", sds_info->name);
              status = -1;
            }
            else if ((sds_info->sds_id = SDselect(sds_info->sd_id, sds_info->sds_index)) == FAIL)
            {
              fprintf(stderr, "Cannot select the SDS %s in set_sel_sds_data\n", sds_info->name);
              status = -1;
            } 
            if (status == -1)
              return status;
        }

	/* MSZ 7/26/2018, L2G for each layer is just 2D */
	row_len_compression = 100;
	if(sds_info->dim_size[0] < 100) row_len_compression = sds_info->dim_size[0];
	c_def.chunk_lengths[0] = c_def.comp.chunk_lengths[0] = row_len_compression; /* Rows */
	c_def.chunk_lengths[1] = c_def.comp.chunk_lengths[1] = sds_info->dim_size[1]; /* Cols */
	
	c_def.comp.comp_type = COMP_CODE_DEFLATE;                   /* GZIP */
	c_def.comp.cinfo.deflate.level = 6;                         /* compression level */
	
	status = SDsetchunk(sds_info->sds_id, c_def, HDF_CHUNK | HDF_COMP);
	if (status == -1)
	   fprintf(stderr, "Error in SDsetchunk in open_sds_compression(), maybe output HDF exists!\n");

        if (SDwritedata(sds_info->sds_id, start, NULL, edge, data) == FAIL)
        {
            fprintf(stderr, "Cannot write the SDS data %s in set_sel_sds_data\n", sds_info->name);
            status = -1;
        }
        if (status == -1) return status;
        else return 0;
    }
}

int open_sd(char *fname, sds_t *sds_info, char open_t)
{
    /* MSZ 1/23/2018. a general function to open HDF4/HDF5
     * either for reading or writing purposes.
     * Note that for reading HDF, sds_info->H5IN must be known;
     * for output HDF4/HDF5, sds_info->H5OUT must be set */
    int status = 1;
    hid_t test;
    int HDF_ver = 0;

    if(fname != NULL)
      sds_info->file_name = fname;   /* file_name just shares the same string of the fname */

    if (open_t == 'R')
    {
	/* open HDF file if it is not open */
        if (sds_info->sd_id == -1)
        {
	  if(fname != NULL)
	  {
	    HDF_ver = HDF_version_check(fname);
	    if(HDF_ver == 4)
		sds_info->H5IN = 0;
	    else if(HDF_ver == 5)
		sds_info->H5IN = 1;
	    else
	    {
		fprintf(stderr, "%s is neither HDF4 or HDF5\n", fname);
		status = -1;
	    }

	    /* open HDF5 or HDF4 file */
            if (sds_info->H5IN && (sds_info->sd_id = (int32)H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
            {
		sds_info->sd_id = -1;
                fprintf(stderr, "Cannot open the hdf5 file %s\n", fname);
                status = -1;
            } 
            else if (!sds_info->H5IN && (sds_info->sd_id = SDstart(fname, DFACC_READ)) == FAIL)
            {
                fprintf(stderr, "Cannot open the hdf file %s\n", fname);
                status = -1;
            }
	  }
	  else
	  {
	    fprintf(stderr, "fname is not provided, exit...\n");
	    exit(0);
	  }
        }
    }
    else
    {
	/* MSZ 2/10/2018.  To correct a potential bug that output HDF already exists */
	FILE *fp = NULL;
	
	if(fname != NULL && (fp = fopen(fname, "r")) != NULL)
	{
	   fclose(fp);

	   int HDF_ver = 0;
	   HDF_ver = HDF_version_check(fname);

	   if(HDF_ver == 4) sds_info->H5OUT = 0;
	   else if(HDF_ver == 5) sds_info->H5OUT = 1;
	   else status = -1;
	}

        if (sds_info->sd_id == -1)
        {
	  if(fname != NULL)
	  {
	    /* MSZ 1/23/2018, update this section by adding the WRITE ACC so that it can be used for creating HDF4/HDF5 */
            if (sds_info->H5OUT)
            {
	      if((sds_info->sd_id = (int32)H5Fopen(fname, H5F_ACC_RDWR, H5P_DEFAULT)) < 0)
	        if((sds_info->sd_id = (int32)H5Fcreate(fname, H5F_ACC_EXCL, H5P_DEFAULT, H5P_DEFAULT)) < 0)
	        {
		  sds_info->sd_id = -1;
                  fprintf(stderr, "Cannot create the hdf file %s due to its existence.\n", fname);
                  status = -1;
	        }
            }
            else if (!sds_info->H5OUT)
	    {
	      if ((sds_info->sd_id = SDstart(fname, DFACC_WRITE)) == FAIL)
	      {
		if ((sds_info->sd_id = SDstart(fname, DFACC_CREATE)) == FAIL)
		{
		  fprintf(stderr, "Cannot create the hdf file %s\n", fname);
		  status = -1;
		}
	      }
            }
	  }
	  else
	  {
	    fprintf(stderr, "fname is not provided, exit...\n");
	    exit(0);
	  }
        }
    }
    
    return status;
}

int open_sds(char *fname, sds_t *sds_info, char open_t)
{
    /* MSZ 1/23/2018. update it so that it can be used for general purpose
     * of creating HDF4/HDF5 */
    int status;
    int p1, len;
    int32 sds_index;
    char sds_name[MAX_SDS_NAME_LEN];
    int HDF_ver = 0;
  
    /* H5 addition */
    H5SDS *h5_sds = NULL;
    hid_t grp_id;
    size_t    sh4_amemsize;/* size of attribute data type in memory. */
    size_t    sh4_asize;  /* size of attribute data type in disk. */
    hid_t     sh5_amemtype; /* HDF5 attribute data type in memory */
    hsize_t  *he5_dims = NULL;
    int i;
    
    status = 1;
    if (open_t == 'R')
    {
	/* open HDF file if it is not open */
        if (sds_info->sd_id == -1 && fname != NULL)
        {
	    sds_info->file_name = (char*)fname;

	    HDF_ver = HDF_version_check(fname);
	    if(HDF_ver == 4)
		sds_info->H5IN = 0;
	    else if(HDF_ver == 5)
		sds_info->H5IN = 1;
	    else
	    {
		fprintf(stderr, "%s is neither HDF4 or HDF5\n", fname);
	        status = -1;
	    }

	    /* open HDF5 or HDF4 file */
            if (sds_info->H5IN && (sds_info->sd_id = (int32)H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
            {
		sds_info->sd_id = -1;
                fprintf(stderr, "Cannot open the hdf file %s\n", fname);
                status = -1;
            } 
            else if (!sds_info->H5IN && (sds_info->sd_id = SDstart(fname, DFACC_READ)) == FAIL)
            {
		sds_info->sd_id = -1;
                fprintf(stderr, "Cannot open the hdf file %s\n", fname);
                status = -1;
            }
        }

	/* get sds_info if HDF file is open */
        if (status != -1)
        {
            if (sds_info->H5IN)
            {
		/* MSZ 1/23/2018. Note that h5_dsetget() just works for DATASETS of HDF5 */
                if ((h5_sds = h5_dsetget(sds_info, sds_info->name)) == NULL)
                {
                    fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_info->name);
                    return -1;
                }
                if (!h5_sds->sds_found)
                {
                  if ((p1 = charpos(sds_info->name, '.', 0)) != -1)
                  {
                        free(h5_sds);
                        strmid(sds_info->name, 0, p1, sds_name);
                        if ((h5_sds = h5_dsetget(sds_info, sds_name)) == NULL)
                        {
                            fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_name);
                            return -1;
                        }
                  }
                }
                if (!h5_sds->sds_found)
                {
                    fprintf(stderr, "Cannot find the SDS %s in file %s\n", sds_name, fname);
                    status = -1;
                }
                if ((grp_id = H5Gopen2(sds_info->sd_id, h5_sds->abs_gnames,H5P_DEFAULT)) < 0)
                {
                    fprintf(stderr,"H5: Failed to open H5 group %s in open_sds()\n",h5_sds->abs_gnames);
                    free(h5_sds);
                    return -1;
                }
                
                if ((sds_info->sds_id = (int32)H5Dopen2(grp_id,h5_sds->name,H5P_DEFAULT)) < 0)
                {
                    fprintf(stderr,"H5: Failed to retrieve dataset %s in group %s open_sds()\n",h5_sds->name, h5_sds->abs_gnames);
                    free(h5_sds);
                    return -1;
                }
                
                sds_info->h5dtype = H5Dget_type((hid_t)sds_info->sds_id);
                sds_info->h5dmtype = H5Tget_native_type(sds_info->h5dtype,H5T_DIR_DEFAULT);
                he5_dims= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
                for (i = 0; i < sds_info->rank; i++)
                    he5_dims[i] = sds_info->dim_size[i];
                if ((sds_info->h5space_id = H5Screate_simple(sds_info->rank, he5_dims, NULL)) < 0)
                {
                    fprintf(stderr, "Cannot create dataspace for new SDS %s\n", sds_info->name);
                    status = -1; 
                }
                if (he5_dims) free(he5_dims);
            }
            else
            {
                if ((sds_index = SDnametoindex(sds_info->sd_id, sds_info->name)) == FAIL)
                {
                    if ((p1 = charpos(sds_info->name, '.', 0)) != -1)
                    {
                          strmid(sds_info->name, 0, p1, sds_name);
                          sds_index = SDnametoindex(sds_info->sd_id, sds_name);
                    }
                }
                if (sds_index == FAIL)
                {
                    fprintf(stderr, "Cannot find the SDS %s in file %s\n", sds_name, fname);
                    status = -1;
                }
                else if ((sds_info->sds_id = SDselect(sds_info->sd_id, sds_index)) == FAIL)
                {
                    fprintf(stderr, "Cannot open the SDS %s in file %s\n", sds_name, fname);
                    status = -1;
                }
            }
        }
    }
    else  /* write or create */
    {
	/* MSZ 2/10/2018.  To correct a potential bug that output HDF already exists */
	FILE *fp = NULL;
	
	if(fname != NULL && (fp = fopen(fname, "r")) != NULL)
	{

	   sds_info->file_name = (char*)fname;

	   fclose(fp);

	   int HDF_ver = 0;
	   HDF_ver = HDF_version_check(fname);

	   if(HDF_ver == 4) sds_info->H5OUT = 0;
	   else if(HDF_ver == 5) sds_info->H5OUT = 1;
	   else status = -1;
	}

        if (sds_info->sd_id == -1)
        {
	    /* MSZ 1/23/2018, update this section by adding the WRITE ACC so that it can be used for creating HDF4/HDF5 */
            if (sds_info->H5OUT)
            {
              if((sds_info->sd_id = (int32)H5Fopen(fname, H5F_ACC_RDWR, H5P_DEFAULT)) < 0)
                if((sds_info->sd_id = (int32)H5Fcreate(fname, H5F_ACC_EXCL, H5P_DEFAULT, H5P_DEFAULT)) < 0)
	        {
		  sds_info->sd_id = -1;
                  fprintf(stderr, "Cannot create the hdf file %s\n", fname);
                  status = -1;
	        }
            }
            else if (!sds_info->H5OUT)
	    {
	      if ((sds_info->sd_id = SDstart(fname, DFACC_WRITE)) == FAIL)
		if ((sds_info->sd_id = SDstart(fname, DFACC_CREATE)) == FAIL)
		{
		  fprintf(stderr, "Cannot create the hdf file %s\n", fname);
		  status = -1;
		}
            }

        }
        
        if (status != -1)
        {
            if (sds_info->H5OUT)
            {
		if((sds_info->sds_id = (int32)(H5Dopen2((hid_t)sds_info->sd_id, sds_info->name, H5P_DEFAULT))) < 0) /* MSZ 4/5/2018, not create SDS if it exists to avoid duplicated SDS */ 
		{
                  he5_dims= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
                  for (i = 0; i < sds_info->rank; i++)
                    he5_dims[i] = sds_info->dim_size[i];
                  /* first create a new dataspace */
                  if ((sds_info->h5space_id = H5Screate_simple(sds_info->rank, he5_dims, NULL)) < 0)
                  {
                    fprintf(stderr, "Cannot create dataspace for new SDS %s\n", sds_info->name);
                    status = -1; 
                  }
                
                  convH4toH5(sds_info->data_type, &sh5_amemtype,&sh4_amemsize,&sh4_asize,&sds_info->h5dtype);
                  sds_info->h5dmtype = H5Tget_native_type(sds_info->h5dtype,H5T_DIR_DEFAULT);
                
                  if ((sds_info->sds_id = (int32)H5Dcreate2((hid_t)sds_info->sd_id, sds_info->name,sds_info->h5dtype, sds_info->h5space_id, H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT)) < 0 )
                  {
                    fprintf(stderr, "Cannot create the SDS %s\n", sds_info->name);
                    status = -1; 
                  }
		  free(he5_dims);
		}
                else  /* just write long_name to sds_id */
                {
                    len = (int)strlen(sds_info->name) + 1;
                    if (setH5Attr(DFNT_CHAR8,"long_name",len,(VOIDP)sds_info->name, (hid_t)sds_info->sds_id) == FAIL)
                    fprintf(stderr, "Could not write the attribute 'long_name' in %s\n", sds_info->name);
                }
            }
            else
            {
		/* MSZ 8/11/17 change input HDF5 long absolute sds_name to the short name for HDF4 output */
		if(strstr(sds_info->name, "/"))
		{
		    /* local variables */
		    char *grp_name = NULL, *new_name = NULL;

		    HDF5_split_abs_name_into_path_name(sds_info->name, &grp_name, &new_name);

		    /* clean sds_info->name */
		    for(i=0;i<MAX_SDS_NAME_LEN;i++)
			sds_info->name[i] = '\0';

		    /* refill sds_info->name */
		    strcpy(sds_info->name, new_name);

		    /* free memeory */
		    if(grp_name) free(grp_name);
		    if(new_name) free(new_name);
		}

		if((sds_info->sds_index = SDnametoindex(sds_info->sd_id, sds_info->name)) == FAIL)   /* MSZ 4/5/2018, not create SDS if it exists to avoid duplicated SDS */
		{
                  if ((sds_info->sds_id = SDcreate(sds_info->sd_id, sds_info->name, 
                    sds_info->data_type, sds_info->rank, sds_info->dim_size)) == FAIL)
                  {
                    fprintf(stderr, "Cannot create the SDS %d %s %d %d %d %d\n", sds_info->sd_id, sds_info->name, sds_info->data_type, sds_info->rank, sds_info->dim_size[0], sds_info->dim_size[1]);
                    status = -1;
                  }
		}
                else /* just write long_name to sds_id */
                {
		    sds_info->sds_id = SDselect(sds_info->sd_id, sds_info->sds_index);
                    len = (int)strlen(sds_info->name) + 1;
                    if (SDsetattr(sds_info->sds_id, "long_name", DFNT_CHAR8, len, 
                        (VOIDP)sds_info->name) == FAIL)
                    fprintf(stderr, "Could not write the attribute 'long_name' in %s\n", sds_info->name);
                }
            }
        }
    }
    
    if (sds_info->H5OUT)
    {
        H5Gclose(grp_id);
        if (h5_sds) free(h5_sds);
    }

    return status;
}

int open_sds_compression(char *fname, sds_t *sds_info, int comp_chunk_rows, char open_t)
{
    /* MSZ 1/22/2020, enhanced based on open_sds() so that it can perform internal compression 
     * for write or create condition (open_t != 'R') and HDF4 output.
     * Note that this is used with set_sel_sds_data_compression() */

    /* Note that for internal compression, sds_info->rank and sds_info->dim_size must be known besides dim_size */

    /* MSZ 1/23/2018. update it so that it can be used for general purpose
     * of creating HDF4/HDF5 */


    int status;
    int p1, len;
    int32 sds_index;
    char sds_name[MAX_SDS_NAME_LEN];
    int HDF_ver = 0;
  
    /* H5 addition */
    H5SDS *h5_sds = NULL;
    hid_t grp_id;
    size_t    sh4_amemsize;/* size of attribute data type in memory. */
    size_t    sh4_asize;  /* size of attribute data type in disk. */
    hid_t     sh5_amemtype; /* HDF5 attribute data type in memory */
    hsize_t  *he5_dims = NULL;
    int i;

    HDF_CHUNK_DEF c_def;
    
    status = 1;
    if (open_t == 'R')
    {
	/* open HDF file if it is not open */
        if (sds_info->sd_id == -1 && fname != NULL)
        {
	    sds_info->file_name = (char*)fname;

	    HDF_ver = HDF_version_check(fname);
	    if(HDF_ver == 4)
		sds_info->H5IN = 0;
	    else if(HDF_ver == 5)
		sds_info->H5IN = 1;
	    else
	    {
		fprintf(stderr, "%s is neither HDF4 or HDF5\n", fname);
	        status = -1;
	    }

	    /* open HDF5 or HDF4 file */
            if (sds_info->H5IN && (sds_info->sd_id = (int32)H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
            {
		sds_info->sd_id = -1;
                fprintf(stderr, "Cannot open the hdf file %s\n", fname);
                status = -1;
            } 
            else if (!sds_info->H5IN && (sds_info->sd_id = SDstart(fname, DFACC_READ)) == FAIL)
            {
		sds_info->sd_id = -1;
                fprintf(stderr, "Cannot open the hdf file %s\n", fname);
                status = -1;
            }
        }

	/* get sds_info if HDF file is open */
        if (status != -1)
        {
            if (sds_info->H5IN)
            {
		/* MSZ 1/23/2018. Note that h5_dsetget() just works for DATASETS of HDF5 */
                if ((h5_sds = h5_dsetget(sds_info, sds_info->name)) == NULL)
                {
                    fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_info->name);
                    return -1;
                }
                if (!h5_sds->sds_found)
                {
                  if ((p1 = charpos(sds_info->name, '.', 0)) != -1)
                  {
                        free(h5_sds);
                        strmid(sds_info->name, 0, p1, sds_name);
                        if ((h5_sds = h5_dsetget(sds_info, sds_name)) == NULL)
                        {
                            fprintf(stderr,"failed in h5_dsetget for sdsname %s\n", sds_name);
                            return -1;
                        }
                  }
                }
                if (!h5_sds->sds_found)
                {
                    fprintf(stderr, "Cannot find the SDS %s in file %s\n", sds_name, fname);
                    status = -1;
                }
                if ((grp_id = H5Gopen2(sds_info->sd_id, h5_sds->abs_gnames,H5P_DEFAULT)) < 0)
                {
                    fprintf(stderr,"H5: Failed to open H5 group %s in open_sds_compression()\n",h5_sds->abs_gnames);
                    free(h5_sds);
                    return -1;
                }
                
                if ((sds_info->sds_id = (int32)H5Dopen2(grp_id,h5_sds->name,H5P_DEFAULT)) < 0)
                {
                    fprintf(stderr,"H5: Failed to retrieve dataset %s in group %s open_sds_compression()\n",h5_sds->name, h5_sds->abs_gnames);
                    free(h5_sds);
                    return -1;
                }
                
                sds_info->h5dtype = H5Dget_type((hid_t)sds_info->sds_id);
                sds_info->h5dmtype = H5Tget_native_type(sds_info->h5dtype,H5T_DIR_DEFAULT);
                he5_dims= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
                for (i = 0; i < sds_info->rank; i++)
                    he5_dims[i] = sds_info->dim_size[i];
                if ((sds_info->h5space_id = H5Screate_simple(sds_info->rank, he5_dims, NULL)) < 0)
                {
                    fprintf(stderr, "Cannot create dataspace for new SDS %s\n", sds_info->name);
                    status = -1; 
                }
                if (he5_dims) free(he5_dims);
            }
            else
            {
                if ((sds_index = SDnametoindex(sds_info->sd_id, sds_info->name)) == FAIL)
                {
                    if ((p1 = charpos(sds_info->name, '.', 0)) != -1)
                    {
                          strmid(sds_info->name, 0, p1, sds_name);
                          sds_index = SDnametoindex(sds_info->sd_id, sds_name);
                    }
                }
                if (sds_index == FAIL)
                {
                    fprintf(stderr, "Cannot find the SDS %s in file %s\n", sds_name, fname);
                    status = -1;
                }
                else if ((sds_info->sds_id = SDselect(sds_info->sd_id, sds_index)) == FAIL)
                {
                    fprintf(stderr, "Cannot open the SDS %s in file %s\n", sds_name, fname);
                    status = -1;
                }
            }
        }
    }
    else  /* write or create */
    {
	/* MSZ 2/10/2018.  To correct a potential bug that output HDF already exists */
	FILE *fp = NULL;
	
	if(fname != NULL && (fp = fopen(fname, "r")) != NULL)
	{

	   sds_info->file_name = (char*)fname;

	   fclose(fp);

	   int HDF_ver = 0;
	   HDF_ver = HDF_version_check(fname);

	   if(HDF_ver == 4) sds_info->H5OUT = 0;
	   else if(HDF_ver == 5) sds_info->H5OUT = 1;
	   else status = -1;
	}

        if (sds_info->sd_id == -1)
        {
	    /* MSZ 1/23/2018, update this section by adding the WRITE ACC so that it can be used for creating HDF4/HDF5 */
            if (sds_info->H5OUT)
            {
              if((sds_info->sd_id = (int32)H5Fopen(fname, H5F_ACC_RDWR, H5P_DEFAULT)) < 0)
                if((sds_info->sd_id = (int32)H5Fcreate(fname, H5F_ACC_EXCL, H5P_DEFAULT, H5P_DEFAULT)) < 0)
	        {
		  sds_info->sd_id = -1;
                  fprintf(stderr, "Cannot create the hdf file %s\n", fname);
                  status = -1;
	        }
            }
            else if (!sds_info->H5OUT)
	    {
	      if ((sds_info->sd_id = SDstart(fname, DFACC_WRITE)) == FAIL)
		if ((sds_info->sd_id = SDstart(fname, DFACC_CREATE)) == FAIL)
		{
		  fprintf(stderr, "Cannot create the hdf file %s\n", fname);
		  status = -1;
		}
            }

        }
        
        if (status != -1)
        {
            if (sds_info->H5OUT)
            {
		if((sds_info->sds_id = (int32)(H5Dopen2((hid_t)sds_info->sd_id, sds_info->name, H5P_DEFAULT))) < 0) /* MSZ 4/5/2018, not create SDS if it exists to avoid duplicated SDS */ 
		{
                  he5_dims= (hsize_t *)calloc(sds_info->rank, sizeof(hsize_t));
                  for (i = 0; i < sds_info->rank; i++)
                    he5_dims[i] = sds_info->dim_size[i];
                  /* first create a new dataspace */
                  if ((sds_info->h5space_id = H5Screate_simple(sds_info->rank, he5_dims, NULL)) < 0)
                  {
                    fprintf(stderr, "Cannot create dataspace for new SDS %s\n", sds_info->name);
                    status = -1; 
                  }
                
                  convH4toH5(sds_info->data_type, &sh5_amemtype,&sh4_amemsize,&sh4_asize,&sds_info->h5dtype);
                  sds_info->h5dmtype = H5Tget_native_type(sds_info->h5dtype,H5T_DIR_DEFAULT);
                
                  if ((sds_info->sds_id = (int32)H5Dcreate2((hid_t)sds_info->sd_id, sds_info->name,sds_info->h5dtype, sds_info->h5space_id, H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT)) < 0 )
                  {
                    fprintf(stderr, "Cannot create the SDS %s\n", sds_info->name);
                    status = -1; 
                  }
		  free(he5_dims);
		}
                else  /* just write long_name to sds_id */
                {
                    len = (int)strlen(sds_info->name) + 1;
                    if (setH5Attr(DFNT_CHAR8,"long_name",len,(VOIDP)sds_info->name, (hid_t)sds_info->sds_id) == FAIL)
                    fprintf(stderr, "Could not write the attribute 'long_name' in %s\n", sds_info->name);
                }
            }
            else
            {
		/* MSZ 8/11/17 change input HDF5 long absolute sds_name to the short name for HDF4 output */
		if(strstr(sds_info->name, "/"))
		{
		    /* local variables */
		    char *grp_name = NULL, *new_name = NULL;

		    HDF5_split_abs_name_into_path_name(sds_info->name, &grp_name, &new_name);

		    /* clean sds_info->name */
		    for(i=0;i<MAX_SDS_NAME_LEN;i++)
			sds_info->name[i] = '\0';

		    /* refill sds_info->name */
		    strcpy(sds_info->name, new_name);

		    /* free memeory */
		    if(grp_name) free(grp_name);
		    if(new_name) free(new_name);
		}

		if((sds_info->sds_index = SDnametoindex(sds_info->sd_id, sds_info->name)) == FAIL)   /* MSZ 4/5/2018, not create SDS if it exists to avoid duplicated SDS */
		{
                  if ((sds_info->sds_id = SDcreate(sds_info->sd_id, sds_info->name, 
                    sds_info->data_type, sds_info->rank, sds_info->dim_size)) == FAIL)
                  {
                    fprintf(stderr, "Cannot create the SDS %d %s %d %d %d %d\n", sds_info->sd_id, sds_info->name, sds_info->data_type, sds_info->rank, sds_info->dim_size[0], sds_info->dim_size[1]);
                    status = -1;
                  }

		  if(sds_info->rank != 1)
		  {
		    /* initialize the compression parameters */
		    for(i = 0; i < MAX_LDOPE_DIMS; i++)
			c_def.chunk_lengths[i] = c_def.comp.chunk_lengths[i] = 1;

		    /* use the below method so that there is no need to use sds_info->interleave in case it is not assigned */
		    /* MSZ 1/21/2020 */
		    if(sds_info->rank == 2 || sds_info->dim_size[0] > sds_info->dim_size[sds_info->rank - 1])   /* 2D or BIP */
		    {
		      /* set rows and cols for the chunk size */
		      c_def.chunk_lengths[0] = c_def.comp.chunk_lengths[0] = comp_chunk_rows;
		      c_def.chunk_lengths[1] = c_def.comp.chunk_lengths[1] = sds_info->dim_size[1];
		    }
		    else                                                                                        /* BSQ */
		    {
		      /* set rows and cols for the chunk size */
		      c_def.chunk_lengths[sds_info->rank - 2] = c_def.comp.chunk_lengths[sds_info->rank - 2] = comp_chunk_rows;
		      c_def.chunk_lengths[sds_info->rank - 1] = c_def.comp.chunk_lengths[sds_info->rank - 1] = sds_info->dim_size[sds_info->rank - 1];
		    }

		    c_def.comp.comp_type = COMP_CODE_DEFLATE;                   /* GZIP */
		    c_def.comp.cinfo.deflate.level = 6;                         /* compression level */

		    status = SDsetchunk(sds_info->sds_id, c_def, HDF_CHUNK | HDF_COMP);
		    if (status == -1)
		       fprintf(stderr, "Error1 in SDsetchunk in open_sds_compression(), maybe info. on rank or dim_size is missing. Or output HDF SDS exists!\n");
		  }
		  else /* for 1D compression */
		  {
		    c_def.chunk_lengths[0] = c_def.comp.chunk_lengths[0] = sds_info->dim_size[0];

		    c_def.comp.comp_type = COMP_CODE_DEFLATE;                   /* GZIP */
		    c_def.comp.cinfo.deflate.level = 6;                         /* compression level */

		    status = SDsetchunk(sds_info->sds_id, c_def, HDF_CHUNK | HDF_COMP);
		    if (status == -1)
		       fprintf(stderr, "Error1 in SDsetchunk in open_sds_compression(), maybe info. on rank or dim_size is missing. Or output HDF SDS exists!\n");
		    
		  }
		}
                else /* just write long_name to sds_id */
                {
		  sds_info->sds_id = SDselect(sds_info->sd_id, sds_info->sds_index);

		  if(sds_info->rank != 1)
		  {
		    /* initialize the compression parameters */
		    for(i = 0; i < MAX_LDOPE_DIMS; i++)
			c_def.chunk_lengths[i] = c_def.comp.chunk_lengths[i] = 1;

		    /* use the below method so that there is no need to use sds_info->interleave in case it is not assigned */
		    /* MSZ 1/21/2020 */
		    if(sds_info->rank == 2 || sds_info->dim_size[0] > sds_info->dim_size[sds_info->rank - 1])   /* 2D or BIP */
		    {
		      /* set rows and cols for the chunk size */
		      c_def.chunk_lengths[0] = c_def.comp.chunk_lengths[0] = comp_chunk_rows;
		      c_def.chunk_lengths[1] = c_def.comp.chunk_lengths[1] = sds_info->dim_size[1];
		    }
		    else                                                                                        /* BSQ */
		    {
		      /* set rows and cols for the chunk size */
		      c_def.chunk_lengths[sds_info->rank - 2] = c_def.comp.chunk_lengths[sds_info->rank - 2] = comp_chunk_rows;
		      c_def.chunk_lengths[sds_info->rank - 1] = c_def.comp.chunk_lengths[sds_info->rank - 1] = sds_info->dim_size[sds_info->rank - 1];
		    }

		    c_def.comp.comp_type = COMP_CODE_DEFLATE;                   /* GZIP */
		    c_def.comp.cinfo.deflate.level = 6;                         /* compression level */

		    status = SDsetchunk(sds_info->sds_id, c_def, HDF_CHUNK | HDF_COMP);
		    if (status == -1)
		       fprintf(stderr, "Error2 in SDsetchunk in open_sds_compression(), maybe info. on rank or dim_size is missing. Or output HDF SDS exists!\n");
		  }
		  else /* for 1D compression */
		  {
		    c_def.chunk_lengths[0] = c_def.comp.chunk_lengths[0] = sds_info->dim_size[0];

		    c_def.comp.comp_type = COMP_CODE_DEFLATE;                   /* GZIP */
		    c_def.comp.cinfo.deflate.level = 6;                         /* compression level */

		    status = SDsetchunk(sds_info->sds_id, c_def, HDF_CHUNK | HDF_COMP);
		    if (status == -1)
		       fprintf(stderr, "Error2 in SDsetchunk in open_sds_compression(), maybe info. on rank or dim_size is missing. Or output HDF SDS exists!\n");
		    
		  }

                  len = (int)strlen(sds_info->name) + 1;
                  if (SDsetattr(sds_info->sds_id, "long_name", DFNT_CHAR8, len, 
                      (VOIDP)sds_info->name) == FAIL)
                  fprintf(stderr, "Could not write the attribute 'long_name' in %s\n", sds_info->name);
                }
            }
        }
    }
    
    if (sds_info->H5OUT)
    {
        H5Gclose(grp_id);
        if (h5_sds) free(h5_sds);
    }

    return status;
}


void close_hdf(sds_t *sds_info, int H5IO)
{
    if (H5IO)
    {
	if(strlen(sds_info->name) > 0 && sds_info->sds_id != -1)
	   if (H5Dclose(sds_info->sds_id) < 0)
		fprintf(stderr, "Cannot close the HDF5 SDS %s in %s in function close_hdf\n", sds_info->name, sds_info->file_name);
   
	if(sds_info->sd_id >= 0) 	 
	   if (H5Fclose(sds_info->sd_id) < 0)
              fprintf(stderr, "Cannot close the HDF5 %s in function close_hdf\n", sds_info->file_name);
    }
    else
    {
	if(strlen(sds_info->name) > 0 && sds_info->sds_id != -1)
	   if (SDendaccess(sds_info->sds_id) == FAIL)
		fprintf(stderr, "Cannot close the HDF4 SDS %s in %s in function close_hdf\n", sds_info->name, sds_info->file_name);
     
	if(sds_info->sd_id >= 0) 
	   if (SDend(sds_info->sd_id) == FAIL)
		fprintf(stderr, "Cannot close the HDF4 %s\n in function close_hdf", sds_info->file_name);
    }

  sds_info->sd_id = sds_info->sds_id = -1;
}

void close_sds(sds_t *sds_info, int H5IO)
{
    if (H5IO)
    {
	if(strlen(sds_info->name) > 0 && sds_info->sds_id != -1)
	    if (H5Dclose(sds_info->sds_id) < 0)
		fprintf(stderr, "Cannot close the HDF5 SDS %s in %s in function close_sds\n", sds_info->name, sds_info->file_name);
    }
    else
	if(strlen(sds_info->name) > 0 && sds_info->sds_id != -1)
	    if (SDendaccess(sds_info->sds_id) == FAIL)
		fprintf(stderr, "Cannot close the HDF4 SDS %s in %s in function close_sds\n", sds_info->name, sds_info->file_name);

    sds_info->sds_id = -1;
	
}

void close_sd(sds_t *sds_info, int H5IO)
{
    if (H5IO)
    {
	if(sds_info->sd_id > -1)
	    if (H5Fclose(sds_info->sd_id) < 0)
		fprintf(stderr, "Cannot close the HDF5 file %s in function close_sd\n", sds_info->file_name);
    }
    else
	if(sds_info->sd_id > -1)
	    if (SDend(sds_info->sd_id) == FAIL)
		fprintf(stderr, "Cannot close the HDF4 file %s in function close_sd\n", sds_info->file_name);

    sds_info->sd_id = -1;
}

int get_l2g_sds_names(char *fname, char **sds_names)
{
    int i;
    int len;
    int32 dt;
    int32 rank;
    int32 nsds;
    int32 nattr;
    int32 sd_id;
    int32 sds_id;
    int32 dim_size[3];
    int sds_cnt;
    char tmp[5];
    char name[100];
    /* HDF5 addition */
     H5SDS h5_sds;
    int HDF_ver = 0;
    int H5IN;

    /* MSZ HDF5 or HDF4 */
    HDF_ver = HDF_version_check(fname);

    if(HDF_ver == 4)
        H5IN = 0;
    else if(HDF_ver == 5)
        H5IN = 1;
    else
    {
        fprintf(stderr, "%s is neither HDF4 nor HDF5, exit...\n", fname);
        exit(0);
    }
     
    sds_cnt = 0;
    if (H5IN)
    {
        if ((sd_id = H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
        {
	    sd_id = -1;
            fprintf(stderr, "get_non1d_sds_number sds_info->sd_id = %d\n",sd_id);
            fprintf(stderr, "Cannot open the HDF file %s\n", fname);
            return sds_cnt;
        }
        h5_sds.srch_type = 4; /* for getting all sds names */
        h5_sds.nsds = 0;
        if ((h5_sds.h5sdsnames = (char **)Calloc2D(MAX_NUM_PARAM, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
        {
          fprintf(stderr, "Error allocating memory for in_sdsnames in alloc_memory_for_param()\n");
          return sds_cnt;
        }
        H5Ovisit(sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5O_iterate_t)H5dataset_search, (void *)&h5_sds);
        
        for (i=0; i<h5_sds.nsds; i++)
        {
            len = (int)strlen(h5_sds.h5sdsnames[i]);
            strmid(h5_sds.h5sdsnames[i], len-2, 2, tmp);
            if (strcmp(tmp, "_1") == 0)
            { 
              strcpy(sds_names[sds_cnt], h5_sds.h5sdsnames[i]);
              sds_names[sds_cnt][len-2] = '\0';
              sds_cnt++; 
            }
        }
        Free2D((void **)h5_sds.h5sdsnames);
        H5Fclose(sd_id);
    }
    else
    {
        if ((sd_id = SDstart(fname, DFACC_READ)) == FAIL)
          fprintf(stderr, "Cannot open the HDF file %s\n", fname);
        else
        {
          if (SDfileinfo(sd_id, &nsds, &nattr) == FAIL)
            fprintf(stderr, "Cannot read the information of file %s\n", fname);
          else for (i=0; i<nsds; i++)
          {
            if ((sds_id = SDselect(sd_id, i)) != FAIL)
            {
              if (SDgetinfo(sds_id, name, &rank, dim_size, &dt, &nattr) != FAIL)
              {
                len = (int)strlen(name);
            strmid(name, len-2, 2, tmp);
            if (strcmp(tmp, "_1") == 0)
                { 
              strcpy(sds_names[sds_cnt], name);
              sds_names[sds_cnt][len-2] = '\0';
              sds_cnt++; 
            }
              }
              SDendaccess(sds_id);
            }
          }
          SDend(sd_id);
        }
    }
  return sds_cnt;
}

int get_H5_full_sds_names(char *fname, char **sds_names)
{
  /* Developer: Maosheng Zhao, SSAI
   * date: 2/9/2018
   * Purpose: get all sds full names in an HDF5 with numerical H5 data type
   */

  hid_t h5_sd_id, h5_sds_id, h5dtype;
  H5T_class_t h5_type_class;
  herr_t  ret;
  hsize_t i;
  size_t len_strs;
  H5_OBJS_TYPES *h5_obj_types = NULL;
  void *meta_attr_buf = NULL;
  int sds_cnt = 0;

  if ((h5_sd_id = H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
  {
    fprintf(stderr, "Cannot open the HDF file %s\n", fname);
    return sds_cnt;
  }

  /* get objs_links */
  h5_obj_types = (H5_OBJS_TYPES*)calloc(1, sizeof(H5_OBJS_TYPES));

  /* The reason I don't use the existing
   * H5Ovisit(sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5O_iterate_t)H5dataset_search, (void *)&h5_sds);
   * is that call-back function H5dataset_search removed the groups from the name of DATASETs but
   * I need full name of DATASETs here. So I used my developed H5Fget_objs_links */ 
  ret = H5Fget_objs_links(h5_sd_id, h5_obj_types);

  /* check if there is meta_name */
  for(i = 0; i < h5_obj_types->num_objs_links; i++)
    if(h5_obj_types->obj_type[i] == H5O_TYPE_DATASET)
    {
      /* if the dataset type class is numerical */
      if((h5_sds_id = H5Dopen2(h5_sd_id, h5_obj_types->obj_name[i], H5P_DEFAULT)) < 0)
	fprintf(stderr, "cannot H5Dopen2 %s\n", h5_obj_types->obj_name[i]);
      else
      {
	h5dtype = H5Dget_type(h5_sds_id);
	if(H5Tget_class(h5dtype) == H5T_INTEGER || H5Tget_class(h5dtype) == H5T_FLOAT)
	{
	   strcpy(sds_names[sds_cnt], h5_obj_types->obj_name[i]);
	   sds_cnt++;
	}
	H5Tclose(h5dtype);
	H5Dclose(h5_sds_id);
      }
    }

  for(i = 0; i < h5_obj_types->num_objs_links; i++)
    if(h5_obj_types->obj_name[i]) free(h5_obj_types->obj_name[i]);

  if(h5_obj_types->obj_name) free(h5_obj_types->obj_name);
  if(h5_obj_types->obj_type) free(h5_obj_types->obj_type);
  if(h5_obj_types)free(h5_obj_types);

  H5Fclose(h5_sd_id);

  return sds_cnt;
}

int get_sds_names(char *fname, char **sds_names)
{
    int i;
    int32 dt;
    int32 rank;
    int32 nsds;
    int32 nattr;
    int32 sd_id;
    int32 sds_id;
    int32 dim_size[3];
    int sds_cnt;
    char name[80];
    /* HDF5 addition */
    H5SDS h5_sds;
    int HDF_ver = 0;
    int H5IN;

    /* MSZ HDF5 or HDF4 */
    HDF_ver = HDF_version_check(fname);

    if(HDF_ver == 4)
        H5IN = 0;
    else if(HDF_ver == 5)
        H5IN = 1;
    else
    {
        fprintf(stderr, "%s is neither HDF4 nor HDF5, exit...\n", fname);
        exit(0);
    }
    
    sds_cnt = 0;
    if (H5IN)
    {
        if ((sd_id = (int32)H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
        {
	    sd_id = -1;
            fprintf(stderr, "get_non1d_sds_number sds_info->sd_id = %d\n",sd_id);
            fprintf(stderr, "Cannot open the HDF file %s\n", fname);
            return sds_cnt;
        }
        h5_sds.srch_type = 4; /* for getting all sds names */
        h5_sds.nsds = 0;
        if ((h5_sds.h5sdsnames = (char **)Calloc2D(MAX_NUM_PARAM, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
        {
          fprintf(stderr, "Error allocating memory for in_sdsnames in alloc_memory_for_param()\n");
          return sds_cnt;
        }
       
        /* MSZ check if the root path are added to each objects */ 
        H5Ovisit(sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5O_iterate_t)H5dataset_search, (void *)&h5_sds);

	/* how many Metadata */
        sds_cnt = h5_sds.nsds;
        for (i=0; i<sds_cnt; i++)
        {
	  /* MSZ 1/24/2018, split the h5sdsname into absolute path name and short name */
	  char *abs_path = NULL, *short_name = NULL;
	  HDF5_split_abs_name_into_path_name(h5_sds.h5sdsnames[i], &abs_path, &short_name);

	  /* we just use the short name */
	  strcpy(sds_names[i], short_name);
	  if(abs_path) free(abs_path);
	  if(short_name) free(short_name);
        }
        Free2D((void **)h5_sds.h5sdsnames);
        H5Fclose(sd_id);
    }
    else
    {
        if ((sd_id = SDstart(fname, DFACC_READ)) == FAIL)
        {
          fprintf(stderr, "Cannot open the HDF file %s\n", fname);
          return sds_cnt;
        }
        if (SDfileinfo(sd_id, &nsds, &nattr) == FAIL)
          fprintf(stderr, "Cannot read information for HDF file %s\n", fname);
        else for (i=0; i<nsds; i++)
        {
          if ((sds_id = SDselect(sd_id, i)) != FAIL)
          {
            if (SDgetinfo(sds_id, name, &rank, dim_size, &dt, &nattr) != FAIL)
              strcpy(sds_names[sds_cnt++], name);
            SDendaccess(sds_id);
          }
        }
        SDend(sd_id);
    }
  return sds_cnt;
}

int get_sds_names_exclusions(char *fname, char **sds_names, char **sds_names_exclusion, int num_names_exclusion)
{
    /* Maosheng Zhao, 4/5/2018. 
     * get sds short name list but exclude the number of num_names_exclusio
     * names from sds_names_exclusion. For example, for HDF5, we don't want
     * StructMetadata.0 to be listed as a dataset in some cases */
    int i;
    int32 dt;
    int32 rank;
    int32 nsds;
    int32 nattr;
    int32 sd_id;
    int32 sds_id;
    int32 dim_size[3];
    int sds_cnt;
    char name[80];
    /* HDF5 addition */
    H5SDS h5_sds;
    int HDF_ver = 0;
    int H5IN;
    int j, k, yes, num_exclusions = 0;

    /* MSZ HDF5 or HDF4 */
    HDF_ver = HDF_version_check(fname);

    if(HDF_ver == 4)
        H5IN = 0;
    else if(HDF_ver == 5)
        H5IN = 1;
    else
    {
        fprintf(stderr, "%s is neither HDF4 nor HDF5, exit...\n", fname);
        exit(0);
    }
    
    sds_cnt = 0;
    if (H5IN)
    {
        if ((sd_id = (int32)H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
        {
	    sd_id = -1;
            fprintf(stderr, "get_non1d_sds_number sds_info->sd_id = %d\n",sd_id);
            fprintf(stderr, "Cannot open the HDF file %s\n", fname);
            return sds_cnt;
        }
        h5_sds.srch_type = 4; /* for getting all sds names */
        h5_sds.nsds = 0;
        if ((h5_sds.h5sdsnames = (char **)Calloc2D(MAX_NUM_PARAM, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
        {
          fprintf(stderr, "Error allocating memory for in_sdsnames in alloc_memory_for_param()\n");
          return sds_cnt;
        }
       
        /* MSZ check if the root path are added to each objects */ 
        H5Ovisit(sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5O_iterate_t)H5dataset_search, (void *)&h5_sds);

	/* how many Metadata */
        sds_cnt = h5_sds.nsds;
	k = 0;
        for (i=0; i<sds_cnt; i++)
        {
	  /* MSZ 1/24/2018, split the h5sdsname into absolute path name and short name */
	  char *abs_path = NULL, *short_name = NULL;
	  HDF5_split_abs_name_into_path_name(h5_sds.h5sdsnames[i], &abs_path, &short_name);

	  /* we just use the short name */
	  yes = 0;
	  if(num_names_exclusion > 0)
	    for(j=0; j<num_names_exclusion; j++)
	      if(strcmp(short_name, sds_names_exclusion[j]) == 0)
	      {
		yes = 1;
		break;
	      }

	  if(yes)
	    num_exclusions++;
	  else
	    strcpy(sds_names[k++], short_name);

	  if(abs_path) free(abs_path);
	  if(short_name) free(short_name);
        }
        Free2D((void **)h5_sds.h5sdsnames);
        H5Fclose(sd_id);

	/* adjust sds_cnt */
	sds_cnt -= num_exclusions;
    }
    else
    {
        if ((sd_id = SDstart(fname, DFACC_READ)) == FAIL)
        {
          fprintf(stderr, "Cannot open the HDF file %s\n", fname);
          return sds_cnt;
        }
        if (SDfileinfo(sd_id, &nsds, &nattr) == FAIL)
          fprintf(stderr, "Cannot read information for HDF file %s\n", fname);
        else for (i=0; i<nsds; i++)
        {
          if ((sds_id = SDselect(sd_id, i)) != FAIL)
          {
            if (SDgetinfo(sds_id, name, &rank, dim_size, &dt, &nattr) != FAIL)
	    {
		yes = 0;
		if(num_names_exclusion > 0)
		  for(j=0; j<num_names_exclusion; j++)
		    if(strcmp(name, sds_names_exclusion[j]) == 0)
		    {
		      yes = 1;
		      break;
		    }

		if(!yes)
                  strcpy(sds_names[sds_cnt++], name);
	    }
            SDendaccess(sds_id);
          }
        }
        SDend(sd_id);
    }
  return sds_cnt;
}


int get_nsds(char *fname)
{
    /* MSZ 2/7/2018. developed by modifying get_sds_names() */
    int i;
    int32 dt;
    int32 rank;
    int32 nsds;
    int32 nattr;
    int32 sd_id;
    int32 sds_id;
    int32 dim_size[3];
    char name[80];
    int NSDS;
    /* HDF5 addition */
    H5SDS h5_sds;
    int HDF_ver = 0;
    int H5IN;

    /* MSZ HDF5 or HDF4 */
    HDF_ver = HDF_version_check(fname);

    if(HDF_ver == 4)
        H5IN = 0;
    else if(HDF_ver == 5)
        H5IN = 1;
    else
    {
        fprintf(stderr, "%s is neither HDF4 nor HDF5, exit...\n", fname);
        exit(0);
    }
    
    NSDS = 0;
    if (H5IN)
    {
        if ((sd_id = (int32)H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
        {
	    sd_id = -1;
            fprintf(stderr, "get_non1d_sds_number sds_info->sd_id = %d\n",sd_id);
            fprintf(stderr, "Cannot open the HDF file %s\n", fname);
            return 0;
        }
        h5_sds.srch_type = 4; /* for getting all sds names including H5T_STRING */
        h5_sds.nsds = 0;
        if ((h5_sds.h5sdsnames = (char **)Calloc2D(MAX_NUM_PARAM, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
        {
          fprintf(stderr, "Error allocating memory for in_sdsnames in alloc_memory_for_param()\n");
          return 0;
        }
       
        H5Ovisit(sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5O_iterate_t)H5dataset_search, (void *)&h5_sds);

	if(h5_sds.h5sdsnames) Free2D((void **)h5_sds.h5sdsnames);

        H5Fclose(sd_id);

	NSDS = h5_sds.nsds;

	return NSDS;
    }
    else
    {
        if ((sd_id = SDstart(fname, DFACC_READ)) == FAIL)
        {
          fprintf(stderr, "Cannot open the HDF file %s\n", fname);
          return 0;
        }
        if (SDfileinfo(sd_id, &nsds, &nattr) == FAIL)
	{
          fprintf(stderr, "Cannot read information for HDF file %s\n", fname);
	  return 0;
	}
	else
	{
	  SDend(sd_id);
	  NSDS = nsds;

	  return NSDS;
	}
    }
}

int sds_name_existence(char *fname, char *sds_name)
{
   /* Maosheng Zhao 2/15/2018, For read_l2g and read_l2g_lite */
   /* existence: return 1; otherwise return 0 */
   int nsds = 0;
   int existence = 0;
   int i;
   char **sds_names = NULL;

   nsds = get_nsds(fname);

   if(nsds)
   {
	if ((sds_names = (char **)Calloc2D(nsds, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
	{
	   fprintf(stderr, "Cannot allocate memory for sds_names in sds_name_existence(), exit...\n");
	   exit(0);
	}
	nsds = get_sds_names(fname, sds_names);
	if(nsds)
	   for(i = 0; i < nsds; i++)
	      if(strcmp(sds_names[i], sds_name) == 0)
	      {
		existence = 1;
		break;
	      }

	if(sds_names) Free2D((void**)sds_names);
    }

   return existence;
}
int sds_name_existence_soft(char *fname, char *sds_name)
{
   /* Maosheng Zhao 2/15/2018, For read_l2g and read_l2g_lite */
   /* existence: return 1; otherwise return 0 */
   int nsds = 0;
   int existence = 0;
   int i;
   char **sds_names = NULL;

   nsds = get_nsds(fname);

   if(nsds)
   {
	if ((sds_names = (char **)Calloc2D(nsds, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
	{
	   fprintf(stderr, "Cannot allocate memory for sds_names in sds_name_existence(), exit...\n");
	   exit(0);
	}
	nsds = get_sds_names(fname, sds_names);
	if(nsds)
	   for(i = 0; i < nsds; i++)
	      if(strcasestr(sds_names[i], sds_name) != '\0')
	      {
		existence = 1;
		break;
	      }

	if(sds_names) Free2D((void**)sds_names);
    }

   return existence;
}

int get_sd_attr_names(const char *HDF_fname, char **attr_names)
{
   /* Maosheng Zhao, SSAI, 2/15/2018 */
   /* get the file attr names and return count of file attrs */
   /* Calloc2D attr_names outside first before calling the function */
   /* if failed, return -1 */
   int HDF_ver = 0;
   int attr_cnt;
   char attr_name[255] = {'\0'};

   HDF_ver = HDF_version_check(HDF_fname);
   if( !(HDF_ver == 4 || HDF_ver == 5) )
   {
	fprintf(stderr, "%s is neither HDF4 nor HDF5, exit...\n", HDF_fname);
	exit(0);
   }

   if(HDF_ver == 5)
   {
	/*###############################################*/
	hid_t   h5_file_id, group_id;
	hid_t   attr;
	hid_t   atype;    /* Attribute type */
	H5T_class_t  type_class;
	herr_t  ret;                /* Return value */
	H5O_info_t oinfo;           /* Object info */
	unsigned i;                 /* Counters */
	ssize_t attr_name_len;

	/* Open the file and group */
	if((h5_file_id = H5Fopen(HDF_fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
	{
	  attr_cnt = -1;
	  fprintf(stderr, "Cannot open the HDF5 file %s\n", HDF_fname);
	}

	if((group_id = H5Gopen2(h5_file_id, "/", H5P_DEFAULT)) < 0)
	{
	  attr_cnt = -1;
	  fprintf(stderr, "Cannot H5Gopen2 for the root group for file %s\n", HDF_fname);
	}

	/* get attribute by idx for the current obj */
	ret = H5Oget_info(group_id, &oinfo);

	if(ret < 0) attr_cnt = -1;
	else attr_cnt = (int)oinfo.num_attrs;

	if(ret >= 0 && (unsigned)oinfo.num_attrs > 0)
	for(i = 0; i < (unsigned)oinfo.num_attrs; i++)
	{
	  /* MSZ 2/21/2018  Note that sometimes H5Aget_name_by_idx() fails to work to get the attribute names */
	  /*attr_name_len = H5Aget_name_by_idx(group_id, ".", H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)i, attr_name, 0, H5P_DEFAULT);*/
          attr = H5Aopen_by_idx(group_id, ".", H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)i, H5P_DEFAULT, H5P_DEFAULT);
          attr_name_len = H5Aget_name(attr, 0, attr_name);

	  if(attr_name_len > 0) strcpy(attr_names[i], attr_name);

	  ret = H5Aclose(attr);
	  #if 0
	  attr = H5Aopen_by_idx(group_id, '.', H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)i, H5P_DEFAULT, H5P_DEFAULT);
	  atype = H5Aget_type(attr);
	  type_class = H5Tget_class(atype);
	  if (type_class == H5T_STRING) {
	  /* do something for attribute */
	  }
	  ret = H5Aclose(attr);
	  ret = H5Tclose(atype);
	  #endif
	}

	/* Close the group_id and h5_file_id */
	H5Gclose(group_id);
	H5Fclose(h5_file_id);
	/*###############################################*/
   }
   else if(HDF_ver == 4)
   {
	/*###############################################*/
	int32 sd_id;
	intn status;
	int32 n_datasets,n_file_attrs;
	int32 data_type, attr_index, n_values;

	sd_id = SDstart(HDF_fname, DFACC_READ);
	if(sd_id == -1)
	{
	    attr_cnt = -1;
	    fprintf(stderr, "Cannot open the HDF4 file %s\n", HDF_fname);
	}

	/* if it is HDF-EOS file, we first get the GridName from the StructMetadata.0 of file sttribute */
	status = SDfileinfo(sd_id,&n_datasets,&n_file_attrs);
	if(status == 0)
	{
	  attr_cnt = n_file_attrs;
	  if(attr_cnt > 0)
	    for(attr_index = 0; attr_index < n_file_attrs; attr_index++)
	    {
	      status = SDattrinfo (sd_id, attr_index, attr_name, &data_type, &n_values);
	      strcpy(attr_names[attr_index], attr_name);
	    }
	}
	else attr_cnt = -1;

	status = SDend(sd_id);
	/*###############################################*/
   }

   return attr_cnt;
}
int get_sd_attr_num(const char *HDF_fname)
{
   /* Maosheng Zhao, SSAI, 2/15/2018 */
   /* get the number of file attr names */
   /* if failed, return -1 */
   int HDF_ver = 0;
   int attr_cnt;
   char attr_name[255] = {'\0'};

   HDF_ver = HDF_version_check(HDF_fname);
   if( !(HDF_ver == 4 || HDF_ver == 5) )
   {
	fprintf(stderr, "%s is neither HDF4 nor HDF5, exit...\n", HDF_fname);
	exit(0);
   }

   if(HDF_ver == 5)
   {
	/*###############################################*/
	hid_t   h5_file_id, group_id;
	hid_t   attr;
	hid_t   atype;    /* Attribute type */
	H5T_class_t  type_class;
	herr_t  ret;                /* Return value */
	H5O_info_t oinfo;           /* Object info */
	unsigned i;                 /* Counters */
	ssize_t attr_name_len;

	/* Open the file and group */
	if((h5_file_id = H5Fopen(HDF_fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
	{
	  attr_cnt = -1;
	  fprintf(stderr, "Cannot open the HDF5 file %s\n", HDF_fname);
	}

	if((group_id = H5Gopen2(h5_file_id, "/", H5P_DEFAULT)) < 0)
	{
	  attr_cnt = -1;
	  fprintf(stderr, "Cannot H5Gopen2 for the root group for file %s\n", HDF_fname);
	}

	/* get attribute by idx for the current obj */
	ret = H5Oget_info(group_id, &oinfo);

	if(ret < 0) attr_cnt = -1;
	else attr_cnt = (int)oinfo.num_attrs;

	if(ret >= 0 && (unsigned)oinfo.num_attrs > 0)
	for(i = 0; i < (unsigned)oinfo.num_attrs; i++)
	{
	  /* MSZ 2/21/2018  Note that sometimes H5Aget_name_by_idx() fails to work to get the attribute names */
	  /*attr_name_len = H5Aget_name_by_idx(group_id, ".", H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)i, attr_name, 0, H5P_DEFAULT);*/
          attr = H5Aopen_by_idx(group_id, ".", H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)i, H5P_DEFAULT, H5P_DEFAULT);
          attr_name_len = H5Aget_name(attr, 0, attr_name);

	  ret = H5Aclose(attr);
	  #if 0
	  attr = H5Aopen_by_idx(group_id, '.', H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)i, H5P_DEFAULT, H5P_DEFAULT);
	  atype = H5Aget_type(attr);
	  type_class = H5Tget_class(atype);
	  if (type_class == H5T_STRING) {
	  /* do something for attribute */
	  }
	  ret = H5Aclose(attr);
	  ret = H5Tclose(atype);
	  #endif
	}

	/* Close the group_id and h5_file_id */
	H5Gclose(group_id);
	H5Fclose(h5_file_id);
	/*###############################################*/
   }
   else if(HDF_ver == 4)
   {
	/*###############################################*/
	int32 sd_id;
	intn status;
	int32 n_datasets,n_file_attrs;
	int32 data_type, attr_index, n_values;

	sd_id = SDstart(HDF_fname, DFACC_READ);
	if(sd_id == -1)
	{
	    attr_cnt = -1;
	    fprintf(stderr, "Cannot open the HDF4 file %s\n", HDF_fname);
	}

	/* if it is HDF-EOS file, we first get the GridName from the StructMetadata.0 of file sttribute */
	status = SDfileinfo(sd_id,&n_datasets,&n_file_attrs);
	if(status == 0)
	{
	  attr_cnt = n_file_attrs;
	  if(attr_cnt > 0)
	    for(attr_index = 0; attr_index < n_file_attrs; attr_index++)
	    {
	      status = SDattrinfo (sd_id, attr_index, attr_name, &data_type, &n_values);
	    }
	}
	else attr_cnt = -1;

	status = SDend(sd_id);
	/*###############################################*/
   }

   return attr_cnt;
}

int file_attr_name_existence(char *fname, char *attr_name)
{
   /* Maosheng Zhao 3/3/2018, For read_l2g and read_l2g_lite */
   /* works for both HDF4/HDF5 */
   /* existence: return 1; otherwise return 0 */
   int nattr = 0;
   int existence = 0;
   int i;
   char **file_attr_names = NULL;

   nattr = get_sd_attr_num((const char*)fname);

   if(nattr)
   {
     if ((file_attr_names = (char **)Calloc2D(nattr, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
     {
	fprintf(stderr, "Cannot allocate memory for file_attr_names in read_l2g_lite(main)\n");
	exit(0);
     }
     else
     {
	nattr = get_sd_attr_names(fname, file_attr_names);
	if(nattr)
	   for(i = 0; i < nattr; i++)
	      if(strcasestr(file_attr_names[i], attr_name) !='\0')
	      {
		existence = 1;
		break;
	      }

	if(file_attr_names) Free2D((void**)file_attr_names);
     }
   }
   return existence;
}

void *get_sd_attr(sds_t *sds_info, char *attr_name, int32 *attr_type, int32 *attr_cnt)
{
    /* MSZ 2/7/2018 modified from void *get_sds_attr() but this function get file attr */
    int32 attr_size;
    int32 attr_index;
    void *attr_buf = NULL;
    char name[80];
  
    hid_t attrh5_type, attr_space;
    hid_t attr_id, memtype;
    H5T_class_t type_class;
    size_t size;

    if (sds_info->H5IN)
    {
      if( (attr_id = H5Aopen(sds_info->sd_id,attr_name,H5P_DEFAULT)) < 0 )
      {
	/* MSZ 8/21/17. for no "offset" case, and its value is 0. */
	if(strcasestr(attr_name, "offset"))
	{
	  *attr_type = DFNT_FLOAT64;
	  *attr_cnt = 1;
	  attr_size = DFKNTsize(*attr_type);
	  if ((attr_buf = (void *)calloc(*attr_cnt, attr_size)) == NULL)
	    fprintf(stderr, "Cannot allocate memory for attr_buf in get_sd_attr\n");
	  else
	  {
	    double offset = 0.;
	    *(float64*)attr_buf = offset;
	  }
	}
	else
          /*fprintf(stderr, "Could not find the attribute %s for HDF file\n", attr_name)*/;
      }
      else
      {
        /* Get attribute datatype, dataspace, rank, and dimensions.  */
        attr_space = H5Aget_space(attr_id);

	attrh5_type= (int32)H5Aget_type(attr_id);
	*attr_type = H5toH4type(attrh5_type);

	type_class = H5Tget_class(attrh5_type);

	/* MSZ 8/15/17, largely modified the below code */
	size       = H5Tget_size(attrh5_type);

	if(type_class == H5T_STRING)
	  size++; /* Make room for null terminator */

	/*
	if(type_class != -1)
	    fprintf(stderr, "ATTR_NAME=%s type_class=%d H5T_INTEGER=%d H5T_FLOAT=%d size = %d\n", attr_name, type_class, H5T_INTEGER, H5T_FLOAT, size);
	*/

	if(type_class == H5T_INTEGER || type_class == H5T_FLOAT)  /* numerical types */
	{
	    /* Get attribute datatype, dataspace, rank, and dimensions.  */
	    *attr_cnt = H5Sget_simple_extent_npoints(attr_space);

	    if ((attr_buf = (void *)calloc(*attr_cnt, size)) == NULL)
		fprintf(stderr, "Cannot allocate memory for attr_buf in get_sd_attr\n");
	    else
	        if (H5Aread(attr_id, attrh5_type, attr_buf) < 0)
		    fprintf(stderr, "Cannot read sd_attr %s in get_sd_attr\n", attr_name);
		
	}
	else if(type_class == H5T_STRING)        /* string */
	{
	    *attr_cnt = size;

	    if ((attr_buf = (void *)calloc(*attr_cnt, sizeof(char))) == NULL)
		fprintf(stderr, "Cannot allocate memory for attr_buf in get_sd_attr\n");
	    else
		if (H5Aread(attr_id, attrh5_type, attr_buf) < 0)
		    fprintf(stderr, "CANnot read sd_attr %s in get_sd_attr\n", attr_name);
	}
	else  /* we just interested in scale_factor, offset fill value etc not others */
	{
	   *attr_cnt = 0;
	   *attr_type = -1;
	   attr_buf = NULL;
	}

	H5Tclose(attrh5_type);
        H5Sclose(attr_space);
        H5Aclose(attr_id);
      } /* else for if( (attr_id = H5Aopen(sd_info->sd_id,attr_name,H5P_DEFAULT)) < 0 ) */
    }
    else
    {
        if ((attr_index = SDfindattr(sds_info->sd_id, attr_name)) == FAIL)
          /*fprintf(stderr, "Could not find the attribute %s for HDF file\n", attr_name)*/;
        else
        {
          if (SDattrinfo(sds_info->sd_id, attr_index, name, attr_type, attr_cnt) == FAIL)
            fprintf(stderr, "Cannot read the attribute information for %s\n", attr_name);
          else
          {
            attr_size = DFKNTsize(*attr_type);
            if ((attr_buf = (void *)calloc(*attr_cnt, attr_size)) == NULL)
              fprintf(stderr, "Cannot allocate memory for attr_buf in get_sd_attr\n");
            else
            {
		int status;
		status = SDreadattr(sds_info->sd_id, attr_index, attr_buf);
		if(status == -1)
		{
		    fprintf(stderr, "Cannot read sds_attr %s in get_sd_attr\n", attr_name);
		    if(attr_buf) free(attr_buf);
		    attr_buf = NULL;
		}
            }
          }
        }
    }

    return attr_buf;
}

void *find_get_sd_attr(sds_t *sds_info, char *attr_name, int32 *attr_type, int32 *attr_cnt)
{
    /* MSZ 2/7/2018 modified from void *get_sd_attr() but this function first find the
     * exactly attrname using strcasestr() if the input is not the exact attr_name */
    int32 attr_size;
    int32 attr_index;
    void *attr_buf = NULL;
    char name[80] = {'\0'};
    char real_attr_name[80] = {'\0'};
    char temp_name[255] = {'\0'};

    herr_t  ret;                /* Return value */
    H5O_info_t oinfo;           /* Object info */
  
    hid_t attrh5_type, attr_space;
    hid_t group_id, attr_id, memtype;
    H5T_class_t type_class;
    size_t size;
    unsigned h5_idx;
    ssize_t name_len, i;
    int find_it = 0;

    if (sds_info->H5IN)
    {
      if((group_id = H5Gopen2((hid_t)sds_info->sd_id, "/", H5P_DEFAULT)) < 0)
          fprintf(stderr, "Cannot H5Gopen2 for the root group\n");
      else
      {
        /* find the real_attr_name */
        ret = H5Oget_info(group_id, &oinfo);
        if(ret >= 0 && oinfo.num_attrs > 0)
          for(h5_idx = 0; h5_idx < (unsigned)oinfo.num_attrs; h5_idx++)
	  {
	    /* H5Aget_name_by_idx() dones't work here for VNP09GA.A2017240.h09v05.001.2017241195209.h5 with read_l2g_lite.c !!!*/
	    /*name_len = H5Aget_name_by_idx((hid_t)sds_info->sd_id, ".", H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)h5_idx, real_attr_name, 0, H5P_DEFAULT);*/
	    attr_id = H5Aopen_by_idx(group_id, ".", H5_INDEX_CRT_ORDER, H5_ITER_INC, (hsize_t)h5_idx, H5P_DEFAULT, H5P_DEFAULT);
	    name_len = H5Aget_name(attr_id, 0, real_attr_name);

	    if(attr_id < 0 || name_len < 0) real_attr_name[0] = '\0';
	    else
	      if(strcasestr(real_attr_name, attr_name) != '\0')
	      {
		find_it = 1;
	        break;
	      }

	    ret = H5Aclose(attr_id);
	  }

	#if 1
        if(find_it == 0 || (attr_id = H5Aopen(group_id, real_attr_name,H5P_DEFAULT)) < 0 )
        {
          /*fprintf(stderr, "Could not find the attribute %s for HDF file\n", real_attr_name)*/;
	  attr_buf = NULL;
        }
        else
        {
          /* Get attribute datatype, dataspace, rank, and dimensions.  */
          attr_space = H5Aget_space(attr_id);

	  attrh5_type= (int32)H5Aget_type(attr_id);
	  *attr_type = H5toH4type(attrh5_type);

	  type_class = H5Tget_class(attrh5_type);

	  /* MSZ 8/15/17, largely modified the below code */
	  size       = H5Tget_size(attrh5_type);

	  if(type_class == H5T_STRING)
	    size++; /* Make room for null terminator */

	  /*
	  if(type_class != -1)
	    fprintf(stderr, "ATTR_NAME=%s type_class=%d H5T_INTEGER=%d H5T_FLOAT=%d size = %d\n", attr_name, type_class, H5T_INTEGER, H5T_FLOAT, size);
	  */

	  if(type_class == H5T_INTEGER || type_class == H5T_FLOAT)  /* numerical types */
	  {
	    /* Get attribute datatype, dataspace, rank, and dimensions.  */
	    *attr_cnt = H5Sget_simple_extent_npoints(attr_space);

	    if ((attr_buf = (void *)calloc(*attr_cnt, size)) == NULL)
		fprintf(stderr, "Cannot allocate memory for attr_buf in get_sd_attr\n");
	    else
	        if (H5Aread(attr_id, attrh5_type, attr_buf) < 0)
		    fprintf(stderr, "Cannot read sd_attr %s in get_sd_attr\n", attr_name);
		
	  }
	  else if(type_class == H5T_STRING)        /* string */
	  {
	    *attr_cnt = size;

	    if ((attr_buf = (void *)calloc(*attr_cnt, sizeof(char))) == NULL)
		fprintf(stderr, "Cannot allocate memory for attr_buf in get_sd_attr\n");
	    else
		if (H5Aread(attr_id, attrh5_type, attr_buf) < 0)
		    fprintf(stderr, "CANnot read sd_attr %s in get_sd_attr\n", attr_name);
	  }
	  else  /* we just interested in scale_factor, offset fill value etc not others */
	  {
	    *attr_cnt = 0;
	    *attr_type = -1;
	    attr_buf = NULL;
	  }

	  H5Tclose(attrh5_type);
          H5Sclose(attr_space);
          H5Aclose(attr_id);
        } /* else for if( (attr_id = H5Aopen(sd_info->sd_id,attr_name,H5P_DEFAULT)) < 0 ) */
	#endif
	H5Gclose(group_id);
      } /* else */
    }
    else
    {
	int32 nsds, nattr;

	/* find the real_attr_name */
	SDfileinfo(sds_info->sd_id, &nsds, &nattr);

	for (attr_index=0; attr_index<nattr; attr_index++)
	{
	  if (SDattrinfo(sds_info->sd_id, attr_index, real_attr_name, attr_type, attr_cnt) == FAIL)
	  {
	    fprintf(stdout, "Cannot read information for attribute %d\n", attr_index);
	    real_attr_name[0] = '\0';
	  }
	  else
	    if(strcasestr(real_attr_name, attr_name) != '\0')
	    {
		find_it = 1;
		break;
	    }
	}

        if (find_it == 0 || (attr_index = SDfindattr(sds_info->sd_id, real_attr_name)) == FAIL)
	{
          /*fprintf(stderr, "Could not find the attribute %s for HDF file\n", real_attr_name)*/;
	  attr_buf = NULL;
	}
        else
        {
          if (SDattrinfo(sds_info->sd_id, attr_index, name, attr_type, attr_cnt) == FAIL)
            fprintf(stderr, "Cannot read the attribute information for %s\n", real_attr_name);
          else
          {
            attr_size = DFKNTsize(*attr_type);
            if ((attr_buf = (void *)calloc(*attr_cnt, attr_size)) == NULL)
              fprintf(stderr, "Cannot allocate memory for attr_buf in get_sd_attr\n");
            else
            {
		int status;
		status = SDreadattr(sds_info->sd_id, attr_index, attr_buf);
		if(status == -1)
		{
		    fprintf(stderr, "Cannot read sds_attr %s in get_sd_attr\n", real_attr_name);
		    if(attr_buf) free(attr_buf);
		    attr_buf = NULL;
		}
            }
          }
        }
    }

    return attr_buf;
}

void find_get_sd_attr_to_str(char *fname, char *attr_name, char *meta_val)
/* Maosheng Zhao, SSAI, 3/3/2018. the function works for both HDF4/HDF5 */
/*void *find_get_sd_attr_to_str(sds_t *sds_info, char *attr_name, int32 *attr_type, int32 *attr_cnt)*/
/* Note that meta_val must be calloc() or predefined string outside this function */
{
   sds_t in_sds_info;
   int32 attr_type, attr_cnt;
   void *attr_buf = NULL;
   int i;
   char tmpData[500];

   in_sds_info.sd_id = in_sds_info.sds_id = -1;
   if(open_sd(fname, &in_sds_info, 'R') == FAIL)
     fprintf(stderr, "Cannot open output HDF file %s\n", fname);
   else
   {
      if( !(attr_buf = find_get_sd_attr(&in_sds_info, attr_name, &attr_type, &attr_cnt)) )
      {
	fprintf(stderr, "Cannot find the file attribute with key words of %s\n", attr_name);
	meta_val[0] = '\0';
      }
      else
      {
	if(attr_cnt == 0)
	  meta_val[0] = '\0';
	else
	{
	  if(attr_type == DFNT_CHAR8 || attr_type == DFNT_UCHAR8)
	    sprintf(meta_val, "%s", attr_buf);
	  else  /* numerical types */
	  {
	    if(attr_cnt <= 0)
	      meta_val[0] = '\0';
	    else if(attr_cnt == 1)
	    {
		/* I don't use switch here in case that attr_type is obtained during the run time not compile time */
	      if(attr_type == DFNT_INT8)
		sprintf(meta_val, "%d", ((int8 *)attr_buf)[0]);
	      else if(attr_type == DFNT_UINT8)
		sprintf(meta_val, "%d", ((uint8 *)attr_buf)[0]);
	      else if(attr_type == DFNT_INT16)
		sprintf(meta_val, "%d", ((int16 *)attr_buf)[0]);
	      else if(attr_type == DFNT_UINT16)
		sprintf(meta_val, "%d", ((uint16 *)attr_buf)[0]);
	      else if(attr_type == DFNT_INT32)
		sprintf(meta_val, "%d", ((int32 *)attr_buf)[0]);
	      else if(attr_type == DFNT_UINT32)
		sprintf(meta_val, "%u", ((uint32 *)attr_buf)[0]);
	      else if(attr_type == DFNT_FLOAT32)
		sprintf(meta_val, "%f", ((float32 *)attr_buf)[0]);
	      else if(attr_type == DFNT_FLOAT64)
		sprintf(meta_val, "%lf", ((float64 *)attr_buf)[0]);
	      else
	      {
		fprintf(stderr, "Don't know the data type %d, exit...\n", attr_type);
		exit(0);
	      }
	    }
	    else  /* attr_cnt > 1 */
	    {
	      /* I don't use switch here in case that attr_type is obtained during the run time not compile time */
	      if(attr_type == DFNT_INT8)
	      {
		sprintf(meta_val, "%d,", ((int8 *)attr_buf)[0]);
		for (i=1; i<attr_cnt-1; i++)
		{
		  sprintf(tmpData, "%d,", ((int8 *)attr_buf)[i]);
		  strcat(meta_val, tmpData);
		}
		sprintf(tmpData, "%d", ((int8 *)attr_buf)[attr_cnt-1]);
		strcat(meta_val, tmpData);
	      } /* attr_type == DFNT_INT8 */
	      else if(attr_type == DFNT_UINT8)
	      {
		sprintf(meta_val, "%d,", ((uint8 *)attr_buf)[0]);
		for (i=1; i<attr_cnt-1; i++)
		{
		  sprintf(tmpData, "%d,", ((uint8 *)attr_buf)[i]);
		  strcat(meta_val, tmpData);
		}
		sprintf(tmpData, "%d", ((uint8 *)attr_buf)[attr_cnt-1]);
		strcat(meta_val, tmpData);
	      } /* attr_type == DFNT_UINT8 */
	      else if(attr_type == DFNT_INT16)
	      {
		sprintf(meta_val, "%d,", ((int16 *)attr_buf)[0]);
		for (i=1; i<attr_cnt-1; i++)
		{
		  sprintf(tmpData, "%d,", ((int16 *)attr_buf)[i]);
		  strcat(meta_val, tmpData);
		}
		sprintf(tmpData, "%d", ((int16 *)attr_buf)[attr_cnt-1]);
		strcat(meta_val, tmpData);
	      } /* attr_type == DFNT_INT16 */
	      else if(attr_type == DFNT_UINT16)
	      {
		sprintf(meta_val, "%d,", ((uint16 *)attr_buf)[0]);
		for (i=1; i<attr_cnt-1; i++)
		{
		  sprintf(tmpData, "%d,", ((uint16 *)attr_buf)[i]);
		  strcat(meta_val, tmpData);
		}
		sprintf(tmpData, "%d", ((uint16 *)attr_buf)[attr_cnt-1]);
		strcat(meta_val, tmpData);
	      } /* attr_type == DFNT_UINT16 */
	      else if(attr_type == DFNT_INT32)
	      {
		sprintf(meta_val, "%d,", ((int32 *)attr_buf)[0]);
		for (i=1; i<attr_cnt-1; i++)
		{
		  sprintf(tmpData, "%d,", ((int32 *)attr_buf)[i]);
		  strcat(meta_val, tmpData);
		}
		sprintf(tmpData, "%d", ((int32 *)attr_buf)[attr_cnt-1]);
		strcat(meta_val, tmpData);
	      } /* attr_type == DFNT_INT32 */
	      else if(attr_type == DFNT_UINT32)
	      {
		sprintf(meta_val, "%u,", ((uint32 *)attr_buf)[0]);
		for (i=1; i<attr_cnt-1; i++)
		{
		  sprintf(tmpData, "%u,", ((uint32 *)attr_buf)[i]);
		  strcat(meta_val, tmpData);
		}
		sprintf(tmpData, "%u", ((uint32 *)attr_buf)[attr_cnt-1]);
		strcat(meta_val, tmpData);
	      } /* attr_type == DFNT_UINT32 */
	      else if(attr_type == DFNT_FLOAT32)
	      {
		sprintf(meta_val, "%f,", ((float32 *)attr_buf)[0]);
		for (i=1; i<attr_cnt-1; i++)
		{
		  sprintf(tmpData, "%f,", ((float32 *)attr_buf)[i]);
		  strcat(meta_val, tmpData);
		}
		sprintf(tmpData, "%f", ((float32 *)attr_buf)[attr_cnt-1]);
		strcat(meta_val, tmpData);
	      } /* attr_type == DFNT_FLOAT32 */
	      else if(attr_type == DFNT_FLOAT64)
	      {
		sprintf(meta_val, "%lf,", ((float64 *)attr_buf)[0]);
		for (i=1; i<attr_cnt-1; i++)
		{
		  sprintf(tmpData, "%lf,", ((float64 *)attr_buf)[i]);
		  strcat(meta_val, tmpData);
		}
		sprintf(tmpData, "%lf", ((float64 *)attr_buf)[attr_cnt-1]);
		strcat(meta_val, tmpData);
	      } /* attr_type == DFNT_FLOAT64 */
	      else
              {
                fprintf(stderr, "Don't know the data type %d, exit...\n", attr_type);
                exit(0);
              }
	
	    }  /* else attr_cnt > 1 */
	 }
       } /* else numerical types */
     }

    close_sd(&in_sds_info, in_sds_info.H5IN);

    if(attr_buf) free(attr_buf);
  }
}

void *get_sds_attr(sds_t *sds_info, char *attr_name, int32 *attr_type, int32 *attr_cnt)
{
    int32 attr_size;
    int32 attr_index;
    void *attr_buf = NULL;
    char name[80];
  
    hid_t attrh5_type, attr_space;
    hid_t attr_id, memtype;
    H5T_class_t type_class;
    size_t size;

    if (sds_info->H5IN)
    {
      if( (attr_id = H5Aopen(sds_info->sds_id,attr_name,H5P_DEFAULT)) < 0 )
      {
	/* MSZ 8/21/17. for no "offset" case, and its value is 0. */
	if(strcasestr(attr_name, "offset"))
	{
	  *attr_type = DFNT_FLOAT64;
	  *attr_cnt = 1;
	  attr_size = DFKNTsize(*attr_type);
	  if ((attr_buf = (void *)calloc(*attr_cnt, attr_size)) == NULL)
	    fprintf(stderr, "Cannot allocate memory for attr_buf in get_sds_attr\n");
	  else
	  {
	    double offset = 0.;
	    *(float64*)attr_buf = offset;
	  }
	}
	else
          /*fprintf(stderr, "Could not find the attribute %s for sds %s\n", attr_name, sds_info->name)*/;
      }
      else
      {
        /* Get attribute datatype, dataspace, rank, and dimensions.  */
        attr_space = H5Aget_space(attr_id);

	attrh5_type= (int32)H5Aget_type(attr_id);
	*attr_type = H5toH4type(attrh5_type);

	type_class = H5Tget_class(attrh5_type);

	/* MSZ 8/15/17, largely modified the below code */
	size       = H5Tget_size(attrh5_type);

	if(type_class == H5T_STRING)
	  size++; /* Make room for null terminator */

	/*
	if(type_class != -1)
	    fprintf(stderr, "ATTR_NAME=%s type_class=%d H5T_INTEGER=%d H5T_FLOAT=%d size = %d\n", attr_name, type_class, H5T_INTEGER, H5T_FLOAT, size);
	*/

	if(type_class == H5T_INTEGER || type_class == H5T_FLOAT)  /* numerical types */
	{
	    /* Get attribute datatype, dataspace, rank, and dimensions.  */
	    *attr_cnt = H5Sget_simple_extent_npoints(attr_space);

	    if ((attr_buf = (void *)calloc(*attr_cnt, size)) == NULL)
		fprintf(stderr, "Cannot allocate memory for attr_buf in get_sds_attr\n");
	    else
	        if (H5Aread(attr_id, attrh5_type, attr_buf) < 0)
		    fprintf(stderr, "Cannot read sds_attr %s in get_sds_attr\n", attr_name);
		
	}
	else if(type_class == H5T_STRING)        /* string */
	{
	    *attr_cnt = size;

	    if ((attr_buf = (void *)calloc(*attr_cnt, sizeof(char))) == NULL)
		fprintf(stderr, "Cannot allocate memory for attr_buf in get_sds_attr\n");
	    else
		if (H5Aread(attr_id, attrh5_type, attr_buf) < 0)
		    fprintf(stderr, "CANnot read sds_attr %s in get_sds_attr\n", attr_name);
	}
	else  /* we just interested in scale_factor, offset fill value etc not others */
	{
	   *attr_cnt = 0;
	   *attr_type = -1;
	   attr_buf = NULL;
	}

	H5Tclose(attrh5_type);
        H5Sclose(attr_space);
        H5Aclose(attr_id);
      } /* else for if( (attr_id = H5Aopen(sds_info->sds_id,attr_name,H5P_DEFAULT)) < 0 ) */
    }
    else
    {
        if ((attr_index = SDfindattr(sds_info->sds_id, attr_name)) == FAIL)
	{
	  /* MSZ 8/21/17. for no "offset" case, and its value is 0. */
	  if(strcasestr(attr_name, "offset"))
	  {
	     *attr_type = DFNT_FLOAT64;
	     *attr_cnt = 1;
	     attr_size = DFKNTsize(*attr_type);
	     if ((attr_buf = (void *)calloc(*attr_cnt, attr_size)) == NULL)
		fprintf(stderr, "Cannot allocate memory for attr_buf in get_sds_attr\n");
	     else
	     {
		double offset = 0.;
		*(float64*)attr_buf = offset;
	     }
	  }
	  else
              /*fprintf(stderr, "Could not find the attribute %s for sds %s\n", attr_name, sds_info->name)*/;
	}
        else
        {
          if (SDattrinfo(sds_info->sds_id, attr_index, name, attr_type, attr_cnt) == FAIL)
            fprintf(stderr, "Cannot read the attribute information for %s\n", attr_name);
          else
          {
            attr_size = DFKNTsize(*attr_type);
            if ((attr_buf = (void *)calloc(*attr_cnt, attr_size)) == NULL)
              fprintf(stderr, "Cannot allocate memory for attr_buf in get_sds_attr\n");
            else
            {
		int status;
		status = SDreadattr(sds_info->sds_id, attr_index, attr_buf);
		if(status == -1)
		{
		    fprintf(stderr, "Cannot read sds_attr %s in get_sds_attr\n", attr_name);
		    if(attr_buf) free(attr_buf);
		    attr_buf = NULL;
		}
            }
          }
        }
    }

    return attr_buf;
}

void write_attr_fval(int H5OUT, int32 sds_id, int32 fval_type, int c, int attr_val)
{
    int32 val_len;
    int status = 1;
    char attr_name[20] = {'\0'};  /* MSZ 6-05-2019, initialize */
    void *fill_val = NULL;        /* MSZ 6-05-2019, initialize */
  
    val_len = 1;
    strcpy(attr_name, "_FillValue");
    switch (fval_type)
    {
        case 5: if ((fill_val = (float32 *)calloc(1, sizeof(float32))) != NULL)
               ((float32 *)fill_val)[0] = (c == 0) ? (float32)FILL_VALUE_INT32 : (float32)attr_val;
             else status = -1;
             break; 
        case 6: if ((fill_val = (float64 *)calloc(1, sizeof(float64))) != NULL)
               ((float64 *)fill_val)[0] = (c == 0) ? (float32)FILL_VALUE_INT32 : (float64)attr_val;
             else status = -1;
             break; 	     
        case 20: if ((fill_val = (int8 *)calloc(1, sizeof(int8))) != NULL)
               ((int8 *)fill_val)[0] = (c == 0) ? (int8)FILL_VALUE_INT8 : (int8)attr_val;
             else status = -1;
             break; 
        case 21: if ((fill_val = (uint8 *)calloc(1, sizeof(uint8))) != NULL)
               ((uint8 *)fill_val)[0] = (c == 0) ? (uint8)FILL_VALUE_UINT8 : (uint8)attr_val;
             else status = -1;
             break; 
        case 22: if ((fill_val = (int16 *)calloc(1, sizeof(int16))) != NULL)
               ((int16 *)fill_val)[0] = (c == 0) ? (int16)FILL_VALUE_INT16 : (int16)attr_val;
             else status = -1;
             break; 
        case 23: if ((fill_val = (uint16 *)calloc(1, sizeof(uint16))) != NULL)
               ((uint16 *)fill_val)[0] = (c == 0) ? (uint16)FILL_VALUE_UINT16 : (uint16)attr_val;
             else status = -1;
             break; 
        case 24: if ((fill_val = (int32 *)calloc(1, sizeof(int32))) != NULL)
               ((int32 *)fill_val)[0] = (c == 0) ? (int32)FILL_VALUE_INT32 : (int32)attr_val;
             else status = -1;
             break; 
        case 25: if ((fill_val = (uint32 *)calloc(1, sizeof(uint32))) != NULL)
               ((uint32 *)fill_val)[0] = (c == 0) ? (uint32)FILL_VALUE_UINT32 : (uint32)attr_val;
             else status = -1;
             break; 
    }
    if (status == -1)
        fprintf(stderr, "Cannot allocate memory for fill_val\n");
    else
    {
        if (H5OUT)
        {
            /* fprintf(stderr, "fval_type = %d\n", fval_type); */
            if (setH5Attr(fval_type,attr_name,val_len,(VOIDP)fill_val, sds_id) == FAIL)
                fprintf(stderr, "Cannot write the attribute %s in write_attr_fval\n", attr_name);
        }
        else
        {
            if (SDsetattr(sds_id, attr_name, fval_type, val_len, (VOIDP)fill_val) == FAIL)
              fprintf(stderr, "Cannot write the attribute _FillValue in write_attr_fval\n");
        }
        if(fill_val) free(fill_val);
    }
}

void write_attr_fvalf(int H5OUT, int32 sds_id, int32 fval_type, int c, double attr_val)
{
    int32 val_len;
    int status = 1;
    char attr_name[20];
    void *fill_val;
  
    val_len = 1;
    strcpy(attr_name, "_FillValue");
    switch (fval_type)
    {
      case 5: if ((fill_val = (float32 *)calloc(1, sizeof(float32))) != NULL)
             ((float32 *)fill_val)[0] = (c == 0) ? (float32)FILL_VALUE_INT32 : (float32)attr_val;
           else status = -1;
           break; 
      case 6: if ((fill_val = (float64 *)calloc(1, sizeof(float64))) != NULL)
             ((float64 *)fill_val)[0] = (c == 0) ? (float32)FILL_VALUE_INT32 : (float64)attr_val;
           else status = -1;
           break; 	     
      case 20: if ((fill_val = (int8 *)calloc(1, sizeof(int8))) != NULL)
             ((int8 *)fill_val)[0] = (c == 0) ? (int8)FILL_VALUE_INT8 : (int8)attr_val;
           else status = -1;
           break; 
      case 21: if ((fill_val = (uint8 *)calloc(1, sizeof(uint8))) != NULL)
             ((uint8 *)fill_val)[0] = (c == 0) ? (uint8)FILL_VALUE_UINT8 : (uint8)attr_val;
           else status = -1;
           break; 
      case 22: if ((fill_val = (int16 *)calloc(1, sizeof(int16))) != NULL)
             ((int16 *)fill_val)[0] = (c == 0) ? (int16)FILL_VALUE_INT16 : (int16)attr_val;
           else status = -1;
           break; 
      case 23: if ((fill_val = (uint16 *)calloc(1, sizeof(uint16))) != NULL)
             ((uint16 *)fill_val)[0] = (c == 0) ? (uint16)FILL_VALUE_UINT16 : (uint16)attr_val;
           else status = -1;
           break; 
      case 24: if ((fill_val = (int32 *)calloc(1, sizeof(int32))) != NULL)
             ((int32 *)fill_val)[0] = (c == 0) ? (int32)FILL_VALUE_INT32 : (int32)attr_val;
           else status = -1;
           break; 
      case 25: if ((fill_val = (uint32 *)calloc(1, sizeof(uint32))) != NULL)
             ((uint32 *)fill_val)[0] = (c == 0) ? (uint32)FILL_VALUE_UINT32 : (uint32)attr_val;
           else status = -1;
           break; 
    }
    if (status == -1)
      fprintf(stderr, "Cannot allocate memory for fill_val\n");
    else
    {
          if (H5OUT)
          {
              /* fprintf(stderr, "fval_type = %d\n", fval_type); */
              if (setH5Attr(fval_type,attr_name,val_len,(VOIDP)fill_val, sds_id) == FAIL)
                  fprintf(stderr, "Cannot write the attribute %s in write_attr_fvalf\n", attr_name);
          }
          else
          {
              if (SDsetattr(sds_id, attr_name, fval_type, val_len, (VOIDP)fill_val) == FAIL)
                fprintf(stderr, "Cannot write the attribute _FillValue in write_attr_fvalf\n");
          }
          free(fill_val);
    }
}

void write_attr_fsoval(char *fname, int32 sds_id, int32 so_type, float64 attr_sval, float64 attr_oval)
{
  int32 val_len;
  int status = 1;
  char attr_name[50], scale_attr_name[50], offset_attr_name[50];
  void *so_val;
  char gran_id[MAX_GID_LEN]; 
  
  strcpy(gran_id, fname); 	
  rm_path(gran_id);
  
  if (attr_sval != FSO_FILL && so_type != SO_TYPE)
  {
    val_len = 1;
    if (strstr(gran_id, "NPP_") != NULL)  
    {
      if ((strstr(gran_id, "NPP_D8SRF1KM_L3D") != NULL) || (strstr(gran_id, "NPP_D8SRFHKM_L3D") != NULL) ||
          (strstr(gran_id, "NPP_D16BRDF1KM_L3D") != NULL) || (strstr(gran_id, "NPP_D16BRDFHKM_L3D") != NULL) ||
	  (strstr(gran_id, "NPP_D16VI1KM_L3D") != NULL) || (strstr(gran_id, "NPP_D16VIHKM_L3D") != NULL) ||
	  (strstr(gran_id, "NPP_D16BRDF1_L3D") != NULL) || (strstr(gran_id, "NPP_D16BRDF3_L3D") != NULL) ||
	  (strstr(gran_id, "NPP_D16BRDF4_L3D") != NULL)) 
      {
	strcpy(scale_attr_name, "scale_factor");
	strcpy(offset_attr_name, "add_offset");
      }  
      else 
      {
	strcpy(scale_attr_name, "Scale");    
	strcpy(offset_attr_name, "Offset"); 
      }  
    }   
    else 
    {
      strcpy(scale_attr_name, "scale_factor"); 
      strcpy(offset_attr_name, "add_offset"); 
    }
    switch (so_type)
    {
      case 5:  if ((so_val = (float32 *)calloc(1, sizeof(float32))) != NULL)
	          ((float32 *)so_val)[0] = (float32)attr_sval;
	       else status = -1;
	       break; 
      case 6:  if ((so_val = (float64 *)calloc(1, sizeof(float64))) != NULL)
	          ((float64 *)so_val)[0] = (float64)attr_sval;
	       else status = -1;
	       break; 	       
      case 20: if ((so_val = (int8 *)calloc(1, sizeof(int8))) != NULL)
		 ((int8 *)so_val)[0] = (int8)attr_sval;
	       else status = -1;
	       break; 
      case 21: if ((so_val = (uint8 *)calloc(1, sizeof(uint8))) != NULL)
		 ((uint8 *)so_val)[0] = (uint8)attr_sval;
	       else status = -1;
	       break; 
      case 22: if ((so_val = (int16 *)calloc(1, sizeof(int16))) != NULL)
		 ((int16 *)so_val)[0] = (int16)attr_sval;
	       else status = -1;
	       break; 
      case 23: if ((so_val = (uint16 *)calloc(1, sizeof(uint16))) != NULL)
		 ((uint16 *)so_val)[0] = (uint16)attr_sval;
	       else status = -1;
	       break; 
      case 24: if ((so_val = (int32 *)calloc(1, sizeof(int32))) != NULL)
		 ((int32 *)so_val)[0] = (int32)attr_sval;
	       else status = -1;
	       break; 
      case 25: if ((so_val = (uint32 *)calloc(1, sizeof(uint32))) != NULL)
		 ((uint32 *)so_val)[0] = (uint32)attr_sval;
	       else status = -1;
	       break; 
    }
    if (status == -1)
      fprintf(stderr, "Cannot allocate memory for so_val\n");
    else
    {
      if (SDsetattr(sds_id, scale_attr_name, so_type, val_len, (VOIDP)so_val) == FAIL)
	fprintf(stderr, "Cannot write the attribute Scale in write_attr_soval\n");
      free(so_val);
    }

    switch (so_type)
    {
      case 5:  if ((so_val = (float32 *)calloc(1, sizeof(float32))) != NULL)
	          ((float32 *)so_val)[0] = (float32)attr_oval;
	       else status = -1;
	       break; 
      case 6:  if ((so_val = (float64 *)calloc(1, sizeof(float64))) != NULL)
	          ((float64 *)so_val)[0] = (float64)attr_oval;
	       else status = -1;
	       break; 		       
      case 20: if ((so_val = (int8 *)calloc(1, sizeof(int8))) != NULL)
		 ((int8 *)so_val)[0] = (int8)attr_oval;
	       else status = -1;
	       break; 
      case 21: if ((so_val = (uint8 *)calloc(1, sizeof(uint8))) != NULL)
		 ((uint8 *)so_val)[0] = (uint8)attr_oval;
	       else status = -1;
	       break; 
      case 22: if ((so_val = (int16 *)calloc(1, sizeof(int16))) != NULL)
		 ((int16 *)so_val)[0] = (int16)attr_oval;
	       else status = -1;
	       break; 
      case 23: if ((so_val = (uint16 *)calloc(1, sizeof(uint16))) != NULL)
		 ((uint16 *)so_val)[0] = (uint16)attr_oval;
	       else status = -1;
	       break; 
      case 24: if ((so_val = (int32 *)calloc(1, sizeof(int32))) != NULL)
		 ((int32 *)so_val)[0] = (int32)attr_oval;
	       else status = -1;
	       break; 
      case 25: if ((so_val = (uint32 *)calloc(1, sizeof(uint32))) != NULL)
		 ((uint32 *)so_val)[0] = (uint32)attr_oval;
	       else status = -1;
	       break; 
    }
    if (status == -1)
      fprintf(stderr, "Cannot allocate memory for so_val\n");
    else
    {
      if (SDsetattr(sds_id, offset_attr_name, so_type, val_len, (VOIDP)so_val) == FAIL)
	fprintf(stderr, "Cannot write the attribute Offset in write_attr_soval\n");
      free(so_val);
    }  
  }
}

#if 0
void write_sds_attrs(int H5IN, int32 in_sds_id, int H5OUT, int32 out_sds_id, int bn)
{
    int attr_id;
    int32 attr_index, attr_cnt;
    int32 attr_type, attr_size;
    int st_attr_id, end_attr_id;
    char attr_name[40];
    void *attr_buf;
    float32 *attr_val;

    char *attr_names[] = {"units", "valid_range", "_FillValue", "scale_factor", "radiance_scales",
                          "radiance_offsets", "radiance_units", "reflectance_scales",
                          "reflectance_offsets", "reflectance_units", "corrected_counts_scales",
                          "corrected_counts_offsets", "corrected_counts_units" };
    int attr_vect[13] = {0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0};
      
    st_attr_id = (bn < 0) ? 0 : 1;
    end_attr_id = (bn < 0) ? 4 : 13;
    if (bn >= 22) bn -= 22;
    else if (bn >= 7) bn -= 7;
    else if (bn >= 2) bn -= 2;
    
    for (attr_id = st_attr_id; attr_id < end_attr_id; attr_id++)
    {
        if (H5OUT)
        {
            if ((attr_buf = get_sds_attr(H5IN, in_sds_id, attr_names[attr_id], &attr_type, &attr_cnt)) == NULL)
                fprintf(stderr,"Failed to get value for attribute: %s in write_sds_attrs\n", attr_names[attr_id]);
            else
            {
                if (attr_vect[attr_id] == 0)
                    setH5Attr(attr_type,attr_names[attr_id],attr_cnt,(VOIDP)attr_buf, out_sds_id);
                else
                {
                    attr_val = (float32 *)calloc(1, sizeof(float32));
                    attr_val[0] = ((float32 *)attr_buf)[bn];
                    setH5Attr(attr_type,attr_names[attr_id],1,(VOIDP)attr_buf, out_sds_id);
                    free(attr_val);
                }
            }
        }
        else
        {
            if ((attr_index = SDfindattr(in_sds_id, attr_names[attr_id])) != FAIL)
            {
                if (SDattrinfo(in_sds_id, attr_index, attr_name, &attr_type, &attr_cnt) != FAIL)
                {
                    attr_size = DFKNTsize(attr_type);
                    if ((attr_buf = (void *)calloc(attr_cnt, attr_size)) == NULL)
                      fprintf(stderr, "Cannot allocate memory for attr_buf in write_sds_attrs\n");
                    else
                    {
                      if (SDreadattr(in_sds_id, attr_index, attr_buf) != FAIL)
                      {
                        if (attr_vect[attr_id] == 0)
                            SDsetattr(out_sds_id, attr_name, attr_type, attr_cnt, attr_buf);
                    else
                        {
                          attr_val = (float32 *)calloc(1, sizeof(float32));
                          attr_val[0] = ((float32 *)attr_buf)[bn];
                          SDsetattr(out_sds_id, attr_name, attr_type, 1, (VOIDP)attr_val);
                          free(attr_val);
                        }
                      }
                      free(attr_buf);
                    }
                }
            }
        }
    }
}
#endif
int set_sd_attr(int H5OUT, int out_sd_id, char *attr_name, int attr_data_type, int n_values, void *values)
{
    /* MSZ 2/7/2018 */
    if (H5OUT)
    {
      if(setH5Attr(attr_data_type, attr_name, n_values, values, (hid_t)out_sd_id) == FAIL)
      {
        fprintf(stderr, "Cannot write attr %s to sd\n", attr_name);
	return -1;
      }
      else
        return 0;
    }
    else
    {
      if(SDsetattr(out_sd_id, attr_name, attr_data_type, n_values, values) == FAIL)
      {
	fprintf(stderr, "Cannot write attr %s to sd\n", attr_name);
	return -1;
      }
      else
      	return 0;
    }
}

int set_sds_attr(int H5OUT, int out_sds_id, char *attr_name, int attr_data_type, int n_values, void *values)
{
    /* MSZ 2/7/2018 */
    if (H5OUT)
    {
      if(setH5Attr(attr_data_type, attr_name, n_values, values, (hid_t)out_sds_id) == FAIL)
      {
        fprintf(stderr, "Cannot write attr %s to sds\n", attr_name);
	return -1;
      }
      else
      	return 0;
    }
    else
    {
      if(SDsetattr(out_sds_id, attr_name, attr_data_type, n_values, values) == FAIL)
      {
	fprintf(stderr, "Cannot write attr %s to sds\n", attr_name);
	return -1;
      }
      else
      	return 0;
    }
}

void write_all_sds_attr(int H5IN, int32 in_sds_id, int H5OUT, int32 out_sds_id, int32 nattr)
{
    void *attr_buf;
    int32 attr_index, attr_cnt;
    int32 attr_type, attr_size;
    char attr_name[MAX_ATTR_NAME_LEN];
    
    /* HDF5 addition */
    H5_ATTR h5attr;
    
    if (H5IN)
    {
        h5attr.loc_id = out_sds_id;
        h5attr.rdata = NULL;
        h5attr.name[0] = '\0';
        if (H5OUT)
            H5Aiterate2(in_sds_id,H5_INDEX_CRT_ORDER,H5_ITER_INC,NULL,(H5A_operator2_t)H5_attr_write_hdf5,(void *)&h5attr);
        else
            H5Aiterate2(in_sds_id,H5_INDEX_CRT_ORDER,H5_ITER_INC,NULL,(H5A_operator2_t)H5_attr_write_hdf4,(void *)&h5attr);
    }
    else
    {
        for (attr_index=0; attr_index<nattr; attr_index++)
        {
            if (SDattrinfo(in_sds_id, attr_index, attr_name, &attr_type, &attr_cnt) != FAIL)
            {
                attr_size = DFKNTsize(attr_type);
                if ((attr_buf = (void *)calloc(attr_cnt, attr_size)) == NULL)
                    fprintf(stderr, "Cannot allocate memory for attr_buf in write_all_sds_attr\n");
                else
                {
                    if (SDreadattr(in_sds_id, attr_index, attr_buf) != FAIL)
                    {
                        if (H5OUT)
                            setH5Attr(attr_type,attr_name,attr_cnt,(VOIDP)attr_buf, out_sds_id);
                        else
                            SDsetattr(out_sds_id, attr_name, attr_type, attr_cnt, attr_buf);
                    }
                    free(attr_buf);
                }
            }
        }
    }
}

void compute_sds_start_offset(sds_t *sds_info, int n, int m, int *st_c, int *offset)
{
  /* MSZ, added comments here for clarification.
   * n and m are 0-based non-XY dim data layers, please check get_sdsname_dim() in str_op.c. 
   * For 3D or 4D array data, n and m refers to specified start subscripts 
   * of non-YX dimensions which n refers to a low dimension order or a slower
   * dimension than m.  Therefore, if it is 4D case, for BSQ interleave, 
   * the start location is [n][m][Y][X] whereas for BIP [Y][X][n][m].
   *
   * st_c:   the start of data vector (storing multi-dim data) for the specified n data layer.
   * offset: the shift of data vector (storing multi-dim data) for each observation or if n and/or m is plus 1.
   *
   * Note that usually, just st_c is being used. */

  int rank;

  rank = sds_info->rank;
  if ((rank == 2) || ((n == -1) && (m == -1)))            /* not specified n and/or m or just 2D data */
  {
    *st_c = 0;
    *offset = 1;
  }
  else     /* MSZ for >2D or >=3D case */
  {
    if ((n != -1) && (m == -1))
    {
      if (sds_info->dim_size[0] > sds_info->dim_size[rank-1])              /* BIP */
      {
        *st_c = n;
        *offset = sds_info->dim_size[rank-1];
      }
      else                                                                 /* BSQ */
      {
        *st_c = n*sds_info->dim_size[rank-1];
        *offset = 1;
      }
    }
    if ((n != -1) && (m != -1))
    {
      if (sds_info->dim_size[0] > sds_info->dim_size[rank-1])              /* BIP */
      {
        *st_c = n*sds_info->dim_size[rank-1] + m;
        *offset = sds_info->dim_size[rank-1]*sds_info->dim_size[rank-2];
      }
      else                                                                 /* BSQ */
      {
/*
        *st_c = n*sds_info->dim_size[rank-1]*sds_info->dim_size[1] +  m*sds_info->dim_size[1];
*/
        *st_c = n*sds_info->dim_size[rank-1]*sds_info->dim_size[1] +  m*sds_info->dim_size[rank-1];
        *offset = 1;
      }
    }
  }
}

void update_nd_sdsnames(char **sds_names, int *sds_cnt, char *fname)
{
  int i, j, nsds, isds, cnt;
  int p1, p2, p11, len;
  char **new_sds_names;
  char nd_ext[20], md_ext[20];
  char sds_name[MAX_SDS_NAME_LEN];
  char **tmp_sds_names;
  sds_t sds_info;  
      
  if ((tmp_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for tmp_sds_names in update_nd_sdsnames\n");  
  if ((new_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for new_sds_name in update_nd_sdsnames\n");
  if ((tmp_sds_names != NULL) && (new_sds_names != NULL))
  {
    cnt = get_sds_names(fname, tmp_sds_names);
    nd_ext[0] = md_ext[0] = '\0';
    for (isds=0, nsds=0; isds < *sds_cnt; isds++)
    {
      if (charpos(sds_names[isds], '(', 0) == -1)
      { 
	if ((p1 = charpos(sds_names[isds], '.', 0)) != -1)
        {
          strmid(sds_names[isds], 0, p1, sds_name);	  
	  sds_info.sd_id = sds_info.sds_id = -1;
	  strcpy(sds_info.name, sds_name);	  	  	 	  
	  if (get_sds_info(fname, &sds_info) != -1)
	  {	  
	    p1++;
	    len = (int)strlen(sds_names[isds]);
	    if ((p2 = charpos(sds_names[isds], '.', p1)) != -1) 
	    {
	      strmid(sds_names[isds], p1, p2-p1, nd_ext); 
	      p2++;
	      strmid(sds_names[isds], p2, len-p2, md_ext); 
	      create_names(fname, sds_name, nd_ext, md_ext, &nsds, new_sds_names);	
	    }
	    else
	    {
	      strmid(sds_names[isds], p1, len-p1, nd_ext);
	      create_names(fname, sds_name, nd_ext, md_ext, &nsds, new_sds_names);	
	    }	  
	  }
	  else
	  {
	    p11 = charpos(sds_names[isds], '.', p1+1);

	    for (i=0, j=0; i<cnt; i++)
	      if (strcmp(sds_name, tmp_sds_names[i]) != 0) 
		j++;    
	    if (j == cnt)	
	    {
	      if (p11 != -1)
	      {	  
		p1++;
		if ((p1 = charpos(sds_names[isds], '.', p1)) != -1) 
		{		      		      		      
        	  strmid(sds_names[i], 0, p11, sds_name);
		  p1++;
        	  len = (int)strlen(sds_names[isds]);
        	  if ((p2 = charpos(sds_names[isds], '.', p1)) != -1) 
        	  {
		    strmid(sds_names[isds], p1, p2-p1, nd_ext); 
		    p2++;
		    strmid(sds_names[isds], p2, len-p2, md_ext); 
		    create_names(fname, sds_name, nd_ext, md_ext, &nsds, new_sds_names);	
        	  }
        	  else
        	  {
		    strmid(sds_names[isds], p1, len-p1, nd_ext);
		    create_names(fname, sds_name, nd_ext, md_ext, &nsds, new_sds_names);
        	  }	      
		}
	      }  
	      else
	      {	 	    	        	     	     	     	      	                    
		strcpy(new_sds_names[isds], sds_names[isds]);
        	nsds++;      
	      }		    
	    }  
	    else
	    {
	      if (p11 != -1)
	      {	  
		p1 = 0;
		if ((p1 = charpos(sds_names[isds], '.', p1)) != -1) 
		{		      		      		      
        	  p1++;
        	  len = (int)strlen(sds_names[isds]);
        	  if ((p2 = charpos(sds_names[isds], '.', p1)) != -1) 
        	  {
		    strmid(sds_names[isds], p1, p2-p1, nd_ext); 
		    p2++;
		    strmid(sds_names[isds], p2, len-p2, md_ext); 
		    create_names(fname, sds_name, nd_ext, md_ext, &nsds, new_sds_names);	
        	  }
        	  else
        	  {
		    strmid(sds_names[isds], p1, len-p1, nd_ext);
		    create_names(fname, sds_name, nd_ext, md_ext, &nsds, new_sds_names);
        	  }	      
		}
	      }  
	      else
	      {	 	    	        	     	     	     	      	      
        	strcpy(new_sds_names[nsds], sds_names[nsds]);
        	nsds++;
	      }		  
	    }
	  }  	  	  	  	  
        }
        else 
        {
          strcpy(new_sds_names[nsds], sds_names[isds]);
          nsds++;
        }
      }
      else
      {
        strcpy(new_sds_names[nsds], sds_names[isds]);
        nsds++;
      }
    }
    for (isds=0; isds<nsds; isds++)
      strcpy(sds_names[isds], new_sds_names[isds]);
    *sds_cnt = nsds;
    Free2D((void **)new_sds_names);
    Free2D((void **)tmp_sds_names);    
  }
}


/*void update_nd_sdsnames(char **sds_names, int *sds_cnt, char *fname) */
/* 
   Create new sds_names.
        e.g. input sds_name is Surf_refl.1.2-3
             The resulting sds_names will be Surf_refl.1.2
	                                     Surf_refl.1.3
	     The sds_cnt will be set to 2 since there are 2 resulting sds names
   Updated to fix the problem with SDS names containing "." character 
   	by Xiaoping Zhang, Sigma Space Corp. on Jun 2012
*/
/*
{
  int nsds, isds;
  int p1, p2, p11, len;
  char **new_sds_names;
  char nd_ext[20], md_ext[20];
  char sds_name[MAX_SDS_NAME_LEN];
  sds_t sds_info;

  memset( &sds_info, 0, sizeof( sds_t ) );
  if ((new_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for new_sds_name in update_nd_sdsnames\n");
  else
    {
      nd_ext[0] = md_ext[0] = '\0';
      for (isds=0, nsds=0; isds < *sds_cnt; isds++)
	{
	  if (charpos(sds_names[isds], '(', 0) == -1)
	{      
	  if ((p1 = charpos(sds_names[isds], '.', 0)) != -1)
	  {		  
	    sds_info.sd_id = sds_info.sds_id = -1;		  
	    strmid(sds_names[isds], 0, p1, sds_name);
	    strcpy(sds_info.name, sds_name);	    
	    if (get_sds_info(fname, &sds_info) != -1)
	    {	      
	      p1++;
	      len = (int)strlen(sds_names[isds]);
	      if ((p2 = charpos(sds_names[isds], '.', p1)) != -1) 
	      {
		strmid(sds_names[isds], p1, p2-p1, nd_ext); 
		p2++;
		strmid(sds_names[isds], p2, len-p2, md_ext); 
		create_names(fname, sds_name, nd_ext, md_ext, &nsds, new_sds_names);	
	      }
	      else
	      {
		strmid(sds_names[isds], p1, len-p1, nd_ext);
		create_names(fname, sds_name, nd_ext, md_ext, &nsds, new_sds_names);	
	      }
	    } 
	    else
	    {
	      p1++;
	      if ((p11 = charpos(sds_names[isds], '.', p1)) != -1) 
	      {		      		      		      
		strmid(sds_names[isds], 0, p11, sds_name);
		strcpy(sds_info.name, sds_name);		
		len = (int)strlen(sds_names[isds]);
		if (get_sds_info((char *)NULL, &sds_info) != -1)
		{		
		  p11++;
		  if ((p2 = charpos(sds_names[isds], '.', p11)) != -1) 
		  {		    
		    strmid(sds_names[isds], p11, p2-p11, nd_ext); 
		    p2++;
		    strmid(sds_names[isds], p2, len-p2, md_ext); 
		    create_names(fname, sds_name, nd_ext, md_ext, &nsds, new_sds_names);	
		  }
		  else
		  {
		    strmid(sds_names[isds], p11, len-p11, nd_ext);
		    create_names(fname, sds_name, nd_ext, md_ext, &nsds, new_sds_names);	
		  }
	        }
		else
		  fprintf(stderr, "Cannot find the SDS %s \n", sds_info.name);			      
	      }	
	      else
	      {
		strcpy(new_sds_names[nsds], sds_names[isds]);
		nsds++;	      
	      }	    		    		    
	    }		   
	  }
	  else 
	  {
	    strcpy(new_sds_names[nsds], sds_names[isds]);
	    nsds++;
	  }
	}
	else
	{
	  strcpy(new_sds_names[nsds], sds_names[isds]);
	  nsds++;
	}
	if (sds_info.sds_id != -1) SDendaccess(sds_info.sds_id);
	if (sds_info.sd_id != -1) SDendaccess(sds_info.sd_id);	
      }
      for (isds=0; isds<nsds; isds++)
	strcpy(sds_names[isds], new_sds_names[isds]);
      *sds_cnt = nsds;
      
      Free2D((void **)new_sds_names);
    }
}
*/

void create_names(char *fname, char *sds_name, char *nd_ext, char *md_ext, int *nsds, char **sds_names)
{
    int n, m;
    int ndim, mdim;
    int n_cnt, m_cnt;
    int narr[20], marr[20];
    sds_t sds_info;
  
    sds_info.sd_id = sds_info.sds_id = -1;
    strcpy(sds_info.name, sds_name);
    if (get_sds_info(fname, &sds_info) != -1)
    {
      n_cnt = m_cnt = 0;
      if (sds_info.rank > 2)
      {
        if (sds_info.dim_size[0] > sds_info.dim_size[2]) 
      { ndim = sds_info.dim_size[2]; mdim = sds_info.dim_size[3]; }
        else 
      { ndim = sds_info.dim_size[0]; mdim = sds_info.dim_size[1]; }
        get_dim_num(nd_ext, narr, ndim, &n_cnt);
        if (sds_info.rank > 3) get_dim_num(md_ext, marr, mdim, &m_cnt);
      }
      if (sds_info.rank == 3)
        for (n=0; n<n_cnt; n++, ++*nsds)
          sprintf(sds_names[*nsds], "%s.%d", sds_name, narr[n]);
      else if (sds_info.rank == 4)
      {
        for (n=0; n<n_cnt; n++)
          for (m=0; m<m_cnt; m++, ++*nsds)
        sprintf(sds_names[*nsds], "%s.%d.%d", sds_name, narr[n], marr[m]); 
      }
    }
    if (sds_info.H5IN)
    {
      if (sds_info.sds_id != -1) H5Dclose(sds_info.sds_id);
      if (sds_info.sd_id != -1) H5Fclose(sds_info.sd_id);
    }
    else
    {
      if (sds_info.sds_id != -1) SDendaccess(sds_info.sds_id);
      if (sds_info.sd_id != -1) SDendaccess(sds_info.sd_id);
    }
}

void get_dim_num(char *str_arr, int *num_arr, int dim_size, int *cnt)
{
  int len;
  char c;
  int i, j, k;
  int bn1, bn2;
  char nstr[5];

  if (str_arr[0] == '*')
  {
    for (i=0; i<dim_size; i++)
      num_arr[i] = i+1;
    *cnt = dim_size;
  }
  else
  {
    c  = ',';
    len = (int)strlen(str_arr);
    for (i=0, j=0; i<=len; i++)
    {
      if ((i<len) && (str_arr[i] != ',') && (str_arr[i] != '-'))
        nstr[j++] = str_arr[i];    
      else 
      {
	nstr[j] = '\0';
	j = 0;
	if (c == ',') { num_arr[*cnt] = (int)atoi(nstr); ++*cnt; c = str_arr[i]; }
        else if (c == '-') 
        {
	  bn1 = num_arr[*cnt - 1];
	  bn2 = (int)atoi(nstr);
	  for (k=bn1+1; k<=bn2; k++) { num_arr[*cnt] = k; ++*cnt; }
	}	 
      }
    }
  }
} 

void update_l2g_sdsnames(char **sds_names, int *sds_cnt, char *fname, int nobs)
{
    sds_t sds_info;
    int p1, len, n, n_cnt;
    int nsds, isds, narr[255];
    char sds_name[MAX_SDS_NAME_LEN];
    char **new_sds_names, nd_ext[20];
  
    if ((new_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
        fprintf(stderr, "Cannot allocate memory for new_sds_name in sds_rw: update_l2g_sdsnames()\n");
    else
    {
        nd_ext[0] = '\0';
        for (isds=0, nsds=0; isds < *sds_cnt; isds++)
        {
          if ((p1 = charpos(sds_names[isds], '.', 0)) == -1)
        fprintf(stderr, "The l2g SDS name %s is not accepted in sds_rw: update_l2g_sdsnames()\n", 
                sds_names[isds]);
          else
          {
                strmid(sds_names[isds], 0, p1, sds_name);
                len = (int)strlen(sds_names[isds]);
                strmid(sds_names[isds], p1+1, len-p1-1, nd_ext);
                sds_info.sd_id = sds_info.sds_id = -1;
                strcpy(sds_info.name, sds_name);
            strcat(sds_info.name, "_1");
                if (get_sds_info(fname, &sds_info) != -1)
                {
                    n_cnt = 0;
                    get_dim_num(nd_ext, narr, nobs, &n_cnt);
                    for (n=0; n<n_cnt; n++, nsds++)
                        sprintf(new_sds_names[nsds], "%s.%d", sds_name, narr[n]);	      
                }
                if (sds_info.H5IN)
                {
                  if (sds_info.sds_id != -1) H5Dclose(sds_info.sds_id);
                  if (sds_info.sd_id != -1) H5Fclose(sds_info.sd_id);
                }
                else
                {
                    if (sds_info.sds_id != -1) SDendaccess(sds_info.sds_id);
                    if (sds_info.sd_id != -1) SDendaccess(sds_info.sd_id);
                }
          }
        }
        for (isds=0; isds<nsds; isds++)
          strcpy(sds_names[isds], new_sds_names[isds]);
        *sds_cnt = nsds;
        Free2D((void **)new_sds_names);
    }
}

void compute_sds_nrows_ncols(sds_t *sds_info, int *nrows, int *ncols)
{
  int nr, nc;
  int i, rank;

  rank = sds_info->rank;
  if (rank == 2)
  {
    nr = sds_info->dim_size[0];
    nc = sds_info->dim_size[1];
  }
  else
  {
    if (sds_info->dim_size[0] < sds_info->dim_size[rank-1])
    {
      nr = sds_info->dim_size[rank-2];
      nc = sds_info->dim_size[rank-1];
      for (i=0; i<rank-2; i++)
        nc = nc*sds_info->dim_size[i];
    }
    else
    {
      nr = sds_info->dim_size[0];
      nc = sds_info->dim_size[1];
      for (i=2; i<rank; i++)
        nc = nc*sds_info->dim_size[i];
    }                                                     
  }
  *nrows = nr;
  *ncols = nc;
}

int compute_sds_ndata(sds_t *sds_info)
{
  int i, rank, ndata;

  rank = sds_info->rank;
  if (rank == 2)
    ndata = sds_info->dim_size[1];
  else
  {
    if (sds_info->dim_size[0] < sds_info->dim_size[rank-1])
    {
      ndata = sds_info->dim_size[rank-1];
      for (i=0; i<rank-2; i++)
        ndata *= sds_info->dim_size[i];
    }
    else
    {
      ndata = sds_info->dim_size[1];
      for (i=2; i<rank; i++)
        ndata *= sds_info->dim_size[i];
    }
  }
  return ndata;
}

void get_sds_edge(sds_t *sds_info, int32 *edge)
{
  int i, rank;
 
  rank = sds_info->rank;
  for (i=0; i<rank; i++)
    edge[i] = 0;
    
  if (rank == 1)
    edge[0] = sds_info->dim_size[0];
  else if (rank == 2)
  {
    edge[0] = 1;
    edge[1] = sds_info->dim_size[1];
  }
  else if (rank > 2)
  {
    for (i=0; i<rank; i++)
      edge[i] = sds_info->dim_size[i];
    if (sds_info->dim_size[0] < sds_info->dim_size[rank-1])
      edge[rank-2] = 1;
    else edge[0] = 1;
  }
}                                                           

void get_sds_param(sds_t *sds_info, int *n, int *m, int *rank, int *dim_size)
{
  int i, rank1;
  char sds_name[MAX_SDS_NAME_LEN];

  get_sdsname_dim(sds_info->name, sds_name, n, m);
  if ((*n == -1) && (*m == -1))
  {
    *rank = sds_info->rank;
    for (i=0; i < *rank; i++)
      dim_size[i] = sds_info->dim_size[i];
  }
  else
  {
    *rank = 2;
    rank1 = sds_info->rank;
    if (sds_info->dim_size[0] < sds_info->dim_size[rank1-1])
    {
      dim_size[0] = sds_info->dim_size[rank1-2];
      dim_size[1] = sds_info->dim_size[rank1-1];
    }
    else
    {
      dim_size[0] = sds_info->dim_size[0];
      dim_size[1] = sds_info->dim_size[1];
    }
  }
}

void write_metadata(int H5IN, int32 in_sd_id, int H5OUT, int32 out_sd_id)
{
    /* Maosheng Zhao, 4/4/2018. Note that if you don't want to
     * copy StructMetadata.0, use the below writeMetadata()
     */ 
    char attr_name[80];
    void *meta_attr_buf;
    int32 attr_cnt, attr_size;
    int32 nsds, nattr, attr_len;
    int32 attr_index, attr_type;
    /* HDF5 addition */
    H5_ATTR h5attr;                                                                                                        
    fprintf(stderr, "\tWriting all global metadata to output file\n");
    if (H5IN)
    {
        h5attr.loc_id = out_sd_id;
        h5attr.rdata = NULL;
        h5attr.name[0] = '\0';
        if (H5OUT)
            H5Aiterate2(in_sd_id,H5_INDEX_CRT_ORDER,H5_ITER_INC,NULL,(H5A_operator2_t)H5_attr_write_hdf5,(void *)&h5attr);
        else
            H5Aiterate2(in_sd_id,H5_INDEX_CRT_ORDER,H5_ITER_INC,NULL,(H5A_operator2_t)H5_attr_write_hdf4,(void *)&h5attr);
    }
    else
    {  
        if (SDfileinfo(in_sd_id, &nsds, &nattr) == FAIL)
            fprintf(stderr, "Cannot read file information for the input HDF file\n");
        else
        {
            fprintf(stderr, "Total %d attrs for global metadata.\n", nattr);
            for (attr_index=0; attr_index<nattr; attr_index++)
            {
              if (SDattrinfo(in_sd_id, attr_index, attr_name, &attr_type, &attr_cnt) == FAIL)
                fprintf(stderr, "Cannot read information for attribute %d in input HDF file\n", attr_index);
              else
              {   
                attr_size = DFKNTsize(attr_type);
                attr_len = attr_size * attr_cnt + 1;
                if ((meta_attr_buf = (void *)calloc(attr_len, attr_size)) == NULL)
                  fprintf(stderr, "Cannot allocate memory for metadata in write_metadata\n");
                else
                {
                    if (SDreadattr(in_sd_id, attr_index, meta_attr_buf) == FAIL)
                      fprintf(stderr, "Cannot read %s from input HDF file\n", attr_name);
                    else
                    {
                         if (H5OUT &&
                            setH5Attr(attr_type,attr_name,attr_cnt,(VOIDP)meta_attr_buf, out_sd_id) == FAIL)
                            fprintf(stderr, "Cannot write metadata %s to output HDF5 file\n", attr_name);
                        else if (!H5OUT && SDsetattr(out_sd_id, attr_name, attr_type, attr_cnt, meta_attr_buf) == FAIL)
                                fprintf(stderr, "Cannot write metadata %s to output HDF file\n", attr_name);
                    }
                    free(meta_attr_buf);
                }
              }
            } /* for (attr_index=0; . . .      */
        }
    }
}

void writeMetadata(int H5IN, int32 in_sd_id, int H5OUT, int32 out_sd_id)
{
    /* Maosheng Zhao 4/4/2018 */
    /* Note that the difference between above write_metadata() and this writeMetadata()
     * is that this function will not copy StructMetadata.0 */
    char attr_name[80];
    void *meta_attr_buf;
    int32 attr_cnt, attr_size;
    int32 nsds, nattr, attr_len;
    int32 attr_index, attr_type;
    /* HDF5 addition */
    H5_ATTR h5attr;
    
    if (H5IN)
    {
        h5attr.loc_id = out_sd_id;
        h5attr.rdata = NULL;
        strcpy(h5attr.name,"StructMetadata.0");
        if (H5OUT)
            H5Aiterate2(in_sd_id,H5_INDEX_CRT_ORDER,H5_ITER_INC,NULL,(H5A_operator2_t)H5_attr_write_hdf5,(void *)&h5attr);
        else
            H5Aiterate2(in_sd_id,H5_INDEX_CRT_ORDER,H5_ITER_INC,NULL,(H5A_operator2_t)H5_attr_write_hdf4,(void *)&h5attr);
    }
    else
    {
        fprintf(stderr, "\tWriting all global metadata to output file\n");
        if (SDfileinfo(in_sd_id, &nsds, &nattr) == FAIL)
            fprintf(stderr, "Cannot read file information for the input HDF file\n");
        else
        {
            fprintf(stderr, "Total %d attrs for global metadata.\n", nattr);
            for (attr_index=0; attr_index<nattr; attr_index++)
            {
                if (SDattrinfo(in_sd_id, attr_index, attr_name, &attr_type, &attr_cnt) == FAIL)
                    fprintf(stderr, "Cannot read information for attribute %d in input HDF file\n", attr_index);
                else
                {
                    /*	fprintf(stderr, " attr_name =%s\n", attr_name); */
                    if (strcmp(attr_name, "StructMetadata.0") != 0)
                    {
                        attr_size = DFKNTsize(attr_type);
                        attr_len = attr_size * attr_cnt + 1;
                        if ((meta_attr_buf = (void *)calloc(attr_len, attr_size)) == NULL)
                          fprintf(stderr, "Cannot allocate memory for metadata in write_metadata\n");
                        else
                        {
                            if (SDreadattr(in_sd_id, attr_index, meta_attr_buf) == FAIL)
                                fprintf(stderr, "Cannot read %s from input HDF file\n", attr_name);
                            else
                            {
                                if (H5OUT &&
                                    setH5Attr(attr_type,attr_name,attr_cnt,(VOIDP)meta_attr_buf, out_sd_id) == FAIL)
                                    fprintf(stderr, "Cannot write metadata %s to output HDF5 file\n", attr_name);
                                else if (!H5OUT && SDsetattr(out_sd_id, attr_name, attr_type, attr_cnt, meta_attr_buf) == FAIL)
                                    fprintf(stderr, "Cannot write metadata %s to output HDF file\n", attr_name);
                            }
                            free(meta_attr_buf);
                        }
                    }
                }
            } /* for (attr_index=0; . . .      */
        }
    }
}

void get_sds_dim_name(sds_t *sds_info, char **dim_names, char **short_dim_names)

     /* this routine takes input sds_info and output the dimension name and 
	short dimension names with ':'
	eg, As in file type MYD43B1, one dimension name is YDim:MOD_Grid_BRDF,
	the short dimension name is YDim */
{   
    int i, rank, p1;
    int32 cnt, nattrs, id, dt;
    
    /* HDF5 addition */
    ssize_t size;
    char *h5_label = NULL;
    
    rank = sds_info->rank;
    
    for (i=0; i<rank; i++)
    {
        if (sds_info->H5IN)
        {
            /* First get the size of the dimension i */
            size = H5DSget_label(sds_info->sds_id, i, NULL, 0);
            if (size < 0)
            {
                fprintf(stderr,"Cound not retrieve the size of HDF5 dimension %d for dataset: %s in get_sds_dim_name() sds_rw.c\n", i, sds_info->name);
                exit(EXIT_FAILURE);
            }
            else
            {
                h5_label = calloc(size+1, sizeof(char));
                if (H5DSget_label(sds_info->sds_id, i, h5_label, size+1) < 0)
                {
                    fprintf(stderr, "Could not Retrieve Dimension Info from Input\n");
                    exit(EXIT_FAILURE);
                }
                strcpy(dim_names[i], h5_label);
                strcpy(short_dim_names[i], h5_label);
                free(h5_label);
            }
        }
        else
        {
            id = SDgetdimid(sds_info->sds_id, i); 
            if (SDdiminfo(id, dim_names[i], &cnt, &dt, &nattrs) == FAIL)
            {
                  fprintf(stderr, "Error reading %dth dimension info in get_sds_dim_name. Program exit.\n", i);
                  exit(EXIT_FAILURE);
            }
        
              p1 = charpos(dim_names[i], ':', 0);
        
              if (p1 != -1)
            {
              strmid(dim_names[i], 0, p1, short_dim_names[i]);
            }
              else
            {
              strcpy(short_dim_names[i], dim_names[i]);
            }
        }
    }
}  

void display_sds_info_of_file(char* filename)
/* Print all the SDS names in the input file */
{
    sds_t sds_info;
    int32 msds, nattr, rank, dt, dim_size[4];
    int k, isds;
    char dim_str[MAX_STR_LEN];
    char name[MAX_SDS_NAME_LEN], sds_name[MAX_SDS_NAME_LEN];
    
    /* HDF5 additions */
    H5SDS h5_sds, *h5s = NULL;
    hid_t grp_id;
    hid_t h5dtype;
    hid_t dspace;
    herr_t ret;
    int HDF_ver = 0;

    /* MSZ HDF5 or HDF4 */
    HDF_ver = HDF_version_check(filename);

    if(HDF_ver == 4)
        sds_info.H5IN = 0;
    else if(HDF_ver == 5)
        sds_info.H5IN = 1;
    else
    {
        fprintf(stderr, "%s is neither HDF4 nor HDF5, exit...\n", filename);
        exit(0);
    }
    
    msds = 0;
    if (sds_info.H5IN)
    {
        if ((sds_info.sd_id = (int32)H5Fopen(filename, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
	{
	    sds_info.sd_id = -1;
            fprintf(stderr, "Cannot open the HDF file %s\n", filename);
	}
        else
        {
            h5_sds.srch_type = 4; /* for getting all sds names */
            h5_sds.nsds = 0;
            if ((h5_sds.h5sdsnames = (char **)Calloc2D(MAX_NUM_PARAM, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
            {
              fprintf(stderr, "Error allocating memory for in_sdsnames in alloc_memory_for_param()\n");
              exit(EXIT_FAILURE);
            }
           
            H5Ovisit(sds_info.sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5O_iterate_t)H5dataset_search, (void *)&h5_sds);
            msds = h5_sds.nsds;
	    /* Maosheng Zhao, 3/17/2020, added HDF version information for Python usage */
	    /*fprintf(stderr, "Valid SDS names, dimension and data type in file: %s\n", filename); */
	    fprintf(stderr, "Valid SDS names, dimension and data type in HDF5 file: %s\n", filename);
            for (isds=0; isds<msds; isds++)
            {
                if ((h5s = h5_dsetget(&sds_info, h5_sds.h5sdsnames[isds])) == NULL)
                {
                    fprintf(stderr,"failed in h5_dsetget for sdsname %s in display_sds_info_of_file\n", h5_sds.h5sdsnames[isds]);
                    exit(EXIT_FAILURE);
                }
                if (h5s->sds_found)
                {
                    if ((grp_id = H5Gopen2(sds_info.sd_id, h5s->abs_gnames, H5P_DEFAULT)) < 0)
                    {
                        fprintf(stderr,"H5: Failed to open H5 group %s in display_sds_info_of_file()\n",h5s->abs_gnames);
                        free(h5s);
                        exit(EXIT_FAILURE);
                    }
                    if ((sds_info.sds_id = (int32)H5Dopen2(grp_id,h5s->name,H5P_DEFAULT)) < 0)
                    {
                        fprintf(stderr,"H5: Failed to retrieve dataset %s in group %s get_sds_info()\n",h5s->name, h5s->abs_gnames);
                        free(h5s);
                        exit(EXIT_FAILURE);
                    }
                    h5dtype = H5Dget_type((hid_t)sds_info.sds_id);
                    dt = H5toH4type(h5dtype);
		    /* MSZ 1/26/2018, added the below dspace and calloc for h5s->h5_dims */
		    dspace = H5Dget_space((hid_t)sds_info.sds_id);
		    h5s->rank = H5Sget_simple_extent_ndims(dspace);
		    /* MSZ 4/6/2018, correct data type */
		    if(h5s->rank > 0 && dt == DFNT_CHAR8) dt = DFNT_INT8;
		    h5s->h5_dims = calloc(h5s->rank, sizeof(hsize_t));
		    ret = H5Sget_simple_extent_dims(dspace, h5s->h5_dims, NULL);
                    sprintf(sds_name, "%s (%llu", h5s->name, h5s->h5_dims[0]);
                    for (k=1; k<h5s->rank; k++)
                    {
                      sprintf(dim_str, " x %llu", h5s->h5_dims[k]);
                      strcat(sds_name, dim_str);
                    }

		    if(h5s->h5_dims) free(h5s->h5_dims);
                    switch(dt)
                    {
                      case  5: strcat(sds_name, ") FLOAT32"); break;
                      case  6: strcat(sds_name, ") FLOAT64"); break;
                      case 20: strcat(sds_name, ") INT8"); break;
                      case 21: strcat(sds_name, ") UINT8"); break;
                      case 22: strcat(sds_name, ") INT16"); break;
                      case 23: strcat(sds_name, ") UINT16"); break;
                      case 24: strcat(sds_name, ") INT32"); break;
                      case 25: strcat(sds_name, ") UINT32"); break;
                      default: strcat(sds_name, ") Unknown"); break;
                    }
                    fprintf(stderr, "\t%s\n", sds_name);
		    H5Tclose(h5dtype);
		    H5Sclose(dspace);
                    H5Dclose(sds_info.sds_id);
		    H5Gclose(grp_id);

                }
                else fprintf(stderr, "Error opening SDS with index %d\n", isds);

		if(h5s) free(h5s);
            }

	    if(h5_sds.h5sdsnames) Free2D((void **)h5_sds.h5sdsnames);
	    H5Fclose(sds_info.sd_id);
        }
    }
    else
    {
        if ((sds_info.sd_id = SDstart(filename, DFACC_READ)) == FAIL)
            fprintf(stderr, "Cannot open the HDF file %s\n", filename);
        else
        {
            if (SDfileinfo(sds_info.sd_id, &msds, &nattr) == FAIL)
                fprintf(stderr, "Error reading information for HDF file %s\n",filename);
            else
            {
	      /* Maosheng Zhao, 3/17/2020, added HDF version information for Python usage */
	      /* fprintf(stderr, "Valid SDS names, dimension and data type in file: %s\n", filename); */
              fprintf(stderr, "Valid SDS names, dimension and data type in HDF4 file: %s\n", filename);
              for (isds=0; isds<msds; isds++) {
                if ((sds_info.sds_id = SDselect(sds_info.sd_id, isds)) != FAIL) {
                  if (SDgetinfo(sds_info.sds_id, name, &rank, dim_size, &dt, &nattr) != FAIL) {
                sprintf(sds_name, "%s (%d", name, dim_size[0]);
                for (k=1; k<rank; k++) {
                  sprintf(dim_str, " x %d", dim_size[k]);
                  strcat(sds_name, dim_str);
                }
                switch(dt) {
                case  5: strcat(sds_name, ") FLOAT32"); break;
                case  6: strcat(sds_name, ") FLOAT64"); break;
                case 20: strcat(sds_name, ") INT8"); break;
                case 21: strcat(sds_name, ") UINT8"); break;
                case 22: strcat(sds_name, ") INT16"); break;
                case 23: strcat(sds_name, ") UINT16"); break;
                case 24: strcat(sds_name, ") INT32"); break;
                case 25: strcat(sds_name, ") UINT32"); break;
                default: strcat(sds_name, ") Unknown"); break;
                }
                fprintf(stderr, "\t%s\n", sds_name);
                SDendaccess(sds_info.sds_id);
                  }
                }
                else fprintf(stderr, "Error opening SDS with index %d\n", isds);
              }
            }
            SDend(sds_info.sd_id);
        }
    }
    if (msds == 0)
      fprintf(stderr, "Input files do not contain any valid SDS\n");
}

void print_sds_dim_size(sds_t *sds_info)
/* This routine prints out dimension name and dimension size of a SDS. */ 
{
  int i, rank;
  char **dim_names, **short_dim_names;
  
  rank = sds_info->rank;
  
  dim_names = (char **)Calloc2D(rank, MAX_NAME_LENGTH, sizeof(char));
  short_dim_names = (char **)Calloc2D(rank, MAX_DIM_NAME_LEN, sizeof(char));

  fprintf(stderr, "     %d dimensions : \n", rank);

  get_sds_dim_name(sds_info, dim_names, short_dim_names);

  for (i=0; i<rank; i++) 
    {
      fprintf(stderr, "          dimension %i    dimension name : %s    dim size : " LONG_INT_FMT "\n",
           i+1, short_dim_names[i], sds_info->dim_size[i]);    
    }
  Free2D((void **)dim_names);
  Free2D((void **)short_dim_names);
}


/* H5 Object handler H5 equiv of SDnametoindex*/
herr_t H5lnk_handler (hid_t loc_id, const char *name, const H5L_info_t *info, void *operator_data)
{
  /* MSZ 8/10/17 changed the previous method by not using H5Gget_num_objs() and H5Gget_objtype_byidx(),
 * which are deprecated from HDF5 1.8.0. */
  /* MSZ Note that here input loc_id must be file id */
  /* MSZ note here that input "name" is the full name, implies group name + sds short name, but group name has no root path
   * for H5Lvisit */
  H5SDS *sds = (H5SDS*)operator_data;
  H5O_info_t      infobuf;
  herr_t status;
  int i;
  char *full_abs_name = NULL;

  if((full_abs_name = (char*)calloc(strlen(name) + 2, sizeof(char))) == NULL)  /* leading "/" is an extra char for H5Lvisit */
  {
	printf("cannot calloc full_abs_name, exit...\n");
	exit(0);
  }

  if(name[0] == '.') /* because we starts from the root path and this applies to H5Ovisit */
	full_abs_name[0] = '/';
  else   /* for H5Lvisit from the root */
  {
	strcpy(full_abs_name, "/");
	strcat(full_abs_name, name);
  }

  /* get H5O_type_t for the object */
  status = H5Oget_info_by_name (loc_id, name, &infobuf, H5P_DEFAULT);

  if (infobuf.type == H5O_TYPE_DATASET)
  {
	char *abs_path = NULL, *var_name = NULL;

	/* split the full_abs_name into absolute path and name */
	HDF5_split_abs_name_into_path_name(full_abs_name, &abs_path, &var_name);

	/* we surelly found it !!! */
	if(!strcmp(var_name, ((H5SDS*)sds)->name))
	{
	    /* use accumulator so that we can detect if there are more than one dataset with the same var_name */
	    ((H5SDS*)sds)->sds_found += 1;

	    /* let's get the group path */
	    /* free it and replace with the updated */
	    for(i = 0; i < 255; i++)
		((H5SDS*)sds)->abs_gnames[i] = '\0';

	    /* update ((H5SDS*)sds)->name and ((H5SDS*)sds)->abs_gname */
	    strcpy(((H5SDS*)sds)->abs_gnames, abs_path);
	}

	/* free intermediate strings */
	if(abs_path) free(abs_path);
	if(var_name) free(var_name);
  }
  
  /* free full_abs_name */
  if(full_abs_name) free(full_abs_name);

  return 0;
}

/* H% object handler to be used to get count of particular object types */
herr_t H5dataset_search (hid_t loc_id, const char *name, const H5O_info_t *info,
            void *sds)
{
  hid_t sds_id;
  char *tmpname;

    if (info->type == H5O_TYPE_DATASET)
    {
        if (((H5SDS*)sds)->srch_type == 0) /* name only search */
        {
            if (strcasecmp(name, ((H5SDS *)sds)->name) == 0)
            {
              strcpy(((H5SDS *)sds)->name, name);
              return 0;
            }
        }
        else if (((H5SDS*)sds)->srch_type == 1) /* count only search */
        {
            ((H5SDS*)sds)->nsds++;
        }
        else if (((H5SDS*)sds)->srch_type == 2) /* non1D count only search */
        {
            if ((sds_id = H5Dopen2(loc_id,name,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"Cannot open SDS %s in H5dataset_search\n", name);
              return -1;
            }
            else
            {
              if ((((H5SDS*)sds)->h5space_id = H5Dget_space((hid_t)sds_id)) < 0)
              {
                fprintf(stderr,"H5: Failed to retrieve dataspace for sds %s in H5dataset_search()\n",((H5SDS*)sds)->name);
                return -1;
              }
              
              /* fprintf(stderr, "In get_sdsid abs_gnames = %s\n", sds->abs_gnames);
              fprintf(stderr, "In get_sdsid grp_id:sds_id:h5space_id = %d:%d:%d\n", grp_id,sds_info->sds_id, sds->h5space_id); */
              
              if ((((H5SDS*)sds)->rank = H5Sget_simple_extent_dims(((H5SDS*)sds)->h5space_id,NULL,NULL)) < 0)
              {
                fprintf(stderr,"H5: Can't get info from SDS \"%s\" .\n", name);
                return -1;
              }
              if (((H5SDS*)sds)->rank > 1) ((H5SDS*)sds)->nsds++;
              H5Dclose((hid_t)sds_id);
            }
        }
        else if (((H5SDS*)sds)->srch_type == 3) /* non1D sdsnames only search */
        {
            if ((sds_id = H5Dopen2(loc_id,name,H5P_DEFAULT)) < 0)
            {
              fprintf(stderr,"Cannot open SDS %s in H5dataset_search\n", name);
              return -1;
            }
            else
            {
              if ((((H5SDS*)sds)->h5space_id = H5Dget_space((hid_t)sds_id)) < 0)
              {
                fprintf(stderr,"H5: Failed to retrieve dataspace for sds %s in H5dataset_search()\n",name);
                return -1;
              }
              
              if ((((H5SDS*)sds)->rank = H5Sget_simple_extent_dims(((H5SDS*)sds)->h5space_id,NULL,NULL)) < 0)
              {
                fprintf(stderr,"H5: Can't get info from SDS \"%s\" .\n", name);
                return -1;
              }
              if (((H5SDS*)sds)->rank > 1)
              {
                tmpname = strrchr(name,'/');
                if (tmpname != NULL)
                  strcpy(((H5SDS*)sds)->h5sdsnames[((H5SDS*)sds)->nsds],++tmpname);
                else
                  strcpy(((H5SDS*)sds)->h5sdsnames[((H5SDS*)sds)->nsds],name);
                  
                /* fprintf (stderr, "name: count in H5dataset_count = %s:%d\n", ((H5SDS*)sds)->h5sdsnames[((H5SDS*)sds)->nsds],((H5SDS*)sds)->nsds);*/
                ((H5SDS*)sds)->nsds++;
              }
              H5Dclose((hid_t)sds_id);
            }
        }
        else if (((H5SDS*)sds)->srch_type == 4) /* all sdsnames only search */
        {
          if ((sds_id = H5Dopen2(loc_id,name,H5P_DEFAULT)) < 0)
          {
            fprintf(stderr,"Cannot open SDS %s in H5dataset_search\n", name);
            return -1;
          }
          else
          {
            tmpname = strrchr(name,'/');
            if (tmpname != NULL)
              strcpy(((H5SDS*)sds)->h5sdsnames[((H5SDS*)sds)->nsds],++tmpname);
            else
              strcpy(((H5SDS*)sds)->h5sdsnames[((H5SDS*)sds)->nsds],name);
                  
            /* fprintf (stderr, "name: count in H5dataset_count = %s:%d\n", ((H5SDS*)sds)->h5sdsnames[((H5SDS*)sds)->nsds],((H5SDS*)sds)->nsds);*/
            ((H5SDS*)sds)->nsds++;
            H5Dclose((hid_t)sds_id);
          }
        }
    }
  return 0;
}

int32 H5toH4type(hid_t h5type)
{
    if( H5Tequal( H5T_NATIVE_CHAR,        h5type ) )
        return DFNT_CHAR8;
    else if (H5Tget_class(h5type) == H5T_STRING)
        return DFNT_CHAR8;
    else if( H5Tequal( H5T_NATIVE_SCHAR,  h5type ) )
        return DFNT_INT8;
    else if( H5Tequal( H5T_NATIVE_UCHAR,  h5type ) ) 
	return DFNT_UINT8;
    else if( H5Tequal( H5T_NATIVE_SHORT,  h5type ) )
	return DFNT_INT16;
    else if( H5Tequal( H5T_NATIVE_USHORT, h5type ) ) 
	return DFNT_UINT16;
    else if( H5Tequal( H5T_NATIVE_INT,    h5type ) ) 
	return DFNT_INT32;      
    else if( H5Tequal( H5T_NATIVE_UINT,   h5type ) ) 
	return DFNT_UINT32;
    else if( H5Tequal( H5T_NATIVE_FLOAT,  h5type ) ) 
	return DFNT_FLOAT32;
    else if( H5Tequal( H5T_NATIVE_DOUBLE, h5type ) ) 
	return DFNT_FLOAT64;
    else
	return -1;
}


int convH4toH5 (const int32 h4type, hid_t* h5memtype, size_t* h4memsize,
				size_t* h4size, hid_t *h5type)
{

  switch (h4type)
  {
  case DFNT_CHAR8:
    *h4size = 1;
    *h4memsize = sizeof(int8);
    /* assume DFNT_CHAR8 C type character. */
    *h5memtype = H5T_STRING;
    *h5type =  H5T_STRING;
    break;

  case DFNT_UCHAR8:

    *h4size = 1;
    *h4memsize = sizeof(int8);
    *h5memtype = H5T_STRING;
    *h5type = H5T_STRING;
    break;

  case DFNT_INT8:

    *h4size = 1;
    *h5type = H5T_STD_I8LE;
    *h4memsize = sizeof(int8);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype = H5T_NATIVE_SCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
       fprintf(stderr, "WARNING: cannot convert signed INT8");
      return FAIL;
    }
    break;

  case DFNT_UINT8:

    *h4size =1;
    *h5type = H5T_STD_U8LE;
    *h4memsize = sizeof(int8);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
       fprintf(stderr, "WARNING: cannot convert unsigned INT8");
      return FAIL;
    }
    break;

  case DFNT_INT16:
    *h4size = 2;
    *h5type = H5T_STD_I16LE;
    *h4memsize = sizeof(int16);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_CHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      fprintf(stderr, "WARNING: cannot convert signed int16");
      return FAIL;
    }
    break;

  case DFNT_UINT16:
    *h4size = 2;
    *h5type = H5T_STD_U16LE;
    *h4memsize = sizeof(int16);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
      fprintf(stderr, "WARNING: cannot convert unsigned int16");
      return FAIL;
    }
    break;

  case DFNT_INT32:
    *h4size = 4;
    *h5type = H5T_STD_I32LE;
    *h4memsize = sizeof(int32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_CHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      fprintf(stderr, "WARNING: cannot convert signed int32");
      return FAIL;
    }
    break;

  case DFNT_UINT32:
    *h4size = 4;
    *h5type = H5T_STD_U32LE;
    *h4memsize = sizeof(int32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
       fprintf(stderr, "WARNING: cannot convert unsigned int32");
      return FAIL;
    }
    break;

  case DFNT_INT64:
    *h4size = 8;
    *h5type = H5T_STD_I64LE;
    *h4memsize = sizeof(long);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_CHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LLONG))
      *h5memtype = H5T_NATIVE_LLONG;
    else {
      fprintf(stderr, "WARNING: cannot convert signed int64");
      return FAIL;
    }
    break;

  case DFNT_UINT64:
    *h4size = 8;
    *h5type = H5T_STD_U64LE;
    *h4memsize = sizeof(long);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LLONG))
      *h5memtype = H5T_NATIVE_ULLONG;
    else {
       fprintf(stderr, "WARNING: cannot convert unsigned int64");
      return FAIL;
    }
    break;

  case DFNT_FLOAT32:
    *h4size =4;
    *h5type = H5T_IEEE_F32LE;
    *h4memsize = sizeof(float32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_FLOAT))
      *h5memtype = H5T_NATIVE_FLOAT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_DOUBLE))
      *h5memtype = H5T_NATIVE_DOUBLE;
    else {
      fprintf(stderr, "WARNING: cannot convert float32");
      return FAIL;
    }
    break;
  
  case DFNT_FLOAT64:
    *h4size = 8;
    *h5type = H5T_IEEE_F64LE;
    *h4memsize = sizeof(float64);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_FLOAT))
      *h5memtype = H5T_NATIVE_FLOAT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_DOUBLE))
      *h5memtype = H5T_NATIVE_DOUBLE;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LDOUBLE))
      *h5memtype = H5T_NATIVE_LDOUBLE;
    else {
      fprintf(stderr, "WARNING: cannot convert float64");
      return FAIL;
    }
    break;

    default: {
       fprintf(stderr, "cannot find the corresponding datatype in HDF5\n");
    
    return FAIL;
    }
  }
  return SUCCEED;
}

int setH5Attr(int32 att_type, char *att_name, int nval, void *mbuf, hid_t id)
{
  size_t    sh4_amemsize;/* size of attribute data type in memory. */
  size_t    sh4_asize;  /* size of attribute data type in disk. */
  hid_t     sh5a_sid; /* HDF5 attribute space id */
  hid_t     sh5a_id; /* HDF5 attribute id */
  hid_t     h5_atype; /* HDF5 attribute data type in disk */
  hid_t     sh5_amemtype; /* HDF5 attribute data type in memory */
  hid_t     sh5str_type; /* HDF5 attribute data type in disk when the data type is string. */
  hid_t     sh5str_memtype; /* HDF5 attribute data type in memory when the data type is string. */
  hsize_t   *he5_dims = NULL; /* For numeric array attributes like valid_range */
  
  /* convert the attr type to H5 equivalent */
  if (convH4toH5(att_type, &sh5_amemtype,&sh4_amemsize,
		&sh4_asize,&h5_atype) == FAIL)
  {
	fprintf(stderr, "WARNING: cannot convert SDS attribute data type from H4 to H5");
	return -1;
  }
  
  if (h5_atype == H5T_STRING)
  {
	/* Create attribute space */
    sh5a_sid = H5Screate(H5S_SCALAR);
    if (sh5a_sid < 0)
    {
      fprintf(stderr, "WARNING: cannot create HDF5 attribute space\n");
      return -1;
    }
    
    if ((sh5str_type = mkstr((size_t)nval*sh4_asize,
			 H5T_STR_NULLTERM))<0)
    {
      fprintf(stderr, "WARNING: cannot make HDF5 string type\n");
      H5Sclose(sh5a_sid);
      return -1;
    }
	
    if ((sh5str_memtype = mkstr((size_t)nval*sh4_amemsize,
			 H5T_STR_NULLTERM))<0)
    {
      fprintf(stderr, "WARNING: cannot make HDF5 memory string type\n");
      H5Sclose(sh5a_sid);
      return -1;
    }
	
    if ((sh5a_id = H5Acreate2(id,att_name,sh5str_type,
			sh5a_sid,H5P_DEFAULT,H5P_DEFAULT)) < 0)
    {
      fprintf(stderr, "WARNING: cannot create H5 attribute H5Acreate2\n");
      H5Sclose(sh5a_sid);
      return -1;
    }
	
    /* Finally write attribute */
    if (H5Awrite(sh5a_id,sh5str_memtype,mbuf) < 0)
    {
      fprintf(stderr, "WARNING: fail to write attribute %s (H5Awrite)\n", att_name);
      return -1;
    }
	
    H5Sclose(sh5a_sid);
    H5Aclose(sh5a_id);
  }
  else
  { 
	if (nval == 1)
	{
	  /* Create attribute space */
	  sh5a_sid = H5Screate(H5S_SCALAR);
	  if (sh5a_sid < 0)
	  {
		fprintf(stderr, "WARNING: cannot create HDF5 attribute space\n");
		return -1;
	  }
	}
	else
	{   
	  /* Create simple attribute space */
      
          he5_dims= (hsize_t *)calloc(2, sizeof(hsize_t)); /* treat all other arrays as 1 row * nval cols = 2D arrays */
          he5_dims[0] = 1;
          he5_dims[1] = nval;
        
	  sh5a_sid = H5Screate_simple(2, he5_dims, NULL);
	  if (sh5a_sid < 0)
	  {
		fprintf(stderr, "WARNING: cannot create HDF5 attribute space\n");
		return -1;
	  }
	}
	
	if ((sh5a_id = H5Acreate2(id,att_name,h5_atype,
			sh5a_sid,H5P_DEFAULT,H5P_DEFAULT)) < 0)
	{
	  fprintf(stderr, "WARNING: cannot create H5 attribute H5Acreate2\n");
	  H5Sclose(sh5a_sid);
	  return -1;
	}
	
	/* Finally write attribute */
	if (H5Awrite(sh5a_id,sh5_amemtype,mbuf) < 0)
	{
	  fprintf(stderr, "WARNING: writting attribute (H5Awrite)\n");
	  return -1;
	}
	
	H5Sclose(sh5a_sid);
	H5Aclose(sh5a_id);

        if(he5_dims) free(he5_dims);
  }
  
  return 0;
}

hid_t mkstr( size_t size, H5T_str_t pad)
{

  hid_t type; /* HDF5 data type id. */

  if((type=H5Tcopy(H5T_C_S1))<0) {
    fprintf(stderr, "WARNING: cannot make H5 String type\n");
    return -1;
  }
  if((H5Tset_size(type,size))<0) {
    fprintf(stderr, "WARNING: cannot set size for H5 String\n");
    return -1;
  }
  if((H5Tset_strpad(type,pad))<0){
    fprintf(stderr, "WARNING: cannot set H5 String Pad\n");
    return -1;
  }
  return type;
}

herr_t H5_attr_read(hid_t obj_id, const char *attr_name, const H5A_info_t *ainfo,
    void *op_data)
{
    hid_t attr_id, attr_type, attr_space, memtype;
    H5_ATTR *ioparam;
    hsize_t *sdim = NULL, *he5_dims = NULL; 
    herr_t ret;
    int i;
    size_t attr_size, attr_cnt, attr_len;
    int rank;
    htri_t vlen;
    
    hid_t     sh5a_sid; /* HDF5 attribute space id */
    hid_t     sh5a_id; /* HDF5 attribute id */
    H5T_cset_t cset;   /* MSZ 10/25/2018 UTF8 */
  
    ioparam = (H5_ATTR *)op_data;
    
    if (ioparam->name[0] == '\0' || !strcasecmp(ioparam->name, attr_name))
    {
        attr_id = H5Aopen(obj_id,attr_name,H5P_DEFAULT);
        /* Get attribute datatype, dataspace, rank, and dimensions.  */
        attr_type  = H5Aget_type(attr_id);
        attr_space = H5Aget_space(attr_id);
        rank = H5Sget_simple_extent_ndims(attr_space);
        
        if (rank > 1)
        {
            sdim = (hsize_t *)calloc(rank, sizeof(hsize_t));
            if (sdim == NULL)
            {
              fprintf(stderr, "Failed to allocate memory for sdim in H5_attr_read\n");
              return -1;
            }
            ret = H5Sget_simple_extent_dims(attr_space, sdim, NULL);
        }
        
        attr_cnt = H5Sget_simple_extent_npoints(attr_space);
        attr_size = H5Tget_size (attr_type);
        
        /* fprintf(stderr, "In H5_attr_read attr_name = %s\n",attr_name); */
        
        if ((H5T_INTEGER == H5Tget_class(attr_type)) ||
            (H5T_FLOAT == H5Tget_class(attr_type)))
        {
            /* fprintf(stderr, "In H5_attr_read rank:attr_cnt:attr_size = %d:%d:%d\n", rank, (int)attr_cnt, (int)attr_size); */
            
            /* if (rank > 1)
              fprintf(stderr, "In H5_attr_read H5T_INTEGER sdim[0]:sdim[1], rank = %d:%d:%d\n", sdim[0], sdim[1], rank); */
            if (rank == 2)
              ioparam->rdata = (void **) Calloc2D (sdim[0], sdim[1], attr_size);
            else
              ioparam->rdata = (void **) Calloc2D (1, attr_cnt, attr_size);
              
            memtype = H5Tget_native_type(attr_type,H5T_DIR_DEFAULT);
            ret  = H5Aread(attr_id, memtype, ioparam->rdata[0]);
        }
        else if (H5T_STRING == H5Tget_class (attr_type))
        {
            if ((vlen = H5Tis_variable_str(attr_type)) == 1)
            {
              memtype = H5Tcopy (H5T_C_S1);

	      /* Maosheng Zhao, 10/25/2018 get and set character systrem of either ASCII or UTF8*/
	      cset = H5Tget_cset(attr_type);
	      H5Tset_cset(memtype, cset);
	
              ret = H5Tset_size (memtype, H5T_VARIABLE);
              ioparam->rdata = (void **) calloc (attr_cnt, sizeof (char *));
              ret = H5Aread (attr_id, memtype, ioparam->rdata[0]);
            }
            else
            {
              memtype = H5Tget_native_type(attr_type,H5T_DIR_ASCEND);
              if (attr_cnt > 1)
              {
                attr_len = 0;
                ioparam->rdata = (void **)Calloc2D (attr_cnt,attr_size,sizeof(char));
                ret  = H5Aread(attr_id, attr_type, ioparam->rdata[0]);
              }
              else
              {
                attr_len = attr_size + 1;
                /* fprintf(stderr, "attr_size = %d, attr_len = %d\n", attr_size, attr_len); */
                if ((ioparam->rdata = (void **)Calloc2D(1,attr_len,sizeof(char))) == NULL)
                {
                  fprintf(stderr, "Cannot allocate memory for attr_buf in H5_attr_read()\n");
                  return -1;
                }
                else
                {
                  ret  = H5Aread(attr_id, memtype, ioparam->rdata[0]);
                  /* fprintf(stderr, "GOT %s\n", ioparam->rdata[0]); */
                }
              }
              /* fprintf(stderr, "In H5_attr_read rank:attr_cnt:attr_size = %d:%d:%d\n", rank, (int)attr_cnt, (int)attr_size);
              fprintf(stderr, "attr_len = %d\n", attr_len); */
            }
        }
        
        ret = H5Tclose(memtype);
        if (sdim) free(sdim);
        ret = H5Tclose(attr_type);
        ret = H5Sclose(attr_space);
        ret = H5Aclose(attr_id);
    }

    return 0;
}

herr_t H5_attr_write_hdf5(hid_t obj_id, const char *attr_name, const H5A_info_t *ainfo,
    void *op_data)
{
    hid_t attr_id, attr_type, attr_space, memtype;
    H5_ATTR *ioparam;
    hsize_t *sdim = NULL, *he5_dims = NULL; 
    herr_t ret;
    int i;
    size_t attr_size, attr_cnt, attr_len;
    int rank;
    htri_t vlen;
    
    hid_t     sh5a_sid; /* HDF5 attribute space id */
    hid_t     sh5a_id; /* HDF5 attribute id */

    H5T_cset_t cset;   /* MSZ 10/25/2018 UTF8 */
  
    ioparam = (H5_ATTR *)op_data;
    
    if (ioparam->name[0] == '\0' || !strcasecmp(ioparam->name, attr_name))
    {
        attr_id = H5Aopen(obj_id,attr_name,H5P_DEFAULT);
        /* Get attribute datatype, dataspace, rank, and dimensions.  */
        attr_type  = H5Aget_type(attr_id);
        attr_space = H5Aget_space(attr_id);
        rank = H5Sget_simple_extent_ndims(attr_space);
        
        if (rank > 1)
        {
            sdim = (hsize_t *)calloc(rank, sizeof(hsize_t));
            if (sdim == NULL)
            {
              fprintf(stderr, "Failed to allocate memory for sdim in H5_attr_write_hdf5\n");
              return -1;
            }
            ret = H5Sget_simple_extent_dims(attr_space, sdim, NULL);
        }
        
        attr_cnt = H5Sget_simple_extent_npoints(attr_space);
        attr_size = H5Tget_size (attr_type);
        
        /* fprintf(stderr, "In H5_attr_write_hdf5 attr_name = %s\n",attr_name); */
        
        if ((H5T_INTEGER == H5Tget_class(attr_type)) ||
            (H5T_FLOAT == H5Tget_class(attr_type)))
        {
            /* fprintf(stderr, "In H5_attr_write_hdf5 rank:attr_cnt:attr_size = %d:%d:%d\n", rank, (int)attr_cnt, (int)attr_size); */
            
            /* if (rank > 1)
              fprintf(stderr, "In H5_attr_write_hdf5 H5T_INTEGER sdim[0]:sdim[1], rank = %d:%d:%d\n", sdim[0], sdim[1], rank); */
            if (rank == 2)
              ioparam->rdata = (void **) Calloc2D (sdim[0], sdim[1], attr_size);
            else
              ioparam->rdata = (void **) Calloc2D (1, attr_cnt, attr_size);
              
            memtype = H5Tget_native_type(attr_type,H5T_DIR_DEFAULT);
            ret  = H5Aread(attr_id, memtype, ioparam->rdata[0]);
            
            if (attr_cnt == 1)
            {
              /* Create attribute space */
              sh5a_sid = H5Screate(H5S_SCALAR);
              if (sh5a_sid < 0)
              {
                fprintf(stderr, "WARNING: cannot create HDF5 attribute space");
                return -1;
              }
            }
            else
            {   
              /* Create simple attribute space */
              he5_dims= (hsize_t *)calloc(2, sizeof(hsize_t)); /* treat all other arrays as 1 row * nval cols = 2D arrays */
              he5_dims[0] = 1;
              he5_dims[1] = attr_cnt;
              sh5a_sid = H5Screate_simple(2, he5_dims, NULL);
              if (sh5a_sid < 0)
              {
                fprintf(stderr, "WARNING: cannot create HDF5 attribute space");
                return -1;
              }
            }
            
            if (!H5Aexists(ioparam->loc_id,attr_name))
            {
                if ((sh5a_id = H5Acreate2(ioparam->loc_id,attr_name,attr_type,
                        sh5a_sid,H5P_DEFAULT,H5P_DEFAULT)) < 0)
                {
                  fprintf(stderr, "WARNING: cannot create H5 attribute for %s in H5Acreate2\n", attr_name);
                  H5Sclose(sh5a_sid);
                  return -1;
                }
                
                /* fprintf(stderr, "rdata = %d\n", ((int **)ioparam->rdata)[0]); */
                /* Finally write attribute */
                if (H5Awrite(sh5a_id,memtype,ioparam->rdata[0]) < 0)
                {
                  fprintf(stderr, "Cannot write attribute %s in H5_attr_write_hdf5()\n", attr_name);
                  return -1;
                }
            }
            
            if (attr_cnt > 1 && he5_dims) free(he5_dims);
            H5Sclose(sh5a_sid);
            H5Aclose(sh5a_id);
        }
        else if (H5T_STRING == H5Tget_class (attr_type))
        {
            if ((vlen = H5Tis_variable_str(attr_type)) == 1)
            {
              memtype = H5Tcopy (H5T_C_S1);

	      /* Maosheng Zhao, 10/25/2018 get and set character systrem of either ASCII or UTF8*/
	      cset = H5Tget_cset(attr_type);
	      H5Tset_cset(memtype, cset);

              ret = H5Tset_size (memtype, H5T_VARIABLE);
              ioparam->rdata = (void **) calloc (attr_cnt, sizeof (char *));
              ret = H5Aread (attr_id, memtype, ioparam->rdata[0]);
            }
            else
            {
              memtype = H5Tget_native_type(attr_type,H5T_DIR_ASCEND);
              if (attr_cnt > 1)
              {
                attr_len = 0;
                ioparam->rdata = (void **)Calloc2D (attr_cnt,attr_size,sizeof(char));
                ret  = H5Aread(attr_id, attr_type, ioparam->rdata[0]);
              }
              else
              {
                attr_len = attr_size + 1;
                /* fprintf(stderr, "attr_size = %d, attr_len = %d\n", attr_size, attr_len); */
                if ((ioparam->rdata = (void **)Calloc2D(1,attr_len,sizeof(char))) == NULL)
                {
                  fprintf(stderr, "Cannot allocate memory for attr_buf in H5_attr_write_hdf5()\n");
                  return -1;
                }
                else
                {
                  ret  = H5Aread(attr_id, memtype, ioparam->rdata[0]);
                  /* fprintf(stderr, "GOT %s\n", ioparam->rdata[0]); */
                }
              }
              /* fprintf(stderr, "In H5_attr_write_hdf5 rank:attr_cnt:attr_size = %d:%d:%d\n", rank, (int)attr_cnt, (int)attr_size);
              fprintf(stderr, "attr_len = %d\n", attr_len); */
            }
            /* Create attribute space */
            sh5a_sid = H5Screate(H5S_SCALAR);
            if (sh5a_sid < 0)
            {
              fprintf(stderr, "WARNING: cannot create HDF5 attribute space");
              return -1;
            }
            
            if (!H5Aexists(ioparam->loc_id,attr_name))
            {
                if ((sh5a_id = H5Acreate2(ioparam->loc_id,attr_name,attr_type,sh5a_sid,H5P_DEFAULT,H5P_DEFAULT)) < 0)
                {
                  fprintf(stderr, "WARNING: cannot create H5 attribute for %s in H5Acreate2\n", attr_name);
                  H5Sclose(sh5a_sid);
                  return -1;
                }
                /* Finally write attribute */
                if (H5Awrite(sh5a_id,memtype,ioparam->rdata[0]) < 0)
                {
                  fprintf(stderr, "Cannot write attribute %s in H5_attr_write_hdf5()\n", attr_name);
                  return -1;
                }
            }
            
            H5Sclose(sh5a_sid);
            H5Aclose(sh5a_id);
        }
        
        ret = H5Tclose(memtype);
        if (sdim) free(sdim);
        ret = H5Tclose(attr_type);
        ret = H5Sclose(attr_space);
        ret = H5Aclose(attr_id);
        if (ioparam->rdata) Free2D((void **)ioparam->rdata);
        ioparam->rdata = NULL;
    }

    return 0;
}


herr_t H5_attr_write_hdf4(hid_t obj_id, const char *attr_name, const H5A_info_t *ainfo,
    void *op_data)
{
    hid_t attr_id, attr_type, attr_space, memtype;
    H5_ATTR *ioparam;
    hsize_t *sdim = NULL, *he5_dims = NULL; 
    herr_t ret;
    int i;
    size_t attr_size, attr_cnt, attr_len;
    int rank;
    htri_t vlen;
    
    hid_t     sh5a_sid; /* HDF5 attribute space id */
    hid_t     sh5a_id; /* HDF5 attribute id */
  
    ioparam = (H5_ATTR *)op_data;
    
    if (ioparam->name[0] == '\0' || !strcasecmp(ioparam->name, attr_name))
    {
        attr_id = H5Aopen(obj_id,attr_name,H5P_DEFAULT);
        if (attr_id >= 0)
        {
            /* Get attribute datatype, dataspace, rank, and dimensions.  */
            attr_type  = H5Aget_type(attr_id);
            attr_space = H5Aget_space(attr_id);
            rank = H5Sget_simple_extent_ndims(attr_space);
            
            if (rank > 1)
            {
                sdim = (hsize_t *)calloc(rank, sizeof(hsize_t));
                if (sdim == NULL)
                {
                  fprintf(stderr, "Failed to allocate memory for sdim in H5_attr_write_hdf4\n");
                  return -1;
                }
                ret = H5Sget_simple_extent_dims(attr_space, sdim, NULL);
            }
            
            attr_cnt = H5Sget_simple_extent_npoints(attr_space);
            attr_size = H5Tget_size (attr_type);
            
            /* fprintf(stderr, "In H5_attr_write_hdf4 attr_name = %s\n",attr_name); */
            /* fprintf(stderr, "attr_name = %s, attr_cnt = %d, attr_h4type = %d\n", attr_name, attr_cnt, H5toH4type(attr_type)); */
            
            if ((H5T_INTEGER == H5Tget_class(attr_type)) ||
                (H5T_FLOAT == H5Tget_class(attr_type)))
            {
                /* fprintf(stderr, "In H5_attr_write_hdf4 rank:attr_cnt:attr_size = %d:%d:%d\n", rank, (int)attr_cnt, (int)attr_size); */
                
                /* if (rank > 1)
                  fprintf(stderr, "In H5_attr_write_hdf4 H5T_INTEGER sdim[0]:sdim[1], rank = %d:%d:%d\n", sdim[0], sdim[1], rank); */
                if (rank == 2)
                  ioparam->rdata = (void **) Calloc2D (sdim[0], sdim[1], attr_size);
                else
                  ioparam->rdata = (void **) Calloc2D (1, attr_cnt, attr_size);
                  
                memtype = H5Tget_native_type(attr_type,H5T_DIR_DEFAULT);
                ret  = H5Aread(attr_id, memtype, ioparam->rdata[0]);
                if ( SDsetattr(ioparam->loc_id, attr_name, H5toH4type(attr_type), attr_cnt, ioparam->rdata[0]) == FAIL)
                {
                    fprintf(stderr, "WARNING: cannot copy HDF5 attribute %s (H5_attr_write_hdf4)\n", attr_name);
                    return -1; 
                }
            }
            else if (H5T_STRING == H5Tget_class (attr_type))
            {
                memtype = H5Tget_native_type(attr_type,H5T_DIR_ASCEND);
                if (attr_cnt > 1)
                {
                  attr_len = 0;
                  ioparam->rdata = (void **)Calloc2D (attr_cnt,attr_size,sizeof(char));
                  ret  = H5Aread(attr_id, attr_type, ioparam->rdata[0]);
                }
                else
                {
                  attr_len = attr_size + 1;
                  /* fprintf(stderr, "attr_size = %d, attr_len = %d\n", attr_size, attr_len); */
                  if ((ioparam->rdata = (void **)Calloc2D(1,attr_len,sizeof(char))) == NULL)
                  {
                    fprintf(stderr, "Cannot allocate memory for attr_buf in H5_attr_write_hdf4()\n");
                    return -1;
                  }
                  else
                  {
                    ret  = H5Aread(attr_id, memtype, ioparam->rdata[0]);
                    /* fprintf(stderr, "GOT %s\n", ioparam->rdata[0]); */
                  }
                }
                /* fprintf(stderr, "In H5_attr_write_hdf4 rank:attr_cnt:attr_size = %d:%d:%d\n", rank, (int)attr_cnt, (int)attr_size);
                fprintf(stderr, "attr_len = %d\n", attr_len); */
                if ( SDsetattr(ioparam->loc_id, attr_name, H5toH4type(attr_type), attr_len, ioparam->rdata[0]) == FAIL)
                {
                    fprintf(stderr, "WARNING: cannot copy HDF5 attribute %s (H5_attr_write_hdf4)\n", attr_name);
                    return -1; 
                }
            }
            else if (H5T_VLEN == H5Tget_class (attr_type))
                return 0; /* Ignore variable length strings for now */
            
            
            ret = H5Tclose(memtype);
            if (sdim) free(sdim);
            ret = H5Tclose(attr_type);
            ret = H5Sclose(attr_space);
            ret = H5Aclose(attr_id);
            if (ioparam->rdata) Free2D((void **)ioparam->rdata);
            ioparam->rdata = NULL;
        }
    }

    return 0;
}

H5SDS *h5_dsetget(sds_t *sds_info, char *sdsname)
{
    /* MSZ 8/16/17.  Previously, the first input is int32 sd_id and I changed it sds_t *sds_info,
 * because this will make it easy to modify the code for exceptional cases when there are more than one datasets
 * having the same sdsname but under different groups */

    H5SDS *h5_sds = NULL;
    h5_sds = (H5SDS*)calloc(1,sizeof(H5SDS));
    herr_t status;
    
    if (h5_sds != NULL)
    {
        h5_sds->sds_found = 0;

	if(strlen(sdsname) > 1)
          strcpy(h5_sds->name,sdsname);
	else
	  return NULL;

        /*fprintf(stderr, "begin search for SDS %s\n", sdsname);*/
	/* MSZ 9/12/17. It is bettter to use H5Lvisit than H5Ovisit because H5Ovisit cannot find data links whereas H5Lvisit can. */
	/* To understand this, just test HDF5 example code h5ex_g_visit.c for h5ex_g_visit.h5 */
	/* start from the root group and root group is not accounted for in H5Lvisit() */
        /*H5Ovisit((hid_t)sds_info->sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5O_iterate_t)H5lnk_handler, (void*)h5_sds);*/
	status = H5Lvisit((hid_t)sds_info->sd_id, H5_INDEX_NAME, H5_ITER_NATIVE, (H5L_iterate_t)H5lnk_handler, (void*)h5_sds);
	/*fprintf(stderr, "end search for SDS %s\n", sdsname);*/
	
	if(status < 0)
	{
	   fprintf(stderr, "failure of H5Lvisit to search \"%s\"!!!\n", h5_sds->name);
	   return NULL;
	}

	/* MSZ 8/14/17, add this to make error info clear */
	if (((H5SDS*)h5_sds)->sds_found == 0)  /* sdsname doesn't exist */
	{
	    fprintf(stderr, "cannot find the sdsname of \"%s\", please use hdfdump to check the sds names. For HDF5, use the short name not full name.\n", sdsname);
	    return NULL;
	}
	else if(((H5SDS*)h5_sds)->sds_found > 1)   /* more than one sdsname found, more coding work is needed :( */
	{
	    fprintf(stderr, "Uh oh, there are more than one dataset for sdsname of %s, you need to add an if condition in callback function H5lnk_handler and here to make it a special case, exit...\n", sdsname);
	    exit(0);
	}
	else  /* exactly one sdsname found, great! */
	{
            return h5_sds;
	}
    }
    else
    {
        fprintf(stderr,"Failed to allocate memory for H5SDS structure in h5_dsetinfo\n");
        return NULL;
    }
}

int h5_dsetinfo(H5SDS *h5_sds, sds_t *sds_info)
{
    int i;
    
    if ((h5_sds->h5space_id = H5Dget_space((hid_t)sds_info->sds_id)) < 0)
    {
      fprintf(stderr,"H5: Failed to retrieve dataspace for sds %s in h5_dsetinfo()\n",h5_sds->name);
      return -1;
    }
    
    
    if ((h5_sds->rank = H5Sget_simple_extent_dims(h5_sds->h5space_id,NULL,NULL)) < 0)
    {
      fprintf(stderr,"H5: Can't get rank info for SDS \"%s\" in h5_dsetinfo\n", h5_sds->name);
      return -1;
    }
    h5_sds->h5_dims = calloc(h5_sds->rank, sizeof(hsize_t));
    if (H5Sget_simple_extent_dims(h5_sds->h5space_id,h5_sds->h5_dims,NULL) < 0)
    {
      fprintf(stderr,"H5: Can't get size info for SDS \"%s\" in h5_dsetinfo.\n", h5_sds->name);
      return -1;
    }		
    sds_info->h5dtype = H5Dget_type((hid_t)sds_info->sds_id);
    sds_info->h5dmtype = H5Tget_native_type(sds_info->h5dtype,H5T_DIR_DEFAULT);
    sds_info->h5space_id = h5_sds->h5space_id;
    sds_info->rank = h5_sds->rank;
    for (i = 0; i < sds_info->rank; i++) sds_info->dim_size[i] = h5_sds->h5_dims[i];
    sds_info->data_type = H5toH4type(sds_info->h5dtype);
    /* Maosheng Zhao SSAI, 4/6/2018, correct data type issue for DFNT_CHAR8 */
    if (sds_info->rank > 0 && sds_info->data_type == DFNT_CHAR8) sds_info->data_type = DFNT_INT8;
    sds_info->data_size = DFKNTsize(sds_info->data_type);
    
    return 0;
}

static int HDF4_scan_SDSs_info(HDF4_SDSs_INFO *HDF4_SDSs_info)
{
	/* written: Maosheng Zhao, SSAI
	 * purpose: scan HDF4 and get all SDSs name, data type, rank, dim_sizes and n_attrs
 	 * Date: July 3rd & 5th, 2017
 	 */ 	
	size_t max_name_len = 1024;
	size_t name_len;
	intn status;
	int32 index, sds_id, n_file_attrs, dim_sizes[MAX_VAR_DIMS];
	char *name = NULL;
	int i;

	/* calloc for name, an intermediate commonly used name string */ 
	if( ( name = (char*)calloc((size_t)max_name_len, sizeof(char)) ) == NULL )
	{
	    fprintf(stderr, "canot calloc name, exit...\n");   
	    exit(0);
	}	

	/* how many SDS in HDF4 ? */
	status = SDfileinfo(HDF4_SDSs_info->sd_id, &HDF4_SDSs_info->n_SDSs, &n_file_attrs);
	if (status == -1)
	{
	    fprintf(stderr, "can not SDfileinfo in SDS_name_list(), exit...\n");
	    exit(0);
	}

	/* get info for each SDS */
	if(HDF4_SDSs_info->n_SDSs <= 0)
	{
	    fprintf(stderr, "There is no SDS in the file in SDS_name_list(), exit...\n");
	    exit(0);
	}
	else
	{
	    /* calloc memory for each element */
	    if( ( HDF4_SDSs_info->data_type = (int32*)calloc((size_t)HDF4_SDSs_info->n_SDSs, sizeof(int32)) ) == NULL )
	    {
		fprintf(stderr, "canot calloc HDF4_SDSs_info->data_type for index %d, exit...\n", index);
		exit(0);
	    }
	    if( ( HDF4_SDSs_info->rank = (int32*)calloc((size_t)HDF4_SDSs_info->n_SDSs, sizeof(int32)) ) == NULL )
	    {
		fprintf(stderr, "canot calloc HDF4_SDSs_info->rank for index %d, exit...\n", index);
		exit(0);
	    }
	    if( ( HDF4_SDSs_info->n_attrs = (int32*)calloc((size_t)HDF4_SDSs_info->n_SDSs, sizeof(int32)) ) == NULL )
	    {
		fprintf(stderr, "canot calloc HDF4_SDSs_info->n_attrs for index %d, exit...\n", index);
		exit(0);
	    }
	    if( ( HDF4_SDSs_info->SDS_names = (char**)calloc((size_t)HDF4_SDSs_info->n_SDSs, sizeof(char*)) ) == NULL )
	    {
		fprintf(stderr, "canot calloc HDF4_SDSs_info->SDS_names, exit...\n");
		exit(0);
	    }
	    if( ( HDF4_SDSs_info->dim_sizes = (int32**)calloc((size_t)HDF4_SDSs_info->n_SDSs, sizeof(int32*)) ) == NULL )
	    {
		fprintf(stderr, "canot calloc HDF4_SDSs_info->dim_sizes, exit...\n");
		exit(0);
	    }

	    /* obtain values and fill the calloced memories for each element */
	    for (index = 0; index < HDF4_SDSs_info->n_SDSs; index++)
	    {
		sds_id = SDselect (HDF4_SDSs_info->sd_id, index);
		status = SDgetinfo (sds_id, name, &HDF4_SDSs_info->rank[index], dim_sizes, &HDF4_SDSs_info->data_type[index], &HDF4_SDSs_info->n_attrs[index]);
		if (sds_id== -1 || status == -1)
		{
		    fprintf(stderr, "can not SDselect or SDgetinfo in SDS_name_list(), exit...\n");
		    exit(0);
		}
	
		/* put each SDS name into SDS_names */	
		name_len = strlen(name) + 1;
		if(name_len <= 0)
		{
		    fprintf(stderr, "name_len is not greater than 0, exit...\n");
		    exit(0);
		}

		if( ( HDF4_SDSs_info->SDS_names[index] = (char*)calloc((size_t)name_len, sizeof(char)) ) == NULL )
		{
		    fprintf(stderr, "canot calloc HDF4_SDSs_info->SDS_names[%d], exit...\n", index);
		    exit(0);
		}
		else
		    strcpy(HDF4_SDSs_info->SDS_names[index], name);

		/* put each dim size into each SDS's dim_sizes */
		if( ( HDF4_SDSs_info->dim_sizes[index] = (int32*)calloc((size_t)HDF4_SDSs_info->rank[index], sizeof(int32)) ) == NULL )
		{
		    fprintf(stderr, "canot calloc HDF4_SDSs_info->dim_sizes[%d], exit...\n", index);
		    exit(0);
		}
		else
		   for(i = 0; i < HDF4_SDSs_info->rank[index]; i++)
			HDF4_SDSs_info->dim_sizes[index][i] = dim_sizes[i];
	    }
	}

	/* free intermediate commonly used name string */
	if(name) free(name);

	return 0;
}

static int HDF4_free_SDSs_info(HDF4_SDSs_INFO* HDF4_SDSs_info)
{
   /* written: Maosheng Zhao, SSAI
    * purpose: free memory of each pointer elements
    * Date: July 3rd & 5th, 2017
    */ 	
    int32 index;

    if(HDF4_SDSs_info)
    {
	/* free vectors */
	if(HDF4_SDSs_info->data_type) free(HDF4_SDSs_info->data_type);
	if(HDF4_SDSs_info->rank) free(HDF4_SDSs_info->rank);
	if(HDF4_SDSs_info->n_attrs) free(HDF4_SDSs_info->n_attrs);

	/* free matrices */
	for(index=0; index<HDF4_SDSs_info->n_SDSs; index++)
	{
	    if(HDF4_SDSs_info->SDS_names[index]) free(HDF4_SDSs_info->SDS_names[index]);
	    if(HDF4_SDSs_info->dim_sizes[index]) free(HDF4_SDSs_info->dim_sizes[index]);
	}
	if(HDF4_SDSs_info->SDS_names) free(HDF4_SDSs_info->SDS_names);
	if(HDF4_SDSs_info->dim_sizes) free(HDF4_SDSs_info->dim_sizes);
	
    }

    return 0;
}

/*##########################################################################################*/
/* Begin stand alone H5Fget_objs_links */
herr_t H5Fget_objs_links(hid_t fid, H5_OBJS_TYPES *h5_obj_types)
{
/* written: Maosheng Zhao
 * Date:   Aug. 2, 2017
 * Science Systems and Applications, Inc. (SSAI) 
 * maosheng.zhao@ssaihq.com
 * Purpose: write a stand-alone function H5Fget_objs_links() for getting all objects in an HDF5.
 *          Note that H5Ovisit() only scans objects not links, whereas H5Lvisit() scans
 *          both objects and links except root group because all objects and links are
 *          linked to root group. So I use H5Lvisit() but count in root group.
 *          struct H5_OBJS_TYPES contains number and names of all objects and links, including root
 *          group, and their H5O_type_t (group, dataset, named_datatype, etc.).
 * Required C functions:
 * 	 static herr_t H5L_get_num_objs_links(hid_t fid, hsize_t *num_objs_links);
 * 	 static herr_t cb_func_num_objs_links_L(hid_t loc_id, const char *name, const H5L_info_t *info, void *operator_data);
 * 	 static herr_t H5L_get_names_types_objs_links(hid_t fid, H5_OBJS_TYPES *h5_obj_types);
 * 	 static herr_t cb_func_names_objs_links_L(hid_t loc_id, const char *name, const H5L_info_t *info, void *operator_data);
 *
 */
    herr_t          status;
    hsize_t num_objs_links;

    /* first get number of objs including links */
    status = H5L_get_num_objs_links(fid, &num_objs_links);

    /* second set h5_obj_types based on number of objs and links */
    h5_obj_types->obj_name = (char**)calloc(num_objs_links, sizeof(char*));
    h5_obj_types->obj_type = (H5O_type_t*)calloc(num_objs_links, sizeof(H5O_type_t));

    /* get h5_obj_types */
    status = H5L_get_names_types_objs_links(fid, h5_obj_types);

    return 0;
}

static herr_t H5L_get_num_objs_links(hid_t fid, hsize_t *num_objs_links)
{
/* written: Maosheng Zhao
 * Date:   Aug. 2, 2017
 * Science Systems and Applications, Inc. (SSAI) 
 * maosheng.zhao@ssaihq.com
 * Purpose: get number of all objects and links in an HDF5
 *          Note that H5Ovisit() only scans objects not links, whereas H5Lvisit() scans
 *          both objects and links except root group because all objects and links are
 *          linked to root group. So I use H5Lvisit() but account for root group.
 */
    herr_t  status;

    *num_objs_links = 1;   /* start from the root group and root group is not accounted for in H5Lvisit() whereas I count it in */

    status = H5Lvisit (fid, H5_INDEX_NAME, H5_ITER_NATIVE, cb_func_num_objs_links_L, num_objs_links);

    return 0;
}

static herr_t H5L_get_names_types_objs_links(hid_t fid, H5_OBJS_TYPES *h5_obj_types)
{
/* written: Maosheng Zhao
 * Date:   Aug. 2, 2017
 * Science Systems and Applications, Inc. (SSAI) 
 * maosheng.zhao@ssaihq.com
 * Purpose: get names of all objects and links in an HDF5. It requries first get number
 *          of all objects and links by calling H5L_get_num_objs_links().
 *          Note that H5Ovisit() only scans objects not links, whereas H5Lvisit() scans
 *          both objects and links except root group because all objects and links are
 *          linked to root group. So I use H5Lvisit() but count in the root group.
 *          struct H5_OBJS_TYPES contains number and names of all objects and links,
 *          and their H5O_type_t (group, dataset, named_datatype, etc.).
 */
    herr_t  status;

    /* start from the root group and root group is not accounted for in H5Lvisit(), but I put it in h5_obj */
    h5_obj_types->obj_name[0] = (char*)calloc(2, sizeof(char));
    strcpy(h5_obj_types->obj_name[0], "/");
    h5_obj_types->obj_type[0] = H5O_TYPE_GROUP;
    h5_obj_types->num_objs_links = 1;

    status = H5Lvisit (fid, H5_INDEX_NAME, H5_ITER_NATIVE, cb_func_names_objs_links_L, h5_obj_types);

    return 0;
}

static herr_t cb_func_num_objs_links_L(hid_t loc_id, const char *name, const H5L_info_t *info, void *operator_data)
{
/* written: Maosheng Zhao
 * Date:   Aug. 2, 2017
 * Science Systems and Applications, Inc. (SSAI) 
 * maosheng.zhao@ssaihq.com
 * Purpose: callback function for H5Lvisit() in H5L_get_num_objs_links()
 */
    herr_t          status;
    hsize_t *num_objs_links = (hsize_t*)operator_data;

    (*num_objs_links)++;

    return 0;
}

static herr_t cb_func_names_objs_links_L(hid_t loc_id, const char *name, const H5L_info_t *info, void *operator_data)
{
/* written: Maosheng Zhao
 * Date:   Aug. 2, 2017
 * Science Systems and Applications, Inc. (SSAI) 
 * maosheng.zhao@ssaihq.com
 * Purpose: callback function for H5Lvisit() in H5L_get_names_types_objs_links()
 */
    herr_t          status;
    H5O_info_t      infobuf;
    H5_OBJS_TYPES *h5_obj_types = (H5_OBJS_TYPES*)operator_data;

    /* scan increment */
    h5_obj_types->num_objs_links++;

    /* +2 for the root group and terminator of the string */
    h5_obj_types->obj_name[h5_obj_types->num_objs_links - 1] = (char*)calloc(strlen(name) + 2, sizeof(char));

    /* assign obj_name */
    strcpy(h5_obj_types->obj_name[h5_obj_types->num_objs_links - 1],"/");
    strcat(h5_obj_types->obj_name[h5_obj_types->num_objs_links - 1], name);

    /* get H5O_type_t for the object */
    status = H5Oget_info_by_name (loc_id, name, &infobuf, H5P_DEFAULT);
    h5_obj_types->obj_type[h5_obj_types->num_objs_links - 1] = infobuf.type;
     
    return 0;
}
/* End stand alone H5Fget_objs_links */
/*##########################################################################################*/

int HDF_version_check(const char *HDF_name)
{
	/* written: Maosheng Zhao
	 * date: Jun 5, 2017
	 * date: Apr 2, 2018 use Hishdf() for HDF4 to replace the previous file signature method
	 * SSAI
	 * Requires mfhdf.h and hdf5.h and HDF4/HDF5 library
	 */ 
	/* check the HDF version of HDF file based on the HDF file signature
 	 from the leading bytes of the HDF file */
	/* check the file signature */
	if( H5Fis_hdf5(HDF_name) )
		return(5);                         /* HDF5 file format */
	else
	{
		if(Hishdf(HDF_name)) /* just 4 leading bytes */
			return(4);                /* HDF4 file format */
		else
			return(0);                /* NOT HDF file format */
	}
}

int HDF5_split_abs_name_into_path_name(const char *obj_abs_name, char **abs_path, char **name)
{
/* written: Maosheng Zhao
 * Date: 8/11/17. 
 * SSAI
 * purpose: to split the absolute name of an HDF5 object (obj_abs_name) into
 *          absolute path (abs_path) and its name (name)
 *          For example: if char *obj_abs_name = "/HDFEOS/GRIDS/VNP_Grid_1km_2D/Data Fields/SurfReflect_M5_1";
 *          	The resulting *abs_path = "/HDFEOS/GRIDS/VNP_Grid_1km_2D/Data Fields/" and
 *          	*name = "SurfReflect_M5_1"
 * Note: obj_abs_name is the input. abs_path and name are outputs and are calloced within this function
 * 	The abs_path always has "/" in the end and name has no "/" at all and it is the last name. This
 * 	will make it easy for join the abs_name with name if necessary.
 *  2/19/2020 change the method of strtok_r() to strrchr() to make it easier and no warning information when making it.	
 */
   
    char *pch = NULL;
    size_t n_char, str_len; 
	
    if(obj_abs_name[0] == '/')   /* must an HDF5 absolute path name from the root */
    {
	pch=strrchr(obj_abs_name, '/');

	n_char = (size_t)(pch - obj_abs_name + 1);

	str_len = n_char;
	/* add an extra char for terminator */
	*abs_path = (char*)calloc(str_len + 1, sizeof(char));

	str_len = strlen(obj_abs_name) - n_char;
	/* add an extra char for terminator */
	*name = (char*)calloc(str_len + 1, sizeof(char));

	strncpy(*abs_path, obj_abs_name, n_char);

	/* the remainder is the name */
	strncpy(*name, (char *)(obj_abs_name + n_char), strlen(obj_abs_name) - n_char);
    }
    else
    {
	*abs_path = NULL;
	str_len = strlen(obj_abs_name);
	if(str_len > 0)
	{
	  *name = (char*)calloc(str_len + 1, sizeof(char));
	  strcpy(*name, obj_abs_name);
	}
	else
	  *name = NULL;
    }

    return 0;
}

HDF5_DATA_TYPE_INFO HDF5_get_data_type_info(hid_t tid)
{
/* Maosheng Zhao SSAI added the below two function for printout
 * attrs of HDF5 objects */

/* requires funtcions:
 * byte_order()
 */ 
    hid_t       p_type=-1;
    H5T_class_t type_class;
    size_t      size;
    H5T_sign_t  sign;
    HDF5_DATA_TYPE_INFO data_type_info;

    type_class = H5Tget_class(tid);
    size       = H5Tget_size(tid);

    if(type_class == H5T_STRING)
	size++; /* Make room for null terminator */

    sign       = H5Tget_sign(tid);

    data_type_info.HDF5_type = 99;

    if(!byte_order())  /* little-endian */
        data_type_info.endian = 0;
    else
        data_type_info.endian = 1;

    /* use the existing HDF4 predefined DFNT */
    switch(type_class) {
	case H5T_INTEGER:
	    if ( size == 1 && sign == H5T_SGN_2) {
		p_type=H5Tcopy(H5T_NATIVE_SCHAR);
		data_type_info.HDF5_type = DFNT_INT8;
	    }	
	    else if ( size == 2 && sign == H5T_SGN_2) {
		p_type=H5Tcopy(H5T_NATIVE_SHORT);
		data_type_info.HDF5_type = DFNT_INT16;
	    }	
	    else if ( size == 4 && sign == H5T_SGN_2) {
		p_type=H5Tcopy(H5T_NATIVE_INT);
		data_type_info.HDF5_type = DFNT_INT32;
	    }	
	    else if ( size == 8 && sign == H5T_SGN_2) {
		p_type=H5Tcopy(H5T_NATIVE_LONG);
		data_type_info.HDF5_type = DFNT_INT64;
	    }
	    else if ( size == 1 && sign == H5T_SGN_NONE) {
		p_type=H5Tcopy(H5T_NATIVE_UCHAR);
		data_type_info.HDF5_type = DFNT_UINT8;
	    }	
	    else if ( size == 2 && sign == H5T_SGN_NONE) {
		p_type=H5Tcopy(H5T_NATIVE_USHORT);
		data_type_info.HDF5_type = DFNT_UINT16;
	    }	
	    else if ( size == 4 && sign == H5T_SGN_NONE) {
		p_type=H5Tcopy(H5T_NATIVE_UINT);
		data_type_info.HDF5_type = DFNT_UINT32;
	    }	
	    else if ( size == 8 && sign == H5T_SGN_NONE) {
		p_type=H5Tcopy(H5T_NATIVE_ULONG);
		data_type_info.HDF5_type = DFNT_UINT64;
	    }	
	    else {
		fprintf(stderr, "unknown data type in H5T_INTEGER, exit...\n");
		exit(0);
	    }
	    break;

	case H5T_FLOAT:
	    if ( size == 4) {
		p_type=H5Tcopy(H5T_NATIVE_FLOAT);
		data_type_info.HDF5_type = DFNT_FLOAT32;
	    }	
	    else if ( size == 8) {
		p_type=H5Tcopy(H5T_NATIVE_DOUBLE);
		data_type_info.HDF5_type = DFNT_FLOAT64;
	    }
	    else {
	        fprintf(stderr, "unknown data type in H5T_FLOAT, exit...\n");
		exit(0);
	    }
	    break;

	case H5T_TIME:
	case H5T_BITFIELD:
	case H5T_OPAQUE:
	case H5T_STRING:
   	    break;
	case H5T_COMPOUND:
	case H5T_REFERENCE:
	case H5T_ENUM:
	case H5T_VLEN:
	case H5T_ARRAY:
	    break;

	case H5T_NO_CLASS:
	case H5T_NCLASSES:
	default:
	    fprintf(stderr, "undefined datatype class in HDF5_get_data_type_info(), exit...\n");
	    exit(0);

    } /* end switch */

    /*return p_type;*/
    data_type_info.num_type = p_type;
    data_type_info.type_class = type_class;
    data_type_info.size = size;
    data_type_info.sign = sign;

    return data_type_info;
}

static unsigned char byte_order(void)  /* support HDF5_get_data_type_info() */
{
    /* written: Maosheng Zhao */
    /* Date: Jan. 09, 2017 */
    /* check how the bytes of the datatype are layed out in memory for the machine you are using */
    unsigned short Endian = 1;
    char *little = NULL;
    unsigned char bigEndian;

    /* check the byte oder of the native machine */
    little = (char*)&Endian;
    if( *little==0)
    {
	bigEndian=1;  /* it is bigEndian machine, which stores the most significant byte in the smallest address */
	/*fprintf(stderr, "This is a big Endian machine\n");*/
    }
    else
    {
	bigEndian=0;  /* it is little Endian machine, which stores the least significant byte in the smallest address */
	/*fprintf(stderr, "This is a little Endian machine\n");*/
    }	

    return(bigEndian);
}

int L2G_format_check(char *fname)
{ /* Developer: Maosheng Zhao, SSAI,
   * date: 2/21/2018
   * purpose: tell whether the format of an HDF4/HDF5
   * is L2G or L2G-lite file.
   * return -1, failed; return 1, L2G; return 2, L2G-lite
   * Note that some L2G also has iobs_res. So there is no strict
   * rules to tell the HDF is L2G or L2G-lite based on num_observations
   * and iobs_res.  But it is sure that num_observations means it is L2G */
   
  int sds_cnt, i, status = -1;
  char **sds_names = NULL;

  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for sds_names in read_l2g_lite(main)\n");
    status = -1;
  }
  else
  {
    sds_cnt = get_sds_names(fname, sds_names); 
    if(sds_cnt == 0) status = -1;
    else
    {
      /* is it L2G ? */
      for(i = 0; i < sds_cnt; i++)
	if(strcasestr(sds_names[i], "num_observations") != '\0')
	{
	  status = 1;        /* L2G */
	  break;
	}

      /* L2G or L2G lite */
      if(status == 1)
        for(i = 0; i < sds_cnt; i++)
	  if(strcasestr(sds_names[i], "iobs_res") != '\0')
	  {
	    status = 2;      /* L2G or L2G lite */
	    break;
	  }
    }
    if(sds_names) Free2D((void **)sds_names);
  }

  return status;
}

int get_lines_from_chunk(void *data_chunk, int data_rank, int chunk_shape[], int bandinterleave, int element_size, int lines_start, int n_lines, void *data_lines)
{
	/* Developer: Maosheng Zhao 1/11/2020
 	 * Purpose:  get full lines or rows of data across all different dimensions. 
 	 *           To enhance the performance of processing data, for SDSs,  reading data
 	 *           in chunk not line by line but getting the line or lines data after reading
 	 *           the chunk is more effcient.
 	 * NOTE currently, we just assume that there are just two bandinterleave types, BSQ or BIP not BIL.
 	 */

	/* Note that because this is for data_chunk and thus bandinterleave must be provided with
 	 * 1 for bsq and 0 for bip.
 	 */

	/* INPUTs: data_chunk, data_rank, chunk_shape[], bandinterleave, element_size, lines_start, and n_lines
 	 * OUTPUTs: data_lines or data_rows.  Also data_lines must be calloced outof the function.
 	 * Note that lines_start is 0-based row and is referenced for the data_chunk.
 	 */ 

	int i;
	int shape_3D[3] = {1, 1, 1};   /* initialize as 1, for remote sensing data, convert any >=2D data as 3D */

	size_t one_line_bytes, chunk_shift, lines_shift, one_layer_bytes;

	/* check memory variables */	
	if(data_chunk == NULL || data_lines == NULL)
	{
	   fprintf(stderr, "data_chunk and data_lines must be calloced outside of this function, exit...\n");
	   exit(0);
	}

	/* Any >=2D data can be reshaped into 3D to faciliatate data process below */
	if(data_rank < 2)
	{
	   fprintf(stderr, "This is just for >=2D data not others, exit...\n");
	   exit(0);
	}
	else if(data_rank == 2)   /* assume it is BSQ with the first dimension size as 1 */
	{
	   bandinterleave = 1;   /* now it is BSQ */

	   shape_3D[0] = 1;
	   shape_3D[1] = chunk_shape[0];                 /* rows */
	   shape_3D[2] = chunk_shape[1];                 /* cols */
	}
	else
	{
	   if(bandinterleave)  /* 1 is for BSQ */
	   {
	      shape_3D[2] = chunk_shape[data_rank - 1];  /* cols */
	      shape_3D[1] = chunk_shape[data_rank - 2];  /* rows */

	      shape_3D[0] = 1;
	      for(i = 0; i < data_rank - 2; i++)
		 shape_3D[0] *= chunk_shape[i];
	   }
	   else                /* 0 is for BIP */
	   {
	      shape_3D[0] = chunk_shape[0];             /* rows */
	      shape_3D[1] = chunk_shape[1];             /* cols */

	      shape_3D[2] = 1;
	      for(i = 2; i < data_rank; i++)
		 shape_3D[2] *= chunk_shape[i];
	   }
	}

	/* check chunk shape */
	if(shape_3D[0] < 1 || shape_3D[1] < 1 || shape_3D[2] < 1)
	{
	   fprintf(stderr, "there are problems for chunk_shape, because the combined one dimension is less than 1, exit...\n");
	   exit(0);
	}

	/* check element_size */
	if(element_size < 1)
	{
	   fprintf(stderr, "element_size is less than 1, exit...\n");
	   exit(0);
	}

	/* now get lines of data from the data_chunk for the specified lines_start and n_lines */
	if(bandinterleave)    /* BSQ */
	{
	    /* check validity of lines_start and n_lines */
	    if(lines_start + n_lines > shape_3D[1])
	    {
		fprintf(stderr, "The specified lines_start and n_lines are out of range of the BSQ chunk data, exit...\n");
		exit(0);
	    }
	    else
	    {
		/* one layer (full X and Y dim) of bytes */
		one_layer_bytes = (size_t)element_size * (size_t)shape_3D[1] * (size_t)shape_3D[2];

		/* number bytes for one line data for one nonxy combined dim */
		one_line_bytes = (size_t)element_size * (size_t)shape_3D[2];

		/* loop for Z dim (all combined non-XY dims) */
		for(i = 0; i < shape_3D[0]; i++)
		{
		   chunk_shift = (size_t)i * one_layer_bytes + (size_t)lines_start * one_line_bytes; /* source shift */
		   lines_shift = (size_t)i * one_line_bytes * (size_t)n_lines;                       /* dest  shift */

		   /* copy the data from source to dest */
		   memmove((void*)(data_lines + lines_shift), (void*)(data_chunk + chunk_shift), one_line_bytes * (size_t)n_lines);
		}
	    }
		
	}
	else                 /* BIP */
	{
	    /* check validity of lines_start and n_lines */
	    if(lines_start + n_lines > shape_3D[0])
	    {
		fprintf(stderr, "The specified lines_start and n_lines are out of range of the BIP chunk data, exit...\n");
		exit(0);
	    }
	    else
	    {
		/* it is sort of like a 2D data (if combined the X and Z dims together) and it is so easy */
		/* number bytes for one full line data for one nonxy combined dim */
		one_line_bytes = (size_t)element_size * (size_t)shape_3D[1] * (size_t)shape_3D[2];

		chunk_shift = (size_t)lines_start * one_line_bytes;

		/* copy the data from source to dest */
		memmove(data_lines, (void*)(data_chunk + chunk_shift), one_line_bytes * (size_t)n_lines);
	    }
	}

	return 0;
}

